/************************************************************************/
/*									*/
/* Ensemble de macros "parsearg.h" pour reconnaitre des arguments 	*/
/* a l'appel d'applications et executer des actions elementaires:	*/
/* affectation de variables, positionnement de drapeaux.		*/
/*									*/
/* l'utilisation des macros de parsearg.h presuppose que les variables 	*/
/* suivantes ont ete definies dans le programme appellant:		*/
/*									*/
/*	int	i;		variable de compte des arguments 	*/
/*	boolean	Args_Ok;	indicateur de correctude des arguments	*/
/*	boolean Help;		indicateur de l'aide. Sert a temperer 	*/
/*				le precedent				*/
/*									*/
/* Les macros suivantes sont egalement necessaires.			*/
/*									*/
/*	TOOLNAME		le nom de l'application			*/
/*	VERSION			son numero de version			*/
/*									*/
/************************************************************************/
/*									*/
/* La plupart des macros admettent un premier argument "Opt", qui est	*/
/* la chaine de caractere correspondant a l'option.			*/
/*									*/
/* Les macros disponibles sont:						*/
/*									*/
/*	SWITCH( Opt, Var, Val )	positionne la variable Var a la valeur	*/
/*				Val.					*/
/*	GET_INT( Opt, Var )	lit un entier et le met dans Var	*/
/*	GET_REAL( Opt, Var )	lit un double et le met dans Var	*/
/*	GET_STRING( Opt, Var )	lit une chaine de caracteres et la 	*/
/*				copie dans Var (qui doit etre allouee)	*/
/*	GET_INT_SWITCH( Opt, Var, Bool, Val )				*/
/*				lit un entier, et en cas de reussite,  	*/
/*				le met dans Val, et positionne Bool a 	*/
/*				Val					*/
/*	GET_REAL_SWITCH( Opt, Var, Bool, Val )				*/
/*				lit un reel, et en cas de reussite, le 	*/
/*				met dans Val, et positionne Bool a Val	*/
/*	GET_STRING_SWITCH( Opt, Var, Bool, Val )			*/
/*				lit une chaine de caracteres, et en cas */
/*				de reussite, la met dans Val, et        */
/*				positionne Bool a Val			*/
/*									*/
/************************************************************************/

#define GET_INT( Opt, Var ) ( 0 == strcmp( Opt, argv[i] ) ) {  \
  i++; \
  if ( ( i < argc ) && ( 1 == sscanf( argv[i], "%d", &Var ) ) ) \
								  i++; \
  else { printf("%s: integer expected after %s\n", TOOLNAME, Opt ); \
  Args_Ok = FALSE; } \
}

#define GET_REAL( Opt, Var ) ( 0 == strcmp( Opt, argv[i] ) ) {  \
  i++; \
  if ( ( i < argc ) && ( 1 == sscanf( argv[i], "%lf", &Var ) ) ) \
								  i++; \
  else { printf("%s: real expected after %s\n", TOOLNAME, Opt ); \
  Args_Ok = FALSE; } \
}

#define GET_REAL_SWITCH( Opt, Var, Bool, Val ) ( 0 == strcmp( Opt, argv[i] ) ) {  \
  i++; \
  if ( ( i < argc ) && ( 1 == sscanf( argv[i], "%lf", &Var ) ) ) { \
								    i++; Bool = Val; } \
  else { printf("%s: real expected after %s\n", TOOLNAME, Opt ); \
  Args_Ok = FALSE; } \
}
#define GET_INT_SWITCH( Opt, Var, Bool, Val ) ( 0 == strcmp( Opt, argv[i] ) ) {  \
  i++; \
  if ( ( i < argc ) && ( 1 == sscanf( argv[i], "%d", &Var ) ) ) { \
								    i++; Bool = Val; } \
  else { printf("%s: integer expected after %s\n", TOOLNAME, Opt ); \
  Args_Ok = FALSE; } \
}

#define GET_STRING_SWITCH( Opt, Var, Bool, Val ) ( 0 == strcmp( Opt, argv[i] ) ) {  \
  i++; \
  if ( i < argc ) { strcpy( Var, argv[i] ); \
  i++; Bool = Val; } \
  else { printf("%s: argument missing after %s\n", TOOLNAME, Opt ); \
  Args_Ok = FALSE; } \
}

#define GET_STRING( Opt, Var ) ( 0 == strcmp( Opt, argv[i] ) ) {  \
  i++; \
  if ( i < argc ) { strcpy( Var, argv[i] ); \
  i++; } \
  else {  printf("%s: argument missing after %s\n", TOOLNAME, Opt ); \
  Args_Ok = FALSE; } \
}

#define SWITCH( Opt, Var, Val ) ( 0 == strcmp( Opt, argv[i] ) ) { \
  Var = Val;	i++; }

#define IS_VERSION ( 0 == strcmp( "-v", argv[i] ) ) {\
  printf("%s (Version %s)\n", TOOLNAME, VERSION); \
  Args_Ok = FALSE; Help = TRUE; }

