#include <math.h>

#include "eg.h"
#include "pn.h"

#define CODE(x)		fprintf(Out_File,x)
#define CODE1(x,y)	fprintf(Out_File,x,y)
#define CODE2(x,y,z)	fprintf(Out_File,x,y,z)
#define CODE3(x,y,z,t)	fprintf(Out_File,x,y,z,t)
#define CODELN		CODE("\n")

#define SCAN(x,y)	for(x=y;x!=NULL;x=x->Next)
#define IRINT(x)	(int)rint(x)

/* external declarations */
extern char *todays_date(void);

/* forward declaration */
void Write_Address( Address The_Add );

/* local background variables */
FILE	*Out_File = stdout;
char	*TOOLNAME;	/* necessary for error/warning messages		*/


/* Procedure to write the initial population. Works for EG only */
void Write_Init_Mark_EG( Place_Desc The_Place )
{	
  int j;

  CODE1("%1d ", The_Place->Init_Pop);
  for( j=0; j < The_Place->Init_Pop; j++ )
    CODE1("%8.4f ", The_Place->Lag_Times[j]);

}

/* Procedure to write the initial population. Works for PN only */
void Write_Init_Mark_PN( Pl_Des The_Place )
{	
  int j;

  CODE1("%1d ", The_Place.Init_Pop);
  for( j=0; j < The_Place.Init_Pop; j++ )
    CODE1("%8.4f ", The_Place.Lag_Times[j]);

}

/* Procedure to write the routing population. */
void Write_Routing(Pl_Des	The_Place)
{	
  int 		j;
  Routing_Desc	The_Routing;

  The_Routing = The_Place.Routing;

  if ( The_Routing != NULL ) {

    switch( The_Routing->Type ) {
    case 'R':
      CODE("Random [");
      for (j=0; j<The_Place.Outdegree; j++)
	CODE1(" %8.6f", The_Routing->Probas[j] );
      CODE("] ");
      break;
    
    case 'D':
      CODE("Cyclic [ ");
      for (j=0; j<The_Routing->Period; j++) {
	/* 	CODE1(" %d", The_Routing->Cycle[j]->Number + 1 ); */
	Write_Address( The_Routing->Cycle[j] );
	CODE(" ");
      }
      CODE("] ");
      break;

    case 'C':
      CODE("Competition");
      break;

    case 'P':
      CODE("Competition Priority [ ");
      for (j=0; j<The_Routing->Period; j++) {
	/* 	CODE1(" %d", The_Routing->Cycle[j] + 1 ); */
	Write_Address( The_Routing->Cycle[j] );
	CODE(" ");
      }
      CODE("] ");
    }

  }
}

void Write_Opt_String(char	*The_Name,
		      int	Option )
{
  if ( The_Name != NULL ) {
    if ( strlen( The_Name ) > 0 ) {
      if ( Option == PAR )
	CODE("(");
      else if ( Option == BRACKET )
	CODE("[");
      else if ( Option == QUOTE )
	CODE("\"");
      else if ( Option == WHITE_QUOTE )
	CODE(" \"");
      CODE1("%s", The_Name );
      if ( Option == PAR )
	CODE(")");
      else if ( Option == BRACKET )
	CODE("]");
      else if ( Option == QUOTE )
	CODE("\"");
      else if ( Option == WHITE_QUOTE )
	CODE("\" ");
    }
  }
}

#define Write_String( x )	Write_Opt_String( x, WHITE_QUOTE )

void Write_Address(Address	The_Add )
{
  Address	A_Scan;

  if ( The_Add == (Address)NULL )
    CODE("<no-address>");
  else {
    SCAN( A_Scan, The_Add ) {
      switch( A_Scan->Type ) {
      case 'I':
	CODE("IN");
	break;
      case 'O':
	CODE("OUT");
	break;
      default:
	CODE1("%c", A_Scan->Type );
      }
      switch( A_Scan->Type ) {
      case 'P':
      case 'T':
	CODE1("%d", A_Scan->Number + 1 );
	break;
      default:
	CODE1("%d", A_Scan->Number );
      }
      if ( A_Scan->Next != NULL )
	CODE("/");
    }
  }
}
	
boolean Write_EG_As_PNL(
EventGraph      *Reseau,
char		*File_Name,
char		*Tool_Name )
{
  int		i;
  Trans_Desc	The_Trans;
  Liste_Place	P_Scan;
  Place_Desc	The_Place;
  char		*The_Date;

  if ( strlen( File_Name ) != 0 ) {
    Out_File = fopen( File_Name, "w" );
    if ( Out_File == NULL ) 
      return FALSE;
  }
  else
    Out_File = stdout;

  TOOLNAME = Tool_Name;

  CODE1("# pnl code generated by %s\n", TOOLNAME );
  The_Date = todays_date();
  CODE1("# Date: %s\n", The_Date );
  free( The_Date );
  CODE1("Transitions <- %d\n", Reseau->Size );
  CODE1("Places <- %d\n", Reseau->Nb_Places );

  /* generate transitions */
  for ( i=0; i<Reseau->Size; i++ ) {

    CODE1("T %d ", i+1 );
    The_Trans = Reseau->T[i];
    Write_String( The_Trans->Name );
    Write_String( The_Trans->Data );

    Write_Law( Out_File, The_Trans->Trans_Law, NORMAL_PRINT_MODE );

    SCAN( P_Scan, Reseau->T[i]->Outputs ) {
      CODE(", ");
      CODE1("%d ", 1+P_Scan->Val->Place_Id );
    }
    CODE(";\n");
  }

  /* now generate places */
  for ( i=0; i<Reseau->Nb_Places; i++ ) {
    CODE1("P %d ", i+1 );

    The_Place = Reseau->P[i];

    Write_String( The_Place->Name );
    Write_Law( Out_File, The_Place->Place_Law, NORMAL_PRINT_MODE );
    Write_Init_Mark_EG( The_Place );
    CODE(", ");
    CODE1("%d ", The_Place->Trans_Id );
    CODE(";\n");
  }

  if ( Out_File != stdout )
    fclose( Out_File );

  return TRUE;
}

boolean Write_PN_As_PNL(
PetriNet        *Reseau,
char		*File_Name,
char		*Tool_Name )
{
  int		i;
  Tr_Des	The_Trans;
  Pl_Des	The_Place;
  Liste_Link	L_Scan;
  char		*The_Date;

  if ( strlen( File_Name ) != 0 ) {
    Out_File = fopen( File_Name, "w" );
    if ( Out_File == NULL ) 
      return FALSE;
  }
  else
    Out_File = stdout;

  TOOLNAME = Tool_Name;

  CODE1("# pnl code generated by %s\n", TOOLNAME );
  The_Date = todays_date();
  CODE1("# Date: %s\n", The_Date );
  free( The_Date );
  if ( strlen( Reseau->File_Name ) > 0 )
    CODE1("# Input file: %s\n", Reseau->File_Name );
  CODE1("Transitions <- %d\n", Reseau->Nb_Trans );
  CODE1("Places <- %d\n", Reseau->Nb_Places );

  /* generate submodel section */
  for ( i=0; i<Reseau->Nb_Submodels; i++ ) {
    CODE1("Submodel %d", Reseau->S[i].Subm_Id );
    if ( Reseau->S[i].Name != NULL )
      CODE(" ");
    Write_String( Reseau->S[i].Name );
    CODE(" FILE ");
    Write_String( Reseau->S[i].File_Name );
    CODE(";\n");
  }

  /* generate transitions */
  for ( i=0; i<Reseau->Nb_Trans; i++ ) {

    CODE1("T %d ", i + 1 );
    The_Trans = Reseau->T[i];
    Write_String( The_Trans.Name );
    Write_String( The_Trans.Data );

    Write_Law( Out_File, The_Trans.Tr_Law, NORMAL_PRINT_MODE );

    SCAN( L_Scan, The_Trans.Outputs ) {
      CODE(",\n\t");
      CODE1("%d ", L_Scan->Dest + 1 );
      CODE1("%d ", L_Scan->Weight );
      Write_String( L_Scan->Data );
    }
    CODE(";\n");
  }

  /* generate places */
  for ( i=0; i<Reseau->Nb_Places; i++ ) {

    CODE1("P %d ", i + 1 );

    The_Place = Reseau->P[i];

    Write_String( The_Place.Name );

    Write_Law( Out_File, The_Place.Pl_Law, NORMAL_PRINT_MODE );

    Write_Init_Mark_PN( The_Place );

    Write_Routing( The_Place );

    SCAN( L_Scan, The_Place.Outputs ) {
      CODE(",\n\t");
      CODE1("%d ", L_Scan->Dest + 1 );
      CODE1("%d ", L_Scan->Weight );
      Write_String( L_Scan->Data );
      if ( L_Scan->Is_Inhibitor )
	CODE(" I");
    }
    CODE(";\n");    
  }

  /* generate I/O section */
  if ( ( Reseau->Nb_Inputs > 0 ) || ( Reseau->Nb_Outputs > 0 ) )
    CODE("Export\n");
  for ( i=0; i<Reseau->Nb_Inputs; i++ ) {
    CODE1("\tIN  %d IS ", i+1 );
    Write_Address( Reseau->Inputs[i] );
    CODE(";\n");
  }
  for ( i=0; i<Reseau->Nb_Outputs; i++ ) {
    CODE1("\tOUT %d IS ", i+1 );
    Write_Address( Reseau->Outputs[i] );
    CODE(";\n");
  }

  if ( Out_File != stdout )
    fclose( Out_File );

  return TRUE;
}


boolean	Write_EG_As_EGL( 
EventGraph	*Reseau,
char		*File_Name,
char		*Tool_Name )
{

  int		i;
  Liste_Place	P_Scan;
  Trans_Desc	The_Trans;
  Place_Desc	The_Place;
  char		*The_Date;

  if ( strlen( File_Name ) != 0 ) {
    Out_File = fopen( File_Name, "w" );
    if ( Out_File == NULL ) 
      return FALSE;
  }
  else
    Out_File = stdout;

  TOOLNAME = Tool_Name;

  CODE1("# egl code generated by %s\n", TOOLNAME );
  The_Date = todays_date();
  CODE1("# Date: %s\n", The_Date );
  free( The_Date );
  CODE1("Size <- %d\n", Reseau->Size );

  for ( i=0; i<Reseau->Size; i++ ) {
	
    CODE1("%d ", i+1 );

    The_Trans = Reseau->T[i];
    if (NULL != The_Trans->Name)
      CODE1("\"%s\" ", The_Trans->Name);
    if (NULL != The_Trans->Data)
      CODE1("\"%s\" ", The_Trans->Data);
    Write_Law( Out_File, The_Trans->Trans_Law, NORMAL_PRINT_MODE );

    SCAN( P_Scan, The_Trans->Outputs ) {
      CODE("\n\t, ");
      The_Place = P_Scan->Val;
      CODE1("%d ", The_Place->Trans_Id);
      if (NULL != The_Place->Name)
	CODE1("\"%s\" ", The_Place->Name);
      if (NULL != The_Place->Data1)
	CODE1("\"%s\" ", The_Place->Data1);
      if (NULL != The_Place->Data2)
	CODE1("\"%s\" ", The_Place->Data2);
      Write_Law( Out_File, The_Place->Place_Law, NORMAL_PRINT_MODE );
      Write_Init_Mark_EG( The_Place );
    }
    CODE(";\n");
  }

  if ( Out_File != stdout )
    fclose( Out_File );

  return TRUE;
}
