/* Translation program from eg language to eg language.			*/
/* The translation usually transforms the graph into some equivalent.	*/

#include <unistd.h>
#include "version.h"
#include "eg.h"

#define	TOOLNAME	"eg2eg"
#define LIN_PROG_NAME	"eg2eg_lin.lp"
#define RES_FILE_NAME	"eg2eg_lin.res"

#define SCAN(x,y)	for(x=y;x!=NULL;x=x->Next)

#define CODE(x)		fprintf(Lin_Prog,x)
#define CODE1(x,y)	fprintf(Lin_Prog,x,y)
#define CODE2(x,y,z)	fprintf(Lin_Prog,x,y,z)
#define CODE3(x,y,z,t)	fprintf(Lin_Prog,x,y,z,t)
#define CODELN		CODE("\n")
	
/* constants for the "flattening" algorithms */
#define	AMONT		0
#define	AVAL		1
#define	ECLATEMENT	2

/* external declarations */
extern boolean          topological_analysis(EventGraph *Reseau);
extern EventGraph	*Flatten_EG(EventGraph *Reseau, int Algo);
extern Place_Desc	Create_Dummy_Place(Trans_Desc Orig, int Dest_Id,
					   double Lag);
extern void		Store_Place(EventGraph*,Place_Desc);
extern EventGraph	*Read_Network(char*);
extern void		dumpe_network(EventGraph *Reseau);
extern void		Free_Event_Graph(EventGraph*);
extern char		*todays_date(void);

/* external declarations from "transcript.c". 	*/
extern boolean		Write_EG_As_EGL(EventGraph*,char*,char*);

/* forward declarations */
void	Recycle_EG(EventGraph *Reseau);
void	Optimize_EG(EventGraph *Reseau,int Algorithm);
void	Determinize_EG(EventGraph *Reseau);
boolean	Is_Recycled_Transition(EventGraph *Reseau, int Tr_Num);

/* convenience background variables */
FILE	*Lin_Prog;
boolean	Lin_Prog_Alone = FALSE;


extern boolean	Debug;

/************************************************************************/
/* Programme principal, avec parsing des options.			*/
/************************************************************************/
int main(int argc, char *argv[])
{
  static char	File_Name[128] = "";
  static char	Code_File_Name[128] = "";
  EventGraph 	*Reseau;
  EventGraph 	*Flat_Reseau;
  int		i;
  boolean	Help = FALSE;
  boolean	Args_Ok = TRUE;
  boolean	Flatten = FALSE;
  boolean	Recycle = FALSE;
  boolean	Optimize = FALSE;
  boolean	Determinize = FALSE;
  boolean	Close = FALSE;
  int		Algorithm = AVAL;

  i = 1;

  if (argc>1)
    do {
      if SWITCH( "-upstream", Algorithm, AMONT )
      else if SWITCH( "-downstream", Algorithm, AVAL )
      else if SWITCH( "-split", Algorithm, ECLATEMENT )
      else if SWITCH( "-linprog", Lin_Prog_Alone, TRUE )
      else if SWITCH( "-flatten", Flatten, TRUE )
      else if SWITCH( "-recycle", Recycle, TRUE )
      else if SWITCH( "-close", Close, TRUE )
      else if SWITCH( "-debug", Debug, TRUE )
      else if SWITCH( "-optimize", Optimize, TRUE )
      else if SWITCH( "-determinize", Determinize, TRUE )
      else if GET_STRING( "-f", File_Name )
      else if GET_STRING( "-o", Code_File_Name )

      else if IS_VERSION
	
      else if ( ( 0 == strcmp( "-h", argv[i] ) )
		|| ( 0 == strcmp( "-?", argv[i] ) ) ) {
	printf("Usage: %s [-f <file>] [-downstream] [-upstream]", 
	       TOOLNAME );
	printf(" [-split] [-flatten]\n\t[-recycle] [-debug] ");
	printf("[-optimize] [-linprog] [-close] [-determinize]");
	printf("\n\t[-o <file>]\n");
	Args_Ok = FALSE;
	Help = TRUE;
	i++;
      }
      else	i++; /* skip unknown argument */
    }
    while ( Args_Ok && ( i < argc ) );

  if ( Args_Ok ) {

    Reseau = Read_Network(File_Name);

    if ( Reseau != NULL ) {

      if (Recycle)
	Recycle_EG( Reseau );

      topological_analysis( Reseau );

      if (Debug)
	dumpe_network( Reseau );

      if (Optimize) {
	Optimize_EG( Reseau, Algorithm );
      }

      if ( !Lin_Prog_Alone ) {
	if (Flatten) {
	  Flat_Reseau = Flatten_EG( Reseau, Algorithm );
	  Free_Event_Graph( Reseau );
	  if ( NULL != Flat_Reseau ) {
	    if ( Debug ) {
	      topological_analysis( Flat_Reseau );
	      dumpe_network( Flat_Reseau );
	    }
	  }
	}
	else
	  Flat_Reseau = Reseau;

	if ( Determinize )
	  Determinize_EG( Flat_Reseau );

	if ( Write_EG_As_EGL( Flat_Reseau, Code_File_Name, TOOLNAME ) ) {
	  Free_Event_Graph( Flat_Reseau );
	  return 0;
	}
	else {
	  fprintf( stderr, "%s: Error opening file %s.\n",
		   TOOLNAME, Code_File_Name );
	  return -2;
	}
      }
      else
	return 0;
    }
    else {
      fprintf( stderr, "%s: Error reading network. No translation.\n", 
	       TOOLNAME );	
      return -1;
    }
  }
  else if (!Help)
    return -1;
  else
    return 0;
}

void	Determinize_EG(EventGraph *Reseau)
{
  int		i;
  Liste_Place	P_Scan;
  Place_Desc	The_Place;
  double	The_Mean;

  for ( i=0; i<Reseau->Size; i++ ) {
    The_Mean = Mean( Reseau->T[i]->Trans_Law );
    Reseau->T[i]->Trans_Law->Name = 'C';
    Reseau->T[i]->Trans_Law->Parameters[0] = The_Mean;

    SCAN( P_Scan, Reseau->T[i]->Outputs ) {
      The_Place = P_Scan->Val;
      The_Mean = Mean( The_Place->Place_Law );
      The_Place->Place_Law->Name = 'C';
      The_Place->Place_Law->Parameters[0] = The_Mean;
    }
  }
}

void	Recycle_EG(EventGraph	*Reseau)
{
  int		i;

  for ( i=0; i<Reseau->Size; i++ )
    if ( !Is_Recycled_Transition( Reseau, i ) ) {
      /* Create the place and attach it to (downstream) neighbors.	*/
      /* It is not necessary to add it to upstream neighbors, as long	*/
      /* as precedence analysis is performed afterwards.		*/
      Store_Place( Reseau,
		   Create_Dummy_Place( Reseau->T[i], i+1, 0.0 ) );
    }
}

/************************************************************************/
/* modules de calcul du marquage realisant la taille minimale de la	*/
/* matrice de transition dans la representation canonique.		*/
/* Ceci passe par la generation d'un programme lineaire dont la solution*/
/* est cette taille. Les resultats sont analyses pour fournir le 	*/
/* nouveau marquage.							*/
/************************************************************************/

/****************************************************************/
/* Generation des egalites de marquage: M = M0 + A x		*/
/****************************************************************/
void	gen_pop( EventGraph *Reseau )
{
  int		i;
  int		k;
  Liste_Place	P_Scan;

  CODE("/* Marking equalities */\n");

  k = 0;
  for( i=0; i<Reseau->Size; i++ )
    SCAN( P_Scan, Reseau->T[i]->Outputs ) {
    CODE1("m_%d = ", k);
    CODE1("%d ", P_Scan->Val->Init_Pop );
    CODE1("+ x_%d", i );
    CODE1(" - x_%d", P_Scan->Val->Trans_Id-1 );
    CODE(";\n");
    k++;
  }
}


void	gen_N_variables( 
EventGraph	*Reseau,
int		Algorithm )
{
  int		i;
  int		Pl_Nb;
  Liste_Place	P_Scan;

  CODE("/* Max marking inequalities */\n");

  for( i=0; i<Reseau->Size; i++ ) {

    SCAN( P_Scan, Reseau->T[i]->Outputs ) {
      Pl_Nb = P_Scan->Val->Place_Id;
      switch( Algorithm ) {
      case AVAL:
	CODE1("N_%d >= ", i);
	CODE1("m_%d;\n", Pl_Nb );
	break;

      case AMONT:
	CODE1("N_%d >= ", P_Scan->Val->Trans_Id-1 );
	CODE1("m_%d;\n", Pl_Nb );
	break;
      }
    }
  }
}

void	gen_lin_code( EventGraph *Reseau, int Algorithm )
{
  int	i;
  char  *Date;

  if ( Debug )
    printf("Creating the linear program.\n");

  if ( !Lin_Prog_Alone )
    Lin_Prog = fopen( LIN_PROG_NAME, "w" );
  else
    Lin_Prog = stdout;

  CODE("/* Linear program computing the optimal matrix size");
  CODE(" for the canonical representation */\n");
  CODE("/* of the (max,+) recurrence */\n");
  Date = todays_date();
  CODE2("/* Generated by %s, %s */\n", TOOLNAME, Date );
  free( Date );

  CODE("min: ");
  for ( i=0; i<Reseau->Size-1; i++ )
    CODE1("N_%d + ", i );
  CODE1("N_%d;\n", Reseau->Size-1 );

  gen_pop(Reseau);
  gen_N_variables( Reseau, Algorithm);

  if ( !Lin_Prog_Alone )
    fclose( Lin_Prog );
}

void solve_lin_program( void )
{
  char	command[512];

  if ( Debug )
    printf("Solving the linear program.\n");

  sprintf( command, "lp_solve < %s > %s", 
	   LIN_PROG_NAME, RES_FILE_NAME );
  system( command );

  unlink( LIN_PROG_NAME );
}

void parse_lin_results(int	Nb_Places,
		       int	*Marking )
{
  FILE	*Data;
  int	Val;
  char	Var[128];
  int	Ident;
  int	i;

  if ( Debug )
    printf("Parsing the results linear program.\n");

  Data = fopen( RES_FILE_NAME, "r" );

  fscanf( Data, "%[A-z: ]%d", &(Var[0]), &Val );

  if ( Debug )
    printf("Value found for the LP: %d\n", Val );

  for ( i=0; i<Nb_Places; ) {
    fscanf( Data, "\n%[mxN_]%d%d", &(Var[0]), &Ident, &Val );
    if ( !strcmp( Var, "m_" ) ) {
      if ( Debug )
	printf("New marking for place #%d: %d\n", Ident, Val );
      Marking[ Ident ] = Val;
      i++;
    }
  }

  unlink( RES_FILE_NAME );

}

void Change_Marking( 
EventGraph      *Reseau,
int     	*New_Marking )
{
  int		i;
  int		k;
  Liste_Place	P_Scan;

  k = 0;
  for( i=0; i<Reseau->Size; i++ )
    SCAN( P_Scan, Reseau->T[i]->Outputs ) {
    P_Scan->Val->Init_Pop = New_Marking[ k ];
    k++;
  }
}

/* teste si une transition est recyclee (sans utiliser Next) */
boolean	Is_Recycled_Transition( 
EventGraph      *Reseau,
int		Tr_Num )
{
  boolean	Ok = FALSE;
  Liste_Place	P_Scan;

  SCAN( P_Scan, Reseau->T[Tr_Num]->Outputs )
    Ok = ( Ok || ( ( P_Scan->Val->Trans_Id-1 == Tr_Num )
		   && ( P_Scan->Val->Init_Pop == 1 ) ) );

  return	Ok;
}

/* teste si toutes les transitions sont recyclees */
boolean	Is_Recycled_Network( EventGraph *Reseau )
{
  int		i;
  boolean	Ok = TRUE;

  for ( i=0; Ok && (i<Reseau->Size); i++ )
    Ok = ( Ok && Is_Recycled_Transition( Reseau, i ) );

  return	Ok;
}

void Optimize_EG(EventGraph	*Reseau,
		 int		Algorithm )
{
  int	*New_Marking;

  if ( !Is_Recycled_Network( Reseau ) ) {
    fprintf( stderr, "%s: warning: some transitions are not recycled.\n",
	     TOOLNAME );
  }

  New_Marking = (int*) calloc( Reseau->Nb_Places, sizeof(int) );

  gen_lin_code( Reseau, Algorithm );
  if ( !Lin_Prog_Alone ) {
    solve_lin_program();
    parse_lin_results( Reseau->Nb_Places, New_Marking );
    Change_Marking( Reseau, New_Marking );
  }
}

/* EventGraph	*Close_It( EventGraph	Reseau ) */
/* { */
/*     EventGraph*	RdP; */
/*     int		i; */
/*     Transition	*The_Trans; */
/*     Transition	*New_Trans; */

/*     RdP = New_Event_Graph( Reseau->Size ); */
/*     if ( NULL == RdP ) { */
/* 	fprintf( stderr, "Error: no space for the new network.\n" ); */
/* 	return NULL; */
/*     } */
/*     else for ( i=0; i<Reseau->Size; i++ ) { */
/* 	    The_Trans = &(Reseau->T[i]); */
/* 	    New_Trans = &(RdP->T[i]); */
/* 	    New_Trans->Trans_Id = i+1; */
/* 	    New_Trans->Trans_Law = Copy_Law( The_Trans->Trans_Law ); */
/* 	    New_Trans->Name = Copy_Name( The_Trans->Name ); */
/* 	    New_Trans->Outdegree = 0; */
/* 	    New_Trans->Outputs = NULL; */
/* 	    New_Trans->Indegree = 0; */
/* 	    New_Trans->Inputs = NULL; */
/* 	    New_Trans->Next = NULL; */
/* 	} */
/* 	for ( i=Reseau->Size; i<RdP->Size; i++ ) { */

/* 	    New_Trans = &(RdP->T[i]); */
/* 	    New_Trans->Trans_Id = i+1; */
/* 	    New_Trans->Trans_Law = New_Law(); */
/* 	    New_Trans->Trans_Law->Name = 'C'; */
/* 	    New_Trans->Trans_Law->Parameters = (double*)  */
/* 		malloc( sizeof( double ) ); */
/* 	    New_Trans->Trans_Law->Parameters[0] = 0.0; */
/* 	    New_Trans->Name = Copy_Name( "Dummy_Transition" ); */
/* 	    New_Trans->Outdegree = 0; */
/* 	    New_Trans->Outputs = NULL; */
/* 	    New_Trans->Indegree = 0; */
/* 	    New_Trans->Inputs = NULL; */
/* 	    New_Trans->Next = NULL; */
/* 	} */
	
/* 	return RdP; */
/*     } */
/* } */
