/* Programmere/module de generation d'un programme C effectuant		*/
/* la simulation equationelle d'un Graphe d'evenements.			*/

#include "eg.h"

#define MAXINT          0x7FFFFFFF
#define INFINI		1e12

#define max(x,y)        ( (x>y) ? x : y )
#define min(x,y)        ( (x<y) ? x : y )

#define	TOOLNAME	"eg2graph"

#define CODE(x)		printf(x)
#define CODE1(x,y)	printf(x,y)
#define CODE2(x,y,z)	printf(x,y,z)
#define CODE3(x,y,z,t)	printf(x,y,z,t)
#define CODELN		CODE("\n")

#define FIG_BLACK	0
#define FIG_BLUE	1
#define FIG_GREEN	2
#define FIG_CYAN	3
#define FIG_RED		4
#define FIG_MAGENTA	5
#define FIG_YELLOW	6
#define FIG_WHITE	7
#define FIG_SOLID	0
#define FIG_DASH	1
#define FIG_DOTTED	2

/* default dimensions for objects 					*/
#define	WIDTH	50		/* horiz. space betw. nodes		*/
#define	HEIGHT	50		/* vert. space betw. nodes		*/
#define	RADIUS	5		/* node size				*/
#define	X_OFF	50		/* horiz. offset			*/
#define	Y_OFF	50		/* vert. offset				*/

#define	NB_LEVELS	10

/* Global declarations */
extern EventGraph	*Read_Network(char* File_Name);
extern void		dumpe_network(EventGraph*);
extern boolean		topological_analysis(EventGraph*);
extern char		*todays_date(void);

int	*Trans_Col;
int	X_Hop = WIDTH;
int	Y_Hop = HEIGHT;
int	X_Offset = X_OFF;
int	Y_Offset = Y_OFF;
int	Radius = RADIUS;
boolean	Is_Color = TRUE;
int	Nb_Levels = NB_LEVELS;

/* forward declaration */
void	gen_fig_code( EventGraph *Reseau, int Max_Levels );

/* the program itself... */
int main(
int	argc,
char	*argv[] )
{
  static char		File_Name[128] = "";
  static char		Code_File_Name[128] = "";
  EventGraph 		*Reseau;
  int			i;
  boolean		Help = FALSE;
  boolean		Args_Ok = TRUE;
  boolean		File_Ok = TRUE;

  i = 1;

  if (argc>1)
    do {
      if GET_STRING( "-f", File_Name )
      else if GET_STRING( "-o", Code_File_Name )

      else if SWITCH( "-debug", Debug, TRUE )
      else if SWITCH( "-bw", Is_Color, FALSE )
      else if SWITCH( "-color", Is_Color, TRUE )

      else if GET_INT( "-dx", X_Hop )
      else if GET_INT( "-dy", Y_Hop )
      else if GET_INT( "-n", Nb_Levels )
      else if GET_INT( "-r", Radius )

      else if IS_VERSION

      else if ( ( 0 == strcmp( "-h", argv[i] ) )
		|| ( 0 == strcmp( "-?", argv[i] ) ) ) {
	printf("Usage: %s [-dx <number] [-dy <number>] [-f <file>] ", argv[0]);
	printf("[-bw] [-color]\n" );
	printf("\t\t\t[-debug] [-o <file>] [-n <number>]");
	printf(" [-r <number>]\n");
	Args_Ok = FALSE;
	Help = TRUE;
	i++;
      }

      else	i++; /* skip unknown argument */
    }
    while ( Args_Ok && ( i < argc ) );

  if (Args_Ok) {
    
    Reseau = Read_Network(File_Name);

    if ( Reseau != NULL ) {

      if ( topological_analysis(Reseau) ) {
	if (Debug)
	  dumpe_network(Reseau);

	if ( strlen(Code_File_Name) != 0 )
	  File_Ok = ( NULL != freopen( Code_File_Name, "w", stdout ) );

	if ( File_Ok ) {
	  gen_fig_code( Reseau, Nb_Levels );
	  return 0;
	}
	else {
	  fprintf( stderr, "%s: Error opening file %s.\n", 
		   TOOLNAME, Code_File_Name );
	  return -1;
	}
      }
      else {
	fprintf( stderr, "Topological analysis failed: ");
	fprintf( stderr, "look for an empty cycle.\n");
	return -1;
      }
    }
    else {
      fprintf( stderr, "%s: Error reading network. No translation.\n", 
	       TOOLNAME );	
      return -1;
    }
  }
  else if (!Help)
    return -1;
  else
    return 0;
}

/****************************************************************/
/* Initialization of auxiliary data...				*/
/****************************************************************/
void 	init_id_table( 
EventGraph	*Reseau )
{
  int	Num_Layer;
  int	Num_Node;
  int	i = 0;

  Trans_Col = (int*) calloc( (unsigned)Reseau->Size, sizeof(int) );

  for( Num_Layer = 0; Num_Layer < Reseau->Nb_Layers; Num_Layer++)
    for ( Num_Node = 0; 
	  Num_Node < Reseau->Level[Num_Layer]->Nb_Nodes;
	  Num_Node++ )
      {
	Trans_Col[Reseau->Level[Num_Layer]
		 ->Node_Id[Num_Node]] = i;
	i++;
      }
}


/****************************************************************/
/* Procedures for code generation...				*/
/****************************************************************/

void	gen_file_header(
EventGraph	*Reseau )
{
        CODE( "#FIG 2.1\n");
	CODE( "# Fig code for an infinite task graph\n");
	CODE( "# generated by ");
	CODE(TOOLNAME); CODE(" version "); CODE(VERSION);
	CODE1("\n# Date: %s \t\t\t\t\n#\n", todays_date() );

        /* resolution, coordinate system */
        CODE( "80 2\n#\n");
}

void gen_trans_numbers( Reseau )
EventGraph	*Reseau;
{
	int	i;

	CODE("# Transition numbers\n");

	for( i=0; i<Reseau->Size; i++ )
	{	
		CODE( "4 0 0 12 0 -1 0 0.0 2 10 10" );
		CODE2( " %d %d", X_Offset + Trans_Col[i]*X_Hop, 
						Y_Offset * 3 /4 );
	
		CODE2( " %d%c\n", i+1, (char)1 );
	}
}
	
void draw_node( xx, yy )
int	xx, yy;
{
	int	Node_Color;

	if ( Is_Color )
		Node_Color = FIG_BLUE;
	else	Node_Color = FIG_BLACK;

	CODE1( "1 3 0 1 %d 0 -1 21 0.0 1 0.0 ", Node_Color );

	printf( "%d %d %d %d %d %d %d %d\n",
			xx, yy, Radius, Radius, xx, yy, xx-2, yy-2 );
}

void draw_direct_arc( Orig, Dest, Level )
int	Orig;
int	Dest;
int	Level;
{
	int	Color;
	int	x1, y1, x2, y2;

	if ( Is_Color )
		Color = FIG_RED;
	else 	Color = FIG_BLACK;

	/* Position du premier point */
	x1 = X_Offset + Trans_Col[ Orig ] * X_Hop;
	y1 = Y_Offset + Level * Y_Hop;

	/* Position du dernier point */
	x2 = X_Offset + Trans_Col[ Dest ] * X_Hop;
	y2 = y1;

	CODE2( "2 1 %d 1 %d 0 -1 0 0.0 0 1 0\n\t", FIG_SOLID, 
							Color );
	CODE( "0 0 1.0 8.0 16.0\n\t" );
	printf( " %d %d %d %d 9999 9999\n#\n", x1, y1, x2, y2 );

}

void draw_arc(
int	Orig,
int	Dest,
int	Level,
int	Hop )
{
  int	Color;
  int	x1, y1, x2, y2, lx1, ly1, lx2, ly2;

  if ( Is_Color )
    Color = FIG_RED;
  else 	Color = FIG_BLACK;

  /* Position du premier point */
  x1 = X_Offset + Trans_Col[ Orig ] * X_Hop;
  y1 = Y_Offset + Level * Y_Hop;

  /* Position du dernier point */
  x2 = X_Offset + Trans_Col[ Dest ] * X_Hop;
  y2 = Y_Offset + ( Level + Hop ) * Y_Hop;

  /* if the arc is downwards, and "good" draw it direct... */
  if ( ( Hop == 1 ) || 
       ( abs( Trans_Col[Dest] - Trans_Col[Orig] ) == 1 ) ||
       ( ( Hop != 0 ) 
	 && ( ( Trans_Col[Dest] - Trans_Col[Orig] ) % Hop != 0 )
	 && ( Hop % ( Trans_Col[Dest] - Trans_Col[Orig] ) != 0 )
	 )
       ) {
    CODE2( "2 1 %d 1 %d 0 -1 0 0.0 0 1 0\n\t", FIG_SOLID, Color );
    CODE( "0 0 1.0 8.0 16.0\n\t" );
    printf( " %d %d %d %d", x1, y1, x2, y2 );
  }
  else {	/* if not draw a B-spline */
    CODE2( "3 0 %d 1 %d 0 -1 0 0.0 1 0\n\t", FIG_SOLID, Color );
    CODE( "0 0 1.0 8.0 16.0\n\t" );

    /* Position du premier point intermediaire */
    lx1 = ( 2*x1 + x2 ) / 3;
    ly1 = ( 2*y1 + y2 ) / 3 + Y_Hop / 4;

    /* Position du deuxieme point intermediaire */
    lx2 = ( x1 + 2*x2 ) / 3;
    ly2 = ( y1 + 2*y2 ) / 3 + Y_Hop / 4;

    printf( " %d %d %d %d %d %d %d %d", x1, y1, lx1, ly1,
	    lx2, ly2, x2, y2 );
  }

  /* fini */
  CODE( " 9999 9999\n#\n" );
}

void	gen_code_node(
EventGraph	*Reseau,
int		The_Node,
int		Level )
{
  Liste_Place	Outputs;
  Liste_Place	L_Scan;
  int		Dest_Id;
  int		Hop;

  CODE1("# Transition %d\n", The_Node);

  draw_node( X_Offset + Trans_Col[ The_Node ] * X_Hop,
	     Y_Offset + Level * Y_Hop );

  Outputs = Reseau->T[The_Node]->Outputs;

  for( L_Scan = Outputs; L_Scan != NULL; L_Scan = L_Scan->Next ) {

    Dest_Id = L_Scan->Val->Trans_Id - 1;
    Hop = L_Scan->Val->Init_Pop;

    if ( ( Hop == 0 )
	 && ( Trans_Col[Dest_Id] == Trans_Col[ The_Node ] + 1 ) )
      draw_direct_arc( The_Node, Dest_Id, Level );
    else	
      draw_arc( The_Node, Dest_Id, Level, Hop );
  }

  /* vertical arrow */
  /*	draw_arc( The_Node, The_Node, Level, 1 );
   */	
}

void 	gen_level_code(
EventGraph	*Reseau,
int		Level )
{
  int	Num_Layer;
  int	Num_Node;
  int	The_Node;

  CODE1("# Level %d\n#\n", Level );

  for( Num_Layer = 0; Num_Layer < Reseau->Nb_Layers; Num_Layer++)
    for ( Num_Node = 0; 
	  Num_Node < Reseau->Level[Num_Layer]->Nb_Nodes;
	  Num_Node++ ) {
      The_Node = Reseau->Level[Num_Layer]
	->Node_Id[Num_Node];
      gen_code_node( Reseau, The_Node, Level );
    }
}

void	gen_fig_code(
EventGraph	*Reseau, 
int		Max_Levels )
{
  int		The_Level;

  init_id_table( Reseau );

  gen_file_header( Reseau );

  gen_trans_numbers( Reseau );

  for( The_Level = 0; The_Level < Max_Levels; The_Level++ )
    gen_level_code( Reseau, The_Level );

}
