#include <math.h>

#include "eg.h"

#define TOOLNAME	"eg2matrix"
#define EPSILON		-1

#define SCAN(x,y)       for(x=y;x!=NULL;x=x->Next)

/* definitions for the output language */
#define BRUNO	0
#define ALAIN	1
#define SVEN	2
#define SCILAB	3
/* definitions for the contents of the output */
#define COMPLET	0
#define MATRICE 1
#define VECTEUR 2
#define SEQUENCE 3
/* definitions for the format of the matrix */
#define FULL	0
#define SPARSE	1

/* external declarations from the module matrice.c */
extern EventGraph *Le_Reseau;
extern int Get_Event_Graph(char*);
extern boolean Init_Times(int);
extern double Sample(Law_Desc);
extern double max(double,double);
extern void Fill_Matrix(double*);
extern void Fill_Slice(double*,int);
extern void Fill_Initial_Vector( double *Y );
extern void Fill_Initial_Vector_Flat( double* Y, int Delta );

/* forward declarations */
boolean	Is_Integral( double *A, int Size );
void Fill_Initial_Vector_Flagged( double *Y, int N, int Trans_Nb );
void Write_Matrix_As_Full( double *A, int N, int, char*, boolean );
void Write_Matrix_As_Sparse( double *A, int N, int, char*, boolean );
void Write_Vector( double *Y, int N, int, char*, boolean );
void Write_Matrix_Sequence_As_Full( EventGraph*, int, char*, boolean );
void Write_Matrix_Sequence_As_Sparse( EventGraph*, int, char*, boolean );
void Write_Scilab_Sequence( EventGraph*, char*, int, boolean );

/* main program */
int main(int argc, char *argv[])
{
  static char	File_Name[128] = "";
  static char	Out_File_Name[128] = "";
  static char	Mat_Prefix[128] = "";
  int		i;
  int		N;
  int		Height = 1;
  double	*A;
  double	*Y;
  int		Language = BRUNO;
  int		Format = FULL;
  int		Infos = COMPLET;
  boolean	Help = FALSE;
  boolean	Args_Ok = TRUE;
  boolean	File_Ok = TRUE;
  boolean	Integer = FALSE;
  int		Flagged_Trans_Nb = -1;

  i = 1;

  if (argc>1)
    do {

      if GET_STRING( "-f", File_Name )
      else if GET_STRING( "-o", Out_File_Name )
      else if GET_STRING( "-prefix", Mat_Prefix )

      else if SWITCH( "-debug", Debug, TRUE )
      else if SWITCH( "-alain", Language, ALAIN )
		      /* Sven's language obsoletized as of ERS 0.2.1 */
		      /* else if SWITCH( "-sven", Language, SVEN ) */
      else if SWITCH( "-scilab", Language, SCILAB )
      else if SWITCH( "-int", Integer, TRUE )
      else if SWITCH( "-matrix", Infos, MATRICE )
      else if SWITCH( "-vector", Infos, VECTEUR )
      else if SWITCH( "-sequence", Infos, SEQUENCE )
      else if SWITCH( "-sparse", Format, SPARSE )
      else if SWITCH( "-full", Format, FULL )

      else if GET_INT( "-slice", Height )
      else if GET_INT( "-flag_trans", Flagged_Trans_Nb )

      else if IS_VERSION

      else if ( ( 0 == strcmp( "-h", argv[i] ) )
		|| ( 0 == strcmp( "-?", argv[i] ) ) )
	{	printf("Usage: %s [-f <file>] ", TOOLNAME );
	printf("[-debug] [-o <file>] [-slice <int>] ");
	printf("\n\t[-scilab] [-prefix <name>] ");
	printf("[-sparse] [-full] [-matrix] [-vector] [-sequence]");
	printf("\n\t[-int] [-flag_trans <int>]\n");
	Args_Ok = FALSE;
	Help = TRUE;
	i++;
	}

      else
	i++; /* skip unknown argument */
    }
    while ( Args_Ok && ( i < argc ) );

  if (Args_Ok) {

    N = Get_Event_Graph(File_Name);

    if ( N == -1 ) {
      fprintf( stderr, "%s: Error reading network. No translation.\n", 
	       TOOLNAME );
      return -6;
    }
    else if ( N == -3 ) {
      printf("%s: error: Network not FIFO.\n", TOOLNAME );
      return -3;
    }

    /* lack of liveness or flatness is ok if mode is SEQUENCE */
    if ( Infos != SEQUENCE ) {
      if ( N == -2 ) {
	printf("%s: error: Network not live.\n", TOOLNAME );
	return -2;
      }
      else if ( N == -4 ) {
	printf("%s: error: Network not flat.\n", TOOLNAME );
	return -4;
      }
    } 
    else if ( N < -4 ) {
      printf("%s: error: unknown error (code=%d).\n", TOOLNAME, N );
      return N;
    }

    /* redirecting the output if necessary */
    if ( strlen(Out_File_Name) != 0 )
      File_Ok = ( NULL != freopen( Out_File_Name, "w", stdout ) );

    if ( !File_Ok ) {
      fprintf( stderr, "%s: Error opening file %s.\n", 
	       TOOLNAME, Out_File_Name );
      return -1;
    } 
    else {
      /* Output is ok. Switch according to SEQUENCE / single matrix */
      if ( Infos != SEQUENCE ) {
	if ( strlen( Mat_Prefix ) == 0 ) {
	  strcpy( Mat_Prefix, "A" );
	}

	A = (double*) calloc( N*N, sizeof( double ) );

	if ( Height < 0 )
	  Fill_Matrix( A );
	else {
	  Init_Times( Height );
	  Fill_Slice( A, Height );
	}

	/* Dump Matrix if requested */
	if ( ( Infos == MATRICE ) || ( Infos == COMPLET ) ) {

	  if ( Integer )
	    if ( !Is_Integral( A, N*N ) ) {
	      fprintf(stderr, "%s: warning: the matrix ", TOOLNAME );
	      fprintf(stderr, "has non-integer entries. Will be");
	      fprintf(stderr, " rounded off\n");
	    }

	  if ( Format == FULL ) {
	    Write_Matrix_As_Full( A, N, Language, Mat_Prefix, Integer );
	  }
	  else {
	    Write_Matrix_As_Sparse( A, N, Language, Mat_Prefix, Integer );
	  }
	}
	
	/* Dump vector if requested */
	if ( ( Infos == VECTEUR ) || ( Infos == COMPLET ) ) {

	  /* Filling up the initial condition */
	  Y = (double*) calloc( (unsigned)N, sizeof(double) );

	  if ( ( Flagged_Trans_Nb <= 0 ) || ( Flagged_Trans_Nb > N ) ) {
	    /* normal initialization from the PN description */
	    Fill_Initial_Vector( Y );
	  }
	  else {
	    Fill_Initial_Vector_Flat( Y, Flagged_Trans_Nb - 1 );
	  }

	  Write_Vector( Y, N, Language, Mat_Prefix, Integer );

	}
	/* successful translation */
	return 0;
      }
      else {
	/* mode is SEQUENCE */
	/* Scilab will use (A,B,C) matrices when no prefix is given.	*/
	/* Other languages will use "A".				*/
	if ( ( Language != SCILAB ) && ( strlen( Mat_Prefix ) == 0 ) ) {
	  strcpy( Mat_Prefix, "A" );
	}

	if ( Language != SCILAB ) {
	  if ( Format == FULL ) {
	    Write_Matrix_Sequence_As_Full( Le_Reseau,
					   Language, Mat_Prefix, Integer );
	  } 
	  else {
	    Write_Matrix_Sequence_As_Sparse( Le_Reseau,
					     Language, Mat_Prefix, Integer );
	  }
	}
	else {
	    Write_Scilab_Sequence( Le_Reseau, Mat_Prefix, Format, Integer );
	}
	/* successful translation */
	return 0;
      }
    }
  }
  else if (!Help)
    return -1;
  else
    return 0;
}

boolean Is_Integral( double *A, int Size )
{
  int	i;
  boolean Ok = TRUE;

  for( i=0; Ok && (i<Size); i++ )
    Ok = Ok && ( fabs( A[i] - floor( A[i] ) ) < 1e-6 );

  return Ok;
}

void Write_Epsilon( int Language )
{
  switch ( Language ) {
  case BRUNO:
    printf("    eps    ");
    break;
  case ALAIN:
    printf("    oo     ");
    break;
  case SCILAB:
    printf("    %%0     ");
    break;
  }
}

void Write_Matrix_As_Full( double *A, int N, 
			   int Language, char *Mat_Prefix, boolean Integer )
{
  int		i;
  int		j;

  /* Write matrix "header" */
  switch ( Language ) {
  case BRUNO:
    printf("%d\n", N );
    break;
  case ALAIN:
    printf("%s=\n", Mat_Prefix );
    break;
  case SCILAB:
    printf("%s=#([\n\t", Mat_Prefix );
    break;
  }

  /* write matrix body */
  for (i=0;i<N;i++) {
    for (j=0;j<N;j++) {
      if ( A[i*N+j] > EPSILON ) {
	if ( Integer )
	  printf("%10d ",(int)A[i*N+j]);
	else
	  printf("%10.4f ",A[i*N+j]);
      }
      else {
	Write_Epsilon( Language );
      }
    }
   
    /* write line/matrix trailer */
    if ( Language == SCILAB ) {
      if ( i < N-1 )
	printf(";\n\t");
      else
	printf("])\n");
    }
    else
      printf("\n");
  }

}

void Write_Matrix_As_Sparse( double *A, int N, 
			     int Language, char *Mat_Prefix, boolean Integer )
{
  int		i;
  int		j;
  int		NbElts;
  int		NumElt;
  int		LineLength = 0;

  /* Write matrix "header" */
  switch ( Language ) {
  case BRUNO:
    printf("%d\n", N );
    printf("sp\n");
    break;
  case ALAIN:
    printf("%s=\n", Mat_Prefix );
    printf("sp\n");
    break;
  case SCILAB:
    printf("%s=#([\n\t", Mat_Prefix );
    LineLength = 8;
    break;
  }

  /* Scilab's format needs special treatment. Handled apart. */
  if ( Language != SCILAB ) {
    /* write matrix body */
    for (i=0;i<N;i++) {
      for (j=0;j<N;j++) {
	if ( A[i*N+j] > EPSILON ) {
	  switch ( Language ) {
	  case BRUNO:
	  case ALAIN:
	    printf("%4d %4d ", i, j );
	    break;
	  }
	  if ( Integer ) {
	    printf("%10d\n",(int)A[i*N+j]);
	  }
	  else
	    printf("%10.4f\n",A[i*N+j]);
	}
      }
    }
  }
  else {
    /* first count elements */
    NbElts = 0;
    for (i=0;i<N;i++)
      for (j=0;j<N;j++)
	if ( A[i*N+j] > EPSILON )
	  NbElts++;

    /* now write index array */
    NumElt = 0;
    for (i=0;i<N;i++) {
      for (j=0;j<N;j++) {
	if ( A[i*N+j] > EPSILON ) {
	  printf("%3d,%3d", i, j );
	  LineLength += 8;
	  if ( NumElt < NbElts-1 )
	    printf(";");
	  if ( LineLength > 72 ) {
	    printf("\n\t");
	    LineLength = 8;
	  }
	  NumElt ++;
	}
      }
    }
    /* now write value array */
    printf("],\n       [");
    LineLength = 8;
    NumElt = 0;
    for (i=0;i<N;i++) {
      for (j=0;j<N;j++) {
	if ( A[i*N+j] > EPSILON ) {
	  if ( Integer ) {
	    printf("%3d",(int)A[i*N+j]);
	    LineLength += 3;
	  }
	  else {
	    printf("%10.4f",A[i*N+j]);
	    LineLength += 10;
	  }
	  if ( NumElt < NbElts-1 ) {
	    printf(",");
	    LineLength += 1;
	  }
	  if ( LineLength > 72 ) {
	    printf("\n\t");
	    LineLength = 8;
	  }
	  NumElt ++;
	}
      }
    }
  }

  /* write matrix trailer */
  if ( Language == SCILAB )
    printf("])\n");
  else
    printf("end\n");

}

void Write_Vector( double *Y, int N, 
		   int Language, char *Mat_Prefix, boolean Integer )
{
  int		i;

  switch ( Language ) {
  case ALAIN:
    printf("Y(0)=\n");
    break;
  case SCILAB:
    printf("Y0_%s=#([", Mat_Prefix );
    break;
  default:
    /* print nothing */
    break;
  }
		    
  for (i=0;i<N;i++) {
    if ( Y[i] > EPSILON ) {
      if ( Integer )
	printf("%10d ",(int)Y[i] );
      else
	printf("%10.4f ",Y[i] );
    }
    else {
      Write_Epsilon( Language );
    }
  }
  if ( Language == SCILAB )
    printf("])");
  printf("\n");

}

void Write_Matrix_Sequence_As_Full( EventGraph *Reseau,
				    int Language, char *Mat_Prefix, 
				    boolean Integer )
{	
  int		i;
  int		j;
  int		k;
  int		m;
  int		The_Size;
  int		Mu;
  double	x;
  double	*Phi;
  Liste_Place	P_Scan;
  Place_Desc	The_Place;

  The_Size = Reseau->Size;

  /* get a single sample of the n-th transition firing */
  Phi = (double*) calloc( (unsigned)The_Size, sizeof(double) );
  for ( i=0; i<The_Size; i++ ) 
    Phi[i] = Sample( Reseau->T[i]->Trans_Law );

  /* print file header */
  if ( Language == SCILAB ) {
    printf("// Scilab matrices written by %s\n//\n", 
	   TOOLNAME );
  } else {
    printf("%d\n", The_Size );
    /* print the dummy probability line */
    for ( Mu = 0; Mu <= Reseau->Max_Pop_Init; Mu++ )
      printf(" 1");
    printf("\n");
  }

  for ( m = 0; m <= Reseau->Max_Pop_Init; m++ ) {
    
    switch ( Language ) {
    case BRUNO:
      printf("\n");
      break;
    case ALAIN:
      printf("%s%1d=\n", Mat_Prefix, m );
      break;
    case SCILAB:
      printf("%s%1d=#([\n\t", Mat_Prefix, m );
      break;
    }

    /* write matrix body */
    for (i=0;i<The_Size;i++) {
      for (j=0;j<The_Size;j++) {
	x = EPSILON;
	SCAN( P_Scan, Reseau->T[j]->Inputs ) {
	  The_Place = P_Scan->Val;
	  k = The_Place->Input_Id-1;
	  Mu = The_Place->Init_Pop;

	  if ( ( i == k ) && ( Mu == m ) ) {
	    x = max( x, Sample( The_Place->Place_Law ) );
	  }
	}

	x = ( x > EPSILON ? x + Phi[j] : EPSILON );

	if ( x > EPSILON ) {
	  if ( Integer )
	    printf("%10d ",(int) x );
	  else
	    printf("%10.4f ", x );
	}
	else {
	  Write_Epsilon ( Language );
	}
      }
      /* write line/matrix trailer */
      if ( Language == SCILAB ) {
	if ( i < The_Size-1 )
	  printf(";\n\t");
	else
	  printf("])\n");
      }
      else
	printf("\n");
    }
  }
}

void Write_Matrix_Sequence_As_Sparse( EventGraph *Reseau,
				      int Language, char *Mat_Prefix, 
				      boolean Integer )
{	
  int		i;
  int		j;
  int		k;
  int		m;
  int		The_Size;
  int		Mu;
  int		NbElts;
  int		NumElt;
  double	x;
  double	*Phi;
  Liste_Place	P_Scan;
  Place_Desc	The_Place;

  The_Size = Reseau->Size;

  /* get a single sample of the n-th transition firing */
  Phi = (double*) calloc( (unsigned)The_Size, sizeof(double) );
  for ( i=0; i<The_Size; i++ ) 
    Phi[i] = Sample( Reseau->T[i]->Trans_Law );

  /* print file header */
  if ( Language == SCILAB ) {
    printf("// Scilab matrices written by %s\n//\n", 
	   TOOLNAME );
  } else {
    printf("%d\n", The_Size );
    printf("%d\n", Reseau->Max_Pop_Init+1 );
    /* print the dummy probability line */
    for ( Mu = 0; Mu <= Reseau->Max_Pop_Init; Mu++ )
      printf(" 1");
    printf("\n");
  }

  for ( m = 0; m <= Reseau->Max_Pop_Init; m++ ) {
    
    /* Count elements. Useful only for scilab */
    if ( Language == SCILAB ) {
      NbElts = 0;
      for (i=0;i<The_Size;i++) {
	for (j=0;j<The_Size;j++) {
	  x = EPSILON;
	  SCAN( P_Scan, Reseau->T[j]->Inputs ) {
	    The_Place = P_Scan->Val;
	    k = The_Place->Input_Id-1;
	    Mu = The_Place->Init_Pop;

	    if ( ( i == k ) && ( Mu == m ) ) {
	      x = max( x, Sample( The_Place->Place_Law ) );
	    }
	  }

	  if ( x > EPSILON )
	    NbElts++;
	}
      }
    }

    /* Write matrix "header" */
    switch ( Language ) {
    case BRUNO:
      printf("sp\n");
      break;
    case ALAIN:
      printf("%s%1d=\n", Mat_Prefix, m );
      printf("sp\n");
      break;
    case SCILAB:
      printf("%s%1d=spmat([\n\t", Mat_Prefix, m );
      break;
    }

    /* write matrix body */
    NumElt = 0;

    for (i=0;i<The_Size;i++) {
      for (j=0;j<The_Size;j++) {
	x = EPSILON;
	SCAN( P_Scan, Reseau->T[j]->Inputs ) {
	  The_Place = P_Scan->Val;
	  k = The_Place->Input_Id-1;
	  Mu = The_Place->Init_Pop;

	  if ( ( i == k ) && ( Mu == m ) ) {
	    x = max( x, Sample( The_Place->Place_Law ) );
	  }
	}

	x = ( x > EPSILON ? x + Phi[j] : EPSILON );

	if ( x > EPSILON ) {
	  if ( Integer ) {
	    switch ( Language ) {
	    case BRUNO:
	    case ALAIN:
	      printf("%4d %4d %10d\n", i, j, (int) x );
	      break;
	    case SCILAB:
	      printf("%4d,%4d,%10d", i, j, (int) x );
	      break;
	    }
	  }
	  else {
	    switch ( Language ) {
	    case BRUNO:
	    case ALAIN:
	      printf("%4d %4d %10.4f\n", i, j, x );
	      break;
	    case SCILAB:
	      printf("%4d,%4d,%10.4f", i, j, x );
	      break;
	    }
	  }
	  if ( ( Language == SCILAB ) && ( NumElt < NbElts-1 ) ) {
	    printf(";\n\t");
	    NumElt++;
	  }
	}
      }
    }
    /* write line/matrix trailer */
    if ( Language == SCILAB )
      printf("])\n");
    else
      printf("end\n");
  }
}

void Write_Scilab_Element( int i, int j, double x, 
			   int Format, boolean Integer )
{
  if ( x > EPSILON ) {
    if ( Format == SPARSE ) {
      printf("%4d,%4d,", i, j );
    }
    if( Integer )
      printf("%10d ",(int) x );
    else
      printf("%10.4f ", x );
  }
  else if ( Format != SPARSE )
    Write_Epsilon ( SCILAB );

}

/************************************************************************/
/* Procedure to find the maximum marking of places between transitions	*/
/* of two given types.							*/
/* Returns -1 if no such place exists.					*/
/************************************************************************/
int Count_Maximal_Marking( EventGraph *Reseau, 
			   int From_Type, int To_Type, int *Trans_Type )
{
  int		i;
  int		j;
  int		Res;
  int		Mu;
  Liste_Place	P_Scan;
  Place_Desc	The_Place;

  Res = -1;

  for (i=0;i<Reseau->Size;i++) {
    if ( Trans_Type[i] == From_Type ) {
      SCAN( P_Scan, Reseau->T[i]->Outputs ) {
	The_Place = P_Scan->Val;
	j = The_Place->Trans_Id-1;
	Mu = The_Place->Init_Pop;

	if ( Trans_Type[j] == To_Type ) {
	  Res = ( Res > Mu ? Res : Mu );
	}
      }
    }
  }

  return Res;

}

void Gen_Scilab_Matrix( EventGraph *Reseau, 
			char *Mat_Prefix, char *Mat_Postfix, int m, 
			int From_Type, int To_Type, int Rows, int Cols,
			int Format, boolean Integer,
			int *Trans_Type, int *Trans_Number, 
			double *Phi )
{
  int		i;
  int		j;
  int		k;
  int		Mu;
  int		NbElts;
  int		NumElt;
  int		The_Size;
  double	x;
  Liste_Place	P_Scan;
  Place_Desc	The_Place;

  The_Size = Reseau->Size;

  /* count the elements of the matrix. Useful when sparse, also for 	*/
  /* full when matrix empty.						*/
  NbElts = 0;
  for (i=0;i<The_Size;i++) {
    if ( Trans_Type[i] == From_Type ) {
      for (j=0;j<The_Size;j++) {
	if ( Trans_Type[j] == To_Type ) {
	  x = EPSILON;
	  SCAN( P_Scan, Reseau->T[j]->Inputs ) {
	    The_Place = P_Scan->Val;
	    k = The_Place->Input_Id-1;
	    Mu = The_Place->Init_Pop;
	      
	    if ( ( i == k ) && ( Mu == m ) ) {
	      x = max( x, Sample( The_Place->Place_Law ) );
	    }
	  }
	    
	  if ( x > EPSILON )
	    NbElts++;
	}
      }
    }
  }

  /* write matrix header */
  if ( 0 == strlen( Mat_Prefix ) ) 
    printf("%s%1d=", Mat_Postfix, m );
  else 
    printf("%s_%s%1d=", Mat_Prefix, Mat_Postfix, m );

  if ( NbElts == 0 )
    printf("%%zeros(%d,%d)\n", Rows, Cols );
  else {
    /* printf("%s([\n\t", ( Format == FULL ? "#" : "spmat" ) ); */
    printf("([\n\t");

    /* write matrix body */
    NumElt = 0;

    for (i=0;i<The_Size;i++) {
      if ( Trans_Type[i] == From_Type ) {
	for (j=0;j<The_Size;j++) {
	  if ( Trans_Type[j] == To_Type ) {
	    x = EPSILON;
	    SCAN( P_Scan, Reseau->T[j]->Inputs ) {
	      The_Place = P_Scan->Val;
	      k = The_Place->Input_Id-1;
	      Mu = The_Place->Init_Pop;
		
	      if ( ( i == k ) && ( Mu == m ) ) {
		x = max( x, Sample( The_Place->Place_Law ) );
	      }
	    }
	      
	    x = ( x > EPSILON ? x + Phi[j] : EPSILON );
	      
	    Write_Scilab_Element( Trans_Number[i], Trans_Number[j], x,
				  Format, Integer );

	    if ( ( Format == SPARSE ) 
		 && ( x > EPSILON ) 
		 && ( NumElt < NbElts-1 ) ) {
	      printf(";\n\t");
	      NumElt++;
	    }
	  }
	}

	/* write line trailer */
	if ( ( Format == FULL ) && ( Trans_Number[i] < Rows-1 ) )
	  printf(";\n\t");
      }
    }
    /* write matrix trailer */
    printf("])\n");
  }
}

void Write_System_Code( char *Mat_Prefix, int Max_Pop[4] )
{
  int		m;

  printf("%sSystem = [", Mat_Prefix );

  printf( " [" );
  for ( m = 0; m <= Max_Pop[0]; m++ ) {
    printf("%sA%d", Mat_Prefix, m );
    if ( m < Max_Pop[0] )
      printf(",");
  }
  printf("],");
  printf( " [" );
  for ( m = 0; m <= Max_Pop[1]; m++ ) {
    printf("%sB%d", Mat_Prefix, m );
    if ( m < Max_Pop[1] )
      printf(",");
  }
  printf("],");
  printf( " [" );
  for ( m = 0; m <= Max_Pop[2]; m++ ) {
    printf("%sC%d", Mat_Prefix, m );
    if ( m < Max_Pop[2] )
      printf(",");
  }
  printf("],");
  printf( " [" );
  for ( m = 0; m <= Max_Pop[3]; m++ ) {
    printf("%sD%d", Mat_Prefix, m );
    if ( m < Max_Pop[3] )
      printf(",");
  }
  printf("] ]\n");

}

void Write_Restricted_System_Code( char *Mat_Prefix, int Max_Pop[4] )
{
  printf("s=mpsyslin(\n");
  printf("\tsparse(A0(:,[1,2]),A0(:,3)),\n");
  printf("\tsparse(B0(:,[1,2]),B0(:,3)),\n");
  printf("\tsparse(C0(:,[1,2]),C0(:,3)),\n");
  printf("\tsparse(A1(:,[1,2]),A1(:,3)))\n");
}

void Write_Scilab_Sequence( EventGraph *Reseau, char *Mat_Prefix, 
			    int Format, boolean Integer )
{	
  int		i;
  int		m;
  int		The_Size;
  double	*Phi;
  int		*Trans_Type;
  int		*Trans_Number;
  int		Input_Nb;
  int		Output_Nb;
  int		Trans_Nb;
  int		Max_Pop[4];

  The_Size = Reseau->Size;

  /* get a single sample of the n-th transition firing */
  Phi = (double*) calloc( (unsigned)The_Size, sizeof(double) );
  for ( i=0; i<The_Size; i++ ) 
    Phi[i] = Sample( Reseau->T[i]->Trans_Law );

  /* Classification of the transitions as I/O/N and numbering */
#define NORMAL_TRANS 0
#define INPUT_TRANS  1
#define OUTPUT_TRANS 2
 
  Trans_Type = (int*) calloc( (unsigned)The_Size, sizeof(int) );
  Trans_Number = (int*) calloc( (unsigned)The_Size, sizeof(int) );
  Input_Nb = 0;
  Output_Nb = 0;
  Trans_Nb = 0;

  for ( i=0; i<The_Size; i++ ) {
    if ( Reseau->T[i]->Indegree == 0 ) {
      Trans_Type[i] = INPUT_TRANS;
      Trans_Number[i] = Input_Nb;
      Input_Nb++;
    }
    else if ( Reseau->T[i]->Outdegree == 0 ) {
      Trans_Type[i] = OUTPUT_TRANS;
      Trans_Number[i] = Output_Nb;
      Output_Nb++;
    }
    else {
      Trans_Type[i] = NORMAL_TRANS;
      Trans_Number[i] = Trans_Nb;
      Trans_Nb++;
    }
  }

  /* print file header */
  printf("// Scilab matrices written by %s\n", TOOLNAME );
  printf("// %d inputs, %d outputs, %d internal transitions\n//\n",
	 Input_Nb, Output_Nb, Trans_Nb );

  if ( Reseau->Max_Pop_Init > 1 ) {
    fprintf( stderr, "%s: warning: the network is not flat. The (max,+) system may be incomplete.\n", TOOLNAME );
  }

  /* Generate A matrices */
  Max_Pop[0] = Count_Maximal_Marking( Reseau, 
				      NORMAL_TRANS, NORMAL_TRANS, Trans_Type );

  for ( m = 0; m <= Max_Pop[0]; m++ ) {

    Gen_Scilab_Matrix( Reseau,
		       Mat_Prefix, "A", m,
		       NORMAL_TRANS, NORMAL_TRANS, Trans_Nb, Trans_Nb,
		       Format, Integer,
		       Trans_Type, Trans_Number,
		       Phi );

  }

  /* Generate B matrices */
  Max_Pop[1] = Count_Maximal_Marking( Reseau, 
				      INPUT_TRANS, NORMAL_TRANS, Trans_Type );

  for ( m = 0; m <= Max_Pop[1]; m++ ) {

    Gen_Scilab_Matrix( Reseau,
		       Mat_Prefix, "B", m,
		       INPUT_TRANS, NORMAL_TRANS, Input_Nb, Trans_Nb,
		       Format, Integer,
		       Trans_Type, Trans_Number,
		       Phi );
  }

  /* Generate C matrices */
  Max_Pop[2] = Count_Maximal_Marking( Reseau, 
				      NORMAL_TRANS, OUTPUT_TRANS, Trans_Type );

  for ( m = 0; m <= Max_Pop[2]; m++ ) {
    
    Gen_Scilab_Matrix( Reseau,
		       Mat_Prefix, "C", m,
		       NORMAL_TRANS, OUTPUT_TRANS, Trans_Nb, Output_Nb,
		       Format, Integer,
		       Trans_Type, Trans_Number,
		       Phi );
  }

  /* Generate D matrices */
  Max_Pop[3] = Count_Maximal_Marking( Reseau, 
				      INPUT_TRANS, OUTPUT_TRANS, Trans_Type );

  for ( m = 0; m <= Max_Pop[3]; m++ ) {
      
    Gen_Scilab_Matrix( Reseau,
		       Mat_Prefix, "D", m,
		       INPUT_TRANS, OUTPUT_TRANS, Input_Nb, Output_Nb,
		       Format, Integer,
		       Trans_Type, Trans_Number,
		       Phi );
  }

  /* write the "system" variable" */
  printf("\n");
  if ( strlen( Mat_Prefix ) > 0 )
    Mat_Prefix = strncat( Mat_Prefix, "_", 128 );

  /* Write_System_Code( Mat_Prefix, Max_Pop ); */
  Write_Restricted_System_Code( Mat_Prefix, Max_Pop );

}

/************************************************************************/
/* The procedure Write_Maple_Matrices was used to generate A0 and A1 in */
/* Sven Hasenfuss' Maple format. Obsolete but kept in case of.		*/
/************************************************************************/
void Write_Maple_Matrices( void )
{	
  int		i;
  int		k;
  int		l;
  int		m;
  int		The_Size;
  boolean	Is_Immediate;
  int		Mu;
  Law_Desc	The_Law;
  Liste_Place	P_Scan;
  Place_Desc	The_Place;

  The_Size = Le_Reseau->Size;

  printf("#\n# Maple file for event graphs written by %s\n#\n", "eg2matrix" );

  /* Writing the A0 matrix */
  printf("A0 := proc( n )\n");
  printf("local AA;\n");
  printf("AA := matrix( %d, %d, [ eps $%d ] ):\n", 
	 The_Size, The_Size, The_Size * The_Size );

  for( l=0; l<Le_Reseau->Nb_Layers; l++)
    for( m=0; m<Le_Reseau->Level[l]->Nb_Nodes; m++) {

      i = Le_Reseau->Level[l]->Node_Id[m];

      SCAN( P_Scan, Le_Reseau->T[i]->Inputs ) {
	The_Place = P_Scan->Val;
	Mu = The_Place->Init_Pop;
	k = The_Place->Input_Id-1;

	if ( Mu == 0 ) {
	  The_Law = Le_Reseau->T[k]->Trans_Law;
	  Is_Immediate = ( ( The_Law->Name == 'C' ) 
			   && ( The_Law->Parameters[0] == 0.0 ) );

	  if ( !Is_Immediate )
	    printf("AA[%d,%d] := X_%d_.n;\n", i+1, k+1, k+1 );
	  else
	    printf("AA[%d,%d] := 0;\n", i+1, k+1 );
	}
      }
    }
  
  printf("op(AA);\n");
  printf("end;\n\n");

  /* Writing the A1 matrix */
  printf("A1 := proc( n )\n");
  printf("local AA;\n");
  printf("AA := matrix( %d, %d, [ eps $%d ] ):\n", 
	 The_Size, The_Size, The_Size * The_Size );

  for( l=0; l<Le_Reseau->Nb_Layers; l++)
    for( m=0; m<Le_Reseau->Level[l]->Nb_Nodes; m++) {

      i = Le_Reseau->Level[l]->Node_Id[m];

      SCAN( P_Scan, Le_Reseau->T[i]->Inputs ) {
	The_Place = P_Scan->Val;
	Mu = The_Place->Init_Pop;
	k = The_Place->Input_Id-1;

	if ( Mu == 1 ) {
	  The_Law = Le_Reseau->T[k]->Trans_Law;
	  Is_Immediate = ( ( The_Law->Name == 'C' ) 
			   && ( The_Law->Parameters[0] == 0.0 ) );

	  if ( !Is_Immediate )
	    printf("AA[%d,%d] := X_%d_.(n-1);\n", i+1, k+1, k+1 );
	  else
	    printf("AA[%d,%d] := 0;\n", i+1, k+1 );
	}
      }
    }
  
  printf("op(AA);\n");
  printf("end;\n\n");
}

