/* Translation program from eg language to QNAP language.	  */

#include <math.h>
#include "eg.h"

#define	TOOLNAME	"eg2qnap"

#define MAX(x,y)	( x>y ? x : y )

#define SCAN(x,y)       for(x=y;x!=NULL;x=x->Next)

extern EventGraph	*Read_Network(char*);
extern boolean		topological_analysis(EventGraph*);
extern void		Free_Event_Graph(EventGraph*);
extern char		*todays_date(void);

/* forward declaration */
void	dump_qnap_code(EventGraph*,int);

boolean		*Place_Exists;
boolean		Optimize = FALSE;
int		Opt_Level;

int main(int argc, char	*argv[])
{
  static char		File_Name[128] = "";
  static char		Code_File_Name[128] = "";
  EventGraph 		*Reseau;
  int			i;
  int			Sim_Length = 1000;
  boolean		Force = FALSE;
  boolean		Help = FALSE;
  boolean		Args_Ok = TRUE;
  boolean		File_Ok = TRUE;

  i = 1;

  if (argc>1)
    do {

      if GET_STRING( "-f", File_Name )
      else if GET_INT( "-nevt", Sim_Length )
      else if GET_STRING( "-o", Code_File_Name )
      else if SWITCH( "-force", Force, TRUE )

      else if SWITCH( "-debug", Debug, TRUE )

      else if GET_INT_SWITCH( "-O", Opt_Level, Optimize, TRUE )

      else if IS_VERSION

      else if ( ( 0 == strcmp( "-h", argv[i] ) )
		|| ( 0 == strcmp( "-?", argv[i] ) ) ) {
	printf("Usage: %s [-f <file>] ", argv[0]);
	printf("[-nevt <int>] [-O <level>] [-force] [-debug] [-o <file>]\n");
	Args_Ok = FALSE;
	Help = TRUE;
	i++;
      }
      else
	i++; /* skip unknown argument */
    }
    while ( Args_Ok && ( i < argc ) );

  if (Args_Ok) {

    Reseau = Read_Network(File_Name);

    if ( Reseau != NULL ) {

      if ( topological_analysis(Reseau) || Force ) {

	if ( strlen(Code_File_Name) != 0 )
	  File_Ok = ( NULL != freopen( Code_File_Name, "w", stdout ) );

	if ( File_Ok ) {
	  dump_qnap_code( Reseau, Sim_Length );

	  Free_Event_Graph( Reseau );

	  return 0;
	}
	else {
	  fprintf( stderr, "%s: Error opening file %s.\n", 
		   TOOLNAME, Code_File_Name );
	  return -1;
	}
      }
      else {
	fprintf( stderr, "Topological analysis failed: ");
	fprintf( stderr, "look for an empty cycle.\n");
	return -3;
      }
    }
    else {
      fprintf( stderr, "%s: Error reading network. No translation.\n", 
	       TOOLNAME );	
      return -2;
    }
  }
  else if (!Help)
    return -1;
  else
    return 0;
}

/* Procedure to write the law. Adapted from the same from gen_eg.c */
void fprint_QNAP_law(Law_Desc	The_Law,
		     int	Obj_Id )
{
  double	Alpha;
  double	Beta;
  double	Delta;
  double	Lambda;
  double	s1;
  double	s2;

  switch( The_Law->Name ) {
  case 'E':
    printf("EXP( %8.4f );\n", The_Law->Parameters[0] );
    break;
  case 'C':
    printf("CST( %8.4f );\n", The_Law->Parameters[0] );
    break;
  case 'B':
    printf("IF RANDU > %8.6f THEN CST(0.0) ELSE CST(1.0);\n", 
	   The_Law->Parameters[0] );
    break;
  case 'g':
    printf("CST( FIX( EXP( -1/LOG(%8.4f) ) ) );\n", 
	   The_Law->Parameters[0] );
    break;
  case 'G':
    printf("CST( 1.0 + FIX( EXP( -1/LOG(%8.4f) ) ) );\n", 
	   The_Law->Parameters[0] );
    break;
  case 'U':
    printf("UNIFORM( %8.4f, %8.4f );\n",
	   The_Law->Parameters[0],
	   The_Law->Parameters[1] );
    break;
  case 'R':
    printf("ERLANG( %8.4f, %4.0f );\n",
	   The_Law->Parameters[1],
	   The_Law->Parameters[0] );
    break;
  case 'O':
    Alpha = 1.0 / The_Law->Parameters[1];
    Beta = 1.0 / The_Law->Parameters[2];
    Lambda = (1.0+Alpha/Beta) / The_Law->Parameters[0];
    Delta = sqrt( Alpha*Alpha + 2.0*Alpha*(Beta+Lambda)
		  + (Beta-Lambda)*(Beta-Lambda) );
    s1 = -( Alpha+Beta+Lambda+Delta )/2.0;
    s2 = -( Alpha+Beta+Lambda-Delta )/2.0;
    printf("IF RANDU < %8.6f THEN EXP( %8.4f ) ELSE EXP( %8.4f );\n",
	   (-Lambda-s2)/(s1-s2), -1.0/s1, -1.0/s2 );
    break;
   case 'H':
    printf("HEXP( %8.4f, %8.4f );\n",
	   The_Law->Parameters[0],
	   The_Law->Parameters[1] );
    break;
  case 'D':
    fprintf( stderr, "%s: error: can't handle timing law (%c)\n",
	     TOOLNAME, The_Law->Name );
    break;
  case 'P':
    /* this part assumes implicitly that the object is a transition	*/
    printf("BEGIN\n");
    printf("        TR_%d.NextArr := EXP( %8.4f );\n",
	   Obj_Id, The_Law->Parameters[0] );
    printf("        WHILE ( TR_%d.NextArr > TR_%d.NextChg )\n",
	   Obj_Id, Obj_Id );
    printf("        DO BEGIN\n");
    printf("            TR_%d.NextArr := TR_%d.NextArr - TR_%d.NextChg;\n",
	   Obj_Id, Obj_Id, Obj_Id );
    printf("            CST( TR_%d.NextChg );\n", Obj_Id );
    printf("            ");
    fprint_QNAP_law( The_Law->Law_Parameters[1], Obj_Id );
    printf("            TR_%d.NextChg := ", Obj_Id );
    fprint_QNAP_law( The_Law->Law_Parameters[0], Obj_Id );
    printf("        END;\n" );
    printf("        CST( TR_%d.NextArr );\n", Obj_Id );
    printf("        TR_%d.NextChg := TR_%d.NextChg - TR_%d.NextArr;\n", 
	   Obj_Id, Obj_Id, Obj_Id );
    printf("    END;\n" );
    break;
  default:
    fprintf( stderr, "%s: error: unknown timing law (%c)\n",
	     TOOLNAME, The_Law->Name );
  }
}

/* procedure to write the initial population. Improved from gen_eg.c */
/* Initial lags are not taken care of... this is not trivial in QNAP.*/
/* Best thing would be to have them generated by a private source.    */
void fprint_init_mark(
Place_Desc	The_Place,
int		The_Class )
{
  if (The_Place->Init_Pop>0)
    printf("    INIT(CL_%d) = %d;\n", The_Class, The_Place->Init_Pop );

}

/* procedure to write what is the destination of customers exiting	*/
/* "transition" The_Orig to place The_Place which is between The_Orig	*/
/* and The_Dest.							*/
void fprint_dest(int	The_Orig, 
		 int	The_Dest,
		 int	The_Place)
{	
  if (Optimize && !Place_Exists[The_Place])
    printf("        TR_%d,", The_Dest );
  else
    printf("        PL_%d,", The_Place );

}

/****************************************************************/
/* Function 	int	In_Link_Number				*/
/* Computes the number of a P->T link, between 0 and the 	*/
/* indegree of T.		 				*/
/*								*/
/* Parameters:							*/
/*	Reseau:		the network where it is computed	*/
/*	T:		the trans id (betw 0 and Nb_Trans-1)	*/
/*	P:		the place id of the link 		*/
/*			(betw 0 and Nb_Places-1)		*/
/*								*/
/* Return Value:						*/
/*	the number of the link					*/
/****************************************************************/
int	In_Link_Number( EventGraph	*Reseau,
			int		T_Id,
			int 		P_Id )
{
  int		Num;
  Liste_Place	P_Scan;

  Num = 0;
  for ( P_Scan=Reseau->T[T_Id]->Inputs;
	(P_Scan != NULL) && (P_Scan->Val->Place_Id != P_Id);
	P_Scan = P_Scan->Next ) {
    Num++;
  }
 
  if ( Num >= Reseau->T[T_Id]->Indegree ) {
    fprintf( stderr, 
	     "%s: warning: inconsistency in pred structure:",
	     TOOLNAME );
    fprintf( stderr, "no place number %d output of Tr %d\n", P_Id, T_Id );
  }
  
  return Num;
}

void	init_place_status(EventGraph	*Reseau)
{
  int		i;
  Place_Desc	The_Place;
  Liste_Place	P_Scan;

  if ( Optimize ) {
    Place_Exists = (boolean*) calloc( Reseau->Nb_Places, sizeof(boolean) );
    for ( i=0; i<Reseau->Size; i++ ) {
      SCAN( P_Scan, Reseau->T[i]->Outputs ) {
	The_Place = P_Scan->Val;
	Place_Exists[The_Place->Place_Id] =
	  ! ( ( The_Place->Place_Law->Name == 'C' )
	      && ( The_Place->Place_Law->Parameters[0] == 0.0 ) );
	if (Debug && !Place_Exists[The_Place->Place_Id] )
	  fprintf(stderr,"Pl %d=(%d,%d) removed\n", 
		  The_Place->Place_Id, i+1, The_Place->Trans_Id );
      }
    }
  }
}

/************************************************************************/
/* Write declaration of special queue object necessary for interrupted	*/
/* Poisson process.							*/
/************************************************************************/
void Write_Special_Declarations(EventGraph        *Reseau)
{
  boolean	Has_IPP = FALSE;
  int		i;

  for( i=0; !Has_IPP && (i<Reseau->Size); i++ ) {
    Has_IPP = ( Reseau->T[i]->Trans_Law->Name == 'P' );
  }

  if ( Has_IPP ) {
    printf("    QUEUE OBJECT OnOffSrc;\n");
    printf("      REAL NextArr;\n");
    printf("      REAL NextChg;\n");
    printf("    END;\n");
  }
}

void dump_qnap_code(EventGraph  *Reseau,
		    int		Sim_Length )
{
  int		i;
  int		j;
  Trans_Desc	The_Trans;
  Place_Desc	The_Place;
  int		Nb_Places;
  int		Nb_Neighb;
  int		The_Dest;
  int		Max_Deg;
  Liste_Place	P_Scan;
  boolean	Some_Places = TRUE;
  int		Last_Place;
  int		Nb_Real_Places;

  init_place_status( Reseau );

  printf("& QNAP2 code generated by %s\n", TOOLNAME );
  printf("& Date: %s\n", todays_date() );
  /* declarations */
  printf("/DECLARE/\n");

  Write_Special_Declarations( Reseau );

  /* declaration of normal queues */
  printf("&   transitions\n");
  printf("    QUEUE");
  for (i=0; i<Reseau->Size-1; i++) {
    if ( Reseau->T[i]->Trans_Law->Name != 'P' ) {
      printf(" TR_%d",i);
      if ( ( i != Reseau->Size-2 ) 
	   || ( Reseau->T[Reseau->Size-1]->Trans_Law->Name != 'P' ) )
	printf(",");
      if ((i+1) % 10 == 0)
	printf("\n         ");
    }
  }
  if ( Reseau->T[Reseau->Size-1]->Trans_Law->Name != 'P' )
    printf(" TR_%d;\n", Reseau->Size-1 );
  else
    printf(";\n");

  /* declaration of IPP sources */
  for (i=0; i<Reseau->Size; i++) {
    if ( Reseau->T[i]->Trans_Law->Name == 'P' ) {
      printf("    OnOffSrc TR_%d;\n",i);
    }
  }

  /* for places, check if some exist... */
  if ( Optimize ) {
    Some_Places = FALSE;
    for (i=0; i<Reseau->Nb_Places; i++)
      {	Some_Places = ( Some_Places || Place_Exists[i] );
      if ( Place_Exists[i] )
	Last_Place = i;
      }
  }
  else	
    Last_Place = Reseau->Nb_Places - 1;
  if ( Some_Places ) {
    printf("&   places\n");
    printf("    QUEUE");
    Nb_Real_Places = 0;
    for (i=0; i<Last_Place; i++) {
      if ( (!Optimize) || Place_Exists[i] ) {
	printf(" PL_%d,",i);
	Nb_Real_Places++;
      }
      if ((Nb_Real_Places+1) % 10 == 0)
	printf("\n         ");
    }
    printf(" PL_%d;\n", Last_Place );
  }

  printf("&   classes\n");
  /* computation of the maximum indegree */
  Max_Deg = Reseau->T[0]->Indegree;
  for (i=1; i<Reseau->Size; i++)
    Max_Deg = MAX( Max_Deg, Reseau->T[i]->Indegree );

  printf("    CLASS");
  for (i=0; i<Max_Deg-1; i++) {
    printf(" CL_%d,",i);
    if ((i+1) % 10 == 0)
      printf("\n         ");
  }
  printf(" CL_%d;\n", Max_Deg-1 );
  printf("\n");

  /* other variables */
  printf("&   other variables\n");
  printf("    REAL Tps_Cpu;\n");
  printf("    INTEGER Nb_Evts;\n");
  printf("\n");

  /* bloc de controle */
  printf("/CONTROL/\n");
  printf("    OPTION = NSOURCE;\n");
  printf("    TMAX = %d;\n", 1000*Sim_Length );
  printf("\n");

  /* generate transitions */
  for ( i=0; i<Reseau->Size; i++ ) {
    printf("/STATION/\n");
    printf("    NAME = TR_%d;\n", i);
    The_Trans = Reseau->T[i];

    switch (The_Trans->Indegree) {
    case 0:	/* a source that is not recycled */
      printf("    TYPE = SOURCE;\n");
      break;
    case 1: 	/* only one input. Nothing to be done */
      break;
    default:	/* have to make a FUSION */
      printf("    FUSION = (\n");
      P_Scan = The_Trans->Inputs;
      for( j=0; j<The_Trans->Indegree-1; j++ ) {
	The_Place = P_Scan->Val;
	printf("        CL_%d, 1,\n", j );
/* 	printf("        CL_%d, 1,\n",  */
/* 	       In_Link_Number( Reseau,  */
/* 			       The_Place->Trans_Id,  */
/* 			       The_Place->Place_Id ) ); */
	P_Scan = P_Scan->Next;
      }
      The_Place = P_Scan->Val;
      printf("        CL_%d, 1 ) CL_0;\n", j );
/*       printf("        CL_%d, 1,\n",  */
/* 	     In_Link_Number( Reseau,  */
/* 			     The_Place->Trans_Id,  */
/* 			     The_Place->Place_Id ) ); */
    }

    /* transition number 0 rules for the number of events */
    printf("    SERVICE = ");
    if ( i == 0 ) {
      printf("BEGIN\n");
      printf("      IF SERVNB( TR_0 ) = %d THEN STOP;\n", Sim_Length );
      printf("      ");
      fprint_QNAP_law( The_Trans->Trans_Law, The_Trans->Trans_Id-1 );
      printf("    END;\n");
    }
    else
      fprint_QNAP_law( The_Trans->Trans_Law, The_Trans->Trans_Id-1 );

    Nb_Neighb = The_Trans->Outdegree;
    switch (Nb_Neighb) {
    case 0:
      printf("    TRANSIT = OUT;\n");
      break;
    case 1:	
      The_Place = The_Trans->Outputs->Val;
      The_Dest = The_Place->Trans_Id-1;
      printf("    TRANSIT = ");
      fprint_dest( i, The_Dest,  The_Place->Place_Id );
      printf(" CL_%d;\n", In_Link_Number( Reseau, The_Dest,
					  The_Place->Place_Id ) );
      break;
    default:
      P_Scan = The_Trans->Outputs;
      printf("    FISSION = (\n");
      for( j=0; j<Nb_Neighb-1; j++ ) {
	The_Place = P_Scan->Val;
	The_Dest = The_Place->Trans_Id-1;
	fprint_dest( i, The_Dest, The_Place->Place_Id );
	printf(" CL_%d, 1,\n", In_Link_Number( Reseau, The_Dest, 
					       The_Place->Place_Id ) );
	P_Scan = P_Scan->Next;
      }
      The_Place = P_Scan->Val;
      The_Dest = The_Place->Trans_Id-1;
      fprint_dest( i, The_Dest, The_Place->Place_Id );
      printf(" CL_%d, 1 );\n", In_Link_Number( Reseau, The_Dest,
					       The_Place->Place_Id ) );
    }
    printf("\n");
  }

  /* now generate places */
  Nb_Places = 0;
  for ( i=0; i<Reseau->Size; i++ ) {
    SCAN( P_Scan, Reseau->T[i]->Outputs ) {
      Nb_Places++;
      The_Place = P_Scan->Val;

      if ( Optimize && !Place_Exists[The_Place->Place_Id] ) {
	/* if initial population, put it in downstram transition.	*/
	if (The_Place->Init_Pop != 0) {
	  printf("/STATION/\n");
	  printf("    NAME = TR_%d;\n", j );
	  fprint_init_mark( The_Place, In_Link_Number( Reseau, 
						       The_Place->Trans_Id-1,
						       The_Place->Place_Id ) );
	  printf("\n");
	}
      }
      else {

	printf("/STATION/\n");

	printf("    NAME = PL_%d;\n", The_Place->Place_Id );
	printf("    TYPE = SERVER, INFINITE;\n");
	printf("    SERVICE = ");
	/* normally, the second argument of fprint_QNAP_law 	*/
	/* is useless here.					*/
	fprint_QNAP_law( The_Place->Place_Law, 0 );
	fprint_init_mark( The_Place, In_Link_Number( Reseau,
						     The_Place->Trans_Id-1,
						     The_Place->Place_Id ) );
	printf("    TRANSIT = TR_%d;\n", The_Place->Trans_Id-1 );
	printf("\n");

      }
    }
  }

  /* exec bloc */
  printf("/EXEC/\n");
  printf("    BEGIN\n");
  printf("        Tps_Cpu := GETCPUT;\n");
  printf("        Nb_Evts := 0;\n");
  printf("        SIMUL;\n");
  for (i=0; i<Reseau->Size; i++) {

    printf("        WRITELN(\"Cycle time of transition %d:\",", 
	   i);
    printf("TIME/SERVNB(TR_%d) );\n", i );
    printf("        Nb_Evts := Nb_Evts + SERVNB(TR_%d);\n", i );
  }
  printf("        Tps_Cpu := GETCPUT;\n");
  printf("        WRITE(\"Total Execution Time: \", Tps_Cpu );\n" );
  printf("        WRITELN(\"      (\", Nb_Evts / Tps_Cpu, \"evt/sec)\");\n" );
  printf("    END;\n");
  printf("/END/\n");

  exit(0);
}
