/* Translation program from eg language to tg language.	  */

#include "eg.h"

#define	TOOLNAME	"eg2tg"

#define SCAN(x,y)	for(x=y;x!=NULL;x=x->Next)

extern EventGraph	*Read_Network(char*);
extern void		dumpe_network(EventGraph*);
extern void		Free_Event_Graph(EventGraph*);
extern char		*todays_date(void);

/* forward declaration */
void	dump_tgl_code(EventGraph *Reseau);

int main(int argc, char *argv[])
{
  static char		File_Name[128] = "";
  static char		Code_File_Name[128] = "";
  EventGraph 		*Reseau;
  int			i;
  boolean		Help = FALSE;
  boolean		Args_Ok = TRUE;
  boolean		File_Ok = TRUE;

  i = 1;

  if (argc>1)
    do {

      if GET_STRING( "-f", File_Name )
      else if GET_STRING( "-o", Code_File_Name )

      else if SWITCH( "-debug", Debug, TRUE )

      else if IS_VERSION

      else if ( ( 0 == strcmp( "-h", argv[i] ) )
		|| ( 0 == strcmp( "-?", argv[i] ) ) ) {
	printf("Usage: %s [-f <file>] ", TOOLNAME );
	printf("[-debug] [-o <file>]\n");
	Args_Ok = FALSE;
	Help = TRUE;
	i++;
      }
      else
	i++; /* skip unknown argument */
    }
    while ( Args_Ok && ( i < argc ) );

  if (Args_Ok) {
    
    Reseau = Read_Network(File_Name);

    if ( Reseau != NULL ) {

      if ( strlen(Code_File_Name) != 0 )
	File_Ok = ( NULL != freopen( Code_File_Name, "w", stdout ) );

      if ( File_Ok ) {
	dump_tgl_code( Reseau );
	Free_Event_Graph( Reseau );

	return 0;
      }
      else {
	fprintf( stderr, "%s: Error opening file %s.\n", 
		 TOOLNAME, Code_File_Name );
	return -1;
      }
    }
    else {
      fprintf( stderr, "%s: Error reading network. No translation.\n", 
	       TOOLNAME );	
      return -2;
    }
  }
  else if (!Help)
    return -1;
  else
    return 0;
}

/* Procedure to write the law. Adapted from the same from gen_eg.c */
void fprint_tg_law(Law_Desc	The_Law)
{
  char	*The_Name;

  The_Name = (char*) calloc( 4, sizeof(char) );

  switch( The_Law->Name ) {
  case 'B':	strcpy( The_Name, "BER" );
    break;
  case 'C':	strcpy( The_Name, "CST" );
    break;
  case 'E':	strcpy( The_Name, "EXP" );
    break;
  case 'g':	strcpy( The_Name, "GEOM" );
    break;
  case 'G':	strcpy( The_Name, "GEOSHIFT" );
    break;
  case 'L':	strcpy( The_Name, "LOGNORMAL" );
    break;
  case 'p':	strcpy( The_Name, "PARETO" );
    break;
  case 'U':	strcpy( The_Name, "UNIF" );
    break;
  case 'W':	strcpy( The_Name, "WEIBULL" );
    break;
  case 'R':	strcpy( The_Name, "ERL" );
    break;
  default:	fprintf( stderr, "%s: unknown law: '%c'\n",
			 TOOLNAME, The_Law->Name );
  }

  switch( The_Law->Name ) {
  case 'B':
  case 'C':
  case 'E':
  case 'G':
  case 'g':
    printf("%s ( %8.4f )", 
	   The_Name, The_Law->Parameters[0] );
    break;
  case 'L':
  case 'p':
  case 'U':
  case 'W':
    printf("%s ( %8.4f, %8.4f )", The_Name,
	   The_Law->Parameters[0],
	   The_Law->Parameters[1] );
    break;
  case 'R':
    printf("%s ( %4.0f, %8.4f )", The_Name,
	   The_Law->Parameters[0],
	   The_Law->Parameters[1] );
    break;
  }
}

void dump_tgl_code(EventGraph *Reseau)
{
  int		i;
  int		j;
  int		The_Number;
  Trans_Desc	The_Trans;
  Place_Desc	The_Place;
  Liste_Place	P_Scan;
  char		*The_Date;

  printf("# tgl code generated by %s\n", TOOLNAME );
  The_Date = todays_date();
  printf("# Date: %s\n", The_Date );
  free( The_Date );

  /* generate transitions */
  for ( i=0; i<Reseau->Size; i++ ) {

    printf("# transition number %d\n", i );
    printf("TASK Trans_%d", i+1 );
    The_Trans = Reseau->T[i];
    if (NULL != The_Trans->Name)
      printf(" %s", The_Trans->Name);
    printf("\n");
    printf("  DURATION ");
    fprint_tg_law( The_Trans->Trans_Law );
    printf("\n");
    printf("END\n\n");

    /* generate output places of the current transition */
    SCAN( P_Scan, Reseau->T[i]->Outputs ) {
      The_Place = P_Scan->Val;
      The_Number = The_Place->Trans_Id-1;
      printf("LINK Link_%d_%d\n", i+1, The_Number+1 );
      printf("  TYPE ");
      if ( The_Place->Init_Pop == 0 )
	printf("DATA\n");
      else	
	printf("PROGRAM\n");
      printf("  IS NECESSARY\n");
      printf("  FROM Trans_%d\n", i+1 );
      printf("  TO Trans_%d\n", The_Number+1 );
      if ( The_Place->Init_Pop == 0 ) {
	printf("  DURATION ");
	fprint_tg_law( The_Place->Place_Law );
	printf("\n");
      }
      else
	printf("  DEPTH %d\n", The_Place->Init_Pop );

      /* informing about the loss timing */
      if ( ( The_Place->Place_Law->Name != 'C' )
	   || ( The_Place->Place_Law->Parameters[0] != 0.0 ) ) {
	fprintf( stderr, "%s: warning: non-zero place timing at place",
		 TOOLNAME );
	fprintf( stderr, " (%d,%d). Ignored.\n", i, j );
      }

      printf("END\n\n");
    }
  }
}
