/* A program for the simulation of Event Graphs				  */

#include <stdio.h>
#include <sys/time.h>
#include <time.h>
#include "eg.h"

#define TOOLNAME	"eg_sim"
#define max(x,y)        ( (x>y) ? x : y )

extern EventGraph	*Read_Network(char*);
extern void		dumpe_network(EventGraph *);
extern boolean		topological_analysis(EventGraph *);
extern boolean		Compile_EG(EventGraph *Reseau, int Sim_Length,
				   boolean Checking,
				   boolean  Keep, boolean Waiting_Time,
				   int From_Trans, int To_Trans );
extern double		**simulate(EventGraph*,int,int);
extern void		dumpe_simulation(EventGraph*,double**,int,int);
extern void		write_stat_results(EventGraph*,double**,int,int,
					   struct timeval,struct timeval);

int main(int argc, char *argv[])
{
  static char		File_Name[128] = "";
  EventGraph 		*Reseau;
  int			i;
  int			Sim_Length = 1000;
  int			Buf_Size = 1024;
  boolean		Help = FALSE;
  boolean		Compile = FALSE;
  boolean		Keep = FALSE;
  boolean		Checking = FALSE;
  int			From_Trans = -1;
  int			To_Trans = -1;
  boolean		Waiting_Time = FALSE;
  boolean		Args_Ok = TRUE;
  double		**X;	/* Contains the trajectory after simulation */
  struct timeval	t_debut;
  struct timeval	t_fin;

  i = 1;

  if (argc>1)
    do {
      if GET_STRING( "-f", File_Name )
      else if GET_INT( "-bs", Buf_Size )
      else if SWITCH( "-debug", Debug, TRUE )
      else if SWITCH( "-c", Checking, TRUE )
      else if GET_INT_SWITCH( "-from", From_Trans, Waiting_Time, TRUE )
      else if GET_INT_SWITCH( "-to", To_Trans, Waiting_Time, TRUE  )
      else if SWITCH( "-compile", Compile, TRUE )
      else if SWITCH( "-keep", Keep, TRUE )
      else if GET_INT( "-nevt", Sim_Length )
      else if IS_VERSION

      else if ( ( 0 == strcmp( "-h", argv[i] ) )
		|| ( 0 == strcmp( "-?", argv[i] ) ) ) {

	printf("Usage: %s [-f <file>] [-nevt <int>] [-bs <int>]",
	       TOOLNAME );
	printf(" [-compile] [-keep]\n\t[-c] [-from <int>] [-to <int>]" );
	printf(" [-debug]\n");
	Args_Ok = FALSE;
	Help = TRUE;
	i++;
      }
      else	
	i++; /* skip unknown argument */
    }
    while ( Args_Ok && ( i < argc ) );
    
  if (Args_Ok) {

    if ( Checking ) {
      Compile_EG( NULL, 0, Checking, FALSE, FALSE, 0, 0 );
      return 0;
    }
    else {

      Reseau = Read_Network(File_Name);

      if ( NULL != Reseau ) {

	if ( topological_analysis(Reseau) ) {

	  if (Debug)
	    dumpe_network(Reseau);

	  Buf_Size = max( Buf_Size, 1 + Reseau->Max_Pop_Init );

	  if ( Waiting_Time ) {
	    Waiting_Time = ( ( From_Trans > 0 ) 
			     && ( From_Trans <= Reseau->Size ) );
	    if ( !Waiting_Time ) {
	      printf( "%s: error: incorrect parameter for -from (%d)\n",
		      TOOLNAME, From_Trans );
	    }
	    else {
	      Waiting_Time = ( ( To_Trans > 0 ) 
			       && ( To_Trans <= Reseau->Size ) );
	      if ( !Waiting_Time ) {
		printf( "%s: error: incorrect parameter for -to (%d)\n",
			TOOLNAME, To_Trans );
	      }
	    }
	    if (!Waiting_Time) {
	      return -4;
	    }
	    else {
	      if ( !Compile ) {
		printf( "%s: warning: waiting time analysis available for ", 
			TOOLNAME );
		printf("compiled simulation only (assumed).\n" );
		Compile = TRUE;
	      }
	    }
	  }

	  if ( !Compile ) {
	    gettimeofday(&t_debut,(struct timezone *)0);
	    X = simulate(Reseau, Sim_Length, Buf_Size);
	    gettimeofday(&t_fin,(struct timezone *)0);

	    if (Debug)
	      dumpe_simulation(Reseau, X, Sim_Length, Buf_Size);

	    write_stat_results(Reseau, X, Sim_Length, Buf_Size, 
			       t_debut, t_fin);
	  }
	  else {
	    Compile_EG( Reseau, Sim_Length, Checking, Keep,
			Waiting_Time, From_Trans, To_Trans );
	  }

	  	
	  return 0;
	}
	else {
	  /* Topo. analysis failed. Network not live but this has been	*/
	  /* reported already.						*/
	  return -3;
	}
      }
      else {
	fprintf( stderr, "%s: Error reading network from file %s.\n", 
		 TOOLNAME, File_Name );
	return -2;
      }
    }
  }
  else if ( !Help )
    return -1;
  else 
    return 0;
}

