/* Un programme d'analyse de la vivacite des graphes d'evenements.	*/

#include <stdio.h>
#include "eg.h"

#define	TOOLNAME	"live_eg"

/* Definition of sparse matrix structures. See analyze_eg.c */
#ifdef USE_SP_MAT
typedef struct SP_BOOL_MAT {
  int Size;
  int *Nb_Elts;		/* For esch i, nb of "outputs" */
  int *Elts;		/* For each i, table of the j's connected. */
} *SP_Bool_Mat;
#define BOOL_MAT	SP_Bool_Mat
#else
#define BOOL_MAT	boolean*
#endif

/* external declarations*/
extern EventGraph	*Read_Network(char*);
extern boolean		topological_analysis(EventGraph*);
extern BOOL_MAT		Make_Adjacency_Matrix(EventGraph*,boolean);
extern int 		*Find_Cycle(BOOL_MAT,int,boolean);

void dump_cycle_as_gss( int *Cycle)
{
  int	i;

  printf( "# Cycle generated by %s\n", TOOLNAME );
  printf( "\n%d\n_E_O_E\n", Cycle[0] );

  for ( i=1; i <= Cycle[0] ; i++ ) {
    printf( "\n%d\n_E_O_E\n", Cycle[i]+1 );
  }
}

int main(int argc, char *argv[])
{
  static char	In_File_Name[128] = "";
  static char	Out_File_Name[128] = "";
  EventGraph 	*Reseau;
  int		i;
  int		*Cycle;
  boolean	Dump_Cycle = FALSE;
  boolean	Help = FALSE;
  boolean	Args_Ok = TRUE;
  boolean	File_Ok = TRUE;
  BOOL_MAT	Adj;

  i = 1;

  if (argc>1)
    do {

      if GET_STRING( "-f", In_File_Name )
      else if GET_STRING( "-o", Out_File_Name )

      else if SWITCH( "-debug", Debug, TRUE )
      else if SWITCH( "-d", Dump_Cycle, TRUE )
 
      else if IS_VERSION

      else if ( ( 0 == strcmp( "-?", argv[i] ) )
		|| ( 0 == strcmp( "-h", argv[i] ) ) ) {
	printf("Usage: %s [-d] [-f <file>] [-o <file>] [-h] [-?] \n", 
	       TOOLNAME);
	Args_Ok = FALSE;
	Help = TRUE;
	i++;
	}
      else	
	i++; /* skip unknown argument */
    }
    while ( Args_Ok && ( i < argc ) );

  if (Args_Ok) {

    Reseau = Read_Network(In_File_Name);

    if ( Reseau != NULL ) {

      if ( !topological_analysis(Reseau) ) {
	/* if returns TRUE, means that the  EG is live */

	if ( Dump_Cycle ) {

	  if ( strlen(Out_File_Name) != 0 )
	    File_Ok = ( NULL != freopen( Out_File_Name, "w", stdout ) );

	  if ( File_Ok ) {

	    Adj = Make_Adjacency_Matrix( Reseau,
					 /* zero weight only? */ FALSE );
	    Cycle = Find_Cycle( Adj, Reseau->Size, FALSE);

	    dump_cycle_as_gss( Cycle );
	  }
	  else
	    fprintf( stderr, "%s: Error opening file %s.\n", 
		     TOOLNAME, Out_File_Name );

	}
      }
      return 0;
    }
    else {
      fprintf( stderr, "%s: Error reading network. No translation.\n", 
	       TOOLNAME );	
      return -2;
    }
  }
  else if (!Help)
    return -1;
  else
    return 0;
}
