/************************************************************************/
/* This program generates the inputs for the CM simulator               */
/* of free choice nets with constant firing times                       */
/* from the pnl description of the Petri net                            */
/************************************************************************/
/* Written by Bruno Gaujal under the name "matrice2.c"			*/
/************************************************************************/

#include <stdio.h>
#include "pn.h" 		/* Petri Net data structures */

#define TOOLNAME 	"fc2totmatrix"

#define INFTY 0x7FFFFFFF

#define min(a,b) ((a)<(b)?(a):(b))



int **A,**B,**Q,**P;
int *M;
int *Part;
int *rho;
int *T1,*T2;
int *Bullet;
int *Nom;
int K,NT1,NT2;

PetriNet *Petri;




/******** Fonctions qui rendent  un pointeur sur une structure PetriNet *****/

extern PetriNet *Read_Network( char* File_Name );

/******** Function for allocating a square integer matrix *************/
int **Alloc_Matrix( int Size ) {

  int **Res;
  int i;
  int Alloc_Ok = 1;	/* True */

  Res = (int**) calloc( (size_t)Size, sizeof(int*) );
  Alloc_Ok = ( (int**)NULL != Res );

  for( i=0; (i<Size) && (Alloc_Ok); i++ ) {
    Res[i] = (int*) calloc( (size_t)Size, sizeof(int) );
    Alloc_Ok = ( (int*)NULL != Res[i] );
  }

  if ( !Alloc_Ok ) {
    /* there is no garbage collecting... this would not help much anyway! */
    Res = (int**)NULL;
  }

  return Res;
}


/********Conformite du r\'eseau ********/

int conform()
{
  int res, i;
  int pl_name;

  res = 1;

  /** Conformite topologique **/

  for (i = 0; i< Petri->Nb_Trans ; i++) {

    if (Petri->T[i].Indegree == 0) {
      fprintf(stderr,"The network is not in canonical form (transition %d has no input place)\n",i);
      res = 0;
    }
    else {
      pl_name = Petri->T[i].Inputs->Dest;
      if (Petri->P[pl_name].Indegree + Petri->P[pl_name].Outdegree > 2) {

	if (Petri->T[i].Indegree != 1) {
	  fprintf(stderr,"The network is not in canonical form (transition %d has several input routing places\n",i);
	  res = 0;
	}
      }
    }
  }

  return res;
}



/********* partition des transitions  (1 si dans T2 (ie MGC), 0 sinon) ******/

void partition()
{
  int i;
  int res;

  NT1 = NT2 = 0;

  for (i = 0; i< Petri->Nb_Trans ; i++) {

    res = 1;
    if (Petri->T[i].Indegree == 1) {

      res = 0;
    }
    Part[i] = res;
    if (res ==0) {T1[i] = NT1; NT1++;} else {T2[i] = NT2; NT2 ++;}
  }

}



/********* Marquage initial *********/

void Initial_Marking()
{
  int i;
  int pl_name;

  for (i = 0 ; i< Petri->Nb_Trans;i++) {

    M[i] = 0;
    if (!(Part[i])) {

      pl_name = Petri->T[i].Inputs->Dest;
      M[i] = Petri->P[pl_name].Init_Pop;
    }
  }
}



/***** generation des matrices A B Q P (l)  et de Bullet et de name  ********/

void generate_mat()
{
  Liste_Link 	liste_pla;
  Liste_Link 	liste_trans;
  Pl_Des 	pla;
  Tr_Des 	trans;
  int 		pla_name;
  int 		tr_name;


  int i,j;
  for (i = 0; i< NT2;i++)
    {
      for (j = 0; j< NT2;j++)
	{
	
	  A[i][j] = INFTY;
	}
    }

  for (i = 0; i< NT2;i++)
    {
      for (j = 0; j< NT1;j++)
	{
	  B[i][j] = INFTY;

	}
    }

  for (i = 0; i< NT1;i++)
    {
      for (j = 0; j< NT1;j++)
	{

	  P[i][j] = 0;
	}
    }

  for (i = 0; i< NT1;i++)
    {
      for (j = 0; j< NT2;j++)
	{
	  Q[i][j] = 0;

	}
    }

  for (i = 0; i< Petri->Nb_Trans;i++) {

    liste_pla = Petri->T[i].Inputs;
    while (liste_pla != NULL)
      {
	pla_name = liste_pla->Dest;
	pla = Petri->P[pla_name];
	liste_trans = pla.Inputs;
	while (liste_trans != NULL)
	  {
	    tr_name = liste_trans->Dest;
	    trans = Petri->T[tr_name];

	    if (Part[i] && Part[trans.Tr_Id]) 
	      {A[T2[i]][T2[trans.Tr_Id]] = 
		 min(A[T2[i]][T2[trans.Tr_Id]],pla.Init_Pop);
	      }
	      
	    else if(Part[i] && !Part[trans.Tr_Id])
	      {B[T2[i]][T1[trans.Tr_Id]] = 
		 min(B[T2[i]][T1[trans.Tr_Id]],pla.Init_Pop);
	      }
		            
	    else if (!Part[i] && !Part[trans.Tr_Id])
	      {P[T1[i]][T1[trans.Tr_Id]] += 1;}
	      
	    else 	
	      {Q[T1[i]][T2[trans.Tr_Id]] += 1;}

	    liste_trans = liste_trans->Next;
	  }
	liste_pla = liste_pla->Next;
      }
  }
}

/*************************/
void Calcul_bullet()
{
  int i;

  Liste_Link liste_pla;
  Pl_Des pla;
  int pla_name;


  for (i = 0; i< Petri->Nb_Trans;i++) {

    if (Part[i] == 0) {

      liste_pla = Petri->T[i].Inputs;
      pla_name = liste_pla->Dest;
      pla = Petri->P[pla_name];
      Bullet[T1[i]] = pla.Pl_Id;
    }
  }
}


/*********** programme principal *********/

int main(int argc, char *argv[] )
{

  static char		In_File_Name[128] = "";
  static char		Out_File_Name[128] = "";

  int i;
  int j;

  int Args_OK = TRUE;
  int Help    = FALSE;
  int InOut   = TRUE;

  /** Parsing **/

  i = 1;

  if (argc>1)
    do {

      if ( 0 == strcmp( "-f", argv[i] ) )
	{	
	  i++;
	  if ( i < argc )
	    {	
	      strcpy( In_File_Name, argv[i] );
	      i++;
	    }
	  else 
	    {	
	      fprintf(stderr,"%s: usage:filename expected after -f\n",argv[0]);
	      Args_OK = FALSE;
	    }
	}
	
      else if (( 0 == strcmp( "-h", argv[i] )) || 
	       ( 0 == strcmp( "-?", argv[i] )) ||
	       ( 0 == strcmp( "-help", argv[i] ))) {

	i++;
	fprintf(stdout,"usage: %s [-f inputfile] [-o outputfile] \nIf -o is used the ouputfile is reset so old version(if any) is lost.\nThe inputfile must be a description of a petri net in pnl.\n",argv[0]);
	Args_OK = FALSE;
	Help = TRUE;
      }

      else if ( 0 == strcmp( "-o", argv[i] ) ) {
	i++;
	if ( i < argc) {
	  strcpy( Out_File_Name, argv[i] );
	  i++;
	}
	else {
	  fprintf(stderr,"usage: %s: filename expected after -o\n",argv[0]);
	  Args_OK = FALSE;
	}
      }
      else if ( 0 == strcmp( "-v", argv[i] ) ) {
	printf("%s (Version %s)\n", TOOLNAME, VERSION);
	Args_OK = FALSE;
	Help = TRUE;
	i++;
      }
      else 
	i++;
    } while( Args_OK && ( i < argc ) );

  if ( Args_OK ) {

    /* Check if output file is ok */
    if  ( strlen(Out_File_Name) != 0 ) {
      if ( NULL == freopen( Out_File_Name, "w", stdout )) {
	fprintf(stderr,"cannot open file %s\n",Out_File_Name);
	InOut = FALSE;
      }
    }

    /** Call to Read_Network **/

    Petri = Read_Network( In_File_Name );

    /** start of the program if everything is OK **/

    /** test de la conformite du reseau **/

    if (conform()) {

      /** allocation des matrices et vecteurs  **/

      A = Alloc_Matrix( Petri->Nb_Trans );
      B = Alloc_Matrix( Petri->Nb_Trans );
      P = Alloc_Matrix( Petri->Nb_Trans );
      Q = Alloc_Matrix( Petri->Nb_Trans );

      M      = (int*) calloc(Petri->Nb_Trans,sizeof( int ));
      Part   = (int*) calloc(Petri->Nb_Trans,sizeof( int ));
      T1     = (int*) calloc(Petri->Nb_Trans,sizeof( int ));
      T2     = (int*) calloc(Petri->Nb_Trans,sizeof( int ));
      Bullet = (int*) calloc(Petri->Nb_Trans,sizeof( int ));

      if ( ( NULL == A ) || ( NULL == B ) || ( NULL == P ) || ( NULL == Q )
	   || ( NULL == M ) || ( NULL == Part ) || ( NULL == T1 )
	   || ( NULL == T2 ) || ( NULL == Bullet ) ) {
	printf("%s: Error: Not enough memory. Abort.\n", TOOLNAME );
	return -1;
      }

      /** Partition des transitions **/
      partition();


      /** affichage **/

      printf("%d \n",NT2);
      printf(" \n");

      printf("%d \n",NT1);
      printf(" \n");


      generate_mat();
      for (i=0;i<NT2;i++)
	{
	  for (j=0;j<NT2;j++)
	    {
	      printf("%d ",A[i][j]);
	    }
	  printf("\n");
	}

      printf("\n");
      for (i=0;i<NT2;i++)
	{
	  for (j=0;j<NT1;j++)
	    {
	      printf("%d ",B[i][j]);
	    }
	  printf("\n");
	}
      printf("\n");
      for (i=0;i<NT1;i++)
	{
	  for (j=0;j<NT1;j++)
	    {
	      printf("%d ",P[i][j]);
	    }
	  printf("\n");
	}
      printf("\n");
      for (i=0;i<NT1;i++)
	{
	  for (j=0;j<NT2;j++)
	    {
	      printf("%d ",Q[i][j]);
	    }
	  printf("\n");
	}
      printf("\n");


      Initial_Marking();
      for (i=0;i<NT1;i++)
	{printf("%d ",M[i]);}

      printf("\n");

      Calcul_bullet();
      for (i=0;i<NT1;i++)
	{printf("%d ",Bullet[i]);}
      printf("\n");

      /*for (i=0;i<Petri->Nb_Trans;i++)*/
      /*{printf("%s\n",Petri->T[i].Name);}*/

      for (i=0;i<Petri->Nb_Trans;i++)
	{if (Part[i]) {printf("%s\n ",Petri->T[i].Name);}}

      for (i=0;i<Petri->Nb_Trans;i++)
	{if (!Part[i]) {printf("%s\n ",Petri->T[i].Name);}}
      printf("\n");

      printf("\n");
      for (i=0;i<Petri->Nb_Trans;i++)
	{if (Part[i]) {printf("%d ",Petri->T[i].Tr_Id);}}

      for (i=0;i<Petri->Nb_Trans;i++)
	{if (!Part[i]) {printf("%d ",Petri->T[i].Tr_Id);}}
      printf("\n");
    }

    /* generation successful */
    return 0;
  } else if ( Help )
    return 0;
  else
    return -1;
}
