/************************************************************************/
/* This program tests liveness of a FCN                                 */
/* using the matrix method       (see RR ???? INRIA  )                  */
/************************************************************************/
/* Written by Bruno Gaujal under the name "liveness.c"			*/
/************************************************************************/

#include <stdio.h>
#include <stdlib.h>

#define TOOLNAME "live_fc"

#define infty 0x7FFFFFFF
#define TRUE 1
#define FALSE 0



int **A;   /* Cf evol equation pour la signification des mat A,B,C,P,Q, R */
int **B;
int **Q;
int **P;
int **C;
int *R; 


int **Astar;
int **Aplus;

int **Cs;    /* SI-subnet */
int **Ps;  
int **Qs;
int *Rs;


int N,M;   /* tailles des matrices */
int scs;     /* nb de compo fortement connexes */

int *bullet;   /* la place d'input (pour les transitions dans B uniquement) */
int *Vivant;   /* booleen qui teste la vivacite */
int *Dead;     /* booleen qui teste le deadlock  dans un SI */
int *Forget;   /* transitions a oublier */
int *Belong;   /* Belong[i] donne la composante f. c. qui contient i */
int *Order;    /* ordre topologique sur les transitions */

char **Name;     /* name[i] nom de la transition i dans le fichier pnl */
int *TrId;      /* TrId[i] Id de la transition i dans le fichier pnl */

/*************************** procedures *******************************************/

/****************************/
int min(int a, int b)
{
  return((a<b)?a:b);
}


/*************************/
int plus(int a, int b)
{
  if ((a == infty) || (b == infty))  {return(infty);}
  else 
    return(a+b);
}

/*************************/
/* int mult(a,b) */
/* int a,b; */
/* { */
/* if  ( (a*b != 0) &&((a == infty) || (b == infty)))  {return(infty);} */
/* else return(a*b); */
/* } */


/******* Procedure to compute C = A \otimes B , size I x H x J ************/
void Otimes(int **A, int **B, int **C, int I, int J, int H )
{
  int i,j,k;

  for (i = 0 ;i<I;i++) {
    for (j = 0 ;j<J;j++) {
      C[i][j] = infty;
    }
  }

  for (i = 0 ;i<I;i++)
    {
      for (j = 0 ;j<J;j++)
	{
	  for (k = 0 ;k<H;k++)
	    {
	      C[i][j] = min(plus(A[i][k],B[k][j]),C[i][j]);
	    }
	}
    }

}

/******************* algorithme de Floyd  pour calculer A*     ***********/
void calculStar(int **A,int **As)
{

  int i,j,k;
  int aux;

  for(i = 0;i<N;i++) {
    for (j= 0;j<N;j++) {

      As[i][j] = A[i][j];
    }
  }

  for (k=0;k<N;k++)
    {
      for(i = 0;i<N;i++)
	{
	  for (j= 0;j<N;j++)
	    {
	      aux = plus(As[i][k],As[k][j]);
	      if (aux < As[i][j]) As[i][j] = aux;
	    }
	}
    }
}


/********** calcule la decomposition en compo fortement connexes ************/
void calcul_Comp()                 
{

  int i,j;
  int temp;
  int now,p;
  int *val,*stack;

  /************** calcule la compo fortement connexe qui contient k ********/
  int visit(int k,int S) 
  {
    int m,min,i;

    now++;
    val[k] = now;
    min = now;

    stack[p] = k;
    p++;

    for (i = 0;i < S;i++)
      {
	if (Ps[i][k] != 0)   
	  {
	    if (val[i] == 0)
	      {
		m = visit(i,S);
		if (m < min) min = m;
	      }
	    else if (val[i] < min) min = val[i];
	  }
      }
    if (min == val[k])
      {
	scs++;
	do
	  {
	    p--;
	    /*Order[stack[p]] = 1 ;       min;*/
	    Belong[stack[p]] = scs;
	    val[stack[p]] = S+2;
	  }
	while (stack[p] != k);

      }

    return(min);
    /*val[k]= S+2;*/
  }

  /********* fin de visit *********/


  scs = -1;
  now = 0;
  p = 0;



  val =  (int *) calloc(M, sizeof(int));
  stack =  (int *) calloc(M, sizeof(int));

  for (i = 0;i<M;i++)
    {
      val[i] =0;                  /* initialisations a "rien" */
      stack[i] =-1;
    }


  for (i = 0;i<M;i++)
    {
      if (val[i] == 0) temp = visit(i,M);
    }

  /***************** calcul des composantes initiales (Order == 1) ***********/

  /** au depart tous initiaux **/

  for (i = 0;i<M;i++)
    {
      Order[i] = 1;
    }


  for (i = 0;i<M;i++)
    {
      for (j = 0;j<M;j++)
	{
	  if((P[j][i] !=0)&&(Belong[i] != Belong[j])) 
	    Order[j] = 2;
	}
    }
}











/*******************************/
void selection(int j)
  /*oublie tous les i initiaux qui partagent une place avec j */
{
  int i;

  for (i = 0 ; i < M ; i++) 
    if (Order[i] == 1)
      {
	if (bullet[i] == bullet[j]) 
	  {
	    Forget[i] = 1;
	  }
      }

}


/**************************/
void projection( int j )    /*oublie transition j */
{
  Forget[j] = 1;
}

/*****************************/
int column(int j) /* teste si la colonne j est vide pour les i initiaux non oublies */
{
  int i;
  int t = 0;
  for (i = 0; i < M; i++)
    {
      if ((Order[i] == 1) && (Forget[i] == 0)) 
	{
	  if (Ps[i][j] != 0) t=1;
	}
    }
  return(t);
}

/*****************************/
/* int ligne(i) /* teste si la ligne i est vide pour les j initiaux non oublies */ 
/* int i; */
/* { */
/* int j; */
/* int t = 0; */
/* for (j = 0; j < M; j++) */
/*   { */
/*     if ((Order[j] == 1) && (Forget[j] == 0))  */
/*       { */
/* 	if (Ps[i][j] != 0) t=1; */
/*       } */
/*   } */
/* return(t); */
/*}*/


/*************************************************************************/
/**************************** main program *******************************/
/*************************************************************************/
int main(int argc, char *argv[])
{

  static char		In_File_Name[128] = "";
  static char		Out_File_Name[128] = "";

  char str[128];

  int i,j,k;
  int aux;
  int pos,S,Sa,s,sa,L;
  int *SI;
  int new,life;
  int NbNotLive;

  int Args_OK = TRUE;
  int Help = FALSE;
  int InOut = TRUE;
  int gss = FALSE;
  int Number = FALSE;

  int m0 = 0;

  /**************************** Parsing ************************************/

  i = 1;

  if (argc>1)
    do {

      if ( 0 == strcmp( "-f", argv[i] ) ) {	
	i++;
	if ( i < argc ) {
	  strcpy( In_File_Name, argv[i] );
	  i++;
	}
	else {
	  fprintf(stderr,"%s: usage:filename expected after -f\n",argv[0]);
	  Args_OK = FALSE;
	}
      }
	
      else if (( 0 == strcmp( "-h", argv[i] )) || 
	       ( 0 == strcmp( "-?", argv[i] )) ||
	       ( 0 == strcmp( "-help", argv[i] ))) {
	i++;
	fprintf(stdout,"usage: %s [-f inputfile] [-o outputfile] [-gss] [-number].\n",
		TOOLNAME );
	fprintf(stdout,"The inputfile must contain the matrix description of a FCN in the format generated by the command matrice2. Option -gss generates an output adapted to the ers format.\n");
	  Args_OK = FALSE;
	  Help = TRUE;
      }
      else if ( 0 == strcmp( "-gss", argv[i] ) ) {
	i++;
	gss = TRUE;
      }
      else if ( 0 == strcmp( "-number", argv[i] ) ) {
	i++;
	Number = TRUE;
      }
      else if ( 0 == strcmp( "-o", argv[i] ) ) {
	i++;
	if ( i < argc) {
	  strcpy( Out_File_Name, argv[i] );
	  i++;
	}
	else {	
	  fprintf(stderr,"usage: %s: filename expected after -o\n",argv[0]);
	  Args_OK = FALSE;
	}
      }
      else i++;
    } while( Args_OK && ( i < argc ) );

  if (Args_OK) {

    if  ( ( strlen(In_File_Name) != 0 ) &&
	  (NULL == freopen( In_File_Name, "r", stdin ) )) {

      fprintf(stderr,"%s: could not open file %s\n",argv[0], In_File_Name );
      InOut = FALSE;
    }
  }


  if (Args_OK && InOut ) {

    if  ( strlen(Out_File_Name) != 0 ) {

      if ( NULL == freopen( Out_File_Name, "w", stdout )) {

	fprintf(stderr,"cannot open file %s\n",Out_File_Name);
	InOut = FALSE;
      }
    }
  }

  /** start of the program if everything is OK **/

  if (Args_OK && InOut ) {

    /*** read the sizes ***/

    scanf("%d",&N);

    scanf("%d",&M);

    /*************** Memory Allocation ***********************/


    A = (int **) calloc(N, sizeof(int*));
    for(i=0;i<N;i++)
      {
	A[i] = (int *) calloc(N, sizeof(int));
	for (j=0;j<N;j++)
	  {
	    A[i][j] = 0;
	  }
      }


    Astar = (int **) calloc(N, sizeof(int*));
    for(i=0;i<N;i++)
      {
	Astar[i] = (int *) calloc(N, sizeof(int));
	for (j=0;j<N;j++)
	  {
	    A[i][j] = 0;
	  }
      }

    Aplus = (int **) calloc(N, sizeof(int*));
    for(i=0;i<N;i++)
      {
	Aplus[i] = (int *) calloc(N, sizeof(int));
	for (j=0;j<N;j++)
	  {
	    A[i][j] = 0;
	  }
      }

    B = (int **) calloc(N, sizeof(int*));
    for(i=0;i<N;i++)
      {
	B[i] = (int *) calloc(M, sizeof(int));
	for (j=0;j<M;j++)
	  {
	    B[i][j] = 0;
	  }
      }




    P = (int **) calloc(M, sizeof(int*));
    for(i=0;i<M;i++)
      {
	P[i] = (int *) calloc(M, sizeof(int));
	for (j=0;j<M;j++)
	  {
	    P[i][j] = 0;
	  }
      }




    Q = (int **) calloc(M, sizeof(int*));
    for(i=0;i<M;i++)
      {
	Q[i] = (int *) calloc(N, sizeof(int));
	for (j=0;j<N;j++)
	  {
	    Q[i][j] = 0;
	  }
      }


    C = (int **) calloc(N, sizeof(int*));
    for(i=0;i<N;i++)
      {
	C[i] = (int *) calloc(M, sizeof(int));
	for (j=0;j<M;j++)
	  {
	    C[i][j] = 0;
	  }
      }



    Cs = (int **) calloc(N, sizeof(int*));
    for(i=0;i<N;i++)
      {
	Cs[i] = (int *) calloc(M, sizeof(int));
	for (j=0;j<M;j++)
	  {
	    Cs[i][j] = 0;
	  }
      }

    Ps = (int **) calloc(M, sizeof(int*));
    for(i=0;i<M;i++)
      {
	Ps[i] = (int *) calloc(M, sizeof(int));
	for (j=0;j<M;j++)
	  {
	    Ps[i][j] = 0;
	  }

      }


    Qs = (int **) calloc(M, sizeof(int*));
    for(i=0;i<M;i++)
      {
	Qs[i] = (int *) calloc(M, sizeof(int));
	for (j=0;j<M;j++)
	  {
	    Qs[i][j] = 0;
	  }
      }


    Name =  (char **) calloc(M+N, sizeof(char*));
    for(i=0;i<M+N;i++)
      {
	Name[i] = (char *) calloc(128, sizeof(char));
      }

    R = (int *)      calloc(M, sizeof(int));
    for (j=0;j<M;j++)
      {
	R[j] = 0;
      }
    Rs = (int *)     calloc(M, sizeof(int));
    for (j=0;j<M;j++)
      {
	Rs[j] = 0;
      }
    bullet = (int *) calloc(M, sizeof(int));
    for (j=0;j<M;j++)
      {
	bullet[j] = 0;
      }
    Dead  = (int *)  calloc(M, sizeof(int));
    for (j=0;j<M;j++)
      {
	Dead[j] = 0;
      }
    SI  = (int *)    calloc(M, sizeof(int));
    for (j=0;j<M;j++)
      {
	SI[j] = 0;
      }
    Forget = (int *) calloc(M, sizeof(int));
    for (j=0;j<M;j++)
      {
	Forget[j] = 0;
      }
    Order = (int *)  calloc(M, sizeof(int));
    for (j=0;j<M;j++)
      {
	Order[j] = 0;
      }
    Belong  = (int *)calloc(M, sizeof(int));
    for (j=0;j<M;j++)
      {
	Belong[j] = 0;
      }

    TrId = (int *) calloc(N+M, sizeof(int));
    for (j=0;j<N+M;j++)
      {
	TrId[j] = 0;
      }
    Vivant  = (int *) calloc(N+M, sizeof(int));
    for (j=0;j<N+M;j++)
      {
	Vivant[j] = 0;
      }
    /********************* read the inputs *******************/



    /***************** A ************************/

    for (i=0;i<N;i++)
      {
	for (j=0;j<N;j++)
	  {
	    scanf("%d",&(A[i][j]));

	  }
      }



    /***************** B ************************/

    for (i=0;i<N;i++)
      {
	for (j=0;j<M;j++)
	  {
	    scanf("%d",&(B[i][j]));

	  }

      }



    /***************** P ************************/

    for (i=0;i<M;i++)
      {
	for (j=0;j<M;j++)
	  {
	    scanf("%d",&(P[i][j]));

	  }
      }



    /***************** Q ************************/

    for (i=0;i<M;i++)
      {
	for (j=0;j<N;j++)
	  {
	    scanf("%d",&(Q[i][j]));

	  }
      }


    /*********************** R ****************/
    for (j=0;j<M;j++)
      {
	scanf("%d",&(R[j]));

      }

    /******************** bullet *************/   
    for (j=0;j<M;j++)
      {
	scanf("%d",&(bullet[j]));

      }


    /*********************** name ****************/ 
    for (j=0;j<M+N;j++)
      {
	scanf("%s",Name[j]);

      }


    /****************** Id *************************/

    for (j=0;j<M+N;j++)
      {
	scanf("%d",&(TrId[j]));

      }








    /*************************************************************/
    /*****************Initialisations ****************************/
    /*************************************************************/

    /***A priori, tout est vivant ******/
    for (j=0;j<M+N;j++)
      {
	Vivant[j] = TRUE;
      }



    /***********************************************************/
    /***************calcul de C = A* otimes B *****************/
    /**********************************************************/
    calculStar(A,Astar);
    Otimes(A,Astar,Aplus,N,N,N);  

    for (i=0;i<N;i++)
      {
	if (Aplus[i][i] == 0) Vivant[i] = FALSE;
      }


    Otimes(Astar,B,C,N,N,M);

    for (i=0;i<N;i++)
      {
	for (j = 0 ; j<M;j++)
	  {
	    if (C[i][j] != infty) SI[i] ++;
	  }

      }

    S = 1;
    for (i=0;i<N;i++)   /* S: number of SI subnets */
      {
	if (SI[i] !=0) S = S * SI[i];
      }


    /*****************************************************************************/
    /***************************** on passe tous les SIsubnets *******************/
    /*****************************************************************************/
    for (s= 0; s< S;s++)
      {

	/************* on depart on n'oublie aucune transitions  ***********/
	for (j = 0;j<M;j++) {Forget[j] = 0;}

	/************ calcul de Cs ****************/
	sa = s; 
	/*Sa = S;*/
	for (i=0;i<N;i++)
	  {
	    if ( SI[i] != 0)
	      {
		pos = 0;
		for (j = 0 ; j<M;j++)
		  {
		    if (C[i][j] != infty) 
		      {
			pos ++;
			if (pos == sa / SI[i]) 
			  {
			    Cs[i][j] = 1;
			    Rs[i] = C[i][j];
			  }
		      }
	 
		  }
		sa = sa % SI[i];
		/*Sa = Sa / SI[i];*/
	      }
	  }



	/******* calcul de  Ps et Rs *************/
	for (i=0;i<M;i++)
	  {
	    for (j = 0 ; j<M;j++)
	      {
		for (k = 0 ; k< N; k++)
		  {
		    Qs[i][j] += Q[i][k] * Cs[k][j];
		  }

	      }
	    Rs[i] += R[i];
	  }


	for (i=0;i<M;i++)
	  {
	    for (j = 0 ; j<M;j++)
	      {
		Ps[i][j] = P[i][j] + Qs[i][j];

	      }
	  }

	/*******decomposer Ps en composantes fortement connexes  ***********/
	calcul_Comp();




	/***************** calcul de la matrice de passage et de son inverse *****************/
	/* k = 0; */
	/* sc = 0; */

	/* for (d = 0; d<= depth; d++) */
	/* { */
	/*   for (s = 0; s<= scs; s++) */
	/*     { */
	/*       if (Order[s] == d) */
	/* 	{ */
	/* 	  for (j = 0; j< M;j++) */
	/* 	    { */
	/* 	      if (Comp[s][j] != -1)  */
	/* 		{  */
	/* 		  Pass[Comp[s][j]][k] = 1; */
	/* 		  Passi[k][Comp[s][j]] = 1; */
	/* 		  Compo[sc][j] = k; */
	/* 		  Rs[k] = Rs[Comp[s][j]]; */
	/* 		  k++; */
	/* 		} */
	/* 	    } */
	/* 	  sc++; */
	/* 	} */
	/*     } */
	/* } */

	/* /*****************************Changement de base *********************************/ 
	/* Mult(Ps,Pass,Aux,M,M,M); */
	/* Mult(Passi,Aux,Ps,M,M,M); */


	/*************** on selectionne sur  les SCS degeneres ***********************/
	for (j = 0;j<M;j++)
	  {
	    if (column(j) == 0) { selection(j);}
	  }


	/***************************on oublie les transitions non initiales ***************/
	for (j = 0;j<M;j++)
	  {
	    if (Order[j] > 1) {projection(j);}
	  }


	/***********************************************************************************/
	/****************** on enleve iterativement les lignes et colonnes vides ***********/
	/***********************************************************************************/
	new = TRUE;
	while (new)
	  {
	    new = FALSE;

	    /*************************** projection - selection / colonnes  vides ********************/
	    for (j = 0;j <M;j++)
	      {
		if ((Order[j] == 1) && (Forget[j] == 0))
		  {
		    if (column(j) == 0)
		      {
			selection(j);
			projection(j);
			new = TRUE;
		      }
		  }

	      }

	    /*************************** projection / lignes vides ***************************/
	    /* for (i = 0;i < M ; i++) */
	    /* 		{ */
	    /* 		  if ((Order[i] == 1) && (Forget[i] == 0)) */
	    /* 		    { */
	    /* 		      if (ligne(i) == 0) */
	    /* 			{ */
	    /* 			  new = TRUE; */
	    /* 			  projection(i); */
	    /* 			} */
	    /* 		    } */
	    /* 		} */

	  }   /**** fin du while ****/


	/***********a priori tout est mort *************/
	for (i = 0; i<M;i++)
	  {
	    Dead[i] = TRUE;
	  }

	/*****************test de deadlock ****************/
	for (i = 0; i<M;i++)
	  {
	    if ((Order[i] == 1)&& (Forget[i] == 0))
	      {
		m0 = 0;
		for (j = 0; j < M;j++)
		  {
		    if ((Belong[j] == Belong[i])  && (Forget[j] == 0) )
		      {
			m0+=Rs[j];
		      }
		  }
		if (m0 != 0) {Dead[i] = FALSE;}
	      }
	  }

	/** If Dead in one SI subnet, then not live **/
	for (i = 0; i<M;i++)
	  {
	    if ((Order[i] == 1) && (Dead[i] == TRUE)) 
	      Vivant[i+N] = FALSE;
	  }


      }   /** fin de la projection sur tous les SIsubnets **/


    /*********************************************************************/
    /******         on propage la non-vivacite                 ***********/
    /*********************************************************************/


    new = TRUE;

    while(new) {
      /***** on propage la non-vivacite dans les transitions de routage ***/
      new = FALSE;
      for (i = 0; i<M;i++)
	{
	  life = FALSE;
	  for (j = 0;j < M ;j++)
	    {
	      if ((P[i][j] != 0)  && (Vivant[j+N] == TRUE))
		{
		  life = TRUE;
		}
	    }
	  for (j = 0;j < N ;j++)
	    {
	      if ((Q[i][j] != 0)  && (Vivant[j] == TRUE))
		{
		  life = TRUE;
		}
	    }
	  if ((Vivant[i+N]==TRUE) && (life == FALSE)) 
	    {Vivant[i+N] = FALSE;new = TRUE;}
  
	  if ((Vivant[i+N]==FALSE) && (life == TRUE)) 
	    {Vivant[i+N] = TRUE;new = TRUE;}
	}




      /************* on propage la non-vivacite dans les MGC *************/
      for (i = 0; i<N;i++)
	{
	  for (j = 0;j < M ;j++)
	    {
	      if ((B[i][j] != infty)  && (Vivant[j+N] == FALSE))
		{
		  if (Vivant[i]){Vivant[i] = FALSE;new = TRUE;}
		}
	    }
	  for (j = 0;j < N ;j++)
	    {
	      if ((A[i][j] != infty)  && (Vivant[j] == FALSE))
		{
		  if (Vivant[i]){Vivant[i] = FALSE;new = TRUE;}
		}
	    }
	}

    }   /* fin de la propagation */


    /*************** On compte le nombre de transitions vivantes **************/
    NbNotLive = 0;
    for (i = 0; i < N+M; i++)
      {
	if (Vivant[i] == FALSE)  {NbNotLive++;}
      }


    /*************************************************/
    /*********** on affiche le resultat **************/
    /*************************************************/

    if ( Number ) {
      /* only the number of dead transitions */
      printf( "%d\n", NbNotLive );
    }
    else if (gss) { 
      /********** dump Live Transitions as gss *****************/
      printf( "# Live transitions generated by %s\n", TOOLNAME );
      printf( "\n%d\n_E_O_E\n", NbNotLive);
      for ( i=0; i < N+M ; i++ ) {
	if (!Vivant[i]) printf( "\n%d\n_E_O_E\n", TrId[i] );
      }
    }
    else { 
      /****** standard display ********/

      for (i = 0; i < N+M; i++) {
	if (Vivant[i]) {
	  printf(" %s  is live\n", Name[i]);
	}
	else {
	  printf(" %s  is not live\n", Name[i]);
	}
      }
    }

    return 0;

  }  /**** fin du if everything OK ****/
  else if ( Help )
    return 0;
  else
    return -1;

}  /**** fin du main ****/
