/* Implementation of boolean matrice and methods */

#include <malloc.h>
#include "boolmat.h"

#define TOOLNAME	"BoolMat"

/* Constructor */
#ifdef USE_SP_BOOL_MATRICES
Boolean_Matrix New_Boolean_Matrix( int Size )
{
  int		i;
  Boolean_Matrix Res;
  boolean	Alloc_Ok;

  Res = (SP_Bool_Mat) malloc( sizeof( struct SP_BOOL_MAT ) );

  Alloc_Ok = ( (SP_Bool_Mat)NULL != Res );

  if ( Alloc_Ok ) {
    Res->Size = Size;
    Res->Nb_Elts = (int*) calloc( (size_t)Size, sizeof(int) );
    Res->Elts = (int**) calloc( (size_t)Size, sizeof(int*) );
    
    Alloc_Ok = ( ( (int*)NULL != Res->Nb_Elts ) 
		 && ( (int**)NULL != Res->Elts ) );

    for ( i=0; (i<Size) && Alloc_Ok; i++ ) {
      Res->Nb_Elts[i] = 0;
      Res->Elts[i] = (int*)NULL;
    }
  }

  if ( !Alloc_Ok ) {
    Free_Boolean_Matrix( Res );
    Res = (SP_Bool_Mat)NULL;
  }

  return Res;
}
#else
Boolean_Matrix New_Boolean_Matrix( int Size )
{
  int		i;
  Boolean_Matrix Res;
  boolean	Alloc_Ok;

  Res = (Full_Bool_Mat) malloc( sizeof( struct FULL_BOOL_MAT ) );
  Res->Size = Size;

  Alloc_Ok = ( (Full_Bool_Mat)NULL != Res );

  if ( Alloc_Ok ) {
    Res->Size = Size;
    Res->Vals = (int**) calloc ( Size, sizeof(int*) );
    Alloc_Ok = ( (int**)NULL != Res->Vals );
    for ( i=0; (i<Size) && Alloc_Ok; i++ ) {
      Res->Vals[i] = (int*) calloc( (size_t)Size, sizeof(int) );
      Alloc_Ok = ( (int*)NULL != Res->Vals[i] );
      }
  }

  if ( !Alloc_Ok ) {
    Free_Boolean_Matrix( Res );
    Res = (Full_Bool_Mat)NULL;
  }

  return Res;
}
#endif

/* Destructor */
#ifdef USE_SP_BOOL_MATRICES
void Free_Boolean_Matrix( Boolean_Matrix A )
{
  int	i;

  if ( (Boolean_Matrix)NULL != A ) {
    free( A->Nb_Elts );
    if ( (int**)NULL != A->Elts ) {
      for ( i=0; i<A->Size; i++ )
	free( A->Elts[i] );
      free( A->Elts );
    }
    free( A );
  }
}
#else 
void Free_Boolean_Matrix(Boolean_Matrix A)
{
  int		i;

  if ( (Boolean_Matrix)NULL != A ) {
    if ( (int**)NULL != A->Vals ) {
      for ( i=0; i<A->Size; i++ ) {
	if ( (int*)NULL != A->Vals[i] ) {
	  free( A->Vals[i] );
	}
      }
      free( A->Vals );
    }
    free( A );
  }
}
#endif

/* Accessors */
#ifdef USE_SP_BOOL_MATRICES
boolean Is_Adj( Boolean_Matrix A, int From, int To )
{
  static int i;
  static boolean Res;

  Res = FALSE;
  for (i=0; (i<A->Nb_Elts[From]) && !Res; i++)
    Res = (A->Elts[From][i] == To);

  return Res;
}
#else
boolean Is_Adj( Boolean_Matrix A, int From, int To )
{
  return 1 == A->Vals[From][To];
}
#endif

#ifdef USE_SP_BOOL_MATRICES
void Set_Adj( Boolean_Matrix A, int From, int To, int Val )
{
  static int i;
  static int j;

  /* looking if element is already there */
  for ( i=0; 
	( i<A->Nb_Elts[From] ) && ( A->Elts[From][i]!=To ); 
	i++ );
  if ( i<A->Nb_Elts[From] ) {
    /* found it. */
    if ( Val != 0 ) {
      /* nothing to do */
    }
    else {
      /* removing value */
      for ( j=i+1; j<A->Nb_Elts[From]; j++ ) {
	A->Elts[From][j-1] = A->Elts[From][j];
      }
      A->Nb_Elts[From]--;
    }
  }
  else {
    /* not there */
    if ( Val != 0 ) {
      /* Adding it to the structure */
      A->Nb_Elts[From]++;
      A->Elts[From] = 
	(int*)realloc( A->Elts[From],
		       (size_t)A->Nb_Elts[From]*sizeof(int) );
	
      if ( (int*)NULL != A->Elts[From] ) {
	A->Elts[From][A->Nb_Elts[From]-1] = To;
      } 
      else {
	fprintf( stderr, "%s: warning: failed to add element to matrix.\n",
		 TOOLNAME );
      }
    }
  }
}
#else
void Set_Adj( Boolean_Matrix A, int From, int To, int Val )
{
  A->Vals[From][To] = Val;
}
#endif

/* Various properties */
int Max_Degree( Boolean_Matrix A )
{
  int	i;
#ifndef USE_SP_BOOL_MATRICES
  int	j;
#endif
  int	Max;
  int	Deg;

  Max = 0;
  for ( i=0; i<A->Size; i++ ) {
#ifdef USE_SP_BOOL_MATRICES
    Deg = A->Nb_Elts[i];
#else
    Deg = 0;
    for ( j=0; j<A->Size; j++ )
      if ( A->Vals[i][j] != 0 )
	Deg++;
#endif
    if ( Deg > Max )
      Max = Deg;
  }

  return Max;
}

double Avg_Degree(  Boolean_Matrix A )
{
  int	i;
#ifndef USE_SP_BOOL_MATRICES
  int	j;
#endif
  int	Tot;

  Tot = 0;
  for ( i=0; i<A->Size; i++ ) {
#ifdef USE_SP_BOOL_MATRICES
    Tot += A->Nb_Elts[i];
#else
    for ( j=0; j<A->Size; j++ ) {
      if ( A->Vals[i][j] != 0 )
	Tot++;
    }
#endif
  }

  return (double)Tot/(double)A->Size;
}

/* I/O */
/* Writes out a Boolean matrix in full format. For debugging purposes */
void Write_Boolean_Matrix(Boolean_Matrix	A,
			  FILE*			Output)
{
  int i;
  int j;
  
  for (i=0; i<A->Size; i++) {
    for (j=0; j<A->Size; j++)
      if ( Is_Adj(A,i,j) )
	fprintf( Output, "1" );
      else
	fprintf( Output, "0" );
    fprintf( Output, "\n" );
  }
}
