/************************************************************************/
/* This program generates a .fig file and draws eigen vectors           */     
/* of matrices 3x3 in the (max, +) algebra in PR^2.      		*/
/* Copyright (c) 1993 by Bruno Gaujal & Jean Mairesse     		*/
/* (E-mail gaujal@sophia.inria.fr)                             		*/
/* Permission to use, copy, modify this program is granted without fee, */
/* provided that this copyright notice appears.     			*/
/* The program is provided "as is" without any warrranty.     		*/
/************************************************************************/
/* Known originally as "drawfig.c", renamed "eigenfig.c" within ERS	*/
/************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <malloc.h>

#define	TOOLNAME	"eigenfig"
#define VERSION		"1.2 (22/12/1998)"

#define infty -1e8
#define N 3		/* A priori, this is the dimension of the matrix */
#define NC 8
#define x0 4
#define y0 4
#define Ratio 80
#define PRECISION 1e-4
/*#define DEBUG*/

typedef struct LISTE
{
double x;
double y;
double z;
struct LISTE  *next;
} *liste;


typedef struct LISTEP
{
int x;
int y;
struct LISTEP *next;
} *listeP;


typedef int boolean ;


#define TRUE 1
#define FALSE 0

#ifdef DEBUG
boolean Debug = TRUE;
# else
boolean Debug = FALSE;
#endif







/*************************global variables **********************/

char str[128];

liste PtList,PtCur,PtNew;
listeP PtListP,PtCurP,PtNewP;

double **A,**Ab,**Aux,**Abux,**As,**An,**Ap,**I, **P, **X, **Y;
double Dep[3];
double Nxt[3];
double d1[2], d2[2], d3[2];
int EV[3];
double Cy[8][2];
double Z1[3][3],Z2[3][3],Z3[3][3];
int Z1P[2][3],Z2P[2][3], Z3P[2][3];
int **AsP , **PP;
double mi[2];
int kk[3];

double Diff[3];
double nor;

int numberPoints;
boolean traj;

/*************************/
double plus(double a, double b)
{
if ((a <= infty) || (b <= infty))  {return(infty);}
else return(a+b);
}

/****************************/
double Max(double a,double b)
{
return((a>b)?a:b);
}

/***************************/
void Add(double **A, double **B, double **C)
{

int i,j;


for (i = 0 ;i<N;i++)
  {
    for (j = 0 ;j<N;j++)
      {
	C[i][j] = Max(A[i][j],B[i][j]);
      }
  }

}





/****************************************************************************/
void mult(double **A, double **B , double **C)
{
int i,j,k;

for (i = 0 ;i<N;i++)
  {
    for (j = 0 ;j<N;j++)
      {
	C[i][j] = infty;
      }
  }

for (i = 0 ;i<N;i++)
  {
    for (j = 0 ;j<N;j++)
      {
	for (k = 0 ;k<N;k++)
	  {
	    C[i][j] = Max(plus(A[i][k],B[k][j]),C[i][j]);
	  }
      }
  }



}



/***************************************************************************/

/****************************************************************************/
void prodMV(double **A, double *B , double *C)
{
int j,k;


 for (j = 0 ;j<N;j++)
   {
     C[j] = infty;
   }
  


 for (j = 0 ;j<N;j++)
   {
     for (k = 0 ;k<N;k++)
       {
	    C[j] = Max(plus(A[j][k],B[k]),C[j]);
       }
   }
  
}

/****************************************************************************/
void power(double **A, int n, double **B)
{
double **C;
int i,j,k;

C = (double **) calloc(N, sizeof(double));
 for(i=0;i<N;i++)
   {
      C[i] = (double *) calloc(N, sizeof(double));
    }


for (i = 0 ;i<N;i++)
  {
    for (j = 0 ;j<N;j++)
      {
	B[i][j] = A[i][j];
      }
  }

for ( i = 1; i <n ; i++)
  {
    mult(B,A,C);

    for (k = 0 ;k<N;k++)
      {
	for (j = 0 ;j<N;j++)
	  {
	    B[k][j] = C[k][j];
	  }
      }
    
  }

}

/****************************************************************************/
int compar(double **A, double **B)
{
int result = 0;
int i,j;
for (i = 0 ;i<N;i++)
  {
    for (j = 0 ;j<N;j++)
      {
       
	if (fabs(A[i][j]-B[i][j]) > PRECISION) result = 1;
      }
  }
return(result);
}







/*************************************************************************/
boolean in(liste L, double E[3])
{
boolean res;

res = FALSE;

  while(L != NULL)
    {
      if ( (fabs(L->x - E[0]- L->y + E[1])< PRECISION ) && (fabs(L->y - E[1] - L->z + E[2]) < PRECISION ) )
	{
	  res = TRUE;
	}
      L = L->next;
    }
return(res);
}

      












/************************************************************************/

double **Alloc_Matrix( Size )
{
  int i;
  double **Res;

  Res = (double **) calloc( (size_t)Size, sizeof(double) );
  for( i=0; i<Size; i++ )
    Res[i] = (double *) calloc( (size_t)Size, sizeof(double) );

  return Res;
}

void Write_Matrix( double **A ) {
  int i;
  int j;

  for (i=0;i<N;i++) {
    for (j=0;j<N;j++) {
      if ( A[i][j] == infty ) {
	fprintf(stderr,"    eps    ");
      }
      else {
	fprintf(stderr,"%10f ",A[i][j]);
      }
      fflush(stderr);
    }
    fprintf(stderr,"\n");
  }
}

void Write_3x3_Matrix( double A[3][3] ) {
  int i;
  int j;
  int NN=3;

  for (i=0;i<NN;i++) {
    for (j=0;j<NN;j++) {
      if ( A[i][j] == infty ) {
	fprintf(stderr,"    eps    ");
      }
      else {
	fprintf(stderr,"%10f ",A[i][j]);
      }
      fflush(stderr);
    }
    fprintf(stderr,"\n");
  }
}

/****************************************************************************/
int main(int argc, char *argv[])
{

  static char		In_File_Name[128] = "";
  static char		Out_File_Name[128] = "";

  double V1[3],V2[3],V3[3];
  double J[3];
  int V1P[2],V2P[2], V3P[3];

  int check;
  int i,j,k,l;
  int imin,imax,imoy;
  int width;
  int exclusion =0;
  int centerx,centery,axis1x,axis1y,axis2x,axis2y,axis3x,axis3y;
  
  int vect1 = 0;
  int vect2 = 0;
  int vect3 =0;
  int flag = 0;
  int scs = 0;      	/* # strongly connected subgraphs */
  int cyc = 1;      	/* cyclicity */
  double ev;         	/* eigenvalue */
  double min;
  double max;

  int first;
  int M;		/* dimension of the matrix in the input file */

  int nev;  /* non eigenvector */
  boolean  	Args_OK = TRUE;
  boolean  	Help = FALSE;
  boolean	InOut = TRUE;
  boolean	append = FALSE;
  int		Fig_Version = 3;
  int		Scale;

  /************************* Parsing *****************************/
  i = 1;

  if (argc>1)
    do {

      if ( 0 == strcmp( "-f", argv[i] ) ) {
	i++;
	if ( i < argc ) {
	  strcpy( In_File_Name, argv[i] );
	  i++;
	}
	else {
	  fprintf(stderr,"%s: usage: filename expected after -f\n",
		  TOOLNAME );
	  Args_OK = FALSE;
	}
      }
      else if  ( 0 == strcmp( "-a", argv[i] ) ) {
	i++;
	if (( i < argc ) && (exclusion == 0)) {
	  strcpy( Out_File_Name, argv[i] );
	  exclusion =1;
	  append = TRUE;
	  i++;
	}
	else {
	  fprintf(stdout,"%s: usage:filename expected after -a\n",argv[0]);
	  Args_OK = FALSE;
	}
      }
      
      else if ( 0 == strcmp( "-trajectory", argv[i] ) ) 
	{
	  i++;
	  traj = TRUE;
	}
      else if ( 0 == strcmp( "-debug", argv[i] ) ) {
        Debug = TRUE;
	i++;
      }
      else if ( 0 == strcmp( "-v", argv[i] ) ) {
	printf("%s (Version %s)\n", TOOLNAME, VERSION);
	Args_OK = FALSE;
	Help = TRUE;
	i++;
      }
      else if ( 0 == strcmp( "-fig_version", argv[i] ) ) {
	i++;
	if ( i < argc ) {
	  sscanf( argv[i], "%d", &Fig_Version );
	  i++;
	} else {
	  fprintf(stdout,"%s: integer expected after %s\n",
		  TOOLNAME, argv[i-1] );
	  Args_OK = FALSE;
	}
      }
      else if (( 0 == strcmp( "-h", argv[i] )) || 
	       ( 0 == strcmp( "-?", argv[i] )) ||
	       ( 0 == strcmp( "-help", argv[i] ))) {
	i++;
	fprintf(stdout,"usage: %s [-f inputfile] [-a outputfile.fig] [-o outputfile.fig][-trajectory]", TOOLNAME );
	fprintf(stdout, "\n\t[-fig_version <int>] [-debug]\n");
	fprintf(stdout, "Options -a and -o are mutually exclusive.\nIf -a is used outputfile.fig is appended so it must already be a .fig file.\nif -o is used the ouputfile is reset so old version (if any) is lost.\nThe inputfile must contain the 2x2 or 3x3 matrix with the following convention:\nthe string 'eps' represents -infinity and \nnumbers smaller than %f are treated as -infinity.\nIf option -trajectory is used, the input file must contain the 3 coordinates of a starting point\n", 
		infty );
	Args_OK = FALSE;
	Help = TRUE;
      }
      else if ( 0 == strcmp( "-o", argv[i] ) ) {
	i++;
	if (( i < argc) && (exclusion ==0) ) {
	  strcpy( Out_File_Name, argv[i] );
	  exclusion = 1;
	  i++;
	}
	else {
	  fprintf(stderr,"usage: %s: filename expected after -o\n",
		  TOOLNAME );
	  Args_OK = FALSE;
	}
      }
      else i++;
    } while( Args_OK && ( i < argc ) );

  if (Args_OK) {
    if  ( ( strlen(In_File_Name) != 0 ) &&
	  (NULL == freopen( In_File_Name, "r", stdin ) )) {
      fprintf(stderr,"%s: could not open file %s\n",
	      TOOLNAME, In_File_Name );
      InOut = FALSE;
    }
  }


  /** start of the program if everything is OK **/

  if (Args_OK && InOut ) {

    I = Alloc_Matrix( N );
    P = Alloc_Matrix( N );
    A = Alloc_Matrix( N );
    Ab = Alloc_Matrix( N );
    Aux = Alloc_Matrix( N );
    Abux = Alloc_Matrix( N );
    As = Alloc_Matrix( N );
    An = Alloc_Matrix( N );
    Ap = Alloc_Matrix( N );
    X = Alloc_Matrix( N );
    Y = Alloc_Matrix( N );

    AsP = (int **) calloc(2, sizeof(int));
    for(i=0;i<2;i++)
      AsP[i] = (int *) calloc(N, sizeof(int));

    PP = (int **) calloc(2, sizeof(int));
    for(i=0;i<2;i++)
      PP[i] = (int *) calloc(N, sizeof(int));

    /**********coefficients for the  projection **********/
    d2[0] = 0;
    d2[1] = 1;
    d1[0] = 0.8660254;  /* sqrt(3)/2 */
    d1[1] = -0.5;
    d3[0] = -0.8660254;
    d3[1] = -0.5;

    EV[1] = EV[2] = EV[0] = 0;

    /********************* read the input matrix *******************/

    /* Read the size and check that it is less or equal to 3! */
    scanf("%s",str);
    sscanf(str,"%d",&M);
    if ( ( M != 3 ) && ( M != 2 ) ) {
      printf("%s: error: input matrix must be of dimension 2 or 3 (got %d).\n",
	     TOOLNAME, M );
      return -4;
    }

    for (i=0;i<M;i++) {
      for (j=0;j<M;j++) {

	scanf("%s",str);
	
	if (strcmp(str,"eps")==0) {
	  A[i][j] = infty ; 
	  Ab[i][j] = infty;
	}
	else { 
	  sscanf(str,"%lf",&A[i][j]);
	  if (A[i][j] <= infty) {
	    Ab[i][j] = infty;
	    A[i][j] = infty;
	  } 
	  else Ab[i][j] = 0;
	}
      }
    }

    if ( M == 2 ) {
      /* must complete the matrix */
      A[0][2] = A[0][1];
      A[2][0] = A[1][0];
      A[2][2] = A[1][1];
      Ab[0][2] = Ab[0][1];
      Ab[2][0] = Ab[1][0];
      Ab[2][2] = Ab[1][1];
      A[1][2] = infty;
      A[2][1] = infty;
      Ab[1][2] = infty;
      Ab[2][1] = infty;
    }


    if (traj)
      for (i=0;i<M;i++)
	{
	  scanf("%s",str);
	  if (strcmp(str,"eps")==0) 
	    { Dep[i] = infty ;}
	  else { 
	    sscanf(str,"%lf",&Dep[i]);
	   
	  }
	}

    if (M == 2)
      {
	Dep[2] = Dep[1];
      }

      

    if ( Debug ) {
      fprintf( stderr, "Matrix A:\n" );
      Write_Matrix( A );
      fprintf( stderr, "Matrix Ab:\n" );
      Write_Matrix( Ab );
    }




/*********** checking irreducibility *****************/
power(A,3,Aux);

for (i=0;i<N;i++)
  {
    for (j=0;j<N;j++)
      {
	if (Aux[i][j] <= infty)  
	  {
	    fprintf(stderr," The matrix is not irreducible (no output)." );
	    exit(0);
	  }
      }
  }

/*************** check aperiodicity *******************/
check = 0;
for (i=0;i<N;i++)
  {
    for (j=0;j<N;j++)
      {
	if (Ab[i][j] != 0) check = 1;
      }
  }

if (check)
  {
power(Ab,2,Abux);
if (compar(Ab,Abux)==0) 
      {
	fprintf(stderr," The matrix is periodic (no output).");
	exit(0);
      }

power(Ab,3,Abux);
if (compar(Ab,Abux)==0)  
	  {
	    fprintf(stderr," The matrix is periodic (no output).");
	    exit(0);
	  }
  }




/*********** critical graph of A **************/

Cy[0][0] = A[0][0];
Cy[1][0] = A[1][1];
Cy[2][0] = A[2][2];
Cy[5][0] = (A[0][1] + A[1][0])/2;
Cy[4][0] = (A[0][2] + A[2][0])/2;
Cy[3][0] = (A[1][2] + A[2][1])/2;
Cy[6][0] = (A[0][2] + A[2][1] + A[1][0])/3;
Cy[7][0] = (A[2][0] + A[0][1] + A[1][2])/3;

ev = Cy[0][0];
for (i = 0 ; i < NC ; i++)
  {
    if( Cy[i][0] > ev ) { ev = Cy[i][0];}
  }

  if ( Debug ) {
    fprintf (stderr, "Eigenvalue: %f \n",ev);
  }

  for (i = NC-1 ; i >= 0  ; i--) {
    if (Cy[i][0] == ev) {

      Cy[i][1] = 1; 

      if ( Debug ) {
	fprintf (stderr, "critical cycle: %d \n",i);
      }

	if (i >=6) 
	  {
	    EV[0] = EV[1] = EV[2] = 1; 
	    scs =1; 
	    cyc = 3; 
	    flag = 1;
	  }
	
	if ((i >= 3) && (i < 6))
	  {
	    if (EV[i-3] == 1) 
	      {
		EV[0]  = EV[1] = EV[2] = 1;
	      }
	    else  
	      {
		EV[0]  = EV[1] = EV[2] = 1; 
		EV[i-3] = 0;
	      }
	    
	    if (flag == 0) 
	      {
		cyc = 2;
		scs =1;
	      }
	    
	    else cyc = 1;
	     
	  }
	if (i < 3) 
	  {
	    if (EV[i] == 0) {scs += 1;}
	    else cyc =1;
	    EV[i] = 1;
	    
	  }
       }
    else {Cy[i][1] = 0;}
  }
  if ( Debug ) {
    fprintf(stderr," XXXX\n");
    fprintf(stderr," scs1: %d   cyc: %d \n",scs,cyc);
    fprintf(stderr," XXXX\n");
  }

/****************** stationary  matrix **********************/

power(A,cyc,Ap);   /* Ap is of cyclicity 1 */

  if ( Debug ) {
    printf ("Matrix Ap (of cyclicity 1):\n");
    Write_Matrix( Ap );
  }

  for (i = 0; i<N;i++) {
    for (j = 0; j<N;j++) {

	An[i][j] = Ap[i][j] - cyc*ev;
    }
  }

  if ( Debug ) {
    printf ("Matrix An (with eigenvalue 0):\n");
    Write_Matrix( An );
  }
  
  for (i=0;i<N;i++) {
    for (j=0;j<N;j++) {

	X[i][j] = infty;
	Y[i][j] = An[i][j];
    }
  }

  while ( compar(X,Y)) {
    
    for (i=0;i<N;i++) {
      for (j=0;j<N;j++) {
	X[i][j] = Y[i][j];
      }
    }
    mult(X,An,Y);
  }

  /* X: stationary version of the matrix */

  if ( Debug ) {
    fprintf(stderr,"Matrix X (stationary version):\n");
    Write_Matrix( X );

    fprintf(stderr,"Critical nodes:\n");
    for (i = 0 ; i< N; i++) {
      fprintf(stderr,"EV[%d] = %d  ", i, EV[i]);
    }
    fprintf(stderr,"\n");
  }

/******************* construction of the 3 points *********************/


/**************** case 1 ****************/
if (scs + cyc == 4)
  {
    for ( i =0 ; i <N ; i++)
      {
	for ( j = 0 ; j< N; j++)
	  {
	    As[i][j] = X[i][j];
	  }
      }

    if ( cyc == 1)
      {
	/* construct the 3 eigenvectors */
	vect1 = vect2 = vect3 =1;
	 for ( i = 0 ; i <N ;i++)
	  {
	    V1[i] = As[i][0];
	    V2[i] = As[i][1];
	    V3[i] = As[i][2];
	  }
       }
    

    if (cyc ==3 ) 
      {
	/* construct the eigenvector */
	vect1 = 1;
	for ( i = 0 ; i <N ;i++)
	  {
	    V1[i] = Max(As[i][0],Max(As[i][1],As[i][2]));
	  }
      }



    if (cyc == 2 ) 
      {
	/* construct the two eigenvectors */
	vect1 = 1;
	vect2 = 1;
	for ( i = 0 ; i <3 ;i++)
	  {
	    if ( Cy[i][1] ==1)
	      {
		k = i;
		V1[0] = As[0][k];
		V1[1] = As[1][k];
		V1[2] = As[2][k];
	      }
	  }
/** bbb*/
	
	
	    min = max = As[0][(k+1)%3] - As[0][(k+2)%3];
	    imin = imax = 0;
	    for (j = 0 ; j < N; j++)
	    {
		J[j] = As[j][(k+1)%3] - As[j][(k+2)%3];
		if (J[j] < min) {min = J[j];imin = j;}
		if (J[j] > max) {max = J[j];imax = j;}
		
	    }
	
	
	imoy = imin;
	    for (j = 0 ; j < N; j++)
	    {
		if ((J[j] > min ) && ( J[j]< max)) {imoy = j;}
	    }

	
	V2[imin]  = As[imoy][(k+1)%3] - As[imoy][(k+2)%3] + As[imin][(k+2)%3];
	V2[imax]  = As[imax][(k+1)%3];
	V2[imoy]  = As[imoy][(k+1)%3];

	/*V2[0] = Max(As[0][(k+1)%3],As[0][(k+2)%3]);
	V2[1] = Max(As[1][(k+1)%3],As[1][(k+2)%3]);
	V2[2] = Max(As[2][(k+1)%3],As[2][(k+2)%3]);*/
      }

  
  }



/**************** case 2 *******************/
if (scs + cyc == 2)
  {
    /* construct the eigenvector */
    vect1=1;
    for ( i = 0 ; i <N ;i++)
	  {
	    if (EV[i] == 1)
	      {
		 V1[0] = X[0][i];
		 V1[1] = X[1][i];
		 V1[2] = X[2][i];
	      }
	  } 
    
  }


/************** case 3 ********************/


if ( scs + cyc == 3)
  {

    /* case where 2 vectors may coincide */
    if (EV[0] == 1 && EV[1] == 1 && EV[2] == 1)
      {
	
	if (X[0][0] ==X[0][1] && X[1][0] == X[1][1] && X[2][0] ==X[2][1])
	  {
	    EV[1] = 0;
	    
	  }
	if (X[0][1] ==X[0][2] && X[1][1] == X[1][2] && X[2][1] ==X[2][2])
	  {
	    EV[2] = 0;
	    
	  }
	if (X[0][2] ==X[0][0] && X[1][2] == X[1][0] && X[2][2] ==X[2][0])
	  {
	    EV[2] = 0;
	    
	  }
      }
    
    for (i = 0 ; i < N ; i++)
      {
	if ( EV[i] == 0) k = i;
      }
  
     for ( i = 0 ; i < N ; i++)
      {
	if ( i != k) 
	  {
	    
	    for ( j = 0 ; j< N; j++)
	      {
		As[j][i] = X[j][i];
	      }
		
	  }
      }


    l=0;
    for ( i =0 ; i <N ; i++)
      {
	if ( i!= k)
	  {
	    
	    mi[l] = X[0][i]- X[0][k];
	    for (j = 0 ; j <N; j++)
	      {
		if ( mi[l] < X[j][i]- X[j][k])  mi[l] = X[j][i]- X[j][k];
	      }
	    l++;
	  }
      }



    l=0;
    for ( j =0 ; j <N ; j++)
      {
	if ( j !=k)
	  {
	    As[j][k] = -mi[l];
	    l++;
	  }
	
      }
    As[k][k] = 0;

#ifdef DEBUG

  for ( i =0 ; i <N ; i++)
      {
	   fprintf(stderr,"EV[%d] = %d \n", i,EV[i]);
	 for ( j =0 ; j <N ; j++)  
	 {
	     fprintf(stderr,"As[%d][%d] = %f \n", i,j,As[i][j]);
	 }
     }

#endif 
    
    if ( cyc == 1)
      {
	/* construct the two eigenvectors */
	first = 1;
	vect1 = vect2 = 1;
	for ( i = 0 ; i <N ;i++)
	  {
	    if (EV[i] == 1 && first)
	      {
		first = 0;
		 V1[0] = As[0][i];
		 V1[1] = As[1][i];
		 V1[2] = As[2][i];
	      }
	    if (EV[i] == 1 && !first)
	       {
		
		 V2[0] = As[0][i];
		 V2[1] = As[1][i];
		 V2[2] = As[2][i];
	      }
	  }
      }
	      

    if (cyc == 2) 
      {
	/* construct the only eigenvector */
	vect1 = 1;
	l = 0;

	for ( i = 0 ; i <N ;i++)
	  {
	    if (EV[i] == 1)
	      { kk[l++] = i;}
	  }

	/*for ( j = 0 ; j < N ; j++)
	{*/
	    min = max = As[0][kk[0]] - As[0][kk[1]];
	    imin = imax = 0;
	    for (k = 0 ; k < N; k++)
	    {
		J[k] = As[k][kk[0]] - As[k][kk[1]];
		if (J[k] < min) {min = J[k];imin = k;}
		if ((J[k] >= max) && (imin != k )) {max = J[k];imax = k;}
	    }
	
	/*}*/
	imoy = imin;
	    for (k = 0 ; k < N; k++)
	    {
	      /*if ((J[k] > min ) && ( J[k]< max)) {imoy = k;}*/
	      if ((k != imin) && (k!=imax)) {imoy = k;}
	    }

	    if (J[imax]+J[imin] < 2*J[imoy])
	      {
		V1[imin]  = As[imin][kk[1]] + 0.5*(J[imax]+J[imin]);
		V1[imax]  = As[imax][kk[0]];
		V1[imoy]  = As[imoy][kk[0]];
	      }
	    else
	      {
		V1[imin]  = As[imin][kk[1]] + 0.5*(J[imax]+J[imin]);
		V1[imax]  = As[imax][kk[0]];
		V1[imoy]  = As[imoy][kk[1]] + 0.5*(J[imax]+J[imin]);
	      }

    }

}
	
    
/************************* print the 3 points *************************/
    
  if ( Debug ) {
    fprintf(stderr,"Matrix As (characteristic points):\n");
    Write_Matrix( As );
  }

/**********renumerotation des trois points **************/
/*
{
int taken[3],D1[3],D2[3];
double **Rs;
float norm1,norm2;

taken[0] = taken[1] = taken[2] = 0;

Rs = (float **) calloc(N, sizeof(float));
 for(i=0;i<N;i++)
   {
      Rs[i] = (float *) calloc(N, sizeof(float));
    }


for ( i = 0 ; i < N ; i++)
      {
	for ( j = 0 ; j < N ; j++)
	  {
	    D1[j] = As[j][(i+1)%3] - As[j][i];
	    D2[j] = As[j][(i+2)%3] - As[j][i];

	  }
	norm1 = Max(D1[0],Max(D1[1],D1[2]));
	norm2 = Max(D2[0],Max(D2[1],D2[2]));
	for ( j = 0 ; j < N ; j++)
	  {
	    D1[j] -= norm1;
	    D2[j] -= norm2;
	  }
	if ((D1[0] <= 0) && (D2[1] <= 0) && (taken[0] == 0))
	  {
	    for ( j = 0 ; j < N ; j++)
	      {
		Rs[j][0] = As[j][i];
		EV[0] = EV[i];
	      }
	    taken[0] = 1;
	  }
	if ((D1[1] <= 0) && (D2[1] <= 0) && (taken[1] == 0))
	  {
	    for ( j = 0 ; j < N ; j++)
	      {
		Rs[j][1] = As[j][i];
		EV[1] = EV[i];
	      }
	    taken[1] = 1;
	  }
	if ((D1[2] <= 0) && (D2[2] <= 0) && (taken[2] == 0))
	  {
	    for ( j = 0 ; j < N ; j++)
	      {
		Rs[j][2] = As[j][i];
		EV[2] = EV[i];
	      }
	    taken[2] = 1;
	  }
      }

for ( i = 0 ; i < N ; i++)
      {
	for ( j = 0 ; j < N ; j++)
	  {
	    As[i][j] = Rs[i][j];
	  }
      }
	
}
*/
/*********************** construction of the convex hull *******************/




for (i = 0 ; i < N ; i++)
  {
    for ( j = i+1 ; j < N ; j++)
      {
       	min = max = As[0][i] - As[0][j];
	imin = imax = 0;
	for (k = 0 ; k < N; k++)
	  {
	    I[k][i+j-1] = As[k][i] - As[k][j];
	    if (I[k][i+j-1] < min) {min = I[k][i+j-1];imin = k;}
	    if (I[k][i+j-1] >= max) {max = I[k][i+j-1];imax = k;}
	  }

	imoy = imin;
	for (k = 0 ; k < N; k++)
	  {
	    P[k][i+j-1]  = As[k][i];
	  }

	for (k = 0 ; k < N; k++)
	  {
	    /*if ((I[k][i+j-1] > min ) && ( I[k][i+j-1] < max)) {imoy = k;}*/
	    if ((k != imin ) && (k!= imax)) {imoy = k;}
	  }

	P[imin][i+j-1]  = As[imoy][i] - As[imoy][j] + As[imin][j];
	P[imax][i+j-1]  = As[imax][i];
	P[imoy][i+j-1]  = As[imoy][i];

      }
  }

  if ( Debug ) {
    fprintf(stderr,"Matrix P:\n");
    Write_Matrix( P );
  }
	

/**************** construction of the zones of attractions ***************/

kk[1] = kk[0]= kk[2] = -1;

nev = -1;

if (scs + cyc == 3)
  {
    for ( k = 0;k<N ; k++)
	  {
	    if (EV[k] ==0 ) nev = k; /*the only non eigenvector is column k */
	  }

    for ( i = 0 ; i < N ; i++)
      {
	if ( i == nev)
	  {
	    for ( k = 0;k<N ; k++)
	      {
		Z1[k][i] = As[k][i];
		Z2[k][i] = As[k][i];
		Z3[k][i] = As[k][i];
	      }
	  }
	else
	  {
	    if (((P[0][i+ nev -1] ==  As[0][i]) && 
		 (P[1][i+ nev -1] ==  As[1][i]) && 
		 (P[2][i+ nev -1] ==  As[2][i])) || 

		((P[0][i+ nev -1] ==  As[0][nev]) && 
		 (P[1][i+ nev -1] ==  As[1][nev]) && 
		 (P[2][i+ nev -1] ==  As[2][nev]))) 

	      {
		Diff[0] = As[0][i] - As[0][nev];
		Diff[1] = As[1][i] - As[1][nev];
		Diff[2] = As[2][i] - As[2][nev];
		
		nor = Max(Diff[0],Max(Diff[1],Diff[2]));
		Diff[0] -= nor;
		Diff[1] -= nor;
		Diff[2] -= nor;   /* Diff is in the direction nev or i*/
		 

	      if (Diff[(nev+1)%3]  == Diff[(nev+2)%3]) /*diff in direc. nev */
		{
		  for ( k = 0;k<N ; k++)
		    {
		      Z1[k][i] = As[k][nev];
		      Z2[k][i] = As[k][nev];
		      Z3[k][i] = As[k][nev];
		    }
		}
	      else
		  {
		    for ( k = 0;k<N ; k++)
		      { 
			Z1[k][i] = As[k][i];
			Z2[k][i] = As[k][i];
			Z3[k][i] = As[k][i];
		      }
		  }
	      }
		    
	    else 
	    for ( k = 0;k<N ; k++)
	      {
		Z1[k][i] = P[k][i+ nev -1];
		Z2[k][i] = P[k][i+ nev -1];
		Z3[k][i] = P[k][i+ nev -1];
	      }
	  }
      }
  }
    
   

else
      {
	for ( i = 0 ;i < N; i++)
	  {
	    for ( k = 0;k<N ; k++)
	      {
		Z1[k][i] = As[k][i];
		Z2[k][i] = As[k][i];
		Z3[k][i] = As[k][i];
	      } 
	  }
      }

  if ( Debug ) {
    fprintf(stderr,"Matrix Z1:\n");
    Write_3x3_Matrix( Z1 );
    fprintf(stderr,"Matrix Z2:\n");
    Write_3x3_Matrix( Z2 );
    fprintf(stderr,"Matrix Z3:\n");
    Write_3x3_Matrix( Z3 );
  }

l = 0;
for ( i = 0 ;i < N; i++)
  {
    
	if ( EV[i] == 1 )
	  { 
	    Z1[(i+1)%3][i] -= 2;
	    Z2[(i+2)%3][i] -=2;
	    kk[l] = i ;
	    l++;
	  }
	
      
  }

for ( i = 0 ;i < N; i++)
  {
    for ( k = 0;k<N ; k++)
      {
	if ( (EV[i] == 0) && (k != kk[1]) && 
	    (k != kk[0]))
	  { Z1[(k+1)%3][i] -= 2;Z2[(k+2)%3][i] -=2;}
	
      }
  }


/*******************trajectory of a point ***************/

if (traj)
{
  PtList = (liste) malloc(sizeof(struct LISTE));
  
  PtList->x = Dep[0];
  PtList->y = Dep[1];
  PtList->z = Dep[2];
  PtList->next = NULL;
  
  PtCur = PtList;
  
  prodMV(A,Dep,Nxt);
  
  while (!in(PtList,Nxt))
    {
      Dep[0] = Nxt[0];
      Dep[1] = Nxt[1];
      Dep[2] = Nxt[2];
      
      PtNew = (liste) malloc(sizeof(struct LISTE));

      PtCur->next = PtNew;
      
      PtNew->x = Dep[0];
      PtNew->y = Dep[1];
      PtNew->z = Dep[2];
      PtNew->next = NULL;
     
      PtCur = PtNew;
      
      prodMV(A,Dep,Nxt);
    }
}
  
  



/************** projections *******************/


for ( i = 0 ; i < N ; i++)
  {
    AsP[0][i] = floor( (As[0][i]*d1[0] +As[1][i]*d2[0] +As[2][i]*d3[0])*Ratio + x0*Ratio) ;
    AsP[1][i] = floor(-(As[0][i]*d1[1] +As[1][i]*d2[1] +As[2][i]*d3[1])*Ratio + y0*Ratio) ;
  }

for ( i = 0 ; i < N ; i++)
  {
    Z1P[0][i] = floor( (Z1[0][i]*d1[0] +Z1[1][i]*d2[0] +Z1[2][i]*d3[0])*Ratio + x0*Ratio) ;
    Z1P[1][i] = floor(-(Z1[0][i]*d1[1] +Z1[1][i]*d2[1] +Z1[2][i]*d3[1])*Ratio + y0*Ratio) ;
  }
for ( i = 0 ; i < N ; i++)
  {
    Z2P[0][i] = floor( (Z2[0][i]*d1[0] +Z2[1][i]*d2[0] +Z2[2][i]*d3[0])*Ratio + x0*Ratio) ;
    Z2P[1][i] = floor(-(Z2[0][i]*d1[1] +Z2[1][i]*d2[1] +Z2[2][i]*d3[1])*Ratio + y0*Ratio) ;
  }

for ( i = 0 ; i < N ; i++)
  {
    Z3P[0][i] = floor( (Z3[0][i]*d1[0] +Z3[1][i]*d2[0] +Z3[2][i]*d3[0])*Ratio + x0*Ratio) ;
    Z3P[1][i] = floor(-(Z3[0][i]*d1[1] +Z3[1][i]*d2[1] +Z3[2][i]*d3[1])*Ratio + y0*Ratio) ;
  }

for ( i = 0 ; i < N ; i++)
  {
    PP[0][i] = floor( (P[0][i]*d1[0] +P[1][i]*d2[0] +P[2][i]*d3[0])*Ratio + x0*Ratio) ;
    PP[1][i] = floor( -(P[0][i]*d1[1] +P[1][i]*d2[1] +P[2][i]*d3[1])*Ratio + y0*Ratio);
  }


if (vect1) 
  {
    V1P[0] = floor( (V1[0]*d1[0] +V1[1]*d2[0] +V1[2]*d3[0])*Ratio + x0*Ratio) ;
    V1P[1] = floor(-(V1[0]*d1[1] +V1[1]*d2[1] +V1[2]*d3[1])*Ratio + y0*Ratio) ;
  }
if (vect2) 
  {
    V2P[0] = floor( (V2[0]*d1[0] +V2[1]*d2[0] +V2[2]*d3[0])*Ratio + x0*Ratio) ;
    V2P[1] = floor(-(V2[0]*d1[1] +V2[1]*d2[1] +V2[2]*d3[1])*Ratio + y0*Ratio) ;
  }
if (vect3) 
  {
    V3P[0] = floor( (V3[0]*d1[0] +V3[1]*d2[0] +V3[2]*d3[0])*Ratio + x0*Ratio) ;
    V3P[1] = floor(-(V3[0]*d1[1] +V3[1]*d2[1] +V3[2]*d3[1])*Ratio + y0*Ratio) ;
  }


if (traj)
{

  numberPoints = 1;
  PtListP = (listeP) malloc(sizeof(struct LISTEP));

  PtListP->x = floor( ((PtList->x)*d1[0] +(PtList->y)*d2[0] +(PtList->z)*d3[0])*Ratio + x0*Ratio) ;
  PtListP->y = floor( -((PtList->x)*d1[1] +(PtList->y)*d2[1] +(PtList->z)*d3[1])*Ratio + y0*Ratio) ;
  PtListP->next = NULL;


  PtCur = PtList->next;
  PtCurP = PtListP;

  while (PtCur !=  NULL)
    {
      numberPoints++;
      PtNewP = (listeP) malloc(sizeof(struct LISTEP));
      PtCurP->next  = PtNewP;

      PtNewP->x = floor( ((PtCur->x)*d1[0] +(PtCur->y)*d2[0] +(PtCur->z)*d3[0])*Ratio + x0*Ratio) ;
      PtNewP->y = floor( -((PtCur->x)*d1[1] +(PtCur->y)*d2[1] +(PtCur->z)*d3[1])*Ratio + y0*Ratio) ;
      PtNewP->next = NULL;
  
      PtCur = PtCur->next;
      PtCurP = PtNewP;
    }
      
}



/***************** axis ******************/
centerx = x0*Ratio;
centery = y0*Ratio;

axis1x = floor(d1[0]*Ratio + x0*Ratio);
axis1y = floor(-d1[1]*Ratio + y0*Ratio);
axis2x = floor(d2[0]*Ratio + x0*Ratio);
axis2y = floor(-d2[1]*Ratio + y0*Ratio);
axis3x = floor(d3[0]*Ratio + x0*Ratio);
axis3y = floor(-d3[1]*Ratio + y0*Ratio);





/***************** construction of the fig file *****************/

  if  ( strlen(Out_File_Name) != 0 ) {
    if (append) {
      if ( NULL == freopen( Out_File_Name, "a", stdout )) {

	fprintf(stderr,"cannot open file %s\n", Out_File_Name);
	return -2;
      }
    }
    else {
      if ( NULL == freopen( Out_File_Name, "w", stdout )) {

	fprintf(stderr,"cannot open file %s\n", Out_File_Name);
	return -2;
      }
    }
  }

  if ( Fig_Version == 2 ) {
    Scale = 1;

    if (!append) {
      printf( "#FIG 2.1\n");
      printf( "80 2\n");
    } 

    printf( "2 1 0 1 -1 0 0 0 0.000 -1 1 0\n");
    printf( "\t0 0 1.000 4.000 8.000\n");
    printf( "\t%d %d %d %d 9999 9999 \n",
	    centerx*Scale,centery*Scale,axis1x*Scale,axis1y*Scale);
	 
    printf( "2 1 0 1 -1 0 0 0 0.000 -1 1 0\n 0 0 1.000 4.000 8.000\n");
    printf( "\t%d %d %d %d 9999 9999 \n",
	    centerx*Scale,centery*Scale,axis2x*Scale,axis2y*Scale);

    printf( "2 1 0 1 -1 0 0 0 0.000 -1 1 0\n 0 0 1.000 4.000 8.000\n");
    printf( "\t%d %d %d %d 9999 9999 \n",
	    centerx*Scale,centery*Scale,axis3x*Scale,axis3y*Scale);

    if ( scs + cyc != 2) {

      if (cyc == 1 ) width = 3; else width = 1;

      for (i = 0 ; i< N; i++) {
	for (j = i+1 ; j< N; j++) {
	  if ((EV[i] == 1) && (EV[j] == 1) ) {
	    printf( "2 1 0 %d -1 0 0 0 0.000 -1 0 0\n", width );
	    printf( "\t%d %d %d %d %d %d 9999 9999\n",
		    AsP[0][i]*Scale, AsP[1][i]*Scale,
		    PP[0][i+j-1]*Scale, PP[1][i+j-1]*Scale,
		    AsP[0][j]*Scale, AsP[1][j]*Scale);
	  }
	  else {
	    printf( "2 1 1 1 -1 0 0 0 4.000 -1 0 0\n");
	    printf( "\t%d %d %d %d %d %d 9999 9999 \n",
		    AsP[0][i]*Scale, AsP[1][i]*Scale,
		    PP[0][i+j-1]*Scale, PP[1][i+j-1]*Scale,
		    AsP[0][j]*Scale, AsP[1][j]*Scale);
	  }
	}
      }

      for (i = 0 ; i< N; i++) {
	printf( "2 1 1 1 -1 2 0 3 0.000 -1 0 0\n");
	printf( "\t%d %d %d %d %d %d 9999 9999\n",
		Z1P[0][i]*Scale, Z1P[1][i]*Scale,
		Z3P[0][i]*Scale, Z3P[1][i]*Scale,
		Z2P[0][i]*Scale, Z2P[1][i]*Scale);
      }
    }

    if (vect1) {
      printf( "1 3 0 1 -1 0 0 21 0.00000 1 0.000 %d %d %d %d %d %d %d %d \n",
	      V1P[0],V1P[1],3,3,V1P[0],V1P[1],V1P[0]+3,V1P[1]+3);
      printf( "4 0 0 12 0 -1 0 0.00000 4 11 37 %d %d (%3.1f, %3.1f, %3.1f) \n",
	      V1P[0]+5,V1P[1],V1[0],V1[1],V1[2]);
    }
  
    if (vect2) {
      printf( "1 3 0 1 -1 0 0 21 0.00000 1 0.000 %d %d %d %d %d %d %d %d \n",
	      V2P[0],V2P[1],3,3,V2P[0],V2P[1],V2P[0]+3,V2P[1]+3);
      printf( "4 0 0 12 0 -1 0 0.00000 4 11 37 %d %d (%3.1f, %3.1f, %3.1f) \n",
	      V2P[0]+5,V2P[1],V2[0],V2[1],V2[2]);
    }

    if (vect3) {
      printf( "1 3 0 1 -1 0 0 21 0.00000 1 0.000 %d %d %d %d %d %d %d %d \n",
	      V3P[0],V3P[1],3,3,V3P[0],V3P[1],V3P[0]+3,V3P[1]+3);
      printf( "4 0 0 12 0 -1 0 0.00000 4 11 37 %d %d (%3.1f, %3.1f, %3.1f) \n",
	      V3P[0]+5,V3P[1], V3[0],V3[1],V3[2]);
    }

    if ( vect1 && vect2 && cyc == 2 && scs ==2) {
      printf( "2 1 0 3 -1 2 0 3 0.000 -1 0 0\n" );
      printf( "\t%d %d %d %d 9999 9999\n",
	      V1P[0]*Scale,V1P[1]*Scale,V2P[0]*Scale,V2P[1]*Scale);
    }


    if (traj)
      {
	printf( "2 1 0 1 -1 0 0 0 0.000 -1 1 0\n 0 0 1.000 4.000 8.000\n\t");
	
	for(PtCurP = PtListP; PtCurP != NULL ; PtCurP = PtCurP->next)
	  {
	    printf( "%d %d ", (PtCurP->x)*Scale, (PtCurP->y)*Scale);
	  }
	printf( "9999 9999\n");
      }
  }
  else {
    /* Code generation for xfig 3.1 */
    Scale = 15;

    if (!append) {
      printf( "#FIG 3.1\n" );
      printf( "Landscape\n" );
      printf( "Center\n" );
      printf( "Metric\n" );
      printf( "1200 2\n");
    }

    printf( "2 1 0 1 -1 -1 0 0 -1 0.000 0 0 -1 1 0 2\n");
    printf( "\t0 0 1.000 75.00 135.00\n");
    printf( "\t%d %d %d %d\n",
	    centerx*Scale,centery*Scale,axis1x*Scale,axis1y*Scale);
	 
    printf( "2 1 0 1 -1 -1 0 0 -1 0.000 0 0 -1 1 0 2\n");
    printf( "\t0 0 1.000 75.00 135.00\n");
    printf( "\t%d %d %d %d\n",
	    centerx*Scale,centery*Scale,axis2x*Scale,axis2y*Scale);

    printf( "2 1 0 1 -1 -1 0 0 -1 0.000 0 0 -1 1 0 2\n");
    printf( "\t0 0 1.000 75.00 135.00\n");
    printf( "\t%d %d %d %d\n",
	    centerx*Scale,centery*Scale,axis3x*Scale,axis3y*Scale);

    if ( scs + cyc != 2) {

      if (cyc == 1 ) width = 3; else width = 1;

      for (i = 0 ; i< N; i++) {
	for (j = i+1 ; j< N; j++) {
	  if ((EV[i] == 1) && (EV[j] == 1) ) {
	    printf( "2 1 0 %d -1 -1 0 0 -1 0.000 0 0 -1 0 0 3\n",
		    width );
	    printf( "\t%d %d %d %d %d %d\n",
		    AsP[0][i]*Scale, AsP[1][i]*Scale,
		    PP[0][i+j-1]*Scale, PP[1][i+j-1]*Scale,
		    AsP[0][j]*Scale, AsP[1][j]*Scale);
	  }
	  else {
	    printf( "2 1 1 1 -1 -1 0 0 -1 4.000 0 0 -1 0 0 3\n");
	    printf( "\t%d %d %d %d %d %d\n",
		    AsP[0][i]*Scale, AsP[1][i]*Scale,
		    PP[0][i+j-1]*Scale, PP[1][i+j-1]*Scale,
		    AsP[0][j]*Scale, AsP[1][j]*Scale);
	  }
	}
      }

      for (i = 0 ; i< N; i++) {
	printf( "2 1 1 1 -1 -1 2 0 3 0.000 0 0 -1 0 0 3\n");
	printf( "\t%d %d %d %d %d %d\n",
		Z1P[0][i]*Scale, Z1P[1][i]*Scale,
		Z3P[0][i]*Scale, Z3P[1][i]*Scale,
		Z2P[0][i]*Scale, Z2P[1][i]*Scale);
      }
    }

    if (vect1) {
      printf( "1 3 0 1 -1 -1 0 0 20 0.0000 1 0.0000 ");
      printf( "%d %d %d %d %d %d %d %d\n",
	      V1P[0]*Scale,V1P[1]*Scale,3*Scale,3*Scale,
	      V1P[0]*Scale,V1P[1]*Scale,
	      (V1P[0]+3)*Scale,(V1P[1]+3)*Scale);
      printf( "4 0 -1 0 0 0 12 0.0000 4 150 1300 %d %d ",
	      Scale*(V1P[0]+5), Scale*(V1P[1]) );
      printf( "(%3.1f, %3.1f, %3.1f)\\001\n",
	      V1[0], V1[1], V1[2] );
    }
  
    if (vect2) {
      printf( "1 3 0 1 -1 -1 0 0 20 0.0000 1 0.0000 ");
      printf( "%d %d %d %d %d %d %d %d\n",
	      V2P[0]*Scale,V2P[1]*Scale,3*Scale,3*Scale,
	      V2P[0]*Scale,V2P[1]*Scale,
	      (V2P[0]+3)*Scale,(V2P[1]+3)*Scale);
      printf( "4 0 -1 0 0 0 12 0.0000 4 150 1300 %d %d ",
	      Scale*(V2P[0]+5), Scale*(V2P[1]) );
      printf( "(%3.1f, %3.1f, %3.1f)\\001\n",
	      V2[0], V2[1], V2[2] );
    }

    if (vect3) {
      printf( "1 3 0 1 -1 -1 0 0 20 0.0000 1 0.0000 ");
      printf( "%d %d %d %d %d %d %d %d\n",
	      V3P[0]*Scale,V3P[1]*Scale,3*Scale,3*Scale,
	      V3P[0]*Scale,V3P[1]*Scale,
	      (V3P[0]+3)*Scale,(V3P[1]+3)*Scale);
      printf( "4 0 -1 0 0 0 12 0.0000 4 150 1300 %d %d ",
	      Scale*(V3P[0]+5), Scale*(V3P[1]) );
      printf( "(%3.1f, %3.1f, %3.1f)\\001\n",
	      V3[0], V3[1], V3[2] );
    }

    if ( vect1 && vect2 && cyc == 2 && scs ==2) {
      printf( "2 1 0 3 -1 -1 2 0 3 0.000 0 0 -1 0 0 2\n" );
      printf( "\t%d %d %d %d\n",
	      V1P[0]*Scale,V1P[1]*Scale,V2P[0]*Scale,V2P[1]*Scale);
    }

    if (traj)
      {
	printf( "2 1 0 1 -1 -1 0 0 -1 0.000 0 0 -1 1 0 %d\n", numberPoints);
	printf( "\t0 0 1.000 75.00 135.00\n");
	for(PtCurP = PtListP; PtCurP != NULL ; PtCurP = PtCurP->next)
	  {
	    printf( "%d %d ", (PtCurP->x)*Scale, (PtCurP->y)*Scale);
	  }
	printf( "\n");
      }
  }

  return 0;

  }
  else if ( Help ) {
    return 0;
  } 
  else { 
    /* bad args */
    return -1;
  }
}
