/* programme/module de generation d'un programme C effectuant		*/
/* la simulation equationelle d'un graphe d'interaction.		*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "boolean.h"
#include "parseargs.h"
#include "i_graph.h"

#define	TOOLNAME	"gr2c"
#define VERSION		"1.1"

#define max(x,y)        ( (x>y) ? x : y )
#define min(x,y)        ( (x<y) ? x : y )

#define IND		make_indent();
#define CODE(x)		{IND;printf(x);}
#define CODE1(x,y)	{IND;printf(x,y);}
#define CODE2(x,y,z)	{IND;printf(x,y,z);}
#define CODE3(x,y,z,t)	{IND;printf(x,y,z,t);}
#define CODELN		CODE("\n")

#define MAX_NAME_LENGTH	256

/* Global declarations */
extern char	*todays_date( void );

/* Forward declarations */
void sort_probas(int*,int);
void gen_c_code(char File_Name[MAX_NAME_LENGTH],Interaction_Graph A,int);
void make_indent(void);

boolean	Debug = FALSE;

int	indent;

int main(int argc, char *argv[] )
{
  static char	In_File_Name[MAX_NAME_LENGTH] = "";
  static char	Out_File_Name[MAX_NAME_LENGTH] = "";

  int		i;
  int		Sim_Length = 100000;
  boolean	Help = FALSE;
  boolean	Args_Ok = TRUE;
  Interaction_Graph G;

  i = 1;

  if (argc>1)
    do {
      if GET_STRING( "-f", In_File_Name )

      else if GET_STRING( "-o", Out_File_Name )

      else if GET_INT( "-nevt", Sim_Length )

      else if SWITCH( "-debug", Debug, TRUE )
		       
      else if IS_VERSION

      else if ( ( 0 == strcmp( "-h", argv[i] ) )
		|| ( 0 == strcmp( "-?", argv[i] ) ) ) {
	printf("Usage: %s [-f <file>] [-o <file>]", TOOLNAME );
	printf(" [-nevt <int>] [-h] [-?]\n");
	Args_Ok = FALSE;
	Help = TRUE;
	i++;
      }
      else
	i++; /* skip unknown argument */
    } while ( Args_Ok && ( i < argc ) );

  if (Args_Ok) {
    if ( strlen(In_File_Name) != 0 )
      freopen( In_File_Name, "r", stdin );

    G = Read_Interaction_Graph( stdin );
    if ( (Interaction_Graph)NULL != G ) {

      if ( Debug )
	Write_Boolean_Matrix( G->A, stdout );

      sort_probas( G->Permut, G->Size );
      gen_c_code( Out_File_Name, G, Sim_Length );

      Free_Interaction_Graph( G );

      return 0;
    }
    else {
      fprintf( stderr, "%s: bad input data. No code generation.\n", 
	       TOOLNAME );
      return -2;
    }
  }
  else if ( Help )
    return 0;
  else
    return -1;
}

/* Procedure to sort compute a good permutation in order to fasten	*/
/* the drawing of random matrices. Not used at the moment.		*/
void	sort_probas(int *Permut, int Nb_Nodes)
{
  int	i;

  for ( i=0; i<Nb_Nodes; i++ )
    Permut[i] = i;
}

void	gen_file_header(int Sim_Length,
			int Nb_Nodes )
{
  char *The_Date;

  CODE("/* Simulation code for interaction graphs generated by ");
  CODE(TOOLNAME); CODE(" version "); CODE(VERSION);
  CODE("\t\t*/\n");
  The_Date = todays_date();
  CODE1("/* Date: %s \t\t\t\t*/\n\n", The_Date );
  free( The_Date );

  /* write definitions, includes */
  CODE("#include <stdio.h>\n");
  CODE("#include <string.h>\n");
  CODELN;
  CODE1("#define TMAX\t\t%d\n", Sim_Length );
  CODE1("#define NB_NODES\t%d\n", Nb_Nodes );
  CODE("#define EPSILON\t\t-1e28\n" );
  CODE("#define MAX_RANDOM_VALUE\t\t0x7FFFFFFF\n" );
  CODE("#define MAX(x,y)\t( x > y ? x : y )\n");
  CODELN;
  CODE("extern long random();\n");
  CODELN;
  CODE("double U_0_1()\n");
  CODE("{\n  return (double)random()/MAX_RANDOM_VALUE;\n");
  CODE("}\n\n");
  CODELN;
}

void	gen_declarations(Interaction_Graph G)
{
  int		i;

  CODE("int\t\tTime;\n");
  CODE("int\t\tk;\n");
  CODE("double\t\tTmp;\n");
  CODE("double\t\tProba;\n");
  CODE("/* table of potentials */\n");
  CODE("double\t\tY[ NB_NODES ];\n");
  CODE("/* table of cumulated probabilities */\n");
  CODE("double\t\tPr_Cum[ NB_NODES ] = {\n");
  indent+=20;

  for ( i=0; i<G->Size-1; i++ ) {
    CODE1("%f,\n", G->Pr_Cum[i] );
  }
  CODE1("%f };\n", 1.0 );

  indent-=20;

}

void	gen_node_code(int 	The_Node, 
		      Interaction_Graph G)
{
  int	j;
  int	Nb_In;
  int	*Neighbors;
  int	Ind;

  Ind = indent;
  indent = 0;

  Neighbors = (int*) calloc( G->Size, sizeof(int) );

  CODE1("\t\t/* evolution of potential of node %d ", G->Permut[The_Node] ); 
  if ( Is_Adj( G->A, The_Node, The_Node ) ) {
    CODE("(recycled) */\n");
  }
  else {
    CODE("(not recycled) */\n");
  }

  /* counting the number of input arcs */
  Nb_In = 0;
  for ( j=0; j<G->Size; j++ ) 
    if ( Is_Adj( G->A, The_Node, j ) ) {
      Neighbors[Nb_In] = j;
      Nb_In++;
    }

  if ( Nb_In == 0 ) {
    CODE1("\t\tY[%d] = EPSILON;\n", The_Node );
  }
  else if ( Nb_In == 1 ) {
    CODE2("\t\tY[%d] = Y[%d] + 1;\n", The_Node, Neighbors[0] );
  }
  else {
    CODE ("\t\tTmp = MAX( ");
    CODE1("Y[%d], ", Neighbors[0] );
    CODE1("Y[%d] );\n ", Neighbors[1] );
    for ( j=2; j < Nb_In; j++ ) {
      CODE ("\t\tTmp = MAX( Tmp, ");
      CODE1("Y[%d] );\n ", Neighbors[j] );
    }
    CODE1("\t\tY[%d] = Tmp + 1;\n", The_Node );
  }

  CODE("\t\tbreak;\n\n");

  free( Neighbors );

  indent = Ind;
}

void	gen_loop_code(void)
{
  CODELN;
  CODE("int main( int argc, char *argv[] )\n{\n");
  indent +=2;

  CODE("int i = 1;\n");
  CODE("int Sim_Length = TMAX;\n");
  CODELN;
  CODE("if (argc>1)\n");
  indent +=2;
  CODE("do {\n");
  indent +=2;
  CODE("if ( !strcmp( \"-nevt\", argv[i] ) ) {\n");
  CODE("  i++;\n");
  CODE("  if ( i < argc ) {\n");
  CODE("    sscanf( argv[i], \"%%d\", &Sim_Length );\n");
  CODE("    i++;\n");
  CODE("  }\n");
  CODE("  else {\n");
  CODE("    printf( \"number expected after -nevt\\n\" );\n");
  CODE("    return -2;\n");
  CODE("  }\n");
  CODE("}\n");
  CODE("else if ( !strcmp ( argv[i], \"-h\" ) ) {\n");
  CODE("  printf( \"Usage: %%s [-nevt <int>] [-h]\\n\", argv[0] );\n");
  CODE("  return -2;\n");
  CODE("}\n");
  CODE("else\n");
  CODE("  i++;\n");
  indent -=2;
  CODE("} while ( i < argc );\n");
  indent -=2;
  CODELN;
  CODE("for ( Time = 0; Time < Sim_Length; Time ++ ) {\n\n");
  indent +=2;
  CODE("Proba = U_0_1();\n");
  CODE("for ( k=0; ( k < NB_NODES ) && ( Proba > Pr_Cum[k] ); k++ );\n");
  
  CODE("switch( k ) {\n");
  indent +=2;
}

void	gen_file_trailer(void)
{
  indent -= 2;
  CODE("} /* end of switch */\n");
  indent -=2;
  CODE("}; /* end of main loop */\n\n");

  CODE("  if ( Sim_Length > 0 )\n");
  CODE("    for( i=0; i<NB_NODES; i++ ) {\n");
  CODE("      printf( \"gamma[%%d] = \", i );\n");
  CODE("      if ( Y[i] == EPSILON )\n");
  CODE("        printf(\"eps\\n\");\n");
  CODE("      else\n");
  CODE("        printf(\"%%f\\n\", Y[i] / (double)Sim_Length );\n");
  CODE("    }\n");
  CODELN;

  CODE("return 0;\n");
  indent -=2;
  CODE("}\n");
}

void	gen_c_code(char 	Name[MAX_NAME_LENGTH],
		   Interaction_Graph	G,
		   int		Sim_Length )
{
  int 		i;
  boolean	File_Ok = TRUE;

  if ( strlen( Name ) > 0 )
    File_Ok = ( NULL != freopen( Name, "w", stdout ) );

  if ( File_Ok ) {
    gen_file_header( Sim_Length, G->Size);
    gen_declarations( G );
    gen_loop_code();

    for ( i=0; i<G->Size; i++ ) {
      CODE1("case %d:\n", i );
      indent +=2;
      gen_node_code(i, G );
      indent -=2;
    }

    gen_file_trailer();
  }
  else	fprintf( stderr, "%s: error opening file %s\n",
		 TOOLNAME, Name );
  
}

void make_indent( void )
{
  int	i;

  for (i=0; i<indent; i++)
    printf(" ");
}
