/***********************************************************************/
/* Conversion of an Interaction Graph into a Bernoulli matrix product. */
/***********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <math.h>
#include <string.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include "boolean.h"
#include "parseargs.h"
#include "maxpluscore.h"
#include "i_graph.h"

#define TOOLNAME	"gr2mat"
#define VERSION		"1.0"

#define MAX_RANDOM_VALUE        0x7FFFFFFF
#define infty 			-10000

#define MAX_NAME_LENGTH 256

boolean Debug = FALSE;

/***********************************************************************/
/*                Programme Principal                                  */
/***********************************************************************/

int main(int argc, char *argv[], char *argp[] )
{
  static char	In_File_Name[MAX_NAME_LENGTH] = "";
  static char	Out_File_Name[MAX_NAME_LENGTH] = "";

  int 		i;
  boolean	Help = FALSE;
  boolean	Args_Ok = TRUE;
  boolean	File_Ok = TRUE;
  int		Format = FORMAT_SPARSE;
  static char	Mat_Prefix[128] = "A";
  Interaction_Graph	G;
  Bernoulli_Product	BP;

  /******************* Parsing command line args *************************/
  i = 1;

  if (argc>1)
    do {

      if GET_STRING( "-f", In_File_Name )
      else if GET_STRING( "-o", Out_File_Name )
      else if GET_STRING( "-prefix", Mat_Prefix )
      else if SWITCH( "-debug", Debug, TRUE )
      else if IS_VERSION

      else if (( 0 == strcmp( "-h", argv[i] )) || 
	       ( 0 == strcmp( "-?", argv[i] )) ||
	       ( 0 == strcmp( "-help", argv[i] ))) {	
	i++;
	fprintf(stdout,"usage: %s [-f <file>] [-o <file>] ",
		TOOLNAME );
	fprintf(stdout," [-full] [-scilab] [-prefix <string>]");
	fprintf(stdout," [-v] [-h] [-help] [-?]\n");
	Args_Ok = FALSE;
	Help = TRUE;
      }
      else if ( ( 0 == strcmp( "-full", argv[i] ) ) ) {
	i++;
	Format = FORMAT_FULL;
      }
      else if ( ( 0 == strcmp( "-scilab", argv[i] ) ) ) {
	i++;
	Format = FORMAT_SCILAB;
      }
      else 
	/* skip unknown argument */
	i++;
    } while( Args_Ok && ( i < argc ) );

  if (Args_Ok) {

    if ( strlen(In_File_Name) != 0 )
      File_Ok = ( NULL != freopen( In_File_Name, "r", stdin ) );

    if ( !File_Ok ) {
      fprintf(stderr,"%s: could not open file %s\n", TOOLNAME, In_File_Name );
    }

    if ( File_Ok ) {
      if ( strlen(Out_File_Name) != 0 )
	File_Ok = (NULL != freopen( Out_File_Name, "w", stdout ) );

      if ( !File_Ok ) {
	fprintf(stderr,"%s: could not open file %s\n", 
		TOOLNAME, Out_File_Name);
	File_Ok = FALSE;
      }
    }
  } else if ( Help ) {
    return 0;
  } else {
    return -1;
  }

  /************* start of the program if everything is Ok *******************/

  if ( Args_Ok && File_Ok ) {

    /************* read the incidence matrix + prob vector ****************/
    G = Read_Interaction_Graph( stdin );
    /***************** compute synchronization matrices *****************/
    BP = IG_To_BP( G );

    Write_Bernoulli_Product( BP, stdout, Format, Mat_Prefix );

    return 0;
  }
  else if ( !Help ) {
    return -1;
  }
  else
    return 0;
}
