/************************************************************************/
/* Module for constructing a cycle basis for a graph			*/
/* Adapted from a program written by Miguel Canales and Bruno Gaujal, 	*/
/* January 1992.							*/
/*									*/
/* Takes as input a matrix A describing the node/vertex map of the 	*/
/* graph.								*/
/* Results in a matrix containing a description of the cycles.		*/
/* It is known that if the graph is strongly connected, with T nodes	*/
/* and P edges, then the cardinal of the base is: P-T+1.		*/
/************************************************************************/
/*									*/
/*  the codification for the construction of matrix A is as follows:	*/
/*									*/
/*    Aij = -1 if O--->|						*/
/*                i    j						*/
/*									*/
/*          +1 if |--->O						*/
/*                j    i						*/
/*									*/
/*           0 otherwise						*/
/************************************************************************/
#include <stdlib.h>

/* forward declaration */
int Cycle(int a, int b, int *cycle, int P, int T, int **St, int **Bf);

/* int St[P][T], A[P][T], m[T], Cf[P],Bf[P-T+1][P],l=0;	*/
int	l=0;

int	**Cycle_Base(
int	**A,		/* the topology matrix		*/
int	P,		/* number of edges/rows of A	*/
int	T )		/* number of nodes/columns of A	*/
{
  int	n;
  int   p;
  int	t;
  int	i;
  int	j;
  int	t1;
  int	t2;

  int	**St;		/* where the spanning tree is put	*/
  int	**Bf;		/* where the cycle base is put		*/

  int	*m;		/* ? */
  int	*Cf;		/* ? */

  /* Creation of the (P-T+1)xP matrix Bf */
  Bf = (int**) calloc( P-T+1, sizeof(int*) );
  for (i=0; i<P-T+1; i++)
    Bf[i] = (int*) calloc( P, sizeof(int) );

  /* Creation of the P x T matrix St */
  St = (int**) calloc( P, sizeof(int*) );
  for (i=0; i<P; i++)
    St[i] = (int*) calloc( T, sizeof(int) );

  /* creation of auxiliary matrices	*/
  m = (int*) calloc( T, sizeof(int) );
  Cf = (int*) calloc( P, sizeof(int) );

  /* Construction of the spanning tree */
  for(m[0]=n=1; n<T;)
    {
      for(t=0;t<T;t++)
	if( m[t] == 1 )
	  break;
      for(i=0;i<P;i++)
	{
	  if( A[i][t] == 0 )
	    continue;
	  for(j=0;j<T;j++)
	    if( A[i][j] * A[i][t] == -1 && m[j] == 0 )
	      {
		St[i][t] = A[i][t];
		St[i][j] = A[i][j];
		m[j] = 1;
		n++;
	      }
	}  
      m[t] = 2;
    }

  /* Construction of Bf */
  l=0;	/* counts the number of cycles already stored */

  for(p=0;p<P;p++)
    {
      for(t1=0;t1<T;t1++)
	if( St[p][t1] != A[p][t1] )
	  break;
      if( t1 >= T )
	continue;
      /* place p not in the St */
      for(t2=t1+1;t2<T;t2++)
	if( St[p][t2] != A[p][t2] )
	  break;
      Cf[p] = 1;
      if( A[p][t1] == 1 )
	Cycle(t2,t1,Cf,P,T,St,Bf);
      else
	Cycle(t1,t2,Cf,P,T,St,Bf);
      Cf[p] = 0;
    }

  return Bf;
}

int Cycle(int a, 
	  int b, 
	  int *cycle,
	  int	P,
	  int	T,
	  int	**St,
	  int	**Bf)
{
  int *local_cycle,p,t,rc=0;
  int	i;

  local_cycle = (int*) calloc( (size_t)P, sizeof(int) );

  for(p=0;p<P;p++)
    local_cycle[p] = cycle[p];
  for(p=0;p<P && !rc;p++)
    if( St[p][a] != 0 && local_cycle[p] == 0 ) {
	local_cycle[p] = St[p][a];
	for(t=0;t<T;t++)
	  if( St[p][t] * St[p][a] == -1 )
	    break;
	if( t == b ) {
	    for(i=0;i<P;i++)
	      Bf[l][i] = local_cycle[i];
	    l++;

	    free( local_cycle );
	    return 1;
	  }
	else
	  rc = Cycle(t,b,local_cycle,P,T,St,Bf);
	if( !rc )
	  local_cycle[p] = 0;
      }

  free( local_cycle );

  return rc;
}

