/****************************************************************/
/* module pour generer des reseaux ATM  dans le format PNL.	*/
/* buffers en entree et en sortie a chaque port de chaque switch*/
/* sessions avec priorite stricte selon leur ordre dans le      */
/* fichier de description                                       */
/****************************************************************/

#include <math.h>
#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include "law.h"

#define TOOLNAME	"atm2pn"

#ifndef _BOOLEAN_H
#define _BOOLEAN_H
typedef int boolean;

#define TRUE 1
#define FALSE 0
#endif

boolean Debug = FALSE;

int NbSession;
int NbSwitch;    

int   *capaIn;      /* input buffer capacity */
int   *capaOut;     /* output buffer capacity */


int   ***route;     /* routes of each session */
Law_Desc   *SourceLaw;   /* description of the source of each session */
int   **NumSw;      /* names of switches crossed by a session */
int   *NbThrough;   /* nb of switches crossed by a session */
int   *inBufferI;   /* indices of transitions in the buffer In the switch */
int   *outBufferI;
int   *inBufferO;
int   *outBufferO;
int   **inside;     /* inside the switch */

int   *SwitchSize; /* taille des switches */
int   *Layer;     /* nb of layers in a switch */
int   *Nsessions; /* nb of Nsessions in a switch*/
int   **NumSe;     /* names of the sessions in a switch */


FILE  *File;

/*************************************************/
/* une session dans un switch  = un fichier pnl  */
/*************************************************/
void  gen_session(OutFile, index, N)
FILE		*OutFile;
int             index;
int             N;
{
  int	j,jIn,jOut,countT,countP;



  fprintf(OutFile, "Transitions <- %d\n", 9+Layer[NumSw[index][N]]*2);
  fprintf(OutFile, "Places <- %d\n", 6+Layer[NumSw[index][N]]*2);

  /* source of a session through a leaky bucket */
  if (N == 0)
    {
      fprintf(OutFile,"Submodel 1 \"leaky-Bucket\" FILE \"session_leaky.pnl\";\n");
    }



  /* input transition */
  fprintf(OutFile, "T 1 \"input_transition \" C 1.0000 , 1 ;\n");
  countP=1; countT=2;


  /* input buffer of the switch */
  fprintf(OutFile, "P %d \"BufferIn\" C 0.0000 0 ", countP );
  fprintf(OutFile, "Competition Priority [%d %d],%d,%d;\n",
	  countT,countT+1,countT,countT+1);
  fprintf(OutFile, "T %d \"LossBufferIn\" C 0.0000 ;\n",countT+1);
  fprintf(OutFile, "T %d \"AcceptBufferIn\" C 0.0000 , %d;\n",countT,countP+1);
  inBufferI[index]=countT;
  countP++;
  countT=countT+2;
  fprintf(OutFile, "P %d \"WaitService0\" C 1.0000 0 , %d ;\n",countP,countT);
  fprintf(OutFile, "T %d \"BeginService0\" C 0.0000 , %d;\n",countT,countP+1);
  outBufferI[index]=countT;
  countP++;
  countT++;

  fprintf(OutFile, "P %d \"InService0\" C 0.0000 0 , %d ;\n",countP,countT);
  fprintf(OutFile, "T %d \"EndService0\" C 1.0000 , %d;\n",countT,countP+1);
 
  countP++;
  countT++;

  /* inside the switch, number of layer */

      
  for  (j=0;j<Layer[NumSw[index][N]];j++)
    {
      fprintf(OutFile, "P %d \"WaitService%d\" C 0.0000 0 , %d ;\n",
	      countP,j+1,countT);
      fprintf(OutFile, "T %d \"BeginService%d\" C 0.0000 , %d;\n",
	      countT,j+1,countP+1);
      inside[index][j]=countT;
      countP++;
      countT++;
      fprintf(OutFile, "P %d \"InService%d\" C 0.0000 0 , %d ;\n",
	      countP,j+1,countT);
      fprintf(OutFile, "T %d \"EndService%d\" C 1.0000 , %d;\n",
	      countT,j+1,countP+1);
	  
      countP++;
      countT++;
    }

  /*output buffer of the switch */

  fprintf(OutFile, "P %d \"RouteOut\" C 0.0000 0 Competition Priority [%d %d],%d,%d;\n",countP,countT,countT+1,countT,countT+1);

  fprintf(OutFile, "T %d \"LossBufferOut\"C 0.0000 ;\n",countT+1);
  fprintf(OutFile, "T %d \"AcceptBufferOut\" C 0.0000 , %d;\n",
	  countT,countP+1);
  inBufferO[index]=countT;
  countP++;
  countT=countT+2;
  fprintf(OutFile, "P %d \"WaitServiceFinal\" C 1.0000 0 , %d ;\n",
	  countP,countT);
  fprintf(OutFile, "T %d \"BeginServiceFinal\" C 0.0000 , %d;\n",
	  countT,countP+1);
  outBufferO[index]=countT;
  countP++;
  countT++;
  fprintf(OutFile, "P %d \"InServiceFinal\" C 0.0000 0 , %d ;\n",
	  countP,countT);
  fprintf(OutFile, "T %d \"OutputTransition \" C 1.0000 ;\n",countT);
      
  /* bindings */
  if ( N == 0 ) {
    fprintf(OutFile,"# Bindings\n#\n");
    fprintf(OutFile, "S1/OUT1, T1;\n");
  }

  /* exports */ 
  fprintf(OutFile,"# Exports\n#\n");
  fprintf(OutFile, "Export\n");
  if (N==0) 
    {
      fprintf(OutFile, "IN1 is S1/IN1; \n");
    }
  else
    {
      fprintf(OutFile, "IN%d is T%d;\n",1,1);
    }


  jIn=2;jOut=1;

      
  fprintf(OutFile, "IN%d is T%d;\n",jIn,2);jIn++;
  fprintf(OutFile, "IN%d is T%d;\n",jIn,4);jIn++;
  fprintf(OutFile, "IN%d is T%d;\n",jIn,2*Layer[NumSw[index][N]]+6);jIn++;
  fprintf(OutFile, "IN%d is T%d;\n",jIn,2*Layer[NumSw[index][N]]+8);jIn++;
      
  fprintf(OutFile, "OUT%d is T%d;\n",jOut,5);jOut++;
  fprintf(OutFile, "OUT%d is T%d;\n",jOut,2*Layer[NumSw[index][N]]+9);jOut++;
      
  for (j=0;j<Layer[NumSw[index][N]];j++)
    {
      fprintf(OutFile,"IN%d is T%d;\n",jIn,2*j+6);jIn++;
      fprintf(OutFile,"OUT%d is T%d;\n",jOut,2*j+7);jOut++;
    }




  /* link with next switch or exit*/


  /*  lien avec le sous modele sur la transition 1 (attention 1 est fixe)*/
  /*       if (N==0)  */
  /* 	{ */
  /* 	  fprintf(OutFile, "S1/OUT1, T1;\n\n"); */
  /* 	} */
  
  /*    les places et transitions visibles de l'exterieur */
  /*   fprintf(OutFile, "Export\n"); */
  /*   if (N==0)  */
  /*     { */
  /*       fprintf(OutFile, "IN1 is S1/IN1; \n"); */
  /*     } */
  /*   else */
  /*     { */
  /*       fprintf(OutFile, "IN%d is T%d;\n",1,1); */
  /*     } */


  /*   jIn=2;jOut=1; */

      
  /*       fprintf(OutFile, "IN%d is T%d;\n",jIn,inBufferI[index]);jIn++; */
  /*       fprintf(OutFile, "IN%d is T%d;\n",jIn,outBufferI[index]);jIn++; */
  /*       fprintf(OutFile, "IN%d is T%d;\n",jIn,inBufferO[index]);jIn++; */
  /*       fprintf(OutFile, "IN%d is T%d;\n",jIn,outBufferO[index]);jIn++; */
      
  /*       fprintf(OutFile, "OUT%d is T%d;\n",jOut,outBufferI[index]+1);jOut++; */
  /*       fprintf(OutFile, "OUT%d is T%d;\n",jOut,outBufferO[index]+1);jOut++; */
      
  /*       for (j=0;j<Layer[NumSw[index][N]];j++) */
  /* 	{ */
  /* 	  fprintf(OutFile,"IN%d is T%d;\n",jIn,inside[index][j]);jIn++; */
  /* 	  fprintf(OutFile,"OUT%d is T%d;\n",jOut,inside[index][j]+1);jOut++; */
  /* 	} */
            
}

/************************************************************************/
/* Procedure to write the law. Adapted from the same from pn2prosit.c	*/
/************************************************************************/
int		Seed1 = 308;
int		Seed2 = 725;

void print_prosit_law( OutFile, The_Law )
FILE		*OutFile;
Law_Desc	The_Law;
{

  switch( The_Law->Name ) {
  case 'E':
    fprintf( OutFile, "new NegativeExpntl( Scale * %8.4f, new MLCG(%d,%d) )", 
	   The_Law->Parameters[0], ++Seed1, ++Seed2 );
    break;
  case 'C':
    if ( The_Law->Parameters[0] == 0.0 )
      fprintf( OutFile, "Zero");
    else
      fprintf( OutFile, "new Cst( Scale * %8.4f )", The_Law->Parameters[0] );
    break;
  case 'B':
    /* Bernoulli(p) = Binomial( 1, p ) */
    fprintf( OutFile, "new Binomial( 1, %8.4f, new MLCG(%d,%d) )", 
 	   The_Law->Parameters[0], ++Seed1, ++Seed2 );
     break;
   case 'G':
     /* g++ Geometric is called with the parameter, called "mean" */
     fprintf( OutFile, "new Geometric( %8.4f, new MLCG(%d,%d) )",
 	   The_Law->Parameters[0], ++Seed1, ++Seed2 );
     break;
  case 'U':
    fprintf( OutFile, "new Uniform( Scale * %8.4f, Scale * %8.4f",
	   The_Law->Parameters[0], The_Law->Parameters[1] );
    fprintf( OutFile, ", new MLCG(%d,%d) )", 
	     ++Seed1, ++Seed2 );
    break;
  case 'R':
    /* Erlang is called with mean and variance */
    fprintf( OutFile, "new Erlang( Scale * %8.4f, Scale * Scale * %8.4f",
	     The_Law->Parameters[1],
	     The_Law->Parameters[1] * The_Law->Parameters[1] );
    fprintf( OutFile, ", new MLCG(%d,%d) )", 
	     ++Seed1, ++Seed2 );
    break;
  case 'D':
    fprintf( stderr, "%s: warning: unimplemented source law (%c).", 
	    TOOLNAME, The_Law->Name );
    fprintf( stderr, "(constant assumed)\n");
    fprintf( OutFile, "new Uniform( Scale * %8.4f, Scale * %8.4f",
	   The_Law->Parameters[1], The_Law->Parameters[1] );
    fprintf( OutFile, ", new MLCG(%d,%d) )", 
	     ++Seed1, ++Seed2 );
    break;
  case 'O':
    fprintf( OutFile, "new OnOff( Scale * %8.4f, %8.4f, %8.4f",
	   The_Law->Parameters[0], 
	   The_Law->Parameters[1], 
	   The_Law->Parameters[2] );
    fprintf( OutFile, ", new MLCG(%d,%d) )", 
	     ++Seed1, ++Seed2 );
    break;
  case 'H':
  case 'g':
    fprintf( stderr, "%s: error: can't handle source law (%c)\n", 
	    TOOLNAME, The_Law->Name );
    break;
  default:
    fprintf( stderr, "%s: error: unknown source law (%c)\n", 
	    TOOLNAME, The_Law->Name );
  }
}


/**************************************/
/* chaque switch  = un fichier pnl   */
/**************************************/
void gen_ATM(
FILE		*OutFile,
int             k,
boolean		AllFiles )
{

  int i,j,l,c,countP;
  int n;
  int *NbPlaceTot;
  NbPlaceTot = (int*) calloc(NbSwitch, sizeof(int));

  countP=0;

  /**** nb of places and transitions ****/

  fprintf(OutFile, "Transitions <- %d\n",0 );
  fprintf(OutFile, "Places <- %d\n",(4+Layer[k])*SwitchSize[k]);
  


  /**** all corresponding sessions are submodels ****/

  for (j=0;j<NbSession;j++)
    {
      for (n=0;n<NbThrough[j];n++)
	if (NumSw[j][n] == k) {
	  if ( AllFiles ) {
	    fprintf(OutFile,"Submodel %d \"session%d_%d\"  File \"session%d_%d.pnl\";\n",k*1000+j,j,k,j,k);
	  }
	  else {
	    fprintf(OutFile,"Submodel %d \"session%d_%d\"  ", k*1000+j,j,k );
	    fprintf(OutFile,"File \"session_%d%s.pnl\";\n",
		    Layer[k]+1,
		    ( n == 0 ? "+source" : "" ) );
	  }
	}
    }

  /**** generate all its  places ****/

  countP++;

  /*In Buffers In */
  for (i=0;i<SwitchSize[k];i++)
    {
      fprintf(OutFile, "P %d \"bufferIn\" C 0.0000 %d ",
 	      countP,capaIn[k]);
      fprintf(OutFile, "[ 0.0 REP %d ] ", capaIn[k] );
      countP++;
      fprintf(OutFile, "Competition Priority [" );
      for (j=0;j<NbSession;j++)
	{
	  if (route[j][k][0] == i+1)
	    {
		
	      fprintf(OutFile, "S%d/IN%d ",k*1000+j,2);

	    }
	      
	}
      fprintf(OutFile, "]");
	  
      for (j=0;j<NbSession;j++)
	{
	  if (route[j][k][0] == i+1)
	    {
	      fprintf(OutFile, ", S%d/IN%d ",k*1000+j,2);
	    }
	      
	}
      fprintf(OutFile, ";\n");

    }
      
  /*In Buffers Out */
  for (i=0;i<SwitchSize[k];i++)
    {
      fprintf(OutFile, "P %d \"serverIn\" C 0.0000 1 0.0 Competition Priority [",countP);
      countP++;
      for (j=0;j<NbSession;j++)
	{
	  if (route[j][k][0] == i+1)
	    {
	      fprintf(OutFile, "S%d/IN%d ",k*1000+j,3);
	    }
	      
	}
      fprintf(OutFile, "] ");
      for (j=0;j<NbSession;j++)
	{
	  if (route[j][k][0] == i+1)
	    {
	      fprintf(OutFile, ", S%d/IN%d ",k*1000+j,3);
	    }
	      
	}
      fprintf(OutFile, ";\n");
    }

  /*inside the switch */
  for (l=0;l<Layer[k];l++)
    {
      for (i=0;i<SwitchSize[k];i++)
	{
	  fprintf(OutFile,"P %d \"serv_%d_%d\" ", countP, l, i );
	  fprintf(OutFile,"C 0.0000 1 0.0 Competition Priority [" );
	  countP++;
	  for (j=0;j<NbSession;j++)
	    {
	      if (route[j][k][l+1] == i+1)
		{
		  fprintf(OutFile, "S%d/IN%d ",k*1000+j,l+6);
		}
		  
	    }
	  fprintf(OutFile, "] ");
	  for (j=0;j<NbSession;j++)
	    {
	      if (route[j][k][l+1] == i+1)
		{
		  fprintf(OutFile,", S%d/IN%d ",k*1000+j,l+6);
		}
		  
	    }
	  fprintf(OutFile, ";\n");

	}
    }

  /* out buffers in */
  for (i=0;i<SwitchSize[k];i++)
    {
      fprintf(OutFile,"P %d \"bufferOut\" C 0.0000 %d ",countP,capaOut[k]);
      fprintf(OutFile, "[ 0.0 REP %d ] ", capaOut[k] );      
      countP++;
      fprintf(OutFile,"Competition Priority [" );
      for (j=0;j<NbSession;j++)
	{
	  if (route[j][k][Layer[k]+1] == i+1)
	    {
	      fprintf(OutFile, "S%d/IN%d ",k*1000+j,4);
	    }
	      
	}
      fprintf(OutFile, "] ");
      for (j=0;j<NbSession;j++)
	{
	  if (route[j][k][Layer[k]+1] == i+1)
	    {
	      fprintf(OutFile, ", S%d/IN%d ",k*1000+j,4);
	    }
	      
	}
      fprintf(OutFile, ";\n");
    }
      
  /*out  Buffers Out */
  for (i=0;i<SwitchSize[k];i++)
    {
      fprintf(OutFile, "P %d \"serverOut\" C 0.0000 1 0.0 Competition Priority [",countP);
      countP++;
      for (j=0;j<NbSession;j++)
	{
	  if (route[j][k][Layer[k]+1] == i+1)
	    {
	      fprintf(OutFile, "S%d/IN%d ",k*1000+j,5);
	    }
	      
	}
      fprintf(OutFile, "] ");
      for (j=0;j<NbSession;j++)
	{
	  if (route[j][k][Layer[k]+1] == i+1)
	    {
	      fprintf(OutFile, ", S%d/IN%d ",k*1000+j,5);
	    }
	      
	}
      fprintf(OutFile, ";\n");
    }

       
      
  /**** links with the sessions ****/
      
  for (j=0;j<NbSession;j++)
    {
      for (n=0;n<NbThrough[j];n++)
	if (NumSw[j][n] == k)
	  {
	 
	    /* in-buffer out (capacity + server) */
	    fprintf(OutFile, "S%d/OUT%d , P %d , P %d ;\n",
		    k*1000+j,1,route[j][k][0],route[j][k][0]+SwitchSize[k]);


	    /* inside the switch */
	    for (l=1;l<Layer[k]+1;l++)
	      {
		fprintf(OutFile, "S%d/OUT%d , P %d;\n",
			k*1000+j,2+l,route[j][k][l]+(l+1)*SwitchSize[k]);
	      }
	      
	    /* out-buffer out (capacity + server) */
	    fprintf(OutFile, "S%d/OUT%d , P %d , P %d ;\n",
		    k*1000+j,2,
		    route[j][k][Layer[k]+1]+(Layer[k]+2)*SwitchSize[k],
		    route[j][k][Layer[k]+1]+(Layer[k]+3)*SwitchSize[k]);

	  }
    }
  
  fprintf(OutFile, "Export \n");

  /******** input of  the switch  ********/

  c=0;
  for (j=0;j<NbSession;j++)
    {
      for (n=0;n<NbThrough[j];n++)
	if (NumSw[j][n] == k)
	  {
	    c++;
	    fprintf(OutFile, "IN%d is S%d/IN%d ;\n",c,k*1000+j,1);
	  }
    }
	 

  /******** output of the switch  ********/
  c=0;
  for (j=0;j<NbSession;j++)
    {
      for (n=0;n<NbThrough[j];n++)
	if (NumSw[j][n] == k)
	  {
	    c++;
	    fprintf(OutFile, "OUT%d is S%d/OUT%d ;\n",c,k*1000+j,2);
	  }
    }

}





/*************************/
/* generate Prosit code  */
/*************************/

void gen_Prosit(FILE *OutFile, boolean Distributed, boolean Stats )
{
  int i,j,k,c,k1,k2;

  fprintf(OutFile, "//\n");
  fprintf(OutFile, "// PROSIT code generated by %s\t  */\n", TOOLNAME ); 
  fprintf(OutFile, "//\n");
  fprintf(OutFile, "const int INFINI = 100;\n");
  fprintf(OutFile, "const int DUMMY_CLASS = 0;\n");
  fprintf(OutFile, "#include <stdio.h>\n");
  fprintf(OutFile, "#include <iostream.h>\n");
  fprintf(OutFile, "#include <Uniform.h>\n");
  fprintf(OutFile, "#include <Geom.h>\n");
  fprintf(OutFile, "#include <Binomial.h>\n");
  fprintf(OutFile, "#include <NegExp.h>\n");
  fprintf(OutFile, "#include <Erlang.h>\n");
  fprintf(OutFile, "#include \"Cst.hh\"\n");
  fprintf(OutFile, "#include \"OnOff.hh\"\n");
  fprintf(OutFile, "#include <MLCG.h>\n");
  fprintf(OutFile, "#include <bool.h>\n");
  fprintf(OutFile, "#include <fstream.h>\n");

  if ( Distributed ) {
    fprintf(OutFile, "#include \"Cluster.hh\"\n");
    fprintf(OutFile, "#include \"Computer.hh\"\n");
    fprintf(OutFile, "#include \"System.hh\"\n");
    fprintf(OutFile, "#include \"Kernel.hh\"\n");
    fprintf(OutFile, "#include \"CM_Kernel.hh\"\n");
    fprintf(OutFile, "#include \"CM_Auto_Kernel.hh\"\n");
    fprintf(OutFile, "#include \"CM_DD_Kernel.hh\"\n");
    fprintf(OutFile, "#include \"Runtime.hh\"\n");
    fprintf(OutFile, "#include \"Proxy/PCM_Pastec_Sink.hh\"\n");
    fprintf(OutFile, "#include \"Proxy/PCM_Pastec_Server_Normal_FIFO.hh\"\n");
    fprintf(OutFile, "#include \"Proxy/PCM_Pastec_Source.hh\"\n");
  }
  else {
    fprintf(OutFile, "#include \"Pastec_Sink.hh\"\n");
    fprintf(OutFile, "#include \"Pastec_Routing.hh\"\n");
    fprintf(OutFile, "#include \"Pastec_Server_Normal_FIFO.hh\"\n");
    fprintf(OutFile, "//#include \"Pastec_Server_Normal_LIFO.hh\"\n");
    fprintf(OutFile, "//#include \"Pastec_Server_Normal_PRIORITY.hh\"\n");
    fprintf(OutFile, "#include \"Pastec_Source.hh\"\n");
    fprintf(OutFile, "#include \"Kernel.hh\"\n");
    fprintf(OutFile, "#include \"Pastec_Kernel.hh\"\n");
    fprintf(OutFile, "#ifdef PROFILE\n");
    fprintf(OutFile, "#include \"Pastec_PKernel.hh\"\n");
    fprintf(OutFile, "#endif\n");
    fprintf(OutFile, "#include \"Pastec_Client.hh\"\n");
    fprintf(OutFile, "#include \"Pastec_FS_Station.hh\"\n");
    fprintf(OutFile, "#include \"SM_Rule.hh\"\n");
    fprintf(OutFile, "#ifdef PROFILE\n");
    fprintf(OutFile, "Pastec_PKernel* pnoyau = new Pastec_PKernel;\n");
    fprintf(OutFile, "#else\n");
    fprintf(OutFile, "Pastec_Kernel* pnoyau = new Pastec_Kernel;\n");
    fprintf(OutFile, "#endif\n");
    fprintf(OutFile, "Kernel *noyau = pnoyau;\n");
  }

  fprintf(OutFile, "\n\n//files generated by pn2mixt \n");
  for (i=0;i<NbSwitch;i++)
    {
      if ( Distributed ) {
	fprintf(OutFile, "#include \"P_Switch%d.hh\"\n",i);
      }
      fprintf(OutFile, "#include \"Switch%d.hh\"\n",i);
    }

  fprintf(OutFile, "\n\n");
  fprintf(OutFile, "bool Trace;\n\n");

  /* tout le main */  
  if ( Distributed ) {
    fprintf(OutFile, "\n");
    fprintf(OutFile, "extern Kernel *noyau;\n");
    fprintf(OutFile, "\n");
    fprintf(OutFile, "void main_code(int argc, char **argv) {\n\n");
    fprintf(OutFile, "  //creation des clusters\n");
    for ( i=1; i<NbSwitch; i++ ) {
      fprintf(OutFile, "  Cluster *cl%d = new(\"host%d\") Cluster;\n",
	      i, i );
    }
    fprintf(OutFile, "  CM_Kernel::check_virtual_machine();\n");
    fprintf(OutFile, "\n");
    fprintf(OutFile, "  CM_Auto_Kernel *cm_noyau = new CM_Auto_Kernel;\n");
    fprintf(OutFile, "  noyau = cm_noyau;\n");
    fprintf(OutFile, "\n");
    if ( Stats ) {
      fprintf(OutFile, "  cm_noyau->init_stat();\n");
      fprintf(OutFile, "\n");
    }
    fprintf(OutFile, "  // set-up the topology\n");
    fprintf(OutFile, "  //====================\n");
    for ( i=0; i<NbSwitch; i++ )
      for ( j=0; j<NbSwitch; j++ )
	if ( i != j ) {
	  fprintf(OutFile, "  cm_noyau->set_succ(");
	  if ( i == 0 )
	    fprintf(OutFile, "Cluster::current(),");
	  else
	    fprintf(OutFile, "cl%d,", i );
	  if ( j == 0 )
	    fprintf(OutFile, "Cluster::current());\n");
	  else
	    fprintf(OutFile, "cl%d);\n", j );
	}
    fprintf(OutFile,"\n\n");
  }
  else
    fprintf(OutFile, "main(int argc, char *argv[]) {\n\n");

  fprintf(OutFile, "  int       i = 1;\n");
  fprintf(OutFile, "  int       j;\n");
  fprintf(OutFile, "  double    T_Max = 100.0;\n");
  fprintf(OutFile, "  double    Scale = 1.0;   // number multiplying interarrivals\n");
  fprintf(OutFile, "  int       Nb_Gen = 100;  // number of customers at the sources\n");
  fprintf(OutFile, "  bool      Is_Ok;\n\n");
  fprintf(OutFile, "  // analysis of command line arguments\n ");
  fprintf(OutFile, "  if ( argc > 1 )\n");
  fprintf(OutFile, "    do {\n ");
  fprintf(OutFile, "     if ( !strcmp ( argv[i], \"-trace\" ) ) {\n ");
  fprintf(OutFile, "       Trace = TRUE;\n  ");
  fprintf(OutFile, "      i++;\n      }\n ");
  fprintf(OutFile, "     else if ( !strcmp ( argv[i], \"-tmax\" ) ) {\n ");
  fprintf(OutFile, "       i++;\n ");
  fprintf(OutFile, "       if ( 0 == sscanf( argv[i], \"%%lf\", &T_Max ) ) {\n");
  fprintf(OutFile, "          cout << \"Bad argument for -tmax\" << endl;\n");
  fprintf(OutFile, "          exit(-1);\n ");
  fprintf(OutFile, "       }\n ");
  fprintf(OutFile, "       else {\n");
  fprintf(OutFile, "          i++;\n");
  fprintf(OutFile, "        }\n");
  fprintf(OutFile, "      }\n");
  fprintf(OutFile, "      else if ( !strcmp ( argv[i], \"-nb_gen\" ) ) {\n");
  fprintf(OutFile, "        i++;\n");
  fprintf(OutFile, "        if ( 0 == sscanf( argv[i], \"%%d\", &Nb_Gen ) ) {\n");
  fprintf(OutFile, "          cout << \"Bad argument for -nb_gen\" << endl;\n");
  fprintf(OutFile, "          exit(-1);\n");
  fprintf(OutFile, "        }\n");
  fprintf(OutFile, "        else {\n");
  fprintf(OutFile, "          i++;\n");
  fprintf(OutFile, "        }\n");
  fprintf(OutFile, "      }\n");
  fprintf(OutFile, "      else if ( !strcmp ( argv[i], \"-scale\" ) ) {\n");
  fprintf(OutFile, "        i++;\n");
  fprintf(OutFile, "        if ( 0 == sscanf( argv[i], \"%%lf\", &Scale ) ) {\n");
  fprintf(OutFile, "          cout << \"Bad argument for -scale\" << endl;\n");
  fprintf(OutFile, "          exit(-1);\n");
  fprintf(OutFile, "        }\n");
  fprintf(OutFile, "        else {\n");
  fprintf(OutFile, "          i++;\n");
  fprintf(OutFile, "        }\n");
  fprintf(OutFile, "      }\n");
  fprintf(OutFile, "      else if ( !strcmp ( argv[i], \"-h\" ) ) {\n");
  fprintf(OutFile, "        cout << \"Usage: \" << argv[0] << \" [-h] [-trace]\";\n");
  fprintf(OutFile, "        cout << \" [-tmax <real>] [-nb_gen <int>] [-scale <real>]\\n\";\n");
  fprintf(OutFile, "        exit(0);\n");
  fprintf(OutFile, "      }\n");
  fprintf(OutFile, "      else {\n");
  fprintf(OutFile, "        // ignore unknown option\n");
  fprintf(OutFile, "        i++;\n");
  fprintf(OutFile, "      }\n");
  fprintf(OutFile, "    } while ( i < argc );\n\n");

  if ( !Distributed ) {
    fprintf(OutFile, "#ifdef PROFILE\n");
    fprintf(OutFile, "  pnoyau->compute_profile();\n");
    fprintf(OutFile, "#endif\n");
    if ( Stats )
      fprintf(OutFile, "  pnoyau->init_stat();\n");
  }
  

  fprintf(OutFile, "  char Nom[128];\n\n");

  /* all the switches */
  fprintf(OutFile, "\n  // the switches \n\n");

  
  for (i=0;i<NbSwitch;i++)
    {
      fprintf(OutFile, "  %sSwitch%d *Sw_%d;\n",
	      ( Distributed ? "P_" : "" ), i, i );
      if ( Distributed ) {
	fprintf(OutFile, "  Sw_%d = new P_Switch%d(", i, i );
	if ( i == 0 )
	  fprintf(OutFile, "Cluster::current()");
	else
	  fprintf(OutFile, "cl%d", i );
	fprintf(OutFile, ");\n");
      }
      else {
	fprintf(OutFile, "  Sw_%d = new Switch%d( ", i, i );
	fprintf(OutFile, "/* act time*/0.0,/* step */1.0, ");
	fprintf(OutFile, "/* stack sz */ 0x8000 );\n");
      }
      fprintf(OutFile, "  sprintf( Nom,\"Switch_%d\");\n", i );
      fprintf(OutFile, "  Sw_%d->set_name( Nom );\n", i );
      fprintf(OutFile, "  Sw_%d->set_trace_status( Trace );\n", i );
      if ( Stats )
	fprintf(OutFile, "  Sw_%d->compute_stat();\n",i);
    }

  /*all the sources (one per session)*/
  fprintf(OutFile, "\n" );
  fprintf(OutFile, "  // Declaration of the  sources (one per session)\n\n");
  fprintf(OutFile, "  %sPastec_Source *Srce[%d];\n",
	  ( Distributed ? "PCM_" : "" ),
	  NbSession );
  
  for (i=0;i<NbSession;i++)
    {
      if ( Distributed ) {
	fprintf(OutFile, "  Srce[%d] = new PCM_Pastec_Source(", i );
	if ( NumSw[i][0] == 0 )
	  fprintf(OutFile, "Cluster::current());\n");
	else
	  fprintf(OutFile, "cl%d);\n", NumSw[i][0]  );
	fprintf(OutFile, "  Srce[%d]->set_distribution(", i );
	print_prosit_law( OutFile, SourceLaw[i] );
	fprintf(OutFile, ");\n");
      }
      else {
	fprintf(OutFile, "  Srce[%d] = new Pastec_Source(\n\t\t0.0,\n", i );
	fprintf(OutFile, "\t\t" );
	print_prosit_law( OutFile, SourceLaw[i] );
	fprintf(OutFile, ",\n\t\tNb_Gen );\n");
      }
      fprintf(OutFile, "  sprintf( Nom, \"Source_%d\");\n", i );
      fprintf(OutFile, "  Srce[%d]->set_name( Nom );\n", i );
      if ( Stats )
	fprintf(OutFile, "  Srce[%d]->compute_stat();\n", i );
      if ( Distributed ) {
	fprintf(OutFile, "  Srce[%d]->max_generation(Nb_Gen);\n", i );
      }
      fprintf(OutFile, "\n");
    }


  /* all the sinks (one per session ) */
  fprintf(OutFile, "\n  // Declaration of the  sinks (one per session)\n\n");
  fprintf(OutFile, "  %sPastec_Sink *Sink[%d];\n",
	  ( Distributed ? "PCM_" : "" ),
	  NbSession );
  for (i=0;i<NbSession;i++)
    {
      if ( Distributed ) {
	fprintf(OutFile, "  Sink[%d] = new PCM_Pastec_Sink(", i );
	if ( NumSw[i][NbThrough[i]-1] == 0 ) {
	  fprintf(OutFile, "Cluster::current());\n");
	}
	else
	  fprintf(OutFile, "cl%d);\n", NumSw[i][NbThrough[i]-1] );
      }
      else
	fprintf(OutFile, "  Sink[%d] = new Pastec_Sink( 0.0 );\n",i);
      fprintf(OutFile, "  sprintf( Nom, \"Sink_%%d\", %d);\n",i);
      fprintf(OutFile, "  Sink[%d]->set_name( Nom ); \n",i);
      if ( Stats )
	fprintf(OutFile, "  Sink[%d]->compute_stat();\n",i);
    }



  /* all the links between the switches */
  fprintf(OutFile, "\n  // the links \n\n");
  fprintf(OutFile, "  %sPastec_Server_Normal_FIFO *Link[%d][%d];\n",
	  ( Distributed ? "PCM_" : "" ),  
	  NbSession, NbSwitch );
  for (i=0;i<NbSession;i++)
    {
      for (j=0;j<NbThrough[i]-1;j++)
	{
	  /*ATTENTION tous les liens ne sont pas crees */
	  fprintf(OutFile, "  Link[%d][%d] = ", i, NumSw[i][j] );
	  if ( Distributed ) {
	    fprintf(OutFile, "new PCM_Pastec_Server_Normal_FIFO(");
	    if ( NumSw[i][j] == 0 ) {
	      fprintf(OutFile, "Cluster::current());\n");
	    }
	    else
	      fprintf(OutFile, "cl%d);\n", NumSw[i][j] );
	    fprintf(OutFile, "  Link[%d][%d]->set_nb_serveur( INFINI );\n", 
		    i, NumSw[i][j] );
	  }
	  else {
	    fprintf(OutFile, 
		    "new Pastec_Server_Normal_FIFO( 0.0, INFINI );\n" );
	  }
	  fprintf(OutFile, 
		  "  Link[%d][%d]->set_service( new Cst( 100.0 ), 1 );\n",
		  i, NumSw[i][j] );
	  if ( Distributed )
	    fprintf(OutFile, 
		    "  Link[%d][%d]->set_min_service_time( 100.0 );\n",
		    i, NumSw[i][j] );

	  fprintf(OutFile, "  sprintf( Nom, \"Link_%%d_%%d\", %d ,%d);\n",
		  i, NumSw[i][j] );
	  fprintf(OutFile, "  Link[%d][%d]->set_name( Nom );\n",
		  i, NumSw[i][j] );
	  if ( Stats ) 
	    fprintf(OutFile, "  Link[%d][%d]->compute_stat();\n",
		    i, NumSw[i][j] );
	}
    }



  /*** all the connections between the objects ***/
  fprintf(OutFile, "\n  // The connections  between the objects \n\n");

  /* with the sources */
  fprintf(OutFile, "  int Clas[%d];\n",NbSession);


  for (i=0;i<NbSession;i++)
    {
      k=NumSw[i][0];
      c=NumSe[k][i];

      fprintf(OutFile, "  Clas[%d] = Sw_%d->Register_Input( %d );\n",i,k,c+1);
      fprintf(OutFile, "  Srce[%d]->set_succ( Sw_%d, 1.0, \"\", \"\"  );\n",i,k);
      fprintf(OutFile, "  Srce[%d]->customers_class( Clas[%d] );\n",i,i);
    }


  /* in between the switches */
  fprintf(OutFile, "  int Cla[%d][%d];\n",NbSession,NbSwitch);

  for (i=0;i<NbSession;i++)
    {
      for (j=0;j<NbThrough[i]-1;j++) /* do not take the last switch */
	{
	  k1 = NumSw[i][j];
	  k2 = NumSw[i][j+1];
	    
	  fprintf(OutFile, "  Sw_%d->Register_Output( %d, Link[%d][%d] );\n",k1,NumSe[k1][i]+1,i,k1);
	  fprintf(OutFile, "  Cla[%d][%d] = Sw_%d->Register_Input(%d);\n",i,k1,k2,NumSe[k2][i]+1);
	  fprintf(OutFile, "  Link[%d][%d]->set_output_class( Cla[%d][%d] );\n",i,k1,i,k1);
	  fprintf(OutFile, "  Link[%d][%d]->set_succ(Sw_%d,1.0, \"\", \"\" );\n",i,k1,k2);
	}
    }

  /* with the sinks */

  for (i=0;i<NbSession;i++)
    {
      k=NumSw[i][NbThrough[i]-1];
      c=NumSe[k][i];

      fprintf(OutFile, "  Sw_%d->Register_Output( %d, Sink[%d] );\n",k,c+1,i);

    }


  /***** parametre de la simulation ********/

  fprintf(OutFile, "\n  // parameters of the simulation\n");
  if ( Distributed )
    fprintf(OutFile, "  cm_noyau->set_max_simulation_time( T_Max );\n");
  else {
    fprintf(OutFile, "  pnoyau->set_model_name(\"FooModel\");\n");
    fprintf(OutFile, "  pnoyau->set_starting_time(0.0);\n");
    fprintf(OutFile, "  pnoyau->set_ending_time( T_Max );\n");
    fprintf(OutFile, "  // pnoyau->set_animation();\n");
  }
  fprintf(OutFile, "\n\n");
  fprintf(OutFile, "  // computation of statistics\n\n");
  fprintf(OutFile, "  // the simulation itself\n");
  fprintf(OutFile, "  cout << \"Start of the simulation\" << endl;\n");
  fprintf(OutFile, "  noyau->print_registered();\n  noyau->run();\n");
  fprintf(OutFile, "  cout << \"End of the simulation\" << endl;\n");

  /******* display of the results *****/

  if ( Stats ) {
    fprintf(OutFile, "\n  // output of results\n");
    for (k=0;k<NbSwitch;k++) {
      fprintf(OutFile, "  Sw_%d->report_stat();\n", k );
    }

    for (i=0;i<NbSession;i++) {
      fprintf(OutFile, "  Srce[%d]->report_stat();\n",i);
      fprintf(OutFile, "  Sink[%d]->report_stat();\n",i);
    }

    for (i=0;i<NbSession;i++) {
      for (j=0;j<NbThrough[i]-1;j++) {
	k=NumSw[i][j];
	fprintf(OutFile, "  Link[%d][%d]->report_stat();\n",i,k);
      }
    }
  }

  fprintf(OutFile, "\n");
  if ( Distributed )
    fprintf(OutFile, "  cm_noyau->halt();\n");
  else {
    fprintf(OutFile, "#ifdef PROFILE\n");
    fprintf(OutFile, "  pnoyau->report_profile();\n");
    fprintf(OutFile, "#endif\n");
  }
  fprintf(OutFile, "\n");
  fprintf(OutFile, "  exit(0);\n");

  fprintf(OutFile, "}\n");

}

void gen_Top_Network(FILE *OutFile)
{
  int 	i;
  int 	j;
  int	Nb_Links;
  int	Num_Trans;
  int	Num_Place;
  int	The_Switch;
  int	*Input_Count;

  fprintf(OutFile, "#\n");
  fprintf(OutFile, "# PNL code generated by %s\t  */\n", TOOLNAME ); 
  fprintf(OutFile, "#\n");

  /* counting the number of transitions. There are NbSession sources,	*/
  /* NbSession sinks, and a varying number of links.			*/
  Nb_Links = 0;
  for ( i=0; i<NbSession; i++ )
    Nb_Links += ( NbThrough[i] - 1 );

  Input_Count = (int*) calloc( (unsigned)NbSwitch, sizeof(int) );

  fprintf(OutFile, "Transitions <- %d\n", 2 * NbSession + Nb_Links );
  fprintf(OutFile, "Places <- %d\n", 3 * NbSession + 2 * Nb_Links );

  fprintf(OutFile, "# the submodels are the switches.\n");
  for ( i=0; i<NbSwitch; i++ )
    fprintf(OutFile, "Submodel %2d \"Switch%d\" File \"switch%d.pnl\";\n",
	    i, i, i );
    
  Num_Trans = 1;
  Num_Place = 1;

  for ( i=0; i<NbSession; i++ ) {
    fprintf(OutFile, "# Session #%d\n", i );
    fprintf(OutFile, "T %2d \"Srce%d\" ", Num_Trans, i );
    Write_Law( OutFile, SourceLaw[i], NORMAL_PRINT_MODE );
    fprintf(OutFile, ", P %d, P %d;\n", Num_Place, Num_Place+1 );
    fprintf(OutFile, "P %2d \"RecycleSrce%d\" ", Num_Place, i );
    fprintf(OutFile, "C 0.0 1 0.0, T %d;\n", Num_Trans );
    Num_Place ++;
    fprintf(OutFile, "P %2d \"Srce%d_OutBuf\" ", Num_Place, i );
    fprintf(OutFile, "C 0.0 0, S%d/IN%d;\n", 
	    NumSw[i][0], Input_Count[NumSw[i][0]]+1 );
    Input_Count[NumSw[i][0]]++;
    Num_Place++;
    Num_Trans++;
    for ( j=1; j<NbThrough[i]; j++ ) {
      The_Switch = NumSw[i][j];
      fprintf(OutFile, "P %2d \"Link_%d_%d_InBuf\" ", 
	      Num_Place, i, The_Switch );
      fprintf(OutFile, "C 0.0 0, T %d;\n", Num_Trans );
      Num_Place ++;
      fprintf(OutFile, "T %2d \"Link_%d_%d\" ", Num_Trans, i, The_Switch );
      fprintf(OutFile, "C 0.0, P %d;\n", Num_Place );
      Num_Trans++;
      fprintf(OutFile, "P %2d \"Link_%d_%d_OutBuf\" ", 
	      Num_Place, i, The_Switch );
      fprintf(OutFile, "C 0.0 0, S%d/IN%d;\n", 
	      The_Switch, Input_Count[The_Switch]+1 );
      Num_Place ++;
      Input_Count[The_Switch]++;
    }
    fprintf(OutFile, "P %2d \"Sink%d_InBuf\" ", Num_Place, i );
    fprintf(OutFile, "C 0.0 0, T %d;\n", Num_Trans );
    Num_Place ++;
    fprintf(OutFile, "T %2d \"Sink%d\" ", Num_Trans, i );
    fprintf(OutFile, "C 0.0;\n");
    Num_Trans ++;
  }

  /* once again, for internal/external connections... */
  for ( i=0; i<NbSwitch; i++ )
    Input_Count[i] = 0;

  Num_Trans = 1;
  Num_Place = 1;

  fprintf(OutFile, "# Connections\n");
  for ( i=0; i<NbSession; i++ ) {
    Num_Trans ++;
    Num_Place += 2;
    fprintf(OutFile, "# For session %d:\n", i );
    for ( j=0; j<NbThrough[i]; j++ ) {
      The_Switch = NumSw[i][j];
      fprintf(OutFile, "S%d/OUT%d, P %d;\n",
	      The_Switch, Input_Count[The_Switch]+1, Num_Place );
      Input_Count[The_Switch] ++;
      Num_Trans ++;
      Num_Place += ( j == NbThrough[i] - 1 ? 1 : 2 );
    }
  }
}

void Gen_Distributed_Main( FILE *OutFile ) 
{
  int	i;

  fprintf(OutFile, "\n");
  fprintf(OutFile, "#include <iostream.h>\n");
  fprintf(OutFile, "#include <bool.h>\n");
  fprintf(OutFile, "\n");
  fprintf(OutFile, "#include \"Kernel.hh\"\n");
  fprintf(OutFile, "#include \"Runtime.hh\"\n");
  for ( i=0; i<NbSwitch; i++ )
    fprintf(OutFile, "#include \"Switch%d.hh\"\n", i );
  fprintf(OutFile, "#include \"Reconstructor_Comm_Object.hh\"\n");
  fprintf(OutFile, "\n");
  fprintf(OutFile, "// global data\n");
  fprintf(OutFile, "\n");
  fprintf(OutFile, "Kernel *noyau ;\n");
  fprintf(OutFile, "extern void bootstrap();\n");
  fprintf(OutFile, "extern void main_code(int,char **);\n");
  fprintf(OutFile, "extern void spawn_code();\n");
  fprintf(OutFile, "extern void endstrap();\n");
  fprintf(OutFile, "extern init_distributed_runtime();\n");
  fprintf(OutFile, "extern init_distributed_qnlib();\n");
  fprintf(OutFile, "extern init_distributed_examplelib();\n");
  fprintf(OutFile, "\n");
  fprintf(OutFile, "main (int argc, char**argv) {\n");
  fprintf(OutFile, "\n");
  fprintf(OutFile, "	Runtime::init(argc,argv,\"file_host.cnf\",argv[0]);\n");
  fprintf(OutFile, "\n");
  fprintf(OutFile, "\n");
  fprintf(OutFile, "	// Initialisation (communication)\n");
  fprintf(OutFile, "	// ==============================\n");
  fprintf(OutFile, "	init_distributed_runtime();\n");
  fprintf(OutFile, "	init_distributed_qnlib();\n");
  fprintf(OutFile, "	init_distributed_examplelib();\n");
  fprintf(OutFile, "\n");

  for ( i=0; i<NbSwitch; i++ ) {
    fprintf(OutFile, "\tRuntime::memo_comm_object()->store(\n");
    fprintf(OutFile, "\t\t\t\tnew String(\"Switch%d\"),\n", i );
    fprintf(OutFile, "\t\t\t\t&Switch%d::init_cstr);\n", i );
    fprintf(OutFile, "\n");
  }
  fprintf(OutFile, "\n");
  fprintf(OutFile, "	// lancement du programme\n");
  fprintf(OutFile, "	// ======================\n");
  fprintf(OutFile, "	if (Runtime::is_main_cluster())   {\n");
  fprintf(OutFile, "		bootstrap();\n");
  fprintf(OutFile, "		main_code(argc,argv);\n");
  fprintf(OutFile, "		endstrap();\n");
  fprintf(OutFile, "	}\n");
  fprintf(OutFile, "	else {\n");
  fprintf(OutFile, "		bootstrap();\n");
  fprintf(OutFile, "		spawn_code();\n");
  fprintf(OutFile, "		endstrap();\n");
  fprintf(OutFile, "	}\n");
  fprintf(OutFile, "}\n");
}

/************************************************************************/
/* A procedure for reading a stochastic law description and returning	*/
/* a structure describing it.						*/
/************************************************************************/
Law_Desc Get_Law( int Session_Nb )
{
  int	Nb_Params;
  int	i;
  Law_Desc	Res;
  char	Token[4];

  Res = (Law_Desc) malloc( sizeof( struct LAW_DESC ) );
  scanf( "%2s", Token );
  /* 
     printf("Got[%s]\n", Token );
     */
  if ( strlen(Token) > 0 ) {
    Res->Name = Token[0];
    switch( Res->Name ) {
    case 'E':
    case 'C':
    case 'B':
    case 'g':
    case 'G':
      Res->Parameters = (double*) malloc( sizeof(double) );
      scanf( "%lf", &(Res->Parameters[0] ) );
      break;
    case 'R':
    case 'U':
    case 'H':
      Res->Parameters = (double*) malloc ( 2*sizeof(double) );
      scanf( "%lf", &(Res->Parameters[0] ) );
      scanf( "%lf", &(Res->Parameters[1] ) );
      break;
    case 'O':
      Res->Parameters = (double*) malloc( 3*sizeof( double ) );
      scanf( "%lf", &(Res->Parameters[0]) );
      scanf( "%lf", &(Res->Parameters[1]) );
      scanf( "%lf", &(Res->Parameters[2]) );
      break;
    case 'D':
      /* find the length of the (first) parameter list */
      Nb_Params = 0;
      scanf( "%d", &Nb_Params );
      Res->Parameters = (double*) calloc( 1 + 2*Nb_Params, sizeof( double ) );
      Res->Parameters[0] = Nb_Params;
	
      /* Store values of the discrete distribution */
      for( i=1; i<=2*Nb_Params; i++ )
	scanf( "%lf", &(Res->Parameters[i]) );
	
      break;
	
    case 'P':
      /* IPP-type law. */
      Res->Parameters = (double*) calloc( 1, sizeof( double ) );
      scanf( "%lf", &(Res->Parameters[0]) );
      Res->Law_Parameters = (Law_Desc*) calloc( 2, sizeof(Law_Desc) );
      Res->Law_Parameters[0] = Get_Law( Session_Nb );
      Res->Law_Parameters[1] = Get_Law( Session_Nb );
      break;
	
    default:
      printf("%s: unknown law name: %c\n", TOOLNAME, Res->Name );
    }
  }
  else {
    printf("%s: Error reading law name for session number %d.\n",
	   TOOLNAME, Session_Nb );
    free( Res );
    Res = (Law_Desc)NULL;
  }

  return Res;

}

/**************************************/
/* programme principal                */
/**************************************/
int main( int argc, char *argv[] )
{

  char nomPnl[128]= "";	  /*nom des differents fichiers pnl */
  char nomProsit[128]="";   /* nom du fichier Prosit */
  char In_File_Name[128]= "";           /*nom du fichier d'entree */
      

  int  Args_OK=TRUE, InOut = TRUE;
  int i,j,k,J,exp2k;
  int PortIn,PortOut,Port;
  double aa;
  boolean	Distributed = FALSE;
  boolean	Stats = TRUE;
  boolean	AllFiles = FALSE;
  /* variables to handle generation with fewer files */
  int		Nb_Known_Files = 0;
  char		**Known_Files = (char**)NULL;
  boolean	Known;
  
  /************************* Parsing *****************************/
  i = 1;

  if (argc>1)
    do {

      if ( 0 == strcmp( "-f", argv[i] ) ) {
	i++;
	if ( i < argc ) {
	  strcpy( In_File_Name, argv[i] );
	  i++;
	}
	else {
	  fprintf(stderr,"%s: usage:filename expected after -f\n",argv[0]);
	  Args_OK = FALSE;
	}
      }
      else if ( 0 == strcmp( "-allfiles", argv[i] ) ) {
	i++;
	AllFiles = TRUE;
      }
      else if ( 0 == strcmp( "-no_stats", argv[i] ) ) {
	i++;
	Stats = FALSE;
      }
      else if ( 0 == strcmp( "-distributed", argv[i] ) ) {
	i++;
	Distributed = TRUE;
      }
      else if ( 0 == strcmp( "-debug", argv[i] ) ) {
	i++;
	Debug = TRUE;
      }
      else if (( 0 == strcmp( "-h", argv[i] )) || 
	       ( 0 == strcmp( "-?", argv[i] )) ||
	       ( 0 == strcmp( "-help", argv[i] ))) {
	i++;
	fprintf( stdout,
		 "usage: %s [-f inputfile] [-distributed] [-allfiles]",
		 TOOLNAME );
	fprintf( stdout, " [-no_stats]\n\t" );
	fprintf( stdout, 
		 "produces files \n\tsession0_0.pnl session)_1.pnl ,....\n\tswitch0.pnl,switch1.pnl,... \n\tand a file atm_network.cc. \n\tformat of the input file given in:\n\t/u/chin/0/sloop/gaujal/ATM2/exemple_config\n" );
	  Args_OK = FALSE;

	}
      else {
	/* skip silently unknown argument */
	i++;
      }
    } while( i < argc );

  if (Args_OK) {
    if  ( ( strlen(In_File_Name) != 0 ) &&
	  (NULL == freopen( In_File_Name, "r", stdin ) )) {
      fprintf(stderr,"%s: could not open file %s\n",argv[0], In_File_Name );
      InOut = FALSE;
    }
  }


  /************* start of the program if everything is OK ****************/

  if (Args_OK && InOut ) {

    /* read the input and alloc the arrays */
    scanf("%d",&NbSwitch);
    SwitchSize = (int *) calloc(NbSwitch,sizeof(int));
    Layer = (int *) calloc(NbSwitch,sizeof(int));
    capaIn = (int *) calloc(NbSwitch,sizeof(int));
    capaOut= (int *) calloc(NbSwitch,sizeof(int));

    for (i=0;i<NbSwitch;i++) {

      scanf("%d",&(SwitchSize[i]));
      scanf("%d",&(capaIn[i]));
      scanf("%d",&(capaOut[i]));

      /*  test power of 2 */
      for( j=1; j<=SwitchSize[i]; j*=2 );
      if ( (j/2) != SwitchSize[i] ) {
	printf("%s: Error: size of switch %d (=%d) is not",
	       TOOLNAME, i, SwitchSize[i] );
	printf(" a power of 2.\n");
      }

      aa=(double)SwitchSize[i];
      aa=log(aa)/log(2);
      aa=(int)rint(aa);
      Layer[i]=aa-1;
    }


    scanf("%d",&NbSession);

    SourceLaw = (Law_Desc*) calloc( NbSession, sizeof(Law_Desc) );

    NumSw = (int **) calloc(NbSession,sizeof(int*));
    for (i=0;i<NbSession;i++)
      {
	NumSw[i] = (int *) calloc(NbSwitch, sizeof(int));
      }

    NbThrough =(int *) calloc(NbSession, sizeof(int));

    inBufferI=(int *) calloc(NbSession, sizeof(int));

    outBufferI=(int *) calloc(NbSession, sizeof(int));

    inBufferO=(int *) calloc(NbSession, sizeof(int));

    outBufferO=(int *)calloc(NbSession, sizeof(int));

    inside = (int **) calloc(NbSession,sizeof(int*));
    for (i=0;i<NbSession;i++)
      {
	inside[i]=(int *) calloc(NbSwitch, sizeof(int*));
      }


    Nsessions =(int *) calloc(NbSwitch, sizeof(int));
    for (j=0;j<NbSwitch;j++)
      {
	Nsessions[j] = 0;
      }

    NumSe = (int **) calloc(NbSwitch,sizeof(int*));
    for (i=0;i<NbSwitch;i++)
      {
	NumSe[i] = (int *) calloc(NbSession, sizeof(int));
      }

    route = (int ***) calloc(NbSession,sizeof(int**));
    for (i=0;i<NbSession;i++)
      {
	route[i]=(int **) calloc(NbSwitch, sizeof(int*));
    
	for (j=0;j<NbSwitch;j++)
	  {
	    route[i][j]=(int*) calloc(Layer[j]+2, sizeof(int));
	  }
    
      }
    /* init the route array */
    for (i=0;i<NbSession;i++)
      {
	for (j=0;j<NbSwitch;j++)
	  {
	    for (k=0;k<Layer[j]+1;k++)
	      {
		route[i][j][k]=-1;
	      }
	  }
      }

    /* construct the route array, the NumSw and the NumSe arrays */
    for (i=0;i<NbSession;i++)
      {
	/* getting the type of source */
	SourceLaw[i] = Get_Law( i );

	scanf("%d",&NbThrough[i]);
	for (J=0;J<NbThrough[i];J++)
	  {
	    scanf("%d",&j);
	    NumSe[j-1][i] = Nsessions[j-1];
	    Nsessions[j-1]++;
	    NumSw[i][J]=j-1;
	    scanf("%d",&PortIn);
	    scanf("%d",&PortOut);

	    /* XXX test de compatibilite pas fait */

	    /* Sauvegarde de l'ancien algo, au cas ou...
	       route[i][j-1][0] = PortIn;
	       Port=PortIn;
	       for (k=0;k<Layer[j-1];k++)
	       {
	       exp2k=1<<k;
	       if ((PortOut-1) & exp2k)
	       {
	       Port = Port + (1-((Port-1) & exp2k))*exp2k;
	       }
	       else
	       {
	       Port = Port - ((Port-1) & exp2k)*exp2k;
	       }
	       route[i][j-1][k+1] = Port;
	       }
	    */

	    /* Nouvel algo */
	    route[i][j-1][0] = PortIn;
	    Port=PortIn-1;
	    for (k=0;k<Layer[j-1];k++)
	      {
		exp2k=1<<k;
		if ((PortOut-1) & exp2k) {
		  /* PortOut a un 1 dans le bit k */
		  if ( ! ( Port & exp2k ) ) {
		    /* Port a un 0: il faut mettre un 1 */
		    Port = Port + exp2k;
		  }
		}
		else {
		  /* PortOut a un 0 dans le bit k */
		  if ( Port & exp2k ) {
		    /* Port a un 1: il faut l'enlever */
		    Port = Port - exp2k;
		  }
		}
		route[i][j-1][k+1] = Port+1;
	      }
	    route[i][j-1][Layer[j-1]+1] = PortOut;
	  }
      }

    /***** build all the sessions *****/

    if ( Debug ) {
      printf("Routing table (Sw [ row row ... row ]):\n");
      for( i=0;i<NbSession;i++ ) {
	printf("Session %d:", i );
	for ( J=0; J<NbThrough[i]; J++ ) {
	  printf(" Sw %d[ ", NumSw[i][J] );
	  for ( k=0; k<=Layer[NumSw[i][J]]+1; k++ )
	    printf("%d ", route[i][NumSw[i][J]][k] );
	  printf("]");
	}
	printf("\n");
      }
    }
 
    for (i=0;i<NbSession;i++) {

      for (j=0;j<NbThrough[i];j++) {

	if ( AllFiles ) {
	  sprintf(nomPnl,"session%d_%d.pnl",i,NumSw[i][j]);
	  File = fopen(nomPnl, "w");
	  if (File == NULL)
	    {
	      fprintf(stderr,"%s: cannot open file %s\n",argv[0],nomPnl);
	    }
	  gen_session(File,i,j);
	  fclose(File);
	}
	else {
	  /* Check if this configuration has been generated */
	  sprintf( nomPnl, "session_%d%s.pnl", 
		   Layer[NumSw[i][j]]+1,
		   ( j == 0 ? "+source" : "" ) );
	  Known = FALSE;
	  for ( k=0; !Known && (k<Nb_Known_Files); k++ )
	    Known = ( 0 == strcmp( nomPnl, Known_Files[k] ) );
	  if ( !Known ) {
	    Nb_Known_Files++;
	    Known_Files = (char**)realloc( Known_Files,
					   (unsigned)Nb_Known_Files
					   *sizeof( char* ) );
	    Known_Files[ Nb_Known_Files-1 ] = strdup( nomPnl );

	    File = fopen(nomPnl, "w");
	    if (File == NULL) {
	      fprintf(stderr,"%s: cannot open file %s\n",argv[0],nomPnl);
	    }
	    else {
	      gen_session(File, i, j );
	      fclose(File);
	    }
	  }
	}
      }
    }


    /********* build the switches *****/
    for (i=0;i<NbSwitch;i++)
      {
	sprintf(nomPnl,"switch%d.pnl",i);
	File = fopen(nomPnl, "w");
	if (File == NULL)
	  {
	    fprintf(stderr,"%s: cannot open file %s\n",argv[0],nomPnl);
	  }
	gen_ATM(File, i, AllFiles);
	fclose(File);
      }

    /********* build the top network *****/
    sprintf(nomPnl,"top_network.pnl");
    File = fopen(nomPnl, "w");
    if (File == NULL) {
      fprintf(stderr,"%s: cannot open file %s\n",argv[0],nomPnl);
    }
    gen_Top_Network(File);
      

    /************ build the PROSIT program *************/

    sprintf(nomProsit, "atm_network.cc");
    File = fopen(nomProsit, "w");
    if (File == NULL)
      {
	fprintf(stderr,"%s: cannot open file %s\n",argv[0],nomProsit);
      }
    gen_Prosit( File, Distributed, Stats );
    fclose(File);

    if ( Distributed ) {
      sprintf(nomProsit, "main.cc");
      File = fopen(nomProsit, "w");
      if (File == NULL)
	{
	  fprintf(stderr,"%s: cannot open file %s\n",argv[0],nomProsit);
	}
      Gen_Distributed_Main( File );
      fclose(File);
    }

    return 0;

  }
  else
    return -1;

}
