/* Un programme d'analyse de la FIFOite d'un RdP general	*/

#include "pn.h"

#define	TOOLNAME	"fifo_pn"

/* Resources, mais par traitees en tant que tel... */
#define PN_BOUND	"pn_bound"
#define RES_FILE_NAME	"fifo_pn_tmp.res"

#define SCAN(x,y)	for(x=y;x!=NULL;x=x->Next)

#define CODE(x)		fprintf(Code_File,x)
#define CODE1(x,y)	fprintf(Code_File,x,y)
#define CODE2(x,y,z)	fprintf(Code_File,x,y,z)
#define CODE3(x,y,z,t)	fprintf(Code_File,x,y,z,t)
#define CODELN		CODE("\n")

#define NOT_KNOWN		-1
#define INFINITY	-2

extern PetriNet		*Read_Network( char *File_Name );

/* forward declarations */
void Dump_Bad_Trans( int *Table );
void Dump_Bad_Places( int *Table );
void Trans_Are_FIFO( PetriNet* Reseau );
void Places_Are_FIFO( PetriNet* Reseau );

/* global variables */
int	*Place_Bound;
int	*Bad_Trans;
int	*Bad_Places;
FILE	*Code_File;
boolean	Lazy = FALSE;

char	In_File_Name[128] = "";

int main(int argc, char *argv[])
{
  static char		Out_File_Name[128] = "";
  PetriNet 		*Reseau;
  int			i;
  boolean		Dump = FALSE;
  boolean		Help = FALSE;
  boolean		Args_Ok = TRUE;
  boolean		File_Ok = TRUE;
  boolean		Is_FIFO = TRUE;

  i = 1;

  if (argc>1)
    do {

      if GET_STRING( "-f", In_File_Name )
      else if GET_STRING( "-o", Out_File_Name )

      else if SWITCH( "-debug", Debug, TRUE )
      else if SWITCH( "-lazy", Lazy, TRUE )
      else if SWITCH( "-d", Dump, TRUE )

      else if IS_VERSION

      else if ( ( 0 == strcmp( "-?", argv[i] ) )
		|| ( 0 == strcmp( "-h", argv[i] ) ) ) {
	printf("Usage: %s [-d] [-lazy] [-f <file>] [-o <file>] [-h] [-?] \n", 
	       TOOLNAME);
	Args_Ok = FALSE;
	Help = TRUE;
	i++;
      }
      else	
	i++; /* skip unknown argument */
    }
    while ( Args_Ok && ( i < argc ) );

  if (Args_Ok) {

    Reseau = Read_Network( In_File_Name );

    if ( Reseau != NULL ) {

       /* allocation des tableaux necessaires a l'analyse */
       Place_Bound = (int*) calloc( Reseau->Nb_Places, sizeof( int ) );
       Bad_Places = (int*) calloc( Reseau->Nb_Places+1, sizeof( int ) );
       Bad_Trans = (int*) calloc( Reseau->Nb_Trans+1, sizeof( int ) );
       /* initialisation */
       for ( i=0; i<Reseau->Nb_Places; i++ ) {
	 Place_Bound[i] = NOT_KNOWN;
       }
       Bad_Places[0] = 0;
       Bad_Trans[0] = 0;

       /* analyse proprement dite */
       Trans_Are_FIFO( Reseau );
       Is_FIFO = ( Bad_Trans[0] == 0 );

       if ( Is_FIFO || !Lazy ) {
	 Places_Are_FIFO( Reseau );
	 Is_FIFO = ( Bad_Places[0] == 0 );
       }

       if ( Dump ) {
	 if ( strlen(Out_File_Name) != 0 )
	   File_Ok = ( NULL != freopen( Out_File_Name, "w", stdout ) );

	 if ( File_Ok ) {
	   Dump_Bad_Trans( Bad_Trans );
	   Dump_Bad_Places( Bad_Places );
	   return 0;
	 }
	 else {
	   fprintf( stderr, "%s: Error opening file %s.\n", 
		    TOOLNAME, Out_File_Name );
	   return -1;
	 }
       }

       return 0;
     }
    else {
      fprintf( stderr, "%s: Error reading network. No translation.\n", 
	       TOOLNAME );	
      return -2;
    }
  }
  else if (!Help)
    return -1;
  else
    return 0;
}


void Dump_Bad_Trans(
int	*Liste )
{
  int	i;

  printf( "# List of some non-FIFO transitions, generated by %s\n", TOOLNAME );
  printf( "\n%d\n_E_O_E\n", Liste[0] );

  for ( i=1; i <= Liste[0] ; i++ ) {
    printf( "\n%d\n_E_O_E\n", Liste[i]+1 );
  }
}

void Dump_Bad_Places( 
int	*Liste )
{
  int	i;

  printf( "# List of some non-FIFO places, generated by %s\n", TOOLNAME );
  printf( "\n%d\n_E_O_E\n", Liste[0] );

  for ( i=1; i <= Liste[0] ; i++ ) {
    printf( "\n%d\n_E_O_E\n", Liste[i]+1 );
  }
}

/* teste si une transition est recyclee (sans utiliser Next) */
boolean	Is_Recycled_Transition( 
PetriNet      *Reseau,
Tr_Des	      *The_Trans )
{
  Liste_Link	L_Scan;
  boolean	Is_Recycled = FALSE;
  Pl_Des	*The_Place;

  SCAN( L_Scan, The_Trans->Outputs ) {
    The_Place = &(Reseau->P[L_Scan->Dest]);
    Is_Recycled = Is_Recycled ||
      ( ( The_Place->Outdegree == 1 )
	&& ( The_Place->Outputs->Dest == The_Trans->Tr_Id ) );
  }

  return Is_Recycled;
}

int Find_Place_Bound(int	The_Number)
{
  char		command[512];
  FILE		*Res_File;
  int		Val;
  int		Res;

  sprintf( command, "%s -f %s -place %d -Bfile %s > /dev/null", 
	   PN_BOUND, In_File_Name, The_Number+1, RES_FILE_NAME );
  if ( Debug )
    printf( "%s: executing %s\n", TOOLNAME, command );

  system( command );

  Res_File = fopen( RES_FILE_NAME, "r" );
  if ( ( NULL != Res_File ) 
       && ( 2 == fscanf( Res_File, "%d %d", &Res, &Val ) ) ) {
	
    if ( Res == The_Number ) {
      if ( Debug )
	printf("%s: bound for place %d = %d\n", TOOLNAME, 
	       The_Number, Val );

      return Val;
    }
    else {
      printf("%s: Error: incoherent result in bound.\n", TOOLNAME );
      return NOT_KNOWN;
    }
  }
  else {
    /* no bound result. Infinite bound assumed */
    if ( Debug )
      printf("%s: bound for place %d = oo\n", TOOLNAME, The_Number );

    return INFINITY;
  }
}

void Places_Are_FIFO( 
PetriNet* Reseau )
{
  int		i;
  int		The_Num;
  Pl_Des	The_Place;
  boolean	Is_FIFO = TRUE;

  for ( i=0; ( !Lazy || Is_FIFO ) && (i<Reseau->Nb_Places); i++ ) {

    The_Num = i;
    The_Place = Reseau->P[i];

    if ( The_Place.Pl_Law->Name != 'C' ) {
      /* necessity to check the bounds on the place */

      if ( Place_Bound[ The_Num ] == NOT_KNOWN )
	Place_Bound[ The_Num ] = Find_Place_Bound( The_Num );

      Is_FIFO = ( Place_Bound[ The_Num ] != INFINITY ) 
	&& ( Place_Bound[ The_Num ] <= 1 );

      if ( !Is_FIFO ) {
	Bad_Places[0]++;
	Bad_Places[Bad_Places[0]] = i;

	if ( Debug )
	  printf("%s: place #%d (%s) not FIFO\n", TOOLNAME, i, 
		 The_Place.Name);
      }
    }
  }

}

void Trans_Are_FIFO(
PetriNet* Reseau )
{
  int		i;
  Liste_Link	L_Scan;
  boolean	Is_FIFO;
  int		The_Number;
  int		Val;

  for ( i=0; ( !Lazy || Is_FIFO ) && (i<Reseau->Nb_Trans); i++ )
    if ( ( Reseau->T[i].Tr_Law->Name != 'C' )
	 && !Is_Recycled_Transition( Reseau, &(Reseau->T[i]) ) ) {
      /* necessity to check the bounds on the input places */

      Is_FIFO = FALSE;
      for ( L_Scan = Reseau->T[i].Inputs; 
	    !Is_FIFO && ( L_Scan != NULL );
	    L_Scan = L_Scan->Next ) {
	/* find the bound on the place */
	The_Number = L_Scan->Dest;

	Val = Find_Place_Bound( The_Number );
	Place_Bound[The_Number] = Val;
	  
	if ( Val != INFINITY )
	  Is_FIFO = ( Val <= 1 );
      }

      if ( !Is_FIFO ) {
	Bad_Trans[0]++;
	Bad_Trans[Bad_Trans[0]] = i;

	if ( Debug )
	  printf("%s: transition %d not FIFO\n", TOOLNAME, The_Number );
      }
    }
}
