/* modules pour generer des reseaux ATM dans le format de Bruno.	*/

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <math.h>
#include "law.h"

#ifndef _BOOLEAN_H
#define _BOOLEAN_H
typedef int     boolean;
#define TRUE    1
#define FALSE   0
#endif

#define TOOLNAME	"gen_atm"
#define VERSION		"v1.0"

#ifndef ALPHA
extern long random(void);
#endif

void gen_switch_dcl( FILE *OutFile, int Nb_Switch, int Switch_Size,
		     int In_Buf_Size, int Out_Buf_Size );
void gen_session_dcl( FILE *OutFile, int Nb_Session, int Nb_Switch, 
		      int Switch_Size, Law_Desc The_Law, double Rho );
void write_parameters( FILE *OutFile, int argc, char *argv[]);

void display_help(char *Prog_Name)
{	
  printf("Usage: %s [-nb_switch <int>] [-nb_session <int>]",
	 Prog_Name);
  printf(" [-switch_size <int>]\n\t");
  printf("[-in_buf_size <int>] [-out_buf_size <int>]\n\t");
  printf("[-law <law>] [-rho <real>] [-o <file>] [-h] [-?]\n");
  printf("  <law>  : E (def) C B G g R U + parameters (def: 1.0)\n");
}

int main(int argc, char *argv[] )
{	
  int		i;
  static char	File_Name[128] = "";
  int		Switch_Size = 2;
  int		Nb_Switch = 3;
  int		Nb_Session = 10;
  int 		In_Buf_Size = 10;
  int		Out_Buf_Size = 10;
  double	Rho = 0.5;
  Law_Desc	The_Law;
  FILE*		OutFile;
  boolean	Args_Ok = TRUE;
  boolean	Help = FALSE;

  The_Law = (Law_Desc) malloc( sizeof( struct LAW_DESC ) );
  The_Law->Name = 'E';
  The_Law->Parameters = (double*) malloc( sizeof( double ) );
  The_Law->Parameters[0] = 1.0;

  i = 1;
  if (argc>1)
    do {
      if ( 0 == strcmp( "-nb_switch", argv[i] ) ) {
	i++;
	if ( 1 != sscanf( argv[i], "%d", &Nb_Switch ) ) {
	  printf("%s: integer expected after %s\n", TOOLNAME, argv[i-1] );
	  Args_Ok = FALSE;
	}
	else
	  i++;
      }
      else if ( 0 == strcmp( "-nb_session", argv[i] ) ) {
	i++;
	if ( 1 != sscanf( argv[i], "%d", &Nb_Session ) ) {
	  printf("%s: integer expected after %s\n", TOOLNAME, argv[i-1] );
	  Args_Ok = FALSE;
	}
	else
	  i++;
      }
      else if ( 0 == strcmp( "-switch_size", argv[i] ) ) {
	i++;
	if ( 1 != sscanf( argv[i], "%d", &Switch_Size ) ) {
	  printf("%s: integer expected after %s\n", TOOLNAME, argv[i-1] );
	  Args_Ok = FALSE;
	}
	else
	  i++;
      }
      else if ( 0 == strcmp( "-in_buf_size", argv[i] ) ) {
	i++;
	if ( 1 != sscanf( argv[i], "%d", &In_Buf_Size ) ) {
	  printf("%s: integer expected after %s\n", TOOLNAME, argv[i-1] );
	  Args_Ok = FALSE;
	}
	else
	  i++;
      }
      else if ( 0 == strcmp( "-out_buf_size", argv[i] ) ) {
	i++;
	if ( 1 != sscanf( argv[i], "%d", &Out_Buf_Size ) ) {
	  printf("%s: integer expected after %s\n", TOOLNAME, argv[i-1] );
	  Args_Ok = FALSE;
	}
	else
	  i++;
      }
      else if ( 0 == strcmp( "-rho", argv[i] ) ) {
	i++;
	if ( 1 != sscanf( argv[i], "%lf", &Rho ) ) {
	  printf("%s: real expected after %s\n", TOOLNAME, argv[i-1] );
	  Args_Ok = FALSE;
	}
	else
	  i++;
      }
      else if ( 0 == strcmp( "-law", argv[i] ) ) {
	i++;
	switch( The_Law->Name = argv[i][0] ) {
	case 'C':
	case 'E':
	case 'B':
	case 'g':
	case 'G':
	  i++;
	  sscanf( argv[i], "%lf", 
		  &(The_Law->Parameters[0] ) );
	  break;
	case 'R':
	case 'U':
	  i++;
	  The_Law->Parameters = (double*) 
	    realloc ( 
		     The_Law->Parameters, 
		     2*sizeof(double) );
	  sscanf( argv[i], "%lf", 
		  &(The_Law->Parameters[0] ) );
	  i++;
	  sscanf( argv[i], "%lf", 
		  &(The_Law->Parameters[1] ) );
	  break;
	default:
	  printf("%s: unknown law\n", TOOLNAME);
	} /* switch */
	i++;
      }
      else if ( 0 == strcmp( "-o", argv[i] ) )
	{	i++;
	if ( i < argc )
	  strcpy( File_Name, argv[i] );
	else	printf("%s: filename expected\n", TOOLNAME);
	i++;
	}
      else if ( ( 0 == strcmp( "-h", argv[i] ) )
		|| ( 0 == strcmp( "-?", argv[i] ) ) ) {
	i++;
	display_help( TOOLNAME );
	Args_Ok = FALSE;
	Help = TRUE;
      }
      else if ( 0 == strcmp( "-v", argv[i] ) ) {
	i++;
	printf("%s (Version %s)\n",argv[0],VERSION);
	Args_Ok = FALSE;
      }
      else 
	/* skip unknown argument */
	i++;
    }
  while ( ( i < argc ) && Args_Ok );

  if (Args_Ok) {

    if ( strlen(File_Name) != 0 )
      OutFile = fopen( File_Name, "w" );
    else
      OutFile = stdout;

    if ( OutFile != NULL ) {
      gen_switch_dcl( OutFile, Nb_Switch, Switch_Size, 
		      In_Buf_Size, Out_Buf_Size );

      gen_session_dcl( OutFile, Nb_Session, Nb_Switch, Switch_Size, 
		       The_Law, Rho );

      write_parameters( OutFile, argc, argv );

      return 0;
    }
    else {
      fprintf( stderr, "%s: error: could not open file %s for writing.\n",
	       TOOLNAME, File_Name );
      return -2;
    }
  }
  else if (Help)
    return 0;
  else
    return -1;
}

void gen_switch_dcl( FILE *OutFile, int Nb_Switch, int Switch_Size,
		     int In_Buf_Size, int Out_Buf_Size )
{
  int	i;
  int	Rows;

  fprintf( OutFile, "%d\n", Nb_Switch );
  Rows = 1;
  for( i=1; i<= Switch_Size; i++ )
    Rows *= 2;
  for( i=1; i<= Nb_Switch; i++ )
    fprintf( OutFile, "%d %d %d\n", Rows, In_Buf_Size, Out_Buf_Size );

}

#define MAX_RANDOM_VALUE          0x7FFFFFFF
double U_0_1( void )
{
  return (double)random()/MAX_RANDOM_VALUE;
}

/* Discrete uniform distrib between a and b --inclusive-- */
int Unif( int a, int b )
{
  static double x;
  x = (b+1-a)*U_0_1();
  return a + (int)floor( x );
}

void gen_session_dcl( FILE *OutFile, int Nb_Session, int Nb_Switch, 
		      int Switch_Size, Law_Desc The_Law, double Rho )
{
  int	i;
  int	Rows;
  int	A;
  int	Sw_Orig;
  int	Sw_Dest;
  int	Port_In;
  int	Port_Out;
  double	Factor;
  Law_Desc	New_Law;

  fprintf( OutFile, "%d\n", Nb_Session );

  /* computing number of rows, and number of output ports per destination */
  Rows = 1;
  for( i=1; i<= Switch_Size; i++ )
    Rows *= 2;

  if ( Rows < Nb_Switch ) {
    printf("%s: error: too many switches or too small switches.\n",
	   TOOLNAME );
  }
  else {
    A = (int)floor( Rows / Nb_Switch );

    /* perform rescaling of source law */
    Factor = 2.0 * Nb_Session / Rho / Nb_Switch / Rows / Mean( The_Law );
    
    New_Law = Rescale_Law( The_Law, Factor );

    for ( i=0; i<Nb_Session; i++ ) {
      Write_Law( OutFile, New_Law, NORMAL_PRINT_MODE );
      fprintf( OutFile, "\n" );

      /* get route */
      Sw_Orig = Unif( 1, Nb_Switch );
      Sw_Dest = Unif( 1, Nb_Switch );
      if ( Sw_Orig == Sw_Dest )
	fprintf( OutFile, "1\n");
      else
	fprintf( OutFile, "2\n");

      /* write first switch */
      Port_In = Unif( 1+(Nb_Switch-1)*A, Rows );
      if ( Sw_Orig == Sw_Dest )
	Port_Out = Unif( 1+(Nb_Switch-1)*A, Rows );
      else
	Port_Out = A * ( Sw_Dest > Sw_Orig ? Sw_Dest - 2: Sw_Dest - 1 )
	  + Unif( 0, A-1 ) + 1;

      fprintf( OutFile, "%d %d %d\n", Sw_Orig, Port_In, Port_Out );
      /* write second switch */
      if ( Sw_Orig != Sw_Dest ) {
	Port_In = A * ( Sw_Dest > Sw_Orig ? Sw_Orig - 1 : Sw_Orig - 2 )
	  + Unif( 0, A-1 ) + 1;
	Port_Out = Unif( 1+(Nb_Switch-1)*A, Rows );

	fprintf( OutFile, "%d %d %d\n", Sw_Dest, Port_In, Port_Out );
      }
    }
  }
}

void write_parameters( FILE *OutFile, int argc, char *argv[])
{
  int	i;

  fprintf( OutFile, "\n#\n");
  fprintf( OutFile, "# File generated by %s version %s with:\n",
	   TOOLNAME, VERSION );
  fprintf( OutFile, "#\n#");
  for( i=0; i<argc; i++ )
    fprintf( OutFile, " %s", argv[i] );
  fprintf( OutFile, "\n#\n");

}

