/****************************************************************/
/* Partie relative a lex/yacc					*/
/****************************************************************/

#include <stdio.h>
#include <malloc.h>
#include "pn.h"
#include "pnl.tab.h"

#define TOOLNAME	"PN parse"

#define MAX(x,y) ( x > y ? x : y )

#define SCAN(x,y)	for(x=y;x!=NULL;x=x->Next)

extern FILE* yyin;
extern int yydebug;
extern char yytext[];
#ifdef YY_FLEX_LEX_COMPAT
extern int yylineno;
#endif
extern int yyparse();
extern int yylex();
extern int PNL_lineno;

/* forced external declaration */
extern char *strdup(const char *);

/* External procedures */
extern void Write_Opt_String();
extern void Write_Address( Address The_Add );

/* Forward declaration */
void Invert_Network(PetriNet *);
boolean Consistency_Checking(PetriNet *);
void Free_Integer_List(Liste_Entier);
void Free_Address_List(Liste_Address);
void Free_Trans_List(Liste_Tr);
void Free_Place_List(Liste_Pl);
Liste_Subm Read_Submodel_Files(Liste_Subm);

boolean	Debug = FALSE;

/* I'm sick and tired of this nonsense about yyval/yylval. I put the 
   result in a global variable, that's all. */
/*
#ifdef USE_YYLVAL
extern  YYSTYPE yylval;
#define PARSE_RESULT yylval
#else 
*/
PetriNet *The_Petri_Net;
#define PARSE_RESULT The_Petri_Net
/*
#endif
*/
extern YYSTYPE yyval;

int Submodel_Level = 0;

char	**File_Name_Table;
int	Nb_Submodels;
boolean	Parsing_Done;
Binding	Bind_Table = (Binding) NULL;

/************************************************************************/
/* Function returning a FILE descriptor from a name, looking in the	*/
/* environment variable: PNL_PATH.					*/
/* Returns NULL on failure to open a file for reading.			*/
/************************************************************************/
FILE	*Open_PNL_File( char *File_Name )
{
  char	*Pnl_Path;
  char	*All_Path = NULL;
  char	*Sub_Path = NULL;
  char	*Full_File_Name = NULL;
  FILE	*Input_File;

  /* first look if this is a proper file */
  if ( NULL == File_Name )
    return stdin;
  else if ( strlen( File_Name ) == 0 )
    return stdin;
  else {
    /* Looks like a proper file. Check first if it is in the current dir. */
    Input_File = fopen( File_Name, "r" );
    
    if ( NULL != Input_File )
      return Input_File;
    else {
      /* have to look in PNL_PATH... 			*/
      /* At this point, Input_File is NULL 		*/
      Pnl_Path = getenv( "PNL_PATH" );

      if ( NULL != Pnl_Path ) {
	/* duplicate it in order not to modify the original! */
	All_Path = strdup( Pnl_Path );
	
	Sub_Path = strtok( All_Path, ":" );
      }
      /* else, All_Path and Sub_path have their defaults = NULL */

      if ( NULL != Sub_Path ) {
	do {
	  if ( Debug ) {
	    printf( "%s: looking for file %s in dir \"%s\"...",
		    TOOLNAME, File_Name, Sub_Path );
	  }
	  Full_File_Name = realloc( Full_File_Name,
				    (2+strlen(Sub_Path)+strlen(File_Name))*
				    sizeof(char) );

	  Full_File_Name = strcpy( Full_File_Name, Sub_Path );
	  Full_File_Name = strcat( Full_File_Name, "/" );
	  Full_File_Name = strcat( Full_File_Name, File_Name );

	  if ( NULL != ( Input_File = fopen( Full_File_Name, "r" ) ) ) {
	    if ( Debug )
	      printf( "found\n");
	  }
	  else if ( Debug )
	    printf("not there.\n");
	  
	  Sub_Path = strtok( (char*)NULL, ":" );
	} while ( ( NULL == Input_File ) && ( NULL != Sub_Path ) );

      }
      if ( ( NULL == Input_File ) && Debug ) {
	printf( "%s: error: could not open file %s for reading.\n",
		TOOLNAME, File_Name );
      }
      /* some cleanup */
      if ( Full_File_Name != NULL )
	free( Full_File_Name );
      if ( All_Path != NULL )
	free( All_Path );
      
      return Input_File;
    }
  }
}

/************************************************************************/
/* Main Parsing Function						*/
/* Attempts to read a PetriNet in the file File_Name.			*/
/* Files are sought using the environment variable PNL_PATH.		*/
/* But if  File_Name is NULL or empty, the standard input is assumed.	*/
/* If anything fails, NULL is returned.					*/
/************************************************************************/
PetriNet *Read_Network(char *File_Name)
{
  FILE	*Input_File;
  FILE*	Current_Input;
  int	Res;

  Input_File = Open_PNL_File( File_Name );

  if ( NULL != Input_File ) {
    Current_Input = yyin;
    yyin = Input_File;
  
    yydebug = 0;
    Parsing_Done = FALSE;

    Res = yyparse();

    yyin = Current_Input;

    fclose( Input_File );

    if ( Res == 0 ) {
      /* Caution: Parsing may seem succesfull but network may be bad	*/
      /*  for various reasons.						*/
      if ( NULL != PARSE_RESULT ) {
	PARSE_RESULT->File_Name = 
	  (char*)malloc( (unsigned)strlen(File_Name)+1 );
	strcpy( PARSE_RESULT->File_Name, File_Name );
      }
      return PARSE_RESULT;
    }
    else {
      return (PetriNet*)NULL;
    }
  }
  else
    return (PetriNet*)NULL;
}

void yyerror(char *s)
{
#ifdef YY_LEX_FLEX_COMPAT
  printf("---%s---[%s] at line %d\n",s,yytext,yylineno);
#else
  printf("---%s---[%s] at line %d\n",s,yytext,PNL_lineno);
#endif
}

#undef yywrap   /* just in case... */
int yywrap()
{

/*   if ( Debug ) */
/*     printf("Wrapping: Level=%d, Nb_Submodels=%d, Parsing: %s\n", */
/* 	   Submodel_Level, Nb_Submodels,  */
/* 	   ( Parsing_Done ? "done" : "not finished" ) ); */

/*   if ( Submodel_Level > 0 || Parsing_Done ) */
    /* finished... */
    return 1;
/*   else { */
    /* Continue */
/*     return 0; */
/*   } */
}

/****************************************************************/
/* Procedures de construction du reseau				*/
/****************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>

PetriNet *Parse_PN(int		The_T_Size,
		   int		The_P_Size,
		   int		The_Duration,
		   Liste_Pl	The_Places,
		   Liste_Tr	The_Transitions,
		   Binding		The_IOs,
		   Liste_Subm	The_Submodels,
		   Binding		The_Bindings )
{
  PetriNet	*RdP;
  Liste_Tr	T_Scan;
  Liste_Pl	P_Scan;
  Liste_Subm	S_Scan;
  Binding	B_Scan;
  int		Nb_Objs;
  boolean	Parsing_Ok = TRUE;

  /* allouer la place pour le reseau et son tableau de trans. */
  RdP = (PetriNet*) malloc( sizeof ( PetriNet ) );
  if ( NULL == RdP ) {
    E_MEM("PetriNet");
    return RdP;
  }
  else {
    /* Fill in values. */
    RdP->Nb_Trans = The_T_Size;
    RdP->Nb_Places = The_P_Size;
    RdP->T = (Tr_Des*) calloc( RdP->Nb_Trans, sizeof( Tr_Des ) );
    /* WARNING: Simulations control within the PN structure is an	*/
    /* obsolete point of view. Will be ultimately removed. DON'T use	*/
    /* it anymore. 							*/
    RdP->Sim_Params = ( Sim_Control* ) malloc( sizeof( Sim_Control ) );
    RdP->Sim_Params->Nb_Firings = The_Duration;
    RdP->P = (Pl_Des*) calloc( RdP->Nb_Places, sizeof( Pl_Des ) );
    RdP->Max_Pop_Init = 0;
    RdP->Has_Weights = FALSE;
    RdP->Has_Inhibitors = FALSE;
    RdP->File_Name = "<unknown>";

    /* scan the list of trans. and put it in the array T.	*/
    SCAN( T_Scan, The_Transitions ) {
      if ( T_Scan->Val->Tr_Id >= 0 ) {
	if ( T_Scan->Val->Tr_Id < The_T_Size ) {
	  RdP->T[T_Scan->Val->Tr_Id] = *(T_Scan->Val);
	}
	else {
	  fprintf(stderr,"Number of Transition > Trans. Size .");
	  fprintf(stderr," Transition '%d' ignored.\n",
		  T_Scan->Val->Tr_Id+1 );
	}
      }
      else {
	fprintf(stderr,"Error: illegal transition number (%d).",
		P_Scan->Val->Pl_Id+1 );
	fprintf(stderr," Transition ignored.\n");
      }
	
      free( T_Scan->Val );
    }

    /* idem for places */
    SCAN( P_Scan, The_Places) {
      if ( P_Scan->Val->Pl_Id >= 0 ) {
	if ( P_Scan->Val->Pl_Id < The_P_Size ) {
	  RdP->P[P_Scan->Val->Pl_Id] = *(P_Scan->Val);
	}	    
	else {
	  fprintf(stderr,"Number of place > Place Size.");
	  fprintf(stderr," Place '%d' ignored.\n",
		  P_Scan->Val->Pl_Id+1 );
	}
      }
      else {
	fprintf(stderr,"Error: illegal place number (%d). Place ignored.\n",
		P_Scan->Val->Pl_Id+1 );
      }
      free( P_Scan->Val );
    }

    /* idem for I/Os: inputs then outputs. */
    Nb_Objs = 0;
    SCAN( B_Scan, The_IOs )
      if ( B_Scan->From->Type == 'I' )
	Nb_Objs++;
    RdP->Nb_Inputs = Nb_Objs;	
    if ( Nb_Objs > 0 )
      RdP->Inputs = (Address*) calloc( Nb_Objs, sizeof(Address) );
    else
      RdP->Inputs = (Address*) NULL;
    SCAN( B_Scan, The_IOs )
      if ( B_Scan->From->Type == 'I' ) {
	B_Scan->From->Number--;
	if ( ( B_Scan->From->Number < Nb_Objs )
	     && ( 0 <= B_Scan->From->Number ) ) {
	  RdP->Inputs[B_Scan->From->Number] = B_Scan->To;
	}
	else {
	  fprintf(stderr,"%s: warning: Input number %d outside legal interval",
		  TOOLNAME, B_Scan->From->Number+1 );
	  fprintf(stderr,"[%d..%d]. Ignored.\n", 1, Nb_Objs );
	}
      }

    Nb_Objs = 0;
    SCAN( B_Scan, The_IOs )
      if ( B_Scan->From->Type == 'O' )
	Nb_Objs++;
    RdP->Nb_Outputs = Nb_Objs;
    if ( Nb_Objs > 0 )
      RdP->Outputs = (Address*) calloc( Nb_Objs, sizeof(Address) );
    else
      RdP->Outputs = (Address*) NULL;
    SCAN( B_Scan, The_IOs )
      if ( B_Scan->From->Type == 'O' ) {
	B_Scan->From->Number--;
	if ( ( B_Scan->From->Number < Nb_Objs )
	     && ( 0 <= B_Scan->From->Number ) ) {
	  RdP->Outputs[B_Scan->From->Number] = B_Scan->To;
	}
	else {
	  fprintf(stderr,"%s: warning: Output number %d outside legal interval",
		  TOOLNAME, B_Scan->From->Number+1 );
	  fprintf(stderr,"[%d..%d]. Ignored.\n", 1, Nb_Objs );
	}
      }

    /* idem pour les sous-modeles */
    Nb_Objs = 0;
    SCAN( S_Scan, The_Submodels ) {
      Nb_Objs++;
    }
    RdP->Nb_Submodels = Nb_Objs;
    if ( Nb_Objs == 0 ) {
      RdP->S = (Subm_Des*)NULL;
    }
    else {
      The_Submodels = Read_Submodel_Files( The_Submodels );
      if ( NULL == The_Submodels ) {
	/* something went wrong when reading submodels. Abort all.	*/
	Parsing_Ok = FALSE;
      }
      else {
	RdP->S = (Subm_Des*) calloc( Nb_Objs, sizeof( struct SUBM_DES ) );
	Nb_Objs = 0;
	SCAN( S_Scan, The_Submodels ) {
	  RdP->S[Nb_Objs] = *(S_Scan->Val);
	  RdP->S[Nb_Objs].Father = RdP;
	  Nb_Objs++;
	}
      }
    }
    
    if ( Parsing_Ok ) {
      /* put the bindings in */
      RdP->Binding_Table = The_Bindings;
      Invert_Network( RdP );
    
      Parsing_Ok = Consistency_Checking( RdP );

    }

    if ( !Parsing_Ok )
      RdP = (PetriNet*)NULL;

    Free_Trans_List( The_Transitions );
    Free_Place_List( The_Places );

    Parsing_Done = TRUE;
    
    return RdP;
  }
}

Tr_Des *Parse_Transition(int		Trans_Num,
			 Law_Desc	The_Law,
			 char		*The_Name,
			 char		*The_Data )
{
  Tr_Des	*The_Trans;

  The_Trans = (Tr_Des*) malloc( sizeof( Tr_Des ) );

  if ( The_Trans != NULL ) {

      /* internal Ids are ALWAYS between 0 and NB_TRANS-1	*/
      The_Trans->Tr_Id = Trans_Num-1;
      The_Trans->Tr_Law = The_Law;
      /* To get rid of memory management problems: alloc and copy */
      The_Trans->Name = malloc( 1 + strlen( The_Name ) );
      The_Trans->Name = strcpy( The_Trans->Name, The_Name );
      The_Trans->Data = malloc( 1 + strlen( The_Data ) );
      The_Trans->Data = strcpy( The_Trans->Data, The_Data );
      The_Trans->Indegree = 0;
      The_Trans->Outdegree = 0;
      The_Trans->Inputs = (Liste_Link) NULL;
      The_Trans->Outputs = (Liste_Link) NULL;
      The_Trans->Tag = 0;

      if ( Debug ) {
	printf("Transition ");
	if ( strlen( The_Trans->Name ) > 0 )
	  printf("%s ", The_Trans->Name );
	printf("[%d] ", Trans_Num );
	if ( strlen( The_Trans->Data ) > 0 )
	  printf("(a=%s) ", The_Trans->Data );
	printf("reconnue\n");
      }

      /* Cleaning up. But not the strings "" that (may) have not been	*/
      /* malloc-ed.							*/
      if ( strlen( The_Name ) > 0 )
	free( The_Name );
      if ( strlen( The_Data ) > 0 )
	free( The_Data );

    }
  else {
    E_MEM("Transition");
  }

  return The_Trans;
}

Binding Parse_Binding( int		The_Number,
		       Address		The_Address,
		       Liste_Link	The_Neighbors )
{
  Binding	Binding_List;
  Binding	New_Binding;
  Address	New_Address;
  Liste_Link	L_Scan;

  Binding_List = (Binding)NULL;

  SCAN( L_Scan, The_Neighbors ) {

    /* duplicate origin address */
    New_Address = (Address) malloc( sizeof( struct ADDRESS ) );
    New_Binding = (Binding) malloc( sizeof( struct BINDING ) );

    if ( ( NULL != New_Address ) && ( NULL != New_Binding ) ) {

      New_Address->Type = 'S';
      New_Address->Number = The_Number;
      New_Address->Next = The_Address;

      New_Binding->From = New_Address;
      New_Binding->To = L_Scan->A_Dest;

      /* The case of a direct address. Let "To" stay to NULL: this will	*/
      /* be resolved later. The number of the address is in the link	*/
      /* description.							*/

      New_Binding->Link = L_Scan;
      New_Binding->Next = Binding_List;
      Binding_List = New_Binding;

    }
    else {
      E_MEM("Binding+Address");
      /* exit of the SCAN loop */
      L_Scan = (Liste_Link)NULL;
    }

  }

  return( Binding_List );
}

Binding Make_Binding( Address		Address1,
		      Address		Address2,
		      Liste_Link	The_Neighbors )
{
  Binding	New_Binding;

  New_Binding = (Binding) malloc( sizeof( struct BINDING ) );

  if ( NULL != New_Binding ) {

      New_Binding->From = Address1;
      New_Binding->To = Address2;
      New_Binding->Link = The_Neighbors;
      New_Binding->Next = (Binding)NULL;
      
  }
  else {
    E_MEM("Binding");
  }

  return( New_Binding );
}

Binding Append_Binding( Binding List1, Binding List2 )
{
  Binding	B_List;

  if ( List1 == NULL )
    List1 = List2;
  else {
    for ( B_List = List1; B_List->Next != NULL; B_List = B_List->Next);
    B_List->Next = List2;
  }

  return List1;
}

Binding Add_Binding( Binding	The_Table,
		     char	The_Type,
		     int	The_Number,
		     Address	The_Dest,
		     Liste_Link The_Link )
{
  Binding	New_Binding;
  Address	New_Address;

  /* This procedure is called when scanning an output list. Therefore	*/
  /* only "S./IN." are valid bindings.					*/
  if ( The_Dest->Next == NULL ) {
    printf("%s: warning: unexpected address encountered ", TOOLNAME );
    printf("at object %c %d (ignored).\n", The_Type, The_Number );

    return( The_Table );
  }
  else {
    if ( The_Dest->Next->Type == 'I' ) {
      New_Binding = (Binding) malloc( sizeof( struct BINDING ) );
      New_Address = (Address) malloc( sizeof( struct ADDRESS ) );

      New_Address->Type = The_Type;
      New_Address->Number = The_Number;
      New_Address->Next = (Address)NULL;

      New_Binding->From = New_Address;
      New_Binding->To = The_Dest;
      New_Binding->Next = The_Table;
      New_Binding->Link = The_Link;

      return( New_Binding );
    }
    else {
      printf("%s: warning: link to an \"out\" [%c] port encountered ", 
	     TOOLNAME, The_Dest->Next->Type );
      printf("at object %c %d (ignored).\n", The_Type, The_Number );

      return( The_Table );
    }
  }
}
    
Tr_Des *Add_Trans_Sons( Tr_Des		*The_Trans,
			Liste_Link	The_Neighbors )
{	
  Liste_Link	Scan;

  if ( The_Trans != NULL ) {

    The_Trans->Outputs = The_Neighbors;

    for ( Scan = The_Neighbors; Scan != NULL; Scan=Scan->Next ) {
      The_Trans->Outdegree++;
    }
  }
  else {
    fprintf(stderr, 
	    "Cannot add place list to null transition.\n");
  }

  return The_Trans;
}

Pl_Des *Add_Place_Sons( Pl_Des		*The_Place,
			Liste_Link	The_Neighbors )
{	
  Liste_Link	Scan;

  if ( The_Place != NULL ) {
    The_Place->Outputs = The_Neighbors;

    for ( Scan = The_Neighbors; Scan != NULL; Scan=Scan->Next ) {
      The_Place->Outdegree++;
    }
  }
  else {
    fprintf(stderr, "Cannot add transition list to null place.\n");
  }

  return The_Place;
}

Pl_Des *Parse_Place( Liste_Link		Place_Head,
		     Law_Desc		The_Law,
		     Population		*The_Pop,
		     char		*The_Data,
		     Routing_Desc	The_Routing )
{
  Pl_Des	*Resultat;

  Resultat = (Pl_Des*) malloc( sizeof ( Pl_Des ) );

  /* internal Ids are ALWAYS between 0 and NB_TRANS-1	*/
  Resultat->Pl_Id = Place_Head->Dest;
  Resultat->Name = Place_Head->Data;
  Resultat->Data = malloc( 1 + strlen(The_Data) );
  Resultat->Data = strcpy( Resultat->Data, The_Data );
  Resultat->Indegree = 0;
  Resultat->Outdegree = 0;
  Resultat->Inputs = (Liste_Link) NULL;
  Resultat->Outputs = (Liste_Link) NULL;
  Resultat->Pl_Law = The_Law;
  Resultat->Init_Pop = The_Pop->Pop;
  Resultat->Lag_Times = The_Pop->Delais;
  Resultat->Routing = The_Routing;
  Resultat->Tag = 0;

  free( Place_Head );
  free( The_Pop );
  if ( strlen( The_Data ) > 0 )
    free( The_Data );

  return Resultat;
}

Liste_Link Parse_Place_Head( int	Place_Number,
			     char	*Place_Name )
{
  Liste_Link	Resultat;

  Resultat = (Liste_Link) malloc( sizeof ( struct LINK_LIST ) );

  /* Correction of object number done here.		*/
  Resultat->Dest = Place_Number - 1;
  Resultat->Data = Place_Name;

  return Resultat;
}


Listes Append_Transition( Listes The_List, Tr_Des *New_Trans )
{
  Liste_Tr The_Tr_List;

  if ( The_List == NULL ) {
    The_List = (Listes) malloc( sizeof(struct LISTES) );
    The_List->Transitions = NULL;
    The_List->Places = NULL;
  }

  The_Tr_List = (Liste_Tr) malloc( sizeof( struct TR_LIST ) );
  
  if ( The_Tr_List != NULL ) {
    The_Tr_List->Val = New_Trans;
    The_Tr_List->Next = The_List->Transitions;
    The_List->Transitions = The_Tr_List;
  }
  else {
    E_MEM("Transition");
  }

  return The_List;
}

Listes Append_Place( Listes	The_List, Pl_Des	*New_Place )
{
  Liste_Pl The_Pl_List;

  if ( The_List == NULL ) {
    The_List = (Listes) malloc( sizeof(struct LISTES) );
    The_List->Transitions = NULL;
    The_List->Places = NULL;
  }

  The_Pl_List = (Liste_Pl) malloc( sizeof( struct PL_LIST ) );

  if ( The_Pl_List != NULL ) {
    The_Pl_List->Val = New_Place;
    The_Pl_List->Next = The_List->Places;
    The_List->Places = The_Pl_List;
  }
  else {
    E_MEM("Place");
  }

  return The_List;
}

Liste_Entier Append_Integer(Liste_Entier    Integer_List,
			    int		New_Integer )
{
  Liste_Entier	The_List;

  The_List = (Liste_Entier) malloc( sizeof( struct INT_LIST ) );

  if ( The_List != NULL ) {
    The_List->Val = New_Integer;
    The_List->Next = Integer_List;
  }
  else {
    E_MEM("Integer");
  }

  return( The_List );
}

Liste_Link Append_Link(
Liste_Link 	Link_List,
Liste_Link	New_Link )
{
  if ( New_Link != NULL )
    New_Link->Next = Link_List;
  else {
    fprintf(stderr, "Appending null integer to list.\n");
  }

  return( New_Link );
}

Liste_Link Parse_Link( Address		The_Address,
		       int		The_Weight,
		       char		*The_Data )
{
  Liste_Link	The_List;

  The_List = (Liste_Link) malloc( sizeof( struct LINK_LIST ) );

  if ( The_List != NULL ) {
    if ( ( The_Address->Type == 'P' ) 
	 || ( The_Address->Type == 'T' ) 
	 || ( The_Address->Type == 'U' ) ) {
      /* This is a "direct" address. Store it without address structure. */
      The_List->Dest = The_Address->Number;
      The_List->A_Dest = (Address) NULL;
      /* ...but keep type information */
      if ( The_Address->Type == 'P' ) 
	The_List->Link_Type = TRAPLA;
      else if ( The_Address->Type == 'T' ) 
	The_List->Link_Type = PLATRA;
      else 
	The_List->Link_Type = UNKNOWN_TYPE;

      /* then throw address away */
      free( (char*)The_Address );
    }
    else {
      /* This is an address in a submodel. Put it away to handle it later */
      The_List->Dest = -1;	/* conventional "no-address" */
      The_List->A_Dest = The_Address;
    }
    The_List->Weight = The_Weight;
    The_List->Data = The_Data;
    The_List->Is_Inhibitor = FALSE;
    The_List->Next = (Liste_Link) NULL;
  }
  else {
    E_MEM("Link");
  }

  return( The_List );
}

Population *Parse_Init_Pop( 
int		Nb_Tokens,
Liste_Reel	Lags )
{
  Population  *The_Pop;
  int		Nb;
  Liste_Reel  Scan;

  The_Pop = (Population *) malloc(sizeof(struct POPULATION));

  if (The_Pop != NULL) {
    The_Pop->Pop = Nb_Tokens;
    The_Pop->Delais = (double *) calloc(Nb_Tokens,
					sizeof(double));
    if ( ( Nb_Tokens != 0 ) && ( The_Pop->Delais != NULL) ) {
      Nb = 0;
      for (Scan = Lags;
	   (Scan != NULL) && (Nb < Nb_Tokens);
	   Scan = Scan->Next) {
	The_Pop->Delais[Nb] = Scan->Val;
	Nb++;
      }
      if (Nb < Nb_Tokens) {
	fprintf(stderr,
		"Warning: not enough initial lags: %d missing ",
		Nb_Tokens - Nb );
	fprintf(stderr, "(0.0 assumed)\n");
	for (; Nb < Nb_Tokens; Nb++) {
	  The_Pop->Delais[Nb] = 0.0;
	}
      }
      else if (Scan != NULL) {
	fprintf(stderr,
		"Warning: too many initial lags (ignored)\n");
      }
    } 
    else if ( Nb_Tokens != 0 ) {
      E_MEM("real");
    }
  } 
  else {
    E_MEM("population");
  }

  Free_Real_List( Lags );

  return The_Pop;
}

Routing_Desc Parse_Routing(char		The_Type,
			   Liste_Reel	The_Probas,
			   Liste_Address	The_Cycle )
{
  Routing_Desc		The_Routing;
  int			i;
  int             	Nb = 0;
  Liste_Reel      	R_Scan;
  Liste_Address   	I_Scan;

  The_Routing = (Routing_Desc) malloc( sizeof(struct ROUTING) );

  if (The_Routing != NULL) {
    The_Routing->Type = The_Type;

    switch( The_Type ) {
    case 'R':
      /* Parsing random routing description */
      SCAN( R_Scan, The_Probas )
	Nb++;
      The_Routing->Probas = (double *) calloc( (unsigned)Nb, sizeof(double) );
      The_Routing->Cycle = (Address *)NULL;
      The_Routing->Period = 0;
      i = 0;
      /* store the probas. There is NO reversal of the list, on the	*/
      /* assumption that neighbors are ALSO stored in			*/
      /* reverse in the Output list.					*/
      SCAN( R_Scan, The_Probas ) {
	The_Routing->Probas[i] = R_Scan->Val;
	i++;
      }
      break;

    case 'D':
      /* Parsing cyclic routing description */
      SCAN( I_Scan, The_Cycle )
	Nb++;
      The_Routing->Period = Nb;
      The_Routing->Probas = NULL;
      The_Routing->Cycle = (Address *) calloc( (unsigned)Nb, sizeof(Address) );
      i = 0;
      /* store the cycle in the structure. Here it 	*/
      /* is necesary to reverse the order.		*/
      SCAN( I_Scan, The_Cycle ) {
	The_Routing->Cycle[Nb-i-1] = I_Scan->Val;
	/* eliminate the possible "U" in address types 	*/
	if ( The_Routing->Cycle[Nb-i-1]->Type == 'U' )
	  The_Routing->Cycle[Nb-i-1]->Type = 'T';
	i++;
      }
      break;

    case 'P':
      /* Parsing competition/priority routing description */
      SCAN( I_Scan, The_Cycle )
	Nb++;
      The_Routing->Period = Nb;
      The_Routing->Probas = NULL;
      The_Routing->Cycle = (Address *) calloc( (unsigned)Nb, sizeof(Address) );
      i = 0;
      /* store the cycle in the structure. Here it 	*/
      /* is necesary to reverse the order.		*/
      SCAN( I_Scan, The_Cycle ) {
	The_Routing->Cycle[Nb-i-1] = I_Scan->Val;
	/* eliminate the possible "U" in address types 	*/
	if ( The_Routing->Cycle[Nb-i-1]->Type == 'U' )
	  The_Routing->Cycle[Nb-i-1]->Type = 'T';
	i++;
      }
      break;

    case 'C':
      /* Parsing default competition routing description */
      The_Routing->Probas = NULL;
      The_Routing->Cycle = NULL;
      The_Routing->Period = 0;
    }
  }
  else {	fprintf(stderr,"Could not create routing description \n");
  }

  if ( The_Probas != NULL )
    Free_Real_List( The_Probas );

  if ( The_Cycle != NULL )
    Free_Address_List( The_Cycle );
		     
  return (The_Routing);
}

Address Parse_Address( char		The_Type,
		       int		The_Number,
		       Address		The_Sequel )
{
  Address	The_Address;

  The_Address = (Address) malloc( sizeof( struct ADDRESS ) );

  if ( The_Address != NULL ) {
    
    The_Address->Type = The_Type;
    The_Address->Number = The_Number;
    The_Address->Next = The_Sequel;

    /* Correction of the "object number offset" is done here for 	*/
    /* "elementary" objects (places and transitions).			*/
    if ( ( The_Type == 'P' ) || ( The_Type == 'T' ) || ( The_Type == 'U' ) )
      The_Address->Number--;

  }
  else {
    E_MEM("Address");
  }

  return The_Address;
}

Liste_Address	Append_Address( Address The_Add, Liste_Address The_List )
{
  Liste_Address	Res;

  Res = (Liste_Address) malloc( sizeof( struct ADDRESS_LIST ) );

  Res->Val = The_Add;
  Res->Next = The_List;
    
  return Res;
}
  
/************************************************************************/
/* Procedures for parsing submodels.					*/
/* Takes as input a list of (unread) submodels, and attempts to open	*/
/* the files which name is in the field File_Name.			*/
/* Returns the list of submodels with the field Submodel containing 	*/
/* the Petri Nets.							*/
/* If a read fails, the NULL list is returned.				*/
/************************************************************************/
Liste_Subm Read_Submodel_Files( Liste_Subm The_List )
{
  Liste_Subm	S_Scan;
  Subm_Des	*The_Submodel;
  int		i;
  boolean	Read_Ok = TRUE;

  /* Counting the submodels */
  Nb_Submodels = 0;
  SCAN( S_Scan, The_List ) {
    Nb_Submodels++;
  }

  if ( Nb_Submodels == 0 ) {
    if ( Debug )
      printf("No Submodels found.\n" );
    return (Liste_Subm)NULL;
  }
  else {

    if ( Debug )
      printf("Found %d Submodels in total.\n", Nb_Submodels );

    /* "Copying" their file names in a table (just pointer manips) */
    File_Name_Table = (char**) calloc( (unsigned)Nb_Submodels, 
				       sizeof(char*) );
    i = 0;
    SCAN( S_Scan, The_List ) {
      File_Name_Table[i] = S_Scan->Val->File_Name;
      i++;
    }

    /* reading the files */
    SCAN( S_Scan, The_List ) {

      The_Submodel = S_Scan->Val;
      if ( Debug )
	printf("Opening file %s, looking for submodel [level %d]\n", 
	       The_Submodel->File_Name, Submodel_Level );

      Submodel_Level++;
      The_Submodel->Submodel = Read_Network( The_Submodel->File_Name );
      Submodel_Level--;

      if ( NULL != The_Submodel->Submodel ) {
	if ( Debug ) {
	  printf("Submodel ");
	  Write_Opt_String( The_Submodel->File_Name, NO_CHAR );
	  printf("[#%d] ", The_Submodel->Subm_Id );
	  printf("recognized\n");
	}
	Nb_Submodels--;
      }
      else {
	printf("%s: fatal error: Could not get submodel in file %s.\n",
	       TOOLNAME, The_Submodel->File_Name );
	Read_Ok = FALSE;
      }
    }
    /* ok with submodel. Parsing has to continue. */
    Parsing_Done = FALSE;

    if ( Read_Ok )
      return The_List;
    else
      return (Liste_Subm)NULL;
  }
}

Subm_Des *Parse_Submodel( int		The_Id,
			  char		*The_File_Name,
			  char		*The_Name )
{
  Subm_Des	*The_Submodel;

  The_Submodel = (Subm_Des*) malloc( sizeof( Subm_Des ) );

  if ( The_Submodel != NULL ) {
    
    The_Submodel->Subm_Id = The_Id;
    The_Submodel->Submodel = (PetriNet*)NULL;
    The_Submodel->Father = (PetriNet*)NULL;
    The_Submodel->Name = The_Name;
    The_Submodel->File_Name = 
      (char*) malloc( 1+(unsigned)strlen( The_File_Name ) );
    The_Submodel->File_Name = strcpy( The_Submodel->File_Name, The_File_Name );
    The_Submodel->First_Place = 0;
    The_Submodel->First_Trans = 0;

    if ( Debug ) {
      printf("Encountered Submodel " );
      if ( strlen( The_Submodel->File_Name ) > 0 )
	printf("%s ", The_Submodel->File_Name );
      printf("[#%d]\n", The_Submodel->Subm_Id );
    }

    free( The_File_Name );
  }
  else {
    E_MEM("Submodel");
  }

  return The_Submodel;
}

Liste_Subm Append_Submodel( Liste_Subm	The_List,
			    Subm_Des	*New_Submodel )
{
  Liste_Subm	The_Subm_List;

  The_Subm_List = (Liste_Subm) malloc( sizeof( struct SUBM_LIST ) );

  if ( The_Subm_List != NULL ) {
    The_Subm_List->Val = New_Submodel;
    The_Subm_List->Next = The_List;
  }
  else {
    E_MEM("Submodel");
  }

  return( The_Subm_List );
}

/************************************************************************/
/* Procedure 	invert_network						*/
/* This procedure has several functionalities:				*/
/* - trim the lists of downwards links, extracting the "bindings" to	*/
/*   submodels.								*/
/* - construct the lists of upward links.				*/
/* - set some of the "redundant" attributes of a network already	*/
/*   parsed like Max_Pop_Init.						*/
/*									*/
/* Parameters:								*/
/*	RdP:		pointer to the network				*/
/************************************************************************/
void	Invert_Network(PetriNet *RdP)
{
  int		i;
  Pl_Des	*The_Place;
  Tr_Des	*The_Trans;
  Liste_Link	The_List;
  Liste_Link	L_Scan;
  Liste_Link	M_Scan;

  for ( i = 0; i < RdP->Nb_Places; i++ ) {
    The_Place = &(RdP->P[i]);
    RdP->Max_Pop_Init = MAX( The_Place->Init_Pop,
			     RdP->Max_Pop_Init );
    SCAN( L_Scan, The_Place->Outputs ) {
      if ( L_Scan->A_Dest == NULL ) {
	/* This is a local link. Duplicate this link as input to its 	*/
	/* output Place.						*/
	The_List = (Liste_Link) malloc( sizeof( struct LINK_LIST ) );
	The_List->Next = RdP->T[L_Scan->Dest].Inputs;
	The_List->Dest = i;
	The_List->Weight = L_Scan->Weight;
	/* duplicating the data to make things cleaner */
	The_List->Data = malloc( (unsigned)(1+strlen( L_Scan->Data )) );
	The_List->Data = strcpy( The_List->Data, L_Scan->Data );
	The_List->Is_Inhibitor = L_Scan->Is_Inhibitor;
	RdP->T[L_Scan->Dest].Inputs = The_List;
	RdP->T[L_Scan->Dest].Indegree ++;
	if ( The_List->Weight > 1 )
	  RdP->Has_Weights = TRUE;
	if ( The_List->Is_Inhibitor )
	  RdP->Has_Inhibitors = TRUE;
      }
      else {
	/* This is a link into a submodel. First add it to the 		*/
	/* binding table.						*/
	RdP->Binding_Table = Add_Binding( RdP->Binding_Table, 
					  'P', The_Place->Pl_Id, 
					  L_Scan->A_Dest, L_Scan );
	/* Now remove it from the list.					*/
	/* But do NOT decrease Outdegree.				*/
	if ( L_Scan == The_Place->Outputs ) 
	  The_Place->Outputs = L_Scan->Next;
	else {
	  for ( M_Scan = The_Place->Outputs;
		M_Scan->Next != L_Scan;
		M_Scan = M_Scan->Next );
	  M_Scan->Next = L_Scan->Next;
	}
      }
    }
  }

  for ( i = 0; i < RdP->Nb_Trans; i++ ) {
    The_Trans = &(RdP->T[i]);
    SCAN( L_Scan, The_Trans->Outputs ) {
      if ( L_Scan->A_Dest == NULL ) {
	/* This is a local link. Duplicate this link as input to its 	*/
	/* output Transition.						*/
	The_List = (Liste_Link) malloc( sizeof( struct LINK_LIST ) );
	The_List->Next = RdP->P[L_Scan->Dest].Inputs;
	The_List->Dest = i;
	The_List->Weight = L_Scan->Weight;
	/* duplicating the data to make things cleaner */
	The_List->Data = malloc( (unsigned)(1+strlen( L_Scan->Data )) );
	The_List->Data = strcpy( The_List->Data, L_Scan->Data );
	
	RdP->P[L_Scan->Dest].Inputs = The_List;
	RdP->P[L_Scan->Dest].Indegree ++;
      }
      else {
	/* This is a link into a submodel. First add it to the 		*/
	/* binding table.						*/
	RdP->Binding_Table = Add_Binding( RdP->Binding_Table,
					  'T', The_Trans->Tr_Id, 
					  L_Scan->A_Dest, L_Scan );
	/* Now remove it from the list (and throw it away!)		*/
	if ( L_Scan == The_Trans->Outputs ) 
	  The_Trans->Outputs = L_Scan->Next;
	else {
	  for ( M_Scan = The_Trans->Outputs;
		M_Scan->Next != L_Scan;
		M_Scan = M_Scan->Next );
	  M_Scan->Next = L_Scan->Next;
	}
      }
    }
  }
}
/************************************************************************/
/* Stub for upwards compatibility. TO be removed... ultimately.		*/
/************************************************************************/
void	invert_network( PetriNet *RdP )
{
  if ( Debug ) {
    printf("%s: Call to invert_network() obsolete. Update the code.\n",
	   TOOLNAME );
  }
}

/****************************************************************/
/* Procedure de Trace						*/
/****************************************************************/
void dumpe_network( PetriNet *RdP )
{
  int		i;
  int		j;
  Liste_Link	I_Scan;
  boolean	Is_After;
  boolean	Inhibits;
  Binding	B_Scan;

  printf("Reseau de taille %d transitions et %d places\n",
	 RdP->Nb_Trans, RdP->Nb_Places );

  printf("   ");
  for (i=0; i<RdP->Nb_Places+2; i++)
    printf(".");
  printf("\n");
	
  for (i=0; i<RdP->Nb_Trans; i++) {
    printf("%3d.",i+1);
    for (j=0; j<RdP->Nb_Places; j++) {
      Is_After = FALSE;
      SCAN( I_Scan, RdP->T[i].Outputs )
	Is_After = ( Is_After || (j==I_Scan->Dest) );
      if ( Is_After )
	printf("1");
      else
	printf("0");
    }
    printf(".");
    printf("%c",RdP->T[i].Tr_Law->Name);
    SCAN( I_Scan, RdP->T[i].Outputs ) {
      printf("[");
      if ( I_Scan->A_Dest == NULL )
	printf("%d %d %s]", I_Scan->Dest, I_Scan->Weight,
	       I_Scan->Data);
      else {
	Write_Address( I_Scan->A_Dest );
	printf(" %d %s]", I_Scan->Weight, I_Scan->Data);
      }
    }
    printf("\n");
  }

  printf("   ");
  for (i=0; i<MAX(RdP->Nb_Trans,RdP->Nb_Places)+2; i++)
    printf(".");
  printf("\n");

  for (i=0; i<RdP->Nb_Places; i++) {
    printf("%3d.",i+1);
    Is_After = FALSE;
    for (j=0; j<RdP->Nb_Trans; j++) {
      Is_After = FALSE;
      Inhibits = FALSE;
      SCAN( I_Scan, RdP->P[i].Outputs ) {
	Is_After = ( Is_After || (j==I_Scan->Dest) );
	Inhibits = ( Inhibits || 
		     ( Is_After && I_Scan->Is_Inhibitor ) );
      }
      if ( Is_After ) {
	if ( Inhibits )
	  printf("-");
	else
	  printf("1");
      }
      else
	printf("0");
    }
    printf(".");
    printf("%c (%1d)",RdP->P[i].Pl_Law->Name,RdP->P[i].Init_Pop);
    SCAN( I_Scan, RdP->P[i].Outputs )
      printf("[%d %d %s]", I_Scan->Dest, I_Scan->Weight, I_Scan->Data);
    if ( RdP->P[i].Routing != NULL ) {
      switch ( RdP->P[i].Routing->Type ) {
      case 'C':
	printf(" competition");
	break;
      case 'D':
	printf(" cyclic:");
	for ( j=0; j<RdP->P[i].Routing->Period; j++ ) {
	  printf(" %d", RdP->P[i].Routing->Cycle[j]->Number );
	}
	break;
      case 'R':
	printf(" random:");
	for ( j=0; j<RdP->P[i].Outdegree; j++ ) {
	  printf(" %6.4f", RdP->P[i].Routing->Probas[j] );
	}
	break;
      case 'P':
	printf(" priorities:");
	for ( j=0; j<RdP->P[i].Routing->Period; j++ ) {
	  printf(" %d", j<RdP->P[i].Routing->Cycle[j]->Number );
	}
	break;
      }
    }
    printf("\n");
  }

  printf("   ");
  for (i=0; i<RdP->Nb_Trans+2; i++)
    printf(".");
  printf("\n\n");

  printf("population initiale maximale: %d\n",RdP->Max_Pop_Init);
  if ( RdP->Has_Weights )
    printf("Some links have weights\n");
  if ( RdP->Has_Inhibitors )
    printf("There are inhibitor arcs\n");
  printf("\n");

  for (i=0; i<RdP->Nb_Trans; i++) {
    printf("%1d Places amont de %d: ", RdP->T[i].Indegree, i+1);
    SCAN( I_Scan, RdP->T[i].Inputs )
      printf(" %d",I_Scan->Dest+1);
    printf("\n");
  }
  for (i=0; i<RdP->Nb_Places; i++) {
    printf("%1d Transitions amont de %d: ",
	   RdP->P[i].Indegree, i+1);
    SCAN( I_Scan, RdP->P[i].Inputs)
      printf(" %d",I_Scan->Dest+1);
    printf("\n");
  }

  printf("\n");
  printf("Inputs: %d\n", RdP->Nb_Inputs );
  printf("Outputs: %d\n", RdP->Nb_Outputs );
  printf("Bindings:");
  if ( RdP->Binding_Table == NULL ) {
    printf(" none.\n");
  }
  else {
    printf("\n");
    SCAN( B_Scan, RdP->Binding_Table ) {
      Write_Address( B_Scan->From );
      printf(" -> ");
      Write_Address( B_Scan->To );
      printf("[%d %s]\n", B_Scan->Link->Weight, B_Scan->Link->Data);
    }
  }

  printf("Submodels:");
  if ( RdP->Nb_Submodels  == 0 )
    printf(" none.\n");
  else {
    printf(" %d\n", RdP->Nb_Submodels );
    for ( i=0; i<RdP->Nb_Submodels; i++ ) {
      printf("------ Submodel #%d ", i );
      Write_Opt_String( RdP->S[i].Name, BRACKET );
      printf(" in file %s ----------------\n", RdP->S[i].File_Name );
      dumpe_network( RdP->S[i].Submodel );
    }
  }
}

/****************************************************************/
/* Procedures de Verification					*/
/****************************************************************/
char	Find_Type(
PetriNet	*RdP,
Address		The_Address )
{
  int	i;
  char	Res;

  if ( The_Address == NULL )
    Res = 'X';
  else {
    switch( The_Address->Type ) {
    case 'P':
    case 'T':
    case 'I':
    case 'O':
      Res = The_Address->Type;
      break;
    case 'S':
      for ( i=0; (i < RdP->Nb_Submodels) 
	      && (The_Address->Number != RdP->S[i].Subm_Id );
	    i++ );
      if ( The_Address->Number == RdP->S[i].Subm_Id ) {
	/* found the submodel. Exploring it. */
	Res = Find_Type( RdP->S[i].Submodel, The_Address->Next );
      }
      else {
	printf("%s: error: illegal submodel number (%d) found",
	       TOOLNAME, The_Address->Number );
	printf(" in file %s.\n", RdP->File_Name );
	Res = 'X';
      }
      break;
    default:
      printf("%s: error: illegal address type (%c) found",
	     TOOLNAME, The_Address->Type );
      printf(" in file %s.\n", RdP->File_Name );
      Res = 'X';
    }
  }

  return Res;

} 

Address	Expand_Address( PetriNet	*RdP,
			Address		The_Address )
{
  int		i;
  Address	Bound_Address;
  Address	Result;

  Result = The_Address;

  if ( The_Address != NULL ) {
    switch( The_Address->Type ) {
    case 'P':
    case 'T':
      break;
    case 'I':
      if ( The_Address->Number <= RdP->Nb_Inputs ) {
	Bound_Address = RdP->Inputs[The_Address->Number-1];
	if ( ( Bound_Address->Type == 'P' )
	     || ( Bound_Address->Type == 'T' ) ) {
	  /* address is matched. Stop expansion here. */
	  The_Address->Type = Bound_Address->Type;
	  The_Address->Number = Bound_Address->Number;
	  The_Address->Next = (Address)NULL;
	  				/* which it should be already... */
	}
	else if ( Bound_Address->Type == 'S' ) {
	  /* address leads to a submodel. Have to explore it */
	  Result = Expand_Address( RdP, Bound_Address );
	}
      }
      else {
	printf("%s: error: illegal input number (%d) found",
	       TOOLNAME, The_Address->Number );
	printf(" in file %s.\n", RdP->File_Name );
	Result = (Address)NULL;
	/* this is a leak, but who cares... */
      }
      break;
    case 'O':
      if ( The_Address->Number <= RdP->Nb_Outputs ) {
	Bound_Address = RdP->Outputs[The_Address->Number-1];
	if ( ( Bound_Address->Type == 'P' )
	     || ( Bound_Address->Type == 'T' ) ) {
	  /* address is matched. Stop expansion here. */
	  The_Address->Type = Bound_Address->Type;
	  The_Address->Number = Bound_Address->Number;
	  The_Address->Next = (Address)NULL;
	  				/* which it should be already... */
	}
	else if ( Bound_Address->Type == 'S' ) {
	  /* address leads to a submodel. Have to explore it */
	  Result = Expand_Address( RdP, Bound_Address );
	}
      }
      else {
	printf("%s: error: illegal output number (%d) found",
	       TOOLNAME, The_Address->Number );
	printf(" in file %s.\n", RdP->File_Name );
	Result = (Address)NULL;
	/* this is a leak, but who cares... */
      }
      break;
    case 'S':
      for ( i=0; (i < RdP->Nb_Submodels) 
	      && (The_Address->Number != RdP->S[i].Subm_Id );
	    i++ );
      if ( The_Address->Number == RdP->S[i].Subm_Id ) {
	/* found the submodel. Exploring it. */
	Result->Next = Expand_Address( RdP->S[i].Submodel, 
				       The_Address->Next );
      }
      else {
	printf("%s: error: illegal submodel number (%d) found",
	       TOOLNAME, The_Address->Number );
	printf(" in file %s.\n", RdP->File_Name );
	Result = (Address)NULL;
      }
      break;
    default:
      printf("%s: error: illegal address type (%c) found",
	     TOOLNAME, The_Address->Type );
      printf(" in file %s.\n", RdP->File_Name );
      Result = (Address)NULL;
    }
  }

  return Result;
}

boolean Consistency_Checking( PetriNet *RdP )
{
  int		i,j;
  Liste_Link	L_Scan;
  Binding	B_Scan;
  Address	New_Address;
  Address	The_Add;
  Pl_Des	The_Place;
  boolean	Is_Ok;
  boolean	Found;
  boolean	*Visited = NULL;
  char		From_Type;
  char		To_Type;

  /* Checking the validity of bindings.					*/
  /* Valid bindings are:						*/
  /*  - before expansion:  P -> I, T -> I, O -> P, O -> T, O -> I.	*/
  /*  - after expansion:   P -> T, T -> P.				*/
  Is_Ok = TRUE;

  SCAN( B_Scan, RdP->Binding_Table ) {
    /* Checking unexpanded addresses, to see if I/O rules are respected.*/
    From_Type = Find_Type( RdP, B_Scan->From );
    To_Type = Find_Type( RdP, B_Scan->To );
  
    if ( ( From_Type == 'X' ) || ( From_Type == 'I' ) 
	 || ! ( ( ( From_Type == 'P' ) && ( To_Type == 'I' ) )
		|| ( ( From_Type == 'T' ) && ( To_Type == 'I' ) )
		|| ( ( From_Type == 'O' ) && ( To_Type == 'P' ) )
		|| ( ( From_Type == 'O' ) && ( To_Type == 'T' ) )
		|| ( ( From_Type == 'O' ) && ( To_Type == 'I' ) )
		|| ( ( From_Type == 'P' ) && ( To_Type == 'X' ) )
		|| ( ( From_Type == 'T' ) && ( To_Type == 'X' ) )
		|| ( ( From_Type == 'O' ) && ( To_Type == 'X' ) )
		)
	 ) {
      Is_Ok = FALSE;
      printf("%s: Inconsistency detected in binding ", TOOLNAME );
      Write_Address( B_Scan->From );
      printf(" -> ");
      Write_Address( B_Scan->To );
      printf(": %c -> %c\n", From_Type, To_Type );
    }

    if ( Is_Ok ) {
      /* Now expand addresses. First, expand the "from" address.	*/
      if ( Debug ) {
	printf("Expanding address ");
	Write_Address( B_Scan->From );
      }
      B_Scan->From = Expand_Address( RdP, B_Scan->From );
      if ( Debug ) {
	printf(" into ");
	Write_Address( B_Scan->From );
	printf("\n");
      }
      From_Type = Find_Type( RdP, B_Scan->From );

      /* Next, expand "to" address. But check first if it exists!	*/

      if ( B_Scan->To == (Address)NULL ) {
	/* this may happen for bindings of submodel objects to local 	*/
	/* (cf. Parse_Binding). Have to create an address with 		*/
	/* approproate type.						*/
	New_Address = (Address) malloc( sizeof( struct ADDRESS ) );
	if ( From_Type == 'P' )
	  New_Address->Type = 'T';
	else if ( From_Type == 'T' )
	  New_Address->Type = 'P';
	else 
	  New_Address->Type = 'X';
	New_Address->Number = B_Scan->Link->Dest;
	New_Address->Next = (Address)NULL;
	B_Scan->To = New_Address;
      }

      if ( Debug ) {
	printf("Expanding address ");
	Write_Address( B_Scan->To );
      }
      B_Scan->To = Expand_Address( RdP, B_Scan->To );
      if ( Debug ) {
	printf(" into ");
	Write_Address( B_Scan->To );
	printf("\n");
      }
      To_Type = Find_Type( RdP, B_Scan->To );

      if ( ! ( ( From_Type != 'X' ) && ( To_Type != 'X' )
	       && ( ( ( From_Type == 'P' ) && ( To_Type == 'T' ) )
		    || ( ( From_Type == 'T' ) && ( To_Type == 'P' ) )
		    )
	       ) ) {
	Is_Ok = FALSE;
	printf("%s: Inconsistency detected in binding ", TOOLNAME );
	Write_Address( B_Scan->From );
	printf(" -> ");
	Write_Address( B_Scan->To );
	printf(": %c -> %c\n", From_Type, To_Type );
      }
    }
  }

  /* Checking Competition/Priority Declarations */
  for ( i=0; i<RdP->Nb_Places; i++ ) {
    The_Place = RdP->P[i];
    if ( The_Place.Routing != NULL ) {
      Is_Ok = TRUE;

      switch( The_Place.Routing->Type ) {

      case 'P':

	if ( The_Place.Routing->Period != The_Place.Outdegree )
	  Is_Ok = FALSE;
	else {
	  /* Check that priorities and outputs coincide */
	  if ( Visited == NULL )
	    Visited = (boolean*) calloc( (unsigned)RdP->Nb_Trans,
					 sizeof(boolean) );

	  for ( j=0; j<RdP->Nb_Trans; j++ ) 
	    Visited[j] = FALSE;

	  SCAN( L_Scan, The_Place.Outputs ) {
	    if ( Visited[L_Scan->Dest] )
	      Is_Ok = FALSE;
	    else {
	      /* look for output in the "cycle" */
	      Found = FALSE;
	      for ( j=0; !Found && (j<The_Place.Routing->Period); j++ ) {
		The_Add = The_Place.Routing->Cycle[j];
		Found = ( ( ( The_Add->Type == 'P' ) 
			    || ( The_Add->Type == 'T' ) )
			  && ( The_Add->Number == L_Scan->Dest ) );
	      }
	      if ( !Found )
		Is_Ok = FALSE;
	      else
		Visited[L_Scan->Dest] = TRUE;
	    }
	  }
	  /* WARNING: the bindings should be checked too... TO BE DONE */
	}
	break;

      case 'D':
	if ( The_Place.Routing->Period < The_Place.Outdegree )
	  Is_Ok = FALSE;
	else {
	  /* Check that the cycle and the outputs coincide */
	  if ( Visited == NULL )
	    Visited = (boolean*) calloc( (unsigned)RdP->Nb_Trans,
					 sizeof(boolean) );

	  for ( j=0; j<RdP->Nb_Trans; j++ ) 
	    Visited[j] = FALSE;

	  /* checks that outputs are in the cycle */
	  SCAN( L_Scan, The_Place.Outputs ) {
	    Found = FALSE;
	    for ( j=0; !Found && (j<The_Place.Routing->Period); j++ ) {
	      The_Add = The_Place.Routing->Cycle[j];
	      Found = ( ( ( The_Add->Type == 'P' ) 
			  || ( The_Add->Type == 'T' ) )
			&& ( The_Add->Number == L_Scan->Dest ) );
	    }
	    if ( !Found )
	      Is_Ok = FALSE;
	    else
	      Visited[L_Scan->Dest] = TRUE;
	  }
	  /* checks that the cycle is in the output */
	  for ( j=0; j<The_Place.Routing->Period; j++ ) {
	    The_Add = The_Place.Routing->Cycle[j];
	    if ( ( ( The_Add->Type == 'P' ) 
		   || ( The_Add->Type == 'T' ) )
		 && ( ( The_Add->Number < 0 )
		      || ( The_Add->Number >= RdP->Nb_Trans )
		      || !Visited[The_Add->Number] ) )
	      Is_Ok = FALSE;
	  }
	  /* WARNING: the bindings should be checked too... TO BE DONE */
	}
	break;
      } /* end of switch( The_Place.Routing->Type ) */

      if ( !Is_Ok ) {
	printf("%s: Inconsistency detected in routing of place %d ", 
	       TOOLNAME, i );
	if ( strlen( The_Place.Name ) != 0 )
	  printf("(%s)\n", The_Place.Name );
	else
	  printf("\n");
      }
    }
  }

  if ( NULL != Visited )
    free( (char*)Visited );

  return Is_Ok;
}

/************************************************************************/
/* Functions to free the memory allocated to a Petri Net. This is 	*/
/* normally completely useless, but this is to calm down "purify"	*/
/************************************************************************/
void Free_Link_List( Liste_Link 	The_List )
{
  Liste_Link	L_Scan;
  Liste_Link	M_Scan;

  for ( L_Scan = The_List; L_Scan != NULL; ) {
    M_Scan = L_Scan;
    L_Scan = L_Scan->Next;
    if ( M_Scan->Data != NULL ) {
      if ( strlen( M_Scan->Data ) > 0 )
	free( M_Scan->Data );
    }

    free( (char*) M_Scan );
  }
}

void Free_Integer_List( Liste_Entier 	The_List )
{
  Liste_Entier	L_Scan;
  Liste_Entier	M_Scan;

  for ( L_Scan = The_List; L_Scan != NULL; ) {
    M_Scan = L_Scan;
    L_Scan = L_Scan->Next;
    free( (char*) M_Scan );
  }
}

void Free_Address_List( Liste_Address 	The_List )
{
  Liste_Address	L_Scan;
  Liste_Address	M_Scan;

  for ( L_Scan = The_List; L_Scan != NULL; ) {
    M_Scan = L_Scan;
    L_Scan = L_Scan->Next;
    free( (char*) M_Scan );
  }
}

void Free_Trans_List( Liste_Tr	The_List )
{
  Liste_Tr	L_Scan;
  Liste_Tr	M_Scan;

  for ( L_Scan = The_List; L_Scan != NULL; ) {
    M_Scan = L_Scan;
    L_Scan = L_Scan->Next;
    free( (char*) M_Scan );
  }
}
 
void Free_Place_List( Liste_Pl	The_List )
{
  Liste_Pl	L_Scan;
  Liste_Pl	M_Scan;

  for ( L_Scan = The_List; L_Scan != NULL; ) {
    M_Scan = L_Scan;
    L_Scan = L_Scan->Next;
    free( (char*) M_Scan );
  }
}

void Free_Petri_Net( PetriNet	*Reseau )
{
  int		i;

  for ( i=0; i<Reseau->Nb_Trans; i++ ) {
    free( Reseau->T[i].Name );
    free( Reseau->T[i].Data );
    Free_Law( Reseau->T[i].Tr_Law );
    Free_Link_List( Reseau->T[i].Inputs );
    Free_Link_List( Reseau->T[i].Outputs );
  }
  for ( i=0; i<Reseau->Nb_Places; i++ ) {
    free( Reseau->P[i].Name );
    free( Reseau->P[i].Data );
    Free_Law( Reseau->P[i].Pl_Law );
    free( Reseau->P[i].Lag_Times );
    if ( Reseau->P[i].Routing != NULL ) {
      if ( Reseau->P[i].Routing->Probas != NULL )
	free( Reseau->P[i].Routing->Probas );
      if ( Reseau->P[i].Routing->Cycle != NULL )
	free( Reseau->P[i].Routing->Cycle );
      free( Reseau->P[i].Routing );
    }
    Free_Link_List( Reseau->P[i].Inputs );
    Free_Link_List( Reseau->P[i].Outputs );
  }
  for ( i=0; i<Reseau->Nb_Submodels; i++ ) {
    Free_Petri_Net( Reseau->S[i].Submodel );
  }
  if ( Reseau->Nb_Inputs > 0 )
    /* there should be a recursive free of addresses, but well... */
    free( (char*)Reseau->Inputs );

  if ( Reseau->Nb_Outputs > 0 )
    free( Reseau->Outputs );

  free( (char*) Reseau->File_Name );
  free( (char*) Reseau->T );
  free( (char*) Reseau->P );
  if ( NULL != Reseau->S )
    free( (char*) Reseau->S );
  free( (char*) Reseau->Sim_Params );
  free( (char*) Reseau );

}
