/* Structures de donnees specifiques des RdP generaux		 	*/

#include "global.h"

#ifndef _PN_H_
#define _PN_H_
/************************************************************************/
/* structures for general Petri Nets 					*/
/************************************************************************/
typedef struct TR_DES {
	int		Tr_Id;
	char		*Name;
	char		*Data;
	int		Indegree;
	int		Outdegree;
	Liste_Link	Outputs;
	Liste_Link	Inputs;
	Law_Desc	Tr_Law;
  	int		Tag;
} Tr_Des;

typedef struct PL_DES {
	int		Pl_Id;
	char		*Name;
	char		*Data;
	int		Indegree;
	int		Outdegree;
	Liste_Link	Outputs;
	Liste_Link	Inputs;
	Law_Desc	Pl_Law;
	int		Init_Pop;
	double		*Lag_Times;
	Routing_Desc	Routing;
  	int		Tag;
} Pl_Des;

typedef struct PL_LIST {
	Pl_Des		*Val;
	struct PL_LIST	*Next;
} *Liste_Pl;

typedef struct TR_LIST {
	Tr_Des		*Val;
	struct TR_LIST	*Next;
} *Liste_Tr;

typedef struct Sim_Control {
	int		Nb_Firings;
} Sim_Control;

typedef struct PetriNet {
	int		Nb_Trans;	/* number of transitions	*/
	int		Nb_Places;	/* number of places		*/
  	int		Nb_Submodels;	/* number of submodels sons	*/
  	int		Nb_Inputs;	/* number of "visible" inputs	*/
  	int		Nb_Outputs;	/* number of "visible" outputs	*/
	Tr_Des		*T;		/* table of transitions		*/
	Pl_Des		*P;		/* table of places		*/
  	struct SUBM_DES	*S;		/* table of submodels 		*/
  	Address		*Inputs;	/* table of input bindings	*/
  	Address		*Outputs;	/* table of output bindings	*/
	boolean		*Adjacency;	/* matrice adjacence linearisee */
	boolean		*Adjacency0;	/* matrice places vides		*/
	int		Max_Pop_Init;
	Sim_Control	*Sim_Params;	/* ici faute de mieux		*/
  	char		*File_Name;
  	Binding		Binding_Table;
    	boolean		Has_Weights;
    	boolean		Has_Inhibitors;
} PetriNet;

typedef struct SUBM_DES {
  	int		Subm_Id;
	PetriNet	*Submodel;
	PetriNet	*Father;
  	char		*Name;
  	char		*File_Name;
  	int		First_Place;		/* for flattening	*/
  	int		First_Trans;		/* idem.		*/
} Subm_Des;

typedef struct SUBM_LIST {
  	Subm_Des		*Val;
	struct SUBM_LIST	*Next;
} *Liste_Subm;

typedef struct LISTES {
	Liste_Pl	Places;
	Liste_Tr	Transitions;
	Liste_Subm	Submodels;
} *Listes;

#endif
