/* programme/module de generation d'un programme C effectuant		*/
/* la simulation mixte PROSIT/equationelle d'un r\'eseau de Petri.	*/

#include <stdlib.h>
#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <math.h>

#include "pn.h"

#define	TOOLNAME	"pn2mixt"

#define INFINI		1e12
#define FILENAME_LG	256

#define PROSIT_DIR	"/u/chin/2/sloop3/gsiegel/Prosit_28_Jul"
#define SOURCE_CODE	"/u/chin/2/sloop3/gsiegel/Prosit_28_Jul/Src/Queueing/Sequential/Routing_Source.cc"
#define NULL_STRING	"<rien>"

/* tags for sorting out the "zones" for code generation */
#define EQ_OBJ		0
#define PROSIT_OBJ	1

#define max(x,y)        ( (x>y) ? x : y )
#define min(x,y)        ( (x<y) ? x : y )

#define SCAN(x,y)	for(x=y;x!=NULL;x=x->Next)
#define IRINT(x)	(int)rint(x)

typedef struct LAYER_DESC {
  int	Nb_Nodes;
  int	*Node_Id;
} *Layer_Desc;

typedef struct LAYER_STRUCT {
  int		Nb_Levels;
  Layer_Desc	*Level;		/* array of layer descriptions */
} *Layer_Struct;

/* Definitions and declarations for code generation */
int	Indent_Nb = 0;
char	*Prefix;
void	Indent( void );		/* declared below */

#define CODE(x)		printf(x)
#define CODE1(x,y)	printf(x,y)
#define CODE2(x,y,z)	printf(x,y,z)
#define CODE3(x,y,z,t)	printf(x,y,z,t)
#define CODE4(x,y,z,t,u)	printf(x,y,z,t,u)
#define CODELN		CODE("\n")
#define COMLINE		CODE("//============================================================\n");
#define IND2		if ( Prosit_Class ) CODE("  ");
#define IND4		if ( Prosit_Class ) CODE("    ");

/* Extern from ?? */
extern void Write_Address();

/* Externs from the module flatten.o */
extern PetriNet	*Flatten_PN( PetriNet* Reseau, boolean	Expand_Names );

/* Externs from the module parse_pn.o */
extern PetriNet		*Read_Network();
extern void		dumpe_network();
extern char		*todays_date();
extern void		Free_Petri_Net();
extern char		Find_Type();

/* Externs from the module analyze_pn.o */
extern boolean		topological_analysis();
extern boolean		Has_Simulable_Places();
extern boolean		Has_All_Deterministic_Places();
extern boolean		Has_Integral_Transitions(PetriNet *Reseau, 
						 boolean Testing,
						 boolean Sources_Also,
						 int *Bad_Objects );
extern boolean		Has_All_Routing_Places();
extern int		Find_Time_Scale(PetriNet *Reseau, boolean Testing,
					boolean	Sources_Also, int Scale,
					int *Bad_Objects );
extern int		Nb_Available_Tokens();
extern void		Dump_Layer_Struct();
extern double		Max_Lag();
extern double		Max_Delay();
extern void		Free_Layer_Struct();

/* Externs from the module gen_prosit.c */
extern void Gen_Class_File_Header( PetriNet *Reseau, char *Path, 
				   boolean Proxy );
extern void Gen_Class_Methods( PetriNet	*Reseau, int Max_Delay, int Scale,
			      char *Class_Name );
extern int  Gen_Class_Proxy_Prototype( char *Class_Name );
extern void Gen_Class_Prototype( PetriNet *Reseau, int 	Max_Dependency,
				 int Scale, char *Class_Name,
				 boolean Distributed, 
				 boolean Single_Marking_Function );
extern void Gen_Input_Handling(PetriNet	*Reseau, int Max_Dependency,
			       int Scale, char *Class_Name);
extern void Gen_Output_Handling(PetriNet *Reseau, int Max_Dependency,
				int Scale, char *Class_Name,
				boolean Distributed );
extern void Gen_Class_Cons_Des( PetriNet *Reseau, char *Class_Name, 
				boolean Distributed );

/* Forward declarations */
boolean	satisfies_constraints();
boolean is_recycling_place();
void	gen_prosit_code(PetriNet *Reseau, char Tmp_Dir[FILENAME_LG], 
			char File_Name[FILENAME_LG]);
void	gen_prosit_makefile();
void	gen_agat_code();

boolean Is_Dummy();
int	gen_c_code(PetriNet *Reseau, Layer_Struct The_Layer_Struct,
		   char  Tmp_Dir[FILENAME_LG], char  File_Name[FILENAME_LG],
		   boolean Prosit_Class, char  *Class_Name,
		   boolean Single_Marking_Function,
		   boolean QNAP_Style, boolean Distributed );
void	TimeScale( PetriNet *Petri, int Scale, boolean Prosit_Class );

int	Buf_Size = 1024;	/* default buffer size */
int	Scale = 10;		/* for integrality check */

int	*Hold_Time;
int	*Fire_Time;
boolean	*Is_Compet_Pl;		/* flag for places involved in a conflict */
boolean	*Is_Compet_Tr;		/* flag for transitions involved */
boolean	*Is_Source;
boolean	*Is_Recycling;
boolean	*Is_Interface_Pl;
boolean	*Is_Interface_Tr;
boolean	*Already_Visited;
boolean	*Delay_Table;
boolean	Has_Competition = FALSE;
boolean Y_Static = TRUE;
int 	*TSP;          		/* table of place time scales 	*/
int 	*TST;          		/* table of trans. time scales	*/
int	*The_Scales;		/* table of different time sc.	*/

boolean		*Place_Exists;
int		*Trans_Loop_Of;	/* stores the id of the trans a place	*/
				/* recycles (if it does).		*/
boolean		Optimize = FALSE;
int		Opt_Level;
boolean		Agat = FALSE;
boolean		No_Prosit = FALSE;
boolean		No_Stats = FALSE;
int		*Real_Indegree;
int		*Real_Outdegree;

/****************************************************************/
/* Various test functions					*/
/****************************************************************/
int	Nb_EQ_Trans_Succ(
PetriNet	*Reseau, 
int		Place_Id )
{
  Liste_Link	L_Scan;
  int		Nb = 0;

  SCAN( L_Scan, Reseau->P[ Place_Id ].Outputs )
    if ( Reseau->T[ L_Scan->Dest ].Tag == EQ_OBJ )
      Nb++;

  return Nb;
}

boolean	Has_EQ_Trans_Succ( Reseau, Place_Id )
PetriNet	*Reseau;
int		Place_Id;
{
  Liste_Link	L_Scan;
  boolean	Some;

  Some = FALSE;
  SCAN( L_Scan, Reseau->P[ Place_Id ].Outputs )
    Some = Some || ( Reseau->T[ L_Scan->Dest ].Tag == EQ_OBJ );

  return Some;
}

int	Nb_EQ_Place_Succ( Reseau, Trans_Id )
PetriNet	*Reseau;
int		Trans_Id;
{
  Liste_Link	L_Scan;
  int		Nb = 0;

  SCAN( L_Scan, Reseau->T[ Trans_Id ].Outputs )
    if ( Reseau->P[ L_Scan->Dest ].Tag == EQ_OBJ )
      Nb++;

  return Nb;
}

boolean	Has_EQ_Place_Succ( Reseau, Trans_Id )
PetriNet	*Reseau;
int		Trans_Id;
{
  Liste_Link	L_Scan;
  boolean	Some;

  Some = FALSE;
  SCAN( L_Scan, Reseau->T[ Trans_Id ].Outputs )
    Some = Some || ( Reseau->P[ L_Scan->Dest ].Tag == EQ_OBJ );

  return Some;
}

boolean	Has_Prosit_Trans_Succ( Reseau, Place_Id )
PetriNet	*Reseau;
int		Place_Id;
{
  Liste_Link	L_Scan;
  boolean	Some;

  Some = FALSE;
  SCAN( L_Scan, Reseau->P[ Place_Id ].Outputs )
    Some = Some || ( Reseau->T[ L_Scan->Dest ].Tag == PROSIT_OBJ );

  return Some;
}

boolean	Has_Prosit_Place_Succ( Reseau, Trans_Id )
PetriNet	*Reseau;
int		Trans_Id;
{
  Liste_Link	L_Scan;
  boolean	Some;

  Some = FALSE;
  SCAN( L_Scan, Reseau->T[ Trans_Id ].Outputs )
    Some = Some || ( Reseau->P[ L_Scan->Dest ].Tag == PROSIT_OBJ );

  return Some;
}

int Is_Input_Number( Reseau, The_Type, The_Number )
PetriNet	*Reseau;
char		The_Type;
int		The_Number;
{
  int	i;
  int	Res = -1;

  for ( i=0; i<Reseau->Nb_Inputs; i++ )
    if ( ( Reseau->Inputs[i]->Type == The_Type ) 
	 && ( Reseau->Inputs[i]->Number == The_Number ) )
      Res = i;

  return Res;
}

/****************************************************************/
/* Procedure to initialize various tables.			*/
/****************************************************************/
void Init_Generation( Reseau, Max_Delay, Scale )
PetriNet	*Reseau;
int		Scale;
{
  int		i;
  boolean	Some;
  int		Delay;
  int		Pl_Num;
  Liste_Link	L_Scan;
  Routing_Desc	The_Routing;

  Hold_Time = (int*) calloc( (unsigned)Reseau->Nb_Places, sizeof(int) );
  Fire_Time = (int*) calloc( (unsigned)Reseau->Nb_Trans, sizeof(int) );

  Delay_Table = (boolean*) calloc( (unsigned)(1+Max_Delay), 
				   sizeof(boolean) );
  Is_Compet_Pl = (boolean*) calloc( (unsigned)Reseau->Nb_Places, 
				   sizeof(boolean) );
  Is_Compet_Tr = (boolean*) calloc( (unsigned)Reseau->Nb_Trans, 
				   sizeof(boolean) );
  Is_Interface_Pl = (boolean*) calloc( (unsigned)Reseau->Nb_Places, 
				   sizeof(boolean) );
  Is_Interface_Tr = (boolean*) calloc( (unsigned)Reseau->Nb_Trans, 
				   sizeof(boolean) );
  Is_Source = (boolean*) calloc( (unsigned)Reseau->Nb_Trans, 
				   sizeof(boolean) );
  Is_Recycling = (boolean*) calloc( (unsigned)Reseau->Nb_Places, 
					  sizeof( boolean ) );
  Already_Visited = (boolean*) calloc( (unsigned)Reseau->Nb_Places, 
				       sizeof( boolean ) );

  /* initialization of properties of places.				*/
  for ( i=0 ; i<Reseau->Nb_Places; i++ ) {
    Hold_Time[i] = rint( Scale * Reseau->P[i].Pl_Law->Parameters[0] );
    Already_Visited[i] = FALSE;
    Is_Recycling[i] = is_recycling_place( &Reseau->P[i] );

    Some = FALSE;
    if ( Reseau->P[i].Tag == PROSIT_OBJ ) {
      SCAN( L_Scan, Reseau->P[i].Inputs ) {
        Some = Some || ( Reseau->T[L_Scan->Dest].Tag == EQ_OBJ );
      }
    }
    Is_Interface_Pl[i] = Some;
  }

  for ( i=0; i<Reseau->Nb_Trans; i++ ) {

    Is_Compet_Tr[i] = FALSE;
    SCAN( L_Scan, Reseau->T[i].Inputs ) {
      The_Routing = Reseau->P[L_Scan->Dest].Routing;
      if ( The_Routing != NULL ) {
	/* A competition is considered to occur if the routing is 	*/
	/* "Competition Priority" AND if there are at least two outputs.*/
	if ( ( The_Routing->Type == 'P' ) 
	     && ( Reseau->P[L_Scan->Dest].Outdegree > 1 ) ) {
	  Is_Compet_Tr[i] = TRUE;
	  Has_Competition = TRUE;
	}
      }
    }
    Some = FALSE;
    if ( Reseau->T[i].Tag == PROSIT_OBJ ) {
      SCAN( L_Scan, Reseau->T[i].Inputs ) {
        Some = Some || ( Reseau->P[L_Scan->Dest].Tag == EQ_OBJ );
      }
    }
    Is_Interface_Tr[i] = Some;

    /* find out if it is a source or not */
    Is_Source[i] = FALSE;
    if ( Reseau->T[i].Indegree == 0 )
      /* I assume this is meant to be a source... */
      Is_Source[i] = TRUE;
    else if ( Reseau->T[i].Indegree == 1 ) {
      Pl_Num = Reseau->T[i].Inputs->Dest;
      if ( ( Reseau->P[Pl_Num].Indegree == 1 ) 
	   && ( Reseau->P[Pl_Num].Inputs->Dest == i ) )
	Is_Source[i] = TRUE;
    }

    if ( !Is_Source[i] )
      Fire_Time[i] = rint( Scale * Reseau->T[i].Tr_Law->Parameters[0] );
    else
      Fire_Time[i] = 0.0;

    if ( Is_Compet_Tr[i] ) {
      /* if it is, generate Routing() code for **EACH** input transition */
      SCAN( L_Scan, Reseau->T[i].Inputs ) {
	Is_Compet_Pl[L_Scan->Dest] = TRUE;
	Has_Competition = TRUE;
      }
    }
  }

  for ( i=0; i<Max_Delay+1; i++ )
    Delay_Table[i] = FALSE;
  /* Always need Time-1 to ensure firing counts will increase.		 */
  Delay_Table[1] = TRUE;

  for ( i=0; i<Reseau->Nb_Places; i++ ) {
    SCAN( L_Scan, Reseau->P[i].Inputs ) {
      Delay = Hold_Time[ i ];
      if ( !Is_Source[ L_Scan->Dest ] ) 
	Delay += Fire_Time[ L_Scan->Dest ];
      Delay_Table[ Delay ] = TRUE;
    }
  }

  /* structures for time scale analysis */
  TSP = (int *)calloc( (unsigned)Reseau->Nb_Places, sizeof( int ) );
  TST = (int *)calloc( (unsigned)Reseau->Nb_Trans, sizeof( int ) );

}

/* Prosit objects are recognised by the fact that their names begin	*/
/* with the string "PROSIT"						*/
/* If Prosit_Class is true, all objects are tagged "equational".	*/
void Tag_Objects( Reseau, Prosit_Class )
PetriNet	*Reseau;
boolean		Prosit_Class;
{
  int		i;

  for ( i = 0; i < Reseau->Nb_Places; i++ ) {
    if ( !Prosit_Class 
	 && !strncmp( "PROSIT|", Reseau->P[i].Name, 7 ) ) {
      Reseau->P[i].Tag = PROSIT_OBJ;
      if ( Debug )
	printf("Place #%d (%s) tagged Prosit\n", i, Reseau->P[i].Name );
    }
    else 
      Reseau->P[i].Tag = EQ_OBJ;
  }
  for ( i = 0; i < Reseau->Nb_Trans; i++ ) {
    if ( !Prosit_Class 
	 && !strncmp( "PROSIT|", Reseau->T[i].Name, 7 ) ) {
      Reseau->T[i].Tag = PROSIT_OBJ;
      if ( Debug )
	printf("Trans #%d (%s) tagged Prosit\n", i, Reseau->T[i].Name );
    }
    else 
      Reseau->T[i].Tag = EQ_OBJ;
  }
}

/* Cleanup. Solely to please "Purify" */
void Free_Memory( )
{
  free( (char*) Hold_Time );
  free( (char*) Fire_Time );
  free( (char*) Delay_Table );
  free( (char*) Is_Compet_Pl );
  free( (char*) Is_Compet_Tr );
  free( (char*) Is_Interface_Pl );
  free( (char*) Is_Interface_Tr );
  free( (char*) Is_Source );
  free( (char*) Already_Visited );
  free( (char*) TSP );
  free( (char*) TST );
}

/****************************************************************/
/* Procedures for code generation...				*/
/****************************************************************/
void	Gen_EQ_File_Header( Reseau )
PetriNet	*Reseau;
{
  char	*The_Date;

  CODE("/* File for the Equational simulation of a Petri net \t\t*/\n");
  CODE("/* generated by ");
  CODE(TOOLNAME); CODE(" version "); CODE(VERSION);
  if ( strlen( Reseau->File_Name ) > 0 )
    CODE1("from file %s", Reseau->File_Name );
  CODE("\t*/\n");
  The_Date = todays_date();
  CODE1("/* Date: %s \t\t\t\t*/\n\n", The_Date );
  free( The_Date );
}

void	Gen_EQ_Includes(
			PetriNet	*Reseau,
			boolean		Single_Class,
			boolean		Distributed )
{
  /* write definitions, includes */
  CODE("#include <stdio.h>\n");
  CODE("#include <iostream.h>\n");
  CODE("#include <iomanip.h>\n");
  CODE("#include <MLCG.h>\n");
  CODE("#include <Uniform.h>\n");
  if ( ! No_Prosit && ! Distributed )
    CODE("#include \"Routing_Server.hh\"\n" );
  else if ( !Distributed )
    CODE("#include <bool.h>\n");

  /* boolean definitions are implied by the above */
  /*   CODELN; */
  /*   CODE("#define TRUE    1\n"); */
  /*   CODE("#define FALSE   0\n"); */
  CODELN;
  /* These "variables" have no apparent utility, at least in the	*/
  /* single class case.							*/
  if ( !Single_Class ) {
    CODE1("int     TMAX = %d;\n", Scale*Reseau->Sim_Params->Nb_Firings );
    CODE1("int     SCALE = %d;\n", Scale );
  }
}



void Gen_Source_Code( Reseau, Prosit_Class, Class_Name )
PetriNet	*Reseau;
boolean		Prosit_Class;
char		*Class_Name;
{
  int		i;
  int		Delay;
  int		Nb_Tokens;
  int		Nb_Sources;
  double	Proba;
  Law_Desc	The_Law;
  int		Input_Num;
  boolean	Could_Be_Connected;

  /* check first if there are some non tagged sources */
  Nb_Sources = 0;
  for ( i=0; i<Reseau->Nb_Trans; i++ )
    if ( ( Reseau->T[i].Tag == EQ_OBJ ) && Is_Source[i] )
      Nb_Sources++;

  if ( Nb_Sources > 0 ) {
    if ( Prosit_Class ) {
      CODE("// Procedure for computing the number of tokens generated ");
      CODE("by the sources\n");
      CODE1("void %s::", Class_Name );
    }
    else {
      CODE("/* Procedure for computing the number of tokens generated ");
      CODE("by the sources */\n");
      CODE("void ");
    }
    CODE("Re_Init_Sources( int From, int To )\n");
    CODE("{\n  int\ti, j;\n");

    for ( i=0; i<Reseau->Nb_Trans; i++ ) {
      if ( ( Reseau->T[i].Tag == EQ_OBJ ) && Is_Source[i] ) {
	CODE2("  /* Transition %d (%s) */\n", i, Reseau->T[i].Name );
	Input_Num = Is_Input_Number( Reseau, 'T', i );
	Could_Be_Connected = ( Input_Num != -1 ) && Prosit_Class;

	if ( Could_Be_Connected ) {
	  CODE1("  if ( !Is_Input_Trans[%d] )\n", i );
	}
	The_Law = Reseau->T[i].Tr_Law;
	CODE("    for( i = From; i < To; i++ ) {\n");

	if ( Reseau->T[i].Indegree == 0 )
	  /* assume simple loop */
	  Nb_Tokens = 1;
	else 
	  Nb_Tokens = Reseau->P[Reseau->T[i].Inputs->Dest].Init_Pop;

	if ( Nb_Tokens > 1 ) {
	  CODE1("    for( j = 0; j < %d; j ++ ) {\n", Nb_Tokens );
	  Indent_Nb += 2;
	}

	switch( The_Law->Name ) {
	case 'C':
	  Delay = IRINT( The_Law->Parameters[0] * Scale );
	
	  if ( Delay == 0 ) {
	    /* This is a saturating source! */
	    CODE1("    Nb_Tok_%d = Infinity;\n", i );
	  }
	  else {
	    CODE1("    T_Phase_%d --;\n", i );
	    CODE1("    if ( T_Phase_%d == 0 ) {\n", i );
	    CODE1("      Nb_Tok_%d ++;\n", i );
	    CODE2("      T_Phase_%d = %d;\n", i, Delay );
	    CODE ("    }\n");
	  }
	  break;

	case 'G':
	case 'B':	
	  Proba = The_Law->Parameters[0];

	  if ( Scale != 1 ) {
	    CODE1("    T_Phase_%d --;\n", i );
	    CODE1("    if ( T_Phase_%d == 0 ) {\n", i );
	  }
	  CODE1("      Nb_Tok_%d += ", i );
	  if ( The_Law->Name == 'B' )
	    CODE("1 + ");
	  CODE1(" (int)floor( - log((*U_0_1)())/ %12.6f );\n", -log(1.0-Proba) );
	  if ( Scale != 1 ) {
	    CODE2("      T_Phase_%d = %d;\n", i, Scale );
	    CODE("    }\n");
	  }

	  break;

	case 'D':
	  CODE1("    T_Phase_%d --;\n", i );
	  CODE1("    if ( T_Phase_%d == 0 ) {\n", i );
	  CODE1("      Nb_Tok_%d ++;\n", i );
	  CODE1("      if( T_Pr_Zero_%d > 0.0 )\n", i );
	  CODE2("        Nb_Tok_%d += (int)floor( log((*U_0_1)())/ log(T_Pr_Zero_%d ) );\n",
		i, i );
	  CODE( "      U = (*U_0_1)();\n");
	  CODE1("      for( j=0; U > T_Proba_%d[j]; j++ );\n", i );
	  CODE2("      T_Phase_%d = Interarrival_%d[j];\n", i, i );
	  CODE( "    }\n");
	
	  break;

	default:
	  printf("%s: warning: unknown law type at source %d\n",
		 TOOLNAME, i );

	}
	CODE2("    R_%d[i] = Nb_Tok_%d;\n", i, i );

	if ( Nb_Tokens > 1 ) {
	  Indent_Nb -= 2;
	  CODE("    }\n");
	}
	CODE("  }\n");
	CODELN;
      }
    }

    CODE("}\n");
  }
}

void Gen_EQ_Stats_Code( PetriNet	*Reseau,
			boolean	Prosit_Class,
			char	*Class_Name,
			boolean	QNAP_Style )
{
  int		i;
  boolean	stat_ok = TRUE;	/* is there for Prosit compatibility	*/
				/* but is not used yet.			*/
  int		Input_Num;
  boolean	Is_True_Source;
  boolean	Could_Be_Connected;

  CODE("/* Declarations for random number generation");
  CODE("(for routing and sources)*/\n");
  CODE("static Uniform *U_0_1 = new Uniform( 0.0, 1.0, new MLCG( 12, 123 ) );\n");
  CODELN;

  CODE("/* Declaration of procedures for collecting statistics on the\t*/\n");
  CODE("/* EQ object.\t\t\t\t\t\t\t*/\n");

  if ( Prosit_Class )
    CODE1("void %s::", Class_Name );
  else 
    CODE("void ");

  CODE("Compute_Stat( ) {\n");
  CODE("}\n");
  CODELN;

  if ( Prosit_Class )
    CODE1("void %s::", Class_Name );
  else 
    CODE("void ");
  CODE("Report_Trans_Stat(\n");
  CODE("  ostream &Out_File,\n");
  CODE("  int     Id,\n");
  CODE("  char    *Hi_Name,\n");
  CODE("  int	  Count,\n");
  CODE("  int	  Final_Time ) {\n");
  CODELN;
  if ( QNAP_Style ) {
    CODE("  Out_File.setf(ios::left,ios::adjustfield);\n");
    CODE("  Out_File << \" * \";\n");
    CODE("  if ( strlen(Hi_Name) )\n");
    CODE("    Out_File << Hi_Name << \"/\" << setw(59-strlen(Hi_Name))");
    CODE(" << Trans_Name[Id];\n");
    CODE("  else\n");
    CODE("    Out_File << \" * \" << setw(60) << Trans_Name[Id];\n");
    CODE("  Out_File << \" * \";\n");
    CODE("  Out_File.setf(ios::right,ios::adjustfield);\n");
    CODE("  Out_File << setw(7) << Count << \" *\" << endl;\n");
    CODE("  Out_File << \" * \" << setw(60) << \" \";\n");
    CODE("  Out_File << \" *\" << endl;\n");
  }
  else {
    CODE("  Out_File << \"---------------------------------------");
    CODE("-----------\" << endl;\n");
    CODE("  Out_File << \" Station Transition \";\n");
    CODE("  if ( strlen(Hi_Name) )\n");
    CODE("    Out_File << Hi_Name << \"/\" << Trans_Name[Id];\n");
    CODE("  else\n");
    CODE("    Out_File << Trans_Name[Id];\n");
    CODE("  Out_File << \" #EQ/\" << Id << \"   t=\" ");
    CODE("<< Final_Time << endl << endl;\n" );
    CODE("  Out_File << \" Mean departure throughput : \";\n");
    CODE("  Out_File << (double)Count / Final_Time << endl;\n");
    CODE("  Out_File << \" Number of services : \" << Count << endl;\n");
  }

  CODE("}\n");
  CODELN;

  if ( Prosit_Class )
    CODE1("void %s::", Class_Name );
  else 
    CODE("void ");
  
  CODE("Report_Stat( ostream &Out_File ) {\n");
  CODELN;
  CODE("  int\tFinal_Time = Real_EQ_Time - 1;\n");
  CODE1("  int\tFinal_EQ_Time = Final_Time %% %sBUF_SIZE;\n", Prefix);
  CODE("  char\t*S=\"Source \";\n");
  CODE("  char\t*T=\"Transition \";\n");
  CODE("  char\t*G=\"generation \";\n");
  CODE("  char\t*D=\"departure \";\n");
  CODELN;

  for ( i=0; i<Reseau->Nb_Trans; i++ ) 
    if ( Reseau->T[i].Tag == EQ_OBJ ) {

      Input_Num = Is_Input_Number( Reseau, 'T', i );
      Could_Be_Connected = ( Input_Num != -1 ) && Prosit_Class;
      Is_True_Source = Is_Source[i] && !Could_Be_Connected;

      if ( !No_Stats ) {

	if ( !Is_Source[i] ) {
	  CODE1("  Report_Trans_Stat( Out_File, %d,", i );
	  CODE1(" \"%s\", ", ( Prosit_Class ? Class_Name : "" ) );
	  CODE1("\n\t\t\tZ[Final_EQ_Time][%d], Final_Time );\n", i );
	  CODELN;
	}
	else {
	  if ( QNAP_Style ) {
	    CODE1("  Report_Trans_Stat( Out_File, %d,", i );
	    CODE1(" \"%s\", ", ( Prosit_Class ? Class_Name : "" ) );
	    if ( Is_True_Source )
	      CODE1("\n\t\t\tR_%d[Final_EQ_Time], Final_Time );\n", i );
	    else
	      CODE1("\n\t\t\tZ[Final_EQ_Time][%d], Final_Time );\n", i );
	    CODELN;
	  }
	  else {
	    CODE("  Out_File << \"---------------------------------------");
	    CODE("-----------\" << endl;\n");
	    CODE("  Out_File << \" Station \" << ");
	    if ( Is_True_Source )
	      CODE("\"Source \";\n");
	    else if ( !Is_Source[i] )
	      CODE("\"Transition \";\n");
	    else
	      CODE1("( !Is_Input_Trans[%d] ? S : T );\n", i );
	    CODE("  Out_File << \"");
	    if ( Prosit_Class )
	      CODE1("%s/", Class_Name );
	    CODE1("%s ", Reseau->T[i].Name );
	    CODE1("#EQ/%d   t=\";\n", i );
	    CODE("  Out_File << Final_Time << endl << endl;\n" );

	    if( stat_ok ) {
	      CODE("  Out_File << \" Mean \" << ");
	      if ( Is_True_Source ) {
		CODE("\"generation throughput : \";\n");
		CODE1("  Out_File << (double)R_%d[Final_EQ_Time] ", i );
	      }
	      else if ( !Is_Source[i] ) {
		CODE("\"departure throughput : \";\n");
		CODE1("  Out_File << (double)Z[Final_EQ_Time][%d] ", i );
	      }
	      else {
		CODE1("( !Is_Input_Trans[%d] ? G : D )", i );
		CODE(" << \"throughput : \";\n");
		CODE("  Out_File << ");
		CODE2("( !Is_Input_Trans[%d] ? (double)R_%d[Final_EQ_Time]",
		      i, i );
		CODE1(" : (double)Z[Final_EQ_Time][%d] ) ", i );
	      }
	      CODE("/ (double)Final_Time;\n");
	      CODE("  Out_File << endl;\n");
	      CODE("  Out_File << \" Number of ");
	      if ( Is_True_Source ) {
		CODE("generated customers : \";\n");
		CODE1("  Out_File << R_%d[Final_EQ_Time] << endl;\n", i );
	      }
	      else if ( !Is_Source[i] ) {
		CODE("services : \";\n");
		CODE1("  Out_File << Z[Final_EQ_Time][%d] << endl;\n", i );
	      }
	      else {
		CODE("\";\n");
		CODE1("  Out_File << ( !Is_Input_Trans[%d] ? ", i );
		CODE("\"generated customers : \" : \"services : \" );\n");
		CODE1("  Out_File << ( !Is_Input_Trans[%d] ? ", i );
		CODE2("R_%d[Final_EQ_Time] : Z[Final_EQ_Time][%d] );\n", i, i );
		CODE("  Out_File << endl;\n");
	      }
	    }
	    else {
	      CODE("  Out_File << \"Error !! No statistics available\" ");
	      CODE("<< endl;\n");
	    }
	    CODELN;
	  }
        }
      }
      else {
	CODE2("  Out_File << \"%c_%d[\" << Final_Time << \"] =   \";\n",
	     ( Is_True_Source ? 'R' : 'Z' ),
	     i );
	if ( Is_True_Source )
	  CODE1("  Out_File << R_%d[Final_EQ_Time];\n", i );
	else
	  CODE1("  Out_File << Z[Final_EQ_Time][%d];\n", i );
	CODE("  Out_File << \"    Throughput:   \" << ");
	if ( Is_True_Source )
	  CODE1("(double)R_%d[Final_EQ_Time] / (double)Final_Time;\n", i );
	else
	  CODE1("(double)Z[Final_EQ_Time][%d] / (double)Final_Time;\n", i );
	CODE1("  Out_File << \"  (%s)\" << endl;\n", Reseau->T[i].Name );

	CODELN;
      }
    }
  /* no stats for places yet. */
  CODE("}\n");
  CODELN;
}


void Gen_Marking_Computation(
PetriNet	*Reseau,
int		Place_Id,
boolean		Initial,
boolean		Prosit_Class,
boolean		Single_Marking_Function )
{
  Liste_Link	L_Scan;
  Pl_Des	The_Place;
  int		Delay;
  int		Input_Num;
  boolean	Could_Be_Connected;

  The_Place = Reseau->P[Place_Id];

  if ( !Single_Marking_Function ) {
    if ( Initial )
      CODE1("  Res = Avail[%d][ The_Time ];\n", Place_Id );
    else 
      CODE1("  Res = %d;\n", Reseau->P[Place_Id].Init_Pop );
  }
  else if ( !Initial ) {
    if ( Reseau->P[Place_Id].Init_Pop > 0 )
      CODE1("  Res = %d;\n", Reseau->P[Place_Id].Init_Pop );
  }

  SCAN( L_Scan, The_Place.Inputs ) {

    Input_Num = Is_Input_Number( Reseau, 'T', L_Scan->Dest );
    Could_Be_Connected = ( Input_Num != -1 ) && Prosit_Class;

    CODE("  Res += ");
    Delay = Hold_Time[Place_Id];
    if ( !Is_Source[L_Scan->Dest] )
      Delay += Fire_Time[L_Scan->Dest];

    if ( Delay == 0 ) {
      if ( Is_Source[L_Scan->Dest] ) {
	if ( Could_Be_Connected )
	  CODE1("( !Is_Input_Trans[%d] ? ", L_Scan->Dest );	
	CODE2( "R_%d[ %sTime ]", L_Scan->Dest,
	       ( Initial ? "The_" : "EQ_" ) );
	if ( Could_Be_Connected )
	  CODE2( ": Y[ %sTime ][%d] ) ", 
		 ( Initial ? "The_" : "EQ_" ),
		 L_Scan->Dest );
      }
      else if ( Y_Static && !Initial )
	/* "Initial" turns off Ystatic (too messy...) */
	CODE1( "Y_0[%d]", L_Scan->Dest );
      else
 	CODE2( "Y[ %sTime ][%d]", 
	       ( Initial ? "The_" : "EQ_" ),
	       L_Scan->Dest );
    }
    else if ( Initial ) {
      CODE1("( The_Time >= %d ? ", Delay );
      if ( Is_Source[L_Scan->Dest] ) {
	if ( Could_Be_Connected )
	  CODE1("( !Is_Input_Trans[%d] ? ", L_Scan->Dest );
	CODE2( "R_%d[ ( The_Time - %d + ",
	       L_Scan->Dest, Delay );
	CODE2("%sBUF_SIZE ) %% %sBUF_SIZE ]", Prefix, Prefix );
	if ( Could_Be_Connected )
	  CODE4( ": Y[ (The_Time-%d+%sBUF_SIZE) %% %sBUF_SIZE][%d] ) ",
		 Delay, Prefix, Prefix, L_Scan->Dest );
      }
      else {
 	CODE2( "Y[ ( The_Time - %d + %sBUF_SIZE ) ", Delay, Prefix );
	CODE2( "%% %sBUF_SIZE ][%d]", Prefix, L_Scan->Dest );
      }
      CODE(" : 0 )");
    }
    else {
      if ( Is_Source[L_Scan->Dest] ) {
	if ( Could_Be_Connected )
	  CODE1("( !Is_Input_Trans[%d] ? ", L_Scan->Dest );
	CODE2( "R_%d[ EQ_Time_M_%d ]", L_Scan->Dest, Delay );
	if ( Could_Be_Connected )
	  CODE2( ": Y[EQ_Time_M_%d][%d] ) ", Delay, L_Scan->Dest );
      }
      else if ( Y_Static )
	CODE2( "Y_Time_M_%d[%d]", Delay, L_Scan->Dest );
      else
 	CODE2( "Y[ EQ_Time_M_%d ][%d]", Delay, L_Scan->Dest );
    }
    if ( L_Scan->Weight > 1 )
      CODE1(" * %d", L_Scan->Weight );
    CODE(";\n");
  }
  SCAN( L_Scan, The_Place.Outputs ) {
    if ( Y_Static && !Initial )
      CODE1("  Res -= Y_0[%d]", L_Scan->Dest );
    else
      CODE2("  Res -= Y[ %sTime ][%d]",
	    ( Initial ? "The_" : "EQ_" ),
	    L_Scan->Dest );
    if ( L_Scan->Weight > 1 )
      CODE1(" * %d;\n", L_Scan->Weight );
    else
      CODE(";\n");
  }
}

void Gen_Marking_Function( 
PetriNet	*Reseau,
int		Place_Id,
boolean		Initial,
boolean		Prosit_Class,
char		*Class_Name,
boolean		Single_Marking_Function )
{

  if ( Prosit_Class )
    CODE1("int %s::", Class_Name );
  else
    CODE("int ");
  CODE3("%sMarking_%d( %s )\n{\n  register int Res;\n\n", 
	( Initial ? "I_" : "" ),
	Place_Id,
	( Initial ? "int The_Time" : "" ) );

  Gen_Marking_Computation( Reseau, Place_Id, Initial, Prosit_Class,
			   Single_Marking_Function );

  CODE("  return Res;\n}\n\n");
}

void Gen_Single_Marking_Function(
PetriNet	*Reseau,
boolean		Initial,
boolean		Prosit_Class,
char		*Class_Name  )
{
  int	i;

  if ( Prosit_Class )
    CODE("// Function to compute the marking of places\n" );
  else
    CODE("/* Function to compute the marking of places */\n" );
    
  if ( Prosit_Class )
    CODE1("int %s::", Class_Name );
  else
    CODE("int ");

  CODE2("%sMarking( int Place_Id%s )\n", 
	( Initial ? "I_" : "" ),
	( Initial ? ", int The_Time" : "" ) );
  CODE("{\n");
  CODE("  register int Res;\n");
  CODELN;
  if ( Initial )
    CODE("  Res = Avail[ Place_Id ][ The_Time ];\n");
  else
    CODE("  Res = 0;\n");
  CODELN;
  CODE("  switch( Place_Id ) {\n");
  CODELN;

  for ( i=0; i<Reseau->Nb_Places; i++ )
    if ( ( Reseau->P[i].Tag == EQ_OBJ ) 
	 && ( Is_Compet_Pl[i] ) ) {
      CODE2("  case %d: /* place %s */\n", i, Reseau->P[i].Name );
      Gen_Marking_Computation( Reseau, i, Initial, Prosit_Class, TRUE );
      CODE("  break;\n");
      CODELN;
    }

  CODE("  } /* end of switch */\n");
  CODELN;
  CODE("  return Res;\n");
  CODE("}\n");
  CODELN;
}

void Gen_Marking_Code(
PetriNet	*Reseau,
boolean		Prosit_Class,
char		*Class_Name,
boolean		Single_Marking_Function )
{
  int		i;

  if ( Single_Marking_Function ) {
    Gen_Single_Marking_Function( Reseau, TRUE , Prosit_Class, Class_Name );
    Gen_Single_Marking_Function( Reseau, FALSE, Prosit_Class, Class_Name );
  }
  else {
    for ( i=0; i<Reseau->Nb_Places; i++ )
      if ( ( Reseau->P[i].Tag == EQ_OBJ ) 
	   && ( Is_Compet_Pl[i] ) ) {
	/* generate Marking code 					*/
	if ( Prosit_Class )
	  CODE2("// Function to compute the marking of place %d (%s)\n",
		i, Reseau->P[i].Name );
	else
	  CODE2("/* Function to compute the marking of place %d (%s) */\n",
		i, Reseau->P[i].Name );
	Gen_Marking_Function( Reseau, i, TRUE , Prosit_Class, Class_Name,
			      Single_Marking_Function );
	Gen_Marking_Function( Reseau, i, FALSE, Prosit_Class, Class_Name,
			      Single_Marking_Function );
      }

    CODELN;
  }
}

void	Gen_EQ_Variables( Reseau, Max_Delay, Scale, Prosit_Class, Class_Name )
PetriNet	*Reseau;
int		Max_Delay;
int		Scale;
boolean		Prosit_Class;
char		*Class_Name;
{

  int		i;
  int		j;
  int		k;
  int		l;
  int		Nb_Params;
  int		Max_Degree;
  int		Nb_Values;
  int		Nb_Inserted;
  double	Pr0;
  double	Proba;
  double	Cum_Proba;
  double	*The_Probas;
  double	*The_Values;
  Liste_Link	L_Scan;
  Routing_Desc	The_Routing;
  Law_Desc	The_Law;


  IND2;CODE ("/* table Y of initiated firings\t\t\t\t*/\n");
  IND2;CODE2("int     Y[%sBUF_SIZE][%sNB_TRANS];\n", Prefix, Prefix );
  IND2;CODE ("/* table Z of completed firings\t\t\t\t*/\n");
  IND2;CODE2("int     Z[%sBUF_SIZE][%sNB_TRANS];\n", Prefix, Prefix );
  IND2;CODE ("int     Nb_Tokens;\n");

  /* Defining the number of available tokens for the MAX_DELAY first	*/
  /* units of time. Used at the beginning of the simulation.		*/
  if ( Reseau->Nb_Places > 0 ) {
    if ( Prosit_Class ) {
      IND2;CODE("// table of initial availabilities\n");
      IND2;CODE2("int\tAvail[%sNB_PLACES][%sMAX_DELAY];\n",
		 Prefix, Prefix );

      /* C++ forbids initialization here. Done in Init() */
    }
    else {
      CODE("/* table of initial availabilities\t\t\t\t*/\n");
      CODE2("int\t\t\tAvail[%sNB_PLACES][%sMAX_DELAY] = {\n",
	    Prefix, Prefix );

      for( i=0; i<Reseau->Nb_Places-1; i++ ) {
	CODE("\t");
	for( j=0; j<Max_Delay; j++ ) {
	  CODE1("%2d,", Nb_Available_Tokens( Reseau->P[i], j ) );
	}
	CODE("\n");
      }
      CODE("\t");
      i = Reseau->Nb_Places-1;
      for( j=0; j<Max_Delay-1; j++ ) {
	CODE1("%2d,", Nb_Available_Tokens( Reseau->P[i], j ) );
      }
      CODE1("%2d", Nb_Available_Tokens( Reseau->P[i], Max_Delay-1 ) );
      
      CODE("\n};\n\n");
    }
  }

  if ( Reseau->Nb_Trans > 0 ) {
    if ( Prosit_Class ) {
      CODE("  // table of firing times\n");
      CODE1("  int\tFire_Time[%sNB_TRANS];\n", Prefix );
    }
    else {
      CODE("/* table of firing times\t\t\t\t\t*/\n");
      CODE1("int\t\t\tFire_Time[%sNB_TRANS] = {\n", Prefix );

      for( i=0; i<Reseau->Nb_Trans-1; i++ )
	CODE1("\t%2d,\n", Fire_Time[i] );

      i = Reseau->Nb_Trans-1;
      CODE1("\t%2d", Fire_Time[i] );
      CODE("\n};\n\n");
    }
  }

  if ( Reseau->Nb_Trans > 0 ) {
    if ( Prosit_Class ) {
      CODE("  // table of source flags\n");
      CODE1("  int\tIs_Source[%sNB_TRANS];\n", Prefix );
   }
    else {
      CODE("/* table of source flags\t\t\t\t\t*/\n");
      CODE1("int\t\t\tIs_Source[%sNB_TRANS] = {\n", Prefix );

      for( i=0; i<Reseau->Nb_Trans-1; i++ )
	CODE1("\t%s,\n", ( Is_Source[i] ? "TRUE" : "FALSE" ) );

      i = Reseau->Nb_Trans-1;
      CODE1("\t%s\n", ( Is_Source[i] ? "TRUE" : "FALSE" ) );
      CODE("};\n\n");
    }
  }

  if ( Reseau->Nb_Trans > 0 ) {
    if ( Prosit_Class ) {
      CODE("  // table of transition names\n");
      CODE1("  char\t*Trans_Name[%sNB_TRANS];\n", Prefix );
    }
    else {
      CODE("/* table of transition names\t\t\t\t\t*/\n");
      CODE1("char\t\t\t*Trans_Name[%sNB_TRANS] = {\n", Prefix );

      for( i=0; i<Reseau->Nb_Trans-1; i++ )
	if ( strlen( Reseau->T[i].Name ) > 0 )
	  CODE1("\t\"%s\",\n", Reseau->T[i].Name );
	else
	  CODE("\t\"\",\n" );

      i = Reseau->Nb_Trans-1;
      if ( strlen( Reseau->T[i].Name ) > 0 )
	CODE1("\t\"%s\",\n", Reseau->T[i].Name );
      else
	CODE("\t\"\"\n" );
      CODE("\n};\n\n");
    }
  }

  if ( Prosit_Class )
    CODE("  // variable for computing enabling degree of transitions\n");
  else
    CODE("/* variable for computing enabling degree of transitions */\n");
  IND2;CODE("int\tE_Degree;\n");

  if ( Prosit_Class ) {
    CODE("  // variables for use with the circular buffer\n");
    CODE("  int\tReal_EQ_Time;\n");
    CODE("  int\tEQ_Time;\t\t// The time, modulo the buffer size/\n");
  }
  else {
    CODE("/* variables for use with the circular buffer\t\t\t\t*/\n");
    CODE("int\tReal_EQ_Time = 0;\n");
    CODE("int\tEQ_Time = 0;\t\t/* The time, modulo the buffer size\t*/\n");
  }

  for (i=1; i<=Max_Delay; i++ )
    if ( Delay_Table[i] ) {
      IND2;CODE2("int\tEQ_Time_M_%d;\t/* Time - %d */\n", i, i );
    }
  CODELN;
  if ( Y_Static ) {
    IND2;CODE("int\t*Y_0;\t\t/* Y[ EQ_Time ] */\n");
    for (i=1; i<=Max_Delay; i++ )
      if ( Delay_Table[i] || (i == 1 ) ) {
	/* Y_Time_M_1 is needed in all cases because of the scaling system */
	IND2;CODE2("int\t*Y_Time_M_%d;\t/* Y[ EQ_Time - %d ] */\n", i, i );
      }
    CODELN;
  }

  if ( Prosit_Class ) {
    /* In this case, declarations are without initializations.		*/
    /* On the other hand, all objects are supposed to be tagged.	*/

    CODE("  // tables and counts for routing generation\n");
    for( i=0; i<Reseau->Nb_Places; i++ ) {

      The_Routing = Reseau->P[i].Routing;

      if ( The_Routing != NULL ) {
	switch ( The_Routing->Type ) {

	case 'R':
	  CODE1("  int Count_%d;\t", i );
	  CODE2("  double P_Proba_%d[%d];\t", i, Reseau->P[i].Outdegree );
	  CODE2("  int Dest_%d[%d];\n", i, Reseau->P[i].Outdegree );

	  break;

	case 'D':
	  CODE1("  int Count_%d;\t", i );
	  CODE1("  int P_Phase_%d;\t", i );
	  CODE2("  int Cycle_%d[%d];\n", i, The_Routing->Period );

	  break;

	default:
	  if ( Is_Compet_Pl[i] )
	    CODE1("  int Count_%d;\n", i );

	  break;
	}
      } /* if ( Routing != NULL ) */
      else if ( Is_Compet_Pl[i] )
	CODE1("  int Count_%d;\n", i );

      if ( Reseau->P[i].Outdegree > 1 )
	/* Declare the H table for all places with routing.	 	*/
	CODE3("  int H_%d[%sBUF_SIZE][%sMAX_OUTDEGREE];\n", 
	      i, Prefix, Prefix );
    }
  }
  else {
    /* This is a "normal" mixed code generation.			*/
    CODE("/* tables and counts for routing generation\t\t\t*/\n");
    for( i=0; i<Reseau->Nb_Places; i++ ) 
    if ( Reseau->P[i].Tag == EQ_OBJ ) {
      The_Routing = Reseau->P[i].Routing;

      if ( The_Routing != NULL ) {
	switch ( The_Routing->Type ) {

	case 'R':
	  CODE1("int Count_%d = 0;\n", i );
	  CODE2("double P_Proba_%d[%d] = {", i, Reseau->P[i].Outdegree );
      
	  Cum_Proba = 0.0;
	  for ( j=0; j<Reseau->P[i].Outdegree-1 ; j++ ) {
	    Cum_Proba += The_Routing->Probas[j];
	    CODE1(" %6.4f,", Cum_Proba );
	  }
	  CODE(" 1.0 };\n");
	  CODE2("int Dest_%d[%d] = {", i, Reseau->P[i].Outdegree );
	  L_Scan = Reseau->P[i].Outputs;
	  for ( j=0; j<Reseau->P[i].Outdegree-1 ; j++ ) {
	    CODE1(" %d,", L_Scan->Dest );
	    L_Scan = L_Scan->Next;
	  }
	  CODE1(" %d };\n", L_Scan->Dest );

	  break;

	case 'D':
	  CODE1("int Count_%d = 0;\n", i );
	  CODE1("int P_Phase_%d = 0;\n", i );
	  CODE2("int Cycle_%d[%d] = {", i, The_Routing->Period );
	  for ( j=0; j<The_Routing->Period-1 ; j++ ) {
	    CODE2(" NB_P%d_TO_T%d,", i, The_Routing->Cycle[j]->Number );
	  }
	  CODE2(" NB_P%d_TO_T%d };\n",
		i, The_Routing->Cycle[The_Routing->Period-1]->Number );

	  break;

	default:
	  if ( Is_Compet_Pl[i] )
	    CODE1("int Count_%d = 0;\n", i );

	  break;
	}
      } /* if ( Routing != NULL ) */
      else if ( Is_Compet_Pl[i] )
	CODE1("int Count_%d = 0;\n", i );

      if ( ( Reseau->P[i].Outdegree > 1 )
	   || ( Has_Prosit_Trans_Succ( Reseau, i ) ) )
	/* Declare the H table for places with routing and also for 	*/
	/* EQ places with Prosit output					*/
/* 	CODE1("int H_%d[BUF_SIZE][NB_TRANS];\n", i ); */
	CODE3("int H_%d[%sBUF_SIZE][%sMAX_OUTDEGREE];\n",
	      i, Prefix, Prefix );

    }
    else {
      /* the place is a Prosit object */
      if ( Has_EQ_Trans_Succ( Reseau, i ) )
/* 	CODE1("int H_%d[BUF_SIZE][NB_TRANS];\n", i ); */
 	CODE3("int H_%d[%sBUF_SIZE][%sMAX_OUTDEGREE];\n", 
	      i, Prefix, Prefix );
    }
  }

  /* Tables and counts for sources. Declared even if not EQ because 	*/
  /* are needed for computing increments.				*/
  for ( i=0; i<Reseau->Nb_Trans; i++ ) {
    if ( Is_Source[i] ) {
      if ( Reseau->T[i].Tag == EQ_OBJ ) {
	The_Law = Reseau->T[i].Tr_Law;

	if ( Prosit_Class ) {
	  CODE2("  // Variables for source %d (%s)\n", i, Reseau->T[i].Name );
	  CODE1("  int\tNb_Tok_%d;\n", i );
	  CODE2("  int\tR_%d[ %sBUF_SIZE ];\n", i, Prefix );
	}
	else {
	  CODE2("/* Variables for source %d (%s) */\n", i, Reseau->T[i].Name );
	  CODE1("int\tNb_Tok_%d = 0;\n", i );
	  CODE2("int\tR_%d[ %sBUF_SIZE ];\n", i, Prefix );
	}

	switch( The_Law->Name ) {
	case 'C':
	  /* The phase of the deterministic source is initially put to 	*/
	  /* 1+Cycle in order that the first arrival be generated at 	*/
	  /* t=Cycle and not Cycle-1, taking into account the fact that	*/
	  /* the computation starts at t=0.				*/
	  if ( Prosit_Class )
	    CODE1("  int\tT_Phase_%d;\n", i );
	  else
	    CODE2("int\tT_Phase_%d = %d;\n", i,
		  1 + IRINT( The_Law->Parameters[0] ) );
	  break;

	case 'D':
	  Nb_Params = IRINT(The_Law->Parameters[0]);
	  /* We have to eliminate the null interarrivals. In the process,	*/
	  /* we sort the probabilities by decreasing values		*/
	  Pr0 = 0.0;
	  Nb_Values = Nb_Params;
	  for ( j=0; j<Nb_Params ; j++ ) {
	    if ( The_Law->Parameters[1+j] == 0.0 ) {
	      Pr0 += The_Law->Parameters[Nb_Params+1+j];
	      Nb_Values--;
	    }
	  }
	  if ( Pr0 >= 1.0 )
	    printf("%s: error: infinite throughput at source %d\n",
		   TOOLNAME, i );
	  else {
	    The_Probas = (double*) calloc( Nb_Values, sizeof( double ) );
	    The_Values = (double*) calloc( Nb_Values, sizeof( double ) );
	    Nb_Inserted = 0;
	    for ( j=0; j<Nb_Params; j++ )
	      if ( The_Law->Parameters[1+j] != 0.0 ) {
		Proba = The_Law->Parameters[Nb_Params+1+j];
		for( k=0; (k<Nb_Inserted) && (Proba<The_Probas[k]); k++ );
		for( l=Nb_Inserted; l>k; l-- ) {
		  The_Probas[l] = The_Probas[l-1];
		  The_Values[l] = The_Values[l-1];
		}
		The_Probas[k] = Proba;
		The_Values[k] = The_Law->Parameters[1+j] * Scale;
		Nb_Inserted++;
	      }
	    
	    CODE1("int\tT_Phase_%d = 1;\n", i );
	    CODE2("double\tT_Pr_Zero_%d = %6.4f;\n", i, Pr0 );
	    CODE2("double\tT_Proba_%d[%d] = {", i, Nb_Values );
	    Cum_Proba = 0.0;
	    for ( j=0; j<Nb_Values-1 ; j++ ) {
	      Cum_Proba += The_Probas[j];
	      CODE1(" %6.4f,", Cum_Proba/(1.0-Pr0) );
	    }
	    CODE(" 1.0 };\n");
	  }

	  CODE2("double\tInterarrival_%d[%d] = {", i, Nb_Values );
	  for ( j=0; j<Nb_Values-1 ; j++ ) {
	    CODE1(" %6.4f,", The_Values[j] );
	  }
	  CODE1(" %6.4f };\n", The_Values[Nb_Values-1] );

	  free( The_Probas );
	  free( The_Values );

	  break;

	case 'B':
	case 'G':
	  if ( Scale != 1 )
	    CODE2("int\tT_Phase_%d = %d;\n", i, Scale );
	
	  break;

	case 'g':
	  /* nothing to do */
	  break;

	default:
	  fprintf( stderr, "%s: error: can't handle law id (%c)\n",
		   TOOLNAME, The_Law->Name );
	}
      }
    else {
      /* the source is a Prosit object */
      if ( Has_EQ_Place_Succ( Reseau, i ) ) {
	CODE2("int\tR_%d[ %sBUF_SIZE ];\n", i, Prefix );
      }
    }
    }
  }

}

void	Gen_EQ_Defines( PetriNet	*Reseau,
			int		Max_Delay,
			int		Scale,
			boolean		Prosit_Class,
			char		*Class_Name )
{
  int		i;
  int		j;
  int		Max_Degree;
  Liste_Link	L_Scan;
  Law_Desc	The_Law;

  /* General constants of the object.	*/
  if ( Prosit_Class ) {
    CODE2("const int %sNB_TRANS = %d;\n", Prefix, Reseau->Nb_Trans );
    CODE2("const int %sNB_PLACES = %d;\n", Prefix, Reseau->Nb_Places );
    CODE2("const int %sNB_MAX_DEPT = %d;\n", Prefix, 256 );
    CODE2("const int %sBUF_SIZE = %d;\n", Prefix, Buf_Size );
    CODE2("const int %sHALF_BUF_SIZE = %d;\n", Prefix, Buf_Size/2 );
    CODE2("const int %sMAX_DELAY = %d;\n", Prefix, Max_Delay );
    CODE ("#define Infinity\t0x7FFFFFFF\n" );
  }
  else {
    CODE2("#define %sNB_TRANS\t%d\n", Prefix, Reseau->Nb_Trans );
    CODE2("#define %sNB_PLACES\t%d\n", Prefix, Reseau->Nb_Places );
    CODE2("#define %sNB_MAX_DEPT\t%d\n", Prefix, 256 );
    CODE2("#define %sBUF_SIZE\t%d\n", Prefix, Buf_Size );
    CODE2("#define %sHALF_BUF_SIZE\t%d\n", Prefix, Buf_Size/2 );
    CODE2("#define %sMAX_DELAY\t%d\n", Prefix, Max_Delay );
    CODE ("#define Infinity\t0x7FFFFFFF\n" );
  }
  CODELN;
  CODE("#define MIN(x,y)    ( x<y?x:y )\n");
  CODELN;

  /* Compute the max outdegree of places "involved" in EQ. 		*/
  /* For use with the H_* tables.					*/
  Max_Degree = -1;
  for ( i=0; i<Reseau->Nb_Places; i++ ) 
    if ( ( ( Reseau->P[i].Tag == PROSIT_OBJ )
	   && Has_EQ_Trans_Succ( Reseau, i ) )
	 || ( ( Reseau->P[i].Tag == EQ_OBJ )
	      && ( Reseau->P[i].Outdegree > 1 ) ) )
      Max_Degree = max( Max_Degree, Reseau->P[i].Outdegree );
  if ( Max_Degree > 0 )
    CODE2("#define %sMAX_OUTDEGREE\t%d\n", Prefix, Max_Degree );

  /* now generating class ids for the interface Prosit->EQ 		*/
  /* These IDs begin at NB_TRANS.					*/
  i = max( Reseau->Nb_Trans, Reseau->Nb_Places );

  for ( j=0; j<Reseau->Nb_Trans; j++ ) {
    if ( Reseau->T[j].Tag == PROSIT_OBJ ) {
      SCAN( L_Scan, Reseau->T[j].Outputs )
	if ( Reseau->P[L_Scan->Dest].Tag == EQ_OBJ ) {
	  printf("#define CL_T%d_P%d\t%d\n", 
		 j, L_Scan->Dest, i+1 );
	  i++;
	}
    }
  }
  for ( j=0; j<Reseau->Nb_Places; j++ ) {
    if ( Reseau->P[j].Tag == PROSIT_OBJ ) {
      SCAN( L_Scan, Reseau->P[j].Outputs )
	if ( Reseau->T[L_Scan->Dest].Tag == EQ_OBJ ) {
	  printf("#define CL_P%d_T%d\t%d\n", 
		 j, L_Scan->Dest, i+1 );
	  i++;
	}
    }
  }
  /* defining link numbers for use with H_* tables 			*/
  for ( i=0; i<Reseau->Nb_Places; i++ ) {
    if ( ( ( Reseau->P[i].Tag == PROSIT_OBJ )
	   && Has_EQ_Trans_Succ( Reseau, i ) )
	 || ( ( Reseau->P[i].Tag == EQ_OBJ )
	      && ( Has_Prosit_Trans_Succ( Reseau, i ) ) )
	 || ( ( Reseau->P[i].Tag == EQ_OBJ )
	      && ( Reseau->P[i].Outdegree > 1 ) ) ) {
      j = 0;
      SCAN( L_Scan, Reseau->P[i].Outputs ) {
	CODE3("#define %sNB_P%d_TO_T%d", Prefix, i, L_Scan->Dest );
	CODE1("\t%d\n" , j );
	j++;
      }
    }
  }

  CODELN;
  CODE("/* external declarations from the Prosit part */\n");
  CODE("extern bool\tTrace;\n");

  if ( ! No_Prosit ) {
    for ( j=0; j<Reseau->Nb_Trans; j++ ) {
      if ( Reseau->T[j].Tag == EQ_OBJ ) {
	SCAN( L_Scan, Reseau->T[j].Outputs )
	  if ( Reseau->P[L_Scan->Dest].Tag == PROSIT_OBJ ) {
	    CODE1("extern Routing_Server\t*Place_%d;\n", L_Scan->Dest );
	  }
      }
    }
    for ( j=0; j<Reseau->Nb_Places; j++ ) {
      if ( Reseau->P[j].Tag == EQ_OBJ ) {
	SCAN( L_Scan, Reseau->P[j].Outputs )
	  if ( Reseau->T[L_Scan->Dest].Tag == PROSIT_OBJ ) {
	    CODE1("extern Routing_Server\t*Trans_%d;\n", L_Scan->Dest );
	  }
      }
    }
    CODELN;
  }

}
void	Gen_EQ_Declarations( 
PetriNet	*Reseau,
int		Max_Delay,
int		Scale,
boolean		Prosit_Class,
char		*Class_Name,
boolean		Single_Marking_Function )
{
  int		i;

  Gen_EQ_Defines( Reseau, Max_Delay, Scale, Prosit_Class, Class_Name );

  Gen_EQ_Variables( Reseau, Max_Delay, Scale, Prosit_Class, Class_Name );

  Gen_Marking_Code( Reseau, Prosit_Class, Class_Name, 
		    Single_Marking_Function );

  Gen_Source_Code( Reseau, Prosit_Class, Class_Name );

}

void gen_trace_before_firing( Reseau, Trans_Id, Initial )
PetriNet	*Reseau;
int		Trans_Id;
boolean		Initial;
{
  Tr_Des	The_Trans;
  Liste_Link	L_Scan;
  Liste_Link	M_Scan;
  int		Delay;

  The_Trans = Reseau->T[Trans_Id];

  if ( The_Trans.Indegree > 1 )
    CODE("\tCODE(\"BEGIN_MULTIPLE\\n\");\n");
  /* first remove tokens from input places, except dummy ones */
  SCAN( L_Scan, The_Trans.Inputs ) {
    if ( !Is_Dummy( Reseau->P[L_Scan->Dest] ) ) {
      CODE1("  Marking = %d;\n", Reseau->P[L_Scan->Dest].Init_Pop );
      SCAN( M_Scan, Reseau->P[L_Scan->Dest].Inputs ) {
	Delay = Fire_Time[M_Scan->Dest];
	if ( Initial )
	  CODE3("  Marking += ( EQ_Time>%d ? Y[ EQ_Time - %d ][%d] : 0 )", 
	      Delay, Delay, M_Scan->Dest );
	else if ( Y_Static )
	  CODE2("  Marking += Y_Time_M_%d[%d]", Delay, M_Scan->Dest );
	else
 	  CODE2("  Marking += Y[ EQ_Time_M_%d ][%d]", Delay, M_Scan->Dest );

	if ( M_Scan->Weight > 1 )
	  CODE1(" * %d;\n", M_Scan->Weight );
	else
	  CODE(";\n");
      }
      SCAN( M_Scan, Reseau->P[L_Scan->Dest].Outputs ) {
	if ( Y_Static )
	  CODE1("  Marking -= Y_0[%d]", M_Scan->Dest );
	else 
	  CODE1("  Marking -= Y[ EQ_Time ][%d]", M_Scan->Dest );
	if ( M_Scan->Weight > 1 )
	  CODE1(" * %d;\n", M_Scan->Weight );
	else
	  CODE(";\n");
      }
    }
    CODE("  CODE2(\"MODIFY %%s \\\"token\\\" = %%d\\n\",\n");
    CODE1("        gss_place_name[%d], Marking );\n", L_Scan->Dest );
  } 
  if ( The_Trans.Indegree > 1 )
    CODE("  CODE(\"END_MULTIPLE\\n\");\n");

  /* then move the tokens */
  if ( The_Trans.Indegree > 1 )
    CODE("  CODE(\"BEGIN_MULTIPLE\\n\");\n");
  SCAN( L_Scan, The_Trans.Inputs ) {
    if ( !Is_Dummy( Reseau->P[L_Scan->Dest] ) ) {
      CODE("  CODE2(\"MOVE_NEW_OBJECT OF_CLASS ");
      switch( L_Scan->Weight ) {
      case 0:
      case 1:
	CODE("token");
	break;
      case 2:
	CODE("bitoken");
	break;
      case 3:
	CODE("tritoken");
	break;
      default:
	CODE("quadritoken");
      }
      CODE(" FROM %%s TO %%s \",\n");
      CODE1("        gss_place_name[%d], ", L_Scan->Dest );
      CODE1("gss_trans_name[%d] );\n", Trans_Id );
      CODE1("  CODE(\"ON_PATH %s \");\n", L_Scan->Data );
      CODE("  CODE(\"WITH_STEPS 20\\n\");\n");
    }
  }
  if ( The_Trans.Indegree > 1 )
    CODE("  CODE(\"END_MULTIPLE\\n\");\n");

  /* then switch transition status */
  CODE1("  CODE(\"MODIFY %s \\\"status\\\" = 0\\n\");\n",
	The_Trans.Name );
}

void gen_trace_after_firing( Reseau, Trans_Id, Initial )
PetriNet	*Reseau;
int		Trans_Id;
boolean		Initial;
{
  Tr_Des	The_Trans;
  Liste_Link	L_Scan;
  Liste_Link	M_Scan;
  int		Delay;

  The_Trans = Reseau->T[Trans_Id];

  /* move tokens to (non dummy) output places */
  if ( The_Trans.Outdegree > 1 )
    CODE("  CODE(\"BEGIN_MULTIPLE\\n\");\n");
  SCAN( L_Scan, The_Trans.Outputs ) {
    if ( !Is_Dummy( Reseau->P[L_Scan->Dest] ) ) {

      CODE("  CODE2(\"MOVE_NEW_OBJECT OF_CLASS ");
      switch( L_Scan->Weight ) {
      case 0:
      case 1:
	CODE("token");
	break;
      case 2:
	CODE("bitoken");
	break;
      case 3:
	CODE("tritoken");
	break;
      default:
	CODE("quadritoken");
      }
      CODE(" FROM %%s TO %%s \",\n");
      CODE1("        gss_trans_name[%d],", Trans_Id );
      CODE1("gss_place_name[%d] );\n", L_Scan->Dest );
      CODE1("  CODE(\"ON_PATH %s \");\n", L_Scan->Data );
      CODE("  CODE(\"WITH_STEPS 20\\n\");\n");
    }
  }
  if ( The_Trans.Outdegree > 1 )
    CODE("  CODE(\"END_MULTIPLE\\n\");\n");

  /* finally change marking of (non dummy) output places */
  if ( The_Trans.Outdegree > 1 )
    CODE("  CODE(\"BEGIN_MULTIPLE\\n\");\n");
  SCAN( L_Scan, The_Trans.Outputs ) {
    if ( !Is_Dummy( Reseau->P[L_Scan->Dest] ) ) {
      CODE1("  Marking = %d;\n", Reseau->P[L_Scan->Dest].Init_Pop );
      SCAN( M_Scan, Reseau->P[L_Scan->Dest].Outputs ) {
	Delay = Fire_Time[M_Scan->Dest];
	if ( Initial )
	  CODE3("  Marking += ( EQ_Time>%d ? Y[ EQ_Time - %d ][%d] : 0 )", 
	      Delay, Delay, M_Scan->Dest );
	else
	  CODE2("  Marking += Y[ EQ_Time - %d ][%d]", Delay, M_Scan->Dest );

	if ( M_Scan->Weight > 1 )
	  CODE1(" * %d;\n", M_Scan->Weight );
	else
	  CODE(";\n");
      }
      SCAN( M_Scan, Reseau->P[L_Scan->Dest].Outputs ) {
	CODE1("  Marking -= Y[ EQ_Time ][%d]", M_Scan->Dest );
	if ( M_Scan->Weight > 1 )
	  CODE1(" * %d;\n", M_Scan->Weight );
	else
	  CODE(";\n");
      }
    }
    CODE("  CODE2(\"MODIFY %%s \\\"token\\\" = %%d\\n\",\n");
    CODE1("        gss_place_name[%d], Marking );\n", L_Scan->Dest );
  } 
  if ( The_Trans.Outdegree > 1 )
    CODE("  CODE(\"END_MULTIPLE\\n\");\n");

  CODE("  CODE(\"eol+\\n\");\n");
}

void	write_trans_name( The_Trans )
Tr_Des	The_Trans;
{
  if ( NULL != The_Trans.Name )
    printf( "%s", The_Trans.Name );
  else
    printf( "_TR_%d", The_Trans.Tr_Id );
}

void	Write_Token_Total(
PetriNet	*Reseau,
int		Place_Id,
boolean		Initial,
boolean		Prosit_Class )
{
  Liste_Link	L_Scan;
  int		Delay;
  int		Input_Num;
  boolean	Could_Be_Connected;

  if ( Already_Visited[ Place_Id ] ) {
    CODE1("  Nb_Tokens = Count_%d;\n", Place_Id );
  }
  else {
    if ( Prosit_Class )
      CODE("  // ");
    else
      CODE("  /* ");
    CODE2("Computing number of tokens arrived in place %d (%s)",
	  Place_Id, Reseau->P[Place_Id].Name );
    if ( Prosit_Class )
      CODELN;
    else 
      CODE(" */\n");

    if ( Initial )
      CODE1("  Nb_Tokens = Avail[%d][EQ_Time];\n", Place_Id );
    else
      CODE1("  Nb_Tokens = %d;\n", Reseau->P[Place_Id].Init_Pop );

    /* Have to add tokens from external transitions, if any */
    Input_Num = Is_Input_Number( Reseau, 'P', Place_Id );
    Could_Be_Connected = ( Input_Num != -1 ) && Prosit_Class;
    if ( Could_Be_Connected ) {
      CODE1("  if ( Is_Input_Place[%d] )\n", Place_Id );
      /* there is no need for a loop because a place can have only one	*/
      /* registered external "input channel".				*/
      CODE1("    Nb_Tokens += Ext_Y[%d][ EQ_Time ];\n", Place_Id );
    }

    SCAN( L_Scan, Reseau->P[Place_Id].Inputs ) {
      CODE( "  Nb_Tokens += ");

      Input_Num = Is_Input_Number( Reseau, 'T', L_Scan->Dest );
      Could_Be_Connected = ( Input_Num != -1 ) && Prosit_Class;

      if ( Is_Source[L_Scan->Dest] )
	Delay = IRINT( Scale * 
		       Reseau->P[Place_Id].Pl_Law->Parameters[0] );
      else
	Delay = IRINT( Scale * 
		       ( Reseau->P[Place_Id].Pl_Law->Parameters[0]
			 + Reseau->T[L_Scan->Dest].Tr_Law->Parameters[0] ) );

      if ( Initial ) {
	CODE1( "( EQ_Time >= %d ? ", Delay );
	if ( Is_Source[L_Scan->Dest] ) {
	  if ( Could_Be_Connected )
	    CODE1("( !Is_Input_Trans[%d] ? ", L_Scan->Dest );
	  CODE2( "R_%d[EQ_Time-%d] ", L_Scan->Dest, Delay );
	  if ( Could_Be_Connected )
	    CODE2( ": Y[EQ_Time-%d][%d] ) ", Delay, L_Scan->Dest );
	}
	else
	  CODE2( "Y[EQ_Time-%d][%d] ", Delay, L_Scan->Dest );
	if ( L_Scan->Weight > 1 )
	  CODE1("* %d", L_Scan->Weight );
	CODE(" : 0 );\n");
      }
      else {
	if ( Delay == 0 ) {
	  if ( Is_Source[L_Scan->Dest] ) {
	    if ( Could_Be_Connected )
	      CODE1("( !Is_Input_Trans[%d] ? ", L_Scan->Dest );
	    CODE1( "R_%d[ EQ_Time ]", L_Scan->Dest );
	    if ( Could_Be_Connected )
	      CODE1( ": Y[ EQ_Time ][%d] ) ", L_Scan->Dest );
	  }
	  else if ( Y_Static )
	    CODE1( "Y_0[%d]", L_Scan->Dest );
	  else
 	    CODE1( "Y[ EQ_Time ][%d]", L_Scan->Dest );
	}
	else {
	  if ( Is_Source[L_Scan->Dest] ) {
	    if ( Could_Be_Connected )
	      CODE1("( !Is_Input_Trans[%d] ? ", L_Scan->Dest );
	    CODE2( "R_%d[ EQ_Time_M_%d ]", L_Scan->Dest, Delay );
	    if ( Could_Be_Connected )
	      CODE2( ": Y[EQ_Time_M_%d][%d] ) ", Delay, L_Scan->Dest );
	  }
	  else if ( Y_Static )
	    CODE2( "Y_Time_M_%d[%d]", Delay, L_Scan->Dest );
	  else
 	    CODE2( "Y[ EQ_Time_M_%d ][%d]", Delay, L_Scan->Dest );
	}
	if ( L_Scan->Weight > 1 )
	  CODE1(" * %d", L_Scan->Weight );
	CODE(";\n");
      }
    }
  }
}

/************************************************************************/
/* Procedure for writing the adequate code for computing the "marking"	*/
/* of a place that can be used for computing the enabling degree.	*/
/************************************************************************/
void Write_Partial_Enabling( 
PetriNet	*Reseau,
int 		Place_Id, 
int 		Trans_Id,
int		Weight,
boolean		Initial,
boolean		Single_Marking_Function )
{
  Routing_Desc	The_Routing;
  
  The_Routing = Reseau->P[Place_Id].Routing;

  if ( The_Routing == NULL ) {
    /* Code incomplet, intention inconnue... */
    /*     CODE("Y_0[??"); */
    if ( Single_Marking_Function ) {
      CODE3("%sMarking( %d%s )", 
	    ( Initial ? "I_" : "" ), 
	    Place_Id,
	    ( Initial ? ", EQ_Time" : "" ) );
    }
    else
      CODE3("%sMarking_%d( %s )", 
	    ( Initial ? "I_" : "" ), Place_Id,
	    ( Initial ? "EQ_Time" : "" ) );
    if ( Weight > 1 )
      CODE1(" / %d ", Weight );
  }
  else if ( The_Routing->Type == 'R' ) {
    CODE1("H_%d[EQ_Time]", Place_Id );
    CODE3("[ %sNB_P%d_TO_T%d ]" , Prefix, Place_Id, Trans_Id );
    if ( Initial || !Y_Static )
      CODE1(" - Y[EQ_Time][%d]", Trans_Id );
    else
      CODE1(" - Y_0[%d]", Trans_Id );
  }
  else {
    if ( Single_Marking_Function ) {
      CODE3("%sMarking( %d%s )", 
	    ( Initial ? "I_" : "" ), 
	    Place_Id,
	    ( Initial ? ", EQ_Time" : "" ) );
    }
    else
      CODE3("%sMarking_%d( %s )", 
	    ( Initial ? "I_" : "" ), Place_Id,
	    ( Initial ? "EQ_Time" : "" ) );
    if ( Weight > 1 )
      CODE1(" / %d ", Weight );
  }
}

/************************************************************************/
/* Procedure for writing down the code to increment the H functions.	*/
/* Is called from the loop code						*/
/* - the first time the routing place Place_Id is encountered in the	*/
/*   computation of a transition,					*/
/* - for places (even without routing) with Prosit outputs.		*/
/* The procedure is called with a "Trans_Id" parameter which is used	*/
/* in the case of the 'P' routing. It is safe to use any dummy value if	*/
/*  the routing is not 'P'.						*/
/************************************************************************/
void Gen_EQ_H_Function( 
PetriNet	*Reseau,
int		Place_Id,
int		Trans_Id,
boolean		Initial,
boolean		Single_Marking_Function )
{
  Liste_Link	L_Scan;
  Routing_Desc	The_Routing;

  CODE2("  /* Computing the number of tokens routed from place %d (%s) */\n",
	Place_Id, Reseau->P[Place_Id].Name );

  The_Routing = Reseau->P[Place_Id].Routing;

  if ( The_Routing != NULL ) {
    switch( The_Routing->Type ) {

    case 'R':

      CODE2("  for( ; Count_%d < Nb_Tokens; Count_%d++ ) {\n", 
	    Place_Id, Place_Id );
      CODE("    U = (*U_0_1)();\n");
      CODE1("    for( j=0; U > P_Proba_%d[j]; j++ );\n", Place_Id );
/*       CODE2("    H_%d[EQ_Time][Dest_%d[j]]++;\n", Place_Id, Place_Id ); */
      CODE1("    H_%d[EQ_Time][ j ]++;\n", Place_Id );
      CODE("  }\n");

      Already_Visited[ Place_Id ] = TRUE;

      break;

    case 'D':

      CODE2("  for( ; Count_%d < Nb_Tokens; Count_%d++ ) {\n", 
	    Place_Id, Place_Id );
      CODE3("    H_%d[EQ_Time][ Cycle_%d[ P_Phase_%d ] ]++;\n",
	    Place_Id, Place_Id, Place_Id );
      CODE1("    P_Phase_%d++;\n", Place_Id );
      CODE3("    P_Phase_%d = ( P_Phase_%d %% %d );\n", 
	    Place_Id, Place_Id, The_Routing->Period );
      CODE("  }\n");

      Already_Visited[ Place_Id ] = TRUE;

      break;

    case 'P':

      CODE("  /* Compute enabling degree of transition ");
      CODE2("%d */\n", Trans_Id, Reseau->T[Trans_Id].Name );
      CODE("  E_Degree = ( ");

      L_Scan = Reseau->T[Trans_Id].Inputs;
      Write_Partial_Enabling( Reseau, 
			      L_Scan->Dest, Trans_Id, 
			      L_Scan->Weight, Initial,
			      Single_Marking_Function );
      CODE(" );\n");

      SCAN( L_Scan, Reseau->T[Trans_Id].Inputs->Next ) {
	CODE("  E_Degree = MIN( E_Degree, ");
	Write_Partial_Enabling( Reseau,
				L_Scan->Dest, Trans_Id, 
				L_Scan->Weight, Initial,
				Single_Marking_Function );
	CODE(" );\n");
      }
      CODE("  if ( E_Degree > 0 ) {\n");
      CODE1("    H_%d[EQ_Time]", Place_Id );
      CODE3("[ %sNB_P%d_TO_T%d ] += E_Degree;\n", 
	    Prefix, Place_Id, Trans_Id );
      CODE1("    Count_%d += E_Degree;\n", Place_Id );
      CODE("  }\n");

      /* for 'P' routing, the place does not become "visited" because	*/
      /* the computation of H must continue for other output transitions*/

      break;

    case 'C':
      /* competition allowed if one output only. Normally, this 	*/
      /* feature is used only for Prosit places with EQ output.		*/
      if ( Reseau->P[Place_Id].Outdegree == 1 ) {
/* 	CODE2("  H_%d[EQ_Time][%d] = Nb_Tokens;\n",  */
/* 	      Place_Id, */
/* 	      Reseau->P[Place_Id].Outputs->Dest ); */
	CODE1("  H_%d[EQ_Time]", Place_Id );
	CODE3("[ %sNB_P%d_TO_T%d ] = Nb_Tokens;\n", 
	      Prefix, Place_Id, Reseau->P[Place_Id].Outputs->Dest );
      }
      else {
	fprintf( stderr, "%s: error: illegal competition at place",
		 TOOLNAME );
	fprintf( stderr, "#%d (%s)\n", 
		 Place_Id, Reseau->P[Place_Id].Name );
      }
      break;
	    
    default:

      printf("%s: error: unexpected routing policy (%c)\n",
	     TOOLNAME, The_Routing->Type );
    } /* switch(...) */
  } /* if ( Routing != NULL ) */
  else {
    CODE1("  Count_%d = Nb_Tokens;\n", Place_Id );
    Already_Visited[ Place_Id ] = TRUE;
  }
}

/* Procedure for writing down how many tokens were routed from a place	*/
/* to a trans. Result is put in Nb_Tokens.				*/
/* If there is no routing at the said transition, nothing is written	*/
/* because the result is assumed to be already in Nb_Tokens.		*/
void	Write_Token_Routing(
PetriNet	*Reseau,
int		Place_Id,
int		Trans_Id,
boolean		Initial,
boolean		Single_Marking_Function )
{

  if ( Reseau->P[Place_Id].Outdegree > 1 ) {
    if ( !Already_Visited[Place_Id] ) {
      Gen_EQ_H_Function( Reseau, Place_Id, Trans_Id, Initial,
			 Single_Marking_Function );
    }

/*     CODE2("  Nb_Tokens = H_%d[EQ_Time][%d];\n", Place_Id, Trans_Id ); */
    CODE1("  Nb_Tokens = H_%d[EQ_Time]", Place_Id );
    CODE3("[ %sNB_P%d_TO_T%d ];\n", Prefix, Place_Id, Trans_Id );
  }

}

void Write_Token_Reduction( The_Link )
Liste_Link	The_Link;
{
  if ( The_Link->Weight > 1 )
    CODE1("Nb_Tokens /= %d;\n", The_Link->Weight );
}

void	gen_code_node(
PetriNet	*Reseau,
int		The_Node,
boolean		Initial,
boolean		Prosit_Class,
boolean		Single_Marking_Function )
{
  Liste_Link	L_Scan;
  Tr_Des	The_Trans;
  int		Pred_Pl_Id;
  int		Delay;
  int		Input_Num;
  boolean	Could_Be_Connected;

  The_Trans = Reseau->T[ The_Node ];

  if ( Prosit_Class )
    CODE1("  // Evolution of transition #%d: ", The_Node );
  else
    CODE1("  /* Evolution of transition #%d: ", The_Node );
  write_trans_name( Reseau->T[The_Node] );
  if ( Prosit_Class )
    CODELN;
  else
    CODE(" */\n"); 

  Input_Num = Is_Input_Number( Reseau, 'T', The_Node );
  Could_Be_Connected = ( Input_Num != -1 ) && Prosit_Class;

  if ( !Could_Be_Connected && ( The_Trans.Indegree == 0 ) ) {
    /* Actually, we shouldn't be here at all, because sources are	*/
    /* handled somewhere else.				        */
    fprintf( stderr,
	     "%s: error: inconsistent code generation for trans. T %d\n",
	     TOOLNAME, The_Node+1 );
  }
  else if  ( !Could_Be_Connected && ( The_Trans.Indegree == 1 ) ) {
    /* The transition has a single input. No min is necessary. */
    Pred_Pl_Id = The_Trans.Inputs->Dest;
    if ( Reseau->P[Pred_Pl_Id].Tag == EQ_OBJ ) {
      Write_Token_Total( Reseau, Pred_Pl_Id, Initial, Prosit_Class );
      Write_Token_Routing( Reseau, Pred_Pl_Id, The_Node, Initial,
			   Single_Marking_Function );
      Write_Token_Reduction( The_Trans.Inputs );
    }
    else {
      CODE3("  Nb_Tokens = H_%d[EQ_Time][ NB_P%d_TO_T%d ];\n", 
	    Pred_Pl_Id, Pred_Pl_Id, The_Node );
    }

    if ( Initial || !Y_Static )
      CODE1("  Y[EQ_Time][%d] = Nb_Tokens;\n", The_Node );
    else
      CODE1("  Y_0[%d] = Nb_Tokens;\n", The_Node );
  }
  else { 
    /* Here, either the "internal" indegree is >= 2, or external input	*/
    /* is possible (case of a single Prosit class).			*/

    CODE ("  Tmp_Y = Infinity;\n" );

    /* First compute the number of tokens coming from outside.		*/
    if ( Could_Be_Connected ) {
      CODE1("  if ( Is_Input_Trans[%d] ) {\n", The_Node );
      CODE1("    for ( In_Nb=0; In_Nb < Orig_Nb[%d]; In_Nb++ ) {\n", 
	    Input_Num );
      CODE1("      Nb_Tokens = Ext_H[Orig_Table[%d][In_Nb]][ EQ_Time ];\n",
	    Input_Num );
      CODE ("      Tmp_Y = MIN( Tmp_Y, Nb_Tokens );\n" );
      CODE ("    }\n");
      CODE ("  }\n");
    }

    switch( The_Trans.Indegree ) {

    case 0:
      /* nothing more to do... */
      break;

    default:

      for( L_Scan=The_Trans.Inputs; L_Scan != NULL; L_Scan=L_Scan->Next ) {

	Pred_Pl_Id = L_Scan->Dest;

	if ( Reseau->P[Pred_Pl_Id].Tag == EQ_OBJ ) {
	  Write_Token_Total( Reseau, Pred_Pl_Id, Initial, Prosit_Class );
	  Write_Token_Routing( Reseau, Pred_Pl_Id, The_Node, Initial,
			       Single_Marking_Function );
	  Write_Token_Reduction( L_Scan );
	}
	else {
	  CODE3("  Nb_Tokens = H_%d[EQ_Time][ NB_P%d_TO_T%d ];\n", 
		Pred_Pl_Id, Pred_Pl_Id, The_Node );
	}
	
	CODE("  Tmp_Y = MIN( Tmp_Y, Nb_Tokens );\n" );
      }
    } /* switch () */

    /* The number of firings is now in Tmp_Y. Put it where it belongs */
    if ( Initial || !Y_Static )
      CODE1("  Y[EQ_Time][%d] = Tmp_Y;\n", The_Node );
    else
      CODE1("  Y_0[%d] = Tmp_Y;\n", The_Node );

  }

  CODELN;
}

void	gen_z_code( Reseau, Initial )
PetriNet	*Reseau;
boolean		Initial;
{
  int		i;

  CODE("  /* computing the number of completed firings */\n");
  for ( i=0; i<Reseau->Nb_Trans; i++ )
    if ( Reseau->T[i].Tag == EQ_OBJ ) {
      if ( Initial ) {
	if ( Fire_Time[i] > 0 ) {
	  CODE2("  Z[EQ_Time][%d] = ( EQ_Time >= %d ?", i, Fire_Time[i] );
	  CODE2(" Y[EQ_Time-%d][%d] : 0 );\n", Fire_Time[i], i );
	}
	else 
	  CODE2("  Z[EQ_Time][%d] = Y[EQ_Time][%d];\n", i, i );

      }
      else {
	if ( Fire_Time[i] > 0 ) {
	  CODE3("  Z[EQ_Time][%d] = Y[EQ_Time_M_%d][%d];\n",
		i, Fire_Time[i], i );
	}
	else 
	  CODE2("  Z[EQ_Time][%d] = Y[EQ_Time][%d];\n", i, i );
      }
    }
}

void	gen_preparation_code(
PetriNet	*Reseau,
boolean		Initial,
boolean		Prosit_Class,
char		*Class_Name )
{
  int		i;
  int		Nb_Sources;
  Liste_Link	L_Scan;

  /* check first if there are some non tagged sources */
  Nb_Sources = 0;
  for ( i=0; i<Reseau->Nb_Trans; i++ )
    if ( ( Reseau->T[i].Tag == EQ_OBJ ) && Is_Source[i] )
      Nb_Sources++;

  if ( Nb_Sources > 0 ) {
    CODE ("  /* Draw source arrivals when necessary */\n");
    CODE1("  if ( EQ_Time %% %sHALF_BUF_SIZE == 0 ) {\n", Prefix );
    CODE ("    if ( EQ_Time == 0 )\n");
    CODE1("      Re_Init_Sources( 0, %sHALF_BUF_SIZE );\n", Prefix );
    CODE ("    else\n");
    CODE2("      Re_Init_Sources( %sHALF_BUF_SIZE, %sBUF_SIZE );\n", 
	  Prefix, Prefix );
    CODE ("  }\n");
    CODELN;
  }

  if ( Prosit_Class )
    CODE("  // Copy present firing count for everybody\n");
  else
    CODE("  /* Copy present firing count for everybody  */\n");

  /* Is optimized so at to copy only what is necessary, i.e.: EQ objs.	*/
  /* Except that sometimes (in particular, when all the network is 	*/
  /* in a single module), all objects are necessary. In that case, a	*/
  /* loop is better...							*/
  /* Note that Prosit objs watched by the EQ module are dealt with 	*/
  /* at the end of Increment().						*/
  if ( Prosit_Class ) {
    CODE1("  for ( int i=0; i<%sNB_TRANS; i++ )\n", Prefix );
    if ( Initial )
      CODE("    Y[EQ_Time][ i ] = ( EQ_Time>0 ? Y[EQ_Time-1][ i ] : 0 );\n");
    else if ( Y_Static )
      CODE("    Y_0[ i ] = Y_Time_M_1[ i ];\n");
    else
      CODE("    Y[ EQ_Time ][ i ] = Y[ EQ_Time_M_1 ][ i ];\n");
  }
  else {
    for( i=0; i<Reseau->Nb_Trans; i++ )
      if ( !Is_Source[i] ) {
	if ( Reseau->T[i].Tag == EQ_OBJ ) {
	  if ( Initial )
	    CODE2("  Y[EQ_Time][%d] = ( EQ_Time>0 ? Y[EQ_Time-1][%d] : 0 );\n",
		  i, i );
	  else if ( Y_Static )
	    CODE2("  Y_0[%d] = Y_Time_M_1[%d];\n", i, i );
	  else
	    CODE2("  Y[ EQ_Time ][%d] = Y[ EQ_Time_M_1 ][%d];\n", i, i );
	}
	/* else.. nothing. This must be a "pure"  Prosit object.	*/
      }
  }

  CODELN;

/*   if ( Has_Competition ) { */
/*     CODE("/* Prepare enabling count variables for transitions with "); */
/*     CODE("conflicts * /\n"); */

/*     for( i=0; i<Reseau->Nb_Trans; i++ ) */
/*       if ( Is_Compet_Tr[i] ) { */
/* 	if ( Initial ) */
/* 	  CODE2("Y[Time][%d] = ( Time>0 ? Y[Time-1][%d] : 0 );\n", i, i ); */
/* 	else if ( Y_Static ) */
/* 	  CODE2("Y_0[%d] = Y_Time_M_1[%d];\n", i, i ); */
/* 	else */
/*  	CODE2("Y[ Time ][%d] = Y[ Time_M_1 ][%d];\n", i, i ); */
/*       } */

/*     CODELN; */
/*   } */
}


void 	gen_loop_code(
PetriNet	*Reseau,
Layer_Struct	The_Layer_Struct,
boolean		Initial,
boolean		Prosit_Class,
boolean		Single_Marking_Function )
{
  int	Num_Layer;
  int	Num_Node;
  int	The_Node;
  int	Time_Scale;
  int	i;
  int	j;
  int	Nb_Nodes;

  for( i=1; i<=The_Scales[0]; i++ ) {
    Time_Scale = The_Scales[i];

    /* find the number of nodes within this time scale */
    Nb_Nodes = 0;
    for ( j=0; j<Reseau->Nb_Trans; j++ )
      if ( !Is_Source[j] && ( TST[j] == Time_Scale ) )
	Nb_Nodes++;

    if ( Nb_Nodes != 0 ) {
      CODE3("  /* Computing the %d transition%s with time scale %d */\n", 
	    Nb_Nodes,
	    ( Nb_Nodes > 1 ? "s" : "" ),
	    Time_Scale );

      if ( Time_Scale != 1 ) {
	CODE1("  if ( ( Real_EQ_Time %% %d ) == 0 ) {\n\n", Time_Scale );
	Indent_Nb += 2;
      }

      for( Num_Layer = 0; Num_Layer < The_Layer_Struct->Nb_Levels; 
	   Num_Layer++)
	for ( Num_Node = 0; 
	      Num_Node < The_Layer_Struct->Level[Num_Layer]->Nb_Nodes;
	      Num_Node++ ) {
	    
	  The_Node = The_Layer_Struct->Level[Num_Layer]->Node_Id[Num_Node];
	  if ( ( Reseau->T[The_Node].Tag == EQ_OBJ ) 
	       && ( !Is_Source[The_Node] || Prosit_Class )
	       && ( TST[The_Node] == Time_Scale ) )
	    gen_code_node( Reseau, The_Node, Initial, Prosit_Class,
			   Single_Marking_Function );
	}

      if ( Time_Scale != 1 ) {
	CODE1("  } /* end of time scale %d */\n\n", Time_Scale );
	Indent_Nb -= 2;
      }
    }
  }

  /* Now generate code for places. This is necessary for EQ places that	*/
  /* have Prosit output(s). At the same time, check if all routing	*/
  /* places have been visited.						*/
  for ( i=0; i<Reseau->Nb_Places; i++ ) {
    if ( ( Reseau->P[i].Tag == EQ_OBJ ) 
	 && !Already_Visited[i] 
	 && Has_Prosit_Trans_Succ( Reseau, i ) ) {

      Write_Token_Total( Reseau, i, Initial, Prosit_Class );
      /* call Gen_EQ_H_Function with "-1" as trans. id. This is ok as 	*/
      /* long as routing is not 'P', which should be prevented at 	*/
      /* checking time.							*/
      Gen_EQ_H_Function( Reseau, i, -1, Initial,Single_Marking_Function );

    }
  }

}

void Gen_Loop_Switch_Code(
PetriNet	*Reseau,
int		Max_Dependency,
boolean		Prosit_Class,
char		*Class_Name )
{
  int		i;
  Liste_Link	L_Scan;
  boolean	Some;

  if ( Prosit_Class ) {
    COMLINE;
    CODE("// Procedure for incrementing the firing count vector\n");
    CODE("// Switch according to what time it is...\n");
    COMLINE;
  }
  else {
    CODE("/* Procedure for incrementing the firing count vector\t\t\n");
    CODE("/* Switch according to what time it is...\t\t\t*/\n");
  }
  CODELN;

  if ( Prosit_Class )
    CODE1("void %s::", Class_Name );
  else 
    CODE("void ");
    
  CODE("Increment( )\n{\n");
  CODE1("  if ( Real_EQ_Time >= %d )\n", Max_Dependency );
  CODE("    Increment_Regular();\n" );
  CODE("  else\n");
  CODE("    Increment_Initial();\n");
  CODELN;
  CODE("  /* Increment of the time variables */\n");
  CODE("  Real_EQ_Time ++;\n");
  CODE1("  EQ_Time = ( EQ_Time + 1 ) %% %sBUF_SIZE;\n", Prefix );
  for (i=1; i<=Max_Dependency; i++ )
    if ( Delay_Table[i] )
      CODE3("  EQ_Time_M_%d = ( EQ_Time + %d ) %% %sBUF_SIZE;\n", 
	    i, Buf_Size - i, Prefix );
  if ( Y_Static ) {
    CODE("  Y_0 = Y[ EQ_Time ];\n");
    for (i=1; i<=Max_Dependency; i++ )
      if ( Delay_Table[i] )
	CODE2("  Y_Time_M_%d = Y[ EQ_Time_M_%d ];\n", i, i );
    CODELN;
  }

  /* If some external sources/transitions are to be followed, 	      	*/
  /* copy their value.							*/
  Some = FALSE;
  for ( i=0; i<Reseau->Nb_Trans; i++ )
    if ( ( Reseau->T[i].Tag == PROSIT_OBJ )
	 && ( Has_EQ_Place_Succ( Reseau, i ) )
	 || ( Prosit_Class && ( Reseau->Nb_Inputs > 0 ) ) )
      Some = TRUE;
  if ( Some ) {
    if ( Prosit_Class && ( Reseau->Nb_Inputs > 0 ) ) {
      CODE ("  // Setting information for Prosit interface objects\n");
      CODE ("  // propagating external token counts\n");
      CODE ("  for ( int Tr_In_Nb=0; Tr_In_Nb < Nb_Input_Tr; Tr_In_Nb++ )\n" );
      CODE ("    Ext_Y[ Tr_In_Nb ][EQ_Time] = ");
      CODE ("Ext_Y[ Tr_In_Nb ][EQ_Time_M_1];\n");
    }
    else
      CODE("  /* setting information for Prosit interface objects */\n");
    for ( i=0; i<Reseau->Nb_Trans; i++ ) {
      if ( ( Reseau->T[i].Tag == PROSIT_OBJ )
	   && ( Has_EQ_Place_Succ( Reseau, i ) ) ) {
	if ( Is_Source[i] ) {
	  CODE2("  R_%d[ EQ_Time ] = R_%d[ EQ_Time_M_1 ];\n", i, i );
	}
	else {
	  CODE2("  Y[ EQ_Time ][%d] = Y[ EQ_Time_M_1 ][%d];\n", i, i );
	}
      }
    }
  }

  /* if some routing functions are to be kept, copy their value */
  Some = FALSE;
  for ( i=0; i<Reseau->Nb_Places; i++ )
    if ( ( ( Reseau->P[i].Tag == PROSIT_OBJ )
	   && Has_EQ_Trans_Succ( Reseau, i ) )
	 || ( ( Reseau->P[i].Tag == EQ_OBJ )
	      && ( Reseau->P[i].Outdegree > 1 ) )
	 || ( Prosit_Class && ( Reseau->Nb_Inputs > 0 ) )
	 )
      Some = TRUE;
  if ( Some ) {
    if ( Prosit_Class && ( Reseau->Nb_Inputs > 0 ) ) {
      CODE ("  // Setting up routing information\n");
      CODE ("  // propagating external token counts\n");
      CODE ("  for ( int In_Nb=0; In_Nb < Nb_Input_Pl; In_Nb++ ) {\n" );
      CODE ("    Ext_H[ In_Nb ][EQ_Time] = Ext_H[ In_Nb ][EQ_Time_M_1];\n");
      CODE ("  }\n");
    }
    else
      CODE("  /* Setting up routing information */\n");

    for ( i=0; i<Reseau->Nb_Places; i++ ) {
      if ( ( ( Reseau->P[i].Tag == PROSIT_OBJ )
	     && Has_EQ_Trans_Succ( Reseau, i ) )
	   || ( ( Reseau->P[i].Tag == EQ_OBJ )
		&& ( Reseau->P[i].Outdegree > 1 ) ) ) {
	SCAN( L_Scan, Reseau->P[i].Outputs ) {
/* 	  CODE2("  H_%d[EQ_Time][%d] = ", i, L_Scan->Dest ); */
/* 	  CODE2("H_%d[EQ_Time_M_1][%d];\n", i, L_Scan->Dest ); */
	  CODE4("  H_%d[EQ_Time][ %sNB_P%d_TO_T%d ] = ", 
		i, Prefix, i, L_Scan->Dest );
	  CODE4("H_%d[EQ_Time_M_1][ %sNB_P%d_TO_T%d ];\n",
		i, Prefix, i, L_Scan->Dest );
	}
      }
    }
  }

  CODE("}\n\n");
}

void Gen_Loop_Code(
PetriNet	*Reseau,
boolean		Initial,
Layer_Struct	The_Layer_Struct,
int		Max_Dependency,
boolean		Prosit_Class,
char		*Class_Name,
boolean		Single_Marking_Function )
{
  if ( Prosit_Class ) {
    CODE("// Procedure for incrementing the firing count vector\n");
    if ( Initial ) {
      CODE("// Version to handle initial firings\n" );
      CODE1("void %s::Increment_Initial()\n{\n", Class_Name );
    }
    else {
      CODE("// Version to handle regular firings\n" );
      CODE1("void %s::Increment_Regular()\n{\n", Class_Name );
    }
  }
  else {
    CODE("/* Procedure for incrementing the firing count vector\t*/\n");
    if ( Initial ) {
      CODE("/* Version to handle initial firings\t\t\t*/\n" );
      CODE("void Increment_Initial()\n{\n");
    }
    else {
      CODE("/* Version to handle regular firings\t\t\t*/\n" );
      CODE("void Increment_Regular()\n{\n");
    }
  }
  CODE("  static int\tj;\n");
  CODE("  static int\tTmp_Y;\n");
  CODE("  static double\tU;\n");
  if ( Prosit_Class )
    CODE("  static int\tIn_Nb;\n");
  CODELN;

  gen_preparation_code( Reseau, Initial, Prosit_Class, Class_Name );
  gen_loop_code( Reseau, The_Layer_Struct, Initial, Prosit_Class,
		 Single_Marking_Function );
  gen_z_code( Reseau, Initial );

  CODELN;
  CODE("} /* end of loop code */\n\n");
}

void	Gen_Init_Code(
PetriNet	*Reseau,
int		Max_Delay,
boolean		Prosit_Class,
char		*Class_Name,
boolean		Distributed )
{
  int		i;
  int		j;
  int		Line_Lg;
  int		Nb_Sources;
  Law_Desc	The_Law;
  Routing_Desc	The_Routing;
  double	Cum_Proba;
  double	Max_Lag;
  Liste_Link	L_Scan;

  CODELN;
  if ( Prosit_Class )
    CODE1("void %s::", Class_Name );
  else 
    CODE("void ");
    
  CODE("Init( )\n{\n");
  CODE("  int\ti;\n");
  CODELN;

  if ( Prosit_Class 
       && 
       ( Reseau->Nb_Inputs > 0 ) || ( Reseau->Nb_Outputs > 0 ) ) {
    if ( Reseau->Nb_Inputs > 0 ) {
      CODE ("  Nb_Input_Tr = 0;\n");
      CODE ("  Nb_Input_Pl = 0;\n");
      CODE1("  for ( i=0; i<%d; i++ ) {\n", Reseau->Nb_Inputs );
      CODE ("    Orig_Nb[i] = 0;\n");
      CODE ("    Orig_Table[i] = (int*)NULL;\n");
      CODE ("  }\n");
      CODE1("  for ( i=0; i<%d; i++ )\n", Reseau->Nb_Trans );
      CODE ("    Is_Input_Trans[i] = FALSE;\n");
      CODE1("  for ( i=0; i<%d; i++ )\n", Reseau->Nb_Places );
      CODE ("    Is_Input_Place[i] = FALSE;\n");
      CODELN;
    }
    if ( Reseau->Nb_Outputs > 0 ) {
      CODE1("  for ( i=0; i<%d; i++ ) {\n", Reseau->Nb_Outputs );
      CODE ("    Dest_Nb[i] = 0;\n");
      CODE1("    Dest_Table[i] = (%sRouting_Server**)NULL;\n",
	   ( Distributed ? "CM_" : "" ) );
      CODE("  }\n");
      CODE("  Ext_H = (int**)NULL;\n");
      CODE("  Ext_Y = (int**)NULL;\n");
      CODELN;
    }
  }

  if ( Prosit_Class ) {
    CODE("  // Initialization of the initial availabilities\n");
    for( i=0; i<Reseau->Nb_Places; i++ ) {
      /* The initialization mode depends on the number of operations	*/
      /* to perform. The idea is to avoid the enumeration of trivial	*/
      /* initializations to 0.						*/
      if ( Max_Delay < 8 ) {
	/* That'll be two lines. Guess that's ok to enumerate.		*/
	for( j=0; j<Max_Delay; j++ ) {
	  CODE3("  Avail[%2d][%d]=%d;", 
		i, j, Nb_Available_Tokens( Reseau->P[i], j ) );
	  if ( ( (j+1) % 4 == 0 ) && ( (j+1) < Max_Delay ) ) {
	    CODELN;
	  }
	}
	CODELN;
      }
      else {
	/* The simplest case is when all lags are actually 0. This	*/
	/* includes the case when there is no initial marking.		*/
	/* In this case, a single loop will do the trick.		*/
	Max_Lag = 0.0;
	for ( j=0; j < Reseau->P[i].Init_Pop; j++ )
	  Max_Lag = max( Max_Lag, Reseau->P[i].Lag_Times[j] );
	if ( Max_Lag == 0.0 ) {
	  CODE1("  for ( i=0; i<%d; i++ )\n", Max_Delay );
	  CODE2("    Avail[%2d][i] = %d;\n", i, Reseau->P[i].Init_Pop );
	}
	else {
	  /* For the moment, I just to enumeration, but it is possible	*/
	  /* to do better than that.					*/
	  for( j=0; j<Max_Delay; j++ ) {
	    CODE3("  Avail[%2d][%d]=%d;", 
		  i, j, Nb_Available_Tokens( Reseau->P[i], j ) );
	    if ( ( (j+1) % 4 == 0 ) && ( (j+1) < Max_Delay ) ) {
	      CODELN;
	    }
	  }
	  CODELN;
	}
      }
    }
  }

  CODE("  // Initialization of firing times\n");
  for( i=0; i<Reseau->Nb_Trans; i++ ) {
    CODE2("  Fire_Time[%2d]=%2d;", i, Fire_Time[i] );
    if ( ( i+1 < Reseau->Nb_Trans ) && ( (i+1) % 4 == 0 ) ) {
      CODELN;
    }
  }
  CODELN;

  /* When initializing the source flags, I assume that most transitions	*/
  /* are **not** sources!						*/
  CODE("  // Initialization of source flags\n");
  Nb_Sources = 0;
  for( i=0; i<Reseau->Nb_Trans; i++ )
    if ( Is_Source[i] )
      Nb_Sources ++;
  if ( Nb_Sources < ( Reseau->Nb_Trans / 2 ) ) {
    /* in this case, one by one affectation. */
    for( i=0; i<Reseau->Nb_Trans; i++ ) {
      CODE2("  Is_Source[%2d]=%5s;", i, 
	    ( Is_Source[i] ? "TRUE" : "FALSE" ) );
      if ( ( i+1 < Reseau->Nb_Trans ) && ( (i+1) % 3 == 0 ) ) {
	CODELN;
      }
    }
  }
  else {
    /* in this case, a loop and a case by case affectation */
    CODE1("  for ( i=0; i<%sNB_TRANS; i++ )\n", Prefix );
    CODE("    Is_Source[i] = FALSE;\n");
    j = 0;
    for( i=0; i<Reseau->Nb_Trans; i++ ) {
      if ( Is_Source[i] ) {
	CODE1("  Is_Source[%2d]= TRUE;", i );
	j++;
	if ( j % 3 == 0 ) {
	  CODELN;
	  j = 0;
	}
      }
    }
  }
  CODELN;

  CODE("  // Initialization of transition names\n");
  Line_Lg = 0;
  for( i=0; i<Reseau->Nb_Trans; i++ ) {
    j = 19 + strlen(Reseau->T[i].Name);
    if ( j+Line_Lg > 79 ) {
      CODELN;
      Line_Lg = 0;
    }
    CODE2("  Trans_Name[%2d]=\"%s\";", i, Reseau->T[i].Name );
    Line_Lg += j;
  }
  CODELN;

  CODE("  // Initialization of time variables\n");
  CODE("  Real_EQ_Time = 0;\n");
  CODE("  EQ_Time = 0;\n");

  /* source related variables */
  if ( Prosit_Class ) {
    for ( i=0; i<Reseau->Nb_Trans; i++ ) {
      if ( Is_Source[i] ) {
	The_Law = Reseau->T[i].Tr_Law;

	CODE2("  // Variables for source %d (%s)\n", i, Reseau->T[i].Name );
	CODE1("  Nb_Tok_%d = 0;\n", i );
	CODE1("  for ( i=0; i<%sBUF_SIZE; i++ )\n", Prefix );
	CODE1("    R_%d[ i ] = 0;\n", i);  
	if ( The_Law->Name == 'C' ) {
	  CODE2("  T_Phase_%d = %d;\n", i,
		1 + IRINT( The_Law->Parameters[0] ) );	  
	}
      }
    }
  }

  /* routing related variables */
  if ( Prosit_Class ) {
    for( i=0; i<Reseau->Nb_Places; i++ ) {
      The_Routing = Reseau->P[i].Routing;

      if ( The_Routing != NULL ) {
	switch ( The_Routing->Type ) {

	case 'R':
	  CODE1("  Count_%d = 0;\n", i );
	  Cum_Proba = 0.0;
	  for ( j=0; j<Reseau->P[i].Outdegree; j++ ) {
	    Cum_Proba += The_Routing->Probas[j];
	    if ( j == Reseau->P[i].Outdegree-1 )
	      Cum_Proba = 1.0;
	    CODE3("  P_Proba_%d[%d] = %6.4f;", i, j, Cum_Proba );
	    if ( (j+1 ) % 3 == 0 )
	      CODELN;
	  }
	  CODELN;
	  L_Scan = Reseau->P[i].Outputs;
	  for ( j=0; j<Reseau->P[i].Outdegree; j++ ) {
	    CODE3("  Dest_%d[%d] = %d;", i, j, L_Scan->Dest );
	    L_Scan = L_Scan->Next;
	    if ( (j+1 ) % 4 == 0 )
	      CODELN;
	  }
	  break;

	case 'D':
	  CODE1("  Count_%d = 0;\n", i );
	  CODE1("  P_Phase_%d = 0;\n", i );
	  for ( j=0; j<The_Routing->Period; j++ ) {
	    CODE2("  Cycle_%d[%d] = ", i, j );
	    CODE3(" %sNB_P%d_TO_T%d;", 
		  Prefix, i, 
		  The_Routing->Cycle[j]->Number );
	    if ( (j+1) % 2 == 0 )
	      CODELN;
	  }
	  break;

	default:
	  if ( Is_Compet_Pl[i] )
	    CODE1("  Count_%d = 0;\n", i );

	  break;
	}
      } /* if ( Routing != NULL ) */
      else if ( Is_Compet_Pl[i] )
	CODE1("  Count_%d = 0;\n", i );
    }
  }
  CODELN;
  if ( Prosit_Class ) {
    CODE ("  // Initialization of firing counts\n");
    CODE1("  for ( i=0; i<%sNB_TRANS; i++ ) {\n", Prefix );
    CODE ("    Y[0][i] = 0;\n");
    CODE ("    Z[0][i] = 0;\n");
    CODE ("  }\n");
    CODELN;
  }
  CODE("}\n\n");

}

void	Gen_Terminate_Code( Reseau, Prosit_Class, Class_Name )
PetriNet	*Reseau;
boolean		Prosit_Class;
char		*Class_Name;
{
  CODELN;
  if ( Prosit_Class )
    CODE1("void %s::", Class_Name );
  else 
    CODE("void ");
    
  CODE("Terminate( )\n{\n");
  CODE("}\n\n");
}

void	Gen_Arrival_Code( Reseau, Prosit_Class, Class_Name )
PetriNet	*Reseau;
boolean		Prosit_Class;
char		*Class_Name;
{
  int		i;
  int		j;
  int		Nb_Obj;
  int		Place_Id;
  Liste_Link	L_Scan;
  boolean	Some_Place_Arrivals = FALSE;

  if ( Prosit_Class ) {
    COMLINE;
    CODE("// Procedure for incrementing the firing count vector\n");
    CODE("// due to arrivals from the Prosit part of the network\n");
    COMLINE;
  }
  else {
    CODE("/* Procedure for incrementing the firing count vector\t*/\n");
    CODE("/* due to arrivals from the Prosit part of the network\t*/\n");
  }
  CODELN;

  if ( Prosit_Class )
    CODE1("void %s::", Class_Name );
  else 
    CODE("void ");
    
  CODE("Arrival( int The_Class, double PROSIT_Time )\n{\n");

  /* Now generate code according to the type of arrival */
  CODE("  switch( The_Class ) {\n");

  if ( Prosit_Class ) {
    /* In this case, arrivals are given by the registration procedure.	*/
    /* First counting the number of transitions that are inputs */
    Nb_Obj = 0;
    for ( i=0; i<Reseau->Nb_Inputs; i++ ) {
      if ( 'T' == Find_Type( Reseau, Reseau->Inputs[i] ) )
	Nb_Obj++;
    }
    if ( Nb_Obj > 0 ) {
      CODE("  /* arrivals to transitions */\n");
      for ( i=0; i<Reseau->Nb_Inputs; i++ ) {
	if ( 'T' == Find_Type( Reseau, Reseau->Inputs[i] ) )
	  CODE1("  case %2d:\n", i );
      }
      CODE("    Ext_H[ The_Class ][ EQ_Time ]++;\n");
      CODE("    break;\n");
    }
    /* Now counting the number of places that are inputs */
    Nb_Obj = 0;
    for ( i=0; i<Reseau->Nb_Inputs; i++ ) {
      if ( 'P' == Find_Type( Reseau, Reseau->Inputs[i] ) )
	Nb_Obj++;
    }
    if ( Nb_Obj > 0 ) {
      CODE("  /* arrivals to places */\n");
      for ( i=0; i<Reseau->Nb_Inputs; i++ ) {
	if ( 'P' == Find_Type( Reseau, Reseau->Inputs[i] ) )
	  CODE1("  case %2d:\n", i );
      }
      CODE("    Ext_Y[ The_Class ][ EQ_Time ]++;\n");
      CODE("    break;\n");
    }
    CODE("  default:\n");
    CODE("    cout << \"Error! Illegal class number \" << The_Class;\n");
    CODE("    cout << \" for arrival.\";\n");
  }
  else {
  for ( j=0; j<Reseau->Nb_Trans; j++ ) {
    if ( Reseau->T[j].Tag == PROSIT_OBJ ) {
      SCAN( L_Scan, Reseau->T[j].Outputs )
	if ( Place_Exists[L_Scan->Dest]
	     && Reseau->P[L_Scan->Dest].Tag == EQ_OBJ ) {
	  Place_Id = L_Scan->Dest;
	  CODE2("    case CL_T%d_P%d:\n", j, Place_Id );
	  if ( Is_Source[j] )
	    CODE1("      R_%d[ EQ_Time ]++;\n", j );
	  else
	    CODE1("      Y_0[%d]++;\n", j );

	  CODE ("      if ( Trace )\n");
	  CODE2("        printf(\"+++ Arr. at place #%d (%s) ",
		Place_Id+1, Reseau->P[Place_Id].Name );
	  CODE2("from trans #%d (%s)\\n\");\n",
		j+1, Reseau->T[j].Name );
	  CODE("    break;\n");
	}
    }
  }
  for ( j=0; j<Reseau->Nb_Places; j++ ) {
    if ( Reseau->P[j].Tag == PROSIT_OBJ ) {
      SCAN( L_Scan, Reseau->P[j].Outputs )
	if ( Reseau->T[L_Scan->Dest].Tag == EQ_OBJ ) {
	  CODE2("    case CL_P%d_T%d:\n", j, L_Scan->Dest );
/* 	  CODE2("      H_%d[EQ_Time][%d] ++;\n", j, L_Scan->Dest ); */
	  CODE4("      H_%d[EQ_Time][ %sNB_P%d_TO_T%d ] ++;\n",
		j, Prefix, j, L_Scan->Dest );
	  CODE("      if ( Trace )\n");
	  CODE2("        printf(\"+++ Arr. at transition #%d (%s)\\n\");\n",
		L_Scan->Dest+1, Reseau->T[L_Scan->Dest].Name );
	  CODE("    break;\n");
	}
    }
  }
  }

  CODE("  } /* end of switch */\n" );
  CODELN;

  if ( Prosit_Class ) {
    CODE("    if ( Trace ) {\n");
    CODE("      cout << \"+++ [\" << this->get_name() << ");
    CODE("\"] Arr. at t= \";\n");
    CODE("      cout << EQ_Time << \" on channel \" << ");
    CODE("The_Class << endl;\n");
    CODE("    }\n");
    CODELN;
  }
  CODE("}\n\n");

}

void	Gen_Departure_Code(
PetriNet	*Reseau,
boolean		Prosit_Class,
char		*Class_Name,
boolean		Distributed )
{
  int		i;
  int		j;
  int		Place_Id;
  int		Trans_Id;
  int		Obj_Num;
  Liste_Link	L_Scan;

  if ( Prosit_Class ) {
    COMLINE;
    CODE("// Procedure for incrementing the firing count vector\n");
    CODE("// due to arrivals from the Prosit part of the network\n");
    COMLINE;
  }
  else {
    CODE("/* Procedure for incrementing the firing count vector\t*/\n");
    CODE("/* due to arrivals from the Prosit part of the network\t*/\n");
  }
  CODELN;

  if ( No_Prosit )
    CODE("typedef char Routing_Server;\n\n");

  CODE("static int\tPrepare_New_Batch = TRUE;\n\n");

  if ( Prosit_Class )
    CODE2("%sRouting_Server *%s::",
	  ( Distributed ? "CM_" : "" ),
	  Class_Name );
  else 
    CODE("Routing_Server *");

  CODE("Next_Departure( )\n{\n");
  CODELN;

  CODE("  static int\t\tNb_Departures;\n");
  CODE("  static int\t\tIndex;\n");
  if ( Prosit_Class )
    CODE("  static int\t\tDest_Num;\n");
  CODE2("  static %sRouting_Server\t*Dpt[%sNB_MAX_DEPT];\n", 
	( Distributed ? "CM_" : "" ),
	Prefix );
  CODE("  int\t\t\tDp_Time;\n");
  CODE("  int\t\t\tDp_Time_M_1;\n");
  CODELN;

  CODE("  if ( Prepare_New_Batch ) {\n");
  if ( !No_Prosit )
    CODE("    Prepare_New_Batch = FALSE;\n");
  CODE("    Nb_Departures = 0;\n");
  CODE("    /* get --exact-- time */;\n");
  CODE("    Dp_Time = EQ_Time_M_1;\n");
  CODE2("    Dp_Time_M_1 = ( Dp_Time + %d ) %% %sBUF_SIZE;\n", 
	Buf_Size - 1, Prefix );
  CODELN;

  if ( Prosit_Class ) {
    /* In this case, destinations are known from the registration	*/
    /* procedure.							*/
    for ( i=0; i<Reseau->Nb_Outputs; i++ ) {
      CODE1("    // departures from output #%d = ", i+1 );
      Write_Address( Reseau->Outputs[i] );
      CODELN;
      if ( 'T' == Find_Type( Reseau,  Reseau->Outputs[i] ) ) {
	Obj_Num = Reseau->Outputs[i]->Number;
	CODE1("    for ( Index = Z[Dp_Time_M_1][%d];\n", Obj_Num );
	CODE1("          ( Index < Z[Dp_Time][%d] ) && ", Obj_Num );
	CODE1("( Nb_Departures < %sNB_MAX_DEPT );\n", Prefix );
	CODE ("          Index++ ) {\n");
	CODE1("      for ( Dest_Num = 0; Dest_Num < Dest_Nb[%d]; ", i );
	CODE ("Dest_Num++ ) {\n");
	CODE1("        Dpt[ Nb_Departures ] = Dest_Table[%d][Dest_Num];\n", 
	      i );
	CODE ("        Nb_Departures++;\n");
	CODE ("        if ( Trace ) {\n");
	CODE ("          cout << \"+++ [\" << this->get_name() << ");
	CODE ("\"] Dep. at t= \";\n");
	CODE2("          cout << Dp_Time << \" from trans #%d (%s)",
	      Obj_Num, Reseau->T[Obj_Num].Name );
	CODE (" to object \";\n");
	CODE1("          cout << Dest_Table[%d][Dest_Num]->get_name()", i );
	CODE (" << endl;\n");
	CODE ("      }\n");
	CODE ("    }\n");
	CODE ("  }\n");
      }
      else {
	CODE("    // This is a place. We'll see later.\n");
      }
      CODELN;
    }
  }
  else {
    /* In the "mixt" case, destinations are known from the topology	*/
    for ( j=0; j<Reseau->Nb_Trans; j++ ) {
      if ( Reseau->T[j].Tag == EQ_OBJ ) {
	SCAN( L_Scan, Reseau->T[j].Outputs ) {
	  if ( Place_Exists[L_Scan->Dest] 
	       && Reseau->P[L_Scan->Dest].Tag == PROSIT_OBJ ) {
	    Place_Id = L_Scan->Dest;
	    CODE2("    /* departures from transition #%d (%s) */\n",
		  j, Reseau->T[j].Name );
	    CODE1("    for ( Index = Z[Dp_Time_M_1][%d];\n", j );
	    CODE1("          ( Index < Z[Dp_Time][%d] ) && ", j );
	    CODE1("( Nb_Departures < %sNB_MAX_DEPT );\n", Prefix );
	    CODE ("          Index++ ) {\n");
	    if ( !No_Prosit )
	      CODE1("      Dpt[ Nb_Departures ] = Place_%d;\n", Place_Id );
	    CODE ("      Nb_Departures++;\n");
	    CODE ("      if ( Trace )\n");
	    CODE2("        printf(\"+++ Dep. from trans #%d (%s)",
		  j+1, Reseau->T[j].Name );
	    CODE2(" to place #%d (%s)\\n\");\n",
		  Place_Id+1, Reseau->P[Place_Id].Name );
	    CODE("    }\n");
	  }
	}
      }
    }
    for ( j=0; j<Reseau->Nb_Places; j++ ) {
    if ( Place_Exists[j] 
	 && Reseau->P[j].Tag == EQ_OBJ ) {
      SCAN( L_Scan, Reseau->P[j].Outputs )
	if ( Reseau->T[L_Scan->Dest].Tag == PROSIT_OBJ ) {
	  Trans_Id = L_Scan->Dest;
	  CODE2("    /* departures from place #%d (%s) */\n",
		j, Reseau->P[j].Name );
/* 	  CODE2("    for ( Index = H_%d[Dp_Time_M_1][%d];\n", j, Trans_Id ); */
/* 	  CODE2("        ( Index < H_%d[Dp_Time][%d] ) && ", j, Trans_Id ); */
	  CODE4("    for ( Index = H_%d[Dp_Time_M_1][ %sNB_P%d_TO_T%d ];\n",
		j, Prefix, j, Trans_Id );
	  CODE4("        ( Index < H_%d[Dp_Time][ NB_P%d_TO_T%d ] ) && ",
		j, Prefix, j, Trans_Id );
	  CODE1("( Nb_Departures < %sNB_MAX_DEPT );\n", Prefix );
	  CODE ("          Index++ ) {\n");
	  CODE1("      Dpt[ Nb_Departures ] = Trans_%d;\n", Trans_Id );
	  CODE ("      Nb_Departures++;\n");
	  CODE ("      if ( Trace )\n");
	  CODE2("        printf(\"+++ Dep. from place #%d (%s)",
		j, Reseau->P[j].Name );
	  CODE2(" to trans #%d (%s)\\n\");\n",
		Trans_Id, Reseau->T[Trans_Id].Name );
	  CODE("    }\n");
	}
    }
  }
  }
  CODE1("    if ( Nb_Departures >= %sNB_MAX_DEPT ) {\n", Prefix );
  CODE("      printf(\"warning: max number of simultaneous departures");
  CODE(" exceeded.\n\tSome departures ignored.\\n\");\n");
  CODE("    }\n");

  CODELN;
  CODE("  Index = 0;\n");
  CODE("  }\n");

  CODE("  if ( Index == Nb_Departures ) {\n");
  CODE("    Prepare_New_Batch = TRUE;\n");
  CODE("    Index++;\n");
  CODE("    return 0;\n");
  CODE("  }\n");
  CODE("  else {\n");
  CODE("    Index++;\n");
  CODE("    return Dpt[ Index - 1 ];\n");
  CODE("  }\n" );
  CODELN;
  CODE("}\n\n");

}

boolean Is_Dummy( The_Place )
Pl_Des	The_Place;
{
  if ( NULL != The_Place.Name )
    return FALSE;
  else
    return ( 0 == strncmp( The_Place.Name, "_Dummy_Place", 12 ) );
}

int	gen_c_code(
PetriNet	*Reseau,
Layer_Struct	The_Layer_Struct,
char		Tmp_Dir[FILENAME_LG],
char		File_Name[FILENAME_LG],
boolean		Prosit_Class,
char		*Class_Name,
boolean		Single_Marking_Function,
boolean		QNAP_Style,
boolean		Distributed )
{
  int		Max_Dependency;
  int		Max_Source_Dependency;
  int		i;
  char		Path[2*FILENAME_LG];

  Max_Dependency = rint( Scale * max( Max_Delay( Reseau ), 
				      Max_Lag( Reseau ) ) );
  
  /* adjusting buffer size and trace parameters		*/
  Buf_Size = min( Buf_Size, Scale*Reseau->Sim_Params->Nb_Firings + 1 );
  Buf_Size = max( Buf_Size, 2*Max_Dependency );

  /* Buf_Size has to be even because of the "source dependency problem" */
  if ( 1 == ( Buf_Size % 2 ) )
    Buf_Size++;

  /* initializing some data structures */
  Init_Generation( Reseau, Max_Dependency, Scale );

  /* finding more precisely the time scale of the various objects */
  TimeScale( Reseau, Scale, Prosit_Class );

  /* Now the behavior depends on what is generated. But in both cases	*/
  /* we generate first the declarations ("h"), then the code ("c").	*/

  if ( !Prosit_Class ) {
    /* All the EQ part goes into one single C file: preparing it.	*/
    strcpy( Path, Tmp_Dir );
    strncat( Path, File_Name, 2*FILENAME_LG );

    if ( NULL == freopen( Path, "w", stdout ) ) {
      fprintf( stderr, "%s: error opening %s\n", TOOLNAME, Path );
      return FALSE;
    }
  }
  else {
    /* Declarations go into file Class.hh. Preparing it.		*/
    strcpy( Path, Tmp_Dir );
    strncat( Path, Class_Name, 2*FILENAME_LG );
    strncat( Path, ".hh", 2*FILENAME_LG );

    if ( NULL == freopen( Path, "w", stdout ) ) {
      fprintf( stderr, "%s: error: could not open file %s for writing\n", 
	       TOOLNAME, Path );
      return FALSE;
    }
  }

  if ( Prosit_Class ) {
    Prefix = (char*)malloc( (2+strlen(Class_Name))*sizeof(char) );
    Prefix = strcpy( Prefix, Class_Name );
    Prefix = strcat( Prefix, "_" );
  }
  else {
    Prefix = (char*)malloc( sizeof(char) );
    Prefix[0] = '\0';
  }

  /* generating global parameters of the graph */
  if ( Prosit_Class ) {
    Gen_Class_File_Header( Reseau, Path, FALSE );
    Gen_Class_Prototype( Reseau, Max_Dependency, Scale, Class_Name,
			 Distributed, Single_Marking_Function );
  }
  else {
    Gen_EQ_File_Header( Reseau );
    Gen_EQ_Includes( Reseau, FALSE, FALSE );
    Gen_EQ_Declarations( Reseau, Max_Dependency, Scale, 
			 Prosit_Class, Class_Name, Single_Marking_Function );
  }

  if ( Prosit_Class && Distributed ) {
    /* Preparing path for file P_Class.hh				*/
    strcpy( Path, Tmp_Dir );
    strncat( Path, "P_", 2*FILENAME_LG );
    strncat( Path, Class_Name, 2*FILENAME_LG );
    strncat( Path, ".hh", 2*FILENAME_LG );
    if ( NULL == freopen( Path, "w", stdout ) ) {
      fprintf( stderr, "%s: error: could not open file %s for writing\n", 
	       TOOLNAME, Path );
      return FALSE;
    }
    else {
      /* generating proxy header */
      Gen_Class_File_Header( Reseau, Path, TRUE );
      Gen_Class_Proxy_Prototype( Class_Name );
    }
  }

  if ( Prosit_Class ) {
    /* Preparing path for file Class.cc					*/
    strcpy( Path, Tmp_Dir );
    strncat( Path, Class_Name, 2*FILENAME_LG );
    strncat( Path, ".cc", 2*FILENAME_LG );
    if ( NULL == freopen( Path, "w", stdout ) ) {
      fprintf( stderr, "%s: error: could not open file %s for writing\n", 
	       TOOLNAME, Path );
      return FALSE;
    }
    else {
      Gen_Class_File_Header( Reseau, Path, FALSE );
      Gen_EQ_Includes( Reseau, TRUE, Distributed );
      CODELN;
      CODE1("#include \"%s.hh\"\n", Class_Name );
      CODELN;
      Gen_Class_Cons_Des( Reseau, Class_Name, Distributed );
      Gen_Source_Code( Reseau, TRUE, Class_Name );
      Gen_Marking_Code( Reseau, TRUE, Class_Name, Single_Marking_Function );
    }
  }

  /* code for handling of the statistics of the EQ object		*/
  Gen_EQ_Stats_Code( Reseau, Prosit_Class, Class_Name, QNAP_Style );

  /* code for the initial loop.		*/
  Gen_Loop_Code( Reseau, TRUE, The_Layer_Struct, Max_Dependency,
		 Prosit_Class, Class_Name, Single_Marking_Function );

  /* code for the main loop. 			*/
  for ( i=0; i<Reseau->Nb_Places; i++ )
    Already_Visited[i] = FALSE;
  Gen_Loop_Code( Reseau, FALSE, The_Layer_Struct, Max_Dependency,
		 Prosit_Class, Class_Name, Single_Marking_Function );

  /* code for switching between the two loop codes */
  Gen_Loop_Switch_Code( Reseau, Max_Dependency, Prosit_Class, Class_Name );

  /* code for arrivals */
  Gen_Arrival_Code( Reseau, Prosit_Class, Class_Name );

  /* code for departures */
  Gen_Departure_Code( Reseau, Prosit_Class, Class_Name, Distributed );

  /* code for misc. initializations and post-processing		*/
  Gen_Init_Code( Reseau, Max_Dependency, Prosit_Class, 
		 Class_Name, Distributed );
  Gen_Terminate_Code( Reseau, Prosit_Class, Class_Name );

  if ( Prosit_Class ) {
    Gen_Input_Handling( Reseau, Max_Dependency, Scale, Class_Name );
    Gen_Output_Handling( Reseau, Max_Dependency, Scale, Class_Name,
			 Distributed );
  }

  return TRUE;
}

/************************************************************************/
/* Module for computing the time scale of each transition.		*/
/* Author: Bruno Gaujal, March 1996.					*/
/************************************************************************/
int Original_Trans_Time_Scale( Reseau, Trans_Id, Scale, Prosit_Class )
PetriNet	*Reseau;
int		Trans_Id;
int		Scale;
boolean		Prosit_Class;
{
  Law_Desc	The_Law;
  int		Nb_Params;
  double	Val;
  int		j;
  int		Trans_Scale;

  The_Law = Reseau->T[ Trans_Id ].Tr_Law;

  if ( !Is_Source[ Trans_Id ] ) {
    Trans_Scale = Fire_Time[ Trans_Id ];
  }
  else {
    /* if the source is also an input, it may be connected outside and	*/
    /* its time scale has to be 1.					*/
    if ( Prosit_Class )
      Trans_Scale = 1;
    else
    switch( The_Law->Name ) {
    case 'C':
      Val = Scale * The_Law->Parameters[0];
      Trans_Scale = IRINT( Val );

      break;

    case 'D':
      Nb_Params = IRINT( The_Law->Parameters[0] );
      Trans_Scale = Scale;
      for ( j=1; j<= Nb_Params; j++ ) {
	Val = Scale * The_Law->Parameters[j];
	Trans_Scale = gcd( Trans_Scale, IRINT( Val ) );
      }
      break;

    case 'G':
    case 'B':
      Trans_Scale = Scale;
      break;

    default:
      Trans_Scale = 0;
    }
  }

  return Trans_Scale;
}

void TimeScale(
PetriNet	*Petri,
int		Scale,
boolean		Prosit_Class )
{
  Liste_Link 	LinkPt;
  int 		pl_name,tr_name;
  int 		fire,hold;
  int 		p,t,i;
  int		Nb_Scales;
  int		Val;
  int		k;
  int		l;

  /******* remplissage initial ******/
  for (p = 0 ; p <  Petri->Nb_Places; p++) {

    if ( Petri->P[p].Tag == EQ_OBJ )
      TSP[p] = IRINT( Scale * Petri->P[p].Pl_Law->Parameters[0] );
    else 
      TSP[p] = Scale;
  }

  for (t = 0; t <  Petri->Nb_Trans; t++) {

    if ( Petri->T[t].Tag == EQ_OBJ )
      TST[t] =  Original_Trans_Time_Scale( Petri, t, Scale, Prosit_Class );
    else
      TST[t] = Scale;
    }


  /*** iterations ****/
  for (i=0; i< Petri->Nb_Trans + Petri->Nb_Places; i++)
    {
      /** les places **/
      for (p = 0 ; p <  Petri->Nb_Places; p++)
	{

	  fire = 0;

	  LinkPt =  Petri->P[p].Inputs;
	
	  while (LinkPt != NULL) 
	    {
	      tr_name = LinkPt->Dest;
	      if ((fire != 0) && (TST[tr_name] != 0)) 
		{fire = gcd(fire, TST[tr_name]);}
	      else {fire = max(fire,TST[tr_name]);}
	      LinkPt = LinkPt->Next;
	    }
	
	  if ((fire != 0) && (TSP[p] != 0)) {TSP[p] = gcd(TSP[p],fire);}
	  else {TSP[p] = max(fire,TSP[p]);}

	}

      /** les transitions **/
      for (t = 0; t <  Petri->Nb_Trans; t++)
	{

	  hold =0;
	  LinkPt =  Petri->T[t].Inputs;

	  while (LinkPt != NULL) 
	    {
	      pl_name = LinkPt->Dest;
	      if ((hold != 0) && (TSP[pl_name] != 0)) {
		hold = gcd(hold, TSP[pl_name]);
	      }
	      else {
		hold = max(hold,TSP[pl_name]);
	      }
	    
	      LinkPt = LinkPt->Next;
	    }
	
	  if ((TST[t]!= 0)  && (hold != 0)){TST[t] = gcd(TST[t],hold);}
	  else  {TST[t] = max(hold,TST[t]);}
	}
    
    }

  /* Recognition and sort of the different time scales */
  The_Scales = (int*) calloc( (unsigned)(1+Petri->Nb_Trans), sizeof( int ) );
  Nb_Scales = 0;
  for ( i=0; i<Petri->Nb_Trans; i++ )
    /* Sources are not registered into a time scale --unless-- they	*/
    /* may be connected with the outside world.				*/
    if ( !Is_Source[i] ) {
      Val = TST[i];
      for( k=1; (k<1+Nb_Scales) && (Val>The_Scales[k]); k++ );
      if ( ( k == 1+Nb_Scales ) || ( Val < The_Scales[k] ) ) {
	for( l=Nb_Scales+1; l>k; l-- )
	  The_Scales[l] = The_Scales[l-1];
	The_Scales[k] = Val;
	Nb_Scales++;
      }
    }
  The_Scales[0] = Nb_Scales;

  /* Analysis finished. Tracing if needed. */
  if ( Debug ) {
    printf("Found %d different time scales:", Nb_Scales );
    for ( i=1; i<=Nb_Scales; i++ )
      printf(" %d", The_Scales[i] );
    printf("\n");
    printf("Time Scales:\n");
    for ( i = 0 ; i < Petri->Nb_Places; i++) {
      printf("Pl %2d (%8s) = %2d    ",
	     i, Petri->P[i].Name, TSP[i]);
      if ( ( ( i+1 ) % 2 ) == 0 )
	printf("\n");
    }
    printf("\n");
    for ( i = 0 ; i < Petri->Nb_Trans; i++) {
      printf("Tr %2d (%8s) = %2d    ",
	     i, Petri->T[i].Name, TST[i]);
      if ( ( ( i+1 ) % 2 ) == 0 )
	printf("\n");
    }
    printf("\n");
  }

}

/************************************************************************/
/* Procedure for pretty printing. Normally called only by CODE macros	*/
/************************************************************************/
void Indent( void )
{
  printf("%*c", Indent_Nb, ' ' );
}

boolean	is_serial_place( The_Place )
Pl_Des	*The_Place;
{
  return ( (The_Place->Indegree == 1 ) 
	   && (The_Place->Outdegree == 1 ) );
}

/* test if a law is CST(0.0)						*/
boolean	is_null_time( The_Law )
Law_Desc	The_Law;
{
  return ( The_Law->Name == 'C' ) 
    && ( The_Law->Parameters[0] == 0.0 );
}

/* test si une transition est recyclee...				*/
boolean	is_recycled( Reseau, The_Trans )
PetriNet        *Reseau;
Tr_Des		*The_Trans;
{
  Liste_Link	L_Scan;
  boolean	Is_Recycled = FALSE;
  Pl_Des	*The_Place;

  SCAN( L_Scan, The_Trans->Outputs ) {
    The_Place = &(Reseau->P[L_Scan->Dest]);
    Is_Recycled = Is_Recycled ||
      ( ( The_Place->Outdegree == 1 )
	&& ( The_Place->Outputs->Dest == The_Trans->Tr_Id )
	&& ( is_null_time( The_Place->Pl_Law ) ) );
  }

  return Is_Recycled;
}

/* Test is a place is a recycling place.				*/
/* A recycling place may have several tokens.				*/
boolean	is_recycling_place( The_Place )
Pl_Des		*The_Place;
{
  boolean	Is_Recycling = FALSE;

  Is_Recycling = ( ( The_Place->Indegree == 1 ) 
		   && ( The_Place->Outdegree == 1 )
		   && ( The_Place->Inputs->Dest 
			== The_Place->Outputs->Dest )
		   && ( The_Place->Inputs->Weight == 1 )
		   && ( The_Place->Outputs->Weight == 1 )
		   && is_null_time( The_Place->Pl_Law ) );

  return Is_Recycling;
}

int server_degree( Reseau, The_Trans )
PetriNet	*Reseau;
Tr_Des		*The_Trans;
{
  Liste_Link	L_Scan;
  Pl_Des	*The_Place;
  int		Degree = -1;		/* convention for "infinite" */

  SCAN( L_Scan, The_Trans->Outputs ) {
    The_Place = &(Reseau->P[L_Scan->Dest]);
    if ( ( The_Place->Outdegree == 1 )
	 && ( The_Place->Outputs->Dest == The_Trans->Tr_Id )
	 && ( is_null_time( The_Place->Pl_Law ) ) ) {
      /* in case there are several recyclings, take the min degree */
      if ( Degree == -1 )
	Degree = The_Place->Init_Pop;
      else if ( Degree > The_Place->Init_Pop )
	Degree = The_Place->Init_Pop;
    }
  }
		 
  return Degree;
}

/* Check if the network has a sink in the Prosit part 			*/
/* A sink is a transition without outputs.				*/
/* The net has a sink if there is a place without outputs (terminal)	*/
boolean Has_Sinks( Reseau )
PetriNet	*Reseau;
{
  int		i;
  boolean	Has_Sinks = FALSE;

  for ( i=0; !Has_Sinks && (i<Reseau->Nb_Trans); i++ )
    if ( Reseau->T[i].Tag == PROSIT_OBJ )
      Has_Sinks = ( Real_Outdegree[i] == 0 );
  for ( i=0; !Has_Sinks && (i<Reseau->Nb_Places); i++ )
    if ( Reseau->P[i].Tag == PROSIT_OBJ )
      Has_Sinks = ( Reseau->P[i].Outdegree == 0 );

  return Has_Sinks;
}

/* procedure pour initialiser le tableau Place_Exists, utile a 		*/
/* l'optimisation mais aussi elimine les places de recyclage.		*/
void	init_place_status( Reseau )
PetriNet	*Reseau;
{
  int		i;
  Pl_Des	*The_Place;

  /* First, determine if places "exist".				*/
  /* A place exists if is is not a recycling place, and, depending on	*/
  /* optimization, if it is not serial or has a non-zero holding time 	*/
  /* A place which has an Equational object in output is always	present.*/
  /* Terminal places always exist.					*/
  Place_Exists = (boolean*) calloc( (unsigned)Reseau->Nb_Places, 
				    sizeof(int) );
  Trans_Loop_Of = (int*) calloc( (unsigned)Reseau->Nb_Places, 
				 sizeof(int) );

  for ( i=0; i<Reseau->Nb_Places; i++ ) {
    The_Place = &(Reseau->P[i]);
    if ( is_recycling_place( The_Place ) ) {
      Place_Exists[i] = FALSE;
      Trans_Loop_Of[i] = The_Place->Inputs->Dest;
      if (Debug) {
	printf("Recycling place Pl %d ", i );
	printf("(of Tr %d) removed\n", Trans_Loop_Of[i] );
      }
    }
    else if ( Optimize ) {
      Place_Exists[i] = 
	! ( ( The_Place->Outdegree == 1 )
	    && ( The_Place->Indegree == 1 )
	    && ( The_Place->Pl_Law->Name == 'C' )
	    && ( The_Place->Pl_Law->Parameters[0] == 0.0 )
	    && ( Reseau->T[The_Place->Outputs->Dest].Tag == PROSIT_OBJ )
	    );
      if (Debug && !Place_Exists[i])
	printf("Pl %d removed\n", i );
    }
    else
      Place_Exists[i] = TRUE;
  }

  /* Next, compute real degrees of transitions */
  Real_Indegree = (int*) calloc( (unsigned)Reseau->Nb_Trans, sizeof(int) );
  Real_Outdegree = (int*) calloc( (unsigned)Reseau->Nb_Trans, sizeof(int) );
  for ( i=0; i<Reseau->Nb_Trans; i++ ) {
    Real_Indegree[i] = Reseau->T[i].Indegree;
    Real_Outdegree[i] = Reseau->T[i].Outdegree;
    if ( is_recycled( Reseau, &(Reseau->T[i]) ) ) {
      Real_Indegree[i]--;
      Real_Outdegree[i]--;
    }
  }

}

void Dump_Bad_Objects( Reseau, Out_File_Name, Bad_Objects, GSS_Format )
PetriNet	*Reseau;
char		*Out_File_Name;
int		*Bad_Objects;
boolean		GSS_Format;
{
  FILE		*OutFile;
  int		i;
  int		Obj_Nb;

  if ( strlen( Out_File_Name ) != 0 )
    OutFile = fopen( Out_File_Name, "w" );
  else
    OutFile = stdout;

  if ( OutFile == NULL ) {
    fprintf( stderr, "%s: Error opening file %s.\n", 
	     TOOLNAME, Out_File_Name );
  }
  else {

    if ( GSS_Format ) {
      fprintf( OutFile, 
	       "# List of non simulable elements generated by %s\n", 
	       TOOLNAME );
      fprintf( OutFile, "# Codes: transition = 0, place = 1\n#\n");
      fprintf( OutFile, "# Number of non simulable elements:\n#\n" );
      
      fprintf( OutFile, "\n%d\n_E_O_E\n", Bad_Objects[0] );
      for( i=1; i<=Bad_Objects[0]; i++ ) {
	fprintf( OutFile, "\n%d\n_E_O_E\n", Bad_Objects[2*i-1] );
	fprintf( OutFile, "\n%d\n_E_O_E\n", Bad_Objects[2*i] + 1 );
      }
    }
    else {
      for( i=1; i<=Bad_Objects[0]; i++ ) {
	Obj_Nb = Bad_Objects[2*i];
	if ( Bad_Objects[2*i-1] == 0 )
	  fprintf( OutFile, "Bad transition #%d (%s)\n", 
		   Obj_Nb + 1, Reseau->T[Obj_Nb].Name );
	else 
	  fprintf( OutFile, "Bad place #%d (%s)\n", 
		   Obj_Nb + 1, Reseau->P[Obj_Nb].Name );
      }
    }

    fclose( OutFile );
  }
}

/* Find a valid class name from the specifications of the command line	*/
/* and the name of the input file.					*/
/* Assertion: the argument Code_Name cannot be NULL.			*/
boolean Set_Class_Name( Class_Name, Code_Name )
char	**Class_Name;
char	Code_Name[FILENAME_LG];
{
  int	Lg;
  int	Index;

  Lg = ( NULL == (*Class_Name) ? 0 : strlen(*Class_Name) );

  if ( Lg == 0 ) {
    /* Not good. Have to find alternate name! */
    if ( strlen( Code_Name ) == 0 )
      return FALSE;
    else {
      (*Class_Name) = strdup( Code_Name );
      /* now get rid of extension if any */
      for( Index = strlen( Code_Name )-1; 
	   ( Index >= 0 ) && ( Code_Name[Index] != '.' );
	   Index-- );
      if ( Index >= 0 ) {
	/* a period was found. Cut off tail */
	(*Class_Name)[Index] = '\0';
      }
      return TRUE;
    }
  }
  else 
    return TRUE;
}

int main(int	argc,
	 char	*argv[] )
{
  static char		Code_File_Name[FILENAME_LG] = "";
  static char		Out_File_Name[FILENAME_LG] = "";
  static char		EQ_Module_File_Name[FILENAME_LG] = "EQ_Module.c";
  static char		Prosit_Code_File_Name[FILENAME_LG] = "main.cc";
  static char		Tmp_Dir[FILENAME_LG] = "";
  PetriNet 		*Reseau;
  PetriNet 		*New_Reseau;
  int			i;
  boolean		Help = FALSE;
  boolean		Args_Ok = TRUE;
  boolean		File_Ok = TRUE;
  boolean		Network_Ok = TRUE;
  boolean		Testing = FALSE;
  boolean		For_GSS = FALSE;
  boolean		Prosit_Class = FALSE;
  boolean		Single_Marking_Function;
  boolean		QNAP_Style = FALSE;
  boolean		Distributed = FALSE;
  boolean		Split = FALSE;
  int			Max_Time = 1000;
  Layer_Struct		The_Layer_Struct;
  FILE			*OutFile;
  int			*Bad_Objects;
  char			*Class_Name = NULL;

  i = 1;

  if (argc>1)
    do {

      if GET_INT( "-bs", Buf_Size )
      else if GET_INT( "-scaling", Scale )
      else if GET_INT( "-tmax", Max_Time )

      else if GET_STRING( "-f", Code_File_Name )
      else if GET_STRING( "-o", Out_File_Name )			  
      else if GET_STRING( "-oc", EQ_Module_File_Name )
      else if GET_STRING( "-op", Prosit_Code_File_Name )

      else if SWITCH( "-debug", Debug, TRUE )
      else if SWITCH( "-testing", Testing, TRUE )
/*       else if SWITCH( "-gss", For_GSS, TRUE ) */
      else if SWITCH( "-noYstatic", Y_Static, FALSE )

      else if SWITCH( "-debug", Debug, TRUE )
      else if SWITCH( "-agat", Agat, TRUE )
      else if SWITCH( "-noprosit", No_Prosit, TRUE )
      else if SWITCH( "-nostats", No_Stats, TRUE )
      else if SWITCH( "-class", Prosit_Class, TRUE )
      else if SWITCH( "-single_marking", Single_Marking_Function, TRUE )
      else if SWITCH( "-qnap", QNAP_Style, TRUE )
      else if SWITCH( "-distributed", Distributed, TRUE )
      else if SWITCH( "-split", Split, TRUE )

      else if GET_INT_SWITCH( "-O", Opt_Level, Optimize, TRUE )

      else if IS_VERSION

      else if ( ( 0 == strcmp( "-h", argv[i] ) )
		|| ( 0 == strcmp( "-?", argv[i] ) ) ) {
	printf("Usage: %s [-f <file>] [-bs <buf_size>] ", TOOLNAME );
	printf("[-scaling <int>] [-tmax <int>]\n\t[-testing] ");
	printf("[-tmp <dir>] [-o <file>] [-op <file>] [-oc <file>]\n\t");
	printf("[-agat] [-O <int>] ");
	printf("[-nostats] [-noprosit] [-noYstatic]\n\t");
	printf("[-class] [-classname <name>] [-qnap] ");
	printf("[-distributed] [-split] [-debug]\n\t");
	printf("[-single_marking] [-h] [-?]");
	printf("\n");
	Args_Ok = FALSE;
	Help = TRUE;
	i++;
      }
      else if ( 0 == strcmp( "-tmp", argv[i] ) ) {
	i++;
	if ( i < argc ) {
	  strcpy( Tmp_Dir, argv[i] );
	  if ( ( strlen( Tmp_Dir ) != 0 )
	       && (Tmp_Dir[strlen(Tmp_Dir)-1] != '/' ) )
	    strcat( Tmp_Dir, "/" );
	  i++;
	}
      }
      else if ( 0 == strcmp( "-classname", argv[i] ) ) {
	i++;
	if ( i < argc ) {
	  Class_Name = (char*)(malloc) (1+strlen(argv[i])*sizeof(char));
	  Class_Name = strcpy( Class_Name, argv[i] );
	  i++;
	}
	else {  
	  printf("%s: directory expected after %s\n", TOOLNAME, argv[i-1] );
	  Args_Ok = FALSE;
	}
      }
      else
	i++; /* skip unknown argument */
    }
    while ( Args_Ok && ( i < argc ) );
    
  if (Args_Ok) {

    /* taking care of switch dependencies */
    /* Agat implies Stats */
    if ( Agat && No_Stats ) {
      printf("%s: error: -agat and -nostats are incompatible. Choose one.\n",
	     TOOLNAME );
      return -1;
    }
      
    if ( For_GSS )
      Testing = TRUE;

    /* reading in the network */
    Reseau = Read_Network( Code_File_Name );

    if ( Reseau != NULL ) {
      if ( Reseau->Nb_Submodels > 0 ) {
	if ( Debug ) {
	  printf("%s: warning: the network contains submodels. ", TOOLNAME );
	  printf("Flattening it...\n");
	}

	/* flattening with expansion of names */
	New_Reseau = Flatten_PN( Reseau, TRUE );
      }
      else
	New_Reseau = Reseau;

      /* filling the table of "Prosit" objects */
      Tag_Objects( New_Reseau, Prosit_Class );
      
      /* first checking "local" correctness of the network */
      Network_Ok = 
	satisfies_constraints( New_Reseau, Testing, &Bad_Objects );

      if ( Network_Ok ) {
	Scale = Find_Time_Scale( New_Reseau, TRUE, Scale,
				 Testing, Bad_Objects );
	Network_Ok = ( Scale != 0 );
	if ( Debug )
	  printf("%s: time scale found: %d\n", TOOLNAME, Scale );
      }
      else if ( Testing ) {
	Dump_Bad_Objects( New_Reseau, Out_File_Name, Bad_Objects, For_GSS );
      }

      if ( Network_Ok ) {
	/* Initialisation of flags indicating if places are really there. */
	/* done before redirecting output, so that debug is visible 	*/
	init_place_status(New_Reseau);

	if ( strlen(Out_File_Name) != 0 )
	  File_Ok = ( NULL != freopen( Out_File_Name, "w", stdout ) );

	if ( File_Ok ) {

	  /* Go for structural analysis and code generation */
	  The_Layer_Struct = (Layer_Struct) 
	    malloc( sizeof( struct LAYER_STRUCT ) );

	  if ( topological_analysis( New_Reseau, The_Layer_Struct ) ) {

	    if (Debug)
	      Dump_Layer_Struct( The_Layer_Struct );

	    if ( Max_Time >= 0 )
	      New_Reseau->Sim_Params->Nb_Firings = Max_Time;

	    /* Now can **really** start code generation. */

	    if ( Prosit_Class ) {
	      /* Generate the code for a single class */
	      /* First check the validity of the class name */
	      if ( !Set_Class_Name( &Class_Name, Code_File_Name ) ) {
		printf("%s: error: could not generate valid class name.",
		       TOOLNAME );
		printf(" No code.\n");
		return -4;
	      }
	    }

	    gen_c_code( New_Reseau, The_Layer_Struct,
			Tmp_Dir, EQ_Module_File_Name,
			Prosit_Class, Class_Name, 
			Single_Marking_Function,
			QNAP_Style, Distributed );

	    if ( !Prosit_Class && ! No_Prosit ) {
	      gen_prosit_code( New_Reseau, Tmp_Dir,
			       Prosit_Code_File_Name );
	      gen_prosit_makefile( Tmp_Dir, Prosit_Code_File_Name,
				   EQ_Module_File_Name );

	      if ( Agat )
		gen_agat_code( New_Reseau );
	    }

	    /* code generation successful! Cleanup... */
	    Free_Layer_Struct( The_Layer_Struct );
	    if ( Prosit_Class )
	      free( Class_Name );
	    Free_Petri_Net( New_Reseau );
	    Free_Memory();
	    
	    return 0;
	  }
	  else {
	    fprintf( stderr, "Topological analysis failed: ");
	    fprintf( stderr, "look for an zero-time cycle.\n");
	    return -3;
	  }
	}
	else {
	  fprintf( stderr, "%s: Error opening file %s.\n", 
		   TOOLNAME, Code_File_Name );
	  return -2;
	}
      }
      else {
	fprintf( stderr, "%s: the network does not ", TOOLNAME );	
	fprintf( stderr, "satisfy the constraints. No translation.\n" );
	return -3;
      }
    }
    else {
      fprintf( stderr, "%s: Error reading network. No translation.\n", 
	       TOOLNAME );	
      return -3;
    }
  }
  else if (!Help)
    return -1;

  /* if not exited otherwise, it's ok */
  return 0;
}

void print_trans_name( The_Trans )
Tr_Des		*The_Trans;
{
  if ( ( strlen( The_Trans->Name ) == 0 )
       || !strcmp( The_Trans->Name, " " ) ) {
    if ( strlen( The_Trans->Data ) != 0 )
      printf("%s", The_Trans->Data );
    else 
      printf("<no name>");
  }
  else
    printf("%s", The_Trans->Name );
}

void print_place_name( The_Place )
Pl_Des		*The_Place;
{
  if ( ( strlen( The_Place->Name ) == 0 )
       || !strcmp( The_Place->Name, " " ) ) {
    if ( strlen( The_Place->Data ) != 0 )
      printf("%s", The_Place->Data );
    else 
      printf("<no name>");
  }
  else
    printf("%s", The_Place->Name );
}

/* procedure to write the name of the actual destination of customers	*/
/* going out of some --place-- through link The_Link.			*/
void print_prosit_place_output_dest( Reseau, The_Link )
PetriNet	*Reseau;
Liste_Link	The_Link;
{
  int		The_Dest;

  The_Dest = The_Link->Dest;

  if ( Real_Indegree[The_Dest] == 1 )
    /* there is no fusion at this trans */
    printf("Trans_%d", The_Dest );
  else 
    printf("Fus_%d", The_Dest );
}

/* procedure to write the name of the actual destination of customers	*/
/* going out of some --transition-- through link The_Link.		*/
/* This takes into account the possible disapearance of places due to	*/
/* optimisation, and the possible appearance of fusion objects.		*/
void print_prosit_trans_output_dest( Reseau, The_Link )
PetriNet	*Reseau;
Liste_Link	The_Link;
{
  Pl_Des	*The_Place;
  int		The_Dest;

  The_Dest = The_Link->Dest;
  The_Place = &(Reseau->P[The_Dest]);

  if ( Place_Exists[ The_Dest ] ) {
    if ( Reseau->P[ The_Dest ].Tag == PROSIT_OBJ ) {
      printf("Place_%d", The_Dest );
    }
    else 
      printf("The_EQ");
  }
  else {
    /* normally, there is only one output to this place */
    print_prosit_place_output_dest( Reseau, The_Place->Outputs );
  }
}

/* procedure to write what is the actual destination of customers going	*/
/* out of transition Trans_Nb through link The_Link.			*/
void print_prosit_dest( Reseau, The_Link )
PetriNet	*Reseau;
Liste_Link	The_Link;
{
  Pl_Des	*The_Place;
  int		The_Dest;

  The_Dest = The_Link->Dest;
  The_Place = &(Reseau->P[The_Dest]);

  if ( Place_Exists[ The_Dest ] ) {
    if ( Reseau->P[ The_Dest ].Tag == PROSIT_OBJ ) {
      printf("Place_%d", The_Dest );
    }
    else 
      printf("The_EQ");
  }
  else {
    /* normally, there is only one output to this place */
    The_Dest = The_Place->Outputs->Dest;
    if ( Real_Indegree[The_Dest] == 1 )
      printf("Trans_%d", The_Dest );
    else 
      printf("Fus_%d", The_Dest );
  }
}

/* Function to compute the class number of customers arriving to a	*/
/* transition through link The_Link.					*/
int Class_Number( Reseau, The_Link )
PetriNet	*Reseau;
Liste_Link	The_Link;
{
  Pl_Des	*The_Place;
  int		The_Orig;

  The_Orig = The_Link->Dest;
  The_Place = &(Reseau->P[The_Orig]);  

  if ( Place_Exists[ The_Orig ] )
    return The_Orig;
  else {
    /* this place should be input serial, otherwise: confusion... */
    if ( The_Place->Indegree == 1 ) 
      return The_Place->Inputs->Dest;
    else {
      fprintf( stderr, "%s: error: impossible to compute input class",
	       TOOLNAME );
      fprintf( stderr, "at place #%d (%s)\n", The_Orig, The_Place->Name );
      return 0;
    }
  }
}

/* procedure to write the initial population in the case the place	*/
/* does not exist. Requires first the computation of the object which 	*/
/* is output of this place.						*/
void print_initial_pop( Reseau, Place_Num )
PetriNet	*Reseau;
int		Place_Num;
{
  Pl_Des	*The_Place;
  int		The_Dest;

  The_Dest = Reseau->P[Place_Num].Outputs->Dest;
  The_Place = &(Reseau->P[Place_Num]);

  if ( !is_recycling_place( The_Place ) ) {
    /* if everything is consistent, the place has a single output */
    printf("\n\t// code for (nonexistent) place #%d (", Place_Num );
    print_place_name( The_Place );
    printf(")\n");

    printf("\tTrans_%d->insert_customer( DUMMY_CLASS, ", The_Dest );
    printf("0 /* prio */, 0.0 /* Time */, %d );\n", The_Place->Init_Pop );
  }
}

/* Procedure to write the law. Adapted from the same from gen_pn.c */
int		Seed1 = 308;
int		Seed2 = 725;

void print_prosit_law(The_Law)
Law_Desc	The_Law;
{

  switch( The_Law->Name ) {
  case 'E':
    printf("new NegativeExpntl( %8.4f, new MLCG(%d,%d) )", 
	   The_Law->Parameters[0], ++Seed1, ++Seed2 );
    break;
  case 'C':
    if ( The_Law->Parameters[0] == 0.0 )
      printf("Zero");
    else
      printf("new Cst( %8.4f )", The_Law->Parameters[0] );
    break;
  case 'B':
    /* Bernoulli(p) = Binomial( 1, p ) */
    printf("new Binomial( 1, %8.4f, new MLCG(%d,%d) )", 
 	   The_Law->Parameters[0], ++Seed1, ++Seed2 );
     break;
   case 'G':
     /* Geometric is called with mean */
     printf("new Geometric( %8.4f, new MLCG(%d,%d) )",
 	   The_Law->Parameters[0] / ( 1 - The_Law->Parameters[0] ),
	    ++Seed1, ++Seed2 );
     break;
  case 'U':
    printf("new Uniform( %8.4f, %8.4f, new MLCG(%d,%d) )", 
	   The_Law->Parameters[0], The_Law->Parameters[1],
	   ++Seed1, ++Seed2 );
    break;
/*     printf("HEXP( %8.4f, %8.4f );\n", */
/* 	   The_Law->Parameters[0], */
/* 	   The_Law->Parameters[1] ); */
/*     break; */
   case 'R':
     /* Erlang is called with mean and variance */
     printf("new Erlang( %8.4f, %8.4f, new MLCG(%d,%d) )", 
	    The_Law->Parameters[1],
	    The_Law->Parameters[1] * The_Law->Parameters[1] 
	    * The_Law->Parameters[0],
	    ++Seed1, ++Seed2 );
     break;
  case 'D':
    fprintf( stderr, "%s: warning: unimplemented firing law (%c).", 
	    TOOLNAME, The_Law->Name );
    fprintf( stderr, "(constant assumed)\n");
    printf("new Uniform( %8.4f, %8.4f, new MLCG(%d,%d) )", 
	   The_Law->Parameters[1], The_Law->Parameters[1],
	   ++Seed1, ++Seed2 );
    break;
  case 'O':
    printf("new OnOff( %8.4f, %8.4f, %8.4f, new MLCG(%d,%d) )", 
	   The_Law->Parameters[0], 
	   The_Law->Parameters[1], 
	   The_Law->Parameters[2],
	   ++Seed1, ++Seed2 );
    break;
  case 'H':
  case 'g':
    fprintf( stderr, "%s: error: can't handle firing law (%c)\n", 
	    TOOLNAME, The_Law->Name );
    break;
  default:
    fprintf( stderr, "%s: error: unknown firing law (%c)\n", 
	    TOOLNAME, The_Law->Name );
  }
}

void gen_prosit_makefile( Tmp_Dir, Code_File_Name, EQ_File_Name )
char	Tmp_Dir[FILENAME_LG];
char	Code_File_Name[FILENAME_LG];
char	EQ_File_Name[FILENAME_LG];
{
  char		Path[2*FILENAME_LG];
  char		*Prosit_Home;

  strcpy( Path, Tmp_Dir );
  strcat( Path, "Makefile.Prosit" );

  if ( NULL != freopen( Path, "w", stdout ) ) {

    Prosit_Home = getenv( "PROSIT_HOME" );
    if ( Prosit_Home != NULL )
      printf("include %s/Makefile.head\n", Prosit_Home );
    else
      printf("include %s/Makefile.head\n", PROSIT_DIR );

  printf("\n");
  printf("PROGRAM = sim\n");
  printf("ARCH=SUN4\n");
  printf("OPTIONS+=-D_SUN4\n");
  printf("\n");
  printf("CC            = g++\n");
  printf("LINKER        = $(CC)\n");
  printf(" \n");
  printf("CFLAGS =\n");
  printf("CFLAGS += -I$(RUNTIME_SEQUENTIAL_INCLUDE_DIR)\n");
  printf("CFLAGS += -I$(QUEUEING_SEQUENTIAL_INCLUDE_DIR)\n");
  printf("CFLAGS += -I$(EXAMPLES_SEQUENTIAL_INCLUDE_DIR) \n");
  printf("CFLAGS += -I$(THREAD_INCLUDE_DIR)/$(ARCH) \n");
  printf("CFLAGS += -I$(STAT_TOOL_LOCAL_CPP_INCLUDE_DIR)\n");
  printf("CFLAGS += -I$(STAT_TOOL_LOCAL_C_INCLUDE_DIR)  \n");
  /*   printf("CFLAGS += -I$(STAT_TOOL_DIS_CPP_INCLUDE_DIR)\n"); */
  /*   printf("CFLAGS += -I$(STAT_TOOL_DIS_C_INCLUDE_DIR)    \n"); */
  printf("\n");
  printf("EXTLIBS       =\n");
  printf("EXTLIBS       += -l$(EXAMPLES_SEQUENTIAL_LIBRARY)\n");
  printf("EXTLIBS       += -l$(QUEUEING_SEQUENTIAL_LIBRARY)\n");
  printf("EXTLIBS       += -l$(RUNTIME_SEQUENTIAL_LIBRARY)\n");
  printf("EXTLIBS       += -l$(THREAD_LIBRARY) \n");
  /*   printf("EXTLIBS       += -l$(STAT_TOOL_LOCAL_C_LIBRARY)  \n"); */
  printf("EXTLIBS       += -l$(STAT_TOOL_LOCAL_CPP_LIBRARY)\n");
  printf("EXTLIBS       += -l$(AGAT_LIBRARY)\n");
  printf("\n");
  printf("EXTLIBS       += -l$(TRACE_TOOL_KER_LIBRARY)\n");
  printf("EXTLIBS       += -l$(TRACE_TOOL_STD_LIBRARY) \n");
  printf("\n");
  printf("EXTLIBS       +=  -L$(TRACE_TOOL_LIB_DIR)/$(ARCH)\n");
  printf("EXTLIBS       += -L$(LIB_DIR)/$(ARCH)\n");
  printf("EXTLIBS       += -L$(STAT_TOOL_LIB_DIR)/$(ARCH)\n");
  printf("EXTLIBS       += -L$(AGAT_LIB_DIR)/$(ARCH)\n");
  printf("\n");
  printf("LIBS          = -lm -lg++\n");
  printf("\n");
  printf("MAIN          = main.o\n");
  printf("EQ            = EQ_Module.o\n");
  printf("\n");
  printf("EXL = \n");
  printf("\n");
  printf("\n");
  printf("$(PROGRAM):   $(MAIN) $(EQ) $(EXL)\n");
  printf("\t$(LINKER) -o $(PROGRAM) $(MAIN) $(EQ)");
  printf(" $(LDFLAGS) $(EXL) $(LIBS) $(EXTLIBS)\n");
  printf("\n");
  printf("main.o:             %s \n", Code_File_Name );
  printf("\t$(CC) $(OPTIONS) $(CFLAGS) -c %s -o main.o\n", Code_File_Name );
  printf("\n");
  printf("EQ_Module.o:        %s \n", EQ_File_Name );
  /*   printf("\t$(CC) $(OPTIONS) $(CFLAGS) -O2 -c %s -o EQ_Module.o\n",  */
  /* 	 EQ_File_Name ); */
  printf("\t$(CC) $(OPTIONS) $(CFLAGS) -c %s -o EQ_Module.o\n", 
	 EQ_File_Name );
  printf("\n");
  printf("clean:\n");
  printf("\t/bin/rm  -f $(EXL) $(MAIN) $(EQ)\n");
  printf("\n");

  }
  else {
    fprintf( stderr, "%s: Error opening file %s.\n", TOOLNAME, Path );
  }
}

void gen_prosit_place( Reseau, Place_Num )
PetriNet	*Reseau;
int	        Place_Num;
{
  Pl_Des	*The_Place;
  int		Nb_Succ;
  Liste_Link	L_Scan;
  boolean	Is_Immediate;
  boolean	Is_Terminal;

  The_Place = &(Reseau->P[Place_Num]);

  Is_Terminal = ( Reseau->P[Place_Num].Outdegree == 0 );
  Is_Immediate = is_null_time( The_Place->Pl_Law );

  printf("\n\t// code for place #%d (", Place_Num );
  print_place_name( The_Place );
  printf(")\n");

  CODE2("\t%sPlace_%d = new ",
	( Is_Interface_Pl[Place_Num] ? "" :
	  ( Is_Immediate && !Is_Terminal ? 
	    "Pastec_Routing *" : 
	    "Pastec_Server_Normal_FIFO *" )
	  ),
	Place_Num );

  if ( Is_Immediate && !Is_Terminal ) {
    CODE("Pastec_Routing( 0.0 );\n");
  }
  else {
    printf("Pastec_Server_Normal_FIFO( 0.0, INFINI );\n" );
    printf("\tPlace_%d->set_service(\n", Place_Num);
    printf("\t\t");
    if ( Is_Terminal ) {
      printf("new Cst( T_Max )\n");
      fprintf( stderr, "%s: warning: terminal place encountered at ",
	       TOOLNAME );
      fprintf( stderr, "place #%d (%s)\n", 
	       Place_Num, Reseau->P[Place_Num].Name );
    }
    else			 
      print_prosit_law( The_Place->Pl_Law );
    printf(",\n\t\t1 );\n");
  }

  CODE2("\tPlace_%d->set_name(\"%s\");\n", Place_Num, The_Place->Name );
  CODE2("\tPlace_%d->set_gss_name(\"%s\");\n", Place_Num, The_Place->Data );
  if ( The_Place->Init_Pop > 0 ) {
    printf("\tPlace_%d->insert_customer( DUMMY_CLASS, ", Place_Num );
    printf("0 /* prio */, 0.0 /* Time */, %d );\n", The_Place->Init_Pop );
  }

  /* generation of dummy, class changing objects, when necessary */
  Nb_Succ = Nb_EQ_Trans_Succ( Reseau, Place_Num );
  if ( Nb_Succ > 1 ) {
    SCAN( L_Scan, The_Place->Outputs ) {
      if ( Reseau->T[ L_Scan->Dest ].Tag == EQ_OBJ ) {
	CODE("\t// dummy interface between place #%d and trans #%d (");
	print_trans_name( &(Reseau->T[ L_Scan->Dest ] ) );
	CODE(")\n");
	CODE2("\tPastec_Server_Normal_FIFO *Dum_P%d_T%d = ", 
	      Place_Num, L_Scan->Dest );
	CODE("new Pastec_Server_Normal_FIFO( 0.0, INFINI );\n" );
	CODE2("\tDum_P%d_T%d", Place_Num, L_Scan->Dest );
	printf("->set_name(\"Dum_P%d_T%d\");\n", Place_Num, L_Scan->Dest );
	CODE2("\tDum_P%d_T%d", Place_Num, L_Scan->Dest );
	printf("\t->set_gss_name(\"%s\");\n", NULL_STRING );
	CODE2("\tDum_P%d_T%d", Place_Num, L_Scan->Dest );
	printf("->set_service( Zero, 1 );\n", Place_Num);
      }
    }
  }
}

void gen_prosit_trans( Reseau, Trans_Num )
PetriNet	*Reseau;
int		Trans_Num;
{
  Tr_Des	*The_Trans;
  int		Degree;
  int		The_Dest;
  boolean	Is_Immediate;

  The_Trans = &(Reseau->T[Trans_Num]);

  Is_Immediate = is_null_time( The_Trans->Tr_Law );

  printf("\n\t// code for transition #%d (", Trans_Num );
  print_trans_name( The_Trans );
  printf(")\n");

  if ( Real_Indegree[Trans_Num] == 0 ) {
    /* this is a source, recycled or not */
    printf("\tPastec_Source *Src_%d = new Pastec_Source(\n\t\t0.0,\n", 
	   Trans_Num );
    printf("\t\t");
    print_prosit_law( The_Trans->Tr_Law );
    printf(",\n\t\t Nb_Gen );\n" );
    printf("\tSrc_%d->set_name(\"%s\");\n", Trans_Num, The_Trans->Name );
    printf("\tSrc_%d->set_gss_name(\"%s\");\n", Trans_Num, The_Trans->Data );
    printf("\tSrc_%d->customers_priority( 0 );\n", Trans_Num );
    /* If the source has one single (real) output which is EQ, then	*/
    /* the class has to be spacially set. If not, the standard setting	*/
    /* is ok.								*/
    if ( Real_Outdegree[ Trans_Num ] == 1 ) {
      The_Dest = Reseau->T[Trans_Num].Outputs->Dest;
      if ( is_recycling_place( &(Reseau->P[The_Dest]) ) )
	The_Dest = Reseau->T[Trans_Num].Outputs->Next->Dest;

      if ( Reseau->P[The_Dest].Tag == EQ_OBJ )
	CODE3("\tSrc_%d->customers_class( CL_T%d_P%d );\n", 
	      Trans_Num, Trans_Num, The_Dest );
      else
	CODE2("\tSrc_%d->customers_class( CL_%d );\n", 
	      Trans_Num, Trans_Num );
    }
    else
      CODE2("\tSrc_%d->customers_class( CL_%d );\n", Trans_Num, Trans_Num );

    CODELN;
  }
  else { 
    /* if serial and immediate, create a routing server */
    if ( Is_Immediate
	 && ( Real_Indegree[Trans_Num] == 1 )
	 && ( Real_Outdegree[Trans_Num] <= 1 ) ) {
      CODE1("\tPastec_Routing *Trans_%d = ", Trans_Num );
      CODE("new Pastec_Routing( 0.0 );\n");
      CODE2("\tTrans_%d->set_name(\"%s\");\n", 
	    Trans_Num, The_Trans->Name );
      CODE2("\tTrans_%d->set_gss_name(\"%s\");\n", 
	    Trans_Num, The_Trans->Data );
    }
    else {
      if ( Real_Indegree[Trans_Num] > 1 ) {
	/* on doit creer une fusion */
	printf("\tPastec_FS_Station *Fus_%d = new Pastec_FS_Station;\n", 
	       Trans_Num );
	printf("\tFus_%d->set_name(\"Fus_%d\");\n", Trans_Num, Trans_Num );
	printf("\tFus_%d->set_gss_name(\"%s\");\n", 
	       Trans_Num, The_Trans->Data );      }

      /* on cree ensuite une station simple */
      Degree = server_degree( Reseau, The_Trans );
      if ( Is_Interface_Tr[Trans_Num] )
	printf("\tTrans_%d = ", Trans_Num );
      else
	printf("\tPastec_Server_Normal_FIFO *Trans_%d = ", Trans_Num );
      printf("new Pastec_Server_Normal_FIFO( 0.0, ");
      if ( Degree == -1 )
	printf("INFINI );\n");
      else if ( Degree == 0 )
	fprintf(stderr, "%s: warning: transition %d (%s) is not live.\n",
		TOOLNAME, Trans_Num, The_Trans->Name );
      else
	printf("%d );\n", Degree );
      printf("\tTrans_%d->set_name(\"%s\");\n", 
	     Trans_Num, The_Trans->Name );
      printf("\tTrans_%d->set_gss_name(\"%s\");\n", 
	     Trans_Num, The_Trans->Data );
      printf("\tTrans_%d->set_service(\n", Trans_Num);
      printf("\t\t");
      print_prosit_law( The_Trans->Tr_Law );
      printf(",\n\t\t1 );\n");
    }

    if ( Real_Outdegree[Trans_Num] > 1 ) {
      /* on doit creer une fission */
      printf("\tPastec_FS_Station *Fis_%d = new Pastec_FS_Station;\n", 
	     Trans_Num );
      printf("\tFis_%d->set_name(\"Fis_%d\");\n", Trans_Num, Trans_Num );
      printf("\tFis_%d->set_gss_name(\"%s\");\n", Trans_Num, The_Trans->Data );
    }
    else if ( ( Real_Outdegree[Trans_Num] == 1 )
	      && ( Real_Indegree[Trans_Num] != 0 ) ) {
      /* Because there is no fusion, it is necessary to set the class	*/
      /* of customers here. Unless this is a source, which sets its class	*/
      /* when declared. If the output object is EQ, the ad-hoc class is 	*/
      /* used.								*/
      The_Dest = The_Trans->Outputs->Dest;
      if ( is_recycling_place( &(Reseau->P[The_Dest]) ) )
	The_Dest = The_Trans->Outputs->Next->Dest;

      if ( Reseau->P[The_Dest].Tag == EQ_OBJ )
	CODE3("\tTrans_%d->set_output_class( CL_T%d_P%d );\n", 
	      Trans_Num, Trans_Num, The_Dest );
      else
	CODE2("\tTrans_%d->set_output_class( CL_%d );\n", 
	      Trans_Num, Trans_Num );
    }
  }
}

/************************************************************************/
/* Procedure to write out the routing of a place, assumed to be 	*/
/* "existing", that is, either with one output or with a probabilistic	*/
/* routing.								*/
/************************************************************************/
void gen_prosit_place_routing( Reseau, Place_Num )
PetriNet	*Reseau;
int		Place_Num;
{
  int		i;
  Pl_Des	*The_Place;
  int		The_Dest;
  Liste_Link	L_Scan;
  Routing_Desc	The_Routing;

  The_Place = &(Reseau->P[Place_Num]);

  printf("\n\t// routing for place #%d (", Place_Num );
  print_place_name( The_Place );
  printf(")\n");

  if ( The_Place->Outdegree == 0 ) {
    printf("\tPlace_%d->set_succ(", Place_Num );
    printf("The_Sink, 1.0, \"%s\", \"%s\");\n",
	   The_Place->Data, NULL_STRING );
  }
  if ( The_Place->Outdegree == 1 ) {
    The_Dest = The_Place->Outputs->Dest;
    if ( Reseau->T[The_Dest].Tag == PROSIT_OBJ ) {
      /* it is also a Prosit object ==> standard routing */
      printf("\tPlace_%d->set_succ(", Place_Num );
      print_prosit_place_output_dest( Reseau, The_Place->Outputs );
      printf(", 1.0, \"%s\", \"%s\");\n",
	     The_Place->Data, NULL_STRING );
      printf("\tPlace_%d->set_output_class( CL_%d );\n",
	     Place_Num, Place_Num );
    }
    else {
      /* is is in the EQ part ==> route it to the EQ object */
      printf("\tPlace_%d->set_succ( The_EQ, 1.0, \"%s\", \"%s\");\n",
	     Place_Num, The_Place->Data, NULL_STRING );
      printf("\tPlace_%d->set_output_class( CL_P%d_T%d );\n",
	     Place_Num, Place_Num, The_Dest );
    }
  }
  else {
    The_Routing = The_Place->Routing;
    if ( ( The_Routing == NULL ) 
	 || ( The_Routing->Type == 'C' )
	 || ( The_Routing->Type == 'D' )
	 || ( The_Routing->Type == 'P' ) ) {
      fprintf( stderr, "%s: error: bad routing at place #%d (%s).\n",
	       TOOLNAME, Place_Num, The_Place->Name );
    }
    else {
      i = 0;
      SCAN( L_Scan, The_Place->Outputs ) {
	The_Dest = L_Scan->Dest;
	if ( Reseau->T[The_Dest].Tag == PROSIT_OBJ ) {
	  /* it is also a Prosit object ==> standard routing */
	  printf("\tPlace_%d->set_succ(", Place_Num );
	  print_prosit_place_output_dest( Reseau, L_Scan );
	  printf(", %8.6f,\"%s\",\"%s\");\n",
		 The_Routing->Probas[i],
		 The_Place->Data, NULL_STRING );
	  printf("\tPlace_%d->set_output_class( CL_%d );\n",
		 Place_Num, Place_Num );
	}
	else {
	  /* is is in the EQ part ==> route it to the EQ object, or to  */
	  /* the dummy interface object.				*/
	  if ( Nb_EQ_Trans_Succ( Reseau, Place_Num ) <= 1 ) {
	    printf("\tPlace_%d->set_succ( The_EQ, %8.6f, \"%s\", \"%s\");\n",
		   Place_Num, 
		   The_Routing->Probas[i], 
		   The_Place->Data, NULL_STRING );
	    printf("\tPlace_%d->set_output_class( CL_P%d_T_%d );\n",
		   Place_Num, Place_Num, The_Dest );
	  }
	  else {
	    CODE3("\tPlace_%d->set_succ( Dum_P%d_T%d, ", 
		  Place_Num, Place_Num, The_Dest );
	    CODE3("%8.6f, \"%s\", \"%s\");\n",
		  The_Routing->Probas[i], 
		  The_Place->Data, NULL_STRING );
	    CODE2("\tDum_P%d_T%d", Place_Num, The_Dest );
	    CODE2("->set_succ( The_EQ, 1.0, \"%s\", \"%s\" );\n",
		  NULL_STRING, NULL_STRING );
	    CODE2("\tDum_P%d_T%d", Place_Num, The_Dest );
	    CODE2("->set_output_class( CL_P%d_T%d );\n", 
		  Place_Num, The_Dest );
	  }
	}
	i++;
      }
    }
  }
}

void gen_prosit_trans_routing( Reseau, Trans_Num )
PetriNet	*Reseau;
int		Trans_Num;
{
  Tr_Des	*The_Trans;
  Pl_Des	*The_Place;
  int		The_Dest;
  Liste_Link	L_Scan;

  The_Trans = &(Reseau->T[Trans_Num]);

  printf("\n\t// routing for transition #%d (", Trans_Num );
  print_trans_name( The_Trans );
  printf(")\n");

  if ( Real_Indegree[Trans_Num] > 1 ) {
    /* fusion is necessary */
    printf("\tFus_Rule = new SM_Rule;\n");
    SCAN( L_Scan, The_Trans->Inputs ) {
      The_Place = &(Reseau->P[L_Scan->Dest]);
      if ( !is_recycling_place( The_Place ) ) {
	printf("\tFus_Rule->match( CL_%d, %d );\n", 
	       Class_Number( Reseau, L_Scan), L_Scan->Weight );
      }
    }
    printf("\tFus_Rule->split( DUMMY_CLASS, 1, 0, Trans_%d, \"%s\");\n", 
	   Trans_Num, The_Trans->Data );
    printf("\tFus_%d->set_rule( Fus_Rule );\n", Trans_Num );
    printf("\tFus_%d->set_succ(Trans_%d,1.0,\"%s\",\"%s\");\n",
	   Trans_Num, Trans_Num, The_Trans->Data, NULL_STRING );
  }

  if ( Real_Outdegree[Trans_Num] > 1 ) {
    /* fission is necessary */
    printf("\tTrans_%d->set_succ(Fis_%d,1.0,\"%s\",\"%s\");\n", 
	   Trans_Num, Trans_Num, The_Trans->Data, NULL_STRING );
    if ( Real_Indegree[Trans_Num] == 1 ) {
      /* change of class is necessary. */
      printf("\tTrans_%d->set_output_class( DUMMY_CLASS );\n", Trans_Num );
    }
    printf("\n");
    printf("\tFis_Rule = new SM_Rule;\n");
    SCAN( L_Scan, The_Trans->Outputs ) {
      The_Dest = L_Scan->Dest;
      The_Place = &(Reseau->P[The_Dest]);
      if ( !is_recycling_place( The_Place ) ) {
	if ( The_Place->Tag == EQ_OBJ ) {
	  /* The dest place is in the EQ module. Route with the class	*/
	  /* of type CL_Tx_Py. No dummy places necessary.		*/
	  printf("\tFis_Rule->split( CL_T%d_P%d, %d, 0, ", 
		 Trans_Num, L_Scan->Data, L_Scan->Weight );
	  printf(" The_EQ, \"%s\");\n", The_Trans->Data );
	}
	else {
	  /* it must be a Prosit object. Normal routing */
	  printf("\tFis_Rule->split( CL_%d, %d, 0, ", 
		 Trans_Num, L_Scan->Weight );
	  print_prosit_trans_output_dest( Reseau, L_Scan );
	  printf(", \"%s\");\n", The_Trans->Data );
	}
      }
    }
    printf("\tFis_Rule->match( DUMMY_CLASS, 1 );\n" );
    printf("\tFis_%d->set_rule( Fis_Rule );\n", Trans_Num );
  }
  else if ( Real_Outdegree[Trans_Num] == 1 ) {
    /* simple routing */
    /* find the --real-- output place */
    if ( is_recycling_place( &Reseau->P[The_Trans->Outputs->Dest] ) )
      The_Place = &Reseau->P[The_Trans->Outputs->Next->Dest];
    else
      The_Place = &Reseau->P[The_Trans->Outputs->Dest];

    if ( Real_Indegree[Trans_Num] == 0 )
      printf("\tSrc_%d->set_succ(", Trans_Num );
    else
      printf("\tTrans_%d->set_succ(", Trans_Num );

    if ( The_Place->Tag == EQ_OBJ )
      printf("The_EQ");
    else if ( !is_recycling_place( &Reseau->P[The_Trans->Outputs->Dest] ) )
      print_prosit_trans_output_dest( Reseau, The_Trans->Outputs );
    else
      print_prosit_trans_output_dest( Reseau, The_Trans->Outputs->Next );
    printf(",1.0,\"%s\",\"%s\");\n", NULL_STRING, NULL_STRING );
  }
  else {
    /* sink routing */
    printf("\tTrans_%d->set_succ(The_Sink,1.0,\"FooName\",\"FooName\");\n",
	   Trans_Num );
  }
}

void gen_prosit_stats_trans( Trans_Num )
int		Trans_Num;
{
  if ( Real_Indegree[ Trans_Num ] == 0 )
    printf("\tSrc");
  else
    printf("\tTrans");

  printf("_%d->compute_stat();\n", Trans_Num );
}

void gen_prosit_stats_place( Place_Num )
int		Place_Num;
{
  printf("\tPlace_%d->compute_stat();\n", Place_Num );
}

void gen_prosit_stats_EQ( )
{
  printf("\tThe_EQ->compute_stat();\n" );
}

void gen_prosit_outputs_trans( 
			      PetriNet	*Reseau,
			      int       Trans_Num )
{
  if ( No_Stats ) {
    CODE2("\tcout << \"%c_%d[\" << T_Max << \"] =   \";\n", 
	  ( Is_Source[ Trans_Num ] ? 'R' : 'Z' ), Trans_Num );
    CODE("\tcout << ");
    CODE2("%s_%d->",
	  ( Is_Source[ Trans_Num ] ? "Src" : "Trans" ), Trans_Num );
    CODE1("%s() << \"   Throughput:   \";\n", 
	  ( Is_Source[ Trans_Num ] ? "total_population" 
	    : "get_nb_service" ) );
    CODE3("\tcout << %s_%d->%s()/T_Max;\n", 
	  ( Is_Source[ Trans_Num ] ? "Src" : "Trans" ), 
	  Trans_Num,
	  ( Is_Source[ Trans_Num ] ? "total_population" 
	    : "get_nb_service" ) );
    CODE1("\tcout << \"  (%s)\" << endl;\n", Reseau->T[ Trans_Num ].Name );
  }
  else 
    CODE2("\t%s_%d->report_stat();\n", 
	  ( Real_Indegree[ Trans_Num ] == 0 ? "Src" : "Trans" ),
	  Trans_Num );
}

void gen_prosit_outputs_place( Reseau, Place_Num )
PetriNet	*Reseau;
int		Place_Num;
{
  if ( No_Stats ) {
    CODE1("\tcout << \"Z_%d[\" << T_Max << \"] =   \";\n", Place_Num );
    CODE1("\tcout << Place_%d->get_nb_service();\n", 
	  Place_Num );
    CODE1("\tcout << \"   Throughput:    \" << ", Place_Num );
    CODE1("Place_%d->get_nb_service()/T_Max;\n", Place_Num );
    CODE1("\tcout << \"  (%s)\" << endl;\n", Reseau->P[Place_Num].Name );
  }
  else 
    printf("\tPlace_%d->report_stat();\n", Place_Num );
}

void gen_prosit_outputs_EQ()
{
  if ( No_Stats ) {
    CODE("\tReport_Stat( cout );\n");
  }
  else
    CODE("\tThe_EQ->report_stat();\n" );
}

void gen_prosit_code(
PetriNet        *Reseau,
char		Tmp_Dir[FILENAME_LG],
char		File_Name[FILENAME_LG])
{
  int		i;
  int		j;
  boolean	Some_Places;
  int		Max_Dependency;
  int		Max_Source_Dependency;
  char		Path[2*FILENAME_LG];
  char		*Prosit_Home;
  char		*The_Date;
  boolean	EQ_Is_Declared = FALSE;
  Liste_Link	L_Scan;

  strcpy( Path, Tmp_Dir );
  strcat( Path, File_Name );

  if ( NULL != freopen( Path, "w", stdout ) ) {

    Prosit_Home = getenv( "PROSIT_HOME" );

    The_Date = todays_date();
    printf("/******************************************/\n");
    printf("/* PROSIT code generated by %s\t  */\n", TOOLNAME );
    printf("/* Date: %s\t\t  */\n", The_Date );
    printf("/* Model Name: \t\t\t\t  */\n" );
    printf("/******************************************/\n");
    free( The_Date );

    printf("\n");
    printf("#define INFINI 100\n");
    printf("#define DUMMY_CLASS 0\n");
    /* generating class ids for (possible) use in fusions */
    for ( i=0; i< max( Reseau->Nb_Trans, Reseau->Nb_Places ); i++ )
      printf("#define CL_%d\t%d\n", i, i+1 );
    /* now generating class ids for the interface Prosit->EQ */
    for ( j=0; j<Reseau->Nb_Trans; j++ ) {
      if ( Reseau->T[j].Tag == PROSIT_OBJ ) {
	SCAN( L_Scan, Reseau->T[j].Outputs )
	  if ( Reseau->P[L_Scan->Dest].Tag == EQ_OBJ ) {
	    printf("#define CL_T%d_P%d\t%d\n", 
		   j, L_Scan->Dest, i+1 );
	    i++;
	  }
      }
    }
    for ( j=0; j<Reseau->Nb_Places; j++ ) {
      if ( Reseau->P[j].Tag == PROSIT_OBJ ) {
	SCAN( L_Scan, Reseau->P[j].Outputs )
	  if ( Reseau->T[L_Scan->Dest].Tag == EQ_OBJ ) {
	    printf("#define CL_P%d_T%d\t%d\n", 
		   j, L_Scan->Dest, i+1 );
	    i++;
	  }
      }
    }

    printf("\n");

    /* includes */
    printf("#include <stdio.h>\n");
    printf("#include <iostream.h>\n");
    printf("#include <Uniform.h>\n");
    printf("#include <Geom.h>\n");
    printf("#include <Binomial.h>\n");
    printf("#include <NegExp.h>\n");
    printf("#include <Erlang.h>\n");
    printf("#include \"Cst.hh\"\n");
    printf("#include \"OnOff.hh\"\n");
    printf("#include <MLCG.h>\n");
    printf("#include <bool.h>\n");
    printf("#include <fstream.h>\n");
    printf("#include \"Pastec_Sink.hh\"\n");
    printf("#include \"Pastec_Routing.hh\"\n");
    printf("#include \"Pastec_Server_Normal_FIFO.hh\"\n");
    printf("#include \"Pastec_Server_Normal_LIFO.hh\"\n");
    printf("#include \"Pastec_Server_Normal_PRIORITY.hh\"\n");
    printf("#include \"Pastec_Source.hh\"\n");
    printf("#include \"EQ_Thing.hh\"\n");
    printf("#include \"Kernel.hh\"\n");
    printf("#include \"Pastec_Kernel.hh\"\n");
    printf("#include \"Pastec_Client.hh\"\n");
    printf("#include \"Pastec_FS_Station.hh\"\n");
    printf("#include \"SM_Rule.hh\"\n");  
    printf("\n");

    /* declarations */
    printf("Pastec_Kernel* pnoyau = new Pastec_Kernel;\n");
    printf("Kernel *noyau = pnoyau;\n");
    printf("SM_Rule *Fus_Rule;\n");
    printf("SM_Rule *Fis_Rule;\n");
    printf("\n");

    /* if no stats, there is a direct call to Report_Stat() */
    if ( No_Stats ) {
      CODE("// external declaration from EQ_Module.c\n");
      CODE("extern void Report_Stat( ostream &Out_File );\n");
      CODELN;
    }
      
    /* various variables */
    printf("bool\tTrace;\n\n");
    for ( i=0; i<Reseau->Nb_Trans; i++ )
      if ( Is_Interface_Tr[i] ) {
	if ( is_null_time( Reseau->T[i].Tr_Law )
	     && ( Real_Indegree[i] == 1 )
	     && ( Real_Outdegree[i] <= 1 ) ) {
	  printf("Pastec_Routing *Trans_%d;\n", i );
	}
	else 
	  printf("Pastec_Server_Normal_FIFO *Trans_%d;\n", i );
      }
    for ( i=0; i<Reseau->Nb_Places; i++ )
    if ( Is_Interface_Pl[i] ) {
      if ( is_null_time( Reseau->P[i].Pl_Law ) )
	printf("Pastec_Routing *Place_%d;\n", i );
      else
	printf("Pastec_Server_Normal_FIFO *Place_%d;\n", i );
    }
    CODELN;

    /* code of the main routine */
    printf("main(int argc, char *argv[]) {\n\n");
    CODELN;
    CODE("  int\ti = 1;\n");
    CODE1("  double\tT_Max = %f;\n", 
	  (double)Reseau->Sim_Params->Nb_Firings );
    CODE1("  int   \tNb_Gen = %d;\t",
	  IRINT(Reseau->Sim_Params->Nb_Firings) );
    CODE("/* number of customers at the sources */\n" );
    CODELN;
    CODE("  // analysis of command line arguments\n");

    CODE("  if ( argc > 1 )\n");
    CODE("    do {\n");
    CODE("      if ( !strcmp ( argv[i], \"-trace\" ) ) {\n");
    CODE("        Trace = TRUE;\n");
    CODE("        i++;\n");
    CODE("      }\n");
    CODE("      else if ( !strcmp ( argv[i], \"-tmax\" ) ) {\n");
    CODE("        i++;\n");
    CODE("        if ( 0 == sscanf( argv[i], \"%%lf\", &T_Max ) ) {\n");
    CODE("          cout << \"Bad argument for -tmax\" << endl;\n");
    CODE("          exit(-1);\n");
    CODE("        }\n");
    CODE("        else {\n");
    CODE("          i++;\n");
    CODE("        }\n");
    CODE("      }\n");
    CODE("      else if ( !strcmp ( argv[i], \"-nb_gen\" ) ) {\n");
    CODE("        i++;\n");
    CODE("        if ( 0 == sscanf( argv[i], \"%%d\", &Nb_Gen ) ) {\n");
    CODE("          cout << \"Bad argument for -nb_gen\" << endl;\n");
    CODE("          exit(-1);\n");
    CODE("        }\n");
    CODE("        else {\n");
    CODE("          i++;\n");
    CODE("        }\n");
    CODE("      }\n");
    CODE("      else if ( !strcmp ( argv[i], \"-h\" ) ) {\n");
    CODE("        cout << \"Usage: \" << argv[0] << \" [-h] [-trace]\";\n");
    CODE("        cout << \" [-tmax <real>] [-nb_gen <int>]\\n\";\n");
    CODE("        exit(0);\n");
    CODE("      }\n");
    CODE("      else {\n");
    CODE("        // ignore unknown option\n");
    CODE("        i++;\n");
    CODE("      }\n");
    CODE("    } while ( i < argc );\n");
    CODELN;

    /* generate a Zero time law, which is rarely useless */
    CODE("  Cst *Zero = new Cst( 0.0 );\n");
    CODELN;

    /* generate a sink, if necessary */
    if ( Has_Sinks( Reseau ) ) {
      printf("\t// declaration of the sink\n");
      printf("\tPastec_Sink *The_Sink = new Pastec_Sink(0.0);\n");
      printf("\tThe_Sink->set_name(\"The_Sink\");\n");
      CODELN;
    }
    
    printf("\t// configuration of transition objects\n");
    for ( i=0; i<Reseau->Nb_Trans; i++ ) {
      if ( Reseau->T[i].Tag == PROSIT_OBJ )
	gen_prosit_trans( Reseau, i );
      else if ( !EQ_Is_Declared ) {
	printf("\t// declaration of the EQuational Object \n");
	printf("\tEQ_Thing *The_EQ = new EQ_Thing(0.0);\n");
	printf("\tThe_EQ->set_name(\"The_EQ\");\n\n");
	printf("\tThe_EQ->set_trace_status( Trace );\n");
	EQ_Is_Declared = TRUE;
      }
    }

    Some_Places = FALSE;
    for ( i=0; i<Reseau->Nb_Places; i++ )
      if ( Place_Exists[i] )
	Some_Places = TRUE;

    if ( Some_Places ) {
      printf("\t// configuration of place objects\n");
      for ( i=0; i<Reseau->Nb_Places; i++ ) 
	if ( Reseau->P[i].Tag == PROSIT_OBJ ) {
	  if ( Place_Exists[i] )
	    gen_prosit_place( Reseau, i );
	  else if ( Reseau->P[i].Init_Pop != 0 ) {
	    /* il faut generer le marquage initial ou il faut */
	    print_initial_pop( Reseau, i );
	  }
	}
	else if ( !EQ_Is_Declared ) {
	  printf("\t// declaration of the EQuational Object \n");
	  printf("\tEQ_Thing *The_EQ = new EQ_Thing();\n");
	  printf("\tThe_EQ->set_name(\"The_EQ\");\n\n");
	  printf("\tThe_EQ->set_trace_status( Trace );\n");
	  EQ_Is_Declared = TRUE;
	}
    }
    printf("\n");

    printf("\t// configuration of the routing\n");
    for ( i=0; i<Reseau->Nb_Trans; i++ )
      if ( Reseau->T[i].Tag == PROSIT_OBJ )
	gen_prosit_trans_routing( Reseau, i );
    for ( i=0; i<Reseau->Nb_Places; i++ )
      if ( Place_Exists[i] && ( Reseau->P[i].Tag == PROSIT_OBJ ) )
	gen_prosit_place_routing( Reseau, i );

    printf("\n");

    /* parametrage du modele */
    printf("\t// parameters of the simulation\n");
    printf("\tpnoyau->set_model_name(\"FooModel\");\n");
    printf("\tpnoyau->set_starting_time(0.0);\n");
    printf("\tpnoyau->set_ending_time( T_Max );\n");
    printf("\tpnoyau->set_debug(false);\n");
    printf("\tpnoyau->set_animation(%s);\n",
	   ( Agat ? "true" : "false" ) );
    printf("\tpnoyau->set_statistic(%s);\n",
	   ( No_Stats ? "false" : "true" ) );

    printf("\n");

    /* ordres de calcul des statistiques */
    if ( !No_Stats ) {
      printf("\t// computation of statistics\n");
      for ( i=0; i<Reseau->Nb_Trans; i++ )
	if ( Reseau->T[i].Tag == PROSIT_OBJ ) 
	  gen_prosit_stats_trans( i );
      for ( i=0; i<Reseau->Nb_Places; i++ )
	if ( Place_Exists[i] && ( Reseau->P[i].Tag == PROSIT_OBJ ) )
	  gen_prosit_stats_place( i );

      gen_prosit_stats_EQ();

      printf("\n");
    }

    /* execution de la simulation */
    printf("\t// the simulation itself\n");
    printf("\tcout << \"Start of the simulation\" << endl;\n");
    printf("\tnoyau->print_registered();\n");
    printf("\tnoyau->run();\n");
    printf("\tcout << \"End of the simulation\" << endl;\n");
    printf("\n");

    /* sortie des resultats */
    printf("\t// output of results\n");
    for ( i=0; i<Reseau->Nb_Trans; i++ )
      if ( Reseau->T[i].Tag == PROSIT_OBJ ) 
	gen_prosit_outputs_trans( Reseau, i );
    for ( i=0; i<Reseau->Nb_Places; i++ )
      if ( Place_Exists[i] && ( Reseau->P[i].Tag == PROSIT_OBJ ) )
	gen_prosit_outputs_place( Reseau, i );

    gen_prosit_outputs_EQ();

    printf("\n");
    /* c'est fini! */
    printf("}\n");

  }
  else {
    fprintf( stderr, "%s: Error opening file %s.\n", TOOLNAME, Path );
  }
}

void gen_agat_code( Reseau )
PetriNet	*Reseau;
{
  int	i;

  freopen( "sim.agat", "w", stdout );

  for ( i=0; i<Reseau->Nb_Trans; i++ )
    if ( ( Reseau->T[i].Tag == PROSIT_OBJ )
	 && ( !Is_Source[i] ) ) {
      printf("plotAxeLine(%s__Nombre_de_clients__temps,", Reseau->T[i].Name );
      printf("%s__Nombre_de_clients);\n", Reseau->T[i].Name );
    }
  for ( i=0; i<Reseau->Nb_Places; i++ )
    if ( ( Reseau->P[i].Tag == PROSIT_OBJ )
	 && Place_Exists[i] ) {
      printf("plotAxeLine(%s__Nombre_de_clients__temps,", 
	     Reseau->P[i].Name );
      printf("%s__Nombre_de_clients);\n", Reseau->P[i].Name );
    }
}
/************************************************************************/
/* Fonction	satisfies_constraints( Reseau )				*/
/* verifie si le reseau satisfait aux contraintes necessaires pour etre	*/
/* transcrit en un reseau de files d'attente sous Prosit		*/
/* Les contraintes sont pour le moment:					*/
/* - reseau a routage							*/
/* - pas d'arc inhibiteur						*/
/* - pas de routage 'P' impliquant des transitions Prosit		*/
/*									*/
/* Argument: 	Reseau, de type PetriNet				*/
/* Retourne:	TRUE si le reseau satisfait aux contraintes, FALSE sinon*/
/************************************************************************/
boolean	satisfies_constraints( Reseau, Testing, Bad_Objects )
PetriNet	*Reseau;
boolean		Testing;
int		**Bad_Objects;
{
  boolean	Is_Ok = TRUE;
  Pl_Des	*The_Place;
  Tr_Des	*The_Trans;
  Liste_Link	L_Scan;
  int		i;

  if ( Testing ) {
    (*Bad_Objects) = (int*) calloc( (unsigned)(1 + 6*Reseau->Nb_Places +
					    4*Reseau->Nb_Trans), 
				 sizeof( int ) );
    (*Bad_Objects)[0] = 0;
  }
  else 
    (*Bad_Objects) = (int*) NULL;

  if ( Reseau->Has_Inhibitors ) {
    fprintf( stderr, "%s: the Petri network has inhibitor arcs. ", 
	     TOOLNAME );
    fprintf( stderr, "No translation.\n" );
    return FALSE;
  }
  else {
    Is_Ok = Has_Integral_Transitions( Reseau, Testing, TRUE, *Bad_Objects )
      && Has_All_Deterministic_Places( Reseau, Testing, *Bad_Objects )
      && Has_Simulable_Places( Reseau, Testing, *Bad_Objects );

    /* checking that 'P' routing involves only EQ objects		*/
    for (i=0; Is_Ok && (i<Reseau->Nb_Places); i++ )
      if ( Reseau->P[i].Routing != NULL ) {
	if ( Reseau->P[i].Routing->Type == 'P' ) {
	  /* It has already been checked if all output transitions are	*/
	  /* instantaneous. Check here all are EQ			*/
	  Is_Ok = ( Reseau->P[i].Tag == EQ_OBJ );
	  SCAN( L_Scan, Reseau->P[i].Outputs ) {
	    Is_Ok = Is_Ok && ( Reseau->T[L_Scan->Dest].Tag == EQ_OBJ );
	  }
	}
      }

    if ( !Is_Ok && Testing ) {
      /* Store the marker for places: 1, and the place */
      (*Bad_Objects)[0]++;
      (*Bad_Objects)[2*(*Bad_Objects)[0]-1] = 1;
      (*Bad_Objects)[2*(*Bad_Objects)[0]] = i;
    }

    return Is_Ok;
  }

}
