/* Translation program from pn language to C++ PROSIT language.	  	*/

#include <stdlib.h>
#include <string.h>

#include "pn.h"

#define	TOOLNAME	"pn2prosit"

#define FILENAME_LG	256

#define PROSIT_DIR	"/u/chin/2/sloop3/gsiegel/Prosit_28_Jul"
#define SOURCE_CODE	"/u/chin/2/sloop3/gsiegel/Prosit_28_Jul/Src/Queueing/Sequential/Routing_Source.cc"
#define NULL_STRING	"<rien>"

#define max(x,y)        ( (x>y) ? x : y )
#define min(x,y)        ( (x<y) ? x : y )

#define SCAN(x,y)       for(x=y;x!=NULL;x=x->Next)

#define CODE(x)		printf(x)
#define CODE1(x,y)	printf(x,y)
#define CODE2(x,y,z)	printf(x,y,z)
#define CODE3(x,y,z,t)	printf(x,y,z,t)
#define CODELN		CODE("\n")

/* Externs from the module parse_pn.o */
extern PetriNet	*Read_Network();
extern void	dumpe_network();
extern char	*todays_date();

/* forward declarations */
boolean	satisfies_constraints();
void	gen_prosit_code();
void	gen_prosit_makefile();
void	gen_agat_code();

boolean		*Place_Exists;
int		*Trans_Loop_Of;	/* stores the id of the trans a place	*/
				/* recycles (if it does).		*/
boolean		Optimize = FALSE;
int		Opt_Level;
double   	Trace_Start = 0.0;      /* trace control parameters     */
double   	Trace_End = 100.0;      /* with their initial values    */
boolean		Agat = FALSE;
boolean		Trace = FALSE;
boolean		No_Stats = FALSE;
char		Trace_File_Name[FILENAME_LG] = "trace.results";
int		*Real_Indegree;
int		*Real_Outdegree;

boolean	is_serial_place( The_Place )
Pl_Des	*The_Place;
{
  return ( (The_Place->Indegree == 1 ) 
	   && (The_Place->Outdegree == 1 ) );
}

/* test if a law is CST(0.0)						*/
boolean	is_null_time( The_Law )
Law_Desc	The_Law;
{
  return ( The_Law->Name == 'C' ) 
    && ( The_Law->Parameters[0] == 0.0 );
}

/* test si une transition est recyclee...				*/
boolean	is_recycled( Reseau, The_Trans )
PetriNet        *Reseau;
Tr_Des		*The_Trans;
{
  Liste_Link	L_Scan;
  boolean	Is_Recycled = FALSE;
  Pl_Des	*The_Place;

  SCAN( L_Scan, The_Trans->Outputs ) {
    The_Place = &(Reseau->P[L_Scan->Dest]);
    Is_Recycled = Is_Recycled ||
      ( ( The_Place->Outdegree == 1 )
	&& ( The_Place->Outputs->Dest == The_Trans->Tr_Id )
	&& ( is_null_time( The_Place->Pl_Law ) ) );
  }

  return Is_Recycled;
}

/* test si une place est une place de recyclage.			*/
boolean	is_recycling_place( The_Place )
Pl_Des		*The_Place;
{
  boolean	Is_Recycling = FALSE;

  Is_Recycling = ( ( The_Place->Indegree == 1 ) 
		   && ( The_Place->Outdegree == 1 )
		   && ( The_Place->Inputs->Dest 
			== The_Place->Outputs->Dest )
		   && ( The_Place->Inputs->Weight == 1 )
		   && ( The_Place->Outputs->Weight == 1 )
		   && is_null_time( The_Place->Pl_Law ) );

  return Is_Recycling;
}

int server_degree( Reseau, The_Trans )
PetriNet	*Reseau;
Tr_Des		*The_Trans;
{
  Liste_Link	L_Scan;
  Pl_Des	*The_Place;
  int		Degree = -1;		/* convention for "infinite" */

  SCAN( L_Scan, The_Trans->Outputs ) {
    The_Place = &(Reseau->P[L_Scan->Dest]);
    if ( ( The_Place->Outdegree == 1 )
	 && ( The_Place->Outputs->Dest == The_Trans->Tr_Id )
	 && ( is_null_time( The_Place->Pl_Law ) ) ) {
      /* in case there are several recyclings, take the min degree */
      if ( Degree == -1 )
	Degree = The_Place->Init_Pop;
      else if ( Degree > The_Place->Init_Pop )
	Degree = The_Place->Init_Pop;
    }
  }
		 
  return Degree;
}

/* A sink is a transition without outputs.				*/
/* The net has a sink if there is a place without outputs (terminal)	*/
boolean Has_Sinks( Reseau )
PetriNet	*Reseau;
{
  int		i;
  boolean	Has_Sinks = FALSE;

  for ( i=0; !Has_Sinks && (i<Reseau->Nb_Trans); i++ )
    Has_Sinks = ( Real_Outdegree[i] == 0 );
  for ( i=0; !Has_Sinks && (i<Reseau->Nb_Places); i++ )
    Has_Sinks = ( Reseau->P[i].Outdegree == 0 );

  return Has_Sinks;
}

boolean Is_Source( Reseau, Trans_Num )
PetriNet	*Reseau;
int		Trans_Num;
{
  boolean	Is_Src = FALSE;
  int		Pl_Num;

  if ( Reseau->T[Trans_Num].Indegree == 0 )
    /* I assume this is meant to be a source... */
    Is_Src = TRUE;
  else if ( Reseau->T[Trans_Num].Indegree == 1 ) {
    Pl_Num = Reseau->T[Trans_Num].Inputs->Dest;
    if ( ( Reseau->P[Pl_Num].Indegree == 1 ) 
	 && ( Reseau->P[Pl_Num].Inputs->Dest == Trans_Num ) )
      Is_Src = TRUE;
  }

  return Is_Src;
}

/* procedure pour initialiser le tableau Place_Exists, utile a 		*/
/* l'optimisation mais aussi elimine les places de recyclage.		*/
void	init_place_status( Reseau )
PetriNet	*Reseau;
{
  int		i;
  Pl_Des	*The_Place;

  /* First, determine if places "exist".				*/
  /* A place exists if is is not a recycling place, and, depending on	*/
  /* optimization, if it is not serial with a zero holding time.	*/
  /* Terminal places always exist.					*/
  Place_Exists = (boolean*) calloc( (unsigned)Reseau->Nb_Places, 
				    sizeof(int) );
  Trans_Loop_Of = (int*) calloc( (unsigned)Reseau->Nb_Places, 
				 sizeof(int) );

  for ( i=0; i<Reseau->Nb_Places; i++ ) {
    The_Place = &(Reseau->P[i]);
    if ( is_recycling_place( The_Place ) && !Trace ) {
      Place_Exists[i] = FALSE;
      Trans_Loop_Of[i] = The_Place->Inputs->Dest;
      if (Debug) {
	printf("Recycling place Pl %d ", i );
	printf("(of Tr %d) removed\n", Trans_Loop_Of[i] );
      }
    }
    else if ( Optimize ) {
      Place_Exists[i] = 
	! ( ( The_Place->Outdegree == 1 )
	    && ( The_Place->Indegree == 1 )
	    && ( The_Place->Pl_Law->Name == 'C' )
	    && ( The_Place->Pl_Law->Parameters[0] == 0.0 )
	    );
      if (Debug && !Place_Exists[i])
	printf("Pl %d removed\n", i );
    }
    else
      Place_Exists[i] = TRUE;
  }

  /* Next, compute real degrees of transitions */
  Real_Indegree = (int*) calloc( (unsigned)Reseau->Nb_Trans, sizeof(int) );
  Real_Outdegree = (int*) calloc( (unsigned)Reseau->Nb_Trans, sizeof(int) );
  for ( i=0; i<Reseau->Nb_Trans; i++ ) {
    Real_Indegree[i] = Reseau->T[i].Indegree;
    Real_Outdegree[i] = Reseau->T[i].Outdegree;
    if ( is_recycled( Reseau, &(Reseau->T[i]) ) ) {
      Real_Indegree[i]--;
      Real_Outdegree[i]--;
    }
  }

}

main(argc, argv)
int	argc;
char	*argv[];
{
  static char		File_Name[FILENAME_LG] = "";
  static char		Code_File_Name[FILENAME_LG] = "main.cc";
  PetriNet 		*Reseau;
  int			i;
  int			Max_Time = -1;
  boolean		Help = FALSE;
  boolean		Args_Ok = TRUE;
  boolean		File_Ok = TRUE;

  i = 1;

  if (argc>1)
    do {

      if GET_STRING( "-f", File_Name )
      else if GET_STRING( "-o", Code_File_Name )

      else if SWITCH( "-debug", Debug, TRUE )
      else if SWITCH( "-agat", Agat, TRUE )
      else if SWITCH( "-nostats", No_Stats, TRUE )

      else if GET_INT( "-tmax", Max_Time )
      else if GET_INT_SWITCH( "-O", Opt_Level, Optimize, TRUE )

      else if IS_VERSION

	/* trace stuff removed as long as not fixed.			*/
	/*      else if ( 0 == strcmp( "-tb", argv[i] ) ) {
	i++;
	if ( ( i < argc ) 
	     && ( 0 != sscanf( argv[i], "%lf", &Trace_Start ) ) ) {
	  i++;
	  Trace = TRUE;
	  Optimize = FALSE;
	}
	else {  
	  printf("%s: double expected after -tb\n",argv[0]);
	  Args_Ok = FALSE;
	}
      }
      else if ( 0 == strcmp( "-te", argv[i] ) ) {
	i++;
	if ( ( i < argc ) 
	     && ( 0 != sscanf( argv[i], "%lf", &Trace_End ) ) ) {
	  i++;
	  Trace = TRUE;
	  Optimize = FALSE;
	}
	else {  
	  printf("%s: double expected after -tb\n",argv[0]);
	  Args_Ok = FALSE;
	}
      }
      */

      else if ( ( 0 == strcmp( "-h", argv[i] ) )
		|| ( 0 == strcmp( "-?", argv[i] ) ) ) {
	printf("Usage: %s [-f <file>] ", TOOLNAME);
	printf("[-o <file>] [-agat] [-O <int>] [-tmax <int>]\n\t");
	printf("[-nostats] ");
	/* trace stuff removed as long as not fixed.			*/
	/* 	printf("[-tb <time>] [-te <time>] [-tfile <file> ] "); */
	printf("[-debug ] [-h] [?]\n");
	Args_Ok = FALSE;
	Help = TRUE;
	i++;
      }
      else
	i++; /* skip unknown argument */
    }
    while ( Args_Ok && ( i < argc ) );
    
  if (Args_Ok) {

    /* Agat implies Stats */
    if ( Agat && No_Stats ) {
      printf("%s: error: -agat and -nostats are incompatible. Choose one.\n",
	     TOOLNAME );
      return -1;
    }
      
    Reseau = Read_Network( File_Name );

    if ( Reseau != NULL ) {

	if ( satisfies_constraints( Reseau ) ) {

	  if ( Max_Time >= 0 )
	    Reseau->Sim_Params->Nb_Firings = Max_Time;

	  /* done before redirecting output, so that debug is visible */
	  init_place_status(Reseau);

	  if ( strlen(Code_File_Name) != 0 )
	    File_Ok = ( NULL != freopen( Code_File_Name, "w", stdout ) );

	  if ( File_Ok ) {
	    gen_prosit_code(Reseau);
	    gen_prosit_makefile(Code_File_Name);
	    if ( Agat )
	      gen_agat_code( Reseau );

	    return 0;
	  }
	  else {
	    fprintf( stderr, "%s: Error opening file %s.\n", 
		     TOOLNAME, Code_File_Name );
	    return -2;
	  }
	}
	else {
	  fprintf( stderr, "%s: the network does not ", TOOLNAME );	
	  fprintf( stderr, "satisfy the constraints. No translation.\n" );
	  return -3;
	}
      }
      else {
	fprintf( stderr, "%s: Error reading network. No translation.\n", 
		 TOOLNAME );	
	return -3;
      }
  }
  else if (!Help)
    return -1;

  /* if not exited otherwise, it's ok */
  return 0;
}

void print_name( The_Trans )
Tr_Des		*The_Trans;
{
  if ( ( strlen( The_Trans->Name ) == 0 )
       || !strcmp( The_Trans->Name, " " ) ) {
    if ( strlen( The_Trans->Data ) != 0 )
      printf("%s", The_Trans->Data );
    else 
      printf("<no name>");
  }
  else
    printf("%s", The_Trans->Name );
}

void print_place_name( The_Place )
Pl_Des		*The_Place;
{
  if ( ( strlen( The_Place->Name ) == 0 )
       || !strcmp( The_Place->Name, " " ) ) {
    if ( strlen( The_Place->Data ) != 0 )
      printf("%s", The_Place->Data );
    else 
      printf("<no name>");
  }
  else
    printf("%s", The_Place->Name );
}

/* procedure to write the name of the actual destination of customers	*/
/* going out of some --place-- through link The_Link.			*/
void print_prosit_place_output_dest( Reseau, The_Link )
PetriNet	*Reseau;
Liste_Link	The_Link;
{
  int		The_Dest;

  The_Dest = The_Link->Dest;

  if ( Real_Indegree[The_Dest] == 1 )
    /* there is no fusion at this trans */
    printf("Trans_%d", The_Dest );
  else 
    printf("Fus_%d", The_Dest );
}

/* procedure to write the name of the actual destination of customers	*/
/* going out of some --transition-- through link The_Link.		*/
/* This takes into account the possible disapearance of places due to	*/
/* optimisation, and the possible appearance of fusion objects.		*/
void print_prosit_trans_output_dest( Reseau, The_Link )
PetriNet	*Reseau;
Liste_Link	The_Link;
{
  Pl_Des	*The_Place;
  int		The_Dest;

  The_Dest = The_Link->Dest;
  The_Place = &(Reseau->P[The_Dest]);

  if ( Place_Exists[ The_Dest ] ) {
    printf("Place_%d", The_Dest );
  }
  else {
    /* normally, there is only one output to this place */
    print_prosit_place_output_dest( Reseau, The_Place->Outputs );
  }
}

/* Function to compute the class number of customers arriving to a	*/
/* transition through link The_Link.					*/
int Class_Number( Reseau, The_Link )
PetriNet	*Reseau;
Liste_Link	The_Link;
{
  Pl_Des	*The_Place;
  int		The_Orig;

  The_Orig = The_Link->Dest;
  The_Place = &(Reseau->P[The_Orig]);  

  if ( Place_Exists[ The_Orig ] )
    return The_Orig;
  else {
    /* this place should be input serial, otherwise: confusion... */
    if ( The_Place->Indegree == 1 ) 
      return The_Place->Inputs->Dest;
    else {
      fprintf( stderr, "%s: error: impossible to compute input class",
	       TOOLNAME );
      fprintf( stderr, "at place #%d (%s)\n", The_Orig, The_Place->Name );
      return 0;
    }
  }
}

/* procedure to write the initial population in the case the place	*/
/* does not exist. Requires first the computation of the object which 	*/
/* is output of this place.						*/
void print_initial_pop( Reseau, Place_Num )
PetriNet	*Reseau;
int		Place_Num;
{
  Pl_Des	*The_Place;
  int		The_Dest;

  The_Dest = Reseau->P[Place_Num].Outputs->Dest;
  The_Place = &(Reseau->P[Place_Num]);

  if ( !is_recycling_place( The_Place ) ) {
    /* if everything is consistent, the place has a single output */
    printf("\n\t// code for (nonexistent) place #%d (", Place_Num );
    print_place_name( The_Place );
    printf(")\n");

    printf("\tTrans_%d->insert_customer( DUMMY_CLASS, ", The_Dest );
    printf("0 /* prio */, 0.0 /* Time */, %d );\n", The_Place->Init_Pop );
  }
}

/* Procedure to write the law. Adapted from the same from gen_pn.c */
int		Seed1 = 308;
int		Seed2 = 725;

void print_prosit_law(The_Law)
Law_Desc	The_Law;
{

  switch( The_Law->Name ) {
  case 'E':
    printf("new NegativeExpntl( %8.4f, new MLCG(%d,%d) )", 
	   The_Law->Parameters[0], ++Seed1, ++Seed2 );
    break;
  case 'C':
    if ( The_Law->Parameters[0] == 0.0 )
      printf("Zero");
    else
      printf("new Cst( %8.4f )", The_Law->Parameters[0] );
    break;
  case 'B':
    /* Bernoulli(p) = Binomial( 1, p ) */
    printf("new Binomial( 1, %8.4f, new MLCG(%d,%d) )", 
 	   The_Law->Parameters[0], ++Seed1, ++Seed2 );
     break;
   case 'G':
     /* g++ Geometric is called with the parameter, called "mean" */
     printf("new Geometric( %8.4f, new MLCG(%d,%d) )",
 	   The_Law->Parameters[0], ++Seed1, ++Seed2 );
     break;
  case 'U':
    printf("new Uniform( %8.4f, %8.4f, new MLCG(%d,%d) )", 
	   The_Law->Parameters[0], The_Law->Parameters[1],
	   ++Seed1, ++Seed2 );
    break;
  case 'R':
    /* Erlang is called with mean and variance */
    printf("new Erlang( %8.4f, %8.4f, new MLCG(%d,%d) )", 
	   The_Law->Parameters[1],
	   The_Law->Parameters[1] * The_Law->Parameters[1] 
	   * The_Law->Parameters[0],
	   ++Seed1, ++Seed2 );
    break;
  case 'D':
    fprintf( stderr, "%s: warning: unimplemented firing law (%c).", 
	    TOOLNAME, The_Law->Name );
    fprintf( stderr, "(constant assumed)\n");
    printf("new Uniform( %8.4f, %8.4f, new MLCG(%d,%d) )", 
	   The_Law->Parameters[1], The_Law->Parameters[1],
	   ++Seed1, ++Seed2 );
    break;
  case 'O':
    printf("new OnOff( %8.4f, %8.4f, %8.4f, new MLCG(%d,%d) )", 
	   The_Law->Parameters[0], 
	   The_Law->Parameters[1], 
	   The_Law->Parameters[2],
	   ++Seed1, ++Seed2 );
    break;
  case 'H':
  case 'g':
    fprintf( stderr, "%s: error: can't handle firing law (%c)\n", 
	    TOOLNAME, The_Law->Name );
    break;
  default:
    fprintf( stderr, "%s: error: unknown firing law (%c)\n", 
	    TOOLNAME, The_Law->Name );
  }
}


void gen_prosit_makefile( Code_File_Name )
char	Code_File_Name[FILENAME_LG];
{
  freopen( "Makefile.Prosit", "w", stdout );

  printf("include %s/Makefile.head\n", PROSIT_DIR );
  printf("\n");
  printf("PROGRAM = sim\n");
  printf("ARCH=SUN4\n");
  printf("OPTIONS+=-D_SUN4\n");
  printf("\n");
  printf("CC            = g++\n");
  printf("LINKER        = $(CC)\n");
  printf(" \n");
  printf("CFLAGS =\n");
  printf("CFLAGS += -I$(RUNTIME_SEQUENTIAL_INCLUDE_DIR)\n");
  printf("CFLAGS += -I$(QUEUEING_SEQUENTIAL_INCLUDE_DIR)\n");
  printf("CFLAGS += -I$(EXAMPLES_SEQUENTIAL_INCLUDE_DIR) \n");
  printf("CFLAGS += -I$(THREAD_INCLUDE_DIR)/$(ARCH) \n");
  printf("CFLAGS += -I$(STAT_TOOL_LOCAL_CPP_INCLUDE_DIR)\n");
  printf("CFLAGS += -I$(STAT_TOOL_LOCAL_C_INCLUDE_DIR)  \n");
  /*   printf("CFLAGS += -I$(STAT_TOOL_DIS_CPP_INCLUDE_DIR)\n"); */
  /*   printf("CFLAGS += -I$(STAT_TOOL_DIS_C_INCLUDE_DIR)    \n"); */
  printf("\n");
  printf("EXTLIBS       =\n");
  printf("EXTLIBS       += -l$(EXAMPLES_SEQUENTIAL_LIBRARY)\n");
  printf("EXTLIBS       += -l$(QUEUEING_SEQUENTIAL_LIBRARY)\n");
  printf("EXTLIBS       += -l$(RUNTIME_SEQUENTIAL_LIBRARY)\n");
  printf("EXTLIBS       += -l$(THREAD_LIBRARY) \n");
  /*   printf("EXTLIBS       += -l$(STAT_TOOL_LOCAL_C_LIBRARY)  \n"); */
  printf("EXTLIBS       += -l$(STAT_TOOL_LOCAL_CPP_LIBRARY)\n");
  printf("EXTLIBS       += -l$(AGAT_LIBRARY)\n");
  printf("\n");
  printf("EXTLIBS       += -l$(TRACE_TOOL_KER_LIBRARY)\n");
  printf("EXTLIBS       += -l$(TRACE_TOOL_STD_LIBRARY) \n");
  printf("\n");
  printf("EXTLIBS       +=  -L$(TRACE_TOOL_LIB_DIR)/$(ARCH)\n");
  printf("EXTLIBS       += -L$(LIB_DIR)/$(ARCH)\n");
  printf("EXTLIBS       += -L$(STAT_TOOL_LIB_DIR)/$(ARCH)\n");
  printf("EXTLIBS       += -L$(AGAT_LIB_DIR)/$(ARCH)\n");
  printf("\n");
  printf("LIBS          = -lm -lg++\n");
  printf("MAIN  = main.o\n");
  printf("\n");
  printf("EXL = \n");
  printf("\n");
  printf("\n");
  printf("$(PROGRAM):   $(MAIN)  $(EXL)\n");
  printf("\t$(LINKER) -o $(PROGRAM) $(MAIN) $(LDFLAGS) $(EXL) $(LIBS) $(EXTLIBS)\n");
  printf("\n");
  printf("main.o:             %s \n", Code_File_Name );
  printf("\t$(CC) $(OPTIONS) $(CFLAGS) -c %s -o main.o\n", Code_File_Name );
  printf("\n");
  printf("clean:\n");
  printf("\t/bin/rm  -f $(EXL)  $(MAIN)\n");
  printf("\n");
}

void gen_prosit_place( Reseau, Place_Num )
PetriNet	*Reseau;
int	        Place_Num;
{
  Pl_Des	*The_Place;
  boolean	Is_Immediate;
  boolean	Is_Terminal;

  The_Place = &(Reseau->P[Place_Num]);

  Is_Terminal = ( Reseau->P[Place_Num].Outdegree == 0 );
  Is_Immediate = is_null_time( The_Place->Pl_Law );

  printf("\n\t// code for place #%d (", Place_Num );
  print_place_name( The_Place );
  printf(")\n");
  if ( Is_Immediate && !Is_Terminal ) {
    CODE1("\tPastec_Routing *Place_%d = ", Place_Num );
    CODE("new Pastec_Routing( 0.0 );\n");
  }
  else {
    printf("\tPastec_Server_Normal_FIFO *Place_%d = ", Place_Num );
    printf("new Pastec_Server_Normal_FIFO( 0.0, INFINI );\n" );
    printf("\tPlace_%d->set_service(\n", Place_Num);
    printf("\t\t");
    if ( Is_Terminal ) {
      printf("new Cst( T_Max )\n");
      fprintf( stderr, "%s: warning: terminal place encountered at ",
	       TOOLNAME );
      fprintf( stderr, "place #%d (%s)\n", 
	       Place_Num, Reseau->P[Place_Num].Name );
    }
    else			 
      print_prosit_law( The_Place->Pl_Law );
    printf(",\n\t\t1 );\n");
  }
  CODE2("\tPlace_%d->set_name(\"%s\");\n", Place_Num, The_Place->Name );
  CODE2("\tPlace_%d->set_gss_name(\"%s\");\n", Place_Num, The_Place->Data );
  if ( The_Place->Init_Pop > 0 ) {
    printf("\tPlace_%d->insert_customer( DUMMY_CLASS, ", Place_Num );
    printf("0 /* prio */, 0.0 /* Time */, %d );\n", The_Place->Init_Pop );
  }
}

void gen_prosit_transition( Reseau, Trans_Num )
PetriNet	*Reseau;
int		Trans_Num;
{
  Tr_Des	*The_Trans;
  int		Degree;
  boolean	Is_Immediate;

  The_Trans = &(Reseau->T[Trans_Num]);

  Is_Immediate = is_null_time( The_Trans->Tr_Law );

  printf("\n\t// code for transition #%d (", Trans_Num );
  print_name( The_Trans );
  printf(")\n");

  if ( Real_Indegree[Trans_Num] == 0 ) {
    /* this is a source, recycled or not */
    printf("\tPastec_Source *Src_%d = new Pastec_Source(\n\t\t0.0,\n", 
	   Trans_Num);
    printf("\t\t");
    print_prosit_law( The_Trans->Tr_Law );
    printf(",\n\t\t%d );\n", Reseau->Sim_Params->Nb_Firings );
    printf("\tSrc_%d->set_name(\"%s\");\n", Trans_Num, The_Trans->Name );
    printf("\tSrc_%d->set_gss_name(\"%s\");\n", Trans_Num, The_Trans->Data );
    printf("\tSrc_%d->customers_priority( 0 );\n", Trans_Num );
    printf("\tSrc_%d->customers_class( CL_%d );\n", Trans_Num, Trans_Num );
    printf("\n");
  }
  else { 
    /* if serial and immediate, create a routing server */
    if ( Is_Immediate
	 && ( Real_Indegree[Trans_Num] == 1 )
	 && ( Real_Outdegree[Trans_Num] <= 1 ) ) {
      CODE1("\tPastec_Routing *Trans_%d = ", Trans_Num );
      CODE("new Pastec_Routing( 0.0 );\n");
      CODE2("\tTrans_%d->set_name(\"%s\");\n", 
	    Trans_Num, The_Trans->Name );
      CODE2("\tTrans_%d->set_gss_name(\"%s\");\n", 
	    Trans_Num, The_Trans->Data );
    }
    else {
      if ( Real_Indegree[Trans_Num] > 1 ) {
	/* on doit creer une fusion */
	printf("\tPastec_FS_Station *Fus_%d = new Pastec_FS_Station;\n", 
	       Trans_Num );
	printf("\tFus_%d->set_name(\"Fus_%d\");\n", Trans_Num, Trans_Num );
	printf("\tFus_%d->set_gss_name(\"%s\");\n", 
	       Trans_Num, The_Trans->Data );
      }

      /* on cree ensuite une station simple */
      Degree = server_degree( Reseau, The_Trans );
      printf("\tPastec_Server_Normal_FIFO *Trans_%d = ", Trans_Num );
      printf("new Pastec_Server_Normal_FIFO( 0.0, ");
      if ( Degree == -1 )
	printf("INFINI );\n");
      else if ( Degree == 0 )
	fprintf(stderr, "%s: warning: transition %d (%s) is not live.\n",
		TOOLNAME, Trans_Num, The_Trans->Name );
      else
	printf("%d );\n", Degree );
      printf("\tTrans_%d->set_name(\"%s\");\n", Trans_Num, The_Trans->Name );
      printf("\tTrans_%d->set_gss_name(\"%s\");\n", Trans_Num, The_Trans->Data );
      printf("\tTrans_%d->set_service(\n", Trans_Num);
      printf("\t\t");
      print_prosit_law( The_Trans->Tr_Law );
      printf(",\n\t\t1 );\n");
    }
  }

  if ( Real_Outdegree[Trans_Num] > 1 ) {
    /* on doit creer une fission */
    printf("\tPastec_FS_Station *Fis_%d = new Pastec_FS_Station;\n", 
	   Trans_Num );
    printf("\tFis_%d->set_name(\"Fis_%d\");\n", Trans_Num, Trans_Num );
    printf("\tFis_%d->set_gss_name(\"%s\");\n", Trans_Num, The_Trans->Data );
  }
  else if ( ( Real_Outdegree[Trans_Num] == 1 )
	&& ( Real_Indegree[Trans_Num] != 0 ) ) {
    /* faute de fusion, il faut changer dans la transition elle meme	*/
    /* la classe de sortie des clients. Sauf si c'est une source, qui 	*/
    /* specifie la classe de ses clients. 				*/
    printf("\tTrans_%d->set_output_class( CL_%d );\n", 
	   Trans_Num, Trans_Num );
  }
}

/************************************************************************/
/* Procedure to write out the routing of a place, assumed to be 	*/
/* "existing", that is, either with one output or with a probabilistic	*/
/* routing.								*/
/************************************************************************/
void gen_prosit_place_routing( Reseau, Place_Num )
PetriNet	*Reseau;
int		Place_Num;
{
  int		i;
  Pl_Des	*The_Place;
  int		The_Dest;
  Liste_Link	L_Scan;
  Routing_Desc	The_Routing;

  The_Place = &(Reseau->P[Place_Num]);

  printf("\n\t// routing for place #%d (", Place_Num );
  print_place_name( The_Place );
  printf(")\n");

  if ( The_Place->Outdegree == 0 ) {
    printf("\tPlace_%d->set_succ(", Place_Num );
    printf("The_Sink, 1.0, \"%s\", \"%s\");\n",
	   The_Place->Data, NULL_STRING );
  }
  else if ( The_Place->Outdegree == 1 ) {
    The_Dest = The_Place->Outputs->Dest;
    printf("\tPlace_%d->set_succ(", Place_Num );
    print_prosit_place_output_dest( Reseau, The_Place->Outputs );
    printf(", 1.0, \"%s\", \"%s\");\n",
	   The_Place->Data, NULL_STRING );
  }
  else {
    The_Routing = The_Place->Routing;
    if ( ( The_Routing == NULL ) 
	 || ( The_Routing->Type == 'C' )
	 || ( The_Routing->Type == 'D' )
	 || ( The_Routing->Type == 'P' ) ) {
      fprintf( stderr, "%s: error: bad routing at place #%d (%s).\n",
	       TOOLNAME, Place_Num, The_Place->Name );
    }
    else {
      i = 0;
      SCAN( L_Scan, The_Place->Outputs ) {
	The_Dest = L_Scan->Dest;
	printf("\tPlace_%d->set_succ(", Place_Num );
	print_prosit_place_output_dest( Reseau, L_Scan );
	printf(", %8.6f,\"%s\",\"%s\");\n",
	       The_Routing->Probas[i],
	       The_Place->Data, NULL_STRING );
	i++;
      }
    }
  }
  printf("\tPlace_%d->set_output_class( CL_%d );\n",
	 Place_Num, Place_Num );

}

void gen_prosit_transition_routing( Reseau, Trans_Num )
PetriNet	*Reseau;
int		Trans_Num;
{
  Tr_Des	*The_Trans;
  Pl_Des	*The_Place;
  int		The_Dest;
  Liste_Link	L_Scan;

  The_Trans = &(Reseau->T[Trans_Num]);

  printf("\n\t// routing for transition #%d (", Trans_Num );
  print_name( The_Trans );
  printf(")\n");

  if ( Real_Indegree[Trans_Num] > 1 ) {
    /* fusion is necessary */
    printf("\tFus_Rule = new SM_Rule;\n");
    SCAN( L_Scan, The_Trans->Inputs ) {
      The_Place = &(Reseau->P[L_Scan->Dest]);
      if ( !is_recycling_place( The_Place ) ) {
	printf("\tFus_Rule->match( CL_%d, %d );\n", 
	       Class_Number( Reseau, L_Scan ), L_Scan->Weight );
      }
    }
    printf("\tFus_Rule->split( DUMMY_CLASS, 1, 0, Trans_%d, \"%s\");\n", 
	   Trans_Num, The_Trans->Data );
    printf("\tFus_%d->set_rule( Fus_Rule );\n", Trans_Num );
    printf("\tFus_%d->set_succ(Trans_%d,1.0,\"%s\",\"%s\");\n",
	   Trans_Num, Trans_Num, The_Trans->Data, NULL_STRING );
  }
  if ( Real_Outdegree[Trans_Num] > 1 ) {
    /* fission is necessary */
    printf("\tTrans_%d->set_succ(Fis_%d,1.0,\"%s\",\"%s\");\n", 
	   Trans_Num, Trans_Num, The_Trans->Data, NULL_STRING );
    if ( Real_Indegree[Trans_Num] == 1 ) {
      /* change of class is necessary. */
      printf("\tTrans_%d->set_output_class( DUMMY_CLASS );\n", Trans_Num );
    }
    printf("\n");
    printf("\tFis_Rule = new SM_Rule;\n");
    SCAN( L_Scan, The_Trans->Outputs ) {
      The_Dest = L_Scan->Dest;
      The_Place = &(Reseau->P[The_Dest]);
      if ( !is_recycling_place( The_Place ) ) {
	/* Add the place in the fission rule.				*/
	printf("\tFis_Rule->split( CL_%d, %d, 0, ", 
	       Trans_Num, L_Scan->Weight );
	print_prosit_trans_output_dest( Reseau, L_Scan );
	printf(", \"%s\");\n", The_Trans->Data );
      }
    }
    printf("\tFis_Rule->match( DUMMY_CLASS, 1 );\n" );
    printf("\tFis_%d->set_rule( Fis_Rule );\n", Trans_Num );
  }
  else if ( Real_Outdegree[Trans_Num] == 1 ) {
    /* simple routing */
    if ( Real_Indegree[Trans_Num] == 0 )
      printf("\tSrc_%d->set_succ(", Trans_Num );
    else
      printf("\tTrans_%d->set_succ(", Trans_Num );

    /* find the --true-- output link */
    if ( !is_recycling_place( &Reseau->P[The_Trans->Outputs->Dest] ) )
      print_prosit_trans_output_dest( Reseau, The_Trans->Outputs );
    else
      print_prosit_trans_output_dest( Reseau, The_Trans->Outputs->Next );
    printf(",1.0,\"%s\",\"%s\");\n", NULL_STRING, NULL_STRING );
  }
  else {
    /* sink routing */
    printf("\tTrans_%d->set_succ(The_Sink,1.0,\"FooName\",\"FooName\");\n",
	   Trans_Num );
  }
}

void gen_prosit_stats_trans( Trans_Num )
int		Trans_Num;
{
  if ( Real_Indegree[ Trans_Num ] == 0 )
    printf("\tSrc");
  else
    printf("\tTrans");

  printf("_%d->compute_stat();\n", Trans_Num );
}

void gen_prosit_stats_place( Place_Num )
int		Place_Num;
{
  printf("\tPlace_%d->compute_stat();\n", Place_Num );
}

void gen_prosit_outputs_trans( Reseau, Trans_Num )
PetriNet	*Reseau;
int		Trans_Num;
{
  if ( No_Stats ) {
    CODE2("\tcout << \"%c_%d[\" << T_Max << \"] =   \";\n", 
	  ( Is_Source( Reseau, Trans_Num ) ? 'R' : 'Z' ), Trans_Num );
    CODE("\tcout << ");
    CODE2("%s_%d->",
	  ( Is_Source( Reseau, Trans_Num ) ? "Src" : "Trans" ), Trans_Num );
    CODE1("%s() << \"   Cycle time:   \";\n", 
	  ( Is_Source( Reseau, Trans_Num ) 
	    ? "total_population" 
	    : "get_nb_service" ) );
    CODE3("\tcout << %s_%d->%s()/T_Max;\n", 
	  ( Is_Source( Reseau, Trans_Num ) ? "Src" : "Trans" ), 
	  Trans_Num,
	  ( Is_Source( Reseau, Trans_Num )
	    ? "total_population" 
	    : "get_nb_service" ) );
    CODE1("\tcout << \"  (%s)\" << endl;\n", Reseau->T[ Trans_Num ].Name );
  }
  else 
    CODE2("\t%s_%d->report_stat();\n", 
	  ( Real_Indegree[ Trans_Num ] == 0 ? "Src" : "Trans" ),
	  Trans_Num );
}

void gen_prosit_outputs_place( Reseau, Place_Num )
PetriNet	*Reseau;
int		Place_Num;
{
  if ( No_Stats ) {
    CODE1("\tcout << \"Y_%d[\" << T_Max << \"] =   \";\n", Place_Num );
    CODE1("\tcout << Place_%d->get_nb_service();\n", 
	  Place_Num );
    CODE1("\tcout << \"   Cycle time:    \" << ", Place_Num );
    CODE1("Place_%d->get_nb_service()/T_Max;\n", Place_Num );
    CODE1("\tcout << \"  (%s)\" << endl;\n", Reseau->P[Place_Num].Name );
  }
  else 
    printf("\tPlace_%d->report_stat();\n", Place_Num );
}

void gen_prosit_code( Reseau )
PetriNet        *Reseau;
{
  int		i;
  boolean	Some_Places;

  printf("/******************************************/\n");
  printf("/* PROSIT code generated by %s\t  */\n", TOOLNAME );
  printf("/* Date: %s\t\t  */\n", todays_date() );
  printf("/* Model Name: \t\t\t\t  */\n" );
  printf("/******************************************/\n");

  printf("\n");
  printf("const int INFINI = 100;\n");
  printf("const int DUMMY_CLASS = 0;\n");
  for ( i=0; i< max( Reseau->Nb_Trans, Reseau->Nb_Places ); i++ )
    printf("#define CL_%d\t%d\n", i, i+1 );
  printf("\n");

  /* includes */
  printf("#include <stdio.h>\n");
  printf("#include <iostream.h>\n");
  printf("#include <Uniform.h>\n");
  printf("#include <Geom.h>\n");
  printf("#include <Binomial.h>\n");
  printf("#include <NegExp.h>\n");
  printf("#include <Erlang.h>\n");
  printf("#include \"Cst.hh\"\n");
  printf("#include \"OnOff.hh\"\n");
  printf("#include <MLCG.h>\n");
  printf("#include <bool.h>\n");
  printf("#include <fstream.h>\n");
  printf("#include \"Pastec_Sink.hh\"\n");
  printf("#include \"Pastec_Routing.hh\"\n");
  printf("#include \"Pastec_Server_Normal_FIFO.hh\"\n");
  printf("#include \"Pastec_Server_Normal_LIFO.hh\"\n");
  printf("#include \"Pastec_Server_Normal_PRIORITY.hh\"\n");
  printf("#include \"Pastec_Source.hh\"\n");
  printf("#include \"Kernel.hh\"\n");
  printf("#include \"Pastec_Kernel.hh\"\n");
  printf("#include \"Pastec_Client.hh\"\n");
  printf("#include \"Pastec_FS_Station.hh\"\n");
  printf("#include \"SM_Rule.hh\"\n");  
  printf("\n");

  /* declarations */
  printf("Pastec_Kernel* pnoyau = new Pastec_Kernel;\n");
  printf("Kernel *noyau = pnoyau;\n");
  printf("SM_Rule *Fus_Rule;\n");
  printf("SM_Rule *Fis_Rule;\n");
  printf("\n");
  printf("#include \"%s\"\n", SOURCE_CODE );
  printf("template class Routing_Source<Pastec_Client>;\n");
  printf("\n\n");

  /* code du main */
  printf("main(int argc, char *argv[]) {\n\n");
  CODELN;
  CODE("  int\ti = 1;\n");
  CODE1("  double\tT_Max = %f;\n", 
	(double)Reseau->Sim_Params->Nb_Firings );
  CODELN;

  CODE("  // analysis of command line arguments\n");
  CODE("  if ( argc > 1 )\n");
  CODE("    do {\n");
  CODE("      if ( !strcmp ( argv[i], \"-tmax\" ) ) {\n");
  CODE("        i++;\n");
  CODE("        if ( 0 == sscanf( argv[i], \"%%lf\", &T_Max ) ) {\n");
  CODE("          cout << \"Bad argument for -tmax\" << endl;\n");
  CODE("          exit(-1);\n");
  CODE("        }\n");
  CODE("        else {\n");
  CODE("          i++;\n");
  CODE("        }\n");
  CODE("      }\n");
  CODE("      else if ( !strcmp ( argv[i], \"-h\" ) ) {\n");
  CODE("        cout << \"Usage: \" << argv[0] << \" [-h] \";\n");
  CODE("        cout << \" [-tmax <real>]\\n\";\n");
  CODE("        exit(0);\n");
  CODE("      }\n");
  CODE("      else {\n");
  CODE("        // ignore unknown option\n");
  CODE("        i++;\n");
  CODE("      }\n");
  CODE("    } while ( i < argc );\n");
  CODELN;

  /* generate a Zero time law, which is rarely useless */
  printf("  Cst *Zero = new Cst( 0.0 );\n");
  printf("\n");

  /* generate a sink, if necessary */
  if ( Has_Sinks( Reseau ) ) {
    printf("\t// declaration of the sink\n");
    printf("\tPastec_Sink *The_Sink = new Pastec_Sink(0.0);\n");
    printf("\tThe_Sink->set_name(\"The_Sink\");\n\n");
  }
    
  printf("\t// configuration of transition objects\n");
  for ( i=0; i<Reseau->Nb_Trans; i++ )
      gen_prosit_transition( Reseau, i );

  printf("\n");

  Some_Places = FALSE;
  for ( i=0; i<Reseau->Nb_Places; i++ )
    if ( Place_Exists[i] )
      Some_Places = TRUE;

  if ( Some_Places ) {
    printf("\t// configuration of place objects\n");
    for ( i=0; i<Reseau->Nb_Places; i++ ) {
      if ( Place_Exists[i] )
	gen_prosit_place( Reseau, i );
      else if ( Reseau->P[i].Init_Pop != 0 ) {
	/* il faut generer le marquage initial ou il faut */
	print_initial_pop( Reseau, i );
      }
    }
  }
  printf("\n");

  printf("\t// configuration of the routing\n");
  for ( i=0; i<Reseau->Nb_Trans; i++ )
      gen_prosit_transition_routing( Reseau, i );
  for ( i=0; i<Reseau->Nb_Places; i++ )
    if ( Place_Exists[i] )
      gen_prosit_place_routing( Reseau, i );

  printf("\n");

  /* parametrage du modele */
  printf("\t// parameters of the simulation\n");
  printf("\tpnoyau->set_model_name(\"FooModel\");\n");
  printf("\tpnoyau->set_starting_time(0.0);\n");
  printf("\tpnoyau->set_ending_time( T_Max );\n" );
  printf("\tpnoyau->set_debug(false);\n");
  printf("\tpnoyau->set_animation(%s);\n",
	 ( Agat ? "true" : "false" ) );
  printf("\tpnoyau->set_statistic(%s);\n",
	 ( No_Stats ? "false" : "true" ) );

  if ( Trace )
    printf("\tpnoyau->init_trace(\"%s\");\n", Trace_File_Name );

  printf("\n");

  /* ordres de calcul des statistiques */
  if ( !No_Stats ) {
    printf("\t// computation of statistics\n");
    for ( i=0; i<Reseau->Nb_Trans; i++ )
      gen_prosit_stats_trans( i );
    for ( i=0; i<Reseau->Nb_Places; i++ )
      if ( Place_Exists[i] )
	gen_prosit_stats_place( i );
    printf("\n");
  }

  /* execution de la simulation */
  printf("\t// the simulation itself\n");
  printf("\tcout << \"Start of the simulation\" << endl;\n");
  printf("\tnoyau->print_registered();\n");
  printf("\tnoyau->run();\n");
  printf("\tcout << \"End of the simulation\" << endl;\n");
  printf("\n");

  /* sortie des resultats */
  printf("\t// output of results\n");
  for ( i=0; i<Reseau->Nb_Trans; i++ )
      gen_prosit_outputs_trans( Reseau, i );
  for ( i=0; i<Reseau->Nb_Places; i++ )
      if ( Place_Exists[i] )
	gen_prosit_outputs_place( Reseau, i );

  printf("\n");
  /* c'est fini! */
  printf("}\n");

}

void gen_agat_code( Reseau )
PetriNet	*Reseau;
{
  int	i;

  freopen( "sim.agat", "w", stdout );

  for ( i=0; i<Reseau->Nb_Trans; i++ )
    if ( Real_Indegree[i] != 0 ) {
      printf("plotAxeLine(%s__Nombre_de_clients__temps,", Reseau->T[i].Name );
      printf("%s__Nombre_de_clients);\n", Reseau->T[i].Name );
    }
  for ( i=0; i<Reseau->Nb_Places; i++ )
    if ( Place_Exists[i] ) {
      printf("plotAxeLine(%s__Nombre_de_clients__temps,", 
	     Reseau->P[i].Name );
      printf("%s__Nombre_de_clients);\n", Reseau->P[i].Name );
    }
}
/************************************************************************/
/* Fonction	satisfies_constraints( Reseau )				*/
/* verifie si le reseau satisfait aux contraintes necessaires pour etre	*/
/* transcrit en un reseau de files d'attente sous Prosit		*/
/* Les contraintes sont pour le moment:					*/
/* - reseau a routage							*/
/* - pas d'arc inhibiteur						*/
/*									*/
/* Argument: 	Reseau, de type PetriNet				*/
/* Retourne:	TRUE si le reseau satisfait aux contraintes, FALSE sinon*/
/************************************************************************/
boolean	satisfies_constraints( Reseau )
PetriNet	*Reseau;
{
  boolean	Is_Ok = TRUE;
  Pl_Des	*The_Place;
  Tr_Des	*The_Trans;
  int		i;

  if ( Reseau->Has_Inhibitors ) {
    fprintf( stderr, "%s: the Petri network has inhibitor arcs. ", 
	     TOOLNAME );
    fprintf( stderr, "No translation.\n" );
    return FALSE;
  }
  else {
    for ( i=0; (i<Reseau->Nb_Places) && Is_Ok; i++ ) {
      The_Place = &( Reseau->P[i] );
      if ( ( The_Place->Outdegree > 1 )
	   && ( The_Place->Routing != NULL ) ) {
	if ( ( The_Place->Routing->Type == 'C' ) 
	     || ( The_Place->Routing->Type == 'D' )
	     || ( The_Place->Routing->Type == 'P' ) ) {
	  Is_Ok = FALSE;
	  fprintf( stderr, "%s: the Petri network cannot ", TOOLNAME );
	  fprintf( stderr, "be translated.\n\t" );
	  fprintf( stderr, "Place #%d [%s] has a non-probabilistic routing policy.\n",
		   The_Place->Pl_Id, The_Place->Name );
	}
      }
    }
  }
  
  return Is_Ok;
}
