/* Translation program from pn language to eg language.	  */

#include "pn.h"

#define	TOOLNAME	"pn2eg"

#define SCAN(x,y)	for(x=y;x!=NULL;x=x->Next)
#define IRINT(x)	(int)rint(x)

extern PetriNet		*Read_Network(char *File_Name);
extern void		dumpe_network(PetriNet *RdP);
extern char		*todays_date();

/* forward declarations */
boolean	Is_An_Event_Graph(PetriNet *RdP);
void	dump_egl_code(PetriNet *RdP);

int main(
     int	argc,
     char	*argv[] )
{
  static char		File_Name[128] = "";
  static char		Code_File_Name[128] = "";
  PetriNet 		*Reseau;
  int			i;
  boolean		Help = FALSE;
  boolean		Args_Ok = TRUE;
  boolean		File_Ok = TRUE;

  i = 1;

  if (argc>1)
    do {

      if GET_STRING( "-f", File_Name )
      else if GET_STRING( "-o", Code_File_Name )

      else if SWITCH( "-debug", Debug, TRUE )

      else if IS_VERSION

      else if ( ( 0 == strcmp( "-h", argv[i] ) )
		|| ( 0 == strcmp( "-?", argv[i] ) ) ) {
	printf("Usage: %s [-f <file>] ", TOOLNAME );
	printf("[-debug] [-o <file>]\n");
	Args_Ok = FALSE;
	Help = TRUE;
	i++;
      }

      else
	i++; /* skip unknown argument */
    }
    while ( Args_Ok && ( i < argc ) );


  if ( Args_Ok ) {

    Reseau = Read_Network( File_Name );

    if ( Reseau != NULL ) {

	if ( Reseau->Has_Inhibitors ) {
	  printf( "The network has inhibitor arcs. " );
	  printf( "It cannot be an Event Graph.\n" );
	  return -2;
	} 
	else if ( Reseau->Has_Weights ) {
	  printf( "The network has weighted arcs. " );
	  printf( "It cannot be an Event Graph.\n" );
	  return -2;
	}
	else if ( Is_An_Event_Graph(Reseau) ) {

	  if ( strlen(Code_File_Name) != 0 )
	    File_Ok = ( NULL != freopen( Code_File_Name, "w", stdout ) );

	  if ( File_Ok ) {
	    dump_egl_code(Reseau);
	    return 0;
	  }
	  else {
	    fprintf( stderr, "%s: Error opening file %s.\n", 
		     TOOLNAME, Code_File_Name );
	    return -1;
	  }
	}
	else {
	  printf("This network is not an Event Graph. Check it!\n");
	  return -3;
	}
      }
      else {
	fprintf( stderr, "%s: Error reading network. No translation.\n", 
		 TOOLNAME );
	return -2;
      }
    }

  else if (!Help)
    return -1;
  else
    return 0;
}

boolean	Is_An_Event_Graph(PetriNet *Reseau)
{
  boolean	Is_One = TRUE;
  int	i;

  for( i=0; (i<Reseau->Nb_Places) && (Is_One); i++)
    Is_One = ( Is_One
	       && ( Reseau->P[i].Indegree == 1 )
	       && ( Reseau->P[i].Outdegree == 1 ) );

  return Is_One;
}

/* procedure to write the initial population. Improved from gen_eg.c */
void fprint_init_mark(Pl_Des	The_Place)
{
  int j;

  printf("%1d ", The_Place.Init_Pop);
  for( j=0; j < The_Place.Init_Pop; j++ )
    printf("%8.4f ", The_Place.Lag_Times[j]);

}

void dump_egl_code(PetriNet *Reseau)
{
  int		i;
  Tr_Des	The_Trans;
  Pl_Des	The_Place;
  Liste_Link	L_Scan;

  printf("# egl code generated by %s\n", TOOLNAME );
  printf("# Date: %s\n", todays_date() );
  if ( strlen( Reseau->File_Name ) > 0 )
    printf("# Input file: %s\n", Reseau->File_Name );
  printf("Size <- %d\n", Reseau->Nb_Trans );

  /* printf("Length <- %d\n", Reseau->Sim_Params->Nb_Firings); */

  for ( i=0; i<Reseau->Nb_Trans; i++ ) {

    printf("%d ", i+1 );

    The_Trans = Reseau->T[i];
    /* ecriture du nom. Si un champ "Data" existe, on	*/
    /* l'ecrit egalement.					*/
    if (NULL != The_Trans.Name)
      if ( strlen( The_Trans.Name ) > 0 )
	printf("\"%s\" ", The_Trans.Name);
    if (NULL != The_Trans.Data)
      if ( strlen( The_Trans.Data ) > 0 )
	printf("\"%s\" ", The_Trans.Data);

    Write_Law( stdout, The_Trans.Tr_Law, NORMAL_PRINT_MODE );

    if ( The_Trans.Outputs != NULL )
      SCAN( L_Scan, The_Trans.Outputs ) 
	if ( !L_Scan->Is_Inhibitor ) {
	  printf(", ");
	  The_Place = Reseau->P[L_Scan->Dest];
	  printf("%d ", The_Place.Outputs->Dest+1);
		
	  if (NULL != The_Place.Data) {
	    if ( strlen( The_Place.Data ) > 0 )
	      printf("\"%s\" ", The_Place.Data);
	  }
	  else if (NULL != The_Place.Name) {
	    if ( strlen( The_Place.Name ) > 0 )	    
	      printf("\"%s\" ", The_Place.Name);
	  }

	  if ( NULL != L_Scan->Data ) {
	    /* ecrire le nom lien entree */
	    if ( strlen( L_Scan->Data ) > 0 )
	      printf("\"%s\" ", L_Scan->Data );
	    /* ecrire le nom du lien sortie	*/
	    if ( Reseau->P[L_Scan->Dest].Outputs->Data != NULL )
	      if ( strlen( Reseau->P[L_Scan->Dest].Outputs->Data ) > 0 )
		printf("\"%s\" ",
		       Reseau->P[L_Scan->Dest].Outputs->Data );
	  }
	  Write_Law( stdout, The_Place.Pl_Law, NORMAL_PRINT_MODE );
	  fprint_init_mark( The_Place );
	}

    printf(";\n");
  }
}
