/* Translation program from pn language to some graph languages.	  */

#include <values.h>
#include <math.h>
#include "pn.h"

#define	TOOLNAME	"pn2graph"

#define SCAN(x,y)	for(x=y;x!=NULL;x=x->Next)
#define IRINT(x)	(int)rint(x)

#define MODE_DOT	1
#define MODE_DAVINCI	2
#define MODE_SCILAB	3
#define MODE_PNED	4

extern PetriNet		*Read_Network(char *File_Name);
extern void		dumpe_network(PetriNet *RdP);
extern char		*todays_date(void);

/* forward declarations */
void	dump_daVinci_code( PetriNet *Reseau, boolean Fullname );
void	dump_dot_code( PetriNet *Reseau, boolean Fullname );
void	dump_scilab_code( PetriNet *Reseau, boolean Fullname );
void	dump_pned_code( PetriNet *Reseau, boolean Fullname );

int main(int argc, char *argv[])
{
  static char		File_Name[128] = "";
  static char		Code_File_Name[128] = "";
  PetriNet 		*Reseau;
  int			i;
  int			Mode = MODE_DAVINCI;
  boolean		Help = FALSE;
  boolean		Args_Ok = TRUE;
  boolean		Fullname = FALSE;
  boolean		File_Ok = TRUE;

  i = 1;

  if (argc>1)
    do {

      if GET_STRING( "-f", File_Name )
      else if GET_STRING( "-o", Code_File_Name )

      else if SWITCH( "-debug", Debug, TRUE )

      else if SWITCH( "-fullname", Fullname, TRUE )

      else if IS_VERSION

      else if ( ( 0 == strcmp( "-h", argv[i] ) )
		|| ( 0 == strcmp( "-?", argv[i] ) ) ) {
	printf("Usage: %s [-f <file>] [-mode dot|daVinci|scilab|pned]", 
	       TOOLNAME );
	printf(" [-fullname]\n\t[-debug] [-o <file>]\n");
	Args_Ok = FALSE;
	Help = TRUE;
	i++;
      }
      else if ( 0 == strcmp( "-mode", argv[i] ) ) {
	i++;
	if ( i < argc ) {
	  if ( 0 == strcmp( "dot", argv[i] ) )
	    Mode = MODE_DOT;
	  else if ( 0 == strcmp( "daVinci", argv[i] ) )
	    Mode = MODE_DAVINCI;
	  else if ( 0 == strcmp( "scilab", argv[i] ) )
	    Mode = MODE_SCILAB;
	  else if ( 0 == strcmp( "pned", argv[i] ) )
	    Mode = MODE_PNED;
	  else {
	    printf("%s: error: invalid mode name %s.\n", TOOLNAME, argv[i] );
	    Args_Ok = FALSE;
	  }
	  i++;
	}
	else {
	  printf("%s: error: mode name expected after %s.\n", TOOLNAME, 
		 argv[i-1] );
	  Args_Ok = FALSE;
	}
      }

      else
	i++; /* skip unknown argument */
    }
    while ( Args_Ok && ( i < argc ) );


  if ( Args_Ok ) {

    Reseau = Read_Network( File_Name );

    if ( Reseau != NULL ) {

      if ( strlen(Code_File_Name) != 0 )
	File_Ok = ( NULL != freopen( Code_File_Name, "w", stdout ) );

      if ( File_Ok ) {
	switch( Mode ) {
	case MODE_DOT:
	  dump_dot_code( Reseau, Fullname );
	  break;
	case MODE_DAVINCI:
	  dump_daVinci_code( Reseau, Fullname );
	  break;
	case MODE_SCILAB:
	  dump_scilab_code( Reseau, Fullname );
	  break;
	case MODE_PNED:
	  dump_pned_code( Reseau, Fullname );
	  break;
	}
	return 0;
      }
      else {
	fprintf( stderr, "%s: Error opening file %s.\n", 
		 TOOLNAME, Code_File_Name );
	return -1;
      }
    }
    else {
      fprintf( stderr, "%s: Error reading network. No translation.\n", 
	       TOOLNAME );
      return -2;
    }
  }

  else if (!Help)
    return -1;
  else
    return 0;
}

/* procedure to write the initial population. Improved from gen_eg.c */
void fprint_init_mark(Pl_Des The_Place)
{
  int j;

  printf("%1d ", The_Place.Init_Pop);
  for( j=0; j < The_Place.Init_Pop; j++ )
    printf("%8.4f ", The_Place.Lag_Times[j]);

}

void dump_daVinci_code(
PetriNet        *Reseau,
boolean		Fullname )
{
  int		i;
  Tr_Des	The_Trans;
  Pl_Des	The_Place;
  Liste_Link	L_Scan;

  /*
  printf("# DaVinci2.0 code generated by %s\n", TOOLNAME );
  printf("# Date: %s\n", todays_date() );
  */

  printf("[\n");

  for ( i=0; i<Reseau->Nb_Trans; i++ ) {

    printf("\tl(\"T%d", i+1 );

    The_Trans = Reseau->T[i];
    printf("\",n(\"\",[\n");
    /* attributes */
    /* ecriture du nom. */
    printf("\t\ta(\"OBJECT\",\"T%d", i+1 );
    if ( Fullname && ( NULL != The_Trans.Name ) ) {
      if ( strlen( The_Trans.Name ) > 0 )
	printf("\\n%s", The_Trans.Name);
    }
    printf("\"),\n");
    printf("\t\ta(\"COLOR\",\"red\"),\n");
    printf("\t\ta(\"_GO\",\"box\")],\n");

    if ( The_Trans.Outputs != NULL ) {
      printf("\t\t[\n");
      SCAN( L_Scan, The_Trans.Outputs ) {
	printf("\t\t\tl(\"T%d->P%d\",e(\"\",[],r(\"P%d\")))",
	       i, L_Scan->Dest+1, L_Scan->Dest+1 );
	if ( L_Scan->Next != NULL )
	  printf(",");
	printf("\n");
      }
      printf("\t\t]");
    }
    else
      printf("\t\t[]");

    printf("    )),\n");
  }

  for ( i=0; i<Reseau->Nb_Places; i++ ) {

    printf("\tl(\"P%d", i+1 );

    The_Place = Reseau->P[i];
    /* ecriture du nom. Si un champ "Data" existe, on	*/
    /* l'ecrit egalement.				*/
    /*
      if (NULL != The_Place.Name)
      if ( strlen( The_Place.Name ) > 0 )
      printf(":%s", The_Place.Name);
      */
    printf("\",n(\"\",[\n");
    /* attributes */
    /* ecriture du nom. */
    printf("\t\ta(\"OBJECT\",\"P%d", i+1 );
    if ( Fullname && ( NULL != The_Place.Name ) ) {
      if ( strlen( The_Place.Name ) > 0 )
	printf("\\n%s", The_Place.Name);
    }
    printf("\"),\n");
    printf("\t\ta(\"COLOR\",\"lightblue\"),\n");
    printf("\t\ta(\"_GO\",\"circle\")],\n");
      
    if ( The_Place.Outputs != NULL ) {
      printf("\t\t[\n");
      SCAN( L_Scan, The_Place.Outputs ) {
	printf("\t\t\tl(\"P%d->T%d\",e(\"\",[],r(\"T%d\")))",
	       i, L_Scan->Dest+1, L_Scan->Dest+1 );
	if ( L_Scan->Next != NULL )
	  printf(",");
	printf("\n");
      }
      printf("\t\t]");
    }
    else
      printf("\t\t[]");

    printf("))");
    if ( i == Reseau->Nb_Places - 1 )
      printf("\n");
    else
      printf(",\n");
  }

  printf("]\n");
}

void dump_dot_code(
PetriNet        *Reseau,
boolean		Fullname )
{
  int		i;
  Tr_Des	The_Trans;
  Pl_Des	The_Place;
  Subm_Des	The_Submodel;
  Liste_Link	L_Scan;

  /*
  printf("# dot code generated by %s\n", TOOLNAME );
  printf("# Date: %s\n", todays_date() );
  */

  printf("digraph \"Petri Net\" {");

  if ( strlen( Reseau->File_Name ) > 0 ) {
    printf("\tgraph [\n");
    printf("\tlabel = \"%s\",\n", Reseau->File_Name );
    printf("\trankdir = LR" );
    printf("];\n");
  }

  printf("\tnode [\tshape = box,\n");
  printf("\t\tstyle = filled,\n");
  printf("\t\tcolor = red ];\n");

  for ( i=0; i<Reseau->Nb_Trans; i++ ) {

    printf("\tTr%d", i+1 );

    The_Trans = Reseau->T[i];
    /* attributes */
    /* ecriture du nom. */
    printf(" [\tlabel=\"T%d", i+1 );
    if ( Fullname && ( NULL != The_Trans.Name ) ) {
      if ( strlen( The_Trans.Name ) > 0 )
	printf("\\n%s", The_Trans.Name);
    }
    printf("\" ];\n");
  }

  printf("\tnode [\tshape = circle,\n");
  printf("\t\tstyle = filled,\n");
  printf("\t\tcolor = lightblue ];\n");

  for ( i=0; i<Reseau->Nb_Places; i++ ) {

    printf("\tPl%d", i+1 );

    The_Place = Reseau->P[i];
    /* ecriture du nom. Si un champ "Data" existe, on	*/
    /* l'ecrit egalement.				*/
    /*
      if (NULL != The_Place.Name)
      if ( strlen( The_Place.Name ) > 0 )
      printf(":%s", The_Place.Name);
      */
    /* attributes */
    /* ecriture du nom. */
    printf(" [\tlabel=\"P%d", i+1 );
    if ( Fullname && ( NULL != The_Place.Name ) ) {
      if ( strlen( The_Place.Name ) > 0 )
	printf("\\n%s", The_Place.Name);
    }
    printf("\" ];\n");
  }

  printf("\tnode [\tshape = box,\n");
  printf("\t\tstyle = filled,\n");
  printf("\t\tperipheries = 2,\n");
  printf("\t\tcolor = gray50 ];\n");

  for ( i=0; i<Reseau->Nb_Submodels; i++ ) {

    The_Submodel = Reseau->S[i];

    printf("\tSub%d", i+1 );
    printf(" [\tlabel=\"S%d", The_Submodel.Subm_Id );
    if ( strlen( The_Submodel.Name ) > 0 )
      printf("\\n%s", The_Submodel.Name );
    printf("\\n%s\" ];\n", The_Submodel.File_Name );

  }

  for ( i=0; i<Reseau->Nb_Trans; i++ ) {

    The_Trans = Reseau->T[i];
    if ( The_Trans.Outputs != NULL ) {
      SCAN( L_Scan, The_Trans.Outputs ) {
	printf("\tTr%d -> Pl%d", i+1, L_Scan->Dest+1 );
	if ( L_Scan->Weight > 1 ) {
	  printf(" [ label=%d ]", L_Scan->Weight );
	}
	printf(";\n");
      }
    }
  }

  for ( i=0; i<Reseau->Nb_Places; i++ ) {

    The_Place = Reseau->P[i];

    if ( The_Place.Outputs != NULL ) {
      SCAN( L_Scan, The_Place.Outputs ) {
	printf("\tPl%d -> Tr%d", i+1, L_Scan->Dest+1 );
	if ( L_Scan->Weight > 1 ) {
	  printf(" [ label=%d ]", L_Scan->Weight );
	}
	printf(";\n");
      }
    }
  }

  printf("}\n");
}

void dump_scilab_code(
PetriNet        *Reseau,
boolean		Fullname )
{
  int		i;
  int		Arc_Nb;
  Liste_Link	L_Scan;

  /*
  printf("# dot code generated by %s\n", TOOLNAME );
  printf("# Date: %s\n", todays_date() );
  */

  /* Counting Nb of arcs */
  Arc_Nb = 0;
  for ( i=0; i < Reseau->Nb_Trans; i++ )
    Arc_Nb += ( Reseau->T[i].Indegree + Reseau->T[i].Outdegree );

  printf("GRAPH TYPE (0 = UNDIRECTED, 1 = DIRECTED), ");
  printf("DEFAULTS (NODE DIAMETER, NODE BORDER, ARC WIDTH, ");
  printf("HILITED ARC WIDTH, FONTSIZE):\n");
  printf("1 20 2 1 2 12\n" );
  printf("NUMBER OF ARCS:\n");
  printf("%d\n", Arc_Nb );
  printf("NUMBER OF NODES:\n");
  printf("%d\n", Reseau->Nb_Places + Reseau->Nb_Trans );

  printf("****************************************\n");
  printf("DESCRIPTION OF ARCS:\n");
  printf("ARC NAME, TAIL NODE NAME, HEAD NODE NAME, COLOR, WIDTH, HIWIDTH, FONTSIZE\n");
  printf("COST, MIN CAP, MAX CAP, LENGTH, Q WEIGHT, Q ORIGIN, WEIGHT\n");
  printf("\n");

  Arc_Nb = 1;
  for ( i=0; i<Reseau->Nb_Trans; i++ ) {
    SCAN( L_Scan, Reseau->T[i].Outputs ) {
      printf("%d %d %d 0 0 0 0\n",
	     Arc_Nb, i+1, Reseau->Nb_Trans + L_Scan->Dest + 1 );
      printf("0.0 0.0 0.0 0.0 0.0 0.0 0.0\n");
      Arc_Nb++;
    }
    SCAN( L_Scan, Reseau->T[i].Inputs ) {
      printf("%d %d %d 0 0 0 0\n",
	     Arc_Nb, Reseau->Nb_Trans + L_Scan->Dest + 1, i+1 );
      printf("0.0 0.0 0.0 0.0 0.0 0.0 0.0\n");
      Arc_Nb++;
    }
  }

  printf("****************************************\n");
  printf("DESCRIPTION OF NODES:\n");
  printf("NODE NAME, POSSIBLE TYPE (1 = SINK, 2 = SOURCE)\n");
  printf("X, Y, COLOR, DIAMETER, BORDER, FONTSIZE\n");
  printf("DEMAND\n");
  printf("\n");

  for ( i=0; i<Reseau->Nb_Trans; i++ ) {
    printf("%d\n", i+1 );
    printf("0 0 10 20 1 12\n");
    printf("0.0\n");
  }

  for ( i=0; i<Reseau->Nb_Places; i++ ) {
    printf("%d\n", Reseau->Nb_Trans+i+1 );
    printf("0 0 3 30 1 12\n");
    printf("0.0\n");
  }

}

/* Check for __topological__ recycling. A little different from the	*/
/* functional recycling needed for FIFO (see pn2qnap, fifo_pn)		*/
boolean	is_recycling_place(Pl_Des *The_Place)
{
  boolean	Is_Recycling = FALSE;

  Is_Recycling = ( ( The_Place->Indegree == 1 ) 
		   && ( The_Place->Outdegree == 1 )
		   && ( The_Place->Inputs->Dest 
			== The_Place->Outputs->Dest ) );

  return Is_Recycling;
}

void dump_pned_code(
PetriNet        *Reseau,
boolean		Fullname )
 {
  int		i;
  int		Degree;
  double	x;
  double	y;
  Tr_Des	The_Trans;
  Pl_Des	The_Place;
  Liste_Link	L_Scan;
  char		*The_Date;
  double	*Tr_X_Pos;
  double	*Tr_Y_Pos;
  double	*Pl_X_Pos;
  double	*Pl_Y_Pos;

  int		Trans_Height = 30;
  int		Trans_Width = 5;
  int		Place_Radius = 30;

  printf("# pned code generated by %s\n", TOOLNAME );
  The_Date = todays_date();
  printf("# Date: %s\n", The_Date );
  free( The_Date );

  /* Computing the positions */
  Tr_X_Pos = (double*) calloc( (size_t)Reseau->Nb_Trans, sizeof(double) );
  Tr_Y_Pos = (double*) calloc( (size_t)Reseau->Nb_Trans, sizeof(double) );
  Pl_X_Pos = (double*) calloc( (size_t)Reseau->Nb_Places, sizeof(double) );
  Pl_Y_Pos = (double*) calloc( (size_t)Reseau->Nb_Places, sizeof(double) );

  for ( i=0; i<Reseau->Nb_Trans; i++ ) {
    Tr_X_Pos[i] = 250 + 200 * cos(2*M_PI*i/Reseau->Nb_Trans);
    Tr_Y_Pos[i] = 250 + 200 * sin(2*M_PI*i/Reseau->Nb_Trans);
  }

  for ( i=0; i<Reseau->Nb_Trans; i++ ) {
    The_Trans = Reseau->T[i];

    printf("Tr %d ", i );
    printf("%6.2f %6.2f %6.2f %6.2f", 
	   Tr_X_Pos[i], Tr_Y_Pos[i], 
	   Tr_X_Pos[i]+Trans_Width, Tr_Y_Pos[i]+Trans_Height );
    /* name */
    printf(" \"T%d", i+1 );
    if ( Fullname && ( NULL != The_Trans.Name ) ) {
      if ( strlen( The_Trans.Name ) > 0 )
	printf("\\n%s", The_Trans.Name);
    }
    printf("\" ");

    /* timing, if constant */
    if ( The_Trans.Tr_Law->Name == 'C' )
      printf(" %f\n",  The_Trans.Tr_Law->Parameters[0] );
    else 
      printf(" 0\n");
  }

  for ( i=0; i<Reseau->Nb_Places; i++ ) {
    The_Place = Reseau->P[i];

    printf("Pl %d ", i );

    if ( is_recycling_place( &The_Place ) ) {
      /* shift place wrt the transition */
      The_Trans = Reseau->T[The_Place.Outputs->Dest];
      Pl_X_Pos[i] = Tr_X_Pos[The_Trans.Tr_Id];
      Pl_Y_Pos[i] = Tr_Y_Pos[The_Trans.Tr_Id] + 2*Trans_Height;
    }
    else {
      /* Compute barycenter of transitions */
      Degree = The_Place.Indegree+The_Place.Outdegree;
      if ( Degree > 0 ) {
	x = 0.0;
	y = 0.0;
	SCAN( L_Scan, The_Place.Inputs ) {
	  x += Tr_X_Pos[L_Scan->Dest];
	  y += Tr_Y_Pos[L_Scan->Dest];
	}
	SCAN( L_Scan, The_Place.Outputs ) {
	  x += Tr_X_Pos[L_Scan->Dest];
	  y += Tr_Y_Pos[L_Scan->Dest];
	}
	Pl_X_Pos[i] = x / Degree;
	Pl_Y_Pos[i] = y / Degree;
      } 
      else {
	/* put this isolated place in a corner, trying to avoid 	*/
	/* superpositions...					*/
	Pl_X_Pos[i] = 500 + i;
	Pl_Y_Pos[i] = 500 - i;
      }
    }

    printf("%6.2f %6.2f %6.2f %6.2f", 
	   Pl_X_Pos[i]-Place_Radius/2, Pl_Y_Pos[i]-Place_Radius/2, 
	   Pl_X_Pos[i]+Place_Radius/2, Pl_Y_Pos[i]+Place_Radius/2 );

    /* name */
    printf(" \"P%d", i+1 );
    if ( Fullname && ( NULL != The_Place.Name ) ) {
      if ( strlen( The_Place.Name ) > 0 )
	printf("\\n%s", The_Place.Name);
    }
    printf("\" ");

    /* timing, if constant */
    if ( The_Place.Pl_Law->Name == 'C' )
      printf(" %f",  The_Place.Pl_Law->Parameters[0] );
    else 
      printf(" 0");

    /* marking */
    printf(" %d\n", The_Place.Init_Pop );
  }

  for ( i=0; i<Reseau->Nb_Trans; i++ ) {
    SCAN( L_Scan, Reseau->T[i].Outputs ) {
      printf("TrPl %d %d %d", i, L_Scan->Dest, L_Scan->Weight );
      printf(" %6.2f %6.2f %6.2f %6.2f\n", 
	     Tr_X_Pos[i], Tr_Y_Pos[i], 
	     Pl_X_Pos[L_Scan->Dest], Pl_Y_Pos[L_Scan->Dest] );
    }
  }
  for ( i=0; i<Reseau->Nb_Places; i++ ) {
    SCAN( L_Scan, Reseau->P[i].Outputs ) {
      printf("PlTr %d %d %d", i, L_Scan->Dest, L_Scan->Weight );
      printf(" %6.2f %6.2f %6.2f %6.2f\n", 
	     Pl_X_Pos[i], Pl_Y_Pos[i],
	     Tr_X_Pos[L_Scan->Dest], Tr_Y_Pos[L_Scan->Dest] );
    }
  }

  free( Tr_X_Pos );
  free( Tr_Y_Pos );
  free( Pl_X_Pos );
  free( Pl_Y_Pos );
}
