/* programme/module de generation d'un programme lineaire calculant	*/
/* des bornes sur les performances d'un reseau exponentiel		*/

#include <unistd.h>
#include <math.h>

#include "pn.h"

#ifndef LINUX
#define MAXINT          0x7FFFFFFF
#endif
#define INFINI		1e12

#define max(x,y)        ( (x>y) ? x : y )
#define min(x,y)        ( (x<y) ? x : y )

#define	TOOLNAME	"pn2lin"
#define LIN_PROG_NAME	"pn2lin.lp"
#define RES_FILE_NAME	"pn2lin.res"

#define CODE(x)		fprintf(Lin_Prog,x)
#define CODE1(x,y)	fprintf(Lin_Prog,x,y)
#define CODE2(x,y,z)	fprintf(Lin_Prog,x,y,z)
#define CODE3(x,y,z,t)	fprintf(Lin_Prog,x,y,z,t)
#define CODELN		CODE("\n")
	
#define SCAN(x,y)	for(x=y;x!=NULL;x=x->Next)
#define CLOSE(x,y)	fabs( (double)x - (double)y ) < 1e-4

#define	NAME(x,y)	Reseau->x[y].Name

/* predefined objective functions */
#define SUM_TH	0
#define MAX_TH	1
#define MIN_TH	2

/* Global declarations */
extern PetriNet		*Read_Network( char *File_Name );
extern void		dumpe_network( PetriNet *Reseau );
extern char		*todays_date( void );
extern Liste_Link	Append_Link( Liste_Link L1, Liste_Link L2 );
extern int		**Cycle_Base( int **A, int P, int T );

/* Forward declarations */
boolean	check_network( PetriNet *Reseau, int Bound );
void	gen_lin_code( PetriNet *Reseau, char *Objective, boolean Variances );
void	gen_solve_analyze( PetriNet *Reseau, char *Objective,
			   boolean Variances, boolean Check );

/************************************\
* 				     *
*  convenience background variables  *
* 				     *
\************************************/
boolean		Minimum = FALSE;
boolean		Other_Mu = FALSE;
boolean		Global_Bound = FALSE;
boolean		Other_Bounds = FALSE;
boolean		Verbose = FALSE;
char		Mu_File_Name[128] = "";
char		B_File_Name[128] = "";
double		*X;
double		*Q;
double		**Y;
double		*Th;
FILE		*Lin_Prog;
int		*Up_B;		/* table of upper bounds if they exist	*/
double		*Mu;		/* table of transition firing rates,	*/
				/* declared globally as a matter of	*/
				/* convenience. To be filled in later.	*/
int		**Pi;		/* table of Pi vectors			*/

int main(
int	argc,
char	*argv[] )
{
  static char		File_Name[128] = "";
  static char		Code_File_Name[128] = "";
  static char		Objective[256] = "";
  PetriNet 		*Reseau;
  int			i;
  boolean		Help = FALSE;
  boolean		Args_Ok = TRUE;
  boolean		File_Ok = TRUE;
  boolean		Lin_Prog_Alone = FALSE;
  boolean		Variances = TRUE;
  boolean		Check_Tight = FALSE;
  int			The_Bound;

  i = 1;

  if ( argc > 1 )
    do {

      if GET_STRING( "-f", File_Name )
      else if GET_STRING( "-o", Code_File_Name )

      else if SWITCH( "-debug", Debug, TRUE )
      else if SWITCH( "-verbose", Verbose, TRUE )
      else if SWITCH( "-check", Check_Tight, TRUE )
      else if SWITCH( "-novar", Variances, FALSE )
      else if SWITCH( "-linprog", Lin_Prog_Alone, TRUE )
      else if SWITCH( "-min", Minimum, TRUE )

      else if GET_STRING_SWITCH( "-mufile", Mu_File_Name, Other_Mu, TRUE )
      else if GET_STRING_SWITCH( "-Bfile", B_File_Name, Other_Bounds, TRUE )
      else if GET_INT_SWITCH( "-bound", The_Bound, Global_Bound, TRUE )

      else if GET_STRING( "-objective", Objective )

      else if IS_VERSION

      else if ( ( 0 == strcmp( "-h", argv[i] ) )
		|| ( 0 == strcmp( "-?", argv[i] ) ) ) {
	printf("Usage: %s [-f <file>] [-check]", TOOLNAME );
	printf(" [-o <file>] [-min] [-linprog]");
	printf("\n\t[-objective <string>] [-mufile <file>] ");
	printf("[-Bfile <file>] [-bound <int>]\n\t[-verbose] ");
	printf("[-novar] [-debug] [-h] [-?]\n");
	Args_Ok = FALSE;
	Help = TRUE;
	i++;
      }
      else
	i++; /* skip unknown argument */
    }
    while ( Args_Ok && ( i < argc ) );


  if ( Args_Ok ) {

    Reseau = Read_Network( File_Name );

    if ( Reseau != NULL ) {

	if (Debug)
	  dumpe_network(Reseau);

	if ( check_network(Reseau,The_Bound) ) {
	  if ( Lin_Prog_Alone ) {

	    if ( strlen(Code_File_Name) != 0 ) {
	      Lin_Prog = fopen( Code_File_Name, "w" );
	      File_Ok = ( NULL != Lin_Prog );
	    }
	    else {
	      Lin_Prog = stdout;
	      File_Ok = TRUE;
	    }

	    if ( File_Ok )
	      gen_lin_code( Reseau, Objective, Variances );
	    else {
	      fprintf( stderr, "%s: Error opening file %s.\n", 
		       TOOLNAME, Code_File_Name );
	      return -1;
	    }
	  }
	  else {
	    gen_solve_analyze( Reseau, Objective, Variances,
			       Check_Tight );
	  }
	  return 0;
	}
	else {
	  fprintf(stderr, "The network does not satisfy the");
	  fprintf(stderr, " right assumptions. Sorry.\n");
	  return -3;
	}
      }
      else {
	fprintf( stderr, "%s: Error reading network. No translation.\n", 
		 TOOLNAME );
	return -4;
      }
  }
  else if (!Help)
    return -1;
  else
    return 0;
}

/* checking network validity and filling of Mu[] and Up_B[] */
boolean	check_network(
PetriNet        *Reseau,
int		The_Bound )
{
  int		i;
  boolean	Is_Ok = TRUE;
  Law_Desc	The_Law;
  FILE	*Mu_Data, *B_Data;
  int		Trans_Id, Place_Id;
  double	Mu_Val;
  int		B_Val;

  /* checking places have CST(0) timing */
  for ( i=0; Is_Ok && (i<Reseau->Nb_Places); i++) {
    The_Law = Reseau->P[i].Pl_Law;
    Is_Ok = ( Is_Ok && (The_Law->Name == 'C')
	      && (The_Law->Parameters[0] == 0.0 ) );
  }

  /* checking places have inputs and outputs */
  /* to be done!! */

  if ( Is_Ok ) {
    Mu = (double*) calloc( (unsigned)Reseau->Nb_Trans, sizeof(double) );

    /* checking transitions have EXP(-) timing */
    for ( i=0; Is_Ok && (i<Reseau->Nb_Trans); i++) {
      The_Law = Reseau->T[i].Tr_Law;
      Is_Ok = ( ( Is_Ok && (The_Law->Name == 'E') )
		&& ( The_Law->Parameters[0] != 0.0 ) );
      if ( Is_Ok )
	Mu[i] = 1.0 / The_Law->Parameters[0];
    }

    /* reading overriden rates from the file */
    if ( Other_Mu && ( strlen(Mu_File_Name) != 0 ) ) {
      Mu_Data = fopen( Mu_File_Name, "r" );
      while ( EOF != fscanf( Mu_Data, "%d %lf", 
			     &Trans_Id, &Mu_Val ) )
	if ( ( 0>Trans_Id ) || ( Trans_Id<Reseau->Nb_Trans) ) {
	  Mu[Trans_Id] = Mu_Val;
	  if (Debug)
	    printf( "Mu[%d] := %f\n", Trans_Id, Mu_Val );
	}
	else {
	  printf( "%s: warning: Transition Id in mu-file",
		  TOOLNAME );
	  printf( " is not valid (%d): ignored\n", Trans_Id );
	}
    }

    if ( Global_Bound || Other_Bounds )
      Up_B = (int*) calloc( (unsigned)Reseau->Nb_Places, sizeof(int) );
    if ( Global_Bound )
      for ( i=0; i<Reseau->Nb_Places; i++ )
	Up_B[i] = The_Bound;

    if ( Other_Bounds ) {
      for ( i=0; i<Reseau->Nb_Places; i++ )
	Up_B[i] = MAXINT;

      B_Data = fopen( B_File_Name, "r" );
      while ( EOF != fscanf( B_Data, "%d %d", &Place_Id, &B_Val ) )
	if ( ( 0>Place_Id ) || ( Place_Id<Reseau->Nb_Places ) ) {
	  Up_B[Place_Id] = B_Val;
	  if ( Debug )
	    printf( "Up_B[%d] := %d\n", Place_Id, B_Val );
	}
	else {
	  printf( "%s: warning: Place Id in B-file", TOOLNAME );
	  printf( " is not valid (%d): ignored\n", Place_Id );
	}
    }
  }

  return Is_Ok;
}

/* checking if network is an event graph */
boolean	Is_An_Event_Graph(
PetriNet        *Reseau )
{
  int		i;
  boolean		Is_Ok = TRUE;

  for ( i=0; Is_Ok && (i<Reseau->Nb_Places); i++)
    {
      Is_Ok = ( Is_Ok && (Reseau->P[i].Indegree == 1 )
		&& (Reseau->P[i].Outdegree == 1 ) );
    }

  return Is_Ok;
}

void	Print_I_List( Liste_Link L )
{
  Liste_Link	L_Scan;

  SCAN( L_Scan, L )
    printf(" . %d", L_Scan->Dest );
  printf("\n");

}

/****************************************************************/
/* function 	Liste_Link	List_Inter			*/
/* computes the intersection of two arc lists while filtering	*/
/* out inhibitor arcs.						*/
/* Not a real intersection: multiple elements in L1 will be	*/
/* multiple in the result if also in L2.			*/
/* The weight of the result is the product of the original	*/
/* weights, because this is what is used in the program! 	*/
/*								*/
/* All elements of the result are created. The original lists 	*/
/* are not modified.						*/
/* Inputs:							*/
/*	L1, L2		lists to be intersected			*/
/* Result:							*/
/*	the intersection.					*/
/****************************************************************/
Liste_Link L_List_Inter( 
Liste_Link L1, 
Liste_Link L2 )
{
  Liste_Link	L_Scan;
  Liste_Link	M_Scan;
  Liste_Link	Res;
  Liste_Link	Tmp;

  Res = NULL;

  SCAN(L_Scan,L1)
    SCAN(M_Scan,L2)
    if ( ( !L_Scan->Is_Inhibitor )
	 && ( !M_Scan->Is_Inhibitor )
	 && ( L_Scan->Dest == M_Scan->Dest ) ) {
      Tmp = (Liste_Link) malloc( sizeof( struct LINK_LIST ) );
      Tmp->Dest = L_Scan->Dest;
      Tmp->Weight = L_Scan->Weight * M_Scan->Weight;
      Tmp->Data = L_Scan->Data;
      Res = Append_Link( Res, Tmp );
    }

  return Res;
}

/****************************************************************/
/* function 	Liste_Link	Filter_List			*/
/* filters out inhibitor arcs from some list of arcs		*/
/* Parameters:							*/
/*	L:		the lists to be filtered		*/
/* Returns:		the filtered list			*/
/* Notes:	No new elements are created. The ones that are	*/
/*		left out are ***NOT*** freed yet.		*/
/****************************************************************/
Liste_Link Filter_List( Liste_Link L )
{
  Liste_Link	L_Scan;
  Liste_Link	Element;
  Liste_Link	Res;

  Res = (Liste_Link)NULL;

  L_Scan = L;
  do { 
    if ( !L_Scan->Is_Inhibitor ) {
      Element = L_Scan;
      L_Scan = L_Scan->Next;
      Element->Next = Res;
      Res = Element;
    }
    else {
      L_Scan = L_Scan->Next;
    }
  } while ( L_Scan != NULL );

  return Res;
}

/****************************************************************/
/* function 	Liste_Link	I_List_Union			*/
/* computes the union (i.e. appends) two (integer) lists	*/
/* the resulting list is sorted by increasing values		*/
/* Parameters:							*/
/*	L1, L2:		the lists to be sorted			*/
/* Returns:		the merges and sorted list		*/
/****************************************************************/
Liste_Link I_List_Union( 
Liste_Link L1, 
Liste_Link L2 )
{
  Liste_Link	L_Scan, J_Scan, Res, Tmp;

  Res = NULL;

  if ( (L1==NULL) || (L2==NULL) )
    {	fprintf(stderr,"Union of NULL lists not implemented! ");
    fprintf(stderr,"NULL returned.\n");
    }
  else {
    /* creation of a guard, for simplicity */
    Res = (Liste_Link) malloc( sizeof( struct LINK_LIST ) );
    Res->Dest = -1;
    Res->Weight = 1;
    Res->Data = NULL;
    Res->Next = (Liste_Link) malloc( 
				    sizeof( struct LINK_LIST ) );
    Res->Next->Dest = L1->Dest;
    Res->Next->Weight = L1->Weight;
    Res->Next->Data = L1->Data;
    Res->Next->Next = NULL;

    SCAN(L_Scan,L1->Next)
      {	for( J_Scan=Res; 
	     (J_Scan->Next!=NULL) 
	       && (J_Scan->Next->Dest<=L_Scan->Dest);
	     J_Scan=J_Scan->Next );
      if ( ( J_Scan->Next == NULL )
	   || (J_Scan->Next->Dest != L_Scan->Dest) )
	{	Tmp = (Liste_Link) malloc(
					  sizeof( struct LINK_LIST ) );
	Tmp->Next = J_Scan->Next;
	Tmp->Weight = J_Scan->Weight;
	Tmp->Data = J_Scan->Data;
	Tmp->Dest = L_Scan->Dest;
	J_Scan->Next = Tmp;
	}
      }

    SCAN(L_Scan,L2)
      {	for( J_Scan=Res; 
	     (J_Scan->Next!=NULL) 
	       && (J_Scan->Next->Dest<=L_Scan->Dest);
	     J_Scan=J_Scan->Next );
      if ( ( J_Scan->Next == NULL )
	   || (J_Scan->Next->Dest != L_Scan->Dest) )
	{	Tmp = (Liste_Link) malloc(
					  sizeof( struct LINK_LIST ) );
	Tmp->Next = J_Scan->Next;
	Tmp->Dest = L_Scan->Dest;
	Tmp->Weight = L_Scan->Weight;
	Tmp->Data = L_Scan->Data;
	J_Scan->Next = Tmp;
	}
      }
  }

  /*	Print_I_List( Res );
   */	return Res->Next;
}

/****************************************************************/
/* Bounds from event-graph cycle conservation law		*/
/****************************************************************/
/****************************************************************/
/* function	int	*Make_Mig_Matrix			*/
/* Creates a topology matrix (see module Bf for the format)	*/
/* from a Petri Net, suited for cycle base analysis.		*/
/* Inputs:							*/
/*	Reseau		the network				*/
/* Result:							*/
/*	the topology matrix.					*/
/****************************************************************/
int	**Make_Mig_Matrix(
PetriNet	*Reseau )
{
  int		**A;
  int		i;
  Liste_Link	L_Scan;

  A = (int**) calloc( (unsigned)Reseau->Nb_Places, sizeof(int) );
  for ( i=0; i < Reseau->Nb_Places; i++)
    {	A[i] = (int*) calloc( (unsigned)Reseau->Nb_Trans, sizeof(int) );

    SCAN( L_Scan, Reseau->P[i].Inputs )
      A[i][L_Scan->Dest] = 1;
    SCAN( L_Scan, Reseau->P[i].Outputs )
      A[i][L_Scan->Dest] = -1;
    }

  return A;
}

/****************************************************************/
/* procedure		gen_cycle_Bounds			*/
/****************************************************************/
void	gen_cycle_bounds(
PetriNet*	Reseau )
{
  int	**A;
  int	**Bf;
  int	P, T;
  int	Pop;
  int	i, j, k;

  if ( Is_An_Event_Graph(Reseau) ) {
    CODE("/* Cycle population conservation Law */\n");

    P = Reseau->Nb_Places;
    T = Reseau->Nb_Trans;

    A = Make_Mig_Matrix( Reseau );

    Bf = Cycle_Base( A, P, T );

    for( i=0; i<P-T+1; i++ ) {
      Pop = 0;
      for( j=0; j<P; j++ )
	if ( Bf[i][j] == -1 ) {
	  Pop -= Reseau->P[j].Init_Pop;
	  CODE1(" -x%1d", j );
	}
	else if ( Bf[i][j] == 1 ) {
	  Pop += Reseau->P[j].Init_Pop;
	  CODE1(" +x%1d", j );
	}
      CODE1(" = %1d;\n", Pop );
      /* equalities with y */
      for ( k=0; k<Reseau->Nb_Trans; k++ ) {
	for( j=0; j<P; j++ )
	  if ( Bf[i][j] == -1 ) {
	    CODE2(" -y%1d_%1d", j, k );
	  }
	  else if ( Bf[i][j] == 1 ) {
	    CODE2(" +y%1d_%1d", j, k );
	  }
	CODE2(" = %1d q%1d;\n", Pop, k );
      }
    }
  }
  /* else... make the decomposition, and apply to each EG component */

}

/* transition throughputs */
void	gen_throughput( 
    PetriNet        *Reseau )
{
  int	i;

  CODE("/* Throughput definitions */\n");
  for ( i=0; i<Reseau->Nb_Trans; i++ ) {
    /* useful comment */
    CODE2( "/* Of transition %s (#%1d) */\n", Reseau->T[i].Name, i );
	
    CODE3( "t%d = %4.6f q%1d;\n", i, Mu[i], i );
  }
}

/* Place flow balance */
void	gen_flow_balance(
PetriNet	*Reseau )
{
  int		i;
  Liste_Link	L_Scan;

  CODE("/* Place Flow Balance Equations */\n");

  for ( i=0; i<Reseau->Nb_Places; i++ )
    if ( ( Reseau->P[i].Inputs == NULL )
	 || ( Reseau->P[i].Outputs == NULL ) ) {
      fprintf(stderr, "Warning: input/output problem at ");
      fprintf(stderr, "place %d. Place ignored.\n", i );
    }
    else {
      /* useful comment */
      CODE2( "/* At place %s (#%1d) */\n", Reseau->P[i].Name, i );

      L_Scan = Reseau->P[i].Inputs;
      CODE2("%4.6f q%1d", Mu[L_Scan->Dest]*L_Scan->Weight, 
	    L_Scan->Dest );

      SCAN(L_Scan, Reseau->P[i].Inputs->Next) {
	CODE2(" + %4.6f q%1d", Mu[L_Scan->Dest]*L_Scan->Weight,
	      L_Scan->Dest );
      }
      CODE(" = ");

      /* find first non-inhibitor arc */
      for( L_Scan = Reseau->P[i].Outputs; 
	   ( L_Scan != NULL ) && ( L_Scan->Is_Inhibitor );
	   L_Scan = L_Scan->Next );
      if ( L_Scan != NULL ) {
	CODE2("%4.6f q%1d", Mu[L_Scan->Dest]*L_Scan->Weight, 
	      L_Scan->Dest );

	SCAN(L_Scan,L_Scan->Next)
	  if ( !L_Scan->Is_Inhibitor )
	    CODE2(" + %4.6f q%1d", Mu[L_Scan->Dest]*L_Scan->Weight,
		  L_Scan->Dest );
      }
      else {
	fprintf( stderr, 
		 "%s: warning: place %d [%s] has no outputs.\n",
		 TOOLNAME, i, Reseau->P[i].Name );
	CODE( "0" );
      }
      CODE(";\n");
    }
}		

/* Square population invariance */
void	gen_square_pop(
PetriNet	*Reseau )
{
  int	i;
  Liste_Link	L_Scan;
  Liste_Link	The_List;

  CODE("/* Square of population stationarity */\n");

  for ( i=0; i<Reseau->Nb_Places; i++ )
    if ( ( Reseau->P[i].Inputs == NULL )
	 || ( Reseau->P[i].Outputs == NULL ) )
      {	fprintf(stderr, "Warning: input/output problem at ");
      fprintf(stderr, "place %d. Place ignored.\n", i );
      }
    else {	
      L_Scan = Reseau->P[i].Inputs;
      CODE3( "%4.6f y%1d_%1d", 
	     2*Mu[L_Scan->Dest]*L_Scan->Weight, i, L_Scan->Dest );
		
      SCAN(L_Scan,Reseau->P[i].Inputs->Next) {
	CODE3( " + %4.6f y%1d_%1d",
	       2*Mu[L_Scan->Dest]*L_Scan->Weight, i, L_Scan->Dest );
      }

      SCAN(L_Scan,Reseau->P[i].Inputs) {
	CODE2(" + %4.6f q%1d",
	      Mu[L_Scan->Dest]*L_Scan->Weight*L_Scan->Weight, 
	      L_Scan->Dest );
      }

      SCAN(L_Scan,Reseau->P[i].Outputs) {
	if ( !L_Scan->Is_Inhibitor ) {
	  CODE2(" + %4.6f q%1d", 
		Mu[L_Scan->Dest]*L_Scan->Weight*L_Scan->Weight, 
		L_Scan->Dest );
	}
      }
      CODE(" = ");

      /* find first non-inhibitor arc */
      for( L_Scan = Reseau->P[i].Outputs; 
	   ( L_Scan != NULL ) && ( L_Scan->Is_Inhibitor );
	   L_Scan = L_Scan->Next );
      if ( L_Scan != NULL ) {
	CODE3( "%4.6f y%1d_%1d",
	       2*Mu[L_Scan->Dest]*L_Scan->Weight, i, L_Scan->Dest );
	
	SCAN(L_Scan,L_Scan->Next) 
	  if ( !L_Scan->Is_Inhibitor )
	    CODE3( " + %4.6f y%1d_%1d",
		   2*Mu[L_Scan->Dest]*L_Scan->Weight, i, L_Scan->Dest );
      }


      The_List = L_List_Inter( Reseau->P[i].Inputs, 
			       Reseau->P[i].Outputs );
      /* Weight in the intersection are products of original weights */
      SCAN(L_Scan,The_List) {
	CODE2(" + %4.6f q%1d", 2*Mu[L_Scan->Dest]*L_Scan->Weight,
	      L_Scan->Dest );
      }

      CODE(";\n");
    }

}

/* Bounds for q and y variables */
void	gen_q_y_bounds(
PetriNet	*Reseau )
{
  int		i;
  Liste_Link	L_Scan;

  CODE("/* Sample path bounds for q */\n");

  for ( i=0; i<Reseau->Nb_Trans; i++ )
    if ( Reseau->T[i].Indegree > 0 )
      CODE1("q%1d <= 1;\n", i );
    else
      CODE1("q%1d = 1;\n", i );

  CODE("/* Sample path bounds for y */\n");

  for ( i=0; i<Reseau->Nb_Places; i++ ) {
    if ( Verbose )
      CODE2( "/* For place %s (#%1d) */\n", Reseau->P[i].Name, i );

    SCAN(L_Scan, Reseau->P[i].Outputs)
      if ( !L_Scan->Is_Inhibitor ) {
	CODE2( "y%1d_%1d >= ", i, L_Scan->Dest );
	CODE2( "%d q%1d;\n", L_Scan->Weight, L_Scan->Dest );
      }
      else if ( L_Scan->Weight == 1 )
	CODE2( "y%1d_%1d <= 0;\n", i, L_Scan->Dest );
      else {
	CODE2( "y%1d_%1d ", i, L_Scan->Dest );
	CODE2( "<= %d q%1d;\n", L_Scan->Weight-1, L_Scan->Dest );
      }
  }
}	

/* Bounds for x variables */
void	gen_x_bounds( 
PetriNet	*Reseau )
{
  int		i, j;
  Liste_Link	L_Scan;
  double	Mu_Tot;

  Mu_Tot = Mu[0];
  for ( i=1; i<Reseau->Nb_Trans; i++ )
    Mu_Tot += Mu[i];

  CODE("/* Sample path bounds for x */\n");

  for ( i=0; i<Reseau->Nb_Places; i++ ) {
    if ( Verbose )
      CODE2( "/* For place %s (#%1d) */\n", Reseau->P[i].Name, i );

    for ( j=0; j<Reseau->Nb_Trans; j++ )	    
      CODE3( "y%1d_%1d <= x%1d;\n", i, j, i );

    CODE2( "%8.4f x%1d >= ", Mu_Tot, i );
    CODE3( "%8.4f y%1d_%1d", Mu[0], i, 0 );
    for ( j=1; j<Reseau->Nb_Trans; j++ )	    
      CODE3( " + %8.4f y%1d_%1d", Mu[j], i, j );
    CODE(";\n");

    SCAN(L_Scan, Reseau->P[i].Outputs) {
      if ( ( Reseau->T[L_Scan->Dest].Indegree == 1 )
	   && ( !L_Scan->Is_Inhibitor ) ) {		
	if ( Verbose )
	  CODE2( "/* ...and transition %s (#%1d) */\n",
		 NAME(T,L_Scan->Dest), L_Scan->Dest );
	if ( L_Scan->Weight == 1 )
	  CODE3( "x%1d = y%1d_%1d;\n", i, i, L_Scan->Dest );
	else {
	  CODE3( "x%1d <= y%1d_%1d", i, i, L_Scan->Dest );
	  CODE1( " + %d;\n", L_Scan->Weight-1 );
	}
      }
    }
  }
}

/* Enabling bounds */
void Write_Enabling_RHS_1( 
    Tr_Des	The_Trans )
{
  Liste_Link	L_Scan;
  double	Som;

  Som = 1.0;

  SCAN( L_Scan, The_Trans.Inputs )
    if ( !L_Scan->Is_Inhibitor )
      Som -= Up_B[L_Scan->Dest] / (double)
	( Up_B[L_Scan->Dest] + 1 - L_Scan->Weight );

  CODE1( "%10.6f", Som );

  SCAN(L_Scan,The_Trans.Inputs)
    if ( !L_Scan->Is_Inhibitor )
      CODE2(" + %8.6f x%d", 1.0/(double)
	    ( Up_B[L_Scan->Dest] + 1 - L_Scan->Weight ),
	    L_Scan->Dest );

  SCAN(L_Scan,The_Trans.Inputs)
    if ( L_Scan->Is_Inhibitor )
      CODE2(" - %8.6f x%d",
	    1.0/(double) L_Scan->Weight, L_Scan->Dest );

  CODE(";\n");
}

void Write_Enabling_RHS_2( 
    Tr_Des	The_Trans )
{
  Liste_Link	L_Scan;
  double	Som = 0.0;

  SCAN( L_Scan, The_Trans.Inputs )
    if ( L_Scan->Is_Inhibitor )
      Som += Up_B[L_Scan->Dest] / (double)
	( Up_B[L_Scan->Dest] + 1 - L_Scan->Weight );
    
  Som = Som / The_Trans.Indegree;

  L_Scan = The_Trans.Inputs;
  if ( L_Scan->Is_Inhibitor )
    CODE2("- %10.6f x%d",
	  1.0 / ( Up_B[L_Scan->Dest] + 1 - L_Scan->Weight )
	  / The_Trans.Indegree,
	  L_Scan->Dest );
  else
    CODE2("%10.6f x%d", 
	  1.0 / L_Scan->Weight / The_Trans.Indegree,
	  L_Scan->Dest );

  SCAN( L_Scan, The_Trans.Inputs->Next ) {
    if ( L_Scan->Is_Inhibitor )
      CODE2(" - %10.6f x%d",
	    1.0 / ( Up_B[L_Scan->Dest] + 1 - L_Scan->Weight )
	    / The_Trans.Indegree,
	    L_Scan->Dest );
    else
      CODE2(" + %10.6f x%d", 
	    1.0 / L_Scan->Weight / The_Trans.Indegree,
	    L_Scan->Dest );
  }

  if ( Som != 0.0 )
    CODE1("+ %10.6f;\n", Som );
  else 
    CODE(";\n");
}


void	gen_enabling_bound(
PetriNet        *Reseau )
{
  int		i;
  boolean	Has_Bounded_Inputs;
  int		Nb_Inhibitors;
  Liste_Link	L_Scan, M_Scan;
  double	Som;

  if ( Global_Bound || Other_Bounds ) {

    CODE("/* Bounds from enabling constraints */\n");

    for ( i=0; i<Reseau->Nb_Trans; i++ )
      if ( Reseau->T[i].Indegree > 0 ) {

	Has_Bounded_Inputs = TRUE;
	SCAN(L_Scan,Reseau->T[i].Inputs)
	  Has_Bounded_Inputs = Has_Bounded_Inputs &&
	  ( Up_B[L_Scan->Dest] != MAXINT );

	if ( Has_Bounded_Inputs ) {

	  if ( Verbose )
	    CODE2( "/* Of transition %s (#%1d) */\n",
		   Reseau->T[i].Name, i );

	  CODE1("q%d >= ", i );

	  Write_Enabling_RHS_1( Reseau->T[i] );
	}

	/* second bound, with inhibitors */
	Has_Bounded_Inputs = TRUE;
	Nb_Inhibitors = 0;
	SCAN(L_Scan,Reseau->T[i].Inputs)
	  if ( L_Scan->Is_Inhibitor ) {
	    Has_Bounded_Inputs = Has_Bounded_Inputs &&
	      ( Up_B[L_Scan->Dest] != MAXINT );
	    Nb_Inhibitors++;
	  }

	if ( Has_Bounded_Inputs ) {
	  /* we have a potentially large number of contraints 	*/
	  /* to generate...					*/

	  /* here the one with only "x/pi" */
	  CODE1("q%d <= ", i );
	  Write_Enabling_RHS_2( Reseau->T[i] );

	  /* now the ones with only one "x/pi" */

	  if ( Reseau->T[i].Indegree - Nb_Inhibitors > 1 ) {
	    /* first compute constant terms */
	    Som = (double)Reseau->T[i].Indegree - Nb_Inhibitors - 1;
	    SCAN( L_Scan, Reseau->T[i].Inputs )
	      if ( L_Scan->Is_Inhibitor )
		Som += Up_B[L_Scan->Dest] / (double)
		  ( Up_B[L_Scan->Dest] + 1 - L_Scan->Weight );
	    Som = Som / Reseau->T[i].Indegree;

	    SCAN( M_Scan, Reseau->T[i].Inputs->Next )
	      if ( !M_Scan->Is_Inhibitor ) {
		CODE1("q%d <= ", i );
		CODE2("%10.6f x%d", 
		      1.0/ M_Scan->Weight / Reseau->T[i].Indegree,
		      M_Scan->Dest );

		SCAN( L_Scan, Reseau->T[i].Inputs )
		  if ( L_Scan->Is_Inhibitor )
		    CODE2(" - %10.6f x%d",
			  1.0 / ( Up_B[L_Scan->Dest] 
				  + 1 - L_Scan->Weight )
			  / Reseau->T[i].Indegree,
			  L_Scan->Dest );

		if ( Som != 0.0 )
		  CODE1(" + %10.6f;\n", Som );
		else 
		  CODE(";\n");
	      }
	  }
	}
      }
  }
}
		
/* Bounds from Little's law */
void	gen_little_law(
PetriNet	*Reseau )
{
  int		i;
  Liste_Link	L_Scan;
  double	Mu_Tot;

  CODE("/* Bounds from Little's Law */\n");

  for ( i=0; i<Reseau->Nb_Places; i++ ) {
    if ( Verbose )
      CODE2( "/* For place %s (#%1d) */\n", Reseau->P[i].Name, i );

    Mu_Tot = 0.0;
    SCAN(L_Scan, Reseau->P[i].Outputs)
      if ( !L_Scan->Is_Inhibitor )
	Mu_Tot += Mu[L_Scan->Dest];
    CODE2( "%4.6f x%d >= ", Mu_Tot, i );

    L_Scan = Reseau->P[i].Inputs;
    if ( NULL != L_Scan ) {
      CODE2("%4.6f q%d",
	    Mu[L_Scan->Dest]*L_Scan->Weight, L_Scan->Dest );

      SCAN(L_Scan, Reseau->P[i].Inputs->Next )
	CODE2(" + %4.6f q%d",
	      Mu[L_Scan->Dest]*L_Scan->Weight, L_Scan->Dest );

      CODE(";\n");
    }
    else 
      CODE("0;\n");
  }
}		

/* Liveness bound */
void	gen_liveness_bound(
PetriNet	*Reseau )
{
  int	i, j;

  CODE("/* Liveness bounds */\n");

  CODE("q0 ");
  for ( i=1; i<Reseau->Nb_Trans; i++ )
    CODE1("+ q%1d ", i );
  CODE(" >= 1;\n");

  for ( i=0; i<Reseau->Nb_Places; i++ ) {
    CODE3("x%1d <= y%1d_%1d", i, i, 0 );
    for ( j=1; j<Reseau->Nb_Trans; j++ )
      CODE2(" + y%1d_%1d", i, j );
    CODE(";\n");
  }
}		

/* Bounds from boundedness */
void	gen_boundedness(
PetriNet	*Reseau )
{
  int		i, j;
  Liste_Link	L_Scan;

  if ( Global_Bound || Other_Bounds ) {

    CODE("/* Bounds from boundedness */\n");

    for ( i=0; i<Reseau->Nb_Places; i++ )
      if ( Up_B[i] != MAXINT ) {

	if ( Verbose )
	  CODE2( "/* For place %s (#%1d) */\n", 
		 Reseau->P[i].Name, i );

	CODE2( "x%d <= %d;\n", i, Up_B[i] );

	SCAN(L_Scan, Reseau->P[i].Outputs) 
	  if ( !L_Scan->Is_Inhibitor ) {
	    if ( Verbose )
	      CODE2( "/* ...and transition %s (#%d) */\n",
		     NAME(T,L_Scan->Dest), L_Scan->Dest );
	    CODE2( "y%d_%d <= ", i, L_Scan->Dest );
	    CODE2( "%d q%d;\n", Up_B[i], L_Scan->Dest );
	  }

	for ( j=0; j<Reseau->Nb_Trans; j++ ) {
	  CODE3( "x%d <= y%d_%d + ", i, i, j );
	  CODE3( "%d - %d q%d;\n", Up_B[i], Up_B[i], j );
	  CODE2( "%f x%1d >= ", 1.0 - 1.0/Up_B[i], i );
	  CODE3( "y%d_%d - q%1d;\n", i, j, j );
	}
      }
  }
}

/* Reachable marking constraints */
void	gen_reachable_marking(
PetriNet	*Reseau )
{
  int		i;
  Liste_Link	L_Scan;

  CODE("/* Reachable marking constraints */\n");

  for ( i=0; i<Reseau->Nb_Places; i++ ) {
    if ( Verbose )
      CODE2( "/* For place %s (#%1d) */\n", Reseau->P[i].Name, i );

    CODE2( "x%d = %d", i, Reseau->P[i].Init_Pop );

    SCAN(L_Scan, Reseau->P[i].Inputs)
      if ( L_Scan->Weight == 1 )
	CODE1( " + u%d", L_Scan->Dest);
      else 
	CODE2( " + %1d u%1d", L_Scan->Weight, L_Scan->Dest);

    SCAN(L_Scan, Reseau->P[i].Outputs)
      if ( !L_Scan->Is_Inhibitor ) {
	if ( L_Scan->Weight == 1 )
	  CODE1( " - u%d", L_Scan->Dest);
	else {
	  CODE2( " - %1d u%1d", L_Scan->Weight, L_Scan->Dest);
	}
      }

    CODE(";\n");
  }
}

/* bounds from place covariances */
boolean	write_y_term( 
int		Place_Id,
Liste_Link	Trans_List,
boolean		Some_Terms )
{
  Liste_Link	L_Scan;
  boolean	Already_Some_Terms;

  if ( Some_Terms ) {
    SCAN( L_Scan, Trans_List ) 
      if ( !L_Scan->Is_Inhibitor ) {      
	CODE3( " + %4.6f y%1d_%1d",
	       Mu[L_Scan->Dest]*L_Scan->Weight, 
	       Place_Id, L_Scan->Dest );
      }
    return TRUE;
  }
  else {
    Already_Some_Terms = Some_Terms;

    SCAN(  L_Scan, Trans_List ) {
      if ( !L_Scan->Is_Inhibitor ) {
	if ( Already_Some_Terms ) {
	  CODE(" + ");
	}
	CODE3( "%4.6f y%1d_%1d", 	       
	       Mu[L_Scan->Dest]*L_Scan->Weight,
	       Place_Id, L_Scan->Dest );
	Already_Some_Terms = TRUE;
      }
    }
    return Already_Some_Terms;
  }
}

boolean	write_q_term( 
Liste_Link	Trans_List,
boolean		Some_Terms )
{
  Liste_Link	L_Scan;

  /* Note: the list Trans_List is the result of an intersection and the	*/
  /* weights are the products of original weights of the two links.	*/
  if ( Some_Terms ) {
    SCAN(L_Scan,Trans_List) {
      CODE2(" + %4.6f q%1d", Mu[L_Scan->Dest]*L_Scan->Weight, 
	    L_Scan->Dest );
    }
    return TRUE;
  }
  else 
    if ( Trans_List != NULL ) {
      L_Scan = Trans_List;

      CODE2(" + %4.6f q%1d", Mu[L_Scan->Dest]*L_Scan->Weight, L_Scan->Dest );
		
      SCAN(L_Scan,Trans_List->Next) {
	CODE2(" + %4.6f q%1d", Mu[L_Scan->Dest]*L_Scan->Weight, L_Scan->Dest );
      }
      return TRUE;
    }
    else 
      return FALSE;
}

void	gen_cov_pop(
PetriNet	*Reseau )
{
  int		i,j;
  Liste_Link	T_List;
  boolean		Some_Terms;

  CODE("/* Population covariance stationarity */\n");

  for ( i=0; i<Reseau->Nb_Places; i++ )
    for ( j=i+1; j<Reseau->Nb_Places; j++ ) {

      CODE2("/* places %d et %d */\n", i, j );
      Some_Terms = FALSE;

      Some_Terms = write_y_term( i, Reseau->P[j].Inputs,
				 Some_Terms );

      Some_Terms = write_y_term( j, Reseau->P[i].Inputs, 
				 Some_Terms );

      T_List = L_List_Inter( Reseau->P[i].Inputs,
			     Reseau->P[j].Inputs );

      Some_Terms = write_q_term( T_List, Some_Terms );

      T_List = L_List_Inter( Reseau->P[i].Outputs,
			     Reseau->P[j].Outputs );

      Some_Terms = write_q_term( T_List, Some_Terms );

      if ( !Some_Terms )
	CODE("0");

      CODE(" = ");
	    
      /* right hand side */
      Some_Terms = FALSE;

      Some_Terms = write_y_term( i, Reseau->P[j].Outputs,
				 Some_Terms );

      Some_Terms = write_y_term( j, Reseau->P[i].Outputs, 
				 Some_Terms );

      T_List = L_List_Inter( Reseau->P[i].Inputs,
			     Reseau->P[j].Outputs );

      Some_Terms = write_q_term( T_List, Some_Terms );

      T_List = L_List_Inter( Reseau->P[i].Outputs,
			     Reseau->P[j].Inputs );

      Some_Terms = write_q_term( T_List, Some_Terms );

      if ( !Some_Terms )
	CODE("0");

      CODE(";\n");

    }
}

void	gen_lin_code( 
    PetriNet	*Reseau,
    char	*Objective,
    boolean	Variances )
{
  int	i;

  CODE("/* Linear program computing an upper bound on the throughput");
  CODE(" of a Petri Net */\n");
  CODE2("/* Generated by %s, %s */\n", TOOLNAME, todays_date() );

  if ( Minimum ) {
    CODE("min:");
    if ( strlen( Objective ) > 0 )
      CODE1("%s;\n", Objective);
    else {

      /*		CODE("th;\n");
			for ( i=0; i<Reseau->Nb_Trans; i++ )
			CODE1("t%d <= th;\n", i );
			*/
      CODE( "t0" );
      for ( i=1; i<Reseau->Nb_Trans; i++ )
	CODE1(" + t%d", i );
      CODE(";\n");
    }		
  }
  else {
    CODE("max:");
    if ( strlen( Objective ) > 0 )
      CODE1("%s;\n", Objective);
    else {

      /*		CODE("th;\n");
			for ( i=0; i<Reseau->Nb_Trans; i++ )
			CODE1("t%d >= th;\n", i );
			*/
      CODE( "t0" );
      for ( i=1; i<Reseau->Nb_Trans; i++ )
	CODE1(" + t%d", i );
      CODE(";\n");
    }
  }

  gen_throughput(Reseau);
  gen_flow_balance(Reseau);		/* (1) +w +i */
  if ( Variances ) {
    if ( Reseau->Has_Weights ) {
      fprintf( stderr, "%s: warning: weights not compatible", TOOLNAME );
      fprintf( stderr, "with second moment constraints (yet).\n" );
      fprintf( stderr, "Results probably incorrects.\n");
    }
    gen_square_pop(Reseau); 	/* (2) + ~w */
    gen_cov_pop(Reseau);		/* (3) */
  }
  gen_liveness_bound(Reseau);		/* (11) */
  gen_boundedness(Reseau);		/* (9) +i */
  gen_reachable_marking(Reseau);	/* +w +i */
  gen_q_y_bounds(Reseau);		/* (4) and (5) +w +i */
  gen_x_bounds(Reseau);		/* (6) +w +i */
  gen_enabling_bound(Reseau);		/* (12) +w +i */
  gen_little_law(Reseau);		/* +w +i */
  gen_cycle_bounds(Reseau);		/* (9) */
}

double parse_lin_results(
int	Nb_Trans,
int	Nb_Places )
{
  FILE	*Data;
  double	Val;
  double	Thput;
  char	Variable;
  int	I_Trans, I_Place;
  int	Status;
  int	i;

  if ( Debug )
    printf("Parsing the results of the linear program.\n");

  Data = fopen( RES_FILE_NAME, "r" );

  Status = fscanf( Data, "%*[A-z: ]%lf", &Thput );
  switch( Status ) {
  case EOF:
    printf("No solution found in by the linear program.\n");
    Thput = -1.0;
    break;
  case 0:
    printf("Error reading linear program results.\n");
    Thput = -1.0;
    break;	    
  default:
    X = (double*) calloc( (unsigned)Nb_Places, sizeof(double) );
    Q = (double*) calloc( (unsigned)Nb_Trans, sizeof(double) );
    Th = (double*) calloc( (unsigned)Nb_Trans, sizeof(double) );
    Y = (double**) calloc( (unsigned)Nb_Places, sizeof(double*) );
    for ( i=0; i<Nb_Places; i++ )
      Y[i] = (double*) calloc( (unsigned)Nb_Trans, sizeof(double) );

    while( EOF != fscanf( Data, "%c", &Variable ) )
      {
	switch( Variable ) {
	case 'x':
	  fscanf( Data, "%d %lf\n", &I_Trans, &Val );
	  X[I_Trans] = Val;
	  if ( Debug )
	    printf("Val of X[%d] = %f\n", I_Trans, Val );
	  break;
	case 'q':
	  fscanf( Data, "%d %lf\n", &I_Trans, &Val );
	  Q[I_Trans] = Val;
	  if ( Debug )
	    printf("Val of Q[%d] = %f\n", I_Trans, Val );
	  break;
	case 't':   /* il faut regler le probleme de "th" */
	  if ( 2 == fscanf( Data, "%d %lf\n", 
			    &I_Trans, &Val ) )
	    {   Th[I_Trans] = Val;
	    if ( Debug )
	      printf("Val of Th[%d] = %f\n", I_Trans, Val );
	    }
	  break;
	case 'y':
	  fscanf( Data, "%d_%d %lf\n", &I_Place, &I_Trans, 
		  &Val );
	  Y[I_Place][I_Trans] = Val;
	  if ( Debug )
	    printf("Val of Y[%d][%d] = %f\n", 
		   I_Place, I_Trans, Val );
	  break;
	}
      }
  }

  unlink( RES_FILE_NAME );

  return Thput;

}

void	check_th_bounds( 
PetriNet        *Reseau,
double	Throughput )
{
  int	i;

  printf("Checking throughput bounds\n");

  for ( i=0; i<Reseau->Nb_Trans; i++ )
    if( CLOSE( Throughput, Th[i] ) )
      printf("t%d = th\n", i );
}

void	check_liveness_bound( 
PetriNet        *Reseau )
{
  double	Q_Tot = 0.0;
  double	Y_Tot;
  int		i,j;

  printf("Checking liveness bounds\n");

  for ( i=0; i<Reseau->Nb_Trans; i++ )
    Q_Tot += Q[i];

  if ( CLOSE( Q_Tot, 1.0 ) ) {
    printf("\nq0 ");
    for ( i=1; i<Reseau->Nb_Trans; i++ )
      printf("+ q%1d ", i );
    printf(" = 1;\n");
  }

  for ( i=0; i<Reseau->Nb_Places; i++ ) {
    Y_Tot = Y[i][0];
    for ( j=1; j<Reseau->Nb_Trans; j++ )
      Y_Tot += Y[i][j];

    if ( CLOSE( Y_Tot, X[i] ) ) {
      printf("x%1d = y%1d_%1d", i, i, 0 );
      for ( j=1; j<Reseau->Nb_Trans; j++ )
	printf(" + y%1d_%1d", i, j );
      printf(";\n");
    }
  }
}

void	check_boundedness(
PetriNet	*Reseau )
{
  int		i, j;
  Liste_Link	L_Scan;

  if ( Global_Bound || Other_Bounds ) {

    printf("Checking boundedness bounds\n");

    for ( i=0; i<Reseau->Nb_Places; i++ )
      if ( Up_B[i] != MAXINT ) {

	if ( CLOSE( X[i], Up_B[i] ) )
	  printf( "x%d = %d;\n", i, Up_B[i] );

	SCAN(L_Scan, Reseau->P[i].Outputs)
	  if ( !L_Scan->Is_Inhibitor ) {
	    if ( CLOSE( Y[i][L_Scan->Dest], 
			Up_B[i]*Q[L_Scan->Dest] ) ) {
	      printf( "y%d_%d <= ", i, L_Scan->Dest );
	      printf( "%d q%d;\n", Up_B[i], L_Scan->Dest );
	    }
	  }

	for ( j=0; j<Reseau->Nb_Trans; j++ ) {
	  if ( CLOSE( X[i], Y[i][j] + Up_B[i] * ( 1 - Q[j] ) ) ) {
	    printf( "x%d = y%d_%d + ", i, j, j );
	    printf( "%d ( 1 - q%d );\n", Up_B[i], j );
	  }
	  if ( CLOSE( ( 1.0 - 1.0/Up_B[i] )*X[i], 
		      Y[i][j] - Q[j] ) ) {
	    printf( "(1-1/%d) x%1d = ", Up_B[i], i );
	    printf( "y%d_%d - q%1d;\n", i, j, j );
	  }
	}
      }
  }
}

void    check_sample_path( 
PetriNet        *Reseau )
{
  int		i,j;
  Liste_Link	L_Scan;
  double	Mu_Tot;
  double	RHS;

  Mu_Tot = Mu[0];
  for ( i=1; i<Reseau->Nb_Trans; i++ )
    Mu_Tot += Mu[i];

  printf("Checking sample path bounds\n");

  for ( i=0; i<Reseau->Nb_Trans; i++ )
    if ( CLOSE( Q[i], 1.0 ) )
      printf("q%d = 1;\n", i );

  for ( i=0; i<Reseau->Nb_Places; i++ ) {
    for ( j=0; j<Reseau->Nb_Trans; j++ )
      if ( CLOSE( Y[i][j], X[i] ) )
	printf("y%1d_%1d = x%1d;\n", i, j, i );
	
    RHS = Mu[0] * Y[i][0];
    for ( j=1; j<Reseau->Nb_Trans; j++ )	    
      RHS += Mu[j] * Y[i][j];
    if ( CLOSE( Mu_Tot * X[i], RHS ) ) {
      printf( "%8.4f x%1d = ", Mu_Tot, i );
      printf( "%8.4f y%1d_%1d", Mu[0], i, 0 );
      for ( j=1; j<Reseau->Nb_Trans; j++ )	    
	printf( " + %8.4f y%1d_%1d", Mu[j], i, j );
      CODE(";\n");
    }

    SCAN(L_Scan, Reseau->P[i].Outputs) {
      if ( ( Reseau->T[L_Scan->Dest].Indegree == 1 )
	   && ( !L_Scan->Is_Inhibitor )
	   && ( CLOSE( X[i], Y[i][j] + L_Scan->Weight - 1 ) ) ) {
	if ( L_Scan->Weight == 1 )
	  printf( "x%1d = y%1d_%1d;\n", i, i, L_Scan->Dest );
	else {
	  printf( "x%1d = y%1d_%1d", i, i, L_Scan->Dest );
	  printf( " + %d;\n", L_Scan->Weight-1 );
	}
      }
    }
  }

  for ( i=0; i<Reseau->Nb_Places; i++ )
    SCAN(L_Scan, Reseau->P[i].Outputs) {
    j = L_Scan->Dest;
    if ( !L_Scan->Is_Inhibitor ) {
      if ( CLOSE( Y[i][j], Q[j]*L_Scan->Weight ) )
	printf("y%d_%d = %d q%d\n", i, j, L_Scan->Weight, j );
    }
    else if ( CLOSE( Y[i][j], Q[j]*(L_Scan->Weight)-1 ) )
      printf("y%d_%d = %d q%d\n", i, j, L_Scan->Weight-1, j );
  }
}

void	check_enabling_bound(
PetriNet        *Reseau )
{
  int		i;
  boolean	Has_Bounded_Inputs;
  Liste_Link	L_Scan, M_Scan;
  double	Som;
  double	Right_Hand;
  Tr_Des	The_Trans;
  int		Nb_Inhibitors;

  printf("Checking enabling bounds\n");

  /* to be able to use the functions Write_Enabling_* */
  Lin_Prog = stdout;

  for ( i=0; i<Reseau->Nb_Trans; i++ ) {

    The_Trans = Reseau->T[i];

    Has_Bounded_Inputs = TRUE;
    SCAN(L_Scan,The_Trans.Inputs)
      Has_Bounded_Inputs = Has_Bounded_Inputs &&
      ( Up_B[L_Scan->Dest] != MAXINT );

    if ( Has_Bounded_Inputs ) {

      Right_Hand = 1.0;

      SCAN(L_Scan,The_Trans.Inputs)
	if ( !L_Scan->Is_Inhibitor )
	  Right_Hand -= (Up_B[L_Scan->Dest]- X[L_Scan->Dest])
	    /(double)( Up_B[L_Scan->Dest] + 1 - 
		       L_Scan->Weight );
	else
	  Right_Hand -= X[L_Scan->Dest]/L_Scan->Weight;

      if ( CLOSE( Right_Hand, Q[i] ) ) {

	printf("q%d = ", i );
	Write_Enabling_RHS_1( The_Trans );
      }
    }

    Has_Bounded_Inputs = TRUE;
    Nb_Inhibitors = 0;
    SCAN(L_Scan,The_Trans.Inputs)
      if ( L_Scan->Is_Inhibitor ) {
	Has_Bounded_Inputs = Has_Bounded_Inputs &&
	  ( Up_B[L_Scan->Dest] != MAXINT );
	Nb_Inhibitors++;
      }

    if ( Has_Bounded_Inputs ) {

      Right_Hand = 0.0;

      SCAN(L_Scan,The_Trans.Inputs)
	if ( !L_Scan->Is_Inhibitor )
	  Right_Hand += X[L_Scan->Dest] / L_Scan->Weight;
	else
	  Right_Hand += (Up_B[L_Scan->Dest]- X[L_Scan->Dest])
	    /(double)( Up_B[L_Scan->Dest] + 1 - 
		       L_Scan->Weight );

      Right_Hand /= The_Trans.Indegree;

      if ( CLOSE( Right_Hand, Q[i] ) ) {
	printf("q%d = ", i );
	Write_Enabling_RHS_2( The_Trans );
      }

      if ( The_Trans.Indegree - Nb_Inhibitors > 1 ) {
	/* first compute constant terms */

	Som = (double)The_Trans.Indegree - Nb_Inhibitors - 1;
	SCAN( L_Scan, The_Trans.Inputs )
	  if ( L_Scan->Is_Inhibitor )
	    Som += Up_B[L_Scan->Dest] / (double)
	      ( Up_B[L_Scan->Dest] + 1 - L_Scan->Weight );
	Som = Som / The_Trans.Indegree;

	SCAN( M_Scan, The_Trans.Inputs->Next )
	  if ( !M_Scan->Is_Inhibitor ) {
	    Right_Hand = X[M_Scan->Dest] / M_Scan->Weight;

	    SCAN( L_Scan, The_Trans.Inputs )
	      if ( L_Scan->Is_Inhibitor )
		Right_Hand -= 
		  X[L_Scan->Dest]
		  / ( Up_B[L_Scan->Dest] + 1 
		      - L_Scan->Weight );

	    Right_Hand = Som + Right_Hand / The_Trans.Indegree;

	    if ( CLOSE( Q[i], Right_Hand ) ) {
	      printf("q%d = ", i );
				
	      printf("%10.6f x%d", 
		     1.0/ M_Scan->Weight / Reseau->T[i].Indegree,
		     M_Scan->Dest );

	      SCAN( L_Scan, Reseau->T[i].Inputs )
		if ( L_Scan->Is_Inhibitor )
		  printf("- %10.6f x%d",
			 1.0 / ( Up_B[L_Scan->Dest] 
				 + 1 - L_Scan->Weight )
			 / Reseau->T[i].Indegree,
			 L_Scan->Dest );

	      if ( Som != 0.0 )
		printf("+ %10.6f;\n", Som );
	      else 
		printf(";\n");
	    }
	  }
      }
    }
  }
}
		
void	check_little_law(
PetriNet	*Reseau )
{
  int		i;
  Liste_Link	L_Scan;
  double	Mu_Tot;
  double	Right_Hand, Left_Hand;

  printf("Checking Little's Law\n");

  for ( i=0; i<Reseau->Nb_Places; i++ ) {
    Mu_Tot = 0.0;
    SCAN(L_Scan, Reseau->P[i].Outputs)
      if ( !L_Scan->Is_Inhibitor )
	Mu_Tot += Mu[L_Scan->Dest];

    Right_Hand = Mu_Tot * X[i];

    Left_Hand = 0.0;
    SCAN(L_Scan, Reseau->P[i].Inputs )
      Left_Hand += Mu[L_Scan->Dest] * Q[L_Scan->Dest] * L_Scan->Weight;

    if ( CLOSE( Left_Hand, Right_Hand ) ) {
      printf( "%4.6f x%d = ", Mu_Tot, i );
      L_Scan = Reseau->P[i].Inputs;
      if ( NULL != L_Scan ) {
	printf("%4.6f q%d",
	       Mu[L_Scan->Dest]*L_Scan->Weight, L_Scan->Dest );

	SCAN(L_Scan, Reseau->P[i].Inputs->Next )
	  printf(" + %4.6f q%d",
		 Mu[L_Scan->Dest]*L_Scan->Weight, L_Scan->Dest );

	printf(";\n");
      }
      else 
	printf("0;\n");
    }
  }
}	

void	gen_solve_analyze( 
    PetriNet	*Reseau,
    char	*Objective,
    boolean	Variances,
    boolean	Check )
{
  char	command[512];
  double	Value;

  Lin_Prog = fopen( LIN_PROG_NAME, "w" );

  if (Debug)
    printf("Generating Linear Program\n");

  gen_lin_code( Reseau, Objective, Variances );

  fclose( Lin_Prog );

  sprintf( command, "lp_solve < %s > %s 2>/dev/null", 
	   LIN_PROG_NAME, RES_FILE_NAME );
  system( command );

  unlink( LIN_PROG_NAME );

  if ( 0 <= ( Value = 
	      parse_lin_results( Reseau->Nb_Trans, Reseau->Nb_Places ) ) )
    {
      printf("Value found for ");
      if ( strlen( Objective ) > 0 )
	printf("%s", Objective );
      else
	printf("th");
      printf(": %4.6f\n", Value );

      if ( Check )
	{
	  if ( strlen( Objective ) == 0 )
	    check_th_bounds( Reseau, Value );
	  check_liveness_bound( Reseau );
	  check_boundedness( Reseau );		/* +i */
	  check_sample_path( Reseau );		/* +w +i */
	  if ( Global_Bound || Other_Bounds )
	    check_enabling_bound( Reseau );	/* +w +i */
	  check_little_law( Reseau ); 		/* +w +1 */
	}

    }
}
