/* Translation program from eg language to QNAP language.	  */

#include <math.h>
#include "pn.h"

#define	TOOLNAME	"pn2qnap"

#define MAX(x,y)	( x>y ? x : y )

#define SCAN(x,y)       for(x=y;x!=NULL;x=x->Next)

extern PetriNet	*Read_Network(char *File_Name);
extern void	dumpe_network(PetriNet *RdP);
extern char	*todays_date();

/* forward declarations */
boolean	satisfies_constraints(PetriNet *RdP);
void	dump_qnap_code(PetriNet *RdP, int Max_Time);
int	trans_in_link_number(Tr_Des *The_Trans, Pl_Des *The_Place);

boolean		*Place_Exists;
int		*Trans_Loop_Of;	/* stores the id of the trans a place	*/
				/* recycles (if it does).		*/
boolean		Optimize = FALSE;
int		Opt_Level;
float   	Trace_Start = 0.0;      /* trace control parameters     */
float   	Trace_End = 100.0;      /* with their initial values    */
boolean		Trace = FALSE;
char		Trace_File_Name[128] = "trace.results";
boolean		Counters = FALSE;


int main(int argc, char *argv[])
{
  static char		File_Name[128] = "";
  static char		Code_File_Name[128] = "";
  PetriNet 		*Reseau;
  int			i;
  int			Max_Time = 1000;
  boolean		Help = FALSE;
  boolean		Args_Ok = TRUE;
  boolean		File_Ok = TRUE;

  i = 1;

  if (argc>1)
    do {

      if GET_STRING( "-f", File_Name )
      else if GET_STRING( "-o", Code_File_Name )

      else if SWITCH( "-debug", Debug, TRUE )
      else if SWITCH( "-counters", Counters, TRUE )
      else if GET_INT( "-tmax", Max_Time )

      else if IS_VERSION

      else if ( ( 0 == strcmp( "-h", argv[i] ) )
		|| ( 0 == strcmp( "-?", argv[i] ) ) ) {
		printf("Usage: %s [-f <file>] ", TOOLNAME);
	printf("[-O <level>] [-debug] [-o <file>] \n");
	printf("\t[-tb <time>] [-te <time>] [-tfile <file> ] ");
	printf("[-tmax <time>]\n\t[-counters] [-h] [?]\n");
	Args_Ok = FALSE;
	Help = TRUE;
	i++;
      }

      else if ( 0 == strcmp( "-O", argv[i] ) ) {
	i++;
	if ( i < argc ) {
	  Optimize = TRUE;
	  sscanf( argv[i], "%d", &Opt_Level );
	  Trace = FALSE;
	  i++;
	}
	else {  
	  printf("%s: number expected after %s\n", argv[0], argv[i]);
	Args_Ok = FALSE;
	}
      }
      else if ( 0 == strcmp( "-tb", argv[i] ) ) {
	i++;
	if ( ( i < argc ) 
	     && ( 0 != sscanf( argv[i], "%f", &Trace_Start ) ) ) {
	  i++;
	  Trace = TRUE;
	  Optimize = FALSE;
	}
	else {  
	  printf("%s: float expected after -tb\n",argv[0]);
	  Args_Ok = FALSE;
	}
      }
      else if ( 0 == strcmp( "-te", argv[i] ) ) {
	i++;
	if ( ( i < argc ) 
	     && ( 0 != sscanf( argv[i], "%f", &Trace_End ) ) ) {
	  i++;
	  Trace = TRUE;
	  Optimize = FALSE;
	}
	else {  
	  printf("%s: float expected after -te\n",argv[0]);
	  Args_Ok = FALSE;
	}
      }

      else if GET_STRING( "-tfile", Trace_File_Name )

      else if IS_VERSION

      else
	i++; /* skip unknown argument */
    }
    while ( Args_Ok && ( i < argc ) );


  if ( Args_Ok ) {

    Reseau = Read_Network( File_Name );

    if ( Reseau != NULL ) {

	if ( satisfies_constraints( Reseau ) ) {

	  if ( strlen(Code_File_Name) != 0 )
	    File_Ok = ( NULL != freopen( Code_File_Name, "w", stdout ) );

	  if ( File_Ok ) {
	    dump_qnap_code(Reseau, Max_Time);
	    return 0;
	  }
	  else {
	    fprintf( stderr, "%s: Error opening file %s.\n", 
		     TOOLNAME, Code_File_Name );
	    return -1;
	  }
	}
	else {
	  fprintf( stderr, "%s: the network does not ", TOOLNAME );	
	  fprintf( stderr, "satisfy the constraints. No translation.\n" );
	  return -2;
	}
      }
      else {
	fprintf( stderr, "%s: Error reading network. No translation.\n", 
		 TOOLNAME );
	return -1;
      }
    }

  else if (!Help)
    return -1;
  else
    return 0;
}

/************************************************************************/
/* Procedures de test.							*/
/************************************************************************/
/* test if a place is a switch, that is, has a TRUE Cyclic routing,	*/
/* that is, with a period >= 2.						*/
boolean	is_switch_place(Pl_Des	*The_Place)
{
  return ( (The_Place->Routing !=NULL ) 
	   && (The_Place->Routing->Type == 'D' )
	   && (The_Place->Routing->Period >= 2 ) );
}

/* test if a law is CST(0.0)						*/
boolean	is_null_time(Law_Desc	The_Law)
{
  return ( The_Law->Name == 'C' ) && ( The_Law->Parameters[0] == 0.0 );
}

/* calcul du plus long cycle. Retourne 0 si pas de cycles.		*/
int	max_cycle_length(PetriNet	*Reseau)
{
  int	i;
  int	The_Max = 0;

  for ( i=0; i<Reseau->Nb_Places; i++ )
    if ( is_switch_place( &(Reseau->P[i]) ) )
      The_Max = MAX( The_Max, Reseau->P[i].Routing->Period );
  
  return The_Max;
}

/* test si une transition est recyclee...				*/
boolean	is_recycled(
PetriNet        *Reseau,
Tr_Des		*The_Trans)
{
  Liste_Link	L_Scan;
  boolean	Is_Recycled = FALSE;
  Pl_Des	*The_Place;

  SCAN( L_Scan, The_Trans->Outputs ) {
    The_Place = &(Reseau->P[L_Scan->Dest]);
    Is_Recycled = Is_Recycled ||
      ( ( The_Place->Outdegree == 1 )
	&& ( The_Place->Outputs->Dest == The_Trans->Tr_Id )
	&& ( is_null_time( The_Place->Pl_Law ) ) );
    }

  return Is_Recycled;
}

/* test si une place est une place de recyclage.			*/
boolean	is_recycling_place(Pl_Des *The_Place)
{
  boolean	Is_Recycling = FALSE;

  Is_Recycling = ( ( The_Place->Indegree == 1 ) 
		   && ( The_Place->Outdegree == 1 )
		   && ( The_Place->Inputs->Dest 
			== The_Place->Outputs->Dest )
		   && ( The_Place->Inputs->Weight == 1 )
		   && ( The_Place->Outputs->Weight == 1 )
		   && is_null_time( The_Place->Pl_Law ) );

  return Is_Recycling;
}
/************************************************************************/
/* Procedures d'impression.						*/
/************************************************************************/
/* Procedure to write the law. Adapted from the same from gen_eg.c */
void Output_Law_QNAP( Law_Desc The_Law, int Obj_Id )
{
  double	Alpha;
  double	Beta;
  double	Delta;
  double	Lambda;
  double	s1;
  double	s2;

  switch( The_Law->Name ) {
  case 'E':
    printf("EXP( %8.4f );\n", The_Law->Parameters[0] );
    break;
  case 'C':
    printf("CST( %8.4f );\n", The_Law->Parameters[0] );
    break;
  case 'B':
    printf("IF RANDU > %8.6f THEN CST(0.0) ELSE CST(1.0);\n", 
	   The_Law->Parameters[0] );
    break;
  case 'g':
    printf("CST( FIX( EXP( -1/LOG(%8.4f) ) ) );\n", 
	   The_Law->Parameters[0] );
    break;
  case 'G':
    printf("CST( 1.0 + FIX( EXP( -1/LOG(%8.4f) ) ) );\n", 
	   The_Law->Parameters[0] );
    break;
  case 'U':
    printf("UNIFORM( %8.4f, %8.4f );\n",
	   The_Law->Parameters[0],
	   The_Law->Parameters[1] );
    break;
  case 'R':
    printf("ERLANG( %8.4f, %4.0f );\n",
	   The_Law->Parameters[1],
	   The_Law->Parameters[0] );
    break;
  case 'O':
    Alpha = 1.0 / The_Law->Parameters[1];
    Beta = 1.0 / The_Law->Parameters[2];
    Lambda = (1.0+Alpha/Beta) / The_Law->Parameters[0];
    Delta = sqrt( Alpha*Alpha + 2.0*Alpha*(Beta+Lambda)
		  + (Beta-Lambda)*(Beta-Lambda) );
    s1 = -( Alpha+Beta+Lambda+Delta )/2.0;
    s2 = -( Alpha+Beta+Lambda-Delta )/2.0;
    printf("IF RANDU < %8.6f THEN EXP( %8.4f ) ELSE EXP( %8.4f );\n",
	   (-Lambda-s2)/(s1-s2), -1.0/s1, -1.0/s2 );
    break;
   case 'H':
    printf("HEXP( %8.4f, %8.4f );\n",
	   The_Law->Parameters[0],
	   The_Law->Parameters[1] );
    break;
  case 'D':
    fprintf( stderr, "%s: error: can't handle timing law (%c)\n",
	     TOOLNAME, The_Law->Name );
    break;
  case 'P':
    /* this part assumes implicitly that the object is a transition	*/
    printf("BEGIN\n");
    printf("        TR_%d.NextArr := EXP( %8.4f );\n",
	   Obj_Id, The_Law->Parameters[0] );
    printf("        WHILE ( TR_%d.NextArr > TR_%d.NextChg )\n",
	   Obj_Id, Obj_Id );
    printf("        DO BEGIN\n");
    printf("            TR_%d.NextArr := TR_%d.NextArr - TR_%d.NextChg;\n",
	   Obj_Id, Obj_Id, Obj_Id );
    printf("            CST( TR_%d.NextChg );\n", Obj_Id );
    printf("            ");
    Output_Law_QNAP( The_Law->Law_Parameters[1], Obj_Id );
    printf("            TR_%d.NextChg := ", Obj_Id );
    Output_Law_QNAP( The_Law->Law_Parameters[0], Obj_Id );
    printf("        END;\n" );
    printf("        CST( TR_%d.NextArr );\n", Obj_Id );
    printf("        TR_%d.NextChg := TR_%d.NextChg - TR_%d.NextArr;\n", 
	   Obj_Id, Obj_Id, Obj_Id );
    printf("    END;\n" );
    break;
  default:
    fprintf( stderr, "%s: error: unknown timing law (%c)\n",
	     TOOLNAME, The_Law->Name );
  }
}

/* procedure to write the initial population. Improved from gen_eg.c */
/* Initial lags are not taken care of... this is not trivial in QNAP.*/
/* Best thing would be to have them generated by a private source.   */
void fprint_init_mark( 
		      PetriNet	*Reseau,
		      int	The_Orig)	/* a place Id */
{
  int		The_Tr_Id;
  Pl_Des	*The_Place;

  The_Place = &(Reseau->P[The_Orig] );

  if (The_Place->Init_Pop>0) {
    printf("    INIT(");

    if (Optimize && !Place_Exists[The_Orig]) {
      The_Tr_Id = Reseau->P[The_Orig].Outputs->Dest;
      if ( Reseau->T[The_Tr_Id].Indegree == 1 )
	printf("No_Class");
      else {
	printf("CL_%d", 
	       trans_in_link_number( &(Reseau->T[The_Tr_Id]),
				     &(Reseau->P[The_Orig]) ) );
      }
    }
    else
      printf("No_Class");

    printf(") = %d;\n", The_Place->Init_Pop );
    printf("\n");
  }
}

/* procedure to write what is the actual destination of customers going	*/
/* to "place" The_Orig in the network.					*/
void fprint_dest(PetriNet	*Reseau,
		 int		The_Orig )
{
  int	The_Tr_Id;

  if (Optimize && !Place_Exists[The_Orig]) {
    The_Tr_Id = Reseau->P[The_Orig].Outputs->Dest;
    printf("        TR_%d,", The_Tr_Id );
  }
  else
    printf("        PL_%d,", The_Orig );

}

/* procedure to write what is the actual class of customers going	*/
/* to "place" The_Orig in the network.					*/
void fprint_class(PetriNet	*Reseau,
		  int		The_Orig )
{
  int	The_Tr_Id;
  Tr_Des	*The_Trans;

  if (Optimize && !Place_Exists[The_Orig]) {
    The_Tr_Id = Reseau->P[The_Orig].Outputs->Dest;
    The_Trans = &(Reseau->T[The_Tr_Id]);
    if ( ( The_Trans->Indegree == 1 )
	 || ( ( The_Trans->Indegree == 2 ) && 
	      is_recycled( Reseau, The_Trans ) ) )
      printf(" No_Class" );
    else {
      printf(" CL_%d", trans_in_link_number( 
					    &(Reseau->T[The_Tr_Id]),
					    &(Reseau->P[The_Orig]) ) );
    }
  }
  else 	printf(" No_Class");

}

/* procedure to write what is the actual destination of customers going	*/
/* to transition The_Dest in the network.				*/
void fprint_trans_dest(int The_Dest)
{
  printf("        TR_%d,", The_Dest );

}

/* procedure to write what is the actual class of customers going from	*/
/* place The_Orig to transition The_Dest in the network.		*/
void fprint_trans_class(Pl_Des	*The_Orig,
			Tr_Des	*The_Dest )
{
  if ( The_Dest->Indegree == 1 )
    printf(" No_Class" );
  else	printf(" CL_%d", trans_in_link_number( 
					      The_Dest, The_Orig ) );
}

/* procedure to write the code corresponding to some transition.	*/
void fprint_transition(PetriNet	*Reseau,
		       int	Tr_Id,
		       int	Max_Time )
{
  Tr_Des	*The_Trans;
  Pl_Des	*The_Place;
  int		The_Dest;
  int		The_Number;
  int		Nb_Neighb;
  Liste_Link	L_Scan;
  Liste_Link	The_Link;
  boolean	Loop_Is_Last;

  The_Trans = &(Reseau->T[Tr_Id]);

  /* helpful comment... */
  printf( "&\n& transition %s\n", 
	  ( NULL == The_Trans->Name ? "<no name>" : The_Trans->Name ) );
  printf("/STATION/\n");
  printf("    NAME = TR_%d;\n", Tr_Id);

  Nb_Neighb = The_Trans->Indegree;
  if ( is_recycled( Reseau, The_Trans ) )
    Nb_Neighb--;

  switch (Nb_Neighb) {
  case 0: /* Must check if it is a source! */
    if ( is_recycled( Reseau, The_Trans ) )
      printf("    TYPE = SOURCE;\n");
    else 
      /* a source that is not recycled. This is not 	*/
      /* possible: the conditions are violated.	*/
      fprintf( stderr, "%s: Error: bad transition %d.\n",
	       TOOLNAME, Tr_Id );
    break;
  default:	/* have to make a FUSION */
    if ( is_recycled( Reseau, The_Trans ) )
      printf("    TYPE = SERVER, SINGLE;\n");
    else
      printf("    TYPE = SERVER, INFINITE;\n");
  }

  if ( Nb_Neighb > 1 ) {

    printf("    FUSION = (\n");
    SCAN( L_Scan, The_Trans->Inputs )
      Loop_Is_Last = 
      ( (L_Scan->Next==NULL) 
	&& is_recycling_place( &(Reseau->P[L_Scan->Dest]) ) );

    SCAN( L_Scan, The_Trans->Inputs ) {
      The_Place = &(Reseau->P[L_Scan->Dest]);
      if ( !is_recycling_place( The_Place ) ) {
	The_Number = trans_in_link_number( The_Trans, 
					   The_Place );
	if ( ( L_Scan->Next != NULL )
	     && !( ( L_Scan->Next->Next == NULL ) && Loop_Is_Last ) )
	  printf("        CL_%d, 1,\n", The_Number );
	else
	  printf("        CL_%d, 1 ) No_Class;\n", 
		 The_Number );
      }
    }
  }
  else if ( Nb_Neighb == 1 ) {
    /* check if input link has a weight 				*/
    /* find the input place. Beware of the recycling place 	*/
    The_Place = &(Reseau->P[The_Trans->Inputs->Dest]);
    if ( is_recycling_place( The_Place ) ) {
      The_Link = The_Trans->Inputs->Next;
      The_Place = &(Reseau->P[The_Link->Dest]);
    }
    else
      The_Link = The_Trans->Inputs;

    if ( The_Link->Weight > 1 ) {
      printf("    FUSION = (");
/*       The_Number = trans_in_link_number( The_Trans, The_Place ); */
/*       printf("        CL_%d, %1d ) No_Class;\n",  */
/* 	     The_Number, The_Link->Weight ); */
      printf(" No_Class, %1d ) No_Class;\n", The_Link->Weight );
    }    
  }
  /* else: Nb_Neighb == 0 ==> it is a source */

  /* ecriture du service */
  /* dans le cas "compteurs", la transition numero 0 dicte la fin */
  printf("    SERVICE = ");
  if ( ( Counters ) && ( Tr_Id == 0 ) ) {
    printf("BEGIN\n");
    printf("      IF SERVNB( TR_0 ) = %d THEN STOP;\n", Max_Time );
    printf("      ");
    Output_Law_QNAP( The_Trans->Tr_Law, The_Trans->Tr_Id );
    printf("    END;\n");
  }
  else
    Output_Law_QNAP( The_Trans->Tr_Law, The_Trans->Tr_Id );

  /* ecriture du transit */
  Nb_Neighb = The_Trans->Outdegree;
  if ( is_recycled( Reseau, The_Trans ) )
    Nb_Neighb--;

  switch (Nb_Neighb) {
  case 0:	printf("    TRANSIT = OUT;\n");
    break;
  case 1:	
    The_Dest = The_Trans->Outputs->Dest;
    if ( is_recycling_place( &(Reseau->P[The_Dest]) ) ) {
      The_Dest = The_Trans->Outputs->Next->Dest;
      The_Link = The_Trans->Outputs->Next;
    }
    else
      The_Link = The_Trans->Outputs;

    if ( The_Link->Weight == 1 ) {
      printf("    TRANSIT = \n");
      fprint_dest( Reseau, The_Dest );
      fprint_class( Reseau, The_Dest );
      printf(";\n");
    }
    else if ( The_Link->Weight > 1 ) {
      printf("   FISSION = ( " );
      fprint_dest( Reseau, The_Dest );
      fprint_class( Reseau, The_Dest );
      printf( ", %d );\n", The_Link->Weight );
    }
    else { /* else weight = 0 and no customer created... */
      printf("    TRANSIT = OUT;\n");
    }
    break;
  default:
    printf("    FISSION = (\n");
    SCAN( L_Scan, The_Trans->Outputs )
      Loop_Is_Last = ( (L_Scan->Next==NULL) 
		       && is_recycling_place( &(Reseau->P[L_Scan->Dest]) ) );

    SCAN( L_Scan, The_Trans->Outputs ) {
      The_Dest = L_Scan->Dest;
      if ( !is_recycling_place( &(Reseau->P[The_Dest]) ) ) {
	fprint_dest( Reseau, The_Dest );
	fprint_class( Reseau, The_Dest );
	if ( ( L_Scan->Next != NULL ) 
	     && !( L_Scan->Next->Next==NULL && Loop_Is_Last ) )
	  printf( ", %d,\n", L_Scan->Weight );
	else
	  printf( ", %d );\n", L_Scan->Weight );
      }
    }
  }
  printf("\n");
}

/* procedure pour ecrire le service d'une place				*/
void fprint_place_service(PetriNet      *Reseau,
			  Pl_Des  	*The_Place,
			  int		Max_Time )
{
  printf("    SERVICE = ");

  if ( is_switch_place( The_Place ) ) {

    printf("BEGIN\n        ");
    /* normally, the second arg. of fprint_law is useless here...	*/
    Output_Law_QNAP( The_Place->Pl_Law, The_Place->Pl_Id );
    printf("        PL_%d.Num := 1 + MOD ( PL_%d.Num, %d);\n",
	   The_Place->Pl_Id, The_Place->Pl_Id, 
	   The_Place->Routing->Period );
    printf("        TRANSIT( PL_%d.Cycle( PL_%d.Num  ),",
	   The_Place->Pl_Id, The_Place->Pl_Id );
    printf(" PL_%d.Classe( PL_%d.Num ) );\n", 
	   The_Place->Pl_Id, The_Place->Pl_Id );
    printf("    END;\n");
  }
  else if ( The_Place->Outdegree == 0 )	{ /* it's a sink!! */
    printf("CST( %d );\n", Max_Time );
  }
  else
    Output_Law_QNAP( The_Place->Pl_Law, The_Place->Pl_Id );
}

/* procedure pour ecrire le routage d'une place qui n'est pas un switch	*/
void fprint_place_transit(
PetriNet	*Reseau,
Pl_Des  	*The_Place )
{
  int		Nb_Neighb;
  int		j;
  Liste_Link	L_Scan;

  Nb_Neighb = The_Place->Outdegree;

  switch (Nb_Neighb) {
  case 0:	/* no output, no TRANSIT... but QNAP needs one. So: */
    printf("    TRANSIT = OUT;\n");
    break;
  case 1: /* transit irrespective of conflict policy */
    printf("    TRANSIT =\n");
    fprint_trans_dest( The_Place->Outputs->Dest );
    fprint_trans_class( The_Place, 
			&(Reseau->T[The_Place->Outputs->Dest]) );
    printf(";\n");
    break;
  default:
    printf("    TRANSIT =\n");
    j = 0;
    SCAN( L_Scan, The_Place->Outputs ) {

      fprint_trans_dest( L_Scan->Dest );
      fprint_trans_class( The_Place, &(Reseau->T[L_Scan->Dest]) );
      if ( The_Place->Routing != NULL )
	printf(", %7.4f",
	       The_Place->Routing->Probas[j] );
      else	/* assume equal routing */
	printf(", %7.4f",
	       (double) 1.0 / Nb_Neighb );
      j++;
      if ( L_Scan->Next == NULL )
	printf(";\n");
      else
	printf(",\n");
    }
  }
  printf("\n");
}

void fprint_place(PetriNet      *Reseau,
		  int		The_Id,
		  int		Max_Time )
{
  Pl_Des          *The_Place;

  The_Place = &(Reseau->P[The_Id]);

  if ( !is_recycling_place( The_Place ) ) {

    if (Optimize && !Place_Exists[The_Id]) {
      /* if initial population of a non-recylcing and */
      /* non-routing place, put it in downstream	*/
      /* transition.					*/

      if (The_Place->Init_Pop != 0) {
	printf("/STATION/\n");
	printf("    NAME = TR_%d;\n", The_Place->Outputs->Dest );
	fprint_init_mark( Reseau, The_Id );
	printf("\n");
      }
    }
    else {
      /* helpful comment... */
      printf( "&\n& place %s\n", 
	      ( NULL == The_Place->Name ? 
		"<no name>" : 
		The_Place->Name ) );
      printf("/STATION/\n");

      printf("    NAME = PL_%d;\n", The_Id );
      printf("    TYPE = SERVER, INFINITE;\n");

      fprint_place_service( Reseau, The_Place, Max_Time );
	    
      fprint_init_mark( Reseau, The_Id );
	    
      if ( !is_switch_place( The_Place ) )
	fprint_place_transit( Reseau, The_Place );
    }
  }
}

/* procedure pour ecrire les declarations de procedures dans la trace	*/
void fprint_trace_procs( PetriNet *Reseau )
{
  int		i;
  Tr_Des	*The_Trans;
  Liste_Link	L_Scan;
  boolean	Has_Loop;
  int		Nb_Terms;	/* Used for "pretty" printing */

  /* define the function to compute the real marking in places	*/
  printf("    INTEGER FUNCTION POP( Q );\n");
  printf("    REF QUEUE Q;\n");
  printf("    BEGIN\n");
  printf("     ");

  for ( i=0; i<Reseau->Nb_Places; i++ ) {
    Nb_Terms = 0;
    printf(" IF Q=PL_%d\n", i );
    printf("      THEN RESULT := ");
    SCAN( L_Scan, Reseau->P[i].Inputs ) {
      if ( L_Scan->Weight == 1 )
	printf(" SERVNB( TR_%d ) + ", L_Scan->Dest );
      else 
	printf(" %1d * SERVNB( TR_%d ) + ", L_Scan->Weight,
	       L_Scan->Dest );
      Nb_Terms++;
      if ( Nb_Terms >= 3 ) {
	printf("\n");
	Nb_Terms = 0;
      }
    }
    printf("%d ", Reseau->P[i].Init_Pop );
    SCAN( L_Scan, Reseau->P[i].Outputs ) {
      if ( L_Scan->Weight == 1 )
	printf(" - SERVNB( TR_%d )", L_Scan->Dest );
      else
	printf(" - %1d * SERVNB( TR_%d )", L_Scan->Weight,
	       L_Scan->Dest );
      Nb_Terms++;
      if ( Nb_Terms >= 3 ) {
	printf("\n");
	Nb_Terms = 0;
      }
    }

    if ( i >= Reseau->Nb_Places-1 )
      printf(";\n    END;\n\n");
    else if ( (i+1) % 20 == 0 )
      printf(";\n     ");
    else
      printf("\n      ELSE");
  }

  /* define some macros for readability */
  printf("    $MACRO MODIFY( Obj, Attr, Val )\n" );
  printf("       WRITELN( Trace, \"MODIFY \", Obj, \" \"\"\", Attr,");
  printf(" \"\"\" = \" , Val );\n");
  printf("    $END\n\n");

  printf("    $MACRO MODTOK( Pla, Q1 )\n" );
  printf("       WRITELN( Trace, \"MODIFY \", Pla, \" \"\"token\"\" = \", ");
  printf("POP(Q1) );\n");
  printf("    $END\n\n");

  printf("    $MACRO MOVETOK( Orig, Dest, Path )\n" );
  printf("       WRITE( Trace, \"MOVE_NEW_OBJECT OF_CLASS token ");
  printf("FROM \", Orig );\n" );
  printf("       WRITELN( Trace , \" TO \", Dest, \" ON_PATH \", ");
  printf("Path, \" WITH_STEPS 20 \" );\n");
  printf("    $END\n\n");

  if ( Reseau->Has_Weights ) {
    printf("    $MACRO MOVE2TOK( Orig, Dest, Path )\n" );
    printf("       WRITE( Trace, \"MOVE_NEW_OBJECT OF_CLASS bitoken ");
    printf("FROM \", Orig );\n" );
    printf("       WRITELN( Trace , \" TO \", Dest, \" ON_PATH \", ");
    printf("Path, \" WITH_STEPS 20 \" );\n");
    printf("    $END\n\n");

    printf("    $MACRO MOVE3TOK( Orig, Dest, Path )\n" );
    printf("       WRITE( Trace, \"MOVE_NEW_OBJECT OF_CLASS tritoken ");
    printf("FROM \", Orig );\n" );
    printf("       WRITELN( Trace , \" TO \", Dest, \" ON_PATH \", ");
    printf("Path, \" WITH_STEPS 20 \" );\n");
    printf("    $END\n\n");

    printf("    $MACRO MOVE4TOK( Orig, Dest, Path )\n" );
    printf("       WRITE( Trace, \"MOVE_NEW_OBJECT OF_CLASS quadritoken ");
    printf("FROM \", Orig );\n" );
    printf("       WRITELN( Trace , \" TO \", Dest, \" ON_PATH \", ");
    printf("Path, \" WITH_STEPS 20 \" );\n");
    printf("    $END\n\n");
  }

  printf("    $MACRO BMULT\n" );
  printf("       WRITELN( Trace, \"BEGIN_MULTIPLE\" );\n" );
  printf("    $END\n\n");

  printf("    $MACRO EMULT\n" );
  printf("       WRITELN( Trace, \"END_MULTIPLE\" );\n" );
  printf("    $END\n\n");

  printf("    $MACRO EOL\n" );
  printf("       WRITELN( Trace, \"eol+\" );\n" );
  printf("    $END\n\n");

  for (i=0; i<Reseau->Nb_Trans; i++ ) {

    The_Trans = &(Reseau->T[i]);
    Has_Loop = is_recycled( Reseau, The_Trans );

    printf("    PROCEDURE FIRE%d;\n", i );
    printf("    BEGIN\n");
    /*	    printf("      WRITELN( \"Event(%d) \", GETTRACE:NAMECODE(GETTRACE:EVCODE), \" at \", GETTRACE:QPROVOKE );\n", i ); */
    printf("      $MODIFY( \"TheClock\", \"ctime\", TIME )\n");
    /* remove tokens from input places. I have trouble with	*/
    /* recycled transitions, so I skip the recycling places.	*/
    if ( ( Has_Loop && ( The_Trans->Indegree > 2 ) )
	 || ( !Has_Loop && ( The_Trans->Indegree > 1 ) ) )
      printf("      $BMULT\n");
    SCAN( L_Scan, The_Trans->Inputs )
      if ( !is_recycling_place( &(Reseau->P[L_Scan->Dest]) ) ) {

	printf("      $MODTOK( \"%s\"", 
	       Reseau->P[L_Scan->Dest].Data );
	printf(" , PL_%d )\n", L_Scan->Dest );
      }

    if ( ( Has_Loop && ( The_Trans->Indegree > 2 ) )
	 || ( !Has_Loop && ( The_Trans->Indegree > 1 ) ) )
      printf("      $EMULT\n");
    /* move them along paths . null paths probably are		*/
    /* recycled transitions and are silently avoided 		*/
    if ( ( Has_Loop && ( The_Trans->Indegree > 2 ) )
	 || ( !Has_Loop && ( The_Trans->Indegree > 1 ) ) )
      printf("      $BMULT\n");
    SCAN( L_Scan, The_Trans->Inputs ) {
      if ( L_Scan->Data != NULL ) {
	switch ( L_Scan->Weight ) {
	case 1:
	  printf("      $MOVETOK( ");
	  break;
	case 2:
	  printf("      $MOVE2TOK( ");
	  break;
	case 3:
	  printf("      $MOVE3TOK( ");
	  break;
	default:
	  printf("      $MOVE4TOK( ");
	}
	printf( "\"%s\", \"%s\", \"%s\" )\n",
		Reseau->P[L_Scan->Dest].Data,
		Reseau->T[i].Data,
		L_Scan->Data );
      }
    }
    if ( ( Has_Loop && ( The_Trans->Indegree > 2 ) )
	 || ( !Has_Loop && ( The_Trans->Indegree > 1 ) ) )
      printf("      $EMULT\n");
    /* change firing status of transition */
    printf("      IF( CUSTNB( TR_%d ) = 0 ) THEN ", i );
    printf("$MODIFY( \"%s\", \"status\", 0 )\n", Reseau->T[i].Data );
    /* move tokens along output paths. null paths probably are	*/
    /* recycled transitions and are silently avoided 		*/
    if ( ( Has_Loop && ( The_Trans->Outdegree > 2 ) )
	 || ( !Has_Loop && ( The_Trans->Outdegree > 1 ) ) )
      printf("      $BMULT\n");
    SCAN( L_Scan, The_Trans->Outputs ) {
      if ( L_Scan->Data != NULL ) {
	switch ( L_Scan->Weight ) {
	case 1:
	  printf("      $MOVETOK( ");
	  break;
	case 2:
	  printf("      $MOVE2TOK( ");
	  break;
	case 3:
	  printf("      $MOVE3TOK( ");
	  break;
	default:
	  printf("      $MOVE4TOK( ");
	  break;
	}
	printf( "\"%s\", \"%s\", \"%s\" )\n",
		Reseau->T[i].Data,
		Reseau->P[L_Scan->Dest].Data,
		L_Scan->Data );
      }
    }
    if ( ( Has_Loop && ( The_Trans->Outdegree > 2 ) )
	 || ( !Has_Loop && ( The_Trans->Outdegree > 1 ) ) )
      printf("      $EMULT\n");
    /* create new tokens in downstream places */
    if ( ( Has_Loop && ( The_Trans->Outdegree > 2 ) )
	 || ( !Has_Loop && ( The_Trans->Outdegree > 1 ) ) )
      printf("      $BMULT\n");
    SCAN( L_Scan, The_Trans->Outputs )
      if ( !is_recycling_place( &(Reseau->P[L_Scan->Dest]) ) )
	{
	  printf("      $MODTOK( \"%s\"",
		 Reseau->P[L_Scan->Dest].Data );
	  printf(", PL_%d )\n", L_Scan->Dest );
	}
    if ( ( Has_Loop && ( The_Trans->Outdegree > 2 ) )
	 || ( !Has_Loop && ( The_Trans->Outdegree > 1 ) ) )
      printf("      $EMULT\n");
    printf("      $EOL\n");

    printf("    END;\n");
    printf("\n");
  }

  printf("    PROCEDURE FINFIRE;\n");
  printf("    REF QUEUE Obj;\n");
  printf("    BEGIN\n");
  printf("      Obj := GETTRACE:QPROVOKE;\n");
  printf("     ");
  for ( i=0; i<Reseau->Nb_Trans; i++ )
    {   printf(" IF ( Obj = TR_%d ) THEN FIRE%d", i, i );
    if ( i >= Reseau->Nb_Trans-1 )
      printf(";\n    END;\n\n");
    else if ( (i+1) % 20 == 0 )
      printf(";\n     ");
    else
      printf("\n      ELSE");
    }

  printf("    PROCEDURE DEBFIRE;\n");
  printf("    REF QUEUE Obj;\n");
  printf("    BEGIN\n");
  printf("      Obj := GETTRACE:QPROVOKE;\n");
  printf("     ");
  for ( i=0; i<Reseau->Nb_Trans; i++ ) {
    printf(" IF ( Obj = TR_%d ) \n", i);
    printf("      THEN BEGIN $MODIFY( \"%s\", ", Reseau->T[i].Data );
    printf("\"status\", 1 ) $EOL END");
    if ( i >= Reseau->Nb_Trans-1 )
      printf(";\n    END;\n\n");
    else if ( (i+1) % 20 == 0 )
      printf(";\n     ");
    else
      printf("\n      ELSE");
  }
}	    
	    
/****************************************************************/
/* Function 	int	trans_in_link_number			*/
/* Computes the number of a P/T link, between 0 and the indegree*/
/* of T minus one.						*/
/* The order is the one in the Inputs list of the transition.	*/
/*								*/
/* Parameters:							*/
/*	The_Trans:	pointer on the transition		*/
/*	The_Place:	pointer on the place			*/
/*								*/
/* Return Value:						*/
/*	the number of the link					*/
/****************************************************************/
int	trans_in_link_number(Tr_Des *The_Trans, Pl_Des *The_Place)
{
  int		Num = 0;
  Liste_Link	L_Scan;

  for( L_Scan=The_Trans->Inputs; 
       (L_Scan != NULL) && (L_Scan->Dest != The_Place->Pl_Id );
       L_Scan=L_Scan->Next )
    Num++;

  if ( Num >= The_Trans->Indegree )
    {	fprintf( stderr,
		 "%s: warning: inconsistency in prev structure:",
		 TOOLNAME );
    fprintf( stderr, "no place #%d input of trans #%d\n", 
	     The_Place->Pl_Id, The_Trans->Tr_Id );
    }

  return Num;
}

/* procedure pour initialiser le tableau Place_Exists, utile a 		*/
/* l'optimisation mais aussi elimine les places de recyclage.		*/
void	init_place_status(PetriNet	*Reseau)
{
  int		i;
  Pl_Des		*The_Place;

  Place_Exists = (boolean*) calloc( Reseau->Nb_Places, sizeof(int) );
  Trans_Loop_Of = (int*) calloc( Reseau->Nb_Places, sizeof(int) );

  for ( i=0; i<Reseau->Nb_Places; i++ ) {
    The_Place = &(Reseau->P[i]);
    if ( is_recycling_place( The_Place ) && !Trace ) {
      Place_Exists[i] = FALSE;
      Trans_Loop_Of[i] = The_Place->Inputs->Dest;
      if (Debug) {
	printf("Recycling place Pl %d ", i );
	printf("(of Tr %d) removed\n", Trans_Loop_Of[i] );
      }
    }
    else if ( Optimize ) {
      Place_Exists[i] = 
	! ( ( The_Place->Outdegree == 1 )
	    && ( The_Place->Pl_Law->Name == 'C' )
	    && ( The_Place->Pl_Law->Parameters[0] == 0.0 )
	    );
      if (Debug && !Place_Exists[i])
	printf("Pl %d removed\n", i );
    }
    else
      Place_Exists[i] = TRUE;
  }
}

/************************************************************************/
/* Write declaration of special queue object necessary for interrupted	*/
/* Poisson process.							*/
/************************************************************************/
void Write_Special_Declarations(PetriNet *Reseau)
{
  boolean	Has_IPP = FALSE;
  int		Max_Period;
  int		i;

  Max_Period = max_cycle_length( Reseau );
  if ( Max_Period >= 2 ) {		/* les places de periode 1 ont	*/
					/* un routage simplifie.	*/
    printf("&   nouveau type pour places de routage cyclique\n");
    printf("    QUEUE OBJECT Switch;\n");
    printf("      REF QUEUE Cycle(%d);\n", Max_Period );
    printf("      REF CLASS Classe(%d);\n", Max_Period );
    printf("      INTEGER   Period;\n");
    printf("      INTEGER   Num;\n");
    printf("    END;\n");
  }

  for( i=0; !Has_IPP && (i<Reseau->Nb_Trans); i++ ) {
    Has_IPP = ( Reseau->T[i].Tr_Law->Name == 'P' );
  }

  if ( Has_IPP ) {
    printf("    QUEUE OBJECT OnOffSrc;\n");
    printf("      REAL NextArr;\n");
    printf("      REAL NextChg;\n");
    printf("    END;\n");
  }
}

void dump_qnap_code(PetriNet *Reseau, int Max_Time)
{
  int		i;
  int		j;
  Routing_Desc	The_Routing;
  int		Max_Deg;
  int		Last_Place;
  int		Last_Switch;
  int		Nb_Real_Places;
  char		*The_Date;

  init_place_status( Reseau );

  printf("& QNAP2 code generated by %s", TOOLNAME );
  printf(" version %s\n", VERSION );
  if ( strlen( Reseau->File_Name ) > 0 )
    printf("& Origin: %s\n", Reseau->File_Name );
  The_Date = todays_date();
  printf("& Date: %s\n", The_Date );
  free( The_Date );

  /* desactivation du listage et du resultat */
  printf("/CONTROL/\n");
  printf("    OPTION = NSOURCE;\n");
  printf("    OPTION = NRESULT;\n\n");
  
  /* declarations */
  printf("/DECLARE/\n");

  Write_Special_Declarations( Reseau );

  if ( Trace ) {
    printf("&   fichier pour la trace\n");
    printf("    FILE Trace;\n");
  }

  /* declaration of normal queues */
  printf("&   transitions\n");
  printf("    QUEUE");
  for (i=0; i<Reseau->Nb_Trans-1; i++) {
    if ( Reseau->T[i].Tr_Law->Name != 'P' ) {
      printf(" TR_%d",i);
      if ( ( i != Reseau->Nb_Trans-2 ) 
	   || ( Reseau->T[Reseau->Nb_Trans-1].Tr_Law->Name != 'P' ) )
	printf(",");
      if ((i+1) % 10 == 0)
	printf("\n         ");
    }
  }
  if ( Reseau->T[Reseau->Nb_Trans-1].Tr_Law->Name != 'P' )
    printf(" TR_%d;\n", Reseau->Nb_Trans-1 );
  else
    printf(";\n");

  /* declaration of IPP sources */
  for (i=0; i<Reseau->Nb_Trans; i++) {
    if ( Reseau->T[i].Tr_Law->Name == 'P' ) {
      printf("    OnOffSrc TR_%d;\n",i);
    }
  }

  /* other variables */
  printf("&   other variables\n");
  printf("    REAL Tps_Cpu;\n");
  printf("    INTEGER Nb_Evts;\n");
  printf("\n");

  /* compute index of last normal and switch places. 		*/
  /* Used to know if there are such places, and end properly the	*/
  /* declaration with a ";".					*/
  Last_Place = -1;
  Last_Switch = -1;
  for (i=0; i<Reseau->Nb_Places; i++) {
    if ( Place_Exists[i] ) {
      if ( is_switch_place( &(Reseau->P[i]) ) )
	Last_Switch = i;
      else
	Last_Place = i;
    }
  }
  /* list places */
  if ( Last_Place >= 0 ) {
    printf("&   places\n");
    printf("    QUEUE");
    Nb_Real_Places = 0;
    for (i=0; i<Last_Place; i++) {
      if ( !is_switch_place( &(Reseau->P[i]) )
	   && Place_Exists[i] ) {
	printf(" PL_%d,", i);
	Nb_Real_Places++;
	if ( (Nb_Real_Places+1) % 9 == 0)
	  printf("\n         ");
      }
    }
    printf(" PL_%d;\n", Last_Place );
  }
  /* list switches. Switches always exist. */
  if ( Last_Switch >= 0 ) {
    printf("&   switches cycliques\n");
    printf("    Switch");
    Nb_Real_Places = 0;
    for (i=0; i<Last_Switch; i++)
      if ( is_switch_place( &(Reseau->P[i]) ) )
	{	printf(" PL_%d,",i);
	Nb_Real_Places++;
	if ((Nb_Real_Places+1) % 9 == 0)
	  printf("\n         ");
	}
    printf(" PL_%d;\n", Last_Switch );
  }

  printf("&   classes\n");
  /* computation of the maximum indegree */
  Max_Deg = Reseau->T[0].Indegree;
  for (i=1; i<Reseau->Nb_Trans; i++)
    Max_Deg = MAX( Max_Deg, Reseau->T[i].Indegree );

  printf("    CLASS");
  for (i=0; i<Max_Deg; i++)
    {	printf(" CL_%d,",i);
    if ((i+1) % 9 == 0)
      printf("\n         ");
    }
  printf(" No_Class;\n" );
  printf("\n");

  if ( Trace )
    fprint_trace_procs( Reseau );

  /* bloc de controle */
  printf("/CONTROL/\n");
  printf("    CLASS = ALL QUEUE;\n");
  if ( !Counters )
    printf("    TMAX = %d;\n", Max_Time );
  else
    printf("    TMAX = %d;\n", Max_Time*1000 );
  printf("\n");

  /* generate transitions */
  for ( i=0; i<Reseau->Nb_Trans; i++ )
    fprint_transition( Reseau, i, Max_Time );

  /* now generate places */
  for ( i=0; i<Reseau->Nb_Places; i++ )
    fprint_place( Reseau, i, Max_Time );

  /* exec bloc */
  printf("/EXEC/\n");
  printf("    BEGIN\n");
  /* initialisation des cycles dans les switches. */
  for ( i=0; i<Reseau->Nb_Places; i++ )
    if ( is_switch_place( &(Reseau->P[i]) ) ) {
      The_Routing = Reseau->P[i].Routing;
      for ( j=0; j<The_Routing->Period; j++ )
	printf("        PL_%d.Cycle(%d) := TR_%d;\n",
	       i, j+1, The_Routing->Cycle[j]->Number );
      for ( j=0; j<The_Routing->Period; j++ ) {
	printf("        PL_%d.Classe(%d) :=", i, j+1 );
	if ( Reseau->T[The_Routing->Cycle[j]->Number].Indegree == 1 )
	  printf(" No_Class" );
	else
	  printf(" CL_%d", 
		 trans_in_link_number(
				      &(Reseau->T[The_Routing->Cycle[j]->Number]),
				      &(Reseau->P[i]) ) );
	printf(";\n");
      }
      printf("        PL_%d.Period := %d;\n",
	     i, The_Routing->Period );
      printf("        PL_%d.Num := 0;\n", i );
    }

  /* initialization of IPP sources */
  for (i=0; i<Reseau->Nb_Trans; i++) {
    if ( Reseau->T[i].Tr_Law->Name == 'P' ) {
      printf("        TR_%d.NextChg := ",i);
      Output_Law_QNAP( Reseau->T[i].Tr_Law->Law_Parameters[1], i );
      printf(";\n");
    }
  }

  /* initialization of the trace */
  if ( Trace ) {
    printf("        FILASSIGN( Trace, \"%s\" );\n", Trace_File_Name );
    printf("        OPEN( Trace, 3 );\n");
    printf("        WRITELN( Trace, \"# Trace generated by QNAP" );
    printf("\" );\n" );
    printf("        WRITELN( Trace, \"# Get rid of the current " );
    printf("clock and create a new one\" );\n" );
    printf("        WRITELN( Trace, \"KILL_OBJECT TheClock\" );\n" );
    printf("        WRITELN( Trace, \"MAKE_NODE TheClock OF_CLASS " );
    printf("clock AT_POSITION 720 750 \" );\n" );
    printf("        SETTRACE:BOUNDS( %f, %f );\n",
	   Trace_Start, Trace_End );
    printf("        SETTRACE:DEFSET( NIL );\n" );

    /*	    for ( i=0; i<Reseau->Nb_Trans; i++ )
	    {   printf("        SETTRACE:SET( TR_%d, \"TRANSIT\", \"ENDSERV\"",
	    i );
	    printf(", FIRE%d );\n", i );
	    }
	    */
    printf("        SETTRACE:DEFSET( \"ENDSERV\", FINFIRE );\n");
    printf("        SETTRACE:DEFSET( \"FISSION\", FINFIRE );\n");
    printf("        SETTRACE:DEFSET( \"SERVTIME\", DEBFIRE );\n");
    printf("        WRITELN( Trace, \"# Reset populations ");
    printf("and transition status\" );\n");
    for ( i=0; i<Reseau->Nb_Trans; i++ )
      printf("        $MODIFY( \"%s\", \"status\", 0 )\n", 
	     Reseau->T[i].Data );
    for ( i=0; i<Reseau->Nb_Places; i++ )
      if ( !is_recycling_place( &(Reseau->P[i]) ) )
	/*	    {   if ( Trace_Start = 0.0 )
	 */		    printf("        $MODIFY( \"%s\", \"token\", %d )\n", 
				   Reseau->P[i].Data, Reseau->P[i].Init_Pop );
    /*		else
		printf("        $MODTOK( \"%s\", PL_%d )\n", 
		Reseau->P[i].Data, i );
		}
		*/
    printf("        $EOL\n");

  }

  if ( Trace ) {
    printf("        WRITELN( Trace, \"#\" );\n");
    printf("        WRITELN( Trace, \"# Start animation\" );\n");
    printf("        WRITELN( Trace, \"#\" );\n");
  }

  printf("        Tps_Cpu := GETCPUT;\n");
  printf("        Nb_Evts := 0;\n");
  printf("        SIMUL;\n");
  for (i=0; i<Reseau->Nb_Trans; i++) {
    if ( Counters ) {
      printf("        WRITE(\"X_%d[%d]= \", SERVNB(TR_%d), \"",
	     i, Max_Time, i );
      printf("     Cycle time:  \"  );\n" );
    }
    else
      printf("        WRITE(\"Cycle time of transition %d: \" )\n;", 
	     i);

    printf("        IF( SERVNB(TR_%d) > 0 ) THEN ", i );
    printf("WRITE( TIME/SERVNB(TR_%d) ) ", i );
    printf("ELSE WRITE( \"  oo \" );\n" );

    if ( ( Reseau->T[i].Name != NULL )
	 && ( strlen( Reseau->T[i].Name ) > 0 ) )
      printf("        WRITELN( \"      (%s)\" );\n", 
	     Reseau->T[i].Name );
    else 
      printf("           WRITELN;\n");

    printf("        Nb_Evts := Nb_Evts + SERVNB(TR_%d);\n", i);
  }
  printf("        Tps_Cpu := GETCPUT;\n");
  printf("        WRITE(\"Total Execution Time: \", Tps_Cpu );\n" );
  printf("        WRITELN(\"      (\", Nb_Evts / Tps_Cpu, \"evt/sec)\");\n" );
  printf("    END;\n");
  printf("/END/\n");

  exit(0);
}

/************************************************************************/
/* Fonction	satisfies_constraints( Reseau )				*/
/* verifie si le reseau satisfait aux contraintes necessaires pour etre	*/
/* transcrit en un reseau de files d'attente.				*/
/* Les contraintes sont pour le moment:					*/
/* - pas de transition sans entree (source taux infini)			*/
/* - pas de resolution de (vrai) conflit de type Competition		*/
/* - pas d'arc inhibiteur						*/
/*									*/
/* Argument: 	Reseau, de type PetriNet				*/
/* Retourne:	TRUE si le reseau satisfait aux contraintes, FALSE sinon*/
/************************************************************************/
boolean	satisfies_constraints(PetriNet	*Reseau)
{
  boolean	Is_Ok = TRUE;
  Pl_Des	*The_Place;
  Tr_Des	*The_Trans;
  int		i;

  if ( Reseau->Has_Inhibitors ) {
    fprintf( stderr, "%s: the Petri network has inhibitor arcs. ", 
	     TOOLNAME );
    fprintf( stderr, "No translation.\n" );
    return FALSE;
  }
  else {
    for ( i=0; (i<Reseau->Nb_Places) && Is_Ok; i++ ) {
      The_Place = &( Reseau->P[i] );
      if ( ( The_Place->Outdegree > 1 )
	   && ( The_Place->Routing != NULL ) ) {
	if ( ( The_Place->Routing->Type == 'C' ) 
	     || ( The_Place->Routing->Type == 'P' ) ) {
	  Is_Ok = FALSE;
	  fprintf( stderr, "%s: the Petri network cannot ", TOOLNAME );
	  fprintf( stderr, "be translated.\n\t" );
	  fprintf( stderr, "Place #%d [%s] has a Competition policy.\n",
		   The_Place->Pl_Id, The_Place->Name );
	}
      }
    }
	
    for ( i=0; (i<Reseau->Nb_Trans) && Is_Ok; i++ ) {
      The_Trans = &( Reseau->T[i] );
      if ( The_Trans->Indegree == 0 ) {
	Is_Ok = FALSE;
	fprintf( stderr, "%s: the Petri network cannot ", TOOLNAME );
	fprintf( stderr, "be translated.\n\t" );
	fprintf( stderr, "Trans #%d ", The_Trans->Tr_Id );
	if ( ( The_Trans->Name != NULL ) 
	  && ( strlen(The_Trans->Name) > 0 ) )
	  fprintf( stderr, "(%s) ", The_Trans->Name );
	fprintf( stderr, "has no inputs.\n" );
      }
    }
    
    return Is_Ok;
  }
}


