/* programme/module de generation de bornes sur le marquage		*/
/* par generation de programmes lineaires.				*/

#include <stdio.h>
#include <unistd.h>
#include <math.h>

#include "pn.h"

#ifndef LINUX
#define MAXINT          0x7FFFFFFF
#endif
#define INFINI		1e12

#define max(x,y)        ( (x>y) ? x : y )
#define min(x,y)        ( (x<y) ? x : y )

#define	TOOLNAME	"pn_bound"
#define LIN_PROG_NAME	"pn_bound.lp"
#define RES_FILE_NAME	"pn_bound.res"

#define CODE(x)		fprintf(Lin_Prog,x)
#define CODE1(x,y)	fprintf(Lin_Prog,x,y)
#define CODE2(x,y,z)	fprintf(Lin_Prog,x,y,z)
#define CODE3(x,y,z,t)	fprintf(Lin_Prog,x,y,z,t)
#define CODELN		CODE("\n")
	
#define SCAN(x,y)	for(x=y;x!=NULL;x=x->Next)
#define CLOSE(x,y)	fabs( (double)x - (double)y ) < 1e-4

/* Global declarations */
extern PetriNet		*Read_Network(char *File_Name);
extern void		dumpe_network(PetriNet *Reseau);
extern char		*todays_date();
extern Liste_Link	Append_Link();
extern int		**Cycle_Base();

/* Forward declarations */
boolean	check_network();
void	gen_solve_analyze();
void	write_B_file();

/************************************\
* 				     *
*  convenience background variables  *
* 				     *
\************************************/
FILE		*Lin_Prog;
int		*The_Bound;	/* table of bounds if they exist	*/

int main(int argc, char *argv[] )
{
  static char	File_Name[128] = "";
  static char	Code_File_Name[128] = "";
  static char	B_File_Name[128] = "";
  PetriNet 	*Reseau;
  int		i;
  int		Place_Num;
  boolean	Help = FALSE;
  boolean	Args_Ok = TRUE;
  boolean	File_Ok = TRUE;
  boolean	Write_Bfile = FALSE;
  boolean	Tot_Pop = FALSE;
  boolean	Minimum = FALSE;
  boolean	Single_Place = FALSE;

  i = 1;

  if (argc>1)
    do {

      if GET_STRING( "-f", File_Name )
      else if GET_STRING( "-o", Code_File_Name )
      else if GET_STRING_SWITCH( "-Bfile", B_File_Name, Write_Bfile, TRUE )

      else if GET_INT_SWITCH( "-place", Place_Num, Single_Place, TRUE )

      else if SWITCH( "-debug", Debug, TRUE )
      else if SWITCH( "-min", Minimum, TRUE )
      else if SWITCH( "-total", Tot_Pop, TRUE )

      else if IS_VERSION

      else if ( ( 0 == strcmp( "-h", argv[i] ) )
		|| ( 0 == strcmp( "-?", argv[i] ) ) ) {
	printf("Usage: %s [-f <file>] ", argv[0]);
	printf("[-debug] [-o <file>] [-min] [-total]\n\t");
	printf("[-place <int>] [-Bfile <file>]\n");
	Args_Ok = FALSE;
	Help = TRUE;
	i++;
      }
      else
	i++; /* skip unknown argument */
    }
    while ( Args_Ok && ( i < argc ) );

  if ( Args_Ok ) {

    Reseau = Read_Network( File_Name );

    if ( Reseau != NULL ) {

	if (Debug)
	  dumpe_network(Reseau);

	if ( Reseau->Has_Inhibitors || Reseau->Has_Weights ) {
	  printf("%s: the networks has weights and/or inhibitor arcs.",
		 TOOLNAME );
	  printf(" Can't handle it.\n");
	  return -3;
	}
	else {
	  /* total pop supersedes the Bfile */
	  if ( Tot_Pop )
	    Write_Bfile = FALSE;
		
	  if ( strlen(Code_File_Name) != 0 )
	    File_Ok = ( NULL != freopen( Code_File_Name, "w", stdout ) );

	  if ( File_Ok ) {

	    gen_solve_analyze( Reseau, Tot_Pop, Minimum, 
			       Single_Place, Place_Num-1 );

	    if ( Write_Bfile )
	      write_B_file( Reseau, B_File_Name, Single_Place, Place_Num-1 );

	    /* successful execution */
	    return 0;
	  }
	  else {
	    fprintf( stderr, "%s: Error opening file %s.\n", 
		     TOOLNAME, Code_File_Name );
	    return -1;
	  }
	}
      }
      else {
	fprintf( stderr, "%s: Error reading network. No analysis.\n", 
		 TOOLNAME );	
	return -2;
      }
  }
  else if (!Help)
    return -1;
  else
    return 0;
}

/* dumping bounds into a file */
void	write_B_file(
PetriNet        *Reseau,
char		*B_File_Name,
boolean		Single_Place,
int		Place_Num )
{
  FILE	*B_File;
  int	i;

  if ( NULL != ( B_File = fopen( B_File_Name, "w" ) ) ) {
    if ( Single_Place ) {
      if ( The_Bound[ Place_Num ] != MAXINT )
	fprintf( B_File, "%d %d\n",  Place_Num , The_Bound[ Place_Num ] );
      /* else write nothing. File will exist but stay empty */
    }
    else
      for ( i=0; i<Reseau->Nb_Places; i++ )
	if ( The_Bound[i] != MAXINT )
	  fprintf( B_File, "%d %d\n", i, The_Bound[i] );
  }
  else {
    printf("%s: could not open file %s for writing.\n",
	   TOOLNAME, B_File_Name );
  }

  fclose( B_File );
}

/* Reachable marking constraints */
void	gen_reachable_marking(Reseau)
PetriNet	*Reseau;
{
  int		i;
  Liste_Link	L_Scan;

  CODE("/* Reachable marking constraints */\n");

  for ( i=0; i<Reseau->Nb_Places; i++ )
    {
      CODE2( "x%d = %d", i, Reseau->P[i].Init_Pop );

      SCAN(L_Scan, Reseau->P[i].Inputs)
	CODE1( " + u%d", L_Scan->Dest);

      SCAN(L_Scan, Reseau->P[i].Outputs)
	CODE1( " - u%d", L_Scan->Dest);

      CODE(";\n");
    }
}

void	gen_lin_code( Reseau, Place_Nb, Total, Minimum )
PetriNet	*Reseau;
int		Place_Nb;
boolean		Total;
boolean		Minimum;
{
	int	i;

	CODE("/* Linear program computing an upper bound on the marking");
	CODE(" in a Petri Net */\n");
	CODE2("/* Generated by %s, %s */\n", TOOLNAME, todays_date() );

	if ( Minimum )
	    CODE("min:");
	else
	    CODE("max:");

	if ( Total )
	{   CODE(" x0");
	    for ( i=1; i<Reseau->Nb_Places; i++ )
		CODE1(" + x%d", i );
	    CODE(";\n");
	}
	else
	    CODE1(" x%d;\n", Place_Nb );

	gen_reachable_marking(Reseau);
}

double parse_lin_results( int Nb_Trans, int Nb_Places )
{
  FILE		*Data;
  double	Bound;
  int		Status;

  if ( Debug )
    printf("Parsing the results of the linear program.\n");

  Data = fopen( RES_FILE_NAME, "r" );

  Status = fscanf( Data, "%*[A-z: ]%lf", &Bound );
  switch( Status ) {
  case EOF:
    printf("No solution found in by the linear program.\n");
    Bound = -1.0;
    break;
  case 0:
    printf("Error reading linear program results.\n");
    Bound = -1.0;
    break;
  }

  unlink( RES_FILE_NAME );

  return Bound;

}

void	gen_solve_analyze(
PetriNet	*Reseau,
boolean		Total,
boolean		Minimum,
boolean		Single_Place,
int		Place_Num )
{
  char	command[512];
  double	Value;
  int		i;
  boolean	File_Ok;

  The_Bound = (int*) calloc( Reseau->Nb_Places, sizeof( int ) );

  if ( Total ) {

    Lin_Prog = fopen( LIN_PROG_NAME, "w" );
    File_Ok = ( NULL != Lin_Prog );

    if ( File_Ok ) {
      if (Debug)
	printf("Generating Linear Program for total population\n");

      gen_lin_code( Reseau, 0, TRUE, Minimum );

      fclose( Lin_Prog );

      sprintf( command, "lp_solve < %s > %s", 
	       LIN_PROG_NAME, RES_FILE_NAME );
      system( command );

      unlink( LIN_PROG_NAME );

      Value = parse_lin_results( Reseau->Nb_Trans, Reseau->Nb_Places );

      if ( 0 <= Value ) {
	printf("Value found for ");
	if ( Minimum )
	  printf("minimum");
	else
	  printf("maximum");
	printf(" total population: %d\n", (int) floor(Value) );
      }
    }
    else 
      fprintf( stderr, "%s: Error opening file %s.\n", 
	       TOOLNAME, LIN_PROG_NAME );

  }
  else if ( Single_Place ) {

    if ( ( Place_Num < 0 ) || ( Place_Num >= Reseau->Nb_Places ) ) {
      fprintf( stderr, "%s: invalid place number (%d).\n",
	       TOOLNAME, Place_Num+1 );
    }
    else {

      Lin_Prog = fopen( LIN_PROG_NAME, "w" );
      File_Ok = ( NULL != Lin_Prog );

      if ( File_Ok ) {

	if (Debug)
	  printf("Generating Linear Program for single place %d\n", 
		 Place_Num );

	gen_lin_code( Reseau, Place_Num, Total, Minimum );

	fclose( Lin_Prog );

	sprintf( command, "lp_solve < %s > %s", 
		 LIN_PROG_NAME, RES_FILE_NAME );
	system( command );

	unlink( LIN_PROG_NAME );

	Value = parse_lin_results( Reseau->Nb_Trans, Reseau->Nb_Places );
	if ( 0 <= Value ) {
	  The_Bound[Place_Num] = (int) floor(Value);
	  printf("Value found for ");
	  if ( Minimum )
	    printf("lower");
	  else
	    printf("upper");
	  printf(" bound for place %d (%s): %d\n", 
		 Place_Num+1, Reseau->P[Place_Num].Name,
		 The_Bound[Place_Num] );
	}
	else
	  The_Bound[Place_Num] = MAXINT;
      }
      else 
	fprintf( stderr, "%s: Error opening file %s.\n", 
		 TOOLNAME, LIN_PROG_NAME );

    }
  }
  else for ( i=0; i<Reseau->Nb_Places; i++ ) {

    Lin_Prog = fopen( LIN_PROG_NAME, "w" );
    File_Ok = ( NULL != Lin_Prog );

    if ( File_Ok ) {
      if (Debug)
	printf("Generating Linear Program for place %d\n", i);

      gen_lin_code( Reseau, i, Total, Minimum );

      fclose( Lin_Prog );

      sprintf( command, "lp_solve < %s > %s", 
	       LIN_PROG_NAME, RES_FILE_NAME );
      system( command );

      unlink( LIN_PROG_NAME );

      Value = parse_lin_results( Reseau->Nb_Trans, Reseau->Nb_Places );
      if ( 0 <= Value ) {
	The_Bound[i] = (int) floor(Value);
	printf("Value found for ");
	if ( Minimum )
	  printf("lower");
	else
	  printf("upper");
	printf(" bound for place %d (%s): %d\n", 
	       i+1, Reseau->P[i].Name, The_Bound[i] );
      }
      else
	The_Bound[i] = MAXINT;
    }
    else 
      fprintf( stderr, "%s: Error opening file %s.\n", 
	       TOOLNAME, LIN_PROG_NAME );

  }
}
