/************************************************************************/
/* Program to test the random generators.				*/
/************************************************************************/

#include <stdio.h>
#include <malloc.h>
#include <values.h>
#include <math.h>
#include <strings.h>
#include "boolean.h"
#include "parseargs.h"
#include "law.h"

#define	TOOLNAME	"gen_iid"
#define VERSION 	"v1.0"

/* Global variables */
boolean	Debug = FALSE;

/* Parsing procedure */
int Parse_Law_From_Args( char **argv, int index, int argc, Law_Desc The_Law );
extern double Sample( Law_Desc );

/* Forwars declarations */
void Print_Header_Info( Law_Desc The_Law, int Nb_Samples );
void Draw( Law_Desc, int );

int main(int argc, char *argv[])
{
  static char		Result_File_Name[256] = "";
  int			i;
  boolean		Help = FALSE;
  boolean		Args_Ok = TRUE;
  boolean		File_Ok = TRUE;
  int			Nb_Samples = 100000;
  Law_Desc		The_Law;

  /* Various initializations before reading arguments */
  The_Law = New_Law();
  The_Law->Name = 'E';
  The_Law->Parameters[0] = 1.0;

  i = 1;

  if (argc>1) {
    do {
      if GET_STRING( "-o", Result_File_Name )

      else if GET_INT( "-nbsamples", Nb_Samples )

      else if SWITCH( "-debug", Debug, TRUE )
      else if IS_VERSION		     

      else if ( 0 == strcmp( "-law", argv[i] ) ) {
	i++;
	i = Parse_Law_From_Args( argv, i, argc, The_Law );
      }
      else if ( ( 0 == strcmp( "-h", argv[i] ) )
		|| ( 0 == strcmp( "-?", argv[i] ) ) ) {
	printf("Usage: %s [-o <file>] ", TOOLNAME );
	printf("[-law <law>] [-nbsamples <int>]\n\t");
	printf("[-debug] [-h] [-?]\n");
	Args_Ok = FALSE;
	Help = TRUE;
	i++;
      }
      else
	i++; /* skip unknown argument */
    } while ( Args_Ok && ( i < argc ) );
  }

  if (Args_Ok) {

    if ( strlen(Result_File_Name) != 0 )
      File_Ok = ( NULL != freopen( Result_File_Name, "w", stdout ) );

    if ( !File_Ok ) {
      printf("%s: error: could not open file %s\n", 
	     TOOLNAME, Result_File_Name );
      return -1;
    }
    else { 

      if ( Debug )
	Print_Header_Info( The_Law, Nb_Samples );

      Draw( The_Law, Nb_Samples );

      return 0;
    }
  }

  else if (Help)
    return 0;
  else
    return 1;
}

int Parse_Law_From_Args( char **argv, int index, int argc,
			 Law_Desc The_Law )
{
  int		i;
  int		j;
  int		Nb_Params;
  boolean 	Error;
  boolean	Short;
  boolean 	Ok;
  double	Foo;
  
  i = index;
  Error = FALSE;
  Short = FALSE;

  The_Law->Name = argv[i][0];
  i++;

  if ( i < argc ) {
    switch( The_Law->Name ) {
    case 'C':
    case 'E':
    case 'B':
    case 'g':
    case 'G':
      if ( 1 != sscanf( argv[i], "%lf", &(The_Law->Parameters[0] ) ) ) {
	Error = TRUE;
      }
      else
	i++;
      break;
    case 'R':
    case 'U':
    case 'W':
    case 'N':
    case 'L':
    case 'l': /* double lognormale. Loi ad-hoc pour tests Experimental */
    case 'p':
      The_Law->Parameters = (double*) 
	realloc ( The_Law->Parameters, 2*sizeof(double) );
      if ( 1 != sscanf( argv[i], "%lf", &(The_Law->Parameters[0] ) ) ) {
	Error = TRUE;
      } 
      else {
	i++;
	if ( i < argc ) {
	  if ( 1 != sscanf( argv[i], "%lf", &(The_Law->Parameters[1] ) ) ) {
	    Error = TRUE;
	  }
	  else 
	    i++;
	} 
	else {
	  Short = TRUE;
	}
      }
      break;
    case 'O':
      The_Law->Parameters = (double*) 
	realloc ( The_Law->Parameters, 3*sizeof(double) );
      if ( 1 != sscanf( argv[i], "%lf", &(The_Law->Parameters[0] ) ) ) {
	Error = TRUE;
      }
      else {
	i++;
	if ( i < argc ) {
	  if ( 1 != sscanf( argv[i], "%lf", &(The_Law->Parameters[1] ) ) ) {
	    Error = TRUE;
	  } 
	  else {
	    i++;
	    if ( i < argc ) {
	      if ( 1 != sscanf( argv[i], "%lf", &(The_Law->Parameters[2] ) ) ) {
		Error = TRUE;
	      } 
	    }
	  }
	}
	else 
	  Error = TRUE;
      }
      break;
    case 'H':
      /* the number of parameters is not known in advance... have to 	*/
      /* find it first.							*/
      Ok = TRUE;
      for ( j=i; Ok && ( j < argc ); j++ ) {
	Ok = ( 1 == sscanf( argv[j], "%lf", &Foo ) );
      }
      /* counting the number of floats read */
      if ( j < argc )
	j--;
      Nb_Params = (j-i);
      if ( Nb_Params %2 != 0 ) {
	Error = TRUE;
      } else {
	The_Law->Parameters = (double*) 
	  realloc ( The_Law->Parameters, (1+Nb_Params)*sizeof(double) );
	The_Law->Parameters[0] = (double)(Nb_Params);
	for( j=i; j<i+Nb_Params; j++ ) {
	  sscanf( argv[j], "%lf", &(The_Law->Parameters[1+j-i]) );
	}
	i = i+Nb_Params;
      }
      break;
    case 'D':
      /* skip the (a priori) "[" argument */
      i++;
      /* the number of parameters is not known in advance... have to 	*/
      /* find it first, looking for the next "]".			*/
      Ok = TRUE;
      for ( j=i; Ok && ( j < argc ); j++ ) {
	Ok = ( 0 != strcmp( argv[j], "]" ) );
      }
      /* counting the number of floats read */
      if ( j >= argc ) {
	Short = TRUE;
      }
      else {
	j--;
	Nb_Params = (j-i);
	The_Law->Parameters = (double*) 
	  realloc ( The_Law->Parameters, (1+2*Nb_Params)*sizeof(double) );
	The_Law->Parameters[0] = (double)(Nb_Params);
	for( j=0; j<Nb_Params; j++ ) {
	  sscanf( argv[i+j], "%lf", &(The_Law->Parameters[1+j]) );
	}
	/* skip "]" and "[" */
	i += Nb_Params+2;
	/* continue reading. Have to protect against short arg list */
	Ok = TRUE;
	for( j=0; Ok && (j<Nb_Params) && (i+j<argc); j++ ) {
	  Ok = ( 1 == sscanf( argv[i+j], "%lf", 
			      &(The_Law->Parameters[1+Nb_Params+j]) ) );
	}
	if ( !Ok ) {
	  Error = TRUE;
	  i = i+j-1;
	} 
	else if ( i+j >= argc ) {
	  Short = TRUE;
	}
	else {
	  /* skip final "]", hoping it is actually there */
	  i = i + Nb_Params + 1;
	}
      }
      break;
    case 'P':
      /* read first parameter */
      if ( 1 != sscanf( argv[i], "%lf", &(The_Law->Parameters[0] ) ) ) {
	Error = TRUE;
      }
      else {
	i++;
	/* recursively read two laws */
	The_Law->Law_Parameters = (Law_Desc*) calloc( 2, sizeof(Law_Desc) );
	The_Law->Law_Parameters[0] = New_Law();
	The_Law->Law_Parameters[1] = New_Law();
	i = Parse_Law_From_Args( argv, i, argc, The_Law->Law_Parameters[0] );
	i = Parse_Law_From_Args( argv, i, argc, The_Law->Law_Parameters[1] );
      }
    default:
      printf("%s: unknown law id (%c)\n", TOOLNAME, The_Law->Name );
    }
  }
  else {
    Short = TRUE;
  }

  if ( Short ) {
    printf("%s: error parsing law %c: not enough arguments.\n", 
	   TOOLNAME, The_Law->Name );
    i = index;
  }
  else if ( Error ) {
    printf("%s: error parsing law %c at argument %s.\n", 
	   TOOLNAME, The_Law->Name, argv[i] );
    i = index;
  }

  return i;
}

/* Simulation procedure */
void Draw( Law_Desc	The_Law,
	   int		Nb_Samples )
{
  int		i;

  for ( i=0; i<Nb_Samples; i++ ) {

    printf("%f\n", Sample( The_Law ) );

  }

}

void Print_Header_Info( Law_Desc The_Law, int Nb_Samples )
{
    printf("# %d samples of the law ", Nb_Samples );
    Write_Law( stdout, The_Law, 0 );
    printf("\n");
}


