#include <stdio.h>
#include <malloc.h>
#include <values.h>
#include <math.h>
#include <strings.h>

#ifndef BOOLEAN
#define BOOLEAN
typedef short   boolean;
#define TRUE    1
#define FALSE   0
#endif

typedef struct SEQUENCE {
  int		T;	/* length of the transient period 	*/
  int		D;	/* length of the cyclic period 		*/
  double	*B;	/* Values of the transient part		*/
  double	*Inc;	/* Value of increments.			*/
  double	Cycle_Length;	/* Total length of a cycle	*/
  double	Gamma;	/* Average growth rate 			*/
  double	*C;	/* the "extrapolated" sequence 		*/
  double	*CS;	/* C, shifted at point T 		*/
} *Sequence;

#define	TOOLNAME	"mk1"
#define VERSION 	"v1.0"

#define	MAX(x,y)	( (x>y) ? (x) : (y) )

boolean	Debug;
boolean	Queue = FALSE;
boolean Log = FALSE;
boolean	NewLine = TRUE;

boolean	Read_Data(double *Lambda, Sequence *B_Seq, int *Cust_Nb);
void	Prepare_Data(Sequence Seq, double X0);
int	Interval_Nb(double Val, Sequence Seq);
/* Procedures for transient analysis */
void	Compute_Hinge_Coefs(int Cust_Nb, double	Lambda, Sequence Seq,
			    double **Xi);
void	Compute_Hinge_Coefs_Fast(int Cust_Nb, double	Lambda, Sequence Seq,
			    double **Xi);
void	Print_Range(double, double, double, double Lambda, int Cust_Nb,
		    Sequence Seq, double **Xi);
double	Compute_Moment(double Lambda, int Cust_Nb, Sequence Seq,
		       double **Xi, int Order);
/* Procedures for stationary analysis */
void	Compute_H_Coefs();
void	Compute_Phi_Coefs();
void	Compute_Tau_Coefs(double Lambda, Sequence Seq, double **Phi,
			  double *Tau);
void	Print_Stat_Range();
double	Compute_Stat_Moment();

int main(int argc, char *argv[])
{
  static char		Result_File_Name[256] = "";
  static char		Data_File_Name[256] = "";
  int			i;
  int			Cust_Nb = -2;
  int			New_Cust_Nb;
  boolean		Help = FALSE;
  boolean		Args_Ok = TRUE;
  boolean		File_Ok = TRUE;
  boolean		Ok = TRUE;
  boolean		Value = FALSE;
  boolean		Moment = FALSE;
  boolean		Stationary = FALSE;
  double		x_Start = 0.0;
  double		x_Stop = 10.0;
  double		x_Step = 0.1;
  double		x_Val;
  double		Lambda = 0.0;
  double		New_Lambda;
  int			Order = 1;	/* for the moments	*/
  Sequence		B_Seq;		/* Descr. of the sequence	*/
  int			Max_Index;	/* last interesting interval	*/
  double		**Xi; 		/* cf notes N5 et. seq. */
  double		**Phi;
  double		*Tau;
  double		*H;

  i = 1;

  if (argc>1)
    do {
      if ( 0 == strcmp( "-o", argv[i] ) ) {
	i++;
	if ( i < argc ) {
	  strcpy( Result_File_Name, argv[i] );
	  i++;
	}
	else {
	  printf("%s: filename expected\n", TOOLNAME );
	  Args_Ok = FALSE;
	}
      }
      else if ( 0 == strcmp( "-f", argv[i] ) ) {
	i++;
	if ( i < argc ) {
	  strcpy( Data_File_Name, argv[i] );
	  i++;
	}
	else {
	  printf("%s: filename expected\n", TOOLNAME );
	  Args_Ok = FALSE;
	}
      }
      else if ( 0 == strcmp( "-lambda", argv[i] ) ) {
	i++;
	if ( ( i < argc ) && ( 1 == sscanf( argv[i], "%lf", &Lambda ) ) )
	  i++;
	else {
	  printf("%s: real expected after %s\n", TOOLNAME,argv[i-1]);
	  Args_Ok = FALSE;
	}
      }
      else if ( 0 == strcmp( "-customer", argv[i] ) ) {
	i++;
	if ( ( i < argc ) && ( 1 == sscanf( argv[i], "%d", &Cust_Nb ) ) )
	  i++;
	else {
	  printf("%s: integer expected after %s\n", TOOLNAME, argv[i-1]);
	  Args_Ok = FALSE;
	}
      }
      else if ( 0 == strcmp( "-moment", argv[i] ) ) {
	i++;
	Moment = TRUE;
	if ( ( i < argc ) && ( 1 == sscanf( argv[i], "%d", &Order ) ) )
	  i++;
	else {
	  printf("%s: integer expected after %s\n", TOOLNAME, argv[i-1]);
	  Args_Ok = FALSE;
	}
      }
      else if ( 0 == strcmp( "-stationary", argv[i] ) ) {
	i++;
	Stationary = TRUE;
      }
      else if ( 0 == strcmp( "-xfrom", argv[i] ) ) {
	i++;
	if ( ( i < argc ) && ( 1 == sscanf( argv[i], "%lf", &x_Start ) ) )
	  i++;
	else {	printf("%s: real expected after %s\n", TOOLNAME, argv[i-1]);
	Args_Ok = FALSE;
	}
      }
      else if ( 0 == strcmp( "-xto", argv[i] ) ) {
	i++;
	if ( ( i < argc ) && ( 1 == sscanf( argv[i], "%lf", &x_Stop ) ) )
	  i++;
	else {
	  printf("%s: real expected after %s\n", TOOLNAME, argv[i-1]);
	  Args_Ok = FALSE;
	}
      }
      else if ( 0 == strcmp( "-xval", argv[i] ) ) {
	i++;
	if ( ( i < argc ) && ( 1 == sscanf( argv[i], "%lf", &x_Val ) ) ) {
	  i++;
	  Value = TRUE;
	}
	else {
	  printf("%s: real expected after %s\n", TOOLNAME, argv[i-1]);
	  Args_Ok = FALSE;
	}
      }
      else if ( 0 == strcmp( "-xstep", argv[i] ) ) {
	i++;
	if ( ( i < argc ) && ( 1 == sscanf( argv[i], "%lf", &x_Step ) ) )
	  i++;
	else {
	  printf("%s: real expected after %s\n", TOOLNAME, argv[i-1]);
	  Args_Ok = FALSE;
	}
      }
      else if ( 0 == strcmp( "-queue", argv[i] ) ) {
	Queue = TRUE;
	i++;
      }
      else if ( 0 == strcmp( "-log", argv[i] ) ) {
	Log = TRUE;
	i++;
      }
      else if ( 0 == strcmp( "-nonewline", argv[i] ) ) {
	i++;
	NewLine = FALSE;
      }
      else if ( 0 == strcmp( "-debug", argv[i] ) ) {
	Debug = TRUE;
	i++;
      }
      else if ( ( 0 == strcmp( "-h", argv[i] ) )
		|| ( 0 == strcmp( "-?", argv[i] ) ) ) {
	printf("Usage: %s [-f <file>] [-o <file>]\n\t", TOOLNAME );
	printf("[-lambda <real>] [-customer <int>] [-stationary]\n\t");
	printf("[-xfrom <real>] [-xto <real>] [-xval <real>] ");
	printf("[-xstep <real>]\n\t");
	printf("[-queue] [-log] [-moment <int>] ");
	printf("[-debug] [-h] [-?]\n");
	Args_Ok = FALSE;
	Help = TRUE;
	i++;
      }
      else if ( 0 == strcmp( "-v", argv[i] ) ) {
	printf("%s (Version %s)\n", TOOLNAME, VERSION);
	Args_Ok = FALSE;
	i++;
      }
      else	i++; /* skip unknown argument */
    }
    while ( i < argc );

  if (Args_Ok) {

    if ( Value ) {

      x_Start = x_Val;
      x_Stop = x_Val;
    }

    if ( strlen(Data_File_Name) != 0 )
      File_Ok = ( NULL != freopen( Data_File_Name, "r", stdin ) );

    if ( !File_Ok ) {
      printf("%s: error: could not open file %s\n", 
	     TOOLNAME, Data_File_Name );
      return -1;
    }
    else { 
      if ( ! Read_Data( &New_Lambda, &B_Seq, &New_Cust_Nb ) )
	return -1;      

      Prepare_Data( B_Seq, x_Start );

      /* overrides stationarity */
      if ( Stationary )
	Cust_Nb = -1;
      else if ( Cust_Nb == -2 )
	Cust_Nb = New_Cust_Nb;

      /* overrides Lambda */
      if ( Lambda == 0.0 )
	Lambda = New_Lambda;
      
      /* allocate here the memory */
      if ( Cust_Nb >= 0 ) {

	Xi = (double**) calloc( (unsigned) 1+Cust_Nb, sizeof( double* ) );
	Ok = ( NULL != Xi );
	for ( i=0; Ok && (i<=Cust_Nb); i++ ) {
	  Xi[i] = (double*) calloc((unsigned) 1+Cust_Nb, 
				   sizeof( double ) );
	  Ok = ( NULL != Xi[i] );
	}

	if ( ! Ok ) {
	  printf("Error: could not allocate memory.\n");
	  return -1;
	}
	else {

 	  Compute_Hinge_Coefs( Cust_Nb, Lambda, B_Seq, Xi );
	  /* Not used because not so fast... */
	  /* Compute_Hinge_Coefs_Fast( Cust_Nb, Lambda, B_Seq, Xi ); */

	  if ( strlen(Result_File_Name) != 0 )
	    freopen( Result_File_Name, "w", stdout );

	  if ( !Moment )
	    Print_Range( x_Start, x_Stop, x_Step, 
			 Lambda, Cust_Nb, B_Seq, Xi );
	  else {
	    printf("%f\n", 
		   Compute_Moment( Lambda, Cust_Nb, B_Seq, Xi, Order ) );
	  }

	  return 0;
	}
      }
      else if ( Lambda * B_Seq->Gamma >= 1 ) {
	printf("Unstable queue (rho=%8.4f). No stationary distribution.\n",
	       Lambda * B_Seq->Gamma );
	return -2;
      }
      else {
	
	if ( strlen(Result_File_Name) != 0 )
	  File_Ok = ( NULL != freopen( Result_File_Name, "w", stdout ) );

	if ( File_Ok ) {
	  /* Prepare and Fill the H table */
	  H = (double*) calloc( (unsigned)B_Seq->D, sizeof(double) );
	      
	  Compute_H_Coefs( Lambda, B_Seq, H );

	  /* Prepare and fill the Phi/Psi table */
	  if ( !Moment )
	    Max_Index = 1 + Interval_Nb( x_Stop, B_Seq );
	  else 
	    Max_Index = B_Seq->T;

	  Max_Index = ( Max_Index > B_Seq->T ? Max_Index : B_Seq->T );

	  Phi = (double**) calloc( (unsigned)1+Max_Index, sizeof( double ) );
	  Compute_Phi_Coefs( Lambda, B_Seq, H, Phi, Max_Index );

	  /* Prepare and fill the Tau/Pi table */
	  Tau = (double*) calloc( (unsigned)1+B_Seq->T, sizeof( double ) );
	  Compute_Tau_Coefs( Lambda, B_Seq, Phi, Tau );
	  
	  if ( Moment ) {
	    printf("%f\n", 
		   Compute_Stat_Moment( Lambda, B_Seq, Tau, H, Order ) );
	  }
	  else
	    Print_Stat_Range( x_Start, x_Stop, x_Step, Lambda, B_Seq, Phi );
	  
	  return 0;
	}
	else
	  return -1;
      }
    }
  }

  else if (Help)
    return 0;
  else
    return 1;
}

/************************************************************************/
/* Function Dif( alpha, beta, lambda, k ): the k-th derivative of:	*/
/* exp(alpha*s) (lambda/(lambda-s))^beta, taken at point s=0.		*/
/* Useful for transient and stationary moments.				*/
/************************************************************************/
double Dif(double  alpha,
	   int	   beta,
	   double  lambda,
	   int     k )
{
  double	Beta;
  double	Prod;
  double	Sum;
  int		j;

  Beta = (double) beta;

  if ( lambda == 0.0 )
    return 0.0;
  else {
    if ( k == 0 )
      return 1.0;
    else if ( k == 1 )
      return alpha + Beta / lambda;
    else if ( Beta == 0.0 ) {
      /* result is alpha^k, k>0 */
      Prod = alpha;
      for ( j = 2; j <= k; j++ )
	Prod *= alpha;
      return Prod;
    }
    else {
      Prod = Beta / lambda;
      for ( j = 2; j <= k; j++ )
	Prod *= (Beta+j-1)/lambda;
      Sum = Prod;
      for ( j = k-1; j >= 0; j-- ) {
	Prod *= (j+1) / ((double)(k-j)) / (Beta+j) * alpha * lambda;
	Sum += Prod;
      }
      return Sum;
    }
  }
}
/************************************************************************/
/* Utility procedures for handling values and intervals			*/
/************************************************************************/
/************************************************************************/
/* A function that returns the value of the upper bound of the n-th	*/
/* interval.								*/
/************************************************************************/
double Upper_Value(int N, Sequence Seq)
{
  int		T;
  int		D;
  int		Cycle_Nb;
  int   	i;
  int   	Phase;
  double	Shift;

  T = Seq->T;
  D = Seq->D;

  if ( N < T ) 
    return Seq->B[N+1];
  else {
    Cycle_Nb = ( N + 1 - T ) / D;
    Phase = ( N + 1 - T ) - D * Cycle_Nb;
    Shift = 0;
    for ( i=0; i<Phase; i++ )
      Shift += Seq->Inc[i];

    return Seq->B[T] + Cycle_Nb * Seq->Cycle_Length + Shift;
  }
}
/************************************************************************/
/* A function that returns the number of the B-interval the value Val	*/
/* falls in. 								*/
/************************************************************************/
int Interval_Nb(double		Val,
		Sequence	Seq )
{
  int		T;
  int		D;
  int		Cycle_Nb;
  int   	i;
  double	B_Sup;

  T = Seq->T;
  D = Seq->D;

  if ( Val < Seq->B[T] ) {
    for ( i=0; ( i<T ) && ( Val >= Seq->B[i] ); i++ );
    return i-1;
  }
  else {
    Cycle_Nb = (int) floor( ( Val - Seq->B[T] ) / Seq->Cycle_Length );
    B_Sup = Seq->B[T] + ( 1 + Cycle_Nb ) * Seq->Cycle_Length;
    for ( i=D-1; ( i>=0 ) && ( Val < B_Sup-Seq->Inc[i] ); i-- )
      B_Sup -= Seq->Inc[i];
    return T + Cycle_Nb * D + i;
  }
}

/************************************************************************/
/* Procedures for reading and preparing the data.			*/
/************************************************************************/
/* Build_C_Values: procedure for constructing the "extrapolated"	*/
/* Sequence, and its shifted version. Both are used in the computation	*/
/* of stationary values.						*/
/* This computation could be avoided if only transient is asked.	*/
/************************************************************************/
void Build_C_Values(Sequence Seq)
{
  int	m;
  int	r;
  int	D;
  int	T;
  int	Size;

  D = Seq->D;
  T = Seq->T;

  /* Filling the table up to max(D,T), to play safe... */
  Size = ( Seq->D > Seq->T ? Seq->D : Seq->T );

  Seq->C = (double*) calloc( (unsigned)(Size+1), sizeof(double) );
  Seq->CS = (double*) calloc( (unsigned)(D+1), sizeof(double) );

  Seq->C[T] = Seq->B[T];
  for( m=T-1; m>=0; m-- ) {
    Seq->C[m] = Seq->C[m+1] - Seq->Inc[ D - 1 - ( ( T - m - 1 ) % D ) ];
  }
  for( m=T+1; m<=D; m++ ) {
    Seq->C[m] = Seq->C[m-1] + Seq->Inc[ ( m - T - 1 ) % D ];
  }
  if ( Debug ) {
    printf("C=");
    for ( m=0; m<=Size; m++ )
      printf(" %f", Seq->C[m] );
    printf("\n");
  }

  /* now the shifted sequence */
  r = Seq->T % Seq->D;
  Seq->CS[0] = 0.0;
  for( m=1; m<=Seq->D; m++ )
    Seq->CS[m] = Seq->CS[m-1] + Seq->Inc[ ( m + r - 1 ) % Seq->D ];

}

boolean Read_Data(double	*Lambda,
		  Sequence	*B_Seq,
		  int		*Cust_Nb)
{
  boolean	Ok = TRUE;
  int		i;
  char		val[256];

  Ok = ( 1 == scanf( "%lf", Lambda ) );
  Ok = ( (*Lambda) > 0.0 );

  (*B_Seq) = (Sequence) malloc( sizeof( struct SEQUENCE ) );
  Ok = ( NULL != (*B_Seq) );

  if ( Ok ) {
    Ok = ( 1 == scanf( "%d", &((*B_Seq)->T) ) );
    Ok = ( 1 == scanf( "%d", &((*B_Seq)->D) ) );
    if ( (*B_Seq)->T < 0 ) {
      (*B_Seq)->T = 0;
      fprintf(stderr, "%s: warning: transient length set to 0\n", 
	      TOOLNAME );
    }

    if ( !Ok ) {
      fprintf(stderr, "%s: error: bad beginning of data file.\n", 
	      TOOLNAME );
      return FALSE;
    }
	
    Ok = ( NULL != 
	   ( (*B_Seq)->B = (double*) calloc( (unsigned)(*B_Seq)->T+1, 
					     sizeof(double) ) ) );
    if ( !Ok ) {
      fprintf(stderr, "%s: error: could not allocate memory. ", 
	      TOOLNAME );
      fprintf(stderr, "Transients too long.\n");
      return FALSE;
    }
    for ( i=0; (i<=(*B_Seq)->T) && Ok; i++ ) {
      scanf( "%s", val );
      if ( strcmp( val, "eps" ) == 0 ) {
	(*B_Seq)->B[i] = -MAXDOUBLE;
      } 
      else {
	Ok = Ok && ( 1 == sscanf( val, "%lf", &(*B_Seq)->B[i] ) );
      }
    }
    if ( !Ok ) {
      fprintf(stderr, "%s: error reading transients data: found \"%s\".\n", 
	      TOOLNAME, val );
      return FALSE;
    }

    Ok = ( NULL != 
	 ( (*B_Seq)->Inc = (double*) calloc( (unsigned)(*B_Seq)->D, 
					     sizeof(double) ) ) );
    if ( !Ok ) {
      fprintf(stderr, "%s: error: could not allocate memory. ", 
	      TOOLNAME );
      fprintf(stderr, "Period too long.\n");
      return FALSE;
    }
    for ( i=0; (i<(*B_Seq)->D) && Ok; i++ )
      Ok = Ok && ( 1 == scanf( "%lf", &(*B_Seq)->Inc[i] ) );

    if ( !Ok ) {
      fprintf(stderr, "%s: error reading period data.\n", TOOLNAME );
      return FALSE;
    }

    Ok = ( 1 == scanf("%d", Cust_Nb ) );
    if ( (*Cust_Nb) < -1 ) {
      (*Cust_Nb) = -1;
      fprintf(stderr, "%s: warning: stationary ", TOOLNAME );
      fprintf(stderr, "distribution assumed.\n");
    }

    return TRUE;
  }
  else {
    return FALSE;
  }
}

/************************************************************************/
/* Computation of Auxiliary values					*/
/* The parameter X0 is used to replace the "eps" that may be found	*/
/* in the "transient" part of the data.					*/
/************************************************************************/
void Prepare_Data( Sequence Seq, double X0 )
{
  int		i;
  double	G;

  /* eliminate epsilons */
  if ( Debug )
    printf("Transient[%d] = ", Seq->T );

  for ( i=0; i<=Seq->T; i++ ) {
    if ( Seq->B[i] == -MAXDOUBLE ) {
      Seq->B[i] = X0;
      if ( Debug ) 
	printf(" eps (-> %f)", Seq->B[i] );
    }
    else if ( Debug )
      printf(" %f", Seq->B[i] );
  }
  if ( Debug )
    printf("\n");

  /* compute the gamma */
  G = Seq->Inc[0];
  for ( i=1; i<Seq->D; i++ )
    G += Seq->Inc[i];
  Seq->Cycle_Length = G;
  Seq->Gamma = Seq->Cycle_Length / Seq->D;

  Build_C_Values( Seq );

  /* trace the values */
  if ( Debug ) {
    printf("Increments[%d] = ", Seq->D );
    for ( i=0; i<Seq->D; i++ )
      printf(" %f", Seq->Inc[i] );
    printf("\n");
    printf("Cycle length = %f\n", Seq->Cycle_Length );
    printf("Gamma = %f\n", Seq->Gamma );
    printf("C  = ");
    for ( i=0; i<=Seq->D; i++ )
      printf(" %f", Seq->C[i] );
    printf("\n");
    printf("C* = " );
    for ( i=0; i<=Seq->D; i++ )
      printf(" %f", Seq->CS[i] );
    printf("\n");
  }

}

/************************************************************************/
/* procedure pour le calcul d'un terme x^p/p! exp(-x) 			*/
/* p is (silently) assumed to be >=0					*/
/************************************************************************/
double	Poisson(double	x, int	p )
{
  int		i;
  double	Prod;

  if ( x == 0.0 )
    return ( p == 0 ? 1.0 : 0.0 );
  else {
    Prod = -x + (double)p * log( fabs( x ) );

    for ( i=1; i<=p; i++ )
      Prod -= log( (double)i );

    if ( 0 == p % 2 )
      return exp(Prod);
    else if ( x > 0 )
      return exp(Prod);
    else
      return -exp(Prod);
  }
}

/* procedure pour creer la table Xi */
void Compute_Hinge_Coefs(int		Cust_Nb,
			 double		Lambda,
			 Sequence	Seq,
			 double		**Xi )
{
  int		i, m, s;
  double	sum;
  double	prod;
  double	Rho;
  double	Delta;
  int		T;
  double	*B;
  double	*Inc;

  T = Seq->T;
  B = Seq->B;
  Inc = Seq->Inc;

  /* first, initialize Xi */
  for ( m=0; m <= Cust_Nb; m++ ) {
    Xi[Cust_Nb][m] = 1.0;
    if ( Debug )
      printf("%8.4f", Xi[Cust_Nb][m] );
  }
  if ( Debug )
    printf("\n");

  /* second, build recursively the Xis */
  for ( i=Cust_Nb-1; i >= 0; i-- ) {
    /* computation of the increment */
    if ( i < T )
      Delta = B[i+1] - B[i];
    else 
      Delta = Inc[ ( ( i - T ) % Seq->D ) ];

    Rho = Lambda * Delta;
    for ( m=0; m<=i; m++ ) {
      sum = Xi[i+1][m];
      prod = 1.0;
      for ( s=m+1; s<=i; s++ ) {
	prod *= ( Rho / (double) (s-m) );
	sum += prod * Xi[i+1][s];
      }
      Xi[i][m] = sum * exp( -Rho );
      if ( Debug )
	printf("%8.5f", Xi[i][m] );
    }
    if ( Debug )
      printf("\n");
  }
}

/* A faster (?) procedure for creating the Xi */
void Compute_Hinge_Coefs_Fast(int	Cust_Nb,
			      double	Lambda,
			      Sequence	Seq,
			      double	**Xi )
{
  int		i, m, s;
  double	sum;
  double	Rho;
  double	Log_Rho;
  double	Delta;
  int		T;
  double	*B;
  double	*Inc;
  double	*Log_Facto;

  T = Seq->T;
  B = Seq->B;
  Inc = Seq->Inc;

  /* first, initialize Xi */
  for ( m=0; m <= Cust_Nb; m++ ) {
    Xi[Cust_Nb][m] = 1.0;
    if ( Debug )
      printf("%8.4f", Xi[Cust_Nb][m] );
  }
  if ( Debug )
    printf("\n");

  /* Next, build factorials */
  Log_Facto = (double*) calloc( (unsigned)(1+Cust_Nb), sizeof(double) );
  Log_Facto[0] = 0.0;
  for( i=1; i<= Cust_Nb; i++ )
    Log_Facto[i] = Log_Facto[i-1] + log( (double)i );

  /* Finally, build recursively the Xis */
  for ( i=Cust_Nb-1; i >= 0; i-- ) {
    /* computation of the increment */
    if ( i < T )
      Delta = B[i+1] - B[i];
    else 
      Delta = Inc[ ( ( i - T ) % Seq->D ) ];

    Rho = Lambda * Delta;
    if ( Rho == 0.0 ) {
      Log_Rho = log( Lambda * Delta );
    }
    /* else value useless */
    for ( m=0; m<=i; m++ ) {
      sum = Xi[i+1][m];
      if ( Rho > 0 ) {
	for ( s=m+1; s<=i; s++ ) {
	  sum += exp( Log_Rho * (double) (s-m) - Log_Facto[ s-m ] )
	    * Xi[i+1][s];
	}
      }
      Xi[i][m] = sum * exp( -Rho );
      if ( Debug )
	printf("%8.5f", Xi[i][m] );
    }
    if ( Debug )
      printf("\n");
  }
}

/************************************************************************/
/* Computation of the numbers phi_s and tau_s for the stationary 	*/
/* distribution in the case d=1.					*/
/*                                                                      */
/* The computation of tau_s goes through that of 			*/
/* tau_tilde_s = tau_s * exp(-lambda*b-s).				*/
/************************************************************************/
void Compute_Phi_Coefs( Lambda, Seq, H, Phi, Nb_Vals )
double		Lambda;
Sequence	Seq;
double		*H;
double		**Phi;
int		Nb_Vals;
{
  int		i;
  int		p;
  int		r;
  int		s;
  double	Rho;
  double	E_D_Rho;    
  double	Sum;
  double	Prod;
  double	Delta;
  int		D;
  double	*C;
  double	*HS;		/* shifted sequence H */

  D = Seq->D;

  /* various initializations */
  Rho = Lambda * Seq->Gamma;
  E_D_Rho = exp( D * Rho );

  /* preparing the shifted sequences */
  HS = (double*) calloc( (unsigned)D, sizeof(double) );
  C = (double*) calloc( (unsigned)D, sizeof(double) );
  C[0] = 0.0;

  for ( r=0; (r<D) && (r<=Nb_Vals); r++ ) {
    
    /* Store the shifted sequences HS and C */
    for ( i=0; i<D; i++ )
      HS[i] = H[ ( r + i ) % D];
    for ( i=1; i<D; i++ )
      C[i] = C[i-1] + Seq->Inc[ ( r + i + D - 1 ) % D];
    if ( Debug ) {
      printf("H(%d) = ", r );
      for ( i=0; i<D; i++ )
	printf(" %f", HS[i] );
      printf("\n");
      printf("C(%d) = ", r );
      for ( i=0; i<D; i++ )
	printf(" %f", C[i] );
      printf("\n");
    }

    /* allocation of the table Phi[r] (not protected...) */
    Phi[r] = (double*) calloc( (unsigned)(1+Nb_Vals), sizeof(double) );

    /* start of the computation */
    Phi[r][0] = HS[0];
    if ( Debug )
      printf("Phi(%d)=[0]%8.6f", r, Phi[r][0] );

    for ( s=1; ( s <= D ) && ( s <= Nb_Vals ); s++ ) {
      Sum = HS[D-s] * Poisson( -D*Rho + Lambda * C[D-s], 0 );
      for ( p=1; p<s; p++ ) {
	Sum += HS[p+D-s] * Poisson( -D*Rho + Lambda * C[p+D-s], p );
      }
      Phi[r][s] = Sum;
      if ( Debug )
	printf("[%d]%8.6f", s, Phi[r][s] );
    }

    for ( s=D+1; s<=Nb_Vals; s++ ) {
      Sum = Phi[r][s-D];
      Delta = ( -D*Rho );
      Prod = 1.0;
      for ( i=1; i<=s-D-1; i++ ) {
	Prod *= Delta/(double) i;
	Sum += Prod * Phi[r][s-D-i];
      }
      Sum *= E_D_Rho;

      for ( p=0; p<D; p++ ) {
	Sum += Poisson( -D*Rho + Lambda * C[p], s-D+p ) * HS[p];
      }

      Phi[r][s] = Sum;
      if ( Debug )
	printf("[%d]%8.6f", s, Phi[r][s] );
    }
    if ( Debug )
      printf("\n");
  }

}

void Compute_Tau_Coefs(double	Lambda,
		       Sequence	Seq,
		       double	**Phi,
		       double	*Tau )
{
  int		i, j, s;
  double	E_BT;
  double	Sum;
  double	Prod;
  int		T;
  double	*B;

  T = Seq->T;
  /* tau computations use the **original B** sequence */
  B = Seq->B;

  E_BT = exp(-Lambda*B[T]);

  /* build Tau-tilde[] with recursion page N22 */
  Tau[T] = Phi[0][0] * E_BT;
  /* if debug, printing of "tau", not "tau-tilde" */
  if ( Debug )
    printf("Tau=[%d]%8.6f", T, Tau[T]*exp(Lambda*B[T]) );
	
  for ( s=T-1; s>=0; s-- ) {
    Sum = Phi[0][T-s];
    Prod = 1.0;
    for ( i=s+1; i<=T-1; i++ ) {
      Prod *= (Lambda*(B[T]-B[s]))/(double) (i-s);
      Sum += Prod * Phi[0][T-i];
    }
    Sum *= E_BT;
    for ( i=s+1; i<=T-1; i++ ) {
      /* compute Lambda*(B[i]-B[s]) ^ (i-s) / (i-s)! */
      Prod = 1.0;
      for ( j=1; j<=i-s; j++ )
	Prod *= (Lambda*(B[i]-B[s]))/(double) j;
      Sum -= Prod * Tau[i];
    }
    Tau[s] = Sum;
    if ( Debug )
      printf("[%d]%8.6f", s, Tau[s]*exp(Lambda*B[s]) );
  }
  if ( Debug )
    printf("\n");

  /* making the correction for tau[] */
  for ( s=0; s <= T; s++ )
    Tau[s] *= exp( Lambda*B[s] );

}

/* procedure pour imprimer les valeurs de la distribution */
void	Print_Range(double	Start,
		    double	Stop,
		    double	Step,
		    double	Lambda,
		    int		Cust_Nb,
		    Sequence	Seq,
		    double	**Xi )
{
  int		s;
  int		p;
  int		Cur_i;	/* index of the current interval */
  int		Nb_Points;
  double  	Val;
  double	B_Sup;
  double	BN;
  double	prod;
  double	F;

  if ( Start > Stop ) {
    fprintf( stderr, "%s: bad range\n", TOOLNAME );
    exit( 1 );
  }
  else if ( Step < 0.0 ) {
    fprintf( stderr, "%s: bad step value\n", TOOLNAME );
    exit( 1 );
  }

  Nb_Points = ( Stop == Start ?
		1 :
		1 + floor( ( Stop - Start ) / Step ) );

  if ( Debug ) {
    printf("%d-th waiting time distribution for the M/Known/1", Cust_Nb );
    printf(" on range [%6.3f,%8.3f]\n", Start, Stop );
  }

  /* compute first B[N] */
  BN = Upper_Value( Cust_Nb - 1, Seq );

  Cur_i = 0;

  for ( p=0; p<Nb_Points; p++ ) {

    Val = Start + p*Step;
    /* find the interval */

    Cur_i = Interval_Nb( Val, Seq );
    B_Sup = Upper_Value( Cur_i, Seq );

    if ( Debug )
      printf("%10.6f |", B_Sup );

    if ( Val < Seq->B[0] )
      F = 0.0;
    else if ( Val >= BN )
      F = 1.0;
    else if ( Val == B_Sup )
      F = Xi[Cur_i+1][0];
    else {	/* j'applique la convolution de la page N3 */
      F = Xi[Cur_i+1][0];
      prod = 1.0;
      for ( s=1; s<=Cur_i; s++ ) {
	prod *= Lambda * ( B_Sup - Val ) / (double) s;
	F += Xi[Cur_i+1][s] * prod;
      }
      F *= exp( Lambda * (Val - B_Sup) );
    }

    if ( !Queue )
      printf("%10.6f %12.8f", Val, F );
    else {
      if ( !Log )
	printf("%10.6f %12.8f", Val, 1-F );
      else
	printf("%10.6f %12.8f", Val, log(1-F) );
    }
    if ( NewLine )
      printf("\n");
  }
}

/************************************************************************/
/* Procedure for computing the limit "U" parameters needed for the 	*/
/* stationary distribution.						*/
/************************************************************************/
void Compute_U_Coefs( U, Lambda, T, B, Phi ) 
double	**U;
double	Lambda;
int	T;
double	*B;
double	**Phi;
{
  int		j;
  int		m;
  int		s;
  double	Rho;
  double	Som;
  double	Prod;

  for ( m=0; m<= T; m++ )
    U[T][T-m] = Phi[0][m];

  for ( j=T-1; j>=0; j-- ) {
    Rho = Lambda * (B[j+1] - B[j]);

    for ( m=0; m <= j; m++ ) {

      Prod = 1.0;
      Som = U[j+1][m];
      for ( s=1; s <= j-m; s++ ) {
	Prod *= ( Rho / (double) s );
	Som += Prod * U[j+1][m+s];
      }

      U[j][m] = Som * exp( -Rho );
    }
  }

  if ( Debug ) {
    for ( j=T; j>=0; j-- ) {
      printf("U");
      for ( m=0; m <= j; m++ ) {
	printf("[%d,%d]=%8.6f ", j, m, U[j][m] );
      }
      printf("\n");
    }    
  }

}
/************************************************************************/
/* Function computing the k-th moment of the transient distribution.	*/
/************************************************************************/
double	Compute_Moment(double	Lambda,
		       int	Cust_Nb,
		       Sequence	Seq,
		       double	**Xi,
		       int	Order )
{
  int		p;
  double  	Mom;
  double	BN;
  double	Bp;

  if ( Debug ) {
    printf("Moment of order %d of the %d-th waiting time in the M/Known/1",
	   Order, Cust_Nb );
    printf("\n");
  }

  BN = Upper_Value( Cust_Nb-1, Seq );
  if ( Debug )
    printf("B[%d] = %f\n", Cust_Nb, BN );

  if ( Order == 1 ) {
    /* use simplified formula e.g. p. N14 */
    Mom = 0.0;

    for ( p=0; p<=Cust_Nb; p++ ) {
      Mom += Xi[p][p];
      if ( Debug )
	printf( "%12.6f<", Mom );
    }

    Mom = Mom / Lambda;

    Mom += BN - (double)(Cust_Nb+1) / Lambda;

    return Mom;
  }
  else {
    /* use general formula with function Dif(). Cf. p. N14bis	*/
    /* But be careful not to confuse xi and chi!		*/
    Mom = Dif( -BN, Cust_Nb + 1, Lambda, Order );
    if ( Debug )
      printf( "%12.6f<", Mom );
    for ( p=0; p<=Cust_Nb; p++ ) {
      Bp = Upper_Value( p-1, Seq );
      Mom += Xi[p][p] * 
	( Dif( -Bp, p, Lambda, Order ) - Dif( -Bp, p+1, Lambda, Order ) );
      if ( Debug )
	printf( "%12.6f<", Mom );
    }
    if ( ( Order % 2 ) == 0 ) 
      return Mom;
    else
      return -Mom;
  }
}

/************************************************************************/
/* Procedure for printing the stationary distribution in the case d=1.	*/
/************************************************************************/
void	Print_Stat_Range( 
double		Start,
double		Stop,
double		Step,
double		Lambda,
Sequence	Seq,
double		**Phi )
{
  int		i;
  int		s;
  int		r;
  int		p; 
  int		Cur_i;		/* index of the current interval 	*/
  int		Nb_Points;	/* nb of points to compute		*/
  double  	Val;
  double	B_Sup;
  double	Prod;
  double	F;
  double	Rho;
  int		T;
  double	*B;
  double	**U;

  /* renaming for ease of use */
  T = Seq->T;
  B = Seq->B;

  /* taking care of degenerate cases */
  if ( Start > Stop ) {
    fprintf( stderr, "%s: bad range\n", TOOLNAME );
    exit( 1 );
  }
  else if ( Step < 0.0 ) {
    fprintf( stderr, "%s: bad step value\n", TOOLNAME );
    exit( 1 );
  }

  Nb_Points = ( Stop == Start ?
		1 :
		1 + floor( ( Stop - Start ) / Step ) );

  if ( Debug ) {
    printf("Stationary Waiting time distribution for the M/Known/1" );
    printf(" on range [%f,%f]\n", Start, Stop );
  }

  /* preliminary computations */
  Rho = Lambda * Seq->Gamma;

  /* Now compute the U coefficients, U is a triangular array 	*/
  /* Useful only if x_start < B[T]. 				*/
  if ( Start < B[T] ) {
    U = (double**) calloc( (unsigned)T+1, sizeof(double*) );
    for ( i=0; i<=T; i++ )
      U[i] = (double*) calloc( (unsigned)i+1, sizeof(double) );
    
    Compute_U_Coefs( U, Lambda, T, B, Phi );
  }

  Cur_i = 0;

  for ( p=0; p<Nb_Points; p++ ) {

    Val = Start + p*Step;
    /* find the interval */

    Cur_i = Interval_Nb( Val, Seq );
    B_Sup = Upper_Value( Cur_i, Seq );

    if ( Debug )
      printf("%10.6f |", B_Sup );

      if ( Val < B[0] )
	F = 0.0;
      else if ( Cur_i < T ) {
	F = U[Cur_i+1][0];
	Prod = 1.0;
	Rho = Lambda * ( B_Sup - Val );
	for ( s=1; s<=Cur_i; s++ ) {
	  Prod *= Rho / (double) s;
	  F += U[Cur_i+1][s] * Prod;
	}
	F *= exp( -Rho );
      }
	
      else {
	/* Convolution de la page N3/Cor. 3.4, precisee page LIM11 */
	r = ( Cur_i + 1 - T ) % Seq->D;
	F = Phi[r][Cur_i+1];
	Prod = 1.0;
	Rho = Lambda * ( B_Sup - Val );
	for ( s=1; s<=Cur_i; s++ ) {
	  Prod *= Rho / (double) s;
	  F += Phi[r][Cur_i + 1 -s] * Prod;
	}
	F *= exp( -Rho );
      }

      if ( !Queue )
	printf("%10.6f %12.8f", Val, F );
      else {
	if ( !Log )
	  printf("%10.6f %12.8f", Val, 1-F );
	else
	  printf("%10.6f %12.8f", Val, log(1-F) );
      }
      if ( NewLine )
	printf("\n");
    }
}

/************************************************************************/
/* Procedure to compute the moments of the stationary distribution	*/
/* It assumes stability.						*/
/************************************************************************/
double	Compute_Stat_Moment( Lambda, Seq, Tau, H, Order )
double		Lambda;
Sequence	Seq;
double		*Tau;
  double	*H;
int		Order;
{
  int		p;
  int		q;
  double  	Mom;
  double	Rho;
  double	Un_M_Rho;
  double	EW_Pure;
  double	Compens;
  int		D;
  int		T;
  double	*C;
  double	M1;
  double	M2;
  double	B1;
  double	B2;
  double	Bp;
  int		r;	/* the shift to give to H */

  D = Seq->D;
  T = Seq->T;
  C = Seq->C;

  Rho = Lambda * Seq->Gamma;
  Un_M_Rho = 1.0 - Rho;

  if ( Debug ) {
    printf("Moment of order %d of the stationary waiting time ", Order );
    printf("in the M/Known/1\n");
  }
  if ( Order == 0 )
    return 1.0;
  else {
    /* In order to be used here, sequence H has to be shifted wrt the	*/
    /* one computed starting from the sequence of increments.		*/
    /* The shift makes it sure that h_T = Phi[0] = Tau[T].		*/
    r = ( -T % D ) + D;

    M1 = - 0.5 / Un_M_Rho * ( 1.0 + (double)D * Un_M_Rho * Un_M_Rho );

    B1 = 0.0;
    for ( p=0; p<=D-1; p++ )
      B1 += H[(p+r)%D] * ( (p+1) - Lambda*C[p] );
    B1 /= ( D * Un_M_Rho );

    if ( Order == 1 ) {

      Compens = 0.0;
      for ( p=0; p<=T-1; p++ ) {
	Compens += ( Tau[p] - H[(p+r) % D] );
	if ( Debug )
	  printf( "%12.6f<", Compens );
      }
      Compens = Compens / Lambda;

      /* EW for the "pure" sequence */
      /* 	Mom += Rho*Gamma / ( 1.0 - Rho ) / 2.0; */
      /* "Old" way of computing it, kept for the time being */
      EW_Pure = 0.0;
      for ( q=0; q<D; q++ )
	EW_Pure += H[(q+r)%D] * ( Lambda * C[q] - q );
      EW_Pure *= ( 2.0 / (double)D );
      EW_Pure += ( D*(1.0-Rho)*(1.0-Rho) + 2.0 * Rho - 1.0 );
      EW_Pure /= ( 2.0 * Lambda * ( 1.0 - Rho ) );
      /* Comparison of the two ways, just in case... */
      if ( EW_Pure != ( - ( M1 + B1 ) / Lambda ) )
	printf("%s: warning! dubious computation (dif=%e)\n", 
	       TOOLNAME, EW_Pure + ( M1 + B1 ) / Lambda );

      Mom = ( - ( M1 + B1 ) / Lambda ) + Compens;

      return Mom;
    }
    else if ( Order == 2 ) {

      Compens = 0.0;
      for ( p=0; p<=T-1; p++ ) {
	Bp = Upper_Value( p-1, Seq );
	Compens += ( Tau[p] * ( Dif( -Bp, p, Lambda, 2 ) 
				- Dif( -Bp, p+1, Lambda, 2 ) )
		     - H[(p+r)% D] * ( Dif( -C[p], p, Lambda, 2 ) 
				- Dif( -C[p], p+1, Lambda, 2 ) ) );
	if ( Debug )
	  printf( "%12.6f<", Compens );
      }

      M2 = ( D*D - 1.0 + 4.0*Rho*(1.0-D*D) + 6.0*Rho*Rho*D*D 
	     - 4.0*Rho*Rho*Rho*D*D + Rho*Rho*Rho*Rho*D*D ) / 12.0
	/ ( Un_M_Rho * Un_M_Rho );
      B2 = 0.0;
      for ( p=0; p<=D-1; p++ )
	B2 += H[(p+r)%D] * ( (p+2)*(p+1) - 2*(p+1)*Lambda*C[p] 
		       + Lambda*Lambda*C[p]*C[p] );
      B2 *= ( 0.5 / ( D * Un_M_Rho ) );
      
      Mom = 2.0 * ( M2 + M1*B1 + B2 ) / Lambda/Lambda + Compens;

      return Mom;
    }
    else {
      printf("Moments of order > 2 not implemented yet...\n");
      return 0.0;
    }
  }
}

/************************************************************************/
/* Procedures pour le calcul quand d>1				      	*/
/************************************************************************/
double K( l, d, a, z )
int	l;
int	d;
double  a;
double  z;
{
  double	Sum;
  double	Prod;
  double	Term;
  double	r;
  double	d_log_z;
  int		m;
  int		j;
  int		sgn;
  int		M_Max = 10000;	/* max number of iterations */
  boolean	Continue = TRUE;
  double	Epsilon = 1e-8;

  if ( z <= 0 ) {
    printf( "Error: call to k with negative z (%f)\n", z );
    return 0.0;
  }
  else if ( a == 0.0 ) {
    if ( l == 0 )
      return 1.0;
    else
      return 0.0;
  }
  else {
      
    Sum = 0.0;
    Prod = (double)l * log( z );
    d_log_z = (double)d * log( z );

    for( j=1; j<=l; j++ )
      Prod -= log( (double)j );

    for ( m=0; Continue && ( m < M_Max ); m++ ) {

      if ( m*d+l+a == 0.0 ) {
	if ( m*d+l-1 == 0 )
	  Term = z;
	else
	  Term = 0.0;
	Sum += Term;
      }
      else if ( m*d+l+a < 0.0 ) {
	sgn = 1 - 2 * ( (m*d+l+1) % 2 );
	Term = exp( Prod + (m*d+l-1)*log( fabs(m*d+l+a) ) );       
        Sum += sgn * Term;
      }
      else {
	Term = exp( Prod + (m*d+l-1)*log( m*d+l+a ) );
	Sum += Term;
      }

/*       printf("%e", Term ); */

      if ( ( Term < Epsilon ) && ( m*d+a+l > 0.0 ) ) {
	r = exp( d_log_z + d * ( 0.5 + (double)(d-3)/(m*d+a+l)) );	
/* 	printf(" %e", Epsilon * (1-r) ); */
 	Continue = ( Term > Epsilon * ( 1 - r ) );
      }
/*       printf(" %f\n", Sum ); */

      Prod += d_log_z;
      for( j=m*d+l+1; j<=m*d+l+d; j++ )
	Prod -= log( (double)j );

    }
    if ( Continue ) {
      printf( "Warning: computation of k does not converge (val=%f)\n", 
	      Sum );
    }
    return Sum * a;
  }
}

void Dump_Matrix(int	Size,
		 double	**A,
		 double	*X )
{
  int i,j;

  for( i=0; i<Size; i++ ) {
    for( j=0; j<Size; j++ ) {
      printf(" %10f", A[i][j] );
    }
    printf("| %10f\n", X[i] );
  }
}

/************************************************************************/
/* Procedure Compute_H_Coefs()						*/
/*   Fills in the table H with the values H[0]..H[D-1] computed for	*/
/*   the increment sequence given in Seq, that is,starting from b[T]	*/
/************************************************************************/
void Compute_H_Coefs(double	Lambda,
		     Sequence	Seq,
		     double	*H )
{
  int		i;
  int		j;
  int		k;
  int		l;
  double	*Nu;
  double	*TH;
  double	Z;
  double	Rho;
  double	Max_Val;
  int		Max_Arg;
  double	Rat;
  double	Sum;
  double	**B;		/* Here B is the Matrix */
  double	Tmp;
  int		D;
  double	*Inc;

  D = Seq->D;
  Inc = Seq->Inc;

  Rho = Lambda * Seq->Gamma;
  if ( Debug ) 
    printf("Rho = %8.6f\n", Rho );
  Z = Rho * exp( -Rho );

  if ( D == 1 )
    H[0] = 1 - Rho;
  else {
	
    /* allocation and computation of "increment" coefficients */
    Nu = (double*) calloc( (unsigned)D, sizeof(double) );

    Nu[0] = 0.0;
    for( i=1; i<D; i++ ) {
      Nu[i] = Nu[i-1] + Rho - Lambda * Inc[ D-i ];
    }

    if ( Debug ) {
      printf("Vector Delta = ");
      for( i=0; i<D; i++ )
	printf(" %f", Inc[i] );
      printf("\n");
      printf("Vector Nu = ");
      for( i=0; i<D; i++ )
	printf(" %f", Nu[i] );
      printf("\n");
    }
    
    B = (double**) calloc( (unsigned)D, sizeof(double*) );
    for( i=0; i<D; i++ ) {
      B[i] = (double*) calloc( (unsigned)D, sizeof(double) );
    }

    /* filling the matrix */
    for( i=0; i<D; i++ ) {
      for( j=0; j<D; j++ ) {
	B[i][j] = K( ( 2*D - i - j ) % D, D, Nu[j]/Rho, Z ) * exp( -Nu[j] );
      }
    }

    /* Preparing the vector h */
    TH = (double*) calloc( (unsigned)D, sizeof(double) );
    for ( i=0; i<D; i++ ) {
      TH[i] = ( 1.0 - Rho );
    }

    if ( Debug ) {
      printf("Matrix B | Vector H:\n");
      Dump_Matrix( D, B, TH );
    }

    /* checking for "unity" */
    if ( Debug ) {
      printf(" 1.B - 1 = ");
      for ( j=0; j<D; j++ ) {
	Sum = B[0][j];
	for( i=1; i<D; i++ )
	  Sum += B[i][j];
	printf(" %f", Sum - 1.0 );
      }
      printf("\n");
    }

    /* Inverting the matrix with Gauss' method */
    for( j=0; j<D-1; j++ ) {
      /* Finding the pivot */
      Max_Val = fabs( B[j][j] );
      Max_Arg = j;
      for( k=j; k<D; k++ )
	if ( fabs( B[k][j] ) > Max_Val ) {
	  Max_Val = fabs( B[k][j] );
	  Max_Arg = k;
	}
      if ( Max_Val == 0.0 ) {
	printf("Error! Singular matrix B.\n");
	exit(-4);
      }
      if ( Max_Arg != j ) {
	/* need to permute rows. beark. */
	if ( Debug )
	  printf("Swapping rows %d <-> %d\n", j, Max_Arg );
	for ( l=j; l<D; l++ ) {
	  Tmp = B[j][l]; B[j][l] = B[Max_Arg][l]; B[Max_Arg][l] = Tmp;
	}
	Tmp = TH[j]; TH[j] = TH[Max_Arg]; TH[Max_Arg] = Tmp;
      }
      for( i=j+1; i<D; i++ ) {
	Rat = -B[i][j] / B[j][j];
	for ( k=j; k<D; k++ ) {
	  B[i][k] += Rat * B[j][k];
	}
	TH[i] += Rat * TH[j];
      }
      if ( Debug ) {
	printf("Gauss/Matrix B[%d] = \n", j );    
	Dump_Matrix( D, B, TH );
      }
    }
    /* Gauss phase 2: normalizing */
    for( i=0; i<D; i++ ) {
      Rat = 1.0/B[i][i];
      TH[i] *= Rat;
      for ( j=i; j<D; j++ )
	B[i][j] *= Rat;
    }
    if ( Debug ) {
      printf("Gauss/Matrix B/N = \n" );
      Dump_Matrix( D, B, H );
    }
    /* Gauss phase 2: going up! */
    for( j=D-1; j>=0; j-- ) {
      for( i=0; i<j; i++ ) {
	TH[i] -= B[i][j] * TH[j];
      }
      if ( Debug ) {
	printf("Gauss2/Matrix H[%d] =", j );
	for( k=0; k<D; k++ )
	  printf(" %f", TH[k] );
	printf("\n");
      }
    }

    /* finally, taking care of the "distorsion" of H */
    H[0] = TH[0];
    for( i=1; i<D; i++ )
      H[i] = TH[D-i];
    
    if ( Debug ) {
      printf("Vector H = ");
      for( i=0; i<D; i++ )
	printf(" %f", H[i] );
      printf("\n");
    }
  }
}
