/****************************************************************/
/* Partie relative a lex/yacc					*/
/****************************************************************/

#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include "taskgraph.h"
#include "tgl.tab.h"

#define	E_MEM(x) fprintf(stderr,"Warning: not enough memory: %s not created\n",x);
#define W_EXT(x) fprintf(stderr,"Warning: extra parameters for %c (ignored)\n", x );

extern FILE *yyin;
extern int yydebug;
extern char yytext[];
#ifdef YY_FLEX_LEX_COMPAT
extern int yylineno;
#endif
extern int yyparse();
extern int yylex();
extern int TGL_lineno;

/* forced external declaration */
extern char *strdup(const char *);

/* forward declaration */
void dumpe_network(Graph *TG);

char	**TaskIdents;	/* tableau of task identifiers 	*/
char	**LinkIdents;	/* tableau of link identifiers	*/
int	Nb_Known_Tasks = 0;
int	Nb_Known_Links = 0;
boolean	Parse_Debug = FALSE;
boolean	Debug = FALSE;
#ifdef USE_YYLVAL
extern  YYSTYPE yylval;
#else 
Graph   *The_Task_Graph;
#endif

Graph *Read_Network(char *File_Name)
{
  FILE	*Input_File;
  FILE*	Current_Input;
  int	Res;

  if ( strlen( File_Name ) > 0 ) {
    Input_File = fopen( File_Name, "r" );
  }
  else {
    Input_File = stdin;
  }

  if ( NULL != Input_File ) {
    Current_Input = yyin;
    yyin = Input_File;

    yydebug = 0;
    Res = yyparse();

    yyin = Current_Input;

    fclose( Input_File );

    if ( Res == 0 ) 
#ifdef USE_YYLVAL
      return yylval.un_reseau;
#else
      return The_Task_Graph;
#endif
    else
      return (Graph*)NULL;
  }
  else
    return (Graph*)NULL;
}

void yyerror(char *s)
{
#ifdef YY_LEX_FLEX_COMPAT
  printf("---%s---[%s] at line %d\n",s,yytext,yylineno);
#else
  printf("---%s---[%s] at line %d\n",s,yytext,TGL_lineno);
#endif
}

/****************************************************************/
/* Procedures degestion des identificateurs			*/
/****************************************************************/
/* function Task_Number:	returns the number associated 	*/
/* with a given name. If the name does not exist yet, insert	*/
/* it in the table.						*/
int	Task_Number( char	*The_Name )
{	
  int	i;

  for ( i=0; ( i<Nb_Known_Tasks ) 
	  && ( 0 != strcmp( The_Name, TaskIdents[i] ) );
	i++ );
  if ( i == Nb_Known_Tasks ) {
    if ( Nb_Known_Tasks == 0 )
      TaskIdents = (char**) malloc( sizeof( char* ) );
    else	
      TaskIdents = (char**) realloc( TaskIdents, 
				     (1+Nb_Known_Tasks)*sizeof( char* ) );
    TaskIdents[ Nb_Known_Tasks ] = strdup( The_Name );
    if (Parse_Debug)
      printf( "Learning about task #%d (%s)\n",
	      Nb_Known_Tasks, 
	      TaskIdents[ Nb_Known_Tasks ] );
    Nb_Known_Tasks++;
  }
  else if (Parse_Debug) {
    printf("- %s is task number %d\n", The_Name, i );
  }

  return	i;
}

/****************************************************************/
/* function Link_Number:	returns the number associated 	*/
/* with a given name. If the name does not exist yet, insert	*/
/* it in the table.						*/
int	Link_Number( char	*The_Name )
{	
  int	i;

  for ( i=0; ( i<Nb_Known_Links ) 
	  && ( 0 != strcmp( The_Name, LinkIdents[i] ) );
	i++ );
  if ( i == Nb_Known_Links ) {
    if ( Nb_Known_Links == 0 )
      LinkIdents = (char**) malloc( sizeof( char* ) );
    else	
      LinkIdents = (char **) realloc( LinkIdents, 
				      (1+Nb_Known_Links)*sizeof( char* ) );
    LinkIdents[ Nb_Known_Links ] = strdup( The_Name );
    if (Parse_Debug)
      printf( "Learning about link #%d (%s)\n",
	      Nb_Known_Links, 
	      LinkIdents[ Nb_Known_Links ] );
    Nb_Known_Links++;
  }

  return	i;
}

/****************************************************************/
/* Procedures de construction du reseau				*/
/****************************************************************/

Graph *Parse_TG(Liste_Task	The_Tasks,
		Liste_Link	The_Links )
{
  Graph		*TG;
  Liste_Task 	T_Scan;
  Liste_Link	L_Scan;
  Task	 	*The_Task;
  int		The_Size;
  int		The_Orig;
  Liste_Link	The_L_List;

  /* trouver la taille en parcourant la liste des taches		*/
  The_Size = 0;
  for ( T_Scan = The_Tasks; T_Scan != NULL; T_Scan = T_Scan->Next )
    The_Size++;

  /* allouer la place pour le reseau et son tableau de trans. */
  TG = (Graph*) malloc( sizeof ( Graph ) );
  if (TG != NULL ) {

    TG->Size = The_Size;
    TG->T = (Task*) calloc( TG->Size, sizeof( Task ) );
  }

  /* scanner la liste des trans. et la mettre dans le tableau	*/
  for ( T_Scan = The_Tasks; T_Scan != NULL ; T_Scan = T_Scan->Next ) {

    if ( T_Scan->Val->Task_Id <= The_Size ) {

      The_Task = T_Scan->Val;
      TG->T[ The_Task->Task_Id ] = *(The_Task);
    }
    else {
      fprintf(stderr,"Task number > Size.");
      fprintf(stderr," Task '%d' ignored.\n", The_Task->Task_Id );
    }
  }

  /* Write out the list of links. (I'm looking for a difficult bug!) */
  if (Parse_Debug) {
    for ( L_Scan = The_Links; L_Scan != NULL ; L_Scan = L_Scan->Next ) {
      printf("Link number %d name=%s from %d to %d\n",
	     L_Scan->Val->Link_Id,
	     L_Scan->Val->Link_Name,
	     L_Scan->Val->Task_From,
	     L_Scan->Val->Task_To );
    }
  }
  /* scanner la liste des liens et les mettre dans les trans.	*/
  for ( L_Scan = The_Links; L_Scan != NULL ; L_Scan = L_Scan->Next ) {

    The_Orig = L_Scan->Val->Task_From;
    The_L_List = (Liste_Link) malloc( sizeof( struct LINK_LIST ) );
    The_L_List->Val = L_Scan->Val;
    The_L_List->Next = TG->T[The_Orig].Neighbors;
    TG->T[The_Orig].Neighbors = The_L_List;
  }

  return TG;

}

Task *Parse_Task(char		*The_Name,
		 char		*The_Given_Name,
		 Law_Desc	The_Law )
{	
  Task		*The_Task;

  The_Task = (Task*) malloc( sizeof( Task ) );

  if ( The_Task != NULL )	{	
    The_Task->Task_Id = Task_Number( The_Name );
    The_Task->Task_Name = strdup( The_Name );
    if ( NULL != The_Given_Name ) {
      The_Task->Task_Given_Name = strdup( The_Given_Name );
    }
    else {
      The_Task->Task_Given_Name = (char*)NULL;
    }
    The_Task->Task_Law = The_Law;
    The_Task->Neighbors = NULL;
    The_Task->Prev = NULL;
  }
  else {
    E_MEM("Task");
  }

  return The_Task;
}

Link_Desc Parse_Link( 
char		*The_Name,
Link_Desc	The_Link )
{
  The_Link->Link_Id = Link_Number( The_Name );
  The_Link->Link_Name = The_Name;
  
  return The_Link;
}

Link_Desc Parse_Data_Link( 
SType		The_Synchro,
BiPoint		The_Ends,
Law_Desc	The_Law )
{
  Link_Desc	Resultat;

  Resultat = (Link_Desc) malloc( sizeof ( struct LINK_DESC ) );
  Resultat->Link_Type = PRECEDENCE;
  Resultat->Synchro_Type = The_Synchro;
  Resultat->Link_Law = The_Law;
  Resultat->Task_From = The_Ends->Origin;
  Resultat->Task_To = The_Ends->Destination;
  Resultat->Depth = 0;

  free(The_Ends);

  return Resultat;
}

Link_Desc Parse_Program_Link( 
SType		The_Synchro,
BiPoint		The_Ends,
int		The_Depth )
{
  Link_Desc	Resultat;

  Resultat = (Link_Desc) malloc( sizeof ( struct LINK_DESC ) );
  Resultat->Link_Type = ITERATION;
  Resultat->Synchro_Type = The_Synchro;
  Resultat->Link_Law = NULL;
  Resultat->Task_From = The_Ends->Origin;
  Resultat->Task_To = The_Ends->Destination;
  Resultat->Depth = The_Depth;

  return Resultat;
}

Liste_Task Append_Task(Liste_Task Task_List,
		       Task	*New_Task )
{
  Liste_Task The_List;

  The_List = (Liste_Task) malloc( sizeof( struct TASK_LIST ) );

  if ( The_List != NULL ) {
    The_List->Val = New_Task;
    The_List->Next = Task_List;
  }
  else {
    fprintf(stderr,
	    "Pas assez de memoire. Task non cree.\n");
  }

  return( The_List );
}

Liste_Link Append_Link(Liste_Link Link_List,
		       Link_Desc  New_Link )
{
  Liste_Link The_List;

  The_List = (Liste_Link) malloc( sizeof( struct LINK_LIST ) );

  if ( The_List != NULL ) {
    The_List->Val = New_Link;
    The_List->Next = Link_List;
  }
  else {
    fprintf(stderr,
	    "Pas assez de memoire. Link non cree.\n");
  }

  return( The_List );
}

/****************************************************************/
/* Procedure de Trace						*/
/****************************************************************/
void dumpe_network(Graph *TG)
{
  int		i;
  int		j;
  boolean	Is_Neighbor;
  Liste_Link	L_Scan;
  Link_Desc	The_Link;

  printf("Reseau de taille %d\n",TG->Size);

  printf("   ");
  for (i=0; i<2*TG->Size+3; i++)
    printf(".");
  printf("\n");
	
  for (i=0; i<TG->Size; i++) {
    printf("%3d.",i+1);
    for (j=0; j<TG->Size; j++) {
      Is_Neighbor = FALSE;
      for ( L_Scan = TG->T[i].Neighbors; 
	    ( L_Scan != NULL ) && !Is_Neighbor;
	    L_Scan = L_Scan -> Next ) {
	Is_Neighbor = ( 
		       ( j == L_Scan->Val->Task_To )
		       &&  ( L_Scan->Val->Link_Type == PRECEDENCE )
		       );
	if (Is_Neighbor)
	  The_Link = L_Scan->Val;
      }

      if (!Is_Neighbor)
	printf(" ");
      else
	printf("%c",The_Link->Link_Law->Name);
    }
    printf(".");
    for (j=0; j<TG->Size; j++)
      {	Is_Neighbor = FALSE;
      for ( L_Scan = TG->T[i].Neighbors; 
	    ( L_Scan != NULL ) && !Is_Neighbor;
	    L_Scan = L_Scan -> Next ) {
	Is_Neighbor = ( ( j == L_Scan->Val->Task_To )
			&&  ( L_Scan->Val->Link_Type == ITERATION )
			);
	if (Is_Neighbor)
	  The_Link = L_Scan->Val;
      }

      if (!Is_Neighbor)
	printf(" ");
      else
	printf("%1d",The_Link->Depth);
      }
    printf(".\n");
  }

  printf("   ");
  for (i=0; i<2*TG->Size+3; i++)
    printf(".");
  printf("\n\n");

  printf("Dependence maximale: %d\n\n",TG->Max_Dependency);

  for (i=0; i<TG->Nb_Layers; i++)
    {	printf("Level %d : ",i+1);
    for (j=0; j<TG->Level[i]->Nb_Tasks; j++)
      printf(" %s(%d)",
	     TG->T[TG->Level[i]->Task_Id[j]].Task_Name,
	     TG->Level[i]->Task_Id[j]+1);
    printf("\n");
    }
  printf("\n");

  for (i=0; i<TG->Size; i++)
    {	printf("Tasks amont de %d: ",i+1);
    if (TG->T[i].Prev != NULL)
      for (j=0; j<TG->T[i].Prev->Nb_Tasks; j++)
	printf(" %s(%d)",
	       TG->T[TG->T[i].Prev->Task_Id[j]].Task_Name,
	       TG->T[i].Prev->Task_Id[j]+1);
    else{	printf(" inconnues.\n");
    }
    printf("\n");
    }
}

Liste_Link Extract_Links( Liste_Object The_Objects )
{
  Liste_Object	O_Scan;
  Liste_Link	The_Links;
  Liste_Link	Some_Link;

  The_Links = NULL;

  for( O_Scan = The_Objects; O_Scan != NULL; O_Scan = O_Scan->Next ) {
    if ( O_Scan->A_Link != NULL ) {
      Some_Link = (Liste_Link) malloc( sizeof( struct LINK_LIST ) );
      Some_Link->Val = O_Scan->A_Link;
      Some_Link->Next = The_Links;
      The_Links = Some_Link;
    }
  }

  return The_Links;
}

Liste_Task Extract_Tasks( Liste_Object The_Objects )
{
  Liste_Object	O_Scan;
  Liste_Task	The_Tasks, Some_Task;

  The_Tasks = NULL;

  for( O_Scan = The_Objects; O_Scan != NULL; O_Scan = O_Scan->Next )
    if ( O_Scan->A_Task != NULL ) {
      Some_Task = (Liste_Task) malloc( sizeof( struct TASK_LIST ) );
      Some_Task->Val = O_Scan->A_Task;
      Some_Task->Next = The_Tasks;
      The_Tasks = Some_Task;
    }

  return The_Tasks;
}

Liste_Object Task2Object(Task	*The_Task)
{
  Liste_Object	Resultat;

  Resultat = (Liste_Object) malloc( sizeof( struct OBJ_LIST ) );

  Resultat -> A_Task = The_Task;
  Resultat -> A_Link = NULL;

  return	Resultat;
}

Liste_Object Link2Object(Link_Desc	The_Link)
{
  Liste_Object	Resultat;

  Resultat = (Liste_Object) malloc( sizeof( struct OBJ_LIST ) );

  Resultat -> A_Task = NULL;
  Resultat -> A_Link = The_Link;

  return Resultat;
}

BiPoint Parse_Ends(char	*The_Orig,
		   char	*The_Dest )
{
  BiPoint	Resultat;

  Resultat = (BiPoint) malloc( sizeof( struct BIPOINT ) );

  Resultat->Origin = Task_Number( The_Orig );
  Resultat->Destination = Task_Number( The_Dest );

  free( The_Orig );
  free( The_Dest );

  return	Resultat;
}

/* appends the second object to the list */
Liste_Object Append_Object(Liste_Object	The_List,
			   Liste_Object	The_Obj )
{
  The_Obj -> Next = The_List;

  return	The_Obj;
}

/************************************************************************/
/* Procedures to deallocate a Task Graph.				*/
/* Actually not useful before called before the end of the program.	*/
/* Might be if several graphs were used, or if C++ were used.		*/
/************************************************************************/
void Free_Object_List( Liste_Object	The_List )
{
  Liste_Object	O_Scan;
  Liste_Object	P_Scan;

  for ( O_Scan = The_List; O_Scan != NULL; ) {
    P_Scan = O_Scan;
    O_Scan = O_Scan->Next;
    free( P_Scan );
  }
}

void Free_Link_List( Liste_Link	The_List )
{
  Liste_Link	L_Scan;
  Liste_Link	M_Scan;

  for ( L_Scan = The_List; L_Scan != NULL; ) {
    M_Scan = L_Scan;
    L_Scan = L_Scan->Next;
    free( M_Scan->Val->Link_Name );
    /* A link has not always a law */
    if ( (Law_Desc)NULL != M_Scan->Val->Link_Law )
      Free_Law( M_Scan->Val->Link_Law );
    free( M_Scan->Val );
    free( (char*) M_Scan );
  }
}

void Free_Task_Graph( Graph* Reseau )
{
  int	i;

  for (i=0; i<Reseau->Size; i++) {
    free( Reseau->T[i].Task_Name );
    if ( (char*)NULL != Reseau->T[i].Task_Given_Name )
      free( Reseau->T[i].Task_Given_Name );
    Free_Link_List( Reseau->T[i].Neighbors );
    free( Reseau->T[i].Prev->Task_Id);
    free( Reseau->T[i].Prev );
    if ( (Law_Desc)NULL != Reseau->T[i].Task_Law )
      Free_Law( Reseau->T[i].Task_Law );

  }
  free( Reseau->T );
  for (i=0; i<Reseau->Nb_Layers; i++) {
    free( Reseau->Level[i]->Task_Id );
    free( Reseau->Level[i] );
  }
  free( Reseau->Level );
  free( Reseau );

}
