#
# Various utility procedure, property checks etc.
# 
############################################################
# Is_Empty_Store -
# Boolean function telling if the store contains any piece at all.
#
# Parameters: none
#
proc Is_Empty_Store { } {
    global cur_piece_nb

    return [expr $cur_piece_nb <= 0]
}

############################################################
# Is_Empty_Piece -
# Boolean function telling if the given piece (in normalized form) 
# is empty or not.
#
# Parameters:
# - piece	the piece, as a pair hi/lo profile.
#
proc Is_Empty_Piece { piece } {
    global cur_width

    array set lo_prof [lindex $piece 0]
    set empty 1
    for { set i 0 } { $i < $cur_width } { incr i 1 } {
	if { [string compare $lo_prof($i) eps ] != 0 } {
	    set empty 0
	}
    }
    return $empty
}

############################################################
# Is_Interaction_Graph -
# Boolean function telling if the current set of pieces corresponds
# to an interaction graph, that is, has a maximum height of one.
#
# Parameters: none
#
proc Is_Interaction_Graph { } {
    global Piece_Store cur_piece_nb

    if { $cur_piece_nb == 1 } { return 0 }

    set max [high_point $Piece_Store(1)]
    for { set i 2 } { $i <= $cur_piece_nb } { incr i 1 } {
	set hei [high_point $Piece_Store($i)]
	set max [expr $hei > $max ? $hei : $max ]
    }

    if { $max == 1 } {
	return 1
    } else {
	return 0
    }
}
